#!/usr/bin/perl -w -I../lib
require 5.6.0;

=head1 NAME

ad_response.cgi -- Print a report of ad response

=head1 SYNOPSIS

    ad_response.cgi?begin_date=5/1/2000&end_date=6/1/2000
      or
    ad_response.cgi?month_offset=0		# print current month
      or
    ad_response.cgi?year_offset=0		# print current year

=head1 DESCRIPTION

The ad_response.cgi program produces a report of the ad source inquiry counts
and sales. The begin date and and date are asked for if there are no arguments.
A month offset can be specified to pick the results for that month.
A year offset can be specified to pick the results for that year. If the year
offset is 0, results are for the current year to date.

=head1 NOTES

Removing a customer record removes all counts and totals for that customer.
Even the number of inquiries received.

Number of inquiries are counted only for certain customer categories, such as
client, prospect, and ta.

=head1 CHANGE LOG

12/11/00 Initial version.
2/7/01   Added year offset argument
12/27/01 Converted to AcmeBase/PostgreSql

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000,2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Acme::Db;
use Acme::Options;
use Acme::HTML::NavButtons;
use Acme::HTML::Parameters;
use Acme::Date	qw(today);
use Acme::Config;
use Local::Acme::Banners	qw(common_header common_footer);
use CGI;

use Acme::Security;
Acme::Security::check('Access');

our $Styles = Acme::Config->Styles;

$|=1;		# flush output (synchronize errors with output)

$cgi = new CGI;
print $cgi->header(-type=>'text/html') unless @ARGV;

#
# Program Arguments
#

our $year_offset  = $cgi->param('year_offset');
our $month_offset = $cgi->param('month_offset');
our $bgndate      = $cgi->param('bgndate');
our $enddate      = $cgi->param('enddate');
our $sort_order   = $cgi->param('sort_order') || '';

our $restrict_source = $cgi->param('restrict_source');
our @ad_sources = $restrict_source ? $cgi->param('ad_sources') : ('');
our $ad_sources = join ',', @ad_sources;

our $restrict_type = $cgi->param('restrict_type');
our $types = $restrict_type ? $cgi->param('types') : '';

## Hardwire types of customers listed in report. ##
##$types ||= "client,prosp,ta,taINT,taEUR";

unless (defined($month_offset) or defined($year_offset) or
       (defined($bgndate) and defined($enddate))) {
	get_parameters();
	exit 0;		# the parameter form will re-call this routine
}

unless ($bgndate and $enddate) {
	# compute bgndate and enddate from month or year offset

	my ($m,$y) = (localtime)[4,5];
	my @days = (31,28,31,30,31,30,31,31,30,31,30,31);	# days in months

	my ($bm,$bd,$by,$em,$ed,$ey);

	if (defined($year_offset)) {
		my $year_begin_month = Acme::Config->Fiscal_Year_Begin_Month - 1;

		$by = $ey = 1900 + $y + $year_offset;
		if ($year_begin_month > 0) {
			if ($m > $year_begin_month) {
				$ey += 1;
			}
			else {
				$by -= 1;
			}
		}

		$days[1] = (($ey/4 == 0 && $ey/100 != 0) || $ey/400 == 0) ? 29 : 28;
		$bm = $year_begin_month;
		$bd = 1;
		$em = ($year_begin_month > 0) ? ($year_begin_month - 1) : 11;
		$ed = $days[$em];

		$bm++;			# set month to be between 1 and 12
		$em++;			# set month to be between 1 and 12
	}
	else {
		$em = ($m + $month_offset);
		$ey = 1900 + $y;
		$ey += int($em/12) - 1 if ($em < 0);
		$ey += int($em/12) if ($em > 11);
		$em = $em % 12;

		$days[1] = (($ey/4 == 0 && $ey/100 != 0) || $ey/400 == 0) ? 29 : 28;

		$ed = $days[$em];
		$em++;				# set month to be between 1 and 12

		$bm = $em;
		$bd = 1;
		$by = $ey;
	}

	$bgndate = join("/",$bm,$bd,$by);	# set to first day of month
	$enddate = join("/",$em,$ed,$ey);	# set to last day of month

	if (defined($year_offset) and $year_offset == 0) {
		# override enddate to be today's date
		$enddate = &today;
	}
}

our $opt = new Acme::Options;

# Save last used options 
$opt->set('AdResponseSortOrder'      => $sort_order);
$opt->set('AdResponseSources'        => $ad_sources);
$opt->set('AdResponseRestrictSource' => $restrict_source);
$opt->set('AdResponseTypes'          => $types);
$opt->set('AdResponseRestrictType'   => $restrict_type);
$opt->save();

our $b = new Acme::HTML::NavButtons;

	print <<EOS;
<html>
<head>
<title>Ad Response from $bgndate through $enddate</title>
<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<style type="text/css">
        .number       { text-align:right; }
        .date         { text-align:center; }
        .button       { font-family:system; }
        .small-button { font-family:sans-serif; font-size:8pt;
                        text-decoration:none; }
        .pseudo-link  { color:purple }
	.exec_summary { border:solid black 1px; background-color:#DDDDDD; }
	.column-header	{
			font-family:serif;
			font-size:medium;
			font-weight:bold;
			text-decoration:none;
			color:black;
			padding-left:1ex;
			}

	.column-total	{
			font-family:serif;
			font-size:medium;
			font-weight:bold;
			text-decoration:none;
			color:black;
			border-top:solid black 1px;
			padding-left:1ex;
			}

        TH            { font-family:serif; font-size:medium; }
        TD            { font-family:sans-serif; font-size:small; }
        TD.total      { font-family:sans-serif; font-size:medium; }
</style>
</head>
<body onLoad='self.focus()' topmargin=2 bottommargin=2 leftmargin=2 rightmargin=2>
EOS

print $b->buttons();
print common_header();

report::print_ad_response($bgndate,$enddate,$sort_order,$ad_sources,$types);

print common_footer();
print $b->last_buttons();

print "</body>\n";
print "</html>\n";

exit 0;

#------------------------------------------------------------------------------

sub get_parameters {

	$bgndate ||= '';
	$enddate ||= $bgndate;

	Acme::HTML::Parameters->get_parameters(
			title	=> 'Enter Search Parameters',
			params	=> [
				   "bgndate=?d?$bgndate?Begin Date",
				   "enddate=?d?$enddate?End Date",
				   ],
			);
}

#------------------------------------------------------------------------------

package report;

use Acme::HTML	qw(escape_html);
use Acme::Date  qw(mdate meddate ldate dayadd daydiff weekday);
use Acme::Money qw(dollars commafy);

#------------------------------------------------------------------------------

sub print_ad_response {
	local our ($bgndate,$enddate,$sort_order,$ad_sources,$types) = @_;

	$\="\n";	# append newlines to print statements

	our ($trip_source_clause, $response_clause, $ad_source_clause);

	if ($ad_sources) {
		$ad_source_clause = $ad_sources;
		$ad_source_clause =~ s/\s*,\s*/','/g;

		$trip_source_clause =
		  qq[ and trips.ad_source in ('$ad_source_clause')];
		$response_clause =
		  qq[ and ad_response.ad_source in ('$ad_source_clause')];
		$ad_source_clause =
		  qq[ and ad_sources.ad_source in ('$ad_source_clause')];
	}
	else {
		# Set default clause to exclude internal codes
		my $internal_ad_source = qq[
	select ad_sources.ad_source
	from ad_sources,ad_source_types
	where ad_sources.ad_source_type = ad_source_types.ad_source_type
	  and ad_source_types.internal = 'Y'
	];
		# The following SQL clauses are used in HTML links, so
		# change all linefeeds and tabs to single spaces.
		$internal_ad_source =~ s/\s+/ /g;

		$trip_source_clause =
		  qq[ and trips.ad_source not in ($internal_ad_source)];
		$response_clause =
		  qq[ and ad_response.ad_source not in ($internal_ad_source)];
		$ad_source_clause =
		  qq[ and ad_sources.ad_source not in ($internal_ad_source)];
	}
	
	our $type_clause = '';
	($type_clause = $types) =~ s/\s*,\s*/','/g if $types;
	$type_clause = qq( and customers.type in ('$type_clause')) if $type_clause;

	our $order_clause = "order by ad_source_name, ad_source";	# default 
	$order_clause = "order by ad_source_name, ad_source"
	  if $sort_order eq 'Name';
	$order_clause = "order by ad_source, ad_source_name"
	  if $sort_order eq 'Ad Source';
	$order_clause = "order by inquiries desc"
	  if $sort_order eq 'Inquiries';
	$order_clause = "order by sales desc"
	  if $sort_order eq 'Sales';
	$order_clause = "order by volume desc"
	  if $sort_order eq 'Volume';
	$order_clause = "order by cost desc"
	  if $sort_order eq 'Cost';

	#
	# Get totals used for percentages
	#

	local our ($total_inquiries) = Acme::Db->get(<<End_Of_SQL);
select respondent_id as customer_id
into temp t_ad_count
from ad_response
where response_date between '$bgndate' and '$enddate'
  and ad_response.source_table = 'customers'
 $response_clause
;

select count(*)
from customers, t_ad_count
where t_ad_count.customer_id = customers.customer_id
  $type_clause
End_Of_SQL

	local our ($total_sales, $total_volume) = Acme::Db->get(<<End_Of_SQL);
select count(*), sum(trip_total)
from trips
where sale_date between '$bgndate' and '$enddate'
 $trip_source_clause
End_Of_SQL

	local our ($total_cancel_sales) = Acme::Db->get(<<End_Of_SQL);
select count(*)
from trips
where sale_date between '$bgndate' and '$enddate'
  and trips.status = 'X'
 $trip_source_clause
End_Of_SQL

	# Use group trip budgeted income as lost income for canceled sales
	# ignoring the pax trip record of group trip.
	local our ($total_cancel_volume) = Acme::Db->get(<<End_Of_SQL);
select	trip_id,				-- group trip cancellations
	sum(participants_gross) +
	sum(reservations_gross) +
	sum(fees_gross) +
	sum(misc_fees_gross) as cancel_gross,
	sum(budget_income) as budget_income
into temp t_xxl_group_trips
from trips, ad_sources, ad_source_types
where sale_date between '$bgndate' and '$enddate'
  and trips.status = 'X'
  and trips.group_trip = 'Y'
  and (trips.ad_source is null
    or ad_source_types.internal = 'Y'
    )
  and trips.ad_source = ad_sources.ad_source
  and ad_sources.ad_source_type = ad_source_types.ad_source_type
group by trip_id
;

select	trip_id,				-- non group cancellations
	sum(participants_gross) +
	sum(reservations_gross) +
	sum(fees_gross) +
	sum(misc_fees_gross) as cancel_gross,
	sum(budget_income) as budget_income
into temp t_xxl_trips
from trips
where sale_date between '$bgndate' and '$enddate'
  and trips.status = 'X'
 $trip_source_clause
  and trips.group_trip_id not in (
	select trip_id
	from t_xxl_group_trips
	where t_xxl_group_trips.cancel_gross > 0
	   or t_xxl_group_trips.budget_income > 0
	)
group by trip_id
;
select	t_xxl_trips.trip_id,	-- Combine pax and group trip cancellations
	t_xxl_trips.cancel_gross,
	t_xxl_trips.budget_income
into temp t_xxl_trips2
from t_xxl_trips
union all
select	t_xxl_group_trips.trip_id,
	t_xxl_group_trips.cancel_gross,
	t_xxl_group_trips.budget_income
from t_xxl_group_trips
;
select count(*) as cancel_sales, -- Return trip gross amount or the budgeted income
	sum(cancel_gross) as cancel_volume
into temp t_xxl_sums
from t_xxl_trips2
where cancel_gross != 0
union all
select count(*) as cancel_sales,
	sum(budget_income) as cancel_volume
from t_xxl_trips2
where cancel_gross = 0
;
select sum(cancel_volume) as cancel_volume	-- Combine totals
from t_xxl_sums
;
End_Of_SQL

	#
	# Get list of ad sources response
	#

	my $q = Acme::Db->query(<<End_of_SQL);
select respondent_id as customer_id, ad_source
into temp t_ad_group
from ad_response
where response_date between '$bgndate' and '$enddate'
 $response_clause
;

select
	t_ad_group.ad_source,
	count(*) as inquiries
into temp t_ad
from customers, t_ad_group
where t_ad_group.customer_id = customers.customer_id
  $type_clause
group by t_ad_group.ad_source
;

select
	ad_source,
	count(*) as sales,
	sum(trip_total) as volume,
	sum(profit) as profit
into temp t_sales
from trips
where sale_date between '$bgndate' and '$enddate'
 $trip_source_clause
group by ad_source
;

select t_ad.ad_source
into temp t_ad_sources
from t_ad
union
select t_sales.ad_source
from t_sales
union
select ad_sources.ad_source
from ad_sources
where (start_date is not null and start_date between '$bgndate' and '$enddate')
   or (stop_date is not null and stop_date between '$bgndate' and '$enddate')
   or (start_date is null and stop_date is null and
      (retired is null or retired ~* 'N'))
;

select
	t_ad_sources.ad_source,
	ad_sources.ad_source_type,
	ad_sources.ad_source_name,
	ad_sources.qty,
	ad_sources.cost,
	ad_sources.cost_per_inquiry,
	ad_sources.cost_per_sale,
	ad_sources.commission_rate,
	ad_sources.payment_type,
	t_ad.inquiries as inquiries,
	t_sales.sales as sales,
	t_sales.volume as volume,
	t_sales.profit as profit
from (((t_ad_sources left join ad_sources
     on t_ad_sources.ad_source = ad_sources.ad_source)
     left join t_ad
     on t_ad_sources.ad_source = t_ad.ad_source)
     left join t_sales
     on t_ad_sources.ad_source = t_sales.ad_source)
where 1 = 1
 $ad_source_clause
$order_clause
;
End_of_SQL

	$q->bind_columns_to_variables(__PACKAGE__,1);	# bind short names

	no warnings qw(uninitialized once);

	print "<br>";
	print "<br>";

	my ($ad_cost, $ad_cost_formula, $ad_cost_formula1, $ad_cost_formula2);

	our ($sum_inquiries, $sum_sales, $sum_volume, $sum_cost);

	my $count = 0;
	while ($q->next) {
		$count++;

		if ($count == 1) {
			# Initialize
			$title="Ad Response  from " .  ldate($bgndate)
			  . " through " .  ldate($enddate);
			$title .= "<br>for Ad Sources $ad_sources"
			  if ($ad_sources);

			$sum_inquiries = $sum_sales = $sum_volume = $sum_cost = 0;

			# Print first page header
			before_report();

			# Trigger before group events.
			if ($sort_order eq 'Ad Source') {
				before_group_source();
			}
		}
		else {
			# Trigger after group events if needed.
			if ($sort_order eq 'Ad Source' and $ad_source ne $prev_ad_source) {
				after_group_source();
				before_group_source();
			}
		}

		next if ($inquiries == 0 && $sales == 0 && $volume == 0 && $ad_cost == 0);

		#
		# Compute ad cost
		#

		$ad_cost = 0;
		$ad_cost_formula = '';
		$ad_cost_formula1 = '';
		$ad_cost_formula2 = '';
		if ($cost > 0) {
			$ad_cost += $cost;

			$ad_cost_formula1 .= dollars($ad_cost,0);

			$ad_cost_formula2 .= "ad cost";
		}

		if ($cost_per_inquiry > 0) {
			$ad_cost += ($inquiries * $cost_per_inquiry);

			$ad_cost_formula1 .= " + " if $ad_cost_formula1;
			$ad_cost_formula1 .=
			  "($inquiries * "
			  . dollars($cost_per_inquiry,2)
			  . ")";

			$ad_cost_formula2 .= " + " if $ad_cost_formula2;
			$ad_cost_formula2 .=
			  "(inquiries * cost per inquiry)";
		}

		if ($cost_per_sale > 0) {
			$ad_cost += ($sales * $cost_per_sale);

			$ad_cost_formula1 .= " + " if $ad_cost_formula1;
			$ad_cost_formula1 .=
			  "($sales * "
			  . dollars($cost_per_sale,2)
			  . ")";

			$ad_cost_formula2 .= " + " if $ad_cost_formula2;
			$ad_cost_formula2 .=
			  "(sales * cost per sale)";
		}

		if ($commission_rate > 0) {
			if ($payment_type eq 'G' && $trip_total > 0) {
				# commission on gross amount
				$ad_cost += ($commission_rate/100 * $trip_total);

				$ad_cost_formula1 .= " + " if $ad_cost_formula1;
				$ad_cost_formula1 .= "("
				  . sprintf("%.1f%%", $commission_rate)
				  . " * $trip_total)";

				$ad_cost_formula2 .= " + " if $ad_cost_formula2;
				$ad_cost_formula2 .=
				  "(commission * gross)";
			}
			elsif ($payment_type eq 'N' && $profit > 0) {
				# commission on net profit amount
				$ad_cost += ($commission_rate/100 * $profit);

				$ad_cost_formula1 .= " + " if $ad_cost_formula1;
				$ad_cost_formula1 .= "("
				  . sprintf("%.1f%%", $commission_rate)
				  . " * $profit)";

				$ad_cost_formula2 .= " + " if $ad_cost_formula2;
				$ad_cost_formula2 .=
				  "(commission * net profit)";
			}
		}

		$ad_cost_formula = escape_html("$ad_cost_formula1\n$ad_cost_formula2");

		#
		# Print item
		#

		print "<tr>";

		print "<th align=left>$ad_source_name</th>";
		print "<td valign=baseline><nobr>$ad_source</nobr></td>";
		if ($qty > 0) {
			print qq[<td align="right" valign="baseline">],
			  "<nobr>", commafy($qty,0), "</nobr>";
		}
		else {
			print "<td></td>";
		}
		if ($inquiries > 0) {
			print
			  qq[<td align="right" valign="baseline">],
			  qq[<a href="javascript:open_window('],
			  qq[/cgi-bin/list.cgi?table=customers&search='+],
			  escape_html(
qq[escape("Customers Inquiries between $bgndate and $enddate|customers.customer_id in (select ad_response.respondent_id from ad_response, customers where ad_response.response_date between '$bgndate' and '$enddate' and ad_response.ad_source = '$ad_source' and ad_response.respondent_id = customers.customer_id and ad_response.source_table = 'customers')$type_clause")]
				  ),
			  qq[)" title='List the customers who inquired.'>],
			  commafy($inquiries,0),
			  "</a>",
			  "</td>";
		}
		else {
			print qq(<td align="right" valign="baseline">),
			  commafy($inquiries,0),
			  "</td>";
		}
		print "<td align=right valign=baseline>",
		  sprintf("%.1f%%", $total_inquiries ? (($inquiries / $total_inquiries) * 100) : 0),
		  "</td>";
		print qq[<td align=right valign=baseline style="padding-right:3ex">],
		  sprintf("%.1f%%", $qty ? (($inquiries / $qty) * 100) : 0),
		  "</td>";
		if ($sales > 0) {
			print
			  qq[<td align="right" valign="baseline">],
			  qq[<a href="javascript:open_window('],
			  qq[/cgi-bin/list.cgi?table=customers&search='+],
			  escape_html(
qq[escape("Customers Purchasing Trips between $bgndate and $enddate|customers.customer_id in (select customer_id from trips where trips.sale_date between '$bgndate' and '$enddate' and trips.ad_source = '$ad_source')")]
				  ),
			  qq[)" title='List the customers who purchased trips.'>],
			  commafy($sales,0),
			  "</a>",
			  "</td>";
		}
		else {
			print
			  qq(<td align="right" valign="baseline">),
			  commafy($sales,0),
			  "</td>";
		}
		print "<td align=right valign=baseline>",
		  sprintf("%.1f%%", $total_sales ? (($sales / $total_sales) * 100) : 0),
		  "</td>";

		print '<td align=right valign=baseline style="padding-right:3ex">',
		  sprintf("%.1f%%", $inquiries ? (($sales / $inquiries) * 100) : 0),
		  "</td>";

		if ($volume > 0) {
			print
			  qq[<td align="right" valign="baseline">],
			  qq[<a href="javascript:open_window('],
			  qq[/cgi-bin/list.cgi?table=customers&search='+],
			  escape_html(
qq[escape('Customer Trips Purchased between $bgndate and $enddate|table_join("customers inner join trips on trips.customer_id = customers.customer_id")|trips.sale_date between '$bgndate' and '$enddate' and trips.ad_source = '$ad_source'")]
				  ),
			  qq[)" title='List the trips whose gross add up to this number.'>],
			  commafy($volume,0),
			  "</a>",
			  "</td>";
		}
		else {
			print qq(<td align="right" valign="baseline">),
			  commafy($volume,0),
			  "</td>";
		}
		print "<td align=right valign=baseline>",
		  sprintf("%.1f%%", $total_volume ? (($volume / $total_volume) * 100) : 0),
		  "</td>";

		print "<td>&nbsp;</td>   <!-- Spacer -->";

		if ($ad_cost > 0) {
			print qq(<td align="right" valign="baseline">),
			  qq(<span class="pseudo-link" title="$ad_cost_formula">),
			  commafy($ad_cost,0),
			  "</span>",
			  "</td>";
		}
		else {
			print qq(<td align="right" valign="baseline">0</td>);
		}

		print "<td align=right valign=baseline>",
		  sprintf("%.1f%%", $volume ? (($ad_cost / $volume) * 100) : 0),
		  "</td>";
		print "<td align=right valign=baseline>",
		  $inquiries ? commafy(($ad_cost / $inquiries),2) : '0.00',
		  "</td>";
		print "<td align=right valign=baseline>",
		  $sales ? commafy(($ad_cost / $sales),2) : '0.00',
		  "</td>";

		print "</tr>";
	}
	continue {

		#
		# Accumulate sums and counts
		#

		$sum_inquiries += $inquiries;
		$sum_sales += $sales;
		$sum_volume += $volume;
		$sum_cost += $ad_cost;

		#
		# Track previous values of columns with breaks between groups.
		#

		$prev_ad_source = $ad_source;

		#
		# Last record processing
		#

		if ($count == $q->count) {
			# Last record is being processed.
			# Trigger after group events.
			if ($sort_order eq 'Ad Source') {
				after_group_source();
			}

			# Print report trailer
			after_report($count);
		}
	}

	print "<br>";
}

sub before_group_source {}

sub after_group_source {}

sub before_report {
	print "<h2>$title</h2>";

	$response_clause =~ s/'/\\'/g;

	print <<EOF;
<table width="98%" align="center" class="exec_summary"><tr><td valign="top">

<table cellspacing=2 cellpadding=2 border=0>
<tr>
<th align=right>Total Inquiries</th>
<td align=right class=total>
<a href="javascript:open_window('/cgi-bin/list.cgi?table=customers&search='+@{[escape_html(qq`escape("Customers Inquiries between $bgndate and $enddate|customers.customer_id in (select ad_response.respondent_id from ad_response, customers where ad_response.respondent_id=customers.customer_id and ad_response.response_date between '$bgndate' and '$enddate' and ad_response.source_table = 'customers'$response_clause)$type_clause")`)]})"
 title='List customers who inquired.'
 >@{[commafy($total_inquiries,0)]}</a></td>
<td>&nbsp;</td>   <!-- Spacer -->
<td align=right valign=baseline></td>
</tr>

<tr>
<th align=right>Total Sales</th>
<td align=right class=total>
<a href="javascript:open_window('/cgi-bin/list.cgi?table=customers&search='+@{[escape_html(qq`escape("Customers Who Purchased Trips between $bgndate and $enddate|customers.customer_id in (select customer_id from trips where trips.sale_date between '$bgndate' and '$enddate'$trip_source_clause)")`)]})"
 title='List customers who purchased trips.'
 >@{[commafy($total_sales,0)]}</a></td>
<td>&nbsp;</td>   <!-- Spacer -->
<td align=right valign=baseline>
@{[sprintf("%.1f%%", $total_inquiries ? (($total_sales / $total_inquiries) * 100) : 0)]} of inquiries
</td>
</tr>

<tr>
<th align=right>Total Volume</th>
<td align=right class=total>
<a href="javascript:open_window('/cgi-bin/list.cgi?table=customers&search='+@{[escape_html(qq`escape("Customers with Trips Purchased between $bgndate and $enddate|table_join('customers inner join trips on trips.customer_id = customers.customer_id')|trips.sale_date between '$bgndate' and '$enddate'$trip_source_clause")`)]})"
 title='List trips whose gross add up to this number.'
 >@{[dollars($total_volume,0)]}</a></td>
<td>&nbsp;</td>   <!-- Spacer -->
<td align=right valign=baseline></td>
</tr>
</table>

</td><td valign=bottom>

<table cellspacing=2 cellpadding=2 border=0 align=right>
<tr>
<th align=right>Cancellations</th>
<td align="right" valign="baseline">
<a href="javascript:open_window('/cgi-bin/list.cgi?table=customers&search='+@{[escape_html(qq`escape("Customers Who Canceled Trips between $bgndate and $enddate|customers.customer_id in (select customer_id from trips where trips.sale_date between '$bgndate' and '$enddate' and trips.status = 'X'$trip_source_clause)")`)]})"
 title='List customers who canceled trips.'
 >@{[commafy($total_cancel_sales,0)]}</a></td>
<td>&nbsp;</td>   <!-- Spacer -->
<td align=right valign=baseline>
@{[sprintf("%.1f%%", $total_sales ? (($total_cancel_sales / $total_sales) * 100) : 0)]} of sales
</td>
</tr>

EOF
	print <<EOF;
<tr>
<th align=right>Canceled Volume</th>
<td align="right" valign="baseline">
<a href="javascript:open_window('/cgi-bin/list.cgi?table=customers&search='+@{[escape_html(qq`escape("Customers with Trips Canceled between $bgndate and $enddate|table_join('customers inner join trips on trips.customer_id = customers.customer_id')|trips.sale_date between '$bgndate' and '$enddate' and trips.status = 'X'$trip_source_clause")`)]})"
 title='List canceled trips that make up this number.'
 >@{[dollars($total_cancel_volume,0)]}</a></td>
<td>&nbsp;</td>   <!-- Spacer -->
<td align=right valign=baseline>
@{[sprintf("%.1f%%", $total_volume ? (($total_cancel_volume / ($total_volume + $total_cancel_volume)) * 100) : 0)]} of sales
</td>
</tr>

</table>

</td></tr></table>
<br>
EOF

	print <<EOF;
<table align=center width=98%>
<tr>
<td class="column-header" align=center><br>Ad Source</td>
<td class="column-header" align=left>Source<br>Code</td>
<td class="column-header" align=left><br>Pieces</td>
<td class="column-header" align=right><br>Inquiries</td>
<td class="column-header" align=center>%<br>Total</td>
<td class="column-header" align=center>%<br>Response</td>
<td class="column-header" align=right><br>Sales</td>
<td class="column-header" align=center>%<br>Total</td>
<td class="column-header" align=center>%<br>Conversion</td>
<td class="column-header" align=right><br>Volume</td>
<td class="column-header" align=center>%<br>Total</td>
<td class="column-header" align=right>&nbsp;</td>
<td class="column-header" align=right><br><nobr>Ad Cost</nobr></td>
<td class="column-header" align=center>%<br>Volume</td>
<td class="column-header" align=center><nobr>Ad Cost/</nobr><br>Inquiry</td>
<td class="column-header" align=center><nobr>Ad Cost/</nobr><br>Sale</td>
</tr>
EOF
}

sub after_report {
	my ($count) = @_;

	print <<EOF;
<tr>
<td align=center></td>
<td align=left></td>
<td align=left></td>
<td class="column-total" align=right>@{[commafy($sum_inquiries,0)]}</td>
<td align=right>&nbsp;</td>
<td align=right>&nbsp;</td>
<td class="column-total" align=right>@{[commafy($sum_sales,0)]}</td>
<td align=right>&nbsp;</td>
<td align=right>&nbsp;</td>
<td class="column-total" align=right>@{[dollars($sum_volume,0)]}</td>
<td align=right>&nbsp;</td>
<td align=right>&nbsp;</td>
<td class="column-total" align=right><nobr>@{[dollars($sum_cost,0)]}</nobr></td>
<td class="column-total" align=right><nobr>@{[sprintf("%.1f%%", $sum_volume ? (($sum_cost / $sum_volume) * 100) : 0)]}</nobr></td>
<td class="column-total" align=right><nobr>@{[$sum_inquiries ? dollars($sum_cost / $sum_inquiries,2) : '$0.00']}</nobr></td>
<td class="column-total" align=right><nobr>@{[$sum_sales ? dollars($sum_cost / $sum_sales,2) : '$0.00']}</nobr></td>
</tr>
EOF

	# Re-print headers after totals if list is long.

	print <<EOF if $count > 50;
<tr>
<td colspan="16">&nbsp;</td>	<!-- blank line -->
</tr>
<tr>
<td class="column-header" align=center><br>Ad Source</td>
<td class="column-header" align=left>Source<br>Code</td>
<td class="column-header" align=left><br>Pieces</td>
<td class="column-header" align=right><br>Inquiries</td>
<td class="column-header" align=center>%<br>Total</td>
<td class="column-header" align=center>%<br>Response</td>
<td class="column-header" align=right><br>Sales</td>
<td class="column-header" align=center>%<br>Total</td>
<td class="column-header" align=center>%<br>Conversion</td>
<td class="column-header" align=right><br>Volume</td>
<td class="column-header" align=center>%<br>Total</td>
<td class="column-header" align=right>&nbsp;</td>
<td class="column-header" align=right><br><nobr>Ad Cost</nobr></td>
<td class="column-header" align=center>%<br>Volume</td>
<td class="column-header" align=center><nobr>Ad Cost/</nobr><br>Inquiry</td>
<td class="column-header" align=center><nobr>Ad Cost/</nobr><br>Sale</td>
</tr>
EOF

	print <<EOF;
</table>
<br>
EOF
}
__END__
