#!/usr/bin/perl -wT -I../lib
require 5.6.0;
use strict;

=head1 NAME

admin.cgi -- Display a screen to access database information.

=head1 SYNOPSIS

    admin.cgi

    admin.cgi?table=customers&table=customer_types

    admin.cgi?sort=yes

    admin.cgi?sort=type&all

=head1 DESCRIPTION

The C<admin.cgi> program provides a framework to access and edit all the
tables in the database.  It displays a list of tables with a form consisting
of a "List" button, a Search field, an "Edit" button, and a "New" button.

If the sort option is specified, all tables are listed in alphabetical order.
There are two settings for C<sort>, "yes", or "class". If "class" is specified,
they are grouped by table classes, otherwise, they are listed in the order
they appear in the schema.

Specifying an "all" option displays all tables, even the hidden tables.
Specifying "hidden" does the same.

Specifying tables to be administered suppresses the documentation and the
SQL sections of the page. This action is experimental.

=head1 CHANGE LOG

 5/3/00   Initial version.
 5/2/01   Change to Acme::Base::Schema. Edit and List windows are not re-used.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Acme::HTML	qw(escape_html);
use Acme::HTML::NavButtons;
use Acme::HTML::ListMenus;
use Acme::Base::Schema;
use Acme::Options;
use Local::Acme::Banners	qw(common_header common_footer);
use Acme::Config;
use CGI;

use Acme::Security;
Acme::Security::check('Access');

our $Db_Title    = Acme::Config->DBTITLE;	# title of database
our $Base        = Acme::Config->Base;		# http://... root
our $Help        = Acme::Config->Help;		# location of help files
our $Doc         = Acme::Config->Doc;		# location of documentation files
our $Styles      = Acme::Config->Styles;	# location of css styles
our $Scripts     = Acme::Config->Scripts;
our $Scripts_Lib = Acme::Config->Scripts_Lib;

$\="\n";	# append newlines to all print statements
$|=1;		# flush output (synchronize errors with output)

our $cgi = new CGI;
$cgi->cache('no');
print $cgi->header(-type=>'text/html', -expires=>'-1d') unless @ARGV;

sub yesno_param {		# Interpret CGI parameters
	my ($cgi,$p) = @_;

	# Allow CGI parameters without a value to be interpreted as 'yes'.
	# If the parameter is 'no', then the empty string is returned as
	# the parameters value so that it can be interpreted as 'false'
	# in Perl conditional statements.

	my $val = (grep {/^$p$/i} $cgi->param())
	       || (grep {/^$p$/i} $cgi->param('keywords'));
	$val &&= $cgi->param($p) || 'yes';
	$val = '' if lc($val) eq 'no';
	return $val;
}

our $display_all    = yesno_param($cgi,'all');
our $display_hidden = yesno_param($cgi,'hidden');

our $sort           = yesno_param($cgi,'sort');

our @tables = $cgi->param('table');


our $opt = new Acme::Options;

our $schema = new Acme::Base::Schema;
our $t;

our $is_tables_specified = @tables;	# set flag

# List all the tables unless they were specified in CGI parameters.
@tables = $schema->tables unless $is_tables_specified;

our $b = new Acme::HTML::NavButtons;

#
# Display page
#

display_header();

if ($is_tables_specified) {

	#
	# Display lookup screen.
	#

	display_items();
}
else {
	#
	# Display full administration screen.
	#

	display_documentation();

	print $b->buttons;

	display_items();

	if (Acme::Security::allow('access','sql')) {
		print $b->buttons;

		display_sql();
	}
}

display_footer();

exit 0;

#------------------------------------------------------------------------------

sub display_header {

	my $conflag = Acme::Config->Console_Prog =~ /admin.cgi/
		    ? qq(\n<script language="JavaScript">isConsoleLoaded=true;</script>)
		    : '';

# Keep IE6 in IE5 compatablilty mode by not specifying url
#<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
#"http://www.w3.org/TR/html4/loose.dtd">

        print <<EOF;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">

<html>
<head>
<title>admin $Db_Title</title>
<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<link rel="stylesheet" href="$Styles/console.css" type="text/css">
<link rel="stylesheet" href="$Styles/admin.css" type="text/css">
<script src="$Scripts/common.js" type="text/javascript"></script>
<script src="$Scripts_Lib/form_events.js" type="text/javascript"></script>
<script src="$Scripts_Lib/control.js" type="text/javascript"></script>$conflag
</head>
<body>
EOF
	print $b->first_buttons;

	print common_header('Administration');
}

sub display_footer {

	print common_footer('Administration');

	print $b->last_buttons;

	print <<EOF;
</body>
</html>
EOF

}

sub display_items {

	print <<EOF;
<h2>
$Db_Title Database Tables
</h2>

<div>
EOF
	if ($sort) {
		if ($sort =~ /class/i) {
			#
			# Group tables together according to the table class
			#

			my %grouped_tables = ();
			my $class;
			for my $tab (sort @tables) {
				$t = $schema->t($tab);
				if (my $sc = $t->p('security_code')) {
					next unless Acme::Security::allow('access',$sc);
				}

				$class = $t->class;
				$grouped_tables{$class} = [] unless $grouped_tables{$class};
				push @{$grouped_tables{$class}}, $tab;
			}

			#
			# List tables
			#

			# List primary tables
			print "<h3>Primary tables</h3>"; 
			for $class ('master') {
				for my $tab (@{$grouped_tables{$class}}) {
					$t = $schema->t($tab);
					display_item($t)
					  if ! $t->hidden || $display_all || $display_hidden;
				}
				delete $grouped_tables{$class};
			}
			print $b->buttons;

			# List ancillary tables (usually hidden)
			print "<h3>Ancillary tables</h3>"; 
			for $class ('detail','join','contacts','log') {
				print "<h4>", ucfirst lc $class, " tables</h4>"
				  if $grouped_tables{$class};
				for my $tab (@{$grouped_tables{$class}}) {
					$t = $schema->t($tab);
					display_item($t)
					  if ! $t->hidden || $display_all || $display_hidden;
				}
				delete $grouped_tables{$class};
			}
			print $b->buttons;

			# List support tables
			print "<h3>Support tables</h3>"; 
			for $class ('people','code') {
				print "<h4>", ucfirst lc $class, " tables</h4>"
				  if $grouped_tables{$class};
				for my $tab (@{$grouped_tables{$class}}) {
					$t = $schema->t($tab);
					display_item($t)
					  if ! $t->hidden || $display_all || $display_hidden;
				}
				delete $grouped_tables{$class};
			}

			# List all others
			if (%grouped_tables) {
				print $b->buttons;

				print "<h3>Other tables</h3>"; 
				for $class (sort keys %grouped_tables) {
					print "<h4>", ucfirst lc $class, " tables</h4>"
					  if $grouped_tables{$class};
					for my $tab (@{$grouped_tables{$class}}) {
						$t = $schema->t($tab);
						display_item($t)
						  if ! $t->hidden || $display_all || $display_hidden;
					}
				}
			}
		}
		else {
			#
			# Sort tables
			#

			my (@sorted_tables, @index);
			my $prev_c = '';
			my $c;

			for my $tab (sort {uc($a) cmp uc($b)} @tables) {
				$t = $schema->t($tab);
				next if $t->hidden && ! $display_all && ! $display_hidden;
				if (my $sc = $t->p('security_code')) {
					next unless Acme::Security::allow('access',$sc);
				}

				push @sorted_tables, $tab;

				# Build index
				$c = uc(substr($tab,0,1));
				push @index, $c if $c ne $prev_c;
				$prev_c = $c;
			}

			#
			# Display index
			#

			print qq[<div class="index">\n],
			  ( map { qq[&nbsp;<a href="#$_">$_</a>&nbsp;] } @index ),
			  qq[</div>\n]
			  unless $is_tables_specified;

			#
			# Display sorted tables
			#

			my $anchor;
			$c = $prev_c = '';
			for my $tab (@sorted_tables) {

				$anchor = '';
				$c = uc(substr($tab,0,1));
				if ($c ne $prev_c) {
					$anchor = qq[<a name="$c"></a>\n];
					$prev_c = $c;
				}

				$t = $schema->t($tab);
				display_item($t,$anchor);
			}
		}
	}
	else {
		#
		# List tables in order defined in schema
		#

		for my $tab (@tables) {
			$t = $schema->t($tab);
			if (my $sc = $t->p('security_code')) {
				next unless Acme::Security::allow('access',$sc);
			}

			display_item($t) if ! $t->hidden || $display_all || $display_hidden;
		}
	}

	print <<EOF;
</div>
EOF
}

sub display_item {
	my ($t, $anchor) = @_;

	$anchor ||= '';

	my ($prop, $f);
	my $name = $t->name;
	my $label = $t->title;
	my $description = $t->description;

	my $list_prog = $t->list_prog;
	$list_prog = "$Base$list_prog" if $list_prog =~ m|^/|;

	my $list_onclick = qq(onClick=") . escape_html( qq(this.form.action='$list_prog?table=$name';this.form.target='L'+(new Date()).getTime();open_list('',this.form.target);if(!this.form.search.value)this.form.search.value='*';this.form.submit()) ) . qq(");

	my $pick_prog = "$Base/cgi-bin/pick.cgi";	# hardwire to pick records to edit

	my $edit_onclick = qq(onClick=") . escape_html( qq(if(this.form.search.value.match(/^\\s*\$/)){return false};this.form.action='$pick_prog?table=$name&request=Edit';this.form.target='F'+(new Date).getTime();open_form('',this.form.target);this.form.submit()) ) . qq(");

	my $new_prog = $t->new_prog;
	$new_prog = "$Base$new_prog" if $new_prog =~ m|^/|;

	my $new_onclick = qq(onClick=") . escape_html( qq(javascript:open_form('$new_prog?table=$name&request=New')) ) . qq(");

	my $item_anchor = qq(<a name="a_$name"></a>);
	my $up_button = $b->up_button("a_$name");

	print <<EOF;
$anchor$item_anchor<table cellpadding="4" cellspacing="4" class="item">
  <tr>
  <td valign="top">$up_button</td>
  <td valign="baseline" nowrap>
  <!-- Begin Search Form -->
  <form class="search-form" method="POST"><input type="button" value="List" class="x-small-button" $list_onclick><input type="text" name="search" class="search-input" title="Enter search criteria."><input type="button" value="Edit" class="x-small-button" $edit_onclick><input type="button" value="New" class="x-small-button" $new_onclick>
  <br><table cellpadding=0 cellspacing=0 width=100%>
    <tr>
    <td><a href="$Base/cgi-bin/query.cgi?table=$name" onClick="new_target(this,'form')" class="small-link" title="Construct a query by filling in a form of search parameters.">Construct a Search</a><br><a href="$Base/cgi-bin/mkform.cgi?table=$name" onClick="new_target(this)" class="small-link" title="Construct a form that can be used to enter information.">Construct a Form</a></td>
    <td align="right" valign="top"><a href="$Base/cgi-bin/mkschema.cgi?table=$name" onClick="new_target(this)" class="small-link" title="Display the SQL Data Definition Language statements used to construct this table.">DDL</a></td>
    </tr></table></form>
  <!-- End Search Form -->
  </td>
  <td valign=top>
  <nobr><a href="$Base/cgi-bin/schema.cgi?table=$name" onClick="new_target(this,'doc')" title="Display field names and their descriptions."><font class="tablename">$label</font></a> ($name)</nobr>
  <br><font class="tabledesc">$description</font>
  </td>
  </tr>
</table>
EOF
}

sub display_documentation {

	print <<EOF;

<h2>
Documentation
</h2>

<div>
<ul>
<li><a href="$Base/cgi-bin/doc.cgi" onClick="new_target(this,'doc')">Acme Documentation</a></li>
<li><a href="http://www.apache.org/docs/" onClick="new_target(this)">Apache Documentation</a></li>
<li><a href="$Base/cgi-bin/doc.cgi?library=/usr/lib/perl5" onClick="new_target(this,'doc')">Perl Documentation</a></li>
</ul>
<ul>
<li><a href="$Base/cgi-bin/browse.cgi?directory=${Help}&link=$Base/cgi-bin/help.cgi%3Ffile=\$" onClick="new_target(this,'help')">Help Files</a></li>
<li><a href="$Base/cgi-bin/schema.cgi" onClick="new_target(this,'doc')">Database Schema</a></li>
<li><a href="$Base/cgi-bin/doc.cgi?topic=$Doc/AcmeBase.pod" onClick="new_target(this,'doc')">AcmeBase Properties</a></li>
</ul>
</div>
EOF
}

sub display_sql {

	print <<EOF;

<h2>
SQL Queries
</h2>
EOF
	print sql_control();
}

#------------------------------------------------------------------------------

sub sql_control {

	# Prebuild search history list control
	my @a = $opt->list('sqlQuery');
	my $attr = 'class="history" style="width:100%" onChange="if (this.value) self.document.sql_control_form.sql.value = this.value;self.document.sql_control_form.sql.focus()" tabindex="-1"';

	my $sql_history_options = Acme::HTML::ListMenus->Menu(\@a,'sql_history','',70,$attr,'','noundef');

	return <<EOF;
<script language="JavaScript" type="text/javascript">
/* ------------------------------------------------------------------------ */

var wn = Math.floor(Math.random() * 10000);   // initial window number

function sql_dispatch(request) {

	var f = self.document.sql_control_form;
	var h = self.document.sql_history_form.sql_history;
	var s = f.sql.value;

	f.request.value = request;
	f.target = "sql_" + wn++;

	switch (request) {
	  case "List":
		if (s.length > 0) f.save_history.value = enter_history(s,h);
		f.action="/cgi-bin/list.cgi";
		open_list("",f.target);
		f.submit();
		break;

	  case "Count":
		if (s.length <= 0) return false;
		f.save_history.value = enter_history(s,h);
		f.action="/cgi-bin/list.cgi";
		open_list("",f.target);
		f.submit();
		break;

	}
	f.sql.select();	// keep focus in form
}
</script>

<div id="sql_control" style="display:block; text-align:center; margin-top:3pt">

<table border="0" cellspacing="0" cellpadding="2">
  <tr valign="top"> 
    <td width=16></td>

    <td><table cellspacing=0 cellpadding=0 border=0 width=100%>
    <tr>

      <td width=105 align=right valign=top nowrap class="plain-title"><nobr>&nbsp; &nbsp; SQL Script</nobr><br>
  
        <a href="/cgi-bin/report.cgi#SQL" onClick="new_target(this,'report')" class="link">Reports</a><br>
        <a href="/cgi-bin/schema.cgi?sort=yes" onClick="new_target(this,'doc')" class="link">Fields</a><br>
        <a href="/cgi-bin/options.cgi#sql_options" onClick="new_target(this,'options')" class="link">Options</a><br>
      </td>
  
      <td width=640 style="padding-left:1pt"> 
        <form name="sql_control_form"
         action="/cgi-bin/list.cgi"
         method="get"
         target="new_sql"
         style="margin:0px"
        >
  
        <textarea name="sql" class="control">@{[$opt->sqlQuery]}</textarea>
  
        <input type=hidden name="request" value="List">
        <input type=hidden name="save_history" value="yes">
        </form>
  
      </td>
      </tr>
      <tr>
      <td width=105 align=right valign=middle nowrap>
        <font class="history">History</font> 
      </td>
      <td style="padding-left:1pt">
        <form name="sql_history_form"
         style="margin:2px 0 0 0"
        >
	$sql_history_options
        </form>
      </td>

    </tr>
    </table></td>

    <td width=105><table border="0" cellspacing="0" cellpadding="1">
        <tr> 
          <td><button class=control accesskey="L" onClick="sql_dispatch('List')"><u>L</u>ist</button></td>
        </tr>
        <tr> 
          <td><button class=control accesskey="C" onClick="sql_dispatch('Count')"><u>C</u>ount</button></td>
        </tr>
      </table></td>

    <td width=16><img src="/images/blank.gif"></td>
  </tr>
</table>

</div>
<br>
EOF
}
__END__
