#!/usr/bin/perl -w -I../lib
require 5.6.0;
use strict;

=head1 NAME

audit.cgi -- Display database change information.

=head1 SYNOPSIS

    audit.cgi?field=customers__customer_name&keyval=11111
     -or-
    audit.cgi?table=customers&column=customer_name&customer_id=11111
     -or-
    audit.cgi?table=customers&customer_id=11111

    # Print with a specified title.
    audit.cgi?field=customers__customer_name&title=Name%20Audit&print=yes

=head1 DESCRIPTION

The audit.cgi program displays a list of the changes to a database table or to
a specified column of a database table. The C<field> argument is a name of a
form field whose name identifies the table and column name. It must conform to
the format "table__column". If the name of the table's key field is not known,
the argument C<keyval> can be used to specify the value of the key field.

Not all tables are audited. The C<audit> setting needs to be set in the
AcmeBase schema files and the database table needs to be built with the
extra fields used for auditing: the user id and the time of the last update.
A field recording the date the record was added is also kept. It is not 
actually needed as it can be queried from the audit table's last update field,
but it is useful information the user appreciates being displayed along with the
time of the record's last update. It is also used to verify the audit trail
coherency.

For performance and space reasons, only the changes to a database record
is recorded in the audit history table. If a database record has never been
changed, then there are no records in the audit history table.

Fields with null values in the audit table signify "No Change". To record the
case where a user changes a field from "something" to "nothing" (C<Null>), a
special record is recorded in the audit history file. It has an C<audit_type>
field set to "N" and any field in the record that has a value signifies
that the field was set to C<Null> by the user. Note that on this type of
audit record, the field values do not mean anything except that the field
was changed to C<Null>.


=head1 CHANGE LOG

 1/03/02 Initial version using audit tables kept by using Postgres rules.
 6/11/02 Changed to list audit tables as kept by Acme::Form::Saver.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2002 Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

package audit;

use Acme::HTTP			qw(escape);
use Acme::HTML			qw(escape_html display_message);
use Acme::Base::Schema;
use Acme::SQL;
use Acme::Lister;
use Acme::HTML::NavButtons;
use Local::Acme::Banners	qw(common_header common_footer);
use Local::Acme::User		qw(user	user_name);
use Acme::Config;
use CGI;

use Acme::Security;
Acme::Security::check('Access');

$|=1;		# flush output (synchronize errors with output)

our $cgi = new CGI;

#
# Program Constants
#

our $Styles = Acme::Config->Styles;		# location of CSS styles

# Hardwire the style of the displayed list. The $lsn variable is used as a
# prefix to the CSS class definitions in the audit.css file.

our $lsn = 'list';			# default CSS style
#our $lsn = 'graybar';			# default CSS style
#our $lsn = 'greenbar';			# default CSS style

#
# Program Arguments
#

our $field	  = $cgi->url_param('field')	|| $cgi->param('field');

our $table	  = $cgi->url_param('table')	|| $cgi->param('table')   || '';
our $column	  = $cgi->url_param('column')	|| $cgi->param('column')  || '';

our $keyval	  = $cgi->url_param('key')	|| $cgi->param('key')	  ||
		    $cgi->url_param('id')	|| $cgi->param('id')	  ||
		    $cgi->url_param('keyval')	|| $cgi->param('keyval')  || '';

our $title	  = $cgi->url_param('title')	|| $cgi->param('title')   || '';
our $format;

our $request	  = $cgi->url_param('request')	|| $cgi->param('request') || '';
our $Printing     = $cgi->url_param('print')    || $cgi->param('print')   || '';
$Printing         = $Printing =~ /yes/i         ||
		    $request =~ /print/i        ||
		    ($Printing !~ /^no$/i && exists $cgi->{'print'} ? 1 : 0);

#
# Set defaults
#

$field ||= $column if $column && ! $table;  # parse column like field argument

if ($field) {
	# Parse table and column names from form field name.
	($table,$column) = $field =~ m/(\w+?)__(\w+)/;
	($table,$column) = $field =~ m/(\w+?)\.(\w+)/ if ! $table;
}

($table,$column) = $column =~ m/(\w+?)\.(\w+)/ if $column && ! $table;

#
# Missing table name. Get necessary parameters from user.
#

if (! $table) {
	get_parameters();
	exit;
}

#
# Check security on table
#

Acme::Security::check('Access',$table);

#
# Get table properties
#

our $schema = new Acme::Base::Schema;

our $t = $schema->t($table)
  or die "Cannot find a database table named '$table'";

if (! $t->audit) {
	print $cgi->header(-type=>'text/html') unless @ARGV;
	display_message("Table is not being audited.", $title,undef,0);
	exit 1;
}

#
# Check several places for the key value.
#

$keyval ||= $cgi->param($t->key)
	||  $cgi->param("${table}." .$t->key)
	||  $cgi->param("${table}__".$t->key)
	||  '';

#
# Missing key value. Get necessary parameters from user.
#

if (! $keyval) {
	get_parameters();
	exit;
}

#
# Get column names used for audits.
#

our $audit_table  = $t->audit_table;
our $audit_update = $t->audit_update;
our $audit_user   = $t->audit_user;
our $audit_entry  = $t->audit_entry;

our $audit_type      = Acme::Config->audit_type;
our $audit_timestamp = Acme::Config->audit_timestamp;
our $audit_userstamp = Acme::Config->audit_userstamp;

# Build hash of audit field names in data table for quick lookup.
my %is_audit_field = map { $_, 1 } $audit_entry, $audit_update, $audit_user;

my $key = $t->key;

#
# Get audited field(s)
#

our @field_names;
if ($column) {
	@field_names = grep {$column eq $_} $t->fields_audit
	  or die "The table column '$column' is not being audited";
}
else {
	@field_names = grep {! $is_audit_field{$_}} $t->fields_audit
	  or die "You are not allowed to see audit information";
}

#
# Assemble fields names into string for SQL. Fully qualify column names.
#

my $fields       = "$table."       . (join ", $table.", @field_names);
my $audit_fields = "$audit_table." . (join ", $audit_table.", @field_names);

#
# Add audit fields to beginning of SQL field list.
#

# Supply the current time and the current user as the audit_timestamp and the
# audit_userstamp fields selected from the data table, which does not have
# those corresponding fields.

our $user = user();

$fields       = qq['I' as $audit_type]
	      . qq[, current_timestamp as $audit_timestamp]
	      . qq[, '$user' as $audit_userstamp]
	      . qq[, $table.$audit_user]
	      . qq[, $table.$audit_entry]
	      . qq[, $table.$audit_update]
	      . qq[, $fields];

$audit_fields = qq[$audit_table.$audit_type]
	      . qq[, $audit_table.$audit_timestamp]
	      . qq[, $audit_table.$audit_userstamp]
	      . qq[, $audit_table.$audit_user]
	      . qq[, $audit_table.$audit_entry]
	      . qq[, $audit_table.$audit_update]
	      . qq[, $audit_fields];

#
# Construct where clause
#

my $where;
if ($t->id eq $t->key || $t->f($key)->is_number) {
	$where = "$table.$key = $keyval";	# don't use quotes
}
else {
	$where = "$table.$key = '$keyval'";	# use quotes
}

my $audit_where = "$where and $table.$key = $audit_table.$key";

(my $title_where = $where) =~ s/^.*\.//;	# used in title

#
# Add security restrictions 
#

# We join the audit table with the table it is auditing so that we can apply
# the same security restrictions to the audit table as what exists for the
# data table.

if (my $restrict_where = $t->restrict_where) {
	$where = "($where) $restrict_where";
	$audit_where = "($audit_where) $restrict_where";
}

#
# Construct SQL
#

my $sql = <<End_Of_SQL;
	select $fields
	from $table
	where $where

	union all

	select $audit_fields
	from $table, $audit_table
	where $audit_where
	order by 6 desc, 1
	;
End_Of_SQL

#
# Define lister object
#

my $l = new Acme::Lister (
		sql          => $sql,
		before_list  => \&display_list_header,
		before_group => [ 6 => \&init_list_item ],
		every_row    => \&fill_list_item,
		after_group  => [ 6 => \&display_list_item ],
		after_list   => \&display_list_footer,
		sandbox      => 'audit',
		);

print $cgi->header(-type=>'text/html') unless @ARGV;

if ($l->count == 0) {
	die "Cannot find a record in table $table with a key value of '$keyval'";
}

if ($l->count == 1) {
	display_message(
		"No audit records were found. No changes have been made.",
		$title,undef,0
		);
	exit 1;
}

#
# Define global variables for list helper functions
#

our $b = new Acme::HTML::NavButtons print => $Printing;

our $n_rec = 0;				# record number

$title ||= "$column of $table"	if $column;
$title ||= "$table table"	if $table;

our $list_title = "Changes made to "
  . ($column ? "column '$column' of " : '')
  . "table '$table' "
  . "where $title_where";

#
# Display list
#

display_header();

$l->print();

display_footer();

exit 0;

#------------------------------------------------------------------------------
#
# Page header and footer functions
#
#------------------------------------------------------------------------------

sub display_header {

	my $print_cmd = '';
	my $print_stylesheet = '';

	if ($Printing) {
  		$print_cmd = 
  qq[ onLoad="window.print();window.setTimeout('window.close()',1500)"];

		$print_stylesheet =
  qq[<link rel="stylesheet" href="$Styles/audit_print.css" type="text/css">\n];
  	}

# Keep IE6 in IE5 compatablilty mode by not specifying url
#<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
#"http://www.w3.org/TR/html4/loose.dtd">

        print <<EOF;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">

<html>
<head>
<title>audit $title</title>

<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<link rel="stylesheet" href="$Styles/audit.css" type="text/css">
$print_stylesheet
</head>
<body$print_cmd>
EOF
	print $b->first_buttons;
	print common_header("Audit",$Printing);

	print <<EOF if $list_title;
<h3>
$list_title
</h3>
EOF

	print <<EOF;
<div class="list">
EOF
}

#------------------------------------------------------------------------------

sub display_footer {

	print "</div>\n";
	print common_footer($Printing);
	print $b->last_buttons;
	print "</body>\n</html>";
}

#------------------------------------------------------------------------------
#
# List helper functions
#
#------------------------------------------------------------------------------


our (@a);				# array of fields read from database

our (@row, @curval);			# array of row values and current values
					# Note: as list is printed, the @curval
					# array becomes the orginal row values.

our %users;				# users who changed records (for legend)

#------------------------------------------------------------------------------

sub display_list_header {

	print qq(<table class="${lsn}" id="list_table">\n);

	#
	# Redefine format for table list.
	#

	$format = join qq(</td>\n<td class="${lsn}-item">), ("%s") x (@field_names);

	#
	# Display arguments as column headers.
	#

	print qq(<tr class="${lsn}-column-header">\n);
	print qq(<td valign="top" colspan="2"></td>\n);
	my $f;
	for (@field_names) {
		print qq(<th nowrap title=") . escape_html($t->f($_)->description) . qq(">);
		print column_heading($_);
		print qq(</th>\n);
	}
	print qq(</tr>\n);
}

#------------------------------------------------------------------------------

sub display_list_footer {

	#
	# Print the re-assembled original values
	#

	print_row(
		qq(<h3><nobr>Original Value</nobr></h3>) .
		  qq(<div align="right"><nobr>) .
		  qq( by ) . $a[3] .
		  qq( on ) . $a[5] .
		  qq(</nobr></div>),
		@curval[6..$#curval]
		);

	print qq(</table>\n);
	print qq(<p>$n_rec changes listed.</p>\n);

	#
	# Print legend of user names
	#

	print qq(<div class="legend">\n);
	print qq(<p>User IDs referred to above:</p>\n);
	print qq(<table>\n);
	print qq(<tr>);

	for (sort { lc($a) cmp lc($b) } keys %users) {
		print qq(<th>$_</th>\n);
		print qq(<td>).user_name($_).qq(</td>\n);
	}

	print qq(</tr>\n);
	print qq(</table>\n);
	print qq(</div>\n);
}

#------------------------------------------------------------------------------


sub init_list_item {

	#
	# Initialize row array
	#

	@row = ();
}

sub fill_list_item {

	#
	# Assemble row to be printed. Combine 'N' type audit records of null
	# changes with audit record of value changes to fields.
	#

	if ($a[0] eq 'I') {

		#
		# Initialize the record's current values array.
		#

		# The I record is the first record in query which is a
		# row of the current db values.

		@curval = @a;
	}
	elsif ($a[0] eq 'N') {

		#
		# Fill null values if audit_type (first value) is 'N'.
		#

		# Save audit fields
		# (audit_type, audit_timestamp, audit_userstamp, audit_user,
		# audit_entry, audit_update).

		@row[0..5] = @curval[0..5];
		@curval[0..5] = @a[0..5];

		# Defined values symbolize null values here when the
		# audit_type field (first field) is 'N'.

		for (my $i = 6; $i < @a; $i++) {
			if (defined $a[$i]) {
				$row[$i] = '(Null)';	# set to null symbol
				$curval[$i] = '';	# maintain current value
			}
		}
	}
	else {	# $a[0] eq 'V'

		#
		# Fill in array of row changes.
		#

		# Save audit fields
		# (audit_type, audit_timestamp, audit_userstamp,
		# audit_user, audit_entry, audit_update).

		@row[0..5] = @curval[0..5];
		@curval[0..5] = @a[0..5];

		for (my $i = 6; $i < @a; $i++) {
			if (defined $a[$i]) {
				$row[$i] = $curval[$i];
				$curval[$i] = $a[$i];	# maintain current value
			}
		}
	}

	#
	# Record user who changed record for the legend.
	#

	$users{$curval[3]} = 1;
}

sub display_list_item {

	if ($a[0] eq 'I') {

		#
		# The "I" record is the first record which is a row of the
		# current db values.
		#

		print_row(
			qq(<h3><nobr>Current Value</nobr></h3>) .
			  qq(<div align="right"><nobr>) .
			  qq( as of ) . $a[1] .
			  qq(</nobr></div>),
			@curval[6..$#curval]
			);

		return;
	}

	#
	# Check for missing audit records.
	#

	# Check if audit record timestamp is the same as the previous
	# record's last_update time.
	# Note: for reference, first six fields of @a array are:
	# 	audit_type, audit_timestamp, audit_userstamp,
	# 	audit_user, audit_entry, audit_update.

	if ($row[5] ne $curval[1]) {
		
		print_row(<<End_Of_Error);
<span class="error">A missing audit record has been detected.
The audit record below was recorded at $curval[1] but the
the data record has a timestamp of $row[5].</span>
End_Of_Error
	}

	#
	# Skip null values if displaying changes to just 1 column.
	# Null values designate no change.
	#

	return if $column && ! defined $a[$#a];

	#
	# Consistancy check if entry_date was changed.
	#

	# Only detect if printing an audit of all fields.

	if (!$column && $row[4] ne $curval[4]) {
		
		print_row(<<End_Of_Error);
<span class="error">A change in the record's entry date has been detected.
It was changed from "$curval[4]" to "$row[4]" sometime before $curval[1].</span>
End_Of_Error
	}

	#
	# Print record assembled in @row.
	#

	$n_rec++;

	print_rec();
}

#------------------------------------------------------------------------------
#
# Support functions
#
#------------------------------------------------------------------------------

sub column_heading {
	my $h = shift;

	my $colhead = 'FieldTitles';	# hardwire choice

	COLUMN_HEADER: {
		if (! $colhead || $colhead =~ /none/i) {
			# Display header line with blank headers.
			# (Other controls are in header line).

			return '';
		}

		if ($colhead =~ /FieldTitles/i) {
			if ($table) {
				# Get titles from schema
				my $f;
				for ($h) {
					s/.*\.//;	# remove table qualifier

					$f = $t->f($_);
					if ($f) {
						$_ = $f->title;
					}
					elsif ($_ eq $audit_entry) {
						$_ = 'Entered';
					}
					else {
						$_ = '';
					}
				}
			}
			else {
				# Make titles out of field names
				for ($h) {
					s/.*\.//; # remove table name
					s/_/ /g;  # change underscores to spaces
					s/\b(\w+)\b/ucfirst(lc($1))/eg; # capitalize words
					s/\bId\b/ID/; # capitalize id
				}
			}

			return $h;
		}

		if ($colhead =~ /QualifiedFieldNames/i) {
			if ($table) {
				for ($h) { $_ = "$table.$_" unless /\./; }
			}
			return $h;
		}

		if ($colhead =~ /FieldNames/i) {
			for ($h) { s/.*\.//; }	# strip table name
			return $h;
		}
	}
}

#------------------------------------------------------------------------------

sub print_rec {

	#
	# Print record index
	#

	# Do not use a numeric index on records, it is confusing because the
	# it goes 1,2,3,... where as the changes go from current to original.
	# The index should count down, but we don't know the end value yet.
	# Uncomment the following if a numeric index is to be displayed.
	#my $ri = sprintf("%4s", $n_rec);
	my $ri = '  ';		# supply an indent
	$ri =~ s/ /&nbsp;/g;

	#
	# Add user and last update to record index
	#

	$ri .= ' ';
	$ri .= $row[3];		# user
	$ri .= ' on ';
	$ri .= $row[5];		# last update

	print_row($ri,@row[6..$#row]);
}

sub print_row {

	my $ri = shift;		# first column, usually the record index number
	
	my $desc = '';
	if (! $field) {
		
		# Create a title attribute for the <tr> tag poping up who changed
		# the item and when. Its usefull for records scrolling off the
		# right side of the screen.

		$desc = $ri;
		$desc =~ s/<.*?>//gs;		# remove html tags
		$desc =~ s/^(\s|&nbsp;)+//s;	# trim leading spaces
		$desc = qq(Changed by $desc)
		  unless $desc =~ /Current/ || $desc =~ /Original/;
		$desc =~ s/ as of.*//s if $desc =~ /Current/;
		$desc = qq( title="$desc");
	}

	#
	# Print a row in the output table.
	#

	if ($n_rec % 2) {
		print qq(<tr class="${lsn}-row-odd"$desc>\n);
	}
	else {
		print qq(<tr class="${lsn}-row-even"$desc>\n);
	}

	print qq(<td class="${lsn}-item-index"><nobr>$ri</nobr></td>\n);
	print qq(<td class="${lsn}-item-buttons"></td>\n);

	#
	# Print row
	#

	no warnings qw(uninitialized);

	print qq(<td class="${lsn}-item">)	if $format !~ /^<td/i;
	print sprintf($format, @_);	
	print qq(</td>) 			if $format !~ /^<td/i;
	print qq(\n</tr>\n);

	return;
}

#------------------------------------------------------------------------------

sub get_parameters {

	my $prompts = <<End_Of_HTML;
<div style="text-align:center;margin-top:.25in">
<table>
<tr>
<td align="right">Table</td>
<td><input type="text" name="table" value="$table"></td>
</tr>
<tr>
<td align="right">Key / ID</td>
<td><input type="text" name="keyval" value="$keyval"></td>
</tr>
</table>
</div>
End_Of_HTML

	require Acme::HTML::Parameters;

	$cgi->cache('no');
	print $cgi->header(-type=>'text/html', -expires=>'-1d');
	Acme::HTML::Parameters->get_parameters(
			title	=> 'Enter Audit Parameters',
			html	=> $prompts,
			method  => 'GET',
			);
	exit;
}
__END__
