#!/usr/bin/perl -wT -I../lib
require 5.6.0;
use strict;

=head1 NAME

browse.cgi -- Display files in a directory on the server.

=head1 SYNOPSIS

    browse.cgi?directory=customers&filter=.txt

     -or-

    browse.cgi?lists=customers

=head1 DESCRIPTION

The browse.cgi program displays a directory list of files located on the 
server computer -- not the client. It is used in a pop-up window to fill
in a form field with the name of a file.

An argument C<directory> specifies a directory relative to the
web site's document root directory. An argument C<lists> specifies a
directory where the user's saved lists files are located.

Directory files are not displayed and the user cannot migrate around
the file system. Only the regular files that passes the filter pattern
are listed. However, more than one directory can be displayed if listed
in the arguments. Each directory has an anchor with its name at the 
beginning of their listing.

Full path names are not used for security reasons. The program accepts key
words which map to server directories. For example, the C<directory> argument
as used in the link C<browse.cgi?directory=dirname> causes the program to only
look at files in the C<dirname> directory below the web site's document root
directory. The C<lists> argument in the link C<browse.cgi?lists=customers>
causes the program to only look at files in the C<customers> directory
below the C<List_Space> directory as configured in the site configuration
file C<.../etc/site.conf>.

Other key words representing root directories besides C<directory> and
C<lists> can be set up in the site configuration file C<...etc/site.conf>.

The C<filter> parameter is a regular expression. The special characters
need to be escaped using the HTTP conventions.

The C<link> parameter can be used to specify a link to use to display a
file name. A single '$' in the link gets substituted with the file name.

Unless a C<link> is specified, the selected file name is returned to the
calling window like the C<search.cgi> program does. The calling window must
have a JavaScript routine named _browse_return_WINDOWNAME(), where WINDOWNAME
is the name of the window displaying the directory list.

The selected file name is assigned to the JavaScript variable C<result.value>
before the _browse_return_WINDOWNAME() function is called.

Example:

	var _browse_win_EXAMPLE;		// Saved browse window handle

	function _browse_EXAMPLE(options) {	// Open browse window
		var url = '/cgi-bin/browse.cgi?$browse';
		_browse_win_EXAMPLE = new_browse(url,'EXAMPLE',options);
		return true;
	}

	function _browse_return_EXAMPLE() {	// Return from browse window
		var f = self.document.forms['MYFORM'];
		if (_browse_win_EXAMPLE && _browse_win_EXAMPLE.result) {
			var r = _browse_win_EXAMPLE.result;
			// Assign selection values to fields.
			if (f.MYFIELD && r.value) {
				f.MYFIELD.value = r.value;
			}
		} else {
			// No records were found.
			f.MYFIELD.focus();
		}
	}

The above method, used by the Form Builder, is good for having several
Browse buttons on a form with each button having potentially different
actions upon return.

A simpler method can also be used to retreive the file name from the Browse
window. It uses a standard function browse_return() and returns the file name
as one of the arguments. This method also uses the Browse window's name to keep
track of which field to update.

Example:

	var browse_args = new Object;		// Saved arguments for browse_return()

	function browse(list, field) {
		// Open browse window.
		var url = '/cgi-bin/browse.cgi?lists='+ escape(list);
		var window_name = "w_" + field.name;

		// Save arguments for return
		browse_args[window_name] = field;

		// Open pop-up window
		var w = window.open("", window_name, settings);
		w.focus();
		w.location.href = url;
		return true;
	}

	function browse_return(window_name,file_name) {
		var field = browse_args[window_name];

		if (!file_name) return false;	// no return value

		field.value = file_name;

		field.focus();
		if (field.select) field.select();

		return true;
	}

=head1 CHANGE LOG

 1/15/02  Initial version.
 7/5/02   Added simpler return method.
 11/27/02 Added multiple directories.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

package browse;

use Acme::HTML			qw(display_message);
use Acme::HTTP	qw(escape);
use Acme::Config;
use CGI;

use Acme::Security;
Acme::Security::check('Access');

our $DS='/';	# directory separator character

our $Default_Link = q[javascript:browse_return('$')];

our $Browse_Roots = Acme::Config->Browse_Roots || {};
our $Scripts      = Acme::Config->Scripts;
our $Styles       = Acme::Config->Styles;

local $|=1;	# flush output (synchronize errors with output)

our $cgi = new CGI;
print $cgi->header(-type=>'text/html') unless @ARGV;

if (! $cgi->param()) {
	print "<html><body onLoad='self.focus()'>(empty)</body></html>\n";
	exit 0;
}

our $cgi_title  = $cgi->param('title')  || '';	# window title

our $cgi_filter = $cgi->param('filter') || '';
our $cgi_link 	= $cgi->param('link');

our @file_names;

our ($directory, $title, $name, $filter, $link);

our $param_dir;		# directory specified by program parameter

#
# Set usage flags.
#

our $page_header_displayed = 0;		# flag
our $cancel_button_displayed = 0;	# flag

#
# Find first CGI parameter that matches one of the root directories.
#

for my $root ($cgi->param()) {
	next unless $Browse_Roots->{$root};

	Acme::Security::check('Access',$root);

	for $param_dir ($cgi->param($root)) {

		# Don't allow directory names with '../'.
		die "$param_dir is an invalid path name" if $param_dir =~ /\.\.$DS/;

		for ($Browse_Roots->{$root}) {

			#
			# Set global variables for display functions.
			#

			# Note: $directory can be dynamically determined if
			# it is set to a sub reference.

			$directory =
			  (ref $_->{directory}) =~ /CODE/
			  ? &{$_->{directory}}
			  : $_->{directory};

			# Append $param_dir to root directory.
			$directory .=
			  $param_dir =~ m|^$DS|
			  ? $param_dir
			  : "$DS$param_dir";

			$name =
			  ($param_dir eq '.' || $param_dir eq $DS)
			  ? ($_->{name} || $root)
			  : $param_dir;

			$title  = $_->{title} || $cgi_title || '';
			$filter = $_->{filter};
			$link   = $cgi_link   || $_->{link} || $Default_Link;
		}

		get_file_names();

		display_page_header();
		display_directory_header();
		display_directory_list();
		display_directory_footer();
	}
}

die "A directory was not specified"
  . " or it was not registered in the site configuration file. Error"
  unless $directory;

display_page_footer();

exit;

#-------------------------------------------------------------------------------

sub get_file_names {

	#
	# Get list of files
	#

	opendir DIR, $directory or do {
		@file_names = ();
		warn "Warning: Cannot find $name directory $directory: $!";
	};

	@file_names = readdir DIR;

	closedir DIR;
	shift @file_names; shift @file_names;	# remove . and ..
}

#-------------------------------------------------------------------------------

sub display_page_header {

	return if $page_header_displayed;

	#
	# Print header
	#

	my $window_title = $cgi_title
			 ? "browse $cgi_title"
			 : "browse Files";

	print <<EOF;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">

<html>
<head>
<title>$window_title</title>
<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<script src="$Scripts/common.js" type="text/javascript"></script>
<script language="JavaScript" type="text/javascript">
var result = new Object;
function browse_return(filename) {

	// Return to calling window
	var w = window.opener;
	if (!w || w.closed) { self.close(); return; }	// user closed caller already

	// Construct custom return function name
	var return_func = "_browse_return_" + window.name;
	if (eval("w." + return_func)) {

		// Define selection result values
		result.value = unescape(filename);

		// Normal return. At this point hide and let the calling window
		// read the return variables set according to the user's selection.
		self.blur();				// hide window
		w.focus();				// bring to top calling window
		eval("w." + return_func + "()");	// return to calling window func
	}
	else if (w.browse_return) {	// Try standard function name
		w.browse_return(window.name,filename);
	}

	self.close();
	return;	
}
self.focus();
</script>
</head>
<body>
EOF

	if ($title) {
		print <<EOF;
<table width="100%">
<tr>
 <td align="left">
  <span style="font-size:xx-large; text-weight:bold">$title</span>
 </td>

 <td align="right">
  <input type=button value="Cancel" onClick="self.close()" class=button>
 </td>
</tr>
</table>
EOF
		$cancel_button_displayed = 1;	# set flag
	}

	$page_header_displayed = 1;	# set flag
}

#-------------------------------------------------------------------------------

sub display_page_footer {

	print <<EOF;
</body>
</html>
EOF
}

#-------------------------------------------------------------------------------

sub display_directory_header {

	my $cancel_button = $cancel_button_displayed
	  ? ''
	  : qq[<input type=button value="Cancel" onClick="self.close()" class=button>];

	my $anchor = $name;
	$anchor =~ s/"//g;	# clean up name so it won't break anchor
	$anchor = qq[<a name="$anchor">\n];

	print <<EOF;
$anchor<table width="100%">
<tr>
 <td align="left">
  <span style="font-size:x-large; text-weight:bold">$name</span>
 </td>

 <td align="right">
  $cancel_button
 </td>
</tr>
</table>
<hr>
EOF
	$cancel_button_displayed = 1;	# set flag
}

#-------------------------------------------------------------------------------

sub display_directory_footer {

	return;
}

#-------------------------------------------------------------------------------

sub display_directory_list {

	#
	# Print listing, one file to a line, each line a link.
	#

	my $count = 0;
	my $full_link;
	my $encoded_string;

	print qq[<ul style="margin-left:4ex">\n];
	for (sort { lc($a) cmp lc($b) } @file_names) {

		next if $filter     && ! /$filter/;	# filter out files to ignore
		next if $cgi_filter && ! /$cgi_filter/;	# filter out files to ignore
		next unless ( -f "$directory/$_" && -r _ );

		$count++;

		$encoded_string = escape( $param_dir =~ m|$DS$|
					  ? "$param_dir$_"
					  : "$param_dir$DS$_"
					  );

		($full_link = $link) =~ s/\$/$encoded_string/g;

		print qq[<li><a href="$full_link">$_</a></li>\n];
	}
	print qq[</ul>\n];

	print qq[(empty)\n] unless $count;

	print qq[<br>&nbsp;\n];
}
__END__
