#!/usr/bin/perl -wT -I../lib
require 5.6.0;
use strict;

=head1 NAME

calendar.cgi -- Display a calendar

=head1 SYNOPSIS

    calendar.cgi
     -or-
    calendar.cgi?m=1&y=2000&n=3
     -or-
    calendar.cgi?y=2000
     -or-
    calendar.cgi?y=2000&calendar_id=1
     -or-
    calendar.cgi?calendar_id=1&calendar_id=2&calendar_id=3	# superimpose
     -or-
    calendar.cgi?search=Company	# search on calendar titles
     -or-
    calendar.cgi?calendar_id=2&calendar=3,1-1-2000,3-31-2000
     -or-
    calendar.cgi?calendar_id=2,3,4&calendar=3,1-1-2000,3-31-2000&legend=#33ff55,#22aacc,#fafafa
     -or-
    calendar.cgi?calendar_id=4&calendar=3,1-1-2000,3-31-2000&filter=Foo.*

=head1 DESCRIPTION

Display a calendar.

This program displays the calendars listed in the calendars table and the
calendar_events table. A calendar record in the calendars table can refer
to a row in the calendar_reports table which describes a calendar interface
to other database tables. For example: a calendar of order shipment dates.

More than one calendar can be superimposed on one display.

The C<calendar=n,bgndate,enddate> argument is used by the calendar function
buttons to request a calendar in another date range or a different type of
calendar. Those options override other date and calendar style options.
The calendar's JavaScript strips the C<calendar> argument off the end of the
cgi arguments and re-appends it when calling for a new calendar to be rendered.

The C<legend=color1,color2,color3> argument can be used to specify color codes
for each calendar superimposed. Likewise, the C<text=color1,color2,color2>
argument can be used to specify the color of the text of calendar entries.
The existance of the argument C<superimpose> causes the colors to darken for
each different calendar entry on a day.

The C<filter=value1,value2,value3> argument can be used to specify filter
setting to merge into the SQL for querying calendar events.

=head1 CHANGE LOG

 7/30/01 Initial version.
 8/24/01 Added interface to other database calendars through the
	 calendar_reports table definitions.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

package Calendars;

use Acme::Base::Object;

our @ISA = qw(Acme::Base::Object);

sub calendars {
	my $self = shift;
	return @{$self->components};
}


package Calendar;

use Acme::Base::Object;

our @ISA = qw(Acme::Base::Object);


package calendar;

use CGI;
use Acme::HTML		qw(escape_html display_message);
use Acme::HTML::Parameters;
use Acme::HTML::NavButtons;
use Acme::HTML::ListMenus;
use Acme::Base::Schema;
use Acme::Db;
use Acme::Searcher;
use Acme::Calendar::DayMenu;
use Acme::Date		qw(dayadd short_date_range longdate);
use Acme::Options;
use Acme::Config;
use Local::Acme::Banners	qw(common_header common_footer);

use Acme::Security;
Acme::Security::check('Access');

#
# Begin calendar
#

$|=1;		# flush output (synchronize errors with output)

our $cgi = new CGI;
$cgi->cache('no');
print $cgi->header(-type=>'text/html', -expires=>'-1d') unless @ARGV;

#
# Set program variables / constants
#

our $Styles      = Acme::Config->Styles;	# location of CSS style sheets
our $Scripts     = Acme::Config->Scripts;	# location of JavaScripts
our $Scripts_Lib = Acme::Config->Scripts_Lib;	# location of JavaScripts
our $Bullet      = Acme::Config->Bullet;	# edit bullet
our $DS          = Acme::Config->Date_Separator;

our $Default_Calendar_Type = 6;			# default type if none other

#
# Define an array of background colors to use depending upon the
# order that the calendars are read from the database. The first calendar
# uses the first color in @Legend, the second calendar uses the second color....
#

our @Legend_Colors = @{Acme::Config->Calendar_Colors};
our @Text_Colors   = @{Acme::Config->Calendar_Text_Colors};

#
# Define parameters that are to be used to initially select daily items.
#

our @Params = ();

#
# Define filters that are used to restrict the selection of daily items.
#

our @Filters = ();

#
# Totals (and counts) can be kept for each calendar displayed on the legend
#

our @Totals = ();
our @Counts = ();

#
# Get user options
#

our $Opt = new Acme::Options;

#
# Parse cgi arguments
#

our $Request	= $cgi->param('request')	|| '';

our $Printing	= $cgi->param('print')		|| '';
$Printing	= $Printing =~ /yes/i           ||
		    $Request =~ /print/i        ||
		    (($Printing !~ /^no$/i && exists $cgi->{'print'}) ? 1 : 0);

our $Start_Date	= ($cgi->param('start_date') || '') && $Start_Date =~ s|/|-|g;
our $End_Date	= ($cgi->param('end_date')   || '') && $End_Date   =~ s|/|-|g;

our $Title	= $cgi->param('title')  || '';

our $Type	= $cgi->param('type')   || '';

our $Resize	= $cgi->param('resize') || 'yes';

# Allow search to be a list of ID numbers or an array of ID numbers.
my $search = join ",", $cgi->param('search');
$search  ||= join ",", $cgi->param('calendar_id'),
			$cgi->param('calendars__calendar_id'),
			$cgi->param('calendars.calendar_id');
if (! $search) {
	# Set default search parameters so that search of calendars table
	# fails gracefully and just displays an empty calendar.
	$search = Acme::Config->Default_Calendar || '0';
	$Title ||= 'Calendar';		# set to blank default
}

# Superimpose is a flag. It can be a 1 which uses a default color, or it can
# specify a color code to use for superimposing. The color '#ffffff' (white)
# is ignored, effectivily turning off changing colors for superimposed fields.

our $Superimpose = $cgi->param('superimpose') || exists $cgi->{superimpose};
$Superimpose = $Opt->get("CalendarSuperimpose$search") || '1'	# alt: '#ddccbb'
  if $Superimpose eq '1' || $Superimpose eq 'on';
$Superimpose = '0' if lc($Superimpose) eq '#ffffff';	# ignore 'white'

my ($m,$y) = (localtime)[4,5];		# set defaults for today

my $month	 = $cgi->param('m')	|| $cgi->param('month');
my $year	 = $cgi->param('y')	|| $cgi->param('year');
my $Number	 = $cgi->param('n')	|| 0;

$Number     ||= 12 if $year && ! $month;
$month      ||= $m + 1;
$year       ||= $y + 1900;
$Start_Date ||= sprintf("%02d${DS}%02d${DS}%d",$month,1,$year);

# User option settings from last viewing of calendar overrides above arguments.

my $user_number = $Opt->get("CalendarType$search");
$Number = $user_number if $user_number;

my $user_legend = $Opt->get("CalendarLegend$search");
@Legend_Colors = split(/,/,$user_legend) if $user_legend;

my $user_text = $Opt->get("CalendarText$search");
@Text_Colors = split(/,/,$user_text) if $user_text;

# The calendar parameter overrides above arguments.

my $calendar = $cgi->param('calendar');
if ($calendar) {
	my $number = $Number;
	($Number,$Start_Date,$End_Date) = split(/,/,$calendar);
	$Number ||= $number;
}

# Allow legend colors to be specified

my $legend = join ",", $cgi->param('legend');
my $text   = join ",", $cgi->param('text');

# Gather parameters/filters

@Params = $cgi->param('param');
@Filters = $cgi->param('filter');

our @User_Params = split(/,/,$Opt->get("CalendarParams$search"));

#
# Get schema definitions
#

our $Schema = new Acme::Base::Schema;

my $table = 'calendars';

my $t = $Schema->t($table)
  or die "Cannot find database table for '$table'";

#
# Define SQL query for finding the calendar parameter definitions.
#

my $sql = <<End_Of_SQL;
select
	calendars.calendar_id,
	calendars.title,
	calendars.calendar_style,
	calendars.every_year,
	calendars.calendar_color,
	calendars.text_color,
	calendar_reports.calendar_report_id,
	calendar_reports.table_name,
	calendar_reports.key_field,
	calendar_reports.begin_date_field,
	calendar_reports.begin_time_field,
	calendar_reports.end_date_field,
	calendar_reports.end_time_field,
	calendar_reports.event_type_field,
	calendar_reports.event_description_field,
	calendar_reports.event_status_field,
	calendar_reports.button_hilite_field,
	calendar_reports.importance_field,
	calendar_reports.every_year_field,
	calendar_reports.from_clause,
	calendar_reports.where_clause,
	calendar_reports.parameter_prompt,
	calendar_reports.filter_label,
	calendar_reports.filter_clause,
	calendar_reports.filter_list,
	calendar_reports.total_field,
	calendar_reports.total_label,
	calendar_reports.edit_link_url,
	calendar_reports.edit_link_text,
	calendar_reports.view_link_url,
	calendar_reports.view_link_text,
	calendar_reports.action_link_url,
	calendar_reports.action_link_text,
	calendar_reports.action_message,
	calendar_reports.new_link_url,
	calendar_reports.new_link_text,
	calendar_reports.prepare_sql
from calendars, calendar_reports
where ( @{[$t->list_where ]} )
  and calendars.calendar_report_id = calendar_reports.calendar_report_id
order by @{[$t->list_order_by_clause ]};
End_Of_SQL

#
# Evaluate Search script
#

if (length $search) {

	my $sql_search = $search;

	my $id_field	    = $t->id;
	my $name_field	    = $t->name_field;
	my $record_set_name = $t->record_set_name;

	if ($id_field || $name_field) {

		#
		# Preprocess search script
		#

		$sql_search = Acme::Searcher::preprocess($search, {
		  RecordSetName  => $record_set_name,
		  IDField	 => "$table.$id_field",
		  NameField	 => "$table.$name_field",
		});
	}

	#
	# Parse SQL statement
	#

	my $sqlo = new Acme::SQL $sql;

	#
	# Combine sql with search script
	#

	my $srch = new Acme::Searcher $sql_search, $cgi;
	$sql = $srch->as_sql($sqlo);	

	#
	# Redefine title if in script
	#

	## Ignore search title, which is auto-generated if not specified.
	## Use the calendar titles read from the database below.
	##$Title ||= $srch->title if $srch->title;
}

#
# Query calendars
#

$sql = Acme::Security::clean_sql($sql);

my $q = Acme::Db->query($sql);

if ($search && $q->count == 0) {
	display_message("No calendars were found.", "$Title",undef,0);
	exit 1;
}


$q->set_nulls_to_empty;		# eliminate -w warnings on null fields

#
# Set up parameters for querying calendar events
#

our $Sql = '';		# sql for querying calendar entries

our $Cals = new Calendars; # object to store all the calendars and titles

our $Day_Menu_Header = ''; # menu of links to add new items to calendar

my $calendar_ids = '';	# collected calendar IDs of all superimposed calendars
my $titles = '';	# collected titles of all superimposed calendars

my $number = 0;		# used as flag of user setting for number of calendars
my $ncal = 0;		# number of calendars, index for building legend.
my $n_new_links = 0;	# number of "Add ..." links for adding new items
			# This is different than $ncal because of security.

my $prompts = '';	# query prompts to ask the user first
my $need_parameters = 0;# flag set to true if calendar needs user parameters

# Define function for substituting variables found in url strings.

my $tr = sub {
	my $v = shift;
	return '$date'        if $v eq 'date';		# don't substitute here
	return '$filter'      if $v eq 'filter';	# don't substitute here
	return $Params[$ncal] if $v eq 'param';
	no strict 'refs';
	return ${$v};
};

#
# Scan all calendars first
#

while (our (
		$calendar_id,
		$title,
		$calendar_style,
		$every_year,
		$calendar_color,
		$text_color,
		$calendar_report_id,
		$table_name,
		$key_field,
		$begin_date_field,
		$begin_time_field,
		$end_date_field,
		$end_time_field,
		$event_type_field,
		$event_description_field,
		$event_status_field,
		$button_hilite_field,
		$importance_field,
		$every_year_field,
		$from_clause,
		$where_clause,
		$parameter_prompt,
		$filter_label,
		$filter_clause,
		$filter_list,
		$total_field,
		$total_label,
		$edit_link_url,
		$edit_link_text,
		$view_link_url,
		$view_link_text,
		$action_link_url,
		$action_link_text,
		$action_message,
		$new_link_url,
		$new_link_text,
		$prepare_sql
	  ) = $q->next()) {

	$calendar_ids .= ($calendar_ids ? ',' : '') . $calendar_id;
	$titles       .= ($titles ? '<br>' : '')    . $title;
	$titles =~ s/<br>.*/<br>and $ncal Other Calendars/s if $ncal > 2;

	$Legend_Colors[$ncal] = $calendar_color if ! $user_legend && $calendar_color;
	$Text_Colors[$ncal]   = $text_color     if ! $user_text && $text_color;

	# Initialize accumulators

	$Counts[$ncal] = 0;
	$Totals[$ncal] = 0;

	# Gather parameters to ask user
		
	$Params[$ncal] ||= '';
	if ($parameter_prompt && length($Params[$ncal]) == 0) {
		$need_parameters = 1;			# set flag
		my $value = $User_Params[$ncal] || '';
		$prompts .= <<End_Of_HTML;

<tr>
<td>$parameter_prompt</td>
<td><input type="text" name="param" value="$value"></td>
</tr>
End_Of_HTML
	}
	else {
		# Track empty parameters of superimposed calendars
		$prompts .= qq[<input type="hidden" name="param" value="">];
	}

	# Build menu for adding new items to the calendar

	if (Acme::Security::allow('Insert',$table_name)) {
		# Substitute calendar variables except $date.
		$new_link_url =~ s/\$(\w+)/&$tr($1)/eg;

		$new_link_text ||= 'Add ' . lc $title . ' ...';

		$Day_Menu_Header .= <<End_Of_HTML;
<div class="day-menu-header-item"><a href="$new_link_url" onClick="new_target(this,'form')" target="$table_name">$new_link_text</a></div>
End_Of_HTML
		$n_new_links++;
	}
	else {
		# User not allowed to add new items.
		$new_link_url = '';
	}
	
	# Check with security if user can edit items

	if (Acme::Security::allow('Update',$table_name)) {
		# Substitute calendar variables except $date.
		$edit_link_url =~ s/\$(\w+)/&$tr($1)/eg;

		$edit_link_text ||= "Edit item.";
	}
	else {
		# User not allowed to edit items.
		$edit_link_url = '';
	}


	# Susbstitute parameters into where clause and filter

	$where_clause  ||= '';
	$filter_clause ||= '';

	$where_clause  =~ s/\$(\w+)/&$tr($1)/eg;
	$filter_clause =~ s/\$(\w+)/&$tr($1)/eg;

	#
	# Build structure of calendar parameters used to construct the SQL
	# and menu items.
	#

	$ncal++;

	my $cal = new Calendar;
	$cal->set(
		index      		=> $ncal,

 		calendar_id		=> $calendar_id,
 		title			=> $title,
 		calendar_style		=> $calendar_style,
 		every_year		=> $every_year,
 		calendar_color		=> $calendar_color,
 		text_color		=> $text_color,
 		calendar_report_id	=> $calendar_report_id,
 		table_name		=> $table_name,
 		key_field		=> $key_field,
 		begin_date_field	=> $begin_date_field,
 		begin_time_field	=> $begin_time_field,
 		end_date_field		=> $end_date_field,
 		end_time_field		=> $end_time_field,
 		event_type_field	=> $event_type_field,
 		event_description_field	=> $event_description_field,
 		event_status_field	=> $event_status_field,
 		button_hilite_field	=> $button_hilite_field,
 		importance_field	=> $importance_field,
 		every_year_field	=> $every_year_field,
 		from_clause		=> $from_clause,
 		where_clause		=> $where_clause,
 		parameter_prompt	=> $parameter_prompt,
 		filter_label		=> $filter_label,
 		filter_clause		=> $filter_clause,
 		filter_list		=> $filter_list,
 		total_field		=> $total_field,
 		total_label		=> $total_label,
 		edit_link_url		=> $edit_link_url,
 		edit_link_text		=> $edit_link_text,
 		view_link_url		=> $view_link_url,
 		view_link_text		=> $view_link_text,
 		action_link_url		=> $action_link_url,
 		action_link_text	=> $action_link_text,
		action_message		=> $action_message,
 		new_link_url		=> $new_link_url,
 		new_link_text		=> $new_link_text,
 		prepare_sql		=> $prepare_sql
		);

	$Cals->add($calendar_id => $cal);

	#
	# Finish setting up parameters for querying calendar events
	#

	if (! $number) {
		if (! $calendar) {
			$Number ||= $Opt->get("CalendarType$calendar_id");
			$Number ||= $calendar_style || $Default_Calendar_Type;
		}
		$number = $Number;	# set as flag so this does not repeat
	}
}

$Cals->set('n_new_links' => $n_new_links);

#
# Ask for any prompts from the user 
#

if ($need_parameters) {
	$prompts = <<End_Of_HTML;
<div style="text-align:center;margin-top:.25in">
<table>$prompts
</table>
</div>
End_Of_HTML

	Acme::HTML::Parameters->get_parameters(
			title	=> 'Enter Search Parameters',
			html	=> $prompts,
			method  => 'GET',
			);
	exit;
}

#
# Set final defaults
#

# Check if $Number is a calendar type of 'full', 'list' or in future, 'day'.

if ($Number =~ /full/i || $Number eq '0') {
	$Type = 'full';
	$Number = 1;
}
elsif ($Number =~ /list/i) {
	$Type = 'list';
	$Number = 1;
}
else {
	$Type ||= "$Number";
}

$Start_Date ||= sprintf("%02d${DS}%02d${DS}%d",$month,1,$year);

# Assure $Start_Date starts at the beginning of the period.

($month,$year) = $Start_Date =~ /(\d+).\d+.(\d+)/;
$month = 1 + (int(($month-1)/$Number) * $Number);
$Start_Date = sprintf("%02d${DS}%02d${DS}%d",$month,1,$year);

if (! $End_Date) {
	my ($m, $d, $y) = split(/\W+/,$Start_Date);
	$m += $Number > 12 ? 12 : $Number;
	if ($m > 12) { $y++; $m -= 12 }
	# The dayadd() routine uses unix time() so years restricted.
	if ($y >= 1970 && $y < 2038) {
		# Get last day of previous month.
		$End_Date = dayadd(-1,"${m}${DS}${d}${DS}${y}"); 
	}
}

$Title ||= $titles if $titles;
$Title ||= "$year Calendar";
(our $Window_Title = $Title) =~ s/<.*?>/ /g;	# remove html codes

#
# Override colors with cgi line colors.
#

@Legend_Colors = split(/,/,$legend) if $legend;
@Text_Colors   = split(/,/,$text)   if $text;

#
# Save user's calendar options for next time
#

$Opt->set("LastCalendar" => $calendar_ids);
$Opt->set("LastCalendarSuperimpose" => $Superimpose);
$Opt->set("CalendarType$search"   => $Type)
  if $calendar;
$Opt->set("CalendarLegend$search" => (join ",", @Legend_Colors[0..($ncal-1)]))
  if $legend;
$Opt->set("CalendarText$search"   => (join ",", @Text_Colors[0..($ncal-1)]))
  if $text;
$Opt->set("CalendarSuperimpose$search" => $Superimpose)
  if $calendar;
$Opt->set("CalendarParams$search" => (join ",", @Params)) if @Params;
$Opt->save;

#
# Display calendar
#

$b = new Acme::HTML::NavButtons print => $Printing;

display_header();

display_calendar();

# Database items are displayed only if their dates are between 1970 and 2039.
# This is a limitation in Acme::Date::dayadd(), which uses the Unix clock.

my ($start_year) = $Start_Date =~/.*\W(\d+)/;
my ($end_year)   = $End_Date   =~/.*\W(\d+)/;
$end_year ||= 0;	# eliminate -w warnings
			# happens when calendar not specified, so don't query

if ($ncal > 0
&& $start_year >= 1970 && $end_year < 2038
&& $start_year <= $end_year) {

	#
	# Read all calendars and construct SQL for searching daily records.
	#

	for my $calendar_id ($Cals->calendars) {
		$Sql .= ($Sql ? "union\n" : '') . make_sql($calendar_id);
	}

	#
	# Define SQL sort order to be:
	#
	# y1, m1, d1, begin_time, importance desc, event_type
	#

	$Sql .= 'order by 1, 2, 3, 12, 19 desc, 15';

	#
	# Display calendar contents
	#

	display_items();

	display_legend();
}

display_footer();

exit 0;

#-------------------------------------------------------------------------------

sub display_header {

	#
	# Compute window size according to number of calendars displayed.
	# Note: these window sizes are bigger than calendar.js because they
	# include toolbar and scrollbar.
	#

	my $resize = '';
	if ($Resize =~ /[yY][eE][sS]/ && ! $Printing) {
		my ($w, $h);
		if ($Type =~ /\d+/) {
			if ($Number > 6)     { $w=800; $h=1010;}
			elsif ($Number > 3)  { $w=800, $h=785; }
			elsif ($Number > 1)  { $w=800, $h=590; }
			elsif ($Number == 1) { $w=320, $h=440; }
		}
		else {
			$w=800, $h=785;
		}

		$h += 26 if $Number == 1 && @Filters;	# room for scroll bar

		$resize = qq[
<script language="JavaScript" type="text/javascript">window.resizeTo($w,$h)</script>];
	}

	#
	# Set stylesheet depending upon calendar type
	#

	my $stylesheet = '';
	$stylesheet = qq[
<link rel="stylesheet" href="$Styles/calendar_list.css" type="text/css">]
	  if $Type eq 'list';

	$stylesheet = qq[
<link rel="stylesheet" href="$Styles/calendar_full.css" type="text/css">]
	  if $Type eq 'full';

	#
	# Display header
	#

	my $onload = '';
	$onload = qq[ onLoad="window.print();setTimeout('self.close()',3000)"]
	  if $Printing;
		
## Keep IE6 in compatability mode by NOT specifying DOCTYPE.
## It screws up the placement of popup menus.
##	print <<EOF;
##<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
##"http://www.w3.org/TR/html4/loose.dtd">
##EOF
	print <<EOF;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
<head>
<title>cal $Window_Title</title>

<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<link rel="stylesheet" href="$Styles/calendar.css" type="text/css">$stylesheet

<script src="$Scripts_Lib/calendar.js" type="text/javascript"></script>
<script src="$Scripts_Lib/colorwheel.js" type="text/javascript"></script>
$resize
</head>
<body${onload}>
EOF

	if ($Type eq '1') {
		# Single calendars are too small for company header,
		# but we need common JavaScript to open forms and windows
		print qq(<script src="$Scripts/common.js" type="text/javascript"></script>\n);
	}
	else {
		print $b->first_buttons;
		print common_header($Title,$Printing);
	}
}

#-------------------------------------------------------------------------------

sub display_footer {
	
	if ($Type ne '1') {
		# Single calendars are too small for company footer.
		print common_footer($Title,$Printing);
		print $b->last_buttons;
	}

	print qq(</body>\n</html>);
}

#-------------------------------------------------------------------------------

sub display_calendar {

	#
	# Construct calendar in JavaScript.
	#

	my $print = $Printing ? qq[,\n\t'print':         true] : '';

	# Add extra menu items this cgi knows how to handle.
	my $menu_items = <<'EOF';
var menu_items = ''
  + '<a href="javascript:do_request(\'setstyle\',\'full\')">Full Screen</a>\n'
  + '<a href="javascript:do_request(\'setstyle\',\'list\')">List</a>\n'
EOF

	if ($Type eq 'full' || $Type eq 'list') {

		#
		# Build calendar for Full Screen or List calendars
		#

		my $new_link = 'return false;';
		my $callback_url = '';
		my $button_title = '';
		my $callbacks = '';
		my $day_menu = '';
		if ($Cals->n_new_links > 0) {
			if ($Cals->n == 1) {
				$callback_url = 
				    qq[var callback_url = '] 
				  . $Cals->{$calendar_ids}->new_link_url
				  . qq[';\n];
				$new_link = 
				    qq[open_form(]
				  . qq[callback_url.replace(/\\\$\\{?date\\}?/,date_str)]
				  . qq[);];

				$button_title =
				    $Cals->{$calendar_ids}->new_link_text;
				$button_title =
				    qq[,\n\t'button_title': '$button_title'];

				$callbacks =
				    qq[,\n\t'callback':     'calCallback'];
			}
			else {
				$day_menu  =
				    qq[,\n\t'id_prefix': 'day_menu_'];
				$callbacks =
				    qq[,\n\t'click':        'showMenu()']
				  . qq[,\n\t'unclick':      'hideMenu()']
				  . qq[,\n\t'autoclick':     0.5];
			}
		}

		print <<EOF;
<script language="JavaScript" type="text/javascript">
$callback_url
function calCallback(date_obj, date_str, is_active) {
	$new_link
}

$menu_items
displayCalendar( {
	'n_months'   :  1,
	'type'       : '$Type',
	'start_date' : '$Start_Date',
	'window'     :  self,
	'menu'       :  menu_items$callbacks$button_title$day_menu$print} );
</script>
EOF

		if ($Type eq 'list') {

			# Note: this is a header for the entire list.
			# If the list spans more than one month, then
			# the header for each month should be output by
			# Acme::Calendar::DayMenu, not here. At this 
			# time, lists are assume to span only 1 month.

			my ($m,$y) = (split(/ /,longdate($Start_Date)))[1,3];
			print <<EOF;
<span class="month-title">$m / $y</span>
EOF
		}
	}
	else {
		#
		# Build calendar for 1, 3, 6, and 12 months
		#

		print <<EOF;
<script language="JavaScript" type="text/javascript">
$menu_items
displayCalendar( {
	'n_months'   : $Number,
	'start_date' : '$Start_Date',
	'window'     : self,
	'click'      : 'showMenu()',
	'unclick'    : 'hideMenu()',
	'autoclick'  : 0.5,
	'menu'       : menu_items$print} );
</script>
EOF
	}
}

#-------------------------------------------------------------------------------

sub display_legend {
	
	our $No_Tabindex = Acme::Config->No_Tabindex || -1;

	# Legend is displayed in three columns, left is colors, center is
	# filters, and the right is totals.

	print <<End_Of_HTML;

<form name="legend" class="legend">
<div class="legend">
<table class="legend" cellspacing="0" cellpadding="0">
End_Of_HTML

	my ($color,$text_color);
	my $index = 0;
	for my $calendar_id ($Cals->calendars) {

		$color      = $Legend_Colors[$index] || 'red';
		$text_color = $Text_Colors[$index]   || 'black';

		#
		# Print left column of legend colors
		print qq[<tr>\n];
		print qq[<td class="legend-left">];
		print qq[<div class="legend-item" style="color:$text_color">\n];
		print qq[
<input type="hidden" name="legend" value="$color" id="legend_$index">];

		print qq[
<input type="button" 
 id="button_${index}"
 class="color-button"
 style="background-color:$color"
 title="Select a color."
 name="button_$index"
 onClick="colorwheel(document.getElementById('legend_$index'), this)"
 tabindex="$No_Tabindex"
>];

		print qq[
<input type="hidden" name="text" value="$text_color" id="color_$index">];

		print qq[
<a 
 id="text_${index}"
 class="color-text"
 style="color:$text_color"
 title="Select a color."
 href="javascript:void 0"
 onClick="colorwheel(document.getElementById('color_$index'),this,'color')"
 tabindex="$No_Tabindex"
>];

		print $Cals->{$calendar_id}->title;
		
		print qq[</a>];

		print qq[</div>\n];
		print qq[</td>\n\n];

		#
		# Print center column of filter parameters 
		#

		print qq[<td class="legend-center">];
		print qq[<div class="filter-item">\n];

		if ($Cals->{$calendar_id}->filter_clause) {
			my $default = $Filters[$index];
			print $Cals->{$calendar_id}->filter_label ." ";
			if (my $list = $Cals->{$calendar_id}->filter_list) {
				my $attr = 'onChange="this.form.submit()"';
				print Acme::HTML::ListMenus->$list(
				  				'filter',
								$default,
								undef,
								$attr);
			}
			else {
				print qq[<input type="text"]
				  . qq[ name="filter" value="]
				  . escape_html($default)
				  . qq[">];
			}
		}
		else {
			# Need dummy placeholder
			print qq[<input type="hidden" name="filter" value="">];
		}

		print qq[</div>\n];
		print qq[</td>\n\n];

		#
		# Print right column
		#

		print qq[<td class="legend-right">];
		print qq[<div class="total-item">\n];

		if ($Cals->{$calendar_id}->total_field) {
			print qq[<span class="count">]
			  . $Counts[$index]
			  . qq[ items]
			  . qq[</span>]
			  . qq[ <nobr>]
			  . $Cals->{$calendar_id}->total_label
			  . qq[ ]
			  . $Totals[$index]
			  . qq[</nobr>];
		}

		# Track query parameters
		print qq[<input type="hidden"]
		  . qq[ name="param" value="]
		  . escape_html($Params[$index])
		  . qq[">];

		print qq[</div>\n];
		print qq[</td>\n\n];

		print qq[</tr>\n];

	} continue { $index++ }

	if ($index > 1 && substr($Superimpose,0,1) eq '#') {
		$color = $Superimpose;

		print qq[<tr>\n];

		# Center column

		print qq[<td colspan="3">];
		print qq[<div class="legend-item">\n];
		print qq[
<input type="hidden" name="superimpose" value="$color" id="legend_$index">];

		print qq[
<input type="button" 
 id="button_${index}"
 class="color-button"
 style="background-color:$color"
 title="Select a color."
 name="button_$index"
 onClick="colorwheel(document.getElementById('legend_$index'), this)"
 tabindex="$No_Tabindex"
>];

		print "Superimposed, dates with entries from two or more calendars.";
		print qq[</div>\n];
		print qq[</td>\n];

		print qq[</tr>\n];
	}

	print qq[</table>];
	print qq[</div>];

	#
	# Make hidden form fields for saving CGI parameters.
	#

	print Acme::HTML::Parameters::save_cgi_param_except(
							'legend',
							'superimpose',
							'text',
							'filter',
							'param',
							'resize',
							);

	print qq[\n<p>]
	  . qq[<input type="hidden" name="resize" value="no">]
	  . qq[<input type="submit" value="Redraw Calendar" class="button"></p>]
	  if ! $Printing;

	print qq[\n</form>];
}

#------------------------------------------------------------------------------

sub display_items {

	#
	# Database record
	#

	use vars qw(
		$y1 $m1 $d1 $y2 $m2 $d2
		$events__calendar_id
		$events__table_name
		$events__event_id
		$events__event_description
		$events__begin_date
		$events__begin_time
		$events__end_date
		$events__end_time
		$events__event_type
		$events__event_status
		$events__every_year
		$events__button_hilite
		$events__importance
		$events__total
		);

	#
	# Query database
	#

	$Sql = Acme::Security::clean_sql($Sql);

	my $q = Acme::Db->query($Sql);
	$q->bind_columns_to_variables;
	$q->set_nulls_to_empty;
	
	#
	# Initialize object for collecting menu items into appropriate days.
	#

	# Define re to recognize empty menus which won't be output.
	# Empty menus have no items nor any commands to add new items.
	my $empty_menu = qr|\A<div class="day-menu"[^>]+>\n\n</div>\Z|;

	my $cdm = new Acme::Calendar::DayMenu $Start_Date, $empty_menu;

	$cdm->set_menus_header(\&day_menu_header);
	$cdm->set_menus_footer(\&day_menu_footer);

	#
	# Construct menus for each day.
	#

	my $prev_date = $Start_Date;

	my ($menu_item, $sort_key);

	our ($begin_date, $end_date);
	our $count = 0;

	while ($q->next()) {
		$count++;

		# Compute begin and end dates of event. Force those for every
		# year into the year of the calendar.

		$m2 ||= $m1;
		$d2 ||= $d1;
		$y2 ||= $y1;

		$begin_date = "$m1$DS$d1$DS$y1";
		$end_date   = "$m2$DS$d2$DS$y2";

		# Attach blank menu items between previous date and this date.
		if ($prev_date ne $begin_date) {
			$cdm->fill_menus(
					$prev_date,
					$begin_date,
					'',
					0,
					'blank');
		}

		# Periodically flush menus.
		$cdm->print_menus_till($begin_date)
		  unless ($count % 30);
		
		($menu_item, $sort_key) = make_menu_item();
		# Attach menu item to the dates it applies to.
		$cdm->fill_menus(
				$begin_date,
				$end_date,
				$menu_item,
				$sort_key,
				\&codify);
		
		$prev_date = $begin_date;
	}

	if ($prev_date ne $End_Date) {
		$cdm->fill_menus(
				$prev_date,
				$End_Date,
				'',
				0,
				'blank');
	}

	$cdm->print_menus($End_Date);

	return $count;
}

#-------------------------------------------------------------------------------

sub codify {
	my $d = shift;

	# Produce a code (or list of codes) used to classify specified date.
	# This gets called for every record read from the database for the day.

	use vars qw($begin_date $end_date);

	my @codes;
	push @codes, 'start' if $d eq $begin_date;
	push @codes, 'stop ' if $d eq $end_date;
	push @codes, 'occupy';

	push @codes, $events__event_type
	  if $events__event_type;

	push @codes, {image=>$events__button_hilite}
	  if $events__button_hilite;

	# Assign background image.
	push @codes, {images=>{
			$events__button_hilite =>
			($events__importance || 0)}}
	  if $events__button_hilite;

	# Track number of calendars superimposed on this date.
	push @codes, {calendars=>{$events__calendar_id => 1}};

	return @codes;
}

#-------------------------------------------------------------------------------

sub classify {
	my $d = shift;

	# Produce a classification for the given date depending upon its codes.
	# This gets called once for every day rendered.

	# Up to 3 class codes can be returned: 1) a suffix naming a css style
	# for the date, a background color, and/or a background image.
	# Each code generates a call to hiliteDate() to hilite a date, so
	# setting all 3 codes for a date can be slow for 12 month calendars.

	my @classes;		# returned codes

	# Uncommment next if codes from codify() are only single values
	#my %codes = map {$_,1} @_;

	# Uncomment next to allow key/value codes to be set in codify().
	# Make hash of codes or if code is a hash, assign its key and value.
	my %codes;
	for my $c (@_) {
		if (ref $c eq 'HASH') {
			for my $k (keys %$c) { $codes{$k} = $c->{$k} };
		}
		else {
			$codes{$c} = 1;
		}
	}

	# Check codes that can change css styles.
	my $class_suffix = '';

	# Uncomment following to classify upon date range of events.
	#$class_suffix '-lightred'  if $codes{start} && $codes{stop};
	#$class_suffix '-lightred'  if $codes{start};
	#$class_suffix '-lightred'  if $codes{stop};

	push @classes, $class_suffix if $class_suffix;

	# Check codes that change the background color.
	my $color = '';
	if ($codes{calendars}) {

		# The first calendar in the list of calendars that has an entry
		# for this day becomes the 'primary' color of the day.

		my $index = 0;
		for my $calendar_id ($Cals->calendars) {
			if (defined $codes{calendars}->{$calendar_id}) {
				$color = $Legend_Colors[$index] || 'gray';
				last;
			}
			$index++;
		}

		my @cals = keys %{$codes{calendars}};

		if (scalar @cals > 1 && $Superimpose) {

			if (substr($Superimpose,0,1) eq '#') {
				$color = $Superimpose;	# specified color
			}
			else {
				my $n = ($#cals <= $#Legend_Colors)
					? $#cals
					: $#Legend_Colors;
				$color = darken($color) while $n--;
			}
		}
	}

	push @classes, $color if $color;

	# Check codes that change the background image.
	my $image = '';

	# Assign image according to importance ranking
	my $img_rank = -1;
	while ( my ($img,$rank) = each %{$codes{images}}) {
		if ($rank > $img_rank) {
			$image = $img;
			$img_rank = $rank;
		}
	}

	push @classes, $image if $image;

	# Note: @classes ($class_suffix, $color, and $image) can all be color
	# codes, css class suffixes, or image files (they don't work yet
	# in IE5.5) as Calendar.hiliteDate() JavaScript function accepts all
	# three. Their order is important as the following overwrite the
	# previous.

	return @classes;
}

#-------------------------------------------------------------------------------

sub day_menu_header {
	my $date = shift;

	# Substitute current date for $date in header menu items.
	(my $header = $Day_Menu_Header) =~ s/\${?date}?/$date/g;

	if ($Type eq 'full') {
		$header = <<EOF if $header;
<div class="day-menu-header" id="day_menu_$date" onClick="if(window.event)event.cancelBubble=true">
$header</div>
EOF
		# Return only the new items menu on cells that don't have items.
		return $header unless classify($date,@_);

		return <<EOF;
$header<div class="day-menu" id="day_$date" onClick="if(window.event)event.cancelBubble=true">
EOF
	}

	if ($Type eq 'list') {
		# Don't attach day menus on cells that don't have items.
		return '' unless classify($date,@_);

		my ($d) = $date =~ /\W(\d+)\W/;
		$d += 0;	# remove 0 padding
		my $ld = longdate($date);

		my $onclick = qq[return false];
		$onclick = qq[calCallback(this,'$date')]
		  if $Cals->n_new_links == 1;
		$onclick = qq[do_callback(this)]
		  if $Cals->n_new_links > 1;

		my $title = '';
		$title = $Cals->{ ($Cals->calendars)[0] }->new_link_text
		  if $Cals->n == 1;
		$title = qq[title="$title"] if $title;

		$header = <<EOF if $header;
<div class="day-menu-header" id="day_menu_$date" onClick="if(window.event)event.cancelBubble=true">
$header</div>
EOF
		return <<EOF;
$header<div class="day-menu" id="day_$date" onClick="if(window.event)event.cancelBubble=true">
<button class="week-day" id="cal_$date" $title
 onMouseOver="hilite_date(this)"
 onMouseOut="unhilite_date(this)"
 onClick="$onclick"
>$d</button> &nbsp; <span class="day-title">$ld</span>
EOF
	}

	$header = <<EOF if $header;
<div class="day-menu-header" onClick="if(window.event)event.cancelBubble=true">
$header</div>
<hr class="day-menu-divider">
EOF
	return <<EOF;
<div class="day-menu" id="day_$date" onClick="if(window.event)event.cancelBubble=true">
$header
EOF
}

#-------------------------------------------------------------------------------

sub day_menu_footer {
	my $date = shift;

	my @classify_codes = classify($date,@_);

	if ($Type eq 'full' || $Type eq 'list') {
		# Don't attach day menus on cells that don't have items.
		return '' unless @classify_codes;
	}

	my $footer = "</div>\n";

	return $footer unless @classify_codes;

	if ($Type eq 'full') {
		$footer .= <<EOF;
<script language="JavaScript" type="text/javascript">appendDate(document.getElementById('day_$date'),'$date')</script>
EOF
	}

	for (@classify_codes) {
		$footer .= <<EOF if $_;
<script language="JavaScript" type="text/javascript">hiliteDate('$_','$date')</script>
EOF
	}

	return $footer;
}

#-------------------------------------------------------------------------------

sub make_menu_item {

	my ($menu_item, $sort_key, $color);

	my $index = $Cals->{$events__calendar_id}->index - 1;	# array index

	# Remove extra seconds from time specification
	$events__begin_time = sprintf("%d:%0.2d",
	  ($events__begin_time =~ /(\d+):(\d+)/))
	  if $events__begin_time;
	$events__end_time = sprintf("%d:%0.2d",
	  ($events__end_time =~ /(\d+):(\d+)/))
	  if $events__end_time;

	$color = $Text_Colors[$index];

	# Construct menu item.
	$menu_item  = qq[<div class="day-menu-item" style="color:$color">];

	if (my $url = $Cals->{$events__calendar_id}->edit_link_url) {
		# Append id number to url if it ends with =
		$url .= qq[$events__event_id] if substr($url,-1) eq '=';
		my $wname = $Cals->{$events__calendar_id}->table_name
		  . "$events__event_id";
		$menu_item .= make_edit_bullet(
				$url,
				$Cals->{$events__calendar_id}->edit_link_text,
				$wname,
				);
	}
	else {
		$menu_item .= make_edit_bullet();
	}

	$menu_item .= qq[<a href="]
	  . $Cals->{$events__calendar_id}->view_link_url
	  . qq[$events__event_id"]
	  . qq[ onClick="new_target(this,'view')"]
	  . qq[ title="]
	  . $Cals->{$events__calendar_id}->view_link_text
	  . qq[">];

	$menu_item .= qq[$events__begin_time ]
	  if $events__begin_time;
	$menu_item .= qq[- $events__end_time &nbsp; ]
	  if $events__end_time;
	$menu_item .= qq[$events__event_type &nbsp; ]
	  if $events__event_type;

	if ($events__end_date && $events__end_date ne $events__begin_date) {
		$menu_item .= qq[ ] . short_date_range(
						$events__begin_date,
						$events__end_date);
	}

	$menu_item .= qq[</a> ];

	$menu_item .= escape_html($events__event_description);
	$menu_item .= qq[</div>];

	# Tally totals
	$Counts[$index]++;

	if ($Cals->{$events__calendar_id}->total_field) {
		if (defined $events__total) {
			$Totals[$index] += $events__total;
		}
	}

	# Construct sort key to use to sort entries for each day.
	$sort_key = sprintf("%05d%-16.16s%05.5s",
			0 - ($events__importance || 0),
			$events__event_type,
			$events__begin_time);

	return ($menu_item, $sort_key);
}

#------------------------------------------------------------------------------

sub make_edit_bullet {
	my ($href, $title, $wname) = @_;
	return qq[$Bullet&nbsp;] unless $href;
	my $target = $wname ? qq[ target="$wname"] : '';
	return qq[<a href="$href"]
	  . qq[ class="edit-bullet"]
	  . qq[ title="$title"]
	  . qq[ onClick="new_target(this,'form')"$target]
	  . qq[>$Bullet&nbsp;</a>];
}

#------------------------------------------------------------------------------

sub make_sql {

	#
	# Construct SQL for finding calendar events.
	#

	# Note: the algorithm for merging calendar dates that are to appear
	# on every year's calendar only works for calendars of 12 months or
	# less. The calendar can span across a year boundary, e.g. from
	# July to June.

	my $calendar_id = shift;

	my $table = $Cals->{$calendar_id}->table_name;

	my $t = $Schema->t($table);

	#
	# Set parameters to merge into sql.
	#

	my $cal = $Cals->{$calendar_id};

	my $event_id          = $cal->key_field;
	my $event_description = $cal->event_description_field	|| "''";
	my $begin_date        = $cal->begin_date_field		|| "cast(null as date)";
	my $begin_time        = $cal->begin_time_field		|| "cast(null as time)";
	my $end_date          = $cal->end_date_field		|| $begin_date;
	my $end_time          = $cal->end_time_field		|| "cast(null as time)";
	my $event_type        = $cal->event_type_field		|| "''";
	my $event_status      = $cal->event_status_field	|| "''";
	my $every_year        = $cal->every_year_field		|| "'N'";
	my $button_hilite     = $cal->button_hilite_field	|| "''";
	my $importance        = $cal->importance_field		|| '0';
	my $total             = $cal->total_field		|| '0';

	#
	# Cast quoted strings to be text.
	#

	# Note: only those strings that are used in comparisons or sorting need
	# to be cast to a database type for Postgres.

	my $qt = qr/^'.*'$/;		# matches quoted text string
	$event_type        = "cast($event_type as text)"	if $event_type =~ $qt;
	#$event_description = "cast($event_description as text)" if $event_description =~ $qt;
	#$event_status      = "cast($event_status as text)"	if $event_status =~ $qt;
	#$every_year        = "cast($every_year as text)"	if $every_year =~ $qt;
	#$button_hilite     = "cast($button_hilite as text)"	if $button_hilite =~ $qt;
	#$importance        = "cast($importance as text)"	if $importance =~ $qt;
	#$total             = "cast($total as text)"		if $total =~ $qt;

	#
	# Get SQL statement clauses
	#

	my $prepare_sql       = $cal->prepare_sql		|| '';
	$prepare_sql .= ';' if $prepare_sql && $prepare_sql !~ /;\s*$/;

	my $from_clause       = $cal->from_clause 
	  || die "No from clause specified in calendar reports.";
	my $where_clause      = $cal->where_clause  || '1 = 1';
	my $filter_clause     = $cal->filter_clause || '';
	if ($filter_clause) {
		my $filter = $Filters[$Cals->{$calendar_id}->index - 1]; # set arg
		if (defined $filter && "$filter" ne '') {
			# Substitute $filter and ${filter} in $filter_clause
			$filter_clause =~ s/\${?filter}?/$filter/g;

			$filter_clause = "and ( $filter_clause )"
		}
		else {
			$filter_clause = '';
		}
	}

	#
	# Define sql to select dates within calendar's Start_Date and End_Date.
	#

	my $year_sql = <<End_Of_SQL;
$prepare_sql
select
	extract (year  from $begin_date) as y1,
	extract (month from $begin_date) as m1,
	extract (day   from $begin_date) as d1,
	extract (year  from $end_date)   as y2,
	extract (month from $end_date)   as m2,
	extract (day   from $end_date)   as d2,
	'$calendar_id'                   as events__calendar_id,
	'$table'                         as events__table_name,
	$event_id                        as events__event_id,
	$event_description               as events__event_description,
	$begin_date                      as events__begin_date,
	$begin_time                      as events__begin_time,
	$end_date                        as events__end_date,
	$end_time                        as events__end_time,
	$event_type                      as events__event_type,
	$event_status                    as events__event_status,
	$every_year                      as events__every_year,
	$button_hilite                   as events__button_hilite,
	$importance                      as events__importance,
	$total			         as events__total
from $from_clause
where ( $where_clause )
  and ( @{[$t->list_where ]} )
  and ( ($begin_date between '$Start_Date' and '$End_Date'
      or $end_date between '$Start_Date' and '$End_Date'
      or ($begin_date < '$Start_Date' and $end_date > '$End_Date'))
  and ($every_year is null or $every_year != 'Y') )
  $filter_clause
End_Of_SQL

	return $year_sql unless $cal->every_year eq 'Y';

	#
	# Define sql to select dates to appear on every year's calendar.
	#

	my $every_year_sql;

	# Compute begin and end months.
	my ($m1,$d1,$y1) = split(/\W+/,$Start_Date);
	my ($m2,$d2,$y2) = split(/\W+/,$End_Date);
	my $m = $m2 + (($y2 - $y1) * 12);
	if ($m >= $m1) {
		# Calendar is all within one year (normal).
		$every_year_sql = <<End_Of_SQL;

union

-- Every year dates
select
	$y1 as y1,
	extract (month from $begin_date) as m1,
	extract (day   from $begin_date) as d1,
	$y2 as y2,
	extract (month from $end_date) as m2,
	extract (day   from $end_date) as d2,
	'$calendar_id'                   as events__calendar_id,
	'$table'                         as events__table_name,
	$event_id                        as events__event_id,
	$event_description               as events__event_description,
	$begin_date                      as events__begin_date,
	$begin_time                      as events__begin_time,
	$end_date                        as events__end_date,
	$end_time                        as events__end_time,
	$event_type                      as events__event_type,
	$event_status                    as events__event_status,
	$every_year                      as events__every_year,
	$button_hilite                   as events__button_hilite,
	$importance                      as events__importance,
	$total			         as events__total
from $from_clause
where ( $where_clause )
  and ( @{[$t->list_where ]} )
  and ( ( extract (month from $begin_date) between $m1 and $m2
       or extract (month from $end_date) between $m1 and $m2
        ) and $every_year = 'Y' )
  $filter_clause
End_Of_SQL

	}
	else {
		# Calendar crosses year boundary.
		
		# Redefine ending month.
		$m2 = $m;

		$every_year_sql = <<End_Of_SQL;

union

-- Every year dates between 1 and begin month
select
	$y2                              as y1,
	extract (month from $begin_date) as m1,
	extract (day   from $begin_date) as d1,
	$y2                              as y2,
	extract (month from $end_date)   as m2,
	extract (day   from $end_date)   as d2,
	'$calendar_id'                   as events__calendar_id,
	'$table'                         as events__table_name,
	$event_id                        as events__event_id,
	$event_description               as events__event_description,
	$begin_date                      as events__begin_date,
	$begin_time                      as events__begin_time,
	$end_date                        as events__end_date,
	$end_time                        as events__end_time,
	$event_type                      as events__event_type,
	$event_status                    as events__event_status,
	$every_year                      as events__every_year,
	$button_hilite                   as events__button_hilite,
	$importance                      as events__importance,
	$total			         as events__total
from $from_clause
where ( $where_clause )
  and ( @{[$t->list_where ]} )
  and ( ( extract (month from $begin_date) between 1 and $m2
       or extract (month from $end_date) between 1 and $m2
        ) and $every_year = 'Y' )
  $filter_clause

union

-- Every year dates between end month and 12
select
	$y1                              as y1,
	extract (month from $begin_date) as m1,
	extract (day   from $begin_date) as d1,
	$y1                              as y2,
	extract (month from $end_date)   as m2,
	extract (day   from $end_date)   as d2,
	'$calendar_id'                   as events__calendar_id,
	'$table'                         as events__table_name,
	$event_id                        as events__event_id,
	$event_description               as events__event_description,
	$begin_date                      as events__begin_date,
	$begin_time                      as events__begin_time,
	$end_date                        as events__end_date,
	$end_time                        as events__end_time,
	$event_type                      as events__event_type,
	$event_status                    as events__event_status,
	$every_year                      as events__every_year,
	$button_hilite                   as events__button_hilite,
	$importance                      as events__importance,
	$total                           as events__total
from $from_clause
where ( $where_clause )
  and ( @{[$t->list_where ]} )
  and ( ( extract (month from $begin_date) between $m1 and 12
       or extract (month from $end_date) between $m1 and 12
        ) and $every_year = 'Y' )
  $filter_clause
End_Of_SQL
	}

	#
	# Assemble complete sql query for both target years' dates and
	# for those dates which appear on every year.
	#

	return <<End_Of_SQL;
$year_sql
$every_year_sql
End_Of_SQL
}

#------------------------------------------------------------------------------

sub darken {
	my ($color) = @_;

	# Process is to subtract 0x33 from last, middle, then the first
	# color triplet.

	return $color unless $color =~ /#(\w\w)(\w\w)(\w\w)/;

	my $first  = hex $1;
	my $middle = hex $2;
	my $last   = hex $3;

	if ($last > 0) {
		$last -= hex '33';
		$last = $last > 0 ? $last : 0;
	}
	elsif ($middle > 0) {
		$middle -= hex '33';
		$middle = $middle > 0 ? $middle : 0;
	}
	elsif ($first > 0) {
		$first -= hex '33';
		$first = $first > 0 ? $first : 0;
	}

	return sprintf("#%02x%02x%02x",$first,$middle,$last);
}
__END__
