#!/usr/bin/perl -wT -I../lib
require 5.6.0;
use strict;

=head1 NAME

callbacks.cgi -- Report on user callbacks

=head1 SYNOPSIS

    callbacks.cgi?week_offset=0

    callbacks.cgi?bgndate=1/1/2002&enddate=1/31/2002&uid_sort=Y

    callbacks.cgi?bgndate=1/1/2002&enddate=1/31/2002&uid=mack&restrict_uid=Y

=head1 DESCRIPTION

User callbacks are log records that have a callback message scheduled for a
date. Log records are for customers, suppliers, or any table that has
a log table associated to it.

Setting the C<week_offset> parameter to 0 lists the callback messages scheduled
for the current week. The begin date and ending date can be specified with
the C<bgndate> and C<enddate> parameters.

=head1 CHANGE LOG

 5/98    Initial version.
 1/22/01 Converted to CGI and Acme::Lister

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright(c) 1998-2002  Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Acme::HTML;
use Acme::HTML::NavButtons;
use Acme::HTML::Parameters;
use Acme::Lister;
use Acme::Date			qw(today		dayadd		meddate);
use Acme::Config;
use Local::Acme::Banners	qw(common_header	common_footer);

use Local::Acme::Callbacks;	# SQL and record formatting for all tables

use CGI;

use Acme::Security;
Acme::Security::check('Access');

$|=1;		# flush output to synchronize error msgs

our $cgi = new CGI;
die "No arguments\n" unless $cgi->param();

print $cgi->header(-type=>'text/html') unless @ARGV;

our $Scripts	 = Acme::Config->Scripts;	# location of JavaScripts
our $Styles      = Acme::Config->Styles;	# location of CSS styles

# Arguments
our $title	  = $cgi->param('title')	|| '';
our $showbanner	  = $cgi->param('showbanner')	|| 'yes';

our $week_offset  = $cgi->param('week_offset');
our $bgndate	  = $cgi->param('bgndate');
our $enddate	  = $cgi->param('enddate');
our $restrict_uid = $cgi->param('restrict_uid');
our $uid_sort	  = $cgi->param('uid_sort');
our $uid	  = $cgi->param('uid');

$uid = $restrict_uid ? $uid : '.*';

unless (defined $bgndate || defined $week_offset) {

	my $today = today();

	Acme::HTML::Parameters->get_parameters(
				title	=> $title || 'Callbacks List',
				params	=> [
					"\$bgndate=?d?$today?Begin Date",
					"\$enddate=?d?$today?End Date",
					],
				);
	exit 0;		# the parameter form will re-call this routine
}

unless ($bgndate && $enddate) {

	# compute bgndate and enddate (from monday to sunday)
	$week_offset ||= 0;
	my $wday = (localtime)[6];	# day of week
	$bgndate = dayadd( (($week_offset * 7) - $wday + 1));
	$enddate = dayadd(6, $bgndate);
}

my $bd = meddate($bgndate);
my $ed = meddate($enddate);
$title ||= "Callbacks from $bd through $ed";

our $b = new Acme::HTML::NavButtons;

my $order_by_clause = "order by 7, 8, 2";
$order_by_clause    = "order by 10,9,2" if $uid_sort;

our $l = new Acme::Lister (
	title        => $title,		# list title
    	sql          => Local::Acme::Callbacks->sql($bgndate,$enddate,$uid) . $order_by_clause,
	sandbox      => "callbacks", # optional package where list produced
	);

print <<EOF;
<html>
EOF

if ($showbanner =~ /yes/i) {
	print qq(<head>\n);
	print qq(<title>$title</title>\n);
	print qq(</head>\n);
	print qq(<link rel="stylesheet" href="$Styles/common.css" type="text/css">\n);
	print qq(<link rel="stylesheet" href="$Styles/list.css" type="text/css">\n);
	print qq(<body onLoad="self.focus()">\n);
	print $b->buttons;
	print common_header('Callbacks');
}
else {
	print qq(<head>\n);
	print qq(<link rel="stylesheet" href="$Styles/common.css" type="text/css">\n);
	print qq(<link rel="stylesheet" href="$Styles/list.css" type="text/css">\n);
	print qq(<script src="$Scripts/common.js" type="text/javascript"></script>\n);
	print qq(</head>\n);
	print qq(<body>\n);
}

if ($uid_sort) {
	$l->print (		
		before_list  => \&callbacks::display_callback_list_header,
		after_list   => \&callbacks::display_callback_list_footer,
		before_group => [ user_id       => \&callbacks::display_user_id_header,
				  callback_date => 'print "<br>"' ],
		after_group  => [ user_id       => \&callbacks::display_user_id_footer ],
		every_row    => \&callbacks::display_callback,
	    );
}
else {
	$l->print (		
		before_list  => \&callbacks::display_callback_list_header,
		after_list   => \&callbacks::display_callback_list_footer,
		before_group => [ callback_date => 'print "<br>"' ],
		every_row    => \&callbacks::display_callback,
	    );
}

if ($showbanner =~ /yes/i) {
	print common_footer('Callbacks');
	print $b->last_buttons();
}
else {
	print qq(<hr>\n);
}

print qq(</body>\n);
print qq(</html>\n);

exit 0;

#------------------------------------------------------------------------------

package callbacks;


# Note: The display_callback() routine used to print every row is located in
# the Local::Acme::Callbacks module's file $Local/Acme/Callbacks.pm.

sub display_callback_list_header {
	print <<End_Of_HTML;
<h2>$title</h2>
<ol>
End_Of_HTML
}

sub display_callback_list_footer {
	print "\n</ol>\n";
}

sub display_user_id_header {

	no strict;
	no warnings qw(uninitialized once);

	print <<End_Of_HTML;
<h3>
$user_id callbacks from $bgndate to $enddate
</h3>
<ol>
End_Of_HTML
}

sub display_user_id_footer {
	print "</ol>\n";
}
__END__
