#!/usr/bin/perl -wT -I../lib
require 5.6.0;
use strict;

=head1 NAME

console.cgi -- Display user console (main menu)

=head1 SYNOPSIS

    console.cgi

=head1 DESCRIPTION

The console.cgi program displays the user's main menu. Different selections
appear on the console screen depending upon user's security privileges.

=head1 CHANGE LOG

 2/1/00   Initial version.
 8/12/00  Added security checks
 12/16/01 Search controls can be specified in .../etc/site.conf.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Acme::HTML;
use Acme::HTML::SearchControl;
use Acme::HTML::CallbacksControl;
use Acme::HTML::CalendarsControl;
use Acme::HTML::AdControl;
use Acme::DailyPics		qw(pic_of_day);
use Acme::Options;
use Local::Acme::Banners	qw(common_header	common_footer);
use Acme::Config;

use Acme::Security;
Acme::Security::check('Access');

our $Db_Title    = Acme::Config->DBTITLE;	# database (company) title
our $Styles      = Acme::Config->Styles;	# css stylesheets
our $Scripts_Lib = Acme::Config->Scripts_Lib;	# JavaScript library

$|=1;
print "Content-type: text/html\n\n";

our $opt = new Acme::Options;

if ($opt->RandomDailyPicture =~ /yes/i) {
	$opt->set('DailyPicture' => pic_of_day());
	$opt->save;
}

#
# Determine size of console window from number of search tables.
# The numbers are guesses. They depend upon css styles, screen resolution,
# the header height which is influenced by option button picture heights,
# and the number of lines displayed for the fortune text.
#

my $vsize = 515;
for (@{Acme::Config->Search_Tables}) {
	$vsize += 34 if /[^- ]+/;	# search control
	$vsize += 14 if /^-/;		# horizontal line
	$vsize += 14 if /^\s*$/;	# blank
}

#
# Begin Console Screen Output
#

# Keep IE6 in IE5 compatablilty mode by not specifying url
#<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
#"http://www.w3.org/TR/html4/loose.dtd">

print <<End_Of_HTML;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">

<html>
<head>

<title>console $Db_Title</title>

<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<link rel="stylesheet" href="$Styles/console.css" type="text/css">

<script src="$Scripts_Lib/control.js" type="text/javascript"></script>
<script src="$Scripts_Lib/form_validate.js" type="text/javascript"></script>
<script src="$Scripts_Lib/form_events.js" type="text/javascript"></script>

<script language="JavaScript" type="text/javascript">
var isConsoleLoaded = true;	// set flag
window.resizeTo(640,$vsize);	// resize window upon reload
</script>

</head>

<body onLoad="window.name='console'" onUnload="window.name=''">

End_Of_HTML

print common_header();

#
# Print standard controls
#

print qq[<br>\n];

if (Acme::Security::allow('Access','customers')) {
	print new Acme::HTML::SearchControl table => 'customers';
}

if (Acme::Security::allow('Access','callbacks')) {
	print new Acme::HTML::CallbacksControl;
}

if (Acme::Security::allow('Access','calendars')) {
	print new Acme::HTML::CalendarsControl;
}

#
# Print search controls for those tables named in configuration file.
#

my $prefix = "\n<br>";
for my $table (@{Acme::Config->Search_Tables}) {
	if ($table =~ /^\s*$/) {
		# Empty entry in list of tables.
		# Print blank space.
		$prefix .= "\n<br>" if $prefix !~ /\n<br>$/;
		next;
	}
	
	if ($table =~ /^-/) {
		# Print horozonal line
		print "$prefix\n<hr>";
		$prefix = '';
		next;
	}

	if (Acme::Security::allow('Access',"$table")) {
		print "$prefix\n" if $prefix;

		print new Acme::HTML::SearchControl table =>"$table";

		$prefix = '';
	}
}

#
# Print advertising control
#

if (Acme::Security::allow('Access','ad_sources')) {
	print "\n<br>";
	print new Acme::HTML::AdControl;
}

print qq[<br>\n];

#
# Print fortune
#

if ($opt->RandomDailyFortune =~ /yes/i) {
	print qq[\n<div class="fortune">\n], fortune(), qq[\n</div>\n];
}

#
# Construct copyright.
#

# Obtain version date year from the modification time of this file.
my $version = (localtime((stat($0))[9]))[5] + 1900;

my $about_link =
  qq[<a href="/help/about.htm" onClick="new_target(this,'help')" class="about" title="Show software copyrights and versions.">&copy; Mack Pexton, $version</a>];

#
# Construct a link to the administration screen if security allows.
#

my $admin_link = '';
$admin_link = 
  qq[<a href="/cgi-bin/admin.cgi?sort=yes" target="admin" onClick="new_target(this)" class="link" title="Open administration window.">More...</a>]
  if (Acme::Security::allow("Access",'admin.cgi'));

#
# Print copyright and administration screen links
#

print <<End_Of_HTML;
<table width="100%" style="margin-top:1ex"><tr>
<td align="left">$about_link</td>
<td align="right">$admin_link</td>
</tr></table>
End_Of_HTML

print common_footer();

print <<End_Of_HTML;
</body>
</html>
End_Of_HTML

exit 0;

#------------------------------------------------------------------------------

sub fortune {

	# Choose a random line from the file.
	# The algorithm is from the Perl Cookbook by
	# Tom Chrisiansen and Nathan Torkington.
	# Instead of reading every line into memory so that
	# the number of lines are known, each line is read
	# and random number is generated on the current line
	# number ($.) to determine if the line should be
	# saved as a candidate to print.

	my $fortunes = Acme::Config->Fortunes;

	open FORTUNE, $fortunes or return '';

	my $fortune;

	srand;
	rand($.) < 1 && ($fortune = $_) while <FORTUNE>;

	close FORTUNE;

	chop $fortune;
	return $fortune;
}
__END__
