#!/usr/bin/perl -wT -I../../lib
#!/usr/bin/perl -T -I../../lib
require 5.6.0;
###use strict;


=head1 NAME

importcustomers.cgi -- Import customer data

=head1 SYNOPSIS

    importcustomers.cgi?data=...
      or to display a form of parameters
    importcustomers.cgi

=head1 DESCRIPTION

Import customer names into the database. Records in columns with the 
first record containing the names of the columns are parsed into
data fields and are combined with form field values to build a
record to insert into the database. 

=head1 NOTE

This program is intended to be used with small lists, 10,000 rows or less.

=head1 CHANGE LOG

 10/5/02  Initial version. Assimilated from other custom import programs.
 11/12/02 Adapted to AcmeBase customers.

=head1 AUTHOR

Copyright 2002, Mackley F. Pexton.  All rights reserved.

Address bug reports and comments to: mackleyp@montana.net

=cut

package import;
use SelfLoader;

use Acme::HTML		qw(escape_html	display_message);
use Acme::Db;
use Acme::Date		qw(today);
use Local::Acme::User	qw(user);
use Local::Acme::Form::Banners  qw(common_header        common_footer
                                   common_form_header   common_form_footer);

use Acme::Config;

use Text::ParseWords;
use CGI;

#
# Define possible import types and their conversion routines.
#

our %Conversion = (
	'(standard)'		=> '',
	'Custom 1'		=> 'convert_lindblad()',
	'Custom 2'		=> 'convert_adventure_collection()',
	);

$|=1;		# flush output (synchronize errors with output)

$Debug = 0;	# set to 0 to turn off debug
#$Debug = 1;	# set to 1 to turn on debug
sub debug { print "<br>--- ", @_ if $Debug }

#
# Program Arguments
#

$cgi = new CGI;

$cgi_select_code	= $cgi->param('select_code');
$cgi_data 		= $cgi->param('data');
$cgi_file 		= $cgi->param('file');
$cgi_confirm 		= $cgi->param('confirm');
$cgi_import_type	= $cgi->param('import_type');

$cgi_delimiter		= $cgi->param('delimiter');
$cgi_exclude		= $cgi->param('exclude');

print $cgi->header(-type=>'text/html') unless @ARGV;

#
# Global definitions
#

no warnings qw(once);	# several globals used in AUTOLOAD functions

our $user	 = user();

our $CheckGroups = 3;				# number of groups in dupe check

our $Styles      = Acme::Config->Styles;	# location of CSS styles
our $Images      = Acme::Config->Images;	# location of image files
our $Scripts_Lib = Acme::Config->Scripts_Lib;	# location of JavaScripts

our $Delim	 = Acme::Config->Export_Separator	|| "\t";
our $Bullet	 = Acme::Config->Bullet			|| '&bull;';

our $No_Tabindex = Acme::Config->No_Tabindex		|| '-1';


our $today  = today();

our $Import_Dir  = Acme::Config->Import_Space;

our $Log_Dir     = "$Import_Dir/log";
our $SQL_Dir     = "$Import_Dir/sql";
our $Result_Dir  = "$Import_Dir/result";

#
# Get import data if needed.
#

if (! defined($cgi_data) && ! defined($cgi_file)) {
	get_parameters();
	exit 0;		# the parameter form will re-call this routine
}

#
# Make working director and files.
#

unless ( -d $Log_Dir )    { mkdir $Log_Dir;    chmod 0777, $Log_Dir;    }
unless ( -d $SQL_Dir )    { mkdir $SQL_Dir;    chmod 0777, $SQL_Dir;    }
unless ( -d $Result_Dir ) { mkdir $Result_Dir; chmod 0777, $Result_Dir; }

#
# Make file name pattern.
#

my ($import_type_name) = $cgi_import_type =~ /([^\$\@`'"|;]+)/;	# untaint
$import_type_name =~ s/ /_/g;	# change spaces to underlines for file names
$import_type_name =~ s/\(standard\)//;		# nothing for standard imports

#
# Create files.
#

our $logfile    = mkfile($Log_Dir,"$user-$import_type_name",'.dat');
our $sqlfile    = mkfile($SQL_Dir,"$user-$import_type_name",'.sql');
our $resultfile = mkfile($Result_Dir,"$user-$import_type_name",'.txt');

#
# Initialize list checkmark processing.
#

our %excluded_ids;	# row numbers excluded from import

our $n_rows = 0;	# number of displayed checked rows

use warnings qw(once);	# turn warnings back on

#
# Define temporary table name
#

# Note: This program uses a temporary table to store imported records for
# dupe checking. The table is defined with a SERIAL column to keep track
# of import row numbers. Using a temporary table allows multiple users
# to import at the same time, however Postgres uses Sequences to implement
# SERIAL fields, and sequence names are not private to the user. As a 
# workaround, we use a random number for the temporary table's name.

our $temp_table = "t_" . sprintf("%-0d", int(rand 30000) + 1);

#
# Upload data file if specified.
#

# Note: The entire import file is stored in memory. Primarily because it
# is transmitted back to the user for confirmation if all the rows don't 
# pass, and file uploads cannot be "saved" and redone on next pass.
# This procedure is best used for small import files -- 10,000 records or less.

if (! $cgi_data && $cgi_file) {
	
	#
	# Upload data file
	#

        my $bufsize = 5 * 1024;         # 5K buffer size
        my $buffer = '';

        # Read file into memory.
        while (read($cgi_file,$buffer,$bufsize)) {
		$$cgi_data .= $buffer;
        }
}

die "No import records found." if ! $cgi_data;

#
# Write log file of import data.
#

open(LOG, '>', $logfile) or die("Cannot save to log file '$logfile'. $!");
print LOG $cgi_data;
close LOG;

if (! $cgi_confirm) {
	get_confirmation();
	exit 0;		# the parameter form will re-call this routine
}

if ($cgi_confirm ne 'yes') {
	display_message(
	  "Data was <font color=red>not</font> imported into the database.",
	  "Import aborted");
	exit 0;		# the parameter form will re-call this routine
}

#
# Construct SQL insert statements for each record.
#

%excluded_ids = map { $_, 1 } split /,/, $cgi_exclude if $cgi_exclude;

make_sqlfile();

#
# Import the data.
#

if ($Debug) { print "<pre>",`cat "$sqlfile"`,"</pre>\n"; exit 0; }

do_import();

display_message("Import completed. Look for errors.","Import Completed");

exit 0;

#------------------------------------------------------------------------------

sub do_import {

	$ENV{PATH} = '/bin:/usr/bin';                   # untaint
	delete @ENV{ qw(IFS CDPATH ENV BASH_ENV) };     # untaint

	my $dbname   = Acme::Config->DBNAME;
	my $username = Acme::Config->DBUSER;
	$username  &&= "--username $username";
	my $password = Acme::Config->DBPASS;
	$password  &&= "--password $password";

	system(<<EOF) == 0 or die "Cannot execute import SQL: $?";
psql $dbname $username $password < $sqlfile 2>&1 | tee $resultfile
EOF
}

#------------------------------------------------------------------------------

sub make_sqlfile {
	my ($table, $prefix, $suffix) = @_;

	# This routine is used for checking for dupes in addition to 
	# importing the data. The presence of the $table argument is
	# used as a flag, if it is not set then we are posting to the
	# database table.

	my $is_dupe_checking = $table ? 1 : 0;

	#
	# Open the data file for processing.
	#

	open(LOG, '<', $logfile) or die("Cannot open log file. $!");

	#
	# Define fields (columns) to insert into database.
	#

	my @fields = (
		'customer_name',
		'first_name',
		'last_name',
		'address1',
		'address2',
		'city',
		'state',
		'zip',
		'country',
		'address_usage',
		'email',
		'phone',
		'address_usage',
		'ship_address1',
		'ship_address2',
		'ship_city',
		'ship_state',
		'ship_zip',
		'ship_country',
		'ship_address_usage',
		'customer_type',
		'customer_status',
		'customer_status_date',
		'original_ad_source',
		'ad_source',
		'ad_source_date',
		'salesrep',
		'referral',
		'comments',
		'mailing_lists',
		'interests',
		'select_code',
		'entry_date',
		'last_update',
		'last_update_by',
		);

	#
	# First record is a header consisting of a list of the field names.
	# Skip initial blank lines.
	#

	my ($rec, @a, $n);		# record text and its parsed fields

	while (! $rec) {
		$rec = <LOG>;
		$rec =~ s/\r?\n$//;	# trim \r\n
	}

	DELIM: {
		if (! $cgi_delimiter || $cgi_delimiter eq 'auto') {
			
			#
			# Determine the delimiter character by inspecting header
			#

			$Delim  = $rec =~ /\t/	? "\t"	# tab
				: $rec =~ /\|/	? '\|'	# pipe
				: $rec =~ /,/	? ','	# comma
				: $rec =~ /;/	? ';'	# semi-colon
				: $rec =~ /:/	? ':'	# colon
				: ' ';			# white-space
			last DELIM;
		}

		if ($cgi_delimiter eq 'tab')	{ $Delim = "\t"; last DELIM; }
		if ($cgi_delimiter eq 'pipe')	{ $Delim = '\|'; last DELIM; }
		if ($cgi_delimiter eq 'comma')	{ $Delim = ','; last DELIM; }
		if ($cgi_delimiter eq 'semi-colon') { $Delim = ';'; last DELIM; }
		if ($cgi_delimiter eq 'colon')	{ $Delim = ':'; last DELIM; }
	}

	bind_fields_to_array(\@a,parse($rec));
	debug("bind_fields_to_array: ".(join ", ", parse($rec)))
	  if $Debug;

	open(SQL, '>', $sqlfile)
	  or die("Cannot save to SQL log file '$sqlfile'. $!");

	print SQL $prefix if $prefix;

	#
	# Initialize global record counts.
	#

	$count = 0;	
	$skipcount = 0;	

	my $insertindex = 0;

	while ($rec = <LOG>) {
		$rec =~ s/\r?\n$//;	# trim \r\n
		$count ++;

		next if $excluded_ids{$count};

		#
		# Initialize record variables
		#

		initialize_import_fields();

		#
		# Parse the record.
		#

		# Stuff the elements into array to preserve bindings

		$n = 0; for ( parse($rec,(scalar @a)) ) { $a[$n++] = $_ }

		#
		# Convert the import fields to the customer table fields
		#

		if ($cgi_import_type && $Conversion{$cgi_import_type}) {

			if (! eval $Conversion{$cgi_import_type}) {

				#
				# Conversion failed, skip record
				#

				$skipcount++;
				next;
			}
		}

		#
		# Apply standard conversions and set default values.
		#

		apply_standard_conversions();

		#
		# Quote the SQL quote characters.
		#

		for (@a) { s/'/''/g; }

		#
		# Make list of columns and corresponding values for insert
		#

		no strict 'refs';
		###no warnings 'uninitialized';

		my $insert_columns = '';
		my $insert_values  = '';

		my ($f,$v);
		for $f (@fields) {

			#
			# Set empty values to NULL by excluding the
			# field from the insert clause.
			#

			if ($v = eval "\$$f") {
				$v = Acme::Db->escape_sql($v);

				$insert_columns .= qq[,\n$f];
				$insert_values  .= qq[,\n'$v'];
			}

			debug("field f='$f', v='$v'") if $Debug;
		}
		$insert_columns =~ s/^,\n//;	# remove initial comma
		$insert_values  =~ s/^,\n//;	# remove initial comma

		#
		# Save SQL statement
		#

		$table ||= 'customers';
		$insertindex++;

		print SQL <<EOF;

-- Insert $insertindex: Row $count
insert into $table (
$insert_columns

) values (

$insert_values
);
EOF
	}

	close LOG;

	if (! $is_dupe_checking) {
		
		#
		# Add entries to ad_response table for each customer imported.
		#

		my $select_clause = $cgi_select_code
				  ? qq[and customers.select_code = '$cgi_select_code']
				  : qq[and customers.select_code is null];

# Original insert script that gives wrong results.
#insert into ad_response (ad_source, response_date, customer_id)
#select customers.ad_source, customers.entry_date, customers.customer_id
#from customers
#where customers.entry_date = '$today'
#  and customers.last_update_by = '$user'
#  $select_clause
#  and customers.ad_source = '$ad_source'
#  and customers.customer_id not in (
#	select t_ar.customer_id from t_ar
#);

###		print SQL <<EOF;
###
###select customer_id
###from ad_response
###where response_date = '$today'
###  and ad_source = '$ad_source'
###into temp t_ar;
###
###select customer_id
###from customers
###where customers.entry_date = '$today'
###  and customers.last_update_by = '$user'
###  $select_clause
###  and customers.ad_source = '$ad_source'
###into temp t_arc;
###
###delete from t_arc where customer_id in ( select customer_id from t_ar );
###
###insert into ad_response (ad_source, response_date, customer_id)
###select customers.ad_source, customers.entry_date, customers.customer_id
###from customers
###where customers.customer_id in ( select t_arc.customer_id from t_arc )
###;
###EOF
	}

	if ($table ne 'customers' && Acme::Config->DBENGINE =~ /Postgres/i) {

		# Drop sequence Postgres uses for the temporary table serial
		# column which does not automatically go away for temporary
		# tables. It is important that the temporary table sequence
		# number start at 1 everytime because it is equated to
		# row numbers.

		print SQL qq[\ndrop sequence ${temp_table}_customer_id_seq;];
	}

	print SQL $suffix if $suffix;

	close SQL;

	return $count;
}

#------------------------------------------------------------------------------

sub bind_fields_to_array {
	my $aref = shift;
        ($pkg) = caller;

        #
        # Build hash of field name / column number
        #

	my $n = 0;
	my %fields = map { $_, $n++ } @_;

        #
        # Tie each field name variable to it's appropriate
        # element in the record columns array rec.
        #

        my ($k, $v);
        for $k (keys %fields) {
                $v = $fields{$k};     # get array index of field $k

		#
                # Make sure column name is a valid Perl variable name.
		#

                # Perl variables begin with an alpha [A-Za-z] or underline [_]
                # and the rest of the characters must be an alpha [A-Za-z],
                # digit [0-9], or underline [_]. Prepend an underline
                # to the column name if first character is not an alpha, and
                # substitute underlines for every non-valid variable
                # name character in the column name.

		# Change non-characters to single underlines.
		$k =~ s/\W/_/g;

                $k = "_$k" unless $k =~ /^[A-Za-z]/;

                tie ${"${pkg}::$k"}, import::var, \$$aref[$v];
        }
}

#------------------------------------------------------------------------------

sub parse {
	my ($rec,$limit) = @_;

	return unless $rec;

	if ($Delim eq ',') {

		#
		# Parse a "comma separated values" (CSV) record
		# Remove quoted quotes ("")
		#

		return map { s/""/"/g } &quotewords($Delim, 0, $rec);
	}
	else {
		#
		# Simple parse on a delimiter character
		#

		# Note: Perl's split function has a bug (feature) that strips
		# off trailing null fields unless you specify the number of 
		# fields to split using $limit. If a field is not set to null,
		# it retains its previous value, which is something I have
		# never wanted.

		return split /$Delim/, $rec, $limit;
	}
}

#------------------------------------------------------------------------------

sub mkfile {			# make unique file name
	my ($dir,$pattern,$type) = @_;

	#
	# Return unique file name
	#

	$type = ".tmp" unless $type;
	my $unique = 0;
	my $f;

	my ($d,$m,$y) = (localtime)[3,4,5];
	my $mypattern = sprintf("%02d%02d%02d",$y-100, ++$m, $d);

	# iterate until a file name is found which does not exist
	while ( $f = join "",
	  ($dir, "/", $mypattern, '_', $pattern, $unique, $type)
	  and (-e $f or -s $f) ) {
		$unique++;
	}

	return $f;
}

#-----------------------------------------------------------------------------
#
# Database bound variable routines
#
#------------------------------------------------------------------------------

package import::var;

sub TIESCALAR {
        my ($class, $ref) = @_;
	bless $ref, $class;
}

sub FETCH {
        my ($self) = @_;
        return $$self;
}

sub STORE {
        my ($self,$val) = @_;
        return $$self = $val;
}

sub DESTROY {}          # gobble


package import;

__DATA__

#######################################
# Dynamically loaded import functions #
#######################################

#------------------------------------------------------------------------------

sub get_parameters {

	my $title = "Import Customers";
	my $page_title = "import Customers";

	my $import_options = '';
	for (sort keys %Conversion) {
		$import_options .=
		  qq[<option value="].escape_html($_).qq[">$_</option>\n];
	}

	print <<EOF;
<html>
<head>
<title>$page_title</title>

<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<link rel="stylesheet" href="$Styles/form.css" type="text/css">
<style type="text/css">#data { font-size: 8pt; }</style>

<script src=$Scripts_Lib/form_validate.js></script>
<script src=$Scripts_Lib/form_events.js></script>

</head>
<body onLoad=self.document.forms[0].elements[2].focus()>
EOF
	print common_header($title);

	print <<EOF;

<form	name="list_control"
	method="POST"
	action="$ENV{SERVER_URL}$ENV{SCRIPT_NAME}"
>

EOF
	print common_form_header($title, {submit=>'Check Dupes'});

	print <<EOF;
<p>
This form imports names to the customer database.
</p>

<div align="left" class="instructions">
<!-- Start Hiding -->
<script language="JavaScript" type="text/javascript">
function show_instruction_body() {
	var c;
	c = document.getElementById('instruction_body_more_button_section');
	c.style.display = "none";
	c = document.getElementById('instruction_body');
	c.style.display = "block";	
}
function hide_instruction_body(goto_begin) {
	var c;
	c = document.getElementById('instruction_body');
	c.style.display = "none";       
	c = document.getElementById('instruction_body_more_button_section');
	c.style.display = "block";

	if (goto_begin) {
		// Scroll to make beginning button visible at bottom of page.
		c.scrollIntoView(false);
		window.scrollBy(0,50);
	}
}
</script>

<div id="instruction_body_more_button_section" style="display:block">

<!-- Begin More Button -->
<span class="button-bar"><table cellspacing="0" cellpadding="0" border="0"><tr>
 <td><a href="javascript:show_instruction_body()" title="Show instructions." tabindex="-1">Instructions <font class="button-chevron">&raquo;</font></a></td>
</tr></table></span>
<!-- End More Button -->
</div>

<div id="instruction_body" style="display:none">

<!-- Begin Less Button -->
<span class="button-bar"><table cellspacing="0" cellpadding="0" border="0"><tr>
 <td><a href="javascript:hide_instruction_body()" title="Hide instructions." tabindex="-1"><font class="button-chevron">&laquo;</font> Instructions</a></td>
</tr></table></span>
<!-- End Less Button -->

<div style="border:solid black 1px;padding:2ex">
<p style="margin-top:0">
Steps to import data are:
<ol>
<li>Two ways to enter data:
 <ul>
 <li>With Excel, highlight and Copy (CTRL-C) the names to import.</li>
 <li>Paste the records into the Data field below (CTRL-V). Make sure the first line is the names of the columns.</li>
 </ul>
or
 <ul>
 <li>Specify a text file in the "File" field.</li>
 </ul>
</li>
<li>Click "Check Dupes" button to display imported names that match those
already in our database.</li>
<li>Edit the resulting list. Uncheck the
 <img src="$Images/check.gif" border="0">
 check mark in front of the duplicate records to prevent them from being imported.</li>
<li>Fill in the form at the bottom of the list with values to be set on
every imported record.</li>
<li>Click "Import" button to actually import the names.</li>
</ol>
</p>

<p>For a standard type of import file, and most other types of imports, the
first line of the file is a header record consisting of column names to use
to refer to the data columns in the following lines. Standard type import
files have a pre-defined set of column names that correspond to the
database column names.
</p>
</div>
</div>
</div>

<div style="margin: 2ex 1ex 1ex 2ex;">
<table cellspacing=0 width=95%>

<tr>
<th align=right nowrap>Import Type</th>
<td><select name="import_type">
$import_options
</select></td>
</tr>

<tr>
<th align=right nowrap>Delimiter</th>
<td nowrap>
<input type="radio" name="delimiter" value="auto" checked> Auto &nbsp;
<input type="radio" name="delimiter" value="tab"> Tab &nbsp;
<input type="radio" name="delimiter" value="pipe"> Pipe (|) &nbsp;
<input type="radio" name="delimiter" value="comma"> Comma (,) &nbsp;
<input type="radio" name="delimiter" value="semi-colon"> Semi-Colon (;)
</td>
</tr>

<tr><td>&nbsp;</td><td>&nbsp;</td></tr>  <!-- blank line -->

<tr>
<th align=right valign=top>Data</th>
<td>
<textarea name="data" id="data" rows=20 style="width:100%"></textarea>
</td>
</tr>

<tr>
<td></td>
<td>You can import a text file instead of pasting the data above.</td>
</tr>

<tr>
<th align="right" valign="top">File</th>
<td>
<input type="file" name="file" value="" maxlength="256" style="width:100%">
</td>
</tr>

</table>
<input type="hidden" name="confirm" value="">
</div>

<hr>

<center>
<input type=submit value="Check Dupes" class=button>
</center>
</form>

</body>
</html>
EOF
}

#------------------------------------------------------------------------------

sub get_confirmation {

	my $data = escape_html($cgi_data);

	my $title = "Check Customers Import";
	my $page_title = "import Customers";

	print <<EOF;
<html>
<head>
<title>$page_title</title>

<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<link rel="stylesheet" href="$Styles/list.css" type="text/css">$aux_stylesheet
<style type="text/css">
INPUT.button{ font-family:system; background-color:buttonface; border-color:buttonface; }
</style>

<script src="$Scripts_Lib/form_validate.js" type="text/javascript"></script>
<script src="$Scripts_Lib/form_events.js" type="text/javascript"></script>
<script src="$Scripts_Lib/list.js" type="text/javascript"></script>

<script language="JavaScript">
// Override do_dblclick() in list.js. Go to edit form, not a view screen.
function do_dblclick(eid,url) {
        // Edit record

        // This should be invoked only by TD elements of list table.
        // Let the A tags do their own stuff without triggering a click.
        if (window.event &&
            window.event.srcElement &&
            isNodeName(window.event.srcElement,"A")) return;

        //do_click(eid);          // undo click() event that was also fired
	open_form(url);
}

function display_summary() {
        // Display number of checked and unchecked

        var f = document.list_control;
        if (! f) return;
        f.n_checked.value = nchecked;
        f.n_unchecked.value = nunchecked;
	f.n_import.value = nrows - nunchecked;
}

// Overide list.js/flip(). Flip all checkmarks in all list groups.
function flip(e) {

        // Extract row's key value from my Id attribute.
        var keyval = e.id.replace(/^l(\\d)_/,"");
        var group = RegExp.\$1;         // groups of records displayed

        var f = document.import_control;

        if (e.className == "checked") {
                e.className = "unchecked";

                nchecked--;
                nunchecked++;

                exclude_key(keyval);

                // Uncheck all checks in all groups of records
                var oe, i;
                for (i = 1; i <= $CheckGroups; i++) {
                        if (oe = document.getElementById('l'+i+'_'+keyval)) {
                                oe.className = "unchecked";
                        }
                }
        } else if (e.className == "unchecked") {
                e.className = "checked";

                nchecked++;
                nunchecked--;

                include_key(keyval);

                // Check all checks in all groups of records
                var oe, i;
                for (i = 1; i <= $CheckGroups; i++) {
                        if (oe = document.getElementById('l'+i+'_'+keyval)) {
                                oe.className = "checked";
                        }
                }
        }

        display_summary();
}

// Prepare list for submittal.
function prepare_control_form(request) {

        // Fill in hidden form fields before submitting.

        var f = document.list_control;

        f.n_checked.value     = nchecked;
        f.n_unchecked.value   = nunchecked;
	f.n_import.value      = nrows - nunchecked;

        var i;

        if (nexcluded > 0) {
                // find first
                for (i=0; i < excluded_rows.length; i++)
                        if (excluded_rows[i]) {
                                f.exclude.value = excluded_rows[i];
                                break;
                        }
                for (i=++i; i < excluded_rows.length; i++)
                        if (excluded_rows[i])
                                f.exclude.value += "," + excluded_rows[i];
        }
}
</script>
</head>
<body>
EOF
	print common_header($title);

	print <<EOF;

<form	name="list_control"
	method="POST"
	action="$ENV{SERVER_URL}$ENV{SCRIPT_NAME}"
	onSubmit="prepare_control_form()"
>
<input type="hidden" name="select_all" value="">
<input type="hidden" name="invert_all" value="">
EOF

	print common_form_header($title, {submit=>'Import'});

	print <<EOF;
<p>
This page displays records in the database that are suspected of being the
same as those being imported.
Several checks are performed so this process can take several minutes if there are many import records.
</p>

<p>
Checks performed are:
<ul>
<li>Last names are compared to those located close by (those having the same first three digits of their zip code).</li>
<li>Phone numbers are compared to both the home phone and business phone.</li>
<li>Email addresses are compared.</li>
<li>Duplicate names within the import list are looked for.</li>
<li>Records with missing fields are listed.</li>
</ul>
</p>

<input type="hidden" name="confirm" value="yes">
<input type="hidden" name="import_type" value="$cgi_import_type">
<textarea name="data" wrap="virtual" style="display:none">$data</textarea>
<textarea name="exclude" wrap="virtual" style="display:none">$cgi_exclude</textarea>
EOF
	#
	# List duplicates
	#

	my $start_time = time();

	# Define global record counts tracked in list_duplicate_records().
	our ($count, $skipcount);

	list_duplicate_records();

	#
	# Compute elapsed time
	#

	my $s = time() - $start_time;
	my $m = int($s / 60); $s %= 60;
	my $h = int($s / 60); $s %= 60;

	my $elapsed_time = $h ? "$h hrs " : '';
	$elapsed_time .= $m || $h ? "$m min " : '';
	$elapsed_time .= "$s sec";

	display_list_summary();

	print <<EOF;
<hr>
$count import records processed in $elapsed_time.
<br>&nbsp;
<br>
EOF

	display_common_form_fields();

	print <<EOF;
<center>
<input type=submit value=" Import " class="button">
</center>
</form>

<script language="JavaScript">

// Assign globals
nrows        = $count;

nchecked     = $n_rows;
nunchecked   = 0;

exclude_keys($cgi_exclude);
</script>

</body>
</html>
EOF
}

#------------------------------------------------------------------------------

sub list_duplicate_records {
	
	#
	# Execute several queries testing for duplicate records
	#
	
	my $createtempsql = <<EOF;
create temp table $temp_table (
	customer_id serial,
	customer_name varchar(70),

	first_name varchar(34),
	last_name varchar(34),
	address1 varchar(34),
	address2 varchar(34),
	city varchar(24),
	state varchar(2),
	zip varchar(10),
	province varchar(24),
	postal_code varchar(12),
	country varchar(34),
	email varchar(70),
	phone varchar(24),

	address_usage varchar(5),

	ship_address1 varchar(34),
	ship_address2 varchar(34),
	ship_city varchar(24),
	ship_state varchar(2),
	ship_zip varchar(10),
	ship_province varchar(24),
	ship_postal_code  varchar(12),
	ship_country varchar(34),

	customer_type varchar(6),
	customer_status varchar(8),
	customer_status_date date,

	original_ad_source varchar(8),
	ad_source varchar(8),
	ad_source_date date,

	salesrep varchar(14),
	referral varchar(34),
	comments varchar(150),

	mailing_lists varchar(127),
	interests varchar(30),

	select_code char(1),

	entry_date date,
	last_update date,
	last_update_by varchar(14)
);
EOF

#############
#	# Create temporary table
#
#	my $script = '';
#	my $comma= ',';
#	my $id;
#
#	if ($id = $t->id) {
#		# automatically define ID  key field
#		$script .= "$comma\n  $id serial";
#	}
#	for ($t->fields_all) {
#		next if $_ eq $id;	# already defined above
#		$script .= "$comma\n  $_ ".$t->f($_)->type;
#	}
#	if ($t->audit) {
#		$script .= "$comma\n  ".$t->audit_entry .' '.$t->audit_entry_type;
#		$script .= "$comma\n  ".$t->audit_update.' '.$t->audit_update_type;
#		$script .= "$comma\n  ".$t->audit_user  .' '.$t->audit_user_type;
#	}
#
#	my $createtempsql =
#	  qq[\ncreate temp table $temp_table (]
#	. substr($script,1)		# remove beginning comma
#	. qq[\n);\n];
#
##############

	my $checkfields = <<EOF;
customers.customer_id,				-- 2
customers.customer_name,			-- 3

customers.first_name,				-- 4
customers.last_name,				-- 5

customers.address1,				-- 6
customers.address2,				-- 7
customers.city,					-- 8
customers.state,				-- 9
customers.zip,					-- 10
customers.country,				-- 11

customers.email,				-- 12
customers.phone,				-- 13

$temp_table.customer_id	as t_customer_id,	-- 14
$temp_table.customer_name as t_name,		-- 15

$temp_table.first_name	as t_first_name,	-- 16
$temp_table.last_name	as t_last_name,		-- 17

$temp_table.address1	as t_address1,		-- 18
$temp_table.address2	as t_address2,		-- 19
$temp_table.city	as t_city,		-- 20
$temp_table.state	as t_state,		-- 21
$temp_table.zip		as t_zip,		-- 22
$temp_table.country	as t_country,		-- 23

$temp_table.email	as t_email,		-- 24
$temp_table.phone	as t_phone,		-- 25

customers.customer_type,			-- 26
customers.customer_status			-- 27
EOF

	#
	# Duplicate check sql
	#

	# Note: There are three "groups" of checks in the sql below.
	# The first, and primary, is a union of all the fields that
	# are checked for duplicates in the database. The second lists
	# duplicate records in the input file, and the third lists out
	# records with missing fields.

	# In the list, each group has check marks in front of the rows
	# which can be unchecked by the user to exclude them from the import.

	# If adding another group to the sql, change the global $CheckGroups
	# to match the number of groups.

	my $dupechecksql = <<EOF;
select
	1 as group_order,
$checkfields
from $temp_table, customers
where customers.last_name = $temp_table.last_name
  and $temp_table.first_name is not null
  and (customers.first_name = $temp_table.first_name
    or substring(customers.first_name from 1 for 2) ~ '[Mm][Rr]'
      )
union
select
	1 as group_order,
$checkfields
from $temp_table, customers
where $temp_table.last_name = customers.last_name
  and $temp_table.zip is not null 
  and substring($temp_table.zip from 1 for 3) = substring(customers.zip from 1 for 3)
EOF

	$dupechecksql .= <<EOF;
union
select
	1 as group_order,
$checkfields
from $temp_table, customers
where $temp_table.phone is not null
  and $temp_table.phone != ''
  and $temp_table.phone != '999-999-9999'
  and $temp_table.phone = customers.phone
union
select
	1 as group_order,
$checkfields
from $temp_table, customers
where $temp_table.phone is not null
  and $temp_table.phone != ''
  and $temp_table.phone != '999-999-9999'
  and $temp_table.phone = customers.company_phone
EOF

	$dupechecksql .= <<EOF;
union
select
	1 as group_order,
$checkfields
from $temp_table, customers
where $temp_table.email is not null
  and $temp_table.email != ''
  and $temp_table.email = customers.email
EOF

	$dupechecksql .= <<EOF;
union all
select distinct
	2 as group_order,
$checkfields
from $temp_table, $temp_table as customers
where $temp_table.last_name = customers.last_name
  and $temp_table.customer_id != customers.customer_id
  and substring($temp_table.zip from 1 for 3) = substring(customers.zip from 1 for 3)
EOF

	$dupechecksql .= <<EOF;
union all
select distinct
	3 as group_order,
$checkfields
from $temp_table, $temp_table as customers
EOF

# Order by:
# group_order, temp_table.customer_id desc, temp_table.last_name, temp_table.first_name,
# customers.last_name, customers.first_name

# Note: the temp_table.customer_id is used as a row count because the temp table is
# created each time so it's id starts from 1 each time.

	$dupechecksql .= <<EOF;
order by 1, 14, 17, 16, 5, 4
;
EOF
	
	#
	# Execute queries and read their results
	#

	my $c = make_sqlfile($temp_table,$createtempsql,$dupechecksql);

	#
	# Execute the query
	#

	my $sql;
	open(SQL, '<', $sqlfile) or die("Cannot open SQL file '$sqlfile'. $!");
	{
		local $/;	# set to read entire file
		$sql =  <SQL>;
	}
	close(SQL);

	# Trick the server to keep this possibly very long process alive
	# by trickling characters to the browser periodically.

	keep_alive();

	my $q = Acme::Db->query($sql);
	##$q->bind_columns_to_variables;

	alarm 0;	# turn off keep_alive() alarm

	#
	# Print duplicates
	#

	my $recno;		# row number of import record
	my $ri = 0;		# record index (number of dupe)
	my $prev_id = 0;
	my $prev_group_order = 0;

	if ($q->count == 0) {
		print "<h3>No Duplicate Records Found</h3>\n";
		$n_rows = 0;
		return $count;
	}

	print qq[<div class="list">\n];

	while ( (
$group_order,

$customer_id,
$customer_name,

$first_name,
$last_name,

$address1,
$address2,
$city,
$state,
$zip,
$country,

$email,
$phone,

$t_customer_id,
$t_name,

$t_first_name,
$t_last_name,

$t_address1,
$t_address2,
$t_city,
$t_state,
$t_zip,
$t_country,

$t_email,
$t_phone,

$customers__customer_type,
$customers__customer_status
	) = $q->next() ) {

		$dupe_title = 'Customers with Same Last Name, Phone, or Email'
		  if $group_order eq '1';
		$dupe_title = 'Duplicate Records in Import File'
		  if $group_order eq '2';
		$dupe_title = 'Records with Missing Fields'
		  if $group_order eq '3';

		$recno = $t_customer_id + 1;
		if ($group_order != $prev_group_order) {

			#
			# Print group header
			#

			if ($prev_group_order > 0) {

				#
				# Terminate previous group
				#

				print qq[</table>\n];
			}

			print qq[<h3>$dupe_title</h3>\n];
			print qq[<table class="list">\n];

			$prev_group_order = $group_order;
			$ri = 0;
		}

		if ($t_customer_id != $prev_id) {
			#
			# Print customer information
			#

			$ri++;		# record index, number of dupes

			if ($group_order == 3) {

				#
				# Flag missing parts
				#

				$t_last_name    ||= qq[<font color="red">lastname</font>];
				$t_address2 ||= qq[<font color="red">address</font>]
				  if ! $t_address1;
				$t_phone    ||= qq[<font color="red">phone</font>];
				$t_email    ||= qq[<font color="red">email</font>];
			}

			$t_zip ||= qq[<font color="red">zip</font>]
			  if ! $postal_code;

			print qq[\n<tr class="list"]
			    . qq[ onMouseOver="hi(this)"]
			    . qq[ onMouseOut="lo(this)"];
			print qq[ onClick="do_click('l${group_order}_${t_customer_id}')"]
			  if $t_customer_id;
			print qq[>\n];

			# Print record index.
			print qq[<td class="list-item-index">$ri</td>\n];

			print qq[<td class="list-item-selector">];
			print
			    qq[<a]
			  . qq[ href="javascript:void 0"]
			  . qq[ id="l${group_order}_${t_customer_id}"]
			  . qq[ class="checked"]
			  . qq[ onClick="flip(this)"]
			  . qq[></a>];
			print qq[</td>\n];

			print qq[<td colspan="2">];
			print <<EOF;
$t_last_name, $t_first_name (Row $recno) $t_address1 $t_address2, $t_city, $t_state $t_zip, $t_phone $t_email
EOF
			print qq[</td>\n];
			print qq[</tr>\n];

			$prev_id = $t_customer_id;
		}

		#
		# Skip matching records if printing out records with missing
		# fields.
		#

		next if $group_order == 3;

		#
		# Print matching import records
		#

		# Bold the parts that match
		$last_name = qq[<b>$last_name</b>]	if $last_name eq $t_last_name;
		$first_name = qq[<b>$first_name</b>]	if $first_name eq $t_first_name;
		$address1 = qq[<b>$address1</b>]	if $address1 eq $t_address1;
		$address1 = qq[<b>$address1</b>]	if $address1 eq $t_address2;
		$address2 = qq[<b>$address2</b>]	if $address2 eq $t_address2;
		$address2 = qq[<b>$address2</b>]	if $address2 eq $t_address1;
		$city = qq[<b>$city</b>] 		if $city eq $t_city;
		$state = qq[<b>$state</b>] 		if $state eq $t_state;
		$zip = qq[<b>$zip</b>] 			if $zip eq $t_zip;
		$phone = qq[<b>$phone</b>] 		if $phone eq $t_phone;
		$email = qq[<b>$email</b>] 		if $email eq $t_email;

		#
		# Flag missing parameters with red
		#

		$zip ||= qq[<font color="red">zip</font>];
		$phone ||= qq[<font color="red">phone</font>];
		$email ||= qq[<font color="red">email</font>];

		#
		# Print row
		#

		my $edit_url = qq[/cgi-bin/form.cgi?table=customers&customer_id=$customer_id];

		print qq[\n<tr]
		    . qq[ onMouseOver="hi(this)"]
		    . qq[ onMouseOut="lo(this)"];
		print qq[ onDblClick="do_dblclick('$customer_id','$edit_url')"]
		  if $t_customer_id && $edit_url;
		print qq[>\n];

		print qq[<td></td>];	# record index
		print qq[<td></td>];	# record selector

		print <<EOF if $group_order != 2;
<td class="list-item-buttons"><a href="$edit_url" target="_blank" title="Edit customer">$Bullet</a></td>
<td><a href="/cgi-bin/customers/view.cgi?customer_id=$customer_id" target="_blank" title="View customer">$last_name, $first_name</a> ($customer_id, $customers__customer_status $customers__customer_type) $address1 $address2, $city, $state $zip, <nobr>$phone</nobr> $email</td>
EOF
		my $recno = $customer_id + 1;	# 1st rec is header, so add 1

		print <<EOF if $group_order == 2;
<td></td>
<td>$last_name, $first_name (Row $recno) $address1 $address2, $city, $state $zip, <nobr>$phone</nobr> $email</td>
EOF
		print qq[</tr>\n];

	}
	print qq[</table>\n];
	print qq[</div>\n];

	#
	# Adjust global number of rows displayed (those with dupes).
	#

	$n_rows = $ri if $ri > $n_rows;

	#
	# Return the total number of rows processed.
	#

	return $count;
}

#------------------------------------------------------------------------------

sub display_list_summary {

        #
        # Print summary
        #

	my $check_msg = '';
	my $n_checked = $n_rows;
	my $n_unchecked = 0;

	my $n_passed = $count - $skipcount - $n_checked;# rows not displayed
	my $n_import = $n_passed + $n_checked;		# rows to import

        print <<End_Of_HTML;

<h3>Summary</h3>

<div class="summary">
<table cellspacing="0" cellpadding="0" border="0">
<tr>
<td>
 <table cellspacing="0" cellpadding="0" border="0">
 <tr>
 <th nowrap>Total Rows</th>
 <td><input type="text" class="number-readonly" readOnly name="count" value="$count"></td>
 <td class="description">(number of records in import data)</td>
 </tr>
 <tr>
 <th nowrap>Skipped Rows</th>
 <td><input type="text" class="number-readonly" readOnly name="skipcount" value="$skipcount"></td>
 <td class="description">(number of records that did not convert and were skipped)</td>
 </tr>
 <tr>
 <th>Passed</th>
 <td><input type="text" class="number-readonly" readOnly name="n_passed" value="$n_passed"></td>
 <td class="description">(number of records which passed all tests and are not displayed above)</td>
 </tr>
 <tr>
 <th>Checked</th>
 <td><input type="text" class="number-readonly" readOnly name="n_checked" value="$n_checked"></td>
 <td class="description">(number of records displayed above that are checked for import)</td>
 </tr>
 <tr>
 <th>Unchecked</th>
 <td><input type="text" class="number-readonly" readOnly name="n_unchecked" value="$n_unchecked"
></td>
 <td class="description">(number of records displayed above that are not checked and won't be imported)</td>
 </tr>
 <tr>
 <th nowrap>Total Import</th>
 <td><input type="text" class="number-readonly" readOnly name="n_import" value="$n_import"></td>
 <td class="description">(number of records that will be imported)</td>
 </tr>
 </table>
 </td>
 <td valign="top">
 $check_msg
 </td>
 </tr>
 </table>
</div>
End_Of_HTML
}

#------------------------------------------------------------------------------

sub display_common_form_fields {

	#
	# Form entry for remaining fields.
	#

	####
	use Local::Acme::ListMenus;
	my $address_usage_options   = Local::Acme::ListMenus->address_usage_codes('address_usage');
	my $ad_source_options       = Local::Acme::ListMenus->ad_sources('ad_source');
	my $customer_type_options   = Local::Acme::ListMenus->customer_types('customer_type');
	my $customer_status_options = Local::Acme::ListMenus->cust_status_codes('customer_status');
	my $salesrep_options        = Local::Acme::ListMenus->salesreps('','salesreps');
	####

	print <<EOF;
<table>
<tr><td colspan=2>
<!-- Begin form for extra fields -->

<div align="left" class="more-fields">
<!-- Start Hiding -->
<script language="JavaScript" type="text/javascript">
function show_morefields_body() {
	var c;
	c = document.getElementById('morefields_body_more_button_section');
	c.style.display = "none";
	c = document.getElementById('morefields_body');
	c.style.display = "block";	
}
function hide_morefields_body(goto_begin) {
	var c;
	c = document.getElementById('morefields_body');
	c.style.display = "none";       
	c = document.getElementById('morefields_body_more_button_section');
	c.style.display = "block";

	if (goto_begin) {
		// Scroll to make beginning button visible at bottom of page.
		c.scrollIntoView(false);
		window.scrollBy(0,50);
	}
}
</script>

<div id="morefields_body_more_button_section" style="display:block">

<!-- Begin More Button -->
<span class="button-bar"><table cellspacing="0" cellpadding="0" border="0"><tr>
 <td><a href="javascript:show_morefields_body()" title="Show More Fields." tabindex="-1">More Fields<font class="button-chevron">&raquo;</font></a></td>
</tr></table></span>
<!-- End More Button -->
</div>

<div id="morefields_body" style="display:none">

<!-- Begin Less Button -->
<span class="button-bar"><table cellspacing="0" cellpadding="0" border="0"><tr>
 <td><a href="javascript:hide_morefields_body()" title="Hide More Fields." tabindex="-1"><font class="button-chevron">&laquo;</font>Less Fields</a></td>
</tr></table></span>
<!-- End Less Button -->

<div style="border:solid black 1px;padding:2ex">
<p style="margin-top:0">
The values entered below are assigned to every import record. They will
overwrite any setting specified in the import file.
</p>
<table cellspacing="0" border="0" width="90%">

<tr>
<th align=right nowrap>Select Code</th>
<td>
<input type="text" name="select_code" size="1" maxsize="1" style="text-align:center">
</td>
</tr>

<tr>
<th align=right valign=top>Address Type</th>
<td>
$address_usage_options
</td>
</tr>

<tr>
<th valign=top align=right nowrap>Current Source</th>
<td>
$ad_source_options
</td>
</tr>

<tr>
<th valign=top align=right nowrap>Ad Reference</th>
<td>
<input type=text name="ad_source_ref" size=10 maxsize=10 value="">
</td>
</tr>

<tr>
<th valign=top align=right>Category</th>
<td>
$customer_type
</td>
</tr>

<tr>
<th valign=top align=right>Status</th>
<td>
$customer_status_options
</td>
</tr>

<tr>
<th valign=top align=right>Salesrep</th>
<td>
$salesrep_options
</td>
</tr>

<tr>
<th align=right valign=top>Comments</th>
<td>
<textarea name="comments" cols=50 rows=3></textarea>
</td>
</tr>

</table>
</div>
</div>
</div>

<!-- End form for extra fields -->
</td></tr>
</table>
EOF
}

#-----------------------------------------------------------------------------

sub keep_alive {

	# Trick the server to keep this possibly very long process alive
	# by trickling characters to the browser periodically.

	my $TrickleSec = 5;		# repeat interval in seconds
	my $handler = sub {
		print ".\n";
		alarm $TrickleSec;
	};

	$SIG{ALRM} = $handler;
	alarm $TrickleSec;		# set timer
}

################################################################################
#
# Standard Import Conversions  Routine
#

#------------------------------------------------------------------------------

sub initialize_import_fields {

	$first_name		= '';
	$last_name		= '';
	$address1		= '';
	$address2		= '';
	$city			= '';
	$state			= '';
	$zip			= '';
	$province		= '';
	$postal_code		= '';
	$country		= '';
	$email			= '';
	$phone			= '';

	$ship_address1		= '';
	$ship_address2		= '';
	$ship_city		= '';
	$ship_state		= '';
	$ship_zip		= '';
	$ship_province		= '';
	$ship_postal_code	= '';
	$ship_country		= '';

	$address_usage		= '';

	$customer_type		= '';
	$customer_status	= '';
	$customer_status_date	= '';
	$original_ad_source	= '';
	$ad_source		= '';
	$ad_source_date		= '';

	$salesrep		= '';
	$referral		= '';
	$comments		= '';

	$mailing_lists		= '';
	$interests		= '';

	$select_code		= '';

	$entry_date		= '';
	$last_update		= '';
	$last_update_by		= '';
}

#------------------------------------------------------------------------------

sub apply_standard_conversions {

	#
	# These conversions are always performed for every import record.
	# If there is a custom conversion routine registered in %Conversions,
	# then that routine is executed first before this one. That allows
	# the custom routines to simply assign their variables to the
	# standard variables, which are then applied to standard conversions.
	# For example, addresses in all capital letters is converted to
	# title case as a standard conversion.
	#
	# CGI variables are searched for missing import values.
	#

	#
	# Capitalize names if needed
	#

	if ($customer_name && ! $first_name && ! $last_name) {
		($first_name, $last_name) = $customer_name =~ /(.*?) (.*)/;
	}
	if (! $customer_name && $last_name) {
		$customer_name = $first_name ? "$first_name $last_name" : $last_name;
	}

	$first_name =~ s/\b(\w+)\b/ucfirst lc $1/eg
	  if $first_name =~ /^[a-z]/
	  || $first_name =~ /^[A-Z][A-Z]/;
	$last_name =~ s/\b(\w+)\b/ucfirst lc $1/eg
	  if $last_name =~ /^[a-z]/
	  || $last_name =~ /^[A-Z][A-Z]/;
	$address1 =~ s/\b(\w+)\b/ucfirst lc $1/eg
	  if $address1 !~ /^[A-Z]/
	  || $address1 !~ /[a-z]/;
	# change address back NE,NW,SE,SW
	$address1 =~ s/( [NS][ew]$)/uc $1/eg;
	$address2 =~ s/\b(\w+)\b/ucfirst lc $1/eg
	  if $address2 !~ /^[A-Z]/
	  || $address2 !~ /[a-z]/;
	# change address back NE,NW,SE,SW
	$address2 =~ s/( [NS][ew]$)/uc $1/eg;
	$city =~ s/\b(\w+)\b/ucfirst lc $1/eg
	  if $city =~ /^\s*[a-z]/
	  || $city =~ /^\s*[A-Z][A-Z]/;
	$state =~ s/\b(\w+)\b/uc $1/eg
	  if $state =~ /^[a-z]/;
	$zip =~ s/^(\d{5})(\d{4})$/$1-$2/;	# put dash in 9 digit zip
	$province =~ s/\b(\w+)\b/ucfirst lc $1/eg
	  if $province =~ /^[a-z]/
	  || $province =~ /^[A-Z][A-Z]/;
	$country =~ s/\b(\w+)\b/ucfirst lc $1/eg
	  if $country =~ /^[a-z]/
	  || $country =~ /^[A-Z][A-Z]/;

	$email =~ s/\b(\w+)\b/lc($1)/eg
	  if $email =~ /^[A-Z][A-Z]/;
	
	$email = '' if $email eq 'none';

	#
	# Convert other fields' format.
	#

	$country = '' if $country =~ /united states/i || $country =~ /^us$/i;

	$phone =~ s/(\d{3})(\d{3})(\d{4})$/$1-$2-$3/;	# put - in phone numbers
	$phone = '' if ! $phone =~ /\d/;		# must have digit to be a phone number

	#
	# Convert import codes to database codes
	#

	# All these fields need to be defined for import.
	# Set default field values.

	# The following fields are personal and need to be in
	# the import file.

	$first_name		||= '';
	$last_name		||= '';
	$address1		||= '';
	$address2		||= '';
	$city			||= '';
	$state			||= '';
	$zip			||= '';
	$province		||= '';
	$postal_code		||= '';
	$country		||= '';
	$email			||= '';
	$phone			||= '';

	# If only one address line, make it address2.
	if ($address1 && ! $address2) {
		$address2 = $address1;
		$address1 = '';
	}

	$ship_address1		||= $address1	|| '';
	$ship_address2		||= $address2	|| '';
	$ship_city		||= $city	|| '';
	$ship_state		||= $state	|| '';
	$ship_zip		||= $zip	|| '';
	$ship_province		||= $province	|| '';
	$ship_postal_code	||= $postal_code || '';
	$ship_country		||= $country	|| '';

	# The following fields can be specified in the form.
	$address_usage		||= $cgi->param('')		|| '';

	$customer_type		||= $cgi->param('customer_type')	|| 'prosp';
	$customer_status	||= $cgi->param('customer_status')	|| 'warm';
	$customer_status_date	||= $cgi->param('customer_status_date')	|| $today;
	$original_ad_source	||= $cgi->param('original_ad_source')	|| '';
	$ad_source		||= $cgi->param('ad_source')		|| 'unk';
	$ad_source_date		||= $cgi->param('ad_source_date') 	|| $today;

	$salesrep		||= $cgi->param('salesrep')	|| 'sales';
	$referral		||= $cgi->param('referral')	|| '';
	$comments		||= $cgi->param('comments')	|| '';

	$select_code		||= $cgi->param('select_code')	|| '';

	$entry_date		||= $cgi->param('entry_date')	|| $today;
	$last_update		||= $cgi->param('last_update')	|| $today;
	$last_update_by		||= $cgi->param('last_update_by') || $user;
}

################################################################################
#
# Special/Custom Import Conversions  Routines
#
# These routines must be registered by entering them into the %Conversion hash
# defined above.
#

#------------------------------------------------------------------------------

sub convert_lindblad {

	#
	# Set field defaults
	#

	$customer_type		= 'prosp';
	$customer_status		= 'warm';
	$salesrep	= 'aliasonb';
	$referral	= 'Lindblad BRC';

	$original_ad_source = 'RPlincat';
	$address_usage = '';

	$mailing_lists  = 'news';
	$interests = '';

	$comments = "Single (requested only OBP catalog)"
	  if $Type =~ /single/i;
	$comments .= ' ' if $comments && $Geography;
	$comments = $Geography;
	$comments .= ' ' if $comments && $Type_of_Travel;
	$comments = $Type_of_Travel;

	#
	# Assign record variables to "standard" names
	#

	$first_name = $First_Name;
	$last_name = $Last_Name;
	$address1  = '';
	$address2  = $Address_1;
	$city      = $City;
	$state     = $State;
	$zip       = $Zip_Code;
	$country   = '';

	$email     = $E_mail_Address;

	return 1;	# successful conversion
}

#------------------------------------------------------------------------------

sub convert_adventure_collection {

	# Skip record if next trip is > 2 years or never
	return 0 if $nexttrip > 3;

	# Skip record if agegroup is 1 (<30) or 2 (31-40).
	#return 0 if $agegroup < 3;
	return 0 if $agegroup < 2;

	#
	# Set field defaults
	#

	$customer_type		= 'prosp';
	$customer_status		= 'warm';
	$salesrep	= 'sales';
	$referral	= 'AC $10,000 Contest';

	$mailing_lists  = 'news';

	$first_name = $firstname;
	$last_name = $lastname;

	$address2 = $address;
	$state = $stateprov;
	$zip   = $postalcode;

	#
	# Convert their fields into our customer fields
	#

	$phone = "$phone_areacode-$phone_prefix-$phone_suffix";

	#
	# Convert their codes to ours
	#

	AD_SOURCE: {
		# No specified source
		if (! $ad_source) { $ad_source = 'Aacx0902';		last AD_SOURCE; }

		if ($ad_source eq 'nyt')   { $ad_source = 'Aacn0902';	last AD_SOURCE; }
		if ($ad_source eq 'nytbrc'){ $ad_source = 'Aacn0902';	last AD_SOURCE; }
		if ($ad_source eq 'smith') { $ad_source = 'Aacs0802';	last AD_SOURCE; }
		if ($ad_source eq 'tl')    { $ad_source = 'Aact0902';	last AD_SOURCE; }
		if ($ad_source eq 'tlbrc') { $ad_source = 'Aact0902';	last AD_SOURCE; }
		if ($ad_source eq 'cnt')   { $ad_source = 'Aacc0902';	last AD_SOURCE; }
		if ($ad_source eq 'cntbrc'){ $ad_source = 'Aacc0902';	last AD_SOURCE; }

		# Default
		$ad_source = 'Aacx0902';
	}

	ADDRESS_USAGE: {
		if ($noemail){
			$address_usage = 'Lemail';
			$mailing_lists = '';
			last ADDRESS_USAGE;
		}

		# Default
		$address_usage = '';
	}

	$interests = '';
	INTERESTS: {
		if ($specint_corporate)	{ $interests .= ',co' }
		if ($specint_famadvent)	{ $interests .= ',ft' }
		if ($specint_famreunion){ $interests .= ',fr' }
		if ($activity_biking)	{ $interests .= ',cy' }
		if ($activity_boating
		or  $activity_rafting)	{ $interests .= ',ri' }
		if ($activity_culture
		or  $activity_nathistory){ $interests .= ',hy' }
		if ($activity_helihiking
		or  $activity_walking)	{ $interests .= ',hi' }
		if ($activity_heliskiing){ $interests .= ',ds' }
		if ($activity_photo)	{ $interests .= ',ph' }
		if ($activity_seakayak)	{ $interests .= ',sk' }
		if ($activity_touring)	{ $interests .= ',si' }
		if ($activity_whale
		or  $activity_wildlife)	{ $interests .= ',wv' }
	}
	$interests =~ s/^,// if $interests;	# remove beginning comma

	$seasons = '';
	SEASONS: {
		if ($dest_alaska)	{ $seasons .= ',ak' }
		if ($dest_canada)	{ $seasons .= ',cr' }
		if ($specint_custom)	{ $seasons .= ',cu' }
		if ($dest_southam)	{ $seasons .= ',sa' }
		if ($dest_westus)	{ $seasons .= ',nr,sw' }
	}
	$seasons =~ s/^,// if $seasons;		# remove beginning comma

	AGEGROUP: {
		if ($agegroup eq '1'){
			$comments = "AC Contest ID# $id; agegroup: $agegroup (30 or under)";
			last AGEGROUP;
		}

		if ($agegroup eq '2'){
			$comments = "AC Contest ID# $id; agegroup: $agegroup (31-40)";
			last AGEGROUP;
		}

		if ($agegroup eq '3'){
			$comments = "AC Contest ID# $id; agegroup: $agegroup (41-50)";
			last AGEGROUP;
		}

		if ($agegroup eq '4'){
			$comments = "AC Contest ID# $id; agegroup: $agegroup (51-60)";
			last AGEGROUP;
		}

		if ($agegroup eq '5'){
			$comments = "AC Contest ID# $id; agegroup: $agegroup (61-70)";
			last AGEGROUP;
		}

		if ($agegroup eq '6'){
			$comments = "AC Contest ID# $id; agegroup: $agegroup (70+)";
			last AGEGROUP;
		}

		# Default
		$comments = "AC Contest ID# $id";
	}

	return 1;	# successful conversion
}
