#!/usr/bin/perl -wT -I../../../lib

# Customer Log Post routine

# Post information from a form to the database. A message is displayed and
# and OK button is displayed which when clicked closes the window the message
# is displayed in.

use Acme::HTML		qw(display_message);
use Acme::Db;
use Acme::Parse		qw(parse_args);
use Acme::Date		qw(today timestamp);
use Local::Acme::User	qw(user);

use Acme::Security;
Acme::Security::check('Access');

local $user = user();
local $today = today();
local $timestamp = timestamp();

local %args;
if (parse_args(\%args)) {
	local $log_id = $args{log_id};
	delete $args{log_id};	# key field, do not update

	if (defined $args{R_log_id}) {
		# Response record is being posted
		local $R_log_id = $args{R_log_id};
		delete $args{R_log_id};	# key field, do not update

		# remove non-database column names from %args
		delete $args{keep_open};	# flag field, do not update
		delete $args{R_current_user};	# phantom field, do not update

		# Construct record audit info
		$args{R_last_update} = $timestamp;	# system assigned values
		$args{R_last_update_by} = $user;	# system assigned values

		if ($R_log_id and $R_log_id > 0) {
			post_change($R_log_id);
		}
		elsif (length $args{R_log} == 0) {
			for (keys %args) { delete $args{$_} if /^R_/ }
			# No response message, simply acknowledge original log
			post_change($log_id);
		}
		else {
			$args{R_entry_date} = $today;	# system assigned values
			post_new();
		}
	}
	else {
		# Original log record being posted
		# remove non-database column names from %args
		delete $args{keep_open};	# flag field, do not update
		delete $args{current_user};	# phantom field, do not update

		# Construct record audit info
		$args{last_update} = $timestamp;	# system assigned values
		$args{last_update_by} = $user;		# system assigned values

		if ($log_id and $log_id > 0) {
			post_change($log_id);
		}
		else {
			$args{entry_date} = $today;	# system assigned values
			post_new();
		}
	}
}
exit 0;

#------------------------------------------------------------------------------

sub post_change {
	my $log_id = shift;		# log record could also be response

	my ($key, $value, @kv);
	while (($key,$value) = each %args) {

		# response log record field names have R_ prepended to them to 
		# differentiate them from the original log record fields
		$key =~ s/^R_//;	# remove field name R_ prefixes

		if (defined $value && $value ne '') {
			$value = Acme::Db->escape_sql($value);
			push @kv, "$key = '$value'";
		}
		else {
			push @kv, "$key = null";
		}
	}

	my $post_script = join '',
	  "update customer_log set \n",
	  (join ",\n", @kv), 
	  "\nwhere log_id = $log_id;";

	$|=1;
	print "Content-type: text/html","\n\n";

	if (Acme::Db->execute($post_script)) {
		display_message("The customer log record has been changed.",
				"Message",
				"post_cleanup()");
	}
	else {
		# It is assumed that an error message has already been output.
		display_message("The customer log record has <font color='red'>NOT</font> been changed.",
				"Message",'',0);
	}
}

#------------------------------------------------------------------------------

sub post_new {
	foreach $a (keys %args) {
		$args{$a} = Acme::Db->escape_sql($args{$a});
	}

	# Construct SQL script
	# the only arguments in %args should be database columns
	my $c = join ",\n", (keys %args);	# column names
	my $v = join "',\n'", (values %args);	# surround values with quotes

	$v =~ s/'',$/null,/gm;		# change blank entries to null

	# response log record field names have R_ prepended to them to 
	# differentiate them from the original log record fields
	$c =~ s/\bR_//gm;		# remove field name R_ prefixes

	my $post_script = "insert into customer_log (\n$c\n) values (\n'$v'\n);";

	$|=1;
	print "Content-type: text/html","\n\n";

	my $rc = Acme::Db->execute($post_script);

	if ($rc and defined $args{R_response_to_log_id}) {
		# This posting is for a response log entry
		# Post to the original log entry the response date.

		my ($id) = Acme::Db->get(<<EOS);
select max(log_id) as id from customer_log
where response_to_log_id = $args{R_response_to_log_id}
;
EOS

		$rc = Acme::Db->execute(<<EOS);
update customer_log set 
  respond_by = '$user',
  respond_date = current_date,
  respond_log_id = $id
where log_id = $args{R_response_to_log_id}
EOS
	}

	if ($rc and $args{ad_source} and $args{customer_id}) {
		# Post ad source to ad_response log
		# and update customer record.

		my $update_script = <<EOS;
insert into ad_response (
	response_date,
	ad_source,
	source_table,
	respondent_id
) values (
	'$today',
	'$args{ad_source}',
	'customers',
	$args{customer_id}
);

update customers set
	ad_source = '$args{ad_source}',
	ad_source_date = '$today'
where customer_id = $args{customer_id};
EOS
		Acme::Db->execute($update_script);

		# Note: if an error occurs above, an error message is usually
		# output on the screen. If successful, we are silent.
	}

	if ($rc and $args{customer_status} and $args{customer_id}) {
		# Update customer record.
		my $update_script = <<EOS;
update customers set
	customer_status = '$args{customer_status}',
	customer_status_date = '$today'
where customer_id = $args{customer_id};
EOS
		Acme::Db->execute($update_script);

		# Note: if an error occurs above, an error message is usually
		# output on the screen. If successful, we are silent.
	}

	if ($rc) {
		display_message("The customer log record has been added.",
				"Message",
				"post_cleanup()");
	}
	else {
		# It is assumed that an error message has already been output.
		display_message("The customer log record has <font color='red'>NOT</font> been added.",
				"Message",'',0);
	}
}
