#!/usr/bin/perl -wT -I../../lib

=head1 NAME

customers/view/order.pl -- Display orders

=head1 DESCRIPTION

These routines support the customers/view.cgi program.
They list orders on the customer view screen.

=back

=head1 CHANGE LOG

 8/13/02  Rewrite

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000-2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

package customer;

require 5.6.0;
use strict;

use Acme::HTTP		qw(escape);
use Acme::Base::Schema;
use Acme::Lister;
use Acme::SQL;
use Acme::HTML::SectionControl;
use Local::Acme::Format;
use Acme::Config;

use Acme::Security;

our $Schema = new Acme::Base::Schema;
our @headers;

#------------------------------------------------------------------------------

use warnings qw(uninitialized once);

sub display_orders {

	my ($keyval, $display_info) = @_;

	#
	# Set up parameters from globals for listing order.
	#

	my $detail_table = 'orders';
	my $detail_field = 'customer_id';

	my $table = 'customers';
	my $field = 'customer_id';	# link field, usually the key field

	#
	# Define field list and formats for list
	#

	my $order_list = <<End_Of_List;
End_Of_List

	my $order_header = <<End_Of_Format;
End_Of_Format

	my $order_format = <<End_Of_Format;
End_Of_Format

	my $order_sql = <<End_Of_SQL;
End_Of_SQL

	#
	# Construct sql for list of detail records.
	#

	my $t  = $Schema->t($table);
	my $f  = $t->f($field);
	my $dt = $Schema->t($detail_table);

	#
	# Use table defaults for lists, formats, and sql.
	#

	$order_list   = $dt->list_fields;
	$order_header = $dt->list_header;
	$order_format = $dt->list_format;
	$order_sql    = $dt->list_sql;

	#
	# Switch to another package before executing list sql because 
	# several globals are set.
	#

	package order;

	my $detail      = ($f && $f->detail)		 || $t->detail;

	my $detail_key  = $dt->key;
	my $edit_prog   = $dt->edit_prog;

	my $list_fields = $order_list	||
			  ($f && $f->detail_list_fields) || $t->detail_list_fields;

	my $list_format = $order_format ||
			  ($f && $f->detail_format)	 || $t->detail_format;
	my $list_header = $order_header ||
			  ($f && $f->detail_header)	 || $t->detail_header;

	my $detail_sql  = $order_sql	||
			  ($f && $f->detail_sql)	 || $t->detail_sql;

	#
	# Start Control section
	#

	our $secctl =  new Acme::HTML::SectionControl
			name	=> 'orders',
			title	=> 'Orders',
			display	=> $display_info;
	print $secctl->begin;

	#
	# Determine if user has privilages to edit record.
	#

	my $has_edit_priv = Acme::Security::allow('Update',
						  $detail_table,
						  $dt->secure,
						  $dt->restrict);
	#
	# Process SQL query for detail records.
	#

	my $sqlo = new Acme::SQL $detail_sql;

	# Assure detail table key field is in query
	$sqlo->add_column("$detail_table.$detail_key")
	  if ! $sqlo->is_column("$detail_table.$detail_key");

	(my $detail_join_field = $detail) =~ s/.*\.//;

	$sqlo->add_where("$detail_table.$detail_join_field = $keyval");

	$detail_sql = Acme::Security::clean_sqlo($sqlo);

	#
	# Define globals for the list helper functions.
	#

	my $bullet = Acme::Config->Bullet;

	my $row_name = $dt->row_name;

	my $nrec = 0;				# record number

	#
	# Define helper functions to list detail items.
	#

	#----------------------------------------
	my $before_list = sub {

		#
		# Compose list header
		#

		my $header = '';
		if ($list_header) {
			for my $h (grep { s/$detail_table\.// } split(/\s*,\s*/,$list_fields)) {
				if ($dt->{$h}) {
					push @headers, $dt->{$h}->title;
				}
				else {
					# Capitalize field name
					# (happens on ID fields)
					push @headers, capitalize($h);
				}
			}

			$header = Local::Acme::Format::sprintf_cmd($list_header,@headers);

			if ($header !~ m|^\s*<th|i
			||  $header =~ s|^\s*<td>(.*)</td>\s*$|$1|i) {
				$header = qq[<th>$list_header</th>];
			}

			#
			# Prepend extra columns for record index and edit bullet.
			#

			$header = qq[<th>&nbsp;</th>$header]
			  if $has_edit_priv;

			$header = qq[<th>&nbsp;</th>\n$header];

			$header = qq[<tr>\n$header\n</tr>];
		}

		return <<End_Of_HTML;
<div class="detail-list">
<table class="detail-list">
$header
End_Of_HTML
	};

	#----------------------------------------
	my $after_list = sub {
		return <<End_Of_HTML;
</table>
</div>

End_Of_HTML
	};

	#----------------------------------------
	my $before_row = sub {

		$nrec++;

		my $html = '';	# returned value

		$html .= <<End_Of_HTML;
<tr>
<td class="detail-index">$nrec</td>
End_Of_HTML

		#
		# Create edit bullets for editing order items.
		#

		no warnings qw(uninitialized once);	# for @a in Lister
		no strict qw(vars refs);
		my $keyval = Acme::HTTP::escape(${ ${detail_table}.'__'.${detail_key} });


		my $sep    = $edit_prog =~ /\?/ ? '&' : '?';
		my $url    = "$edit_prog${sep}$detail_key=${keyval}&request=edit";

		$html .= <<End_Of_HTML if $has_edit_priv;
<td class="detail-link"><a href="$url" onClick="new_target(this,'form')" title="Edit $row_name." class="edit-bullet">$bullet</a></td>
End_Of_HTML
		return $html;
	};

	#----------------------------------------
	my $after_row = sub {
		return <<End_Of_HTML;
</tr>
End_Of_HTML
	};

	#----------------------------------------
	my $every_row = sub {
		no warnings qw(uninitialized once);	# for @a in Lister
		no strict 'vars';

		my $row = Local::Acme::Format::sprintf_cmd($list_format,@a);

		$row = qq[<td>$row</td>\n]
		  if $row !~ m|^\s*<td|i;
		return $row
	};

	#
	# Construct List
	#

	my $l = new Acme::Lister
			sql => $detail_sql,
			before_row => $before_row,
			after_row  => $after_row,
			every_row  => $every_row,
			sandbox => 'order';

	#
	# Print list
	#

	print &$before_list();
	if ($l->count == 0) {	print "<tr>\n<td>(none)</td>\n</tr>\n";   }
	else {			print $l->list(); }
	print &$after_list();

	#
	# End Control section
	#

	print $secctl->end;

	return $l->count;
}

package order;

sub capitalize {
	my $s = shift;
	$s =~ s/_/ /g;		# change _ to spaces
	$s =~ s/\bid\b/ID/;	# uppercase ID
	$s =~ s/\b(\w+)\b/ucfirst(lc($1))/eg; # capitalize
	return $s;
}

1;
