#!/usr/bin/perl -I../lib

=head1 NAME

    doc.cgi - Display documentation of programs and library modules.

=head1 SYNOPSIS

    cgi-bin/doc.cgi?topic=table_name

    cgi-bin/doc.cgi?topic=table_name&library=dir

    cgi-bin/doc.cgi?topic=/help/help.html

    cgi-bin/doc.cgi?library=/usr/doc/samba-2.0.6

    cgi-bin/doc.cgi?topic=/usr/doc/samba-2.0.6/docs/textdocs/WinNT.txt


=head1 DESCRIPTION

Doc.cgi displays a table of contents of programs and library modules which
have embedded POD (plain old documentation).

If a topic is specified, the documentation is displayed for that topic.

A topic is usually a file name. A library is a known name defined in the
script or it is directory name.

Most topics are stored in libraries -- directories referenced by global
variables in the name, e.g. $Lib/Form.pm.

=head2 NOTE

There are some security issues allowing the display of source for any file
in the system. Set $Doc_Any_File to 0 to disable display of any file outside
the web site's DocumentRoot.

=head1 CHANGE LOG

 2/1/00   Initial version.
 11/6/01  Changed topic definition to accept file names if they start with /.
 	  This allows this program to be used as an Apache action to filter
	  POD files found in the document tree.
 11/11/01 Added source=yes argument to print source code.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000,2001 Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Acme::HTML			qw(escape_html);
use Acme::HTML::NavButtons;
use Local::Acme::Banners	qw(common_header common_footer);
use Acme::Config;
use CGI;

use Acme::Security;
Acme::Security::check('Access');

$|=1;
$\="\n";

# Set following to 0 to disallow displaying any os file
$Doc_Any_File = Acme::Config->Doc_Any_File;
$Doc_Source   = Acme::Config->Doc_Source;

$Bullet       = Acme::Config->Bullet;

$Default_Library = 'AcmeBase';		# default library

$Base         = Acme::Config->Base;
$Root         = Acme::Config->Root;
$DocumentRoot = Acme::Config->DocumentRoot;

$Styles       = Acme::Config->Styles;

# Define symbolic locations (see realfile() below)
$Doc          = $DocumentRoot . Acme::Config->Doc;
$Help         = $DocumentRoot . Acme::Config->Help;
$CGI          = '.';
$Lib          = '../lib';
$Schema       = '../schema';

$Me           = "$Base/cgi-bin/doc.cgi";

# Define files and directories in a library.
%Library = (
	'AcmeBase' => [
		'$Doc',
		'$Help',
		'$CGI',
		'$Lib',
		'$Schema',
		],
	);

$cgi = new CGI;
print $cgi->header(-type=>'text/html') unless @ARGV;

$topic   = $cgi->param('topic')   || $ENV{PATH_INFO} || '';
$library = $cgi->param('library') || $Default_Library;
$source  = $cgi->param('source')  || ''; # set to 'yes' to print source code
$title   = $topic;
$title  =~ s/^\$//;              # remove initial $ on library paths
$title = "doc $title";


if ($topic) {

	if ($topic =~ m|^/| && $topic !~ m|\$|) {
		# topic is a file name beginning with a /
		if (-r "$DocumentRoot$topic") {
			# File is found relative to DocumentRoot
			$topic = "$DocumentRoot$topic"
		}
		elsif ( (my $perltopic = $topic) =~ s|/doc/perl5|/usr/lib/perl5|) {
			die "Cannot document topic '$topic'" unless $Doc_Any_File;
			$perltopic =~ s/\.html$/\.pod/	     unless -r $perltopic;
			$perltopic =~ s/\.pod$/\.pm/	     unless -r $perltopic;
			$perltopic =~ s/\.pm$/\.pl/	     unless -r $perltopic;
			die "Cannot find topic '$topic'"     unless -r $perltopic;
			$topic = $perltopic;
		}
		else {
			# File is an absolute file name.
			die "Cannot document topic '$topic'" unless $Doc_Any_File;
			die "Cannot find topic '$topic'"     unless -r $topic;
		}
	}
	else {
		$topic = realfile($topic);
	}

	# Print topic documentation
	my $t;
	if ($source =~ /yes/i) {
		# Print source code
		open TXT, '<', $topic;		# open text file
		local $/;
		$t = <TXT>;	# read entire input
		close TXT;

		$t = "<pre>".escape_html($t)."</pre>" if $t;
	}
	elsif ($topic =~ /\.htm/) {
		open HTML, '<', $topic;	# open html file
		local $/;
		$t = <HTML>;	# read entire input
		close HTML;
	}
	elsif ($topic =~ /\.txt/) {
		open TXT, '<', $topic;		# open text file
		local $/;
		$t = <TXT>;	# read entire input
		close TXT;

		$t = "<pre>".escape_html($t)."</pre>" if $t;
	}
	elsif ($topic =~ /\.pl/ || $topic =~ /\.cgi/) {

		# Old style perl programs do not have POD. Print only the
		# header comments.
		# Make sure cgi programs do have POD. If they do not, print
		# only the header comments.

		open PL, '<', $topic;
		local $/;
		$t = <PL>;	# read entire input
		close PL;

		if ($t =~ /^=head/m) {
			# There is POD documentation
			# Re-read input through pod2html.
			chdir '/tmp';
			# Make non-absolute path names absolute.
			$topic = "$Root/cgi-bin/$topic" unless $topic =~ m|^/|;
			open POD, "/usr/bin/pod2html"
			  . " --title='$title'"
			  . " --podroot=/usr/lib"
			  . " --podpath=perl5"
			  . " --htmlroot=$Base/doc"
			  . " '$topic' |"
			  or die $!;

			local $/;
			$t = <POD>;	# read entire input
			close POD;
		}
		else {
			if ($t) {
				$t = escape_html($t);
				$t = "<html><body><h2>$topic</h2><pre>$t</pre></body></html>";
			}
		}
	}
	else {
		# Change to a directory that is writeable to pod2html for cache.
		# The cache is sharable to all. It probably needs to be locked
		# or made individual to each cgi call.
		chdir '/tmp';		# change to a writeable directory
		# Make non-absolute path names absolute.
		$topic = "$Root/cgi-bin/$topic" unless $topic =~ m|^/|;
		open POD, "/usr/bin/pod2html"
		  . " --title='$title'"
		  . " --podroot=/usr/lib"
		  . " --podpath=perl5"
		  . " --htmlroot=$Base/doc"
		  . " '$topic' |"
		  or die $!;
		local $/;
		$t = <POD>;	# read entire input
		close POD;
	}

	if ($t) {
		print $t, "\n";
	}
	else {
		die "Cannot find documentation for '$topic'.\n";
	}
}
else {
	# List table of contents

	if ($Library{$library}) {
		push @f, map {  if(-d realfile($_)){lsR($_)}else{$_} } @{$Library{$library}};
	}
	else {
		die "Cannot document library '$library'" unless $Doc_Any_File;
		push @f, lsR($library);
	}

	local $b = new Acme::HTML::NavButtons;


	print <<EOF;
<html>
<head>
<title>doc Index of Topics</title>
<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<link rel="stylesheet" href="$Styles/list.css" type="text/css">
</head>
<body>
EOF
	
	print $b->first_buttons;
	print common_header();

	print <<EOF;
<h1>Documentation Topics</h1>

<h2>$library</h2>

<div style="margin-left:5ex; margin-bottom:2ex">
EOF
	for (@f) {
		$topic = $_;
		s|$library/|| if $library =~ m|^/|;	# strip library from full paths
		s/^\$//;				# strip $ from $Lib

		# Cull dev directories embedded in source tree.
		next if /\bnot_used\b/;		# cull unused programs
		next if /\bts\b/;		# cull test/debug scripts

		print qq(<br>)
		  . make_bullet("$Me?topic=$topic&source=yes",
				"View source code")
		  . qq(<a href="$Me?topic=$topic" target="_blank" title="View documentation">$_</a>);
	}

	print "\n</div>\n";
	
	print common_footer();
	print $b->last_buttons;
	print <<EOF;
</body>
</html>
EOF
}

exit 0;

#------------------------------------------------------------------------------

sub make_bullet {
	my ($href, $title) = @_;
	return qq[$Bullet&nbsp;] if ! $href || ! $Doc_Source;
	return qq[<a href="$href" target="_blank" class="edit-bullet" title="$title">$Bullet&nbsp;</a>];
}

#------------------------------------------------------------------------------

sub realfile {
	# Return real file name of arguments after substituting
	# global variables.

	my $f = shift;
	$f =~ s/\$(\w+)/${$1}/eg;	# substitute global variables
	return $f;
}

#------------------------------------------------------------------------------

sub lsR {
	# Recursively list all files in directory

	my $dir = shift || '.';		# default is current directory

	opendir DIR, realfile($dir)
	  or die "Can't get directory list of '$dir'. $!";

	my @files = readdir DIR;
	closedir DIR;
	shift @files; shift @files;	# remove . and ..

	#
	# Filter files for *.pm, *.pl, *.cgi, *.pod, *.htm, or *.txt files.
	#

	my @ra = grep {  ! -d realfile("$dir/$_") && /(\.pm|\.pl|\.cgi|\.pod|\.htm|\.txt)$/ && ($_ = "$dir/$_")} @files;

	#
	# Decend into subdirectories.
	#

	for (@files) { push @ra, lsR("$dir/$_") if -d realfile("$dir/$_"); }

	return sort @ra;
}
__END__
