#!/usr/bin/perl -wT -I../lib
require 5.6.0;
use strict;

=head1 NAME

list.cgi -- Display database information in a list format.

=head1 SYNOPSIS

    # Display a form to select a table and to enter search criteria.
    list.cgi

    # Display a form to enter search criteria for searching a table.
    list.cgi?table=customers&search

    # List customers whose name matches "Folly".
    list.cgi?table=customers&search=Folly

    # Display a list of previously saved lists for a table.
    list.cgi?table=customers

    # List results of a SQL query.
    list.cgi?sql=select+f1 ...

=head1 DESCRIPTION

The list.cgi program displays rows of data from a database table or from
the sql script passed as an argument. The search parameter is combined
with the query if it is specified.

The "request" parameter can be the code words "List" or "Count" or "Print".
The "List" option is the default. "Count" suppresses the printing of each row
and just prints the number of rows in the list and other summary information.
Setting the request to 'Print' prints the list.

An icon displayed at the beginning of each record. It is a link to select
or deselect the record. Only selected records are operated on by the list
operations like "Export", or "Edit".

A format string may be specified by using the 'format' parameter. It is
a string used by I<Local::Acme::Format>, which is essentially a string to
be formatted by sprintf() and then scanned for functions to execute. The
functions need to be defined in I<Local::Acme::Format> in order to be
recognized in the formatted string.

A list is formatted as an HTML table, with each list item placed between <td>
and </td> tags. That allows the format string to also have <td> and </td>
tags embedded in it to format fields into table columns.

A list can be customized many ways: which columns to display, sort order,
columns to total, and filters to set in order to extract just the records you
want. The list button bar has option menus that can adjust list options. Those
options are saved after every setting and they are used the next time the list
is displayed. In other words, lists are displayed as they were last displayed.

A C<list_id> is used to reference saved list options. The table name is usually
used as the list ID, however, one can be explicitly set in the arguments. The
list ID of "sql" is used if a table is not known and an ID is not specified.

Three parameters are required for lists and list processing: C<list_id>,
C<sql>, and C<key>. However, these parameters are usually inferred by
other parameters. The C<list_id> is usually a database table's name,
C<sql> is usually the table's list as determined by the schema files (see
I<Acme::Base::Schema>), and the fully qualified key name also comes from
the schema files. Most often, this program is called with only a table
name and a search string to use to search the table.

=head1 TODO

 * Group totals.
 * Page breaks.

=head1 CHANGE LOG

 5/04/00  Initial version.
 9/09/00  Introduced list_id to use to identify user options on the list display.
 6/09/01  Converted to Acme::Base::Schema and Acme::SQL.
 6/12/01  Converted to Acme::Lister. Divided into list.cgi and pick.cgi,
          with this version specializing in list operations.
 8/10/02  Added customize menus to button bar.
 11/27/02 Added open/save reports.
 12/22/02 Rewrote selector filter.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000-2002 Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

package list;

use Acme::HTTP			qw(escape	unescape);
use Acme::HTML			qw(escape_html display_message);
use Acme::HTML::Parameters;
use Acme::Base::Schema;
use Acme::Base::Schema::Join	qw(parse_join);
use Acme::Base::Schema::Master;
use Acme::SQL;
use Acme::Searcher;
use Acme::Lister;
use Acme::Lister::Selector;
use Acme::Options;
use Acme::HTML::NavButtons;
use Local::Acme::Banners  	qw(common_header common_footer);
use Local::Acme::Format		qw(sprintf_cmd);
use Local::Acme::User		qw(user		user_folder);
use Acme::Config;
use CGI;

use Acme::Security;
Acme::Security::check('Access');

$|=1;		# flush output (synchronize errors with output)

our $cgi = new CGI;

our $Base	  = Acme::Config->Base;
our $Styles       = Acme::Config->Styles;	# location of CSS styles
our $Images       = Acme::Config->Images;	# location of image files
our $Scripts_Lib  = Acme::Config->Scripts_Lib;	# location of JavaScripts
our $Help         = Acme::Config->Help;		# location of help files
our $Bullet       = Acme::Config->Bullet;	# edit bullet
our $No_Tabindex  = Acme::Config->No_Tabindex || '-1';

our $schema = new Acme::Base::Schema;
our $t;  					# table properties pointer

#
# Program Arguments
#

# Note: We allow arguments to be specified both in the input stream (POST) and
# in the url (GET). Lists can be regenerated with different parameters so we 
# check the POSTed parguments first, unlike the form.cgi program which checks
# their url parameters first.

our $table	  = $cgi->param('table')	|| $cgi->url_param('table')	|| '';

#
# Get the list ID to use for list settings and options.
# The default value is usually the table's name.
#

our $list_id	   = $cgi->param('list_id')	|| $table || 'SQL';

#
# Check security access to table.
#

Acme::Security::check('Access',$list_id);

#
# Get search script.
#

our $search	  = $cgi->param('search')	|| $cgi->url_param('search')	|| '';

#
# Check if we are to open up a report.
#

our $report	   = $cgi->url_param('report')	|| '';
our $report_id     = $cgi->param('report_id')	|| '';

if ($report) {

	our $Report_Space = Acme::Config->Report_Space;	# location of reports

	$report =~ s|^/||;	# remove leading / 

	my $fd;
	my $report_found = 0;	# flag
	for my $filepath ("$list_id/$report","$report") {
		next unless Acme::Security::allow('Access',$filepath);
		next unless open $fd, '<', "$Report_Space/$filepath";
		$report_found = 1;

		#
		# Redefine cgi arguments from saved report file.
		#

		my $report_cgi = new CGI $fd;	# read saved settings
		close $fd;

		#
		# Check if report is to be rendered using a different program.
		#

		if (my $report_url = $report_cgi->param('report_url')) {
			print $cgi->redirect($report_url);
			exit 0;
		}

		#
		# Reassign search parameter if list_id and report_id are the
		# same, otherwise the search script in the report file is used.
		#

		$report_id = $report_cgi->param('report_id')	|| '';


		if ($report_id eq $list_id || $list_id eq 'SQL') {
			if ($search) {

				#
				# Assign current search to report search.
				#

				$report_cgi->param('search',$search);
			}
			else {
				#
				# Get search parameters from user.
				#

				$table ||= $report_cgi->param('table');

				# Skip getting search parameters if the
				# report has a sql statement and no
				# table or search argument.

				if ($table || ! $report_cgi->param('sql')) {
					get_search();
					exit 0;
				}
			}
		
		}

		#
		# Clean out saved cgi parameters that don't apply now.
		#

		# Report settings are not propogated when redrawing to allow
		# the user to customize the report. Only the report_id parameter
		# remains set and passed on as a flag when redrawing the list.
		#$report_cgi->delete('report_id');
		
		$cgi->delete('report');

		$report_cgi->delete('original_search');
		$report_cgi->delete('redrawing');

		$report_cgi->delete('exclude');
		$report_cgi->delete('include');
		$report_cgi->delete('invert_all');
		$report_cgi->delete('select_all');

		#
		# Re-assign $table and $search globals.
		#

		$table   = $report_cgi->param('table')   || '';
		$list_id = $report_cgi->param('list_id') if $report_cgi->param('list_id');
		$search  = $report_cgi->param('search')  || '';

		#
		# Set cgi to the report's cgi.
		#

		$cgi = $report_cgi;
		last;
	}

	die "Cannot open the report '$report' for list Id '$list_id'. $!"
	  if ! $report_found;
}

#
# Check if we are to open up a list
#

if (my $listfile = $cgi->url_param('list')) {

	our $List_Space   = Acme::Config->List_Space;	# location of lists

	$listfile =~ s|^/||;	# remove leading / 

	# If the table parameter is not specified, look at first component
	# of the list file name, as lists are stored in directories with
	# the same name as the table's name.

	($table) = $listfile =~ m|(.+?)/| unless $table;

	die "Table name not specified.\n"
	  . "Lists are categorized by table names, so a table name needs"
	  . " to be specified along with a list name."
	  unless $table;

	# Ensure list_id is set to table if table was just defined above.
	$list_id = $table if ! $list_id || $list_id eq 'SQL';

	my $fd;
	my $list_found = 0;	# flag
	for my $filepath ("$table/$listfile","$listfile") {
		next unless -f "$List_Space/$filepath";
		$list_found = 1;

		#
		# Redefine search argument to read list file.
		#
		
		$t = $schema->t($table)
		  or die "Cannot find a database table named '$table'\n";

		my $key = $t->key;

		$search = qq[List "$filepath";\n]
		  . qq[list_join('$filepath', '$table.$key');\n];

		#
		# Clean out cgi parameters that don't apply now.
		#

		$cgi->delete('list');
		$cgi->delete('list_title');

		$cgi->delete('original_search');
		$cgi->delete('redrawing');

		$cgi->delete('exclude');
		$cgi->delete('include');
		$cgi->delete('invert_all');
		$cgi->delete('select_all');

		last;
	}

	die "Cannot find the list '$listfile' for table '$table'. $!"
	  if ! $list_found;
}

#
# Save the initial version of the search script.
#

our $original_search = $cgi->param('original_search');
if (! defined $original_search) {

	# The original_search is used when saving a report. It has the 
	# original script which might include statements to get parameters
	# before rendering the report.

	# Save it as soon as possible and actually set the $cgi parameter if
	# needed so that it gets propagated when search scripts ask for
	# parameters from the user.

	$original_search = $search;
	$cgi->param('original_search',$original_search);
}

#
# Get list's CGI parameters.
#

our $sql	  = $cgi->param('sql')		|| $cgi->url_param('sql')	|| '';
our $original_sql = $sql;	# save original as $sql gets redefined below.

our $title	  = $cgi->param('title')	|| $cgi->url_param('title')	|| '';

our $request	  = $cgi->param('request')	|| $cgi->url_param('request')	|| 'List';

our $Printing     = $cgi->url_param('print')	|| $cgi->param('print')		|| '';
$Printing         = $Printing =~ /yes/i         ||
		    $request =~ /print/i        ||
		    ($Printing !~ /^no$/i && exists $cgi->{'print'} ? 1 : 0);

if ($table && ! $search) {
	
	#
	# Get previously saved list file for the table.
	#

	# Note: this does a redirect so we must do this before outputting the
	# HTTP header.

	get_list_file();
	exit 0;
}

if ( ! ($table || $sql || $search) ) {
	
	#
	# Get search parameters from user.
	#

	get_search();
	exit 0;
}

print $cgi->header(-type=>'text/html') unless @ARGV;

our $sort	  = $cgi->param('sort')		|| $cgi->url_param('sort')	   || '';

our $format 	  = $cgi->param('format')	|| $cgi->url_param('format')	   || '';
our $list_fields  = $cgi->param('list_fields')	|| $cgi->url_param('list_fields')  || '';

our $list_title   = $cgi->param('list_title')	|| $cgi->url_param('list_title')   || '';
our $annotate     = $cgi->param('annotate')	|| $cgi->url_param('annotate')     || '';

our $record_set_name  = $cgi->param('record_set_name') || $cgi->url_param('record_set_name');

our $id_field	  = $cgi->param('id_field')     || $cgi->url_param('id_field')	   ||
		    $cgi->param('id_field')     || $cgi->url_param('id_fields')    || '';
our $code_field	  = $cgi->param('code_field')   || $cgi->url_param('code_field')   ||
		    $cgi->param('code_field')   || $cgi->url_param('code_fields')  || '';
our $name_field	  = $cgi->param('name_field')   || $cgi->url_param('name_field')   ||
		    $cgi->param('name_field')   || $cgi->url_param('name_fields')  || '';
our $phone_field  = $cgi->param('phone_field')  || $cgi->url_param('phone_field')  ||
		    $cgi->param('phone_field')  || $cgi->url_param('phone_fields') || '';
our $email_field  = $cgi->param('email_field')  || $cgi->url_param('email_field')  ||
		    $cgi->param('email_field')  || $cgi->url_param('email_fields') || '';
our $first_name_field = $cgi->param('first_name_field') || $cgi->url_param('first_name_field');
our $last_name_field  = $cgi->param('last_name_field')  || $cgi->url_param('last_name_field');

our $code_field_match = $cgi->param('code_field_match') || $cgi->url_param('code_field_match');
our $phone_field_match= $cgi->param('phone_field_match')|| $cgi->url_param('phone_field_match');

#
# Get list options that comes from the table's schema.
# They determine the button bar options.
#

# Note: with the exception of $key, the following settings do not have
# corresponding <input> fields in the list form to preserve their settings
# from one invocation to the next unless they are set in the url.

our $key	   = $cgi->param('key') || '';

our $edit_prog     = $cgi->url_param('edit_prog')    || $cgi->param('edit_prog')    || '';
our $view_prog     = $cgi->url_param('view_prog')    || $cgi->param('view_prog')    || '';

our $logall_prog   = $cgi->url_param('logall_prog')  || $cgi->param('logall_prog')  || '';
our $emailall_prog = $cgi->url_param('emailall_prog')|| $cgi->param('emailall_prog')|| '';
our $export_prog   = $cgi->url_param('export_prog')  || $cgi->param('export_prog')  || '';
our $editall_prog  = $cgi->url_param('editall_prog') || $cgi->param('editall_prog') || '';

#
# Get limits
#

# Note: only limits set in the url are used, and they are overwritten by limits
# specified in the schema if they are lower. The (possibly modified) $limit_abort
# value is passed on to list.cgi helper programs using a parameter named "limit".
# The limit_alert email address is added to the list specified in the schema.

our $limit_abort   = $cgi->url_param('limit_abort')  || $cgi->url_param('limit')    || '';
our $limit_warn    = $cgi->url_param('limit_warn')   || '';
our $limit_alert   = $cgi->url_param('limit_alert')  || '';

#
# Get current list item's check settings.
#

# These parameters allow the list to be redrawn in a different format without
# loosing the current rows' checkmark settings.

# Note: the accumulator variables n_rows, n_checked, n_unchecked, ...
# associated with the following are recomputed as the list is being rendered.

our $include      = $cgi->param('include')       || '';
our $exclude      = $cgi->param('exclude')       || '';

our $select_all   = $cgi->param('select_all')    || '';
our $invert_all   = $cgi->param('invert_all')    || '';

$select_all     = '' unless 'yes' =~ /\Q$select_all\E/i; # make binary
$invert_all     = '' unless 'yes' =~ /\Q$invert_all\E/i; # make binary

#
# Get list settings from 1) CGI, 2) list settings, or 3) user option settings.
#

my $list_settings_file = $cgi->param('list_settings')     ||
			 user_folder() . "/${list_id}ListSettings";
our $list_settings = new Acme::Options file => $list_settings_file
  unless $report_id;

#
# Get user options
#

our $opt = new Acme::Options;			# user option settings

#
# Set list options
#

# First initialize flags which tell us if this is the first time the list
# is rendered. The "redrawing" parameter is not intended to be one of the
# list options, but is entered in the list_control form with a constant "yes"
# telling us whe the user has selected the "Redraw" option. This helps us
# know when to reset the options back to "nothing" upon user request.

our $redrawing           = $cgi->param('redrawing') || '';
$redrawing               = '' unless 'yes' =~ /\Q$redrawing\E/i;

# Get customize options (none, previous)

our $customize           = $cgi->param('customize')		||
			   $cgi->param('custom')		|| '';
$customize             ||= $list_settings->get("ListCustomize") if $list_settings;
$customize             ||= $opt->get("${list_id}Customize");

our $nocustomize = 'none' =~ /\Q$customize\E/i;	# set true/false flag

# Turn following options into true/false values ('y*'/nothing).

# nohilo
our $nohilo              = $cgi->param('highlighter')		|| '';
$nohilo                ||= $opt->get("listUseHiliter");	# hardwire 'list'
$nohilo                  = '' unless 'yes' =~ /\Q$nohilo\E/i;
$nohilo		         = $nohilo ? '' : 'yes';		# reverse sense

# list_as_rows
our $list_as_rows        = $customize eq 'none' ? 'yes' : '';
$list_as_rows          ||= $list_settings->get("ListAsRows")
			   if ($customize eq 'previous' || (! $table && ! $redrawing))
			   && $list_settings;
$list_as_rows          ||= $cgi->param('list_as_rows')		|| '';
$list_as_rows          ||= $list_settings->get("ListAsRows") if $list_settings;
$list_as_rows          ||= $opt->get("${list_id}ListAsRows");
$list_as_rows            = '' unless 'yes' =~ /\Q$list_as_rows\E/i;

# list_as_table
our $list_as_table       = $nocustomize ? 'no' : '';
$list_as_table           = $list_as_rows ? 'no' : '';
$list_as_table         ||= $list_settings->get("ListAsTable")
			   if $customize eq 'previous' && $list_settings;
$list_as_table         ||= $cgi->param('list_as_table')		|| '';
$list_as_table         ||= $list_settings->get("ListAsTable") if $list_settings;
$list_as_table         ||= $opt->get("${list_id}ListAsTable");
$list_as_table           = '' unless 'yes' =~ /\Q$list_as_table\E/i;

# list_all_columns
our $list_all_columns   = $nocustomize ? 'no' : '';
$list_all_columns      ||= $list_settings->get("ListAllColumns")
			   if $customize eq 'previous' && $list_settings;
$list_all_columns      ||= $cgi->param('list_all_columns')	|| '';
$list_all_columns      ||= $list_settings->get("ListAllColumns") if $list_settings;
$list_all_columns      ||= $opt->get("${list_id}ListAllColumns");
$list_all_columns        = '' unless 'yes' =~ /\Q$list_all_columns\E/i;

# Set list option codes.

# list_format_style
our $list_format_style   = $nocustomize ? 'list' : '';
$list_format_style     ||= $list_settings->get("ListFormatStyle")
			   if $customize eq 'previous' && $list_settings;
$list_format_style     ||= $cgi->param('list_format_style')	||
			   $cgi->param('list_format')		|| '';
$list_format_style     ||= $list_settings->get("ListFormatStyle")
			   if ! $nocustomize && $list_settings;
$list_format_style     ||= $opt->get("${list_id}ListFormatStyle");
$list_format_style     ||= 'list';

# list_column_headers
our $list_column_headers = $list_settings->get("ListColumnHeaders")
			   if $customize eq 'previous' && $list_settings;
$list_column_headers   ||= $cgi->param('list_column_headers')	|| '';
$list_column_headers   ||= $list_settings->get("ListColumnHeaders")
			   if ! $nocustomize && $list_settings;
$list_column_headers   ||= $opt->get("${list_id}ListColumnHeaders");
$list_column_headers   ||= 'FieldNames';

# columns
our @columns             = split(/,/,$list_settings->get("ListColumns"))
			   if $customize eq 'previous' && $list_settings;
@columns                 = $cgi->param('columns')	if ! @columns;
@columns	  	 = $cgi->param('fields')	if ! @columns;
@columns                 = split(/,/,$list_settings->get("ListColumns"))
			   if ! @columns && $customize ne 'none' && $list_settings
			   && $list_id ne 'SQL';
@columns                 = split(/,/,$opt->get("${list_id}Columns"))
			   if ! @columns && $list_id ne 'SQL';

# sort_columns
our @sort_columns        = split(/,/,$list_settings->get("ListSortColumns"))
			   if $customize eq 'previous' && $list_settings;
@sort_columns            = $cgi->param('sort_columns')	if ! @sort_columns;
@sort_columns	  	 = $cgi->param('sort_fields')	if ! @sort_columns;
@sort_columns            = split(/,/,$sort)  		if ! @sort_columns && $sort;
@sort_columns            = split(/,/,$list_settings->get("ListSortColumns"))
			   if ! @sort_columns && $customize ne 'none' && $list_settings;
@sort_columns            = split(/,/,$opt->get("${list_id}SortColumns"))
			   if ! @sort_columns;

# total_columns
our @total_columns       = split(/,/,$list_settings->get("ListTotalColumns"))
			   if $customize eq 'previous' && $list_settings;
@total_columns           = $cgi->param('total_columns')	if ! @total_columns;;
@total_columns	  	 = $cgi->param('total_fields')	if ! @total_columns;
@total_columns           = split(/,/,$list_settings->get("ListTotalColumns"))
			   if ! @total_columns && $customize ne 'none' && $list_settings
			   && ! $redrawing;
@total_columns           = split(/,/,$opt->get("${list_id}TotalColumns"))
			   if ! @total_columns && ! $redrawing;

# group_columns
our @group_columns       = split(/,/,$list_settings->get("ListGroupColumns"))
			   if $customize eq 'previous' && $list_settings;
@group_columns           = $cgi->param('group_columns')	if ! @group_columns;;
@group_columns	  	 = $cgi->param('group_fields')	if ! @group_columns;
@group_columns           = split(/,/,$list_settings->get("ListGroupColumns"))
			   if ! @group_columns && $customize ne 'none' && $list_settings
			   && ! $redrawing;
@group_columns           = split(/,/,$opt->get("${list_id}GroupColumns"))
			   if ! @group_columns && ! $redrawing;

# filter_columns
our @filter_columns      = split(/,/,$list_settings->get("ListFilterColumns"))
			   if $customize eq 'previous' && $list_settings;
@filter_columns          = $cgi->param('filter_columns')          if ! @filter_columns;
@filter_columns          = split(/,/,$cgi->param('filter') || '') if ! @filter_columns;
# Don't use previous list invocation settings. User sets filters each session.
#@filter_columns          = split(/,/,$list_settings->get("ListFilterColumns"))
#			   if ! @filter_columns && $customize ne 'none' && $list_settings
#			   && ! $redrawing;
# no user default setting for filter (schema setting rules supreme)
#@filter_columns	 = split(/,/,$opt->get("${list_id}FilterColumns"))
#			   if ! @filter_columns && ! $redrawing;

# filter_desc	(elements corresponds to the @filter_columns elements)
our @filter_desc	 = split(/\s*,\s*/, $cgi->param('filter_desc') || '');

# selector
our $selector            = $cgi->param('selector');
# Don't use previous list invocation settings. User sets selector each session.
# Maybe use this if the schema does not specify a selector???
#$selector              ||= $list_settings->get("ListSelector")
#			   if $customize ne 'none' && $list_settings && $redrawing;
# no user default setting for selector (schema setting rules supreme)
#$selector              ||= $opt->get("${list_id}ListSelector");
$selector	       ||= '';

# selector_cond
our $selector_cond       = $cgi->param('selector_cond');
# Don't use previous list invocation settings. User sets selector each session.
# Maybe use this if the schema does not specify a selector???
#$selector_cond         ||= $list_settings->get("ListSelectorCond")
#			   if $customize ne 'none' && $list_settings && $redrawing;
# no user default setting for selector (schema setting rules supreme)
#$selector_cond         ||= $opt->get("${list_id}ListSelectorCond");
$selector_cond         ||= 'not null';

# selector_param1
our $selector_param1     = $cgi->param('selector_param1');
# Don't use previous list invocation settings. User sets selector each session.
# Maybe use this if the schema does not specify a selector???
#$selector_param1       ||= $list_settings->get("ListSelectorParam1")
#			   if $customize ne 'none' && $list_settings && $redrawing;
# no user default setting for selector (schema setting rules supreme)
#$selector_param1       ||= $opt->get("${list_id}ListSelectorParam1");
$selector_param1       ||= '';

# selector_param2
our $selector_param2     = $cgi->param('selector_param2');
# Don't use previous list invocation settings. User sets selector each session.
# Maybe use this if the schema does not specify a selector???
#$selector_param2       ||= $list_settings->get("ListSelectorParam2")
#			   if $customize ne 'none' && $list_settings && $redrawing;
# no user default setting for selector (schema setting rules supreme)
#$selector_param2       ||= $opt->get("${list_id}ListSelectorParam2");
$selector_param2       ||= '';

# selector_param3
our $selector_param3     = $cgi->param('selector_param3');
# Don't use previous list invocation settings. User sets selector each session.
# Maybe use this if the schema does not specify a selector???
#$selector_param3       ||= $list_settings->get("ListSelectorParam3")
#			   if $customize ne 'none' && $list_settings && $redrawing;
# no user default setting for selector (schema setting rules supreme)
#$selector_param3       ||= $opt->get("${list_id}ListSelectorParam3");
$selector_param3       ||= '';

#
# Ensure all "mode" flags are properly set.
#

our $counting = $request =~ /Count/i;		# flag

$list_as_rows ||= 'yes' unless $list_as_table || $counting;
$list_all_columns = '' if $list_as_rows;	# force to 'false'

#
# Reset search if filters not being used anymore.
#

if ($redrawing) {

	#
	# Scan search script for all the filter definitions.
	#

	# Note: we only need to remove those filters in the search
	# scripts that do not correspond to the @filter_columns.
	# Filters are added to the search script on the client side.
	
	my @fltrs = $search =~ m/^\$(FLTR_\w+)/mg;

	# Remove possible "null" keyword from filter column name.
	my %valid_filter = map { s/ null$//i; $_, 1 } map { $_ } @filter_columns;

	for my $fltr (@fltrs) {
		
		(my $col = $fltr) =~ s/FLTR_//; $col =~ s/__/\./g;

		next if $valid_filter{$col};

		#
		# Remove filter from search script.
		#

		$search =~ s/\$$fltr\s*=.*?[;|]\s*//s;	# remove script line
		$search =~ s/(\s+and\s+)?\$$fltr//s;	# remove from where clause

		#
		# Remove description from annotation
		#

		# Note: the text below must match that defined in the
		# Javascript file list.js.

		# Note: the "period space" is used as the terminating
		# delimiter, not just a period.

		$annotate =~ s/A filter on "$col" has excluded the values:.*?\.\s+(<br>)?//s;
		$cgi->param('annotate',$annotate);
	}
}

#
# Setup selector filter used to check/uncheck individual records.
#

our $sel = new Acme::Lister::Selector
		selector	=> $selector,
		cond		=> $selector_cond,
		param1		=> $selector_param1,
		param2		=> $selector_param2,
		param3		=> $selector_param3,
		table		=> $table,
		sandbox		=> 'list';

($selector,$selector_cond,$selector_param1,$selector_param2,$selector_param3) = $sel->selector;

#
# Restrict total/group columns to those that are displayed.
#

if (@columns && (@total_columns || @group_columns)) {
	my %columns = map { $_, 1 } @columns;	# make hash for quick lookups

	@total_columns  = grep { $columns{$_} } @total_columns;
	@group_columns  = grep { $columns{$_} } @group_columns;
}

#
# Save list settings
#

if ($list_settings && $list_id ne 'list') {

	#
	# Save only the non-standard (non-"list") user settings.
	# Standard settings come from the main user option settings.

	# Note: use of "list" as a list_id is experimental. Currently it is not
	# being used. (12/1/02 mp)

	$list_settings->set('ListCustomize',$customize || '');

	unless ($nocustomize) {
		$list_settings->set('ListCustomize',''); # never set 'previous'
		$list_settings->set('ListAsRows',$list_as_rows || 'no');
		$list_settings->set('ListAsTable',$list_as_table || 'no');
		$list_settings->set('ListAllColumns',$list_all_columns || 'no');
		$list_settings->set('ListColumnHeaders',$list_column_headers);
		$list_settings->set('ListFormatStyle',$list_format_style);
		$list_settings->set('ListColumns',(join ',', @columns))
		  if ! $list_as_rows;
		$list_settings->set('ListSortColumns',(join ',', @sort_columns));
		$list_settings->set('ListTotalColumns',(join ',', @total_columns));
		$list_settings->set('ListGroupColumns',(join ',', @group_columns));
		$list_settings->set('ListFilterColumns',(join ',', @filter_columns));

		# Don't save selector settings. User sets selector each session.
		#$list_settings->set('ListSelector',$selector);
		#$list_settings->set('ListSelectorCond',$selector_cond);
		#$list_settings->set('ListSelectorParam1',$selector_param1);
		#$list_settings->set('ListSelectorParam2',$selector_param2);
		#$list_settings->set('ListSelectorParam3',$selector_param3);
	}

	$list_settings->save;
}

#
# Reset the columns if not ListAsRows option.
#

if ($list_as_table && ! $list_all_columns && $list_settings && $list_id ne 'SQL'
&& (my $col = $list_settings->get('ListColumns')) ) {
	@columns = split(/,/,$col);
}

#
# Save history if only a sql query was specified. (see admin.cgi)
#

if ($sql && ! $search && ! $table) {

	if ((my $sh = $cgi->param('save_history')) && ! $redrawing) {

		#
		# Save sql query in history list in user options file.
		#

		$list_id ||= 'SQL';
		if ($sh =~ /yes/i) {
			$opt->enter("${list_id}Query",$sql);
		}

		$cgi->delete('save_history');
	}

	$list_title ||= $sql;

	#
	# Get limits
	#

	my $limit = Acme::Config->SQL_Limit_Abort || Acme::Config->List_Limit_Abort;
	$limit_abort = $limit if ! $limit_abort || $limit < $limit_abort;

	$limit_warn ||= Acme::Config->SQL_Limit_Warn;
}

#
# Initialize accumulators
#

our $n_rows        = 0;
our $n_checked     = 0;
our $n_unchecked   = 0;

# Make hash of all the include/exclude row id's.
our %include = map { s/^(['"])(.*)\1$/$2/; ($_, 1) } split(/,/, $include);
our %exclude = map { s/^(['"])(.*)\1$/$2/; ($_, 1) } split(/,/, $exclude);

#
# Set flags for errors
#

our $nullkey = 0;

# Define a special footer format string that is a degenerate of $format
# used in table lists.
our $footer_format = '';

#
# Construct SQL from table schema
#

our $export_security_code = '';
our $email_security_code  = '';

if ($table) {
	
	$t = $schema->t($table)
	  or die "Cannot find a database table named '$table'\n";

	#
	# Get list_id to use from table name or use default for all tables.
	#

	$list_id ||= $opt->get("${table}ListID") || $table;

	$key ||= $t->key;

	$edit_prog     ||= $t->edit_prog;
	$view_prog     ||= $t->view_prog;

	$logall_prog   ||= $t->logall_prog;
	$export_prog   ||= $t->export_prog;
	$emailall_prog ||= $t->emailall_prog;
	$editall_prog  ||= $t->editall_prog;

	$export_security_code = $t->export_security_code;
	$email_security_code  = $t->email_security_code;

	if ($list_all_columns) {
		$list_fields = join ",\n", available_columns();

		$format = '';	# define as empty to force auto-columns below
	}

	#
	# Set defaults from schema properties
	#

	$list_fields  ||= $t->list_fields;

	if (! $format) {
		$format = $t->list_format;
		$format =~ s/^\s*(["'])(.*)\1\s*$/$2/;	# remove possible quotes
	}

	$title             ||= $t->title;
	$list_title        ||= $t->list_title unless $search;

	$record_set_name   ||= $t->title;

	$id_field	   ||= fully_qualify($table, $t->id_field);
	$code_field	   ||= fully_qualify($table, $t->code_field);
	$name_field	   ||= fully_qualify($table, $t->name_field);
	$first_name_field  ||= fully_qualify($table, $t->first_name_field);
	$last_name_field   ||= fully_qualify($table, $t->last_name_field);
	$phone_field       ||= fully_qualify($table, $t->phone_field);
	$email_field       ||= fully_qualify($table, $t->email_field);

	$code_field_match  ||= $t->code_field_match;
	$phone_field_match ||= $t->phone_field_match;

	#
	# Get limits
	#

	my $limit = $t->limit_abort || Acme::Config->List_Limit_Abort;
	$limit_abort = $limit if ! $limit_abort || $limit < $limit_abort;

	$limit_warn ||= $t->limit_warn || Acme::Config->List_Limit_Warn;
	$limit_alert .= $t->limit_alert;

	#
	# Create default sql to search the table if needed.
	#

	if (! $sql) {

		# Note: this sql should match the default sql in export.cgi.

		$sql ||= $t->list_sql if $t->defined('list_sql');

		if (! $sql) {

			# Construct query here if not explicity defined in schema.
			# This allows customized list of fields such as the
			# ListAllColumns user option above.

			my ($join_expr,$join_clause);

			if ($list_as_rows) {
				# list as rows: use defined list_join property.
				($join_expr,$join_clause) = parse_join($table,$t->list_join);
			}
			else {
				# build custom list of table joins
				($join_expr,$join_clause) = parse_join($table,
				  (join " and ", grep {$_} $t->join, $t->master, $t->list_join));
			}

			$sql = 'select'
			  . ' distinct ' . $list_fields
			  . ' from '	 . $join_expr
			  . ' where '	 . $t->list_where . $join_clause
			  . ' order by ' . $t->list_order_by_clause;
		}
	}

	#
	# Fill in default list customization parameters.
	#

	@columns	= split(/\s*,\s*/, $list_fields)
	  if $customize eq 'none' || $list_as_rows || ! @columns || $list_all_columns;
	@sort_columns	= split(/\s*,\s*/, $t->list_order_by_clause)
	  if $nocustomize || (($list_as_rows || ! @sort_columns) && ! $redrawing);
	@total_columns	= split(/\s*,\s*/, $t->list_total_fields)
	  if $nocustomize || (($list_as_rows || ! @total_columns) && ! $redrawing);
	@group_columns	= split(/\s*,\s*/, $t->list_group_fields)
	  if $nocustomize || (($list_as_rows || ! @group_columns) && ! $redrawing);
	@filter_desc	= split(/\s*,\s*/, $t->list_filter_desc)
	  if $nocustomize || (($list_as_rows || ! @filter_columns) && ! $redrawing);
	  # coupled with filter_columns
	@filter_columns	= split(/\s*,\s*/, $t->list_filter_fields)
	  if $nocustomize || (($list_as_rows || ! @filter_columns) && ! $redrawing);
}

die "Empty search\n" unless $sql;

#
# Parse SQL statement
#

our $sqlo = new Acme::SQL $sql;

#
# Change SQL according to the list customization parameters.
#

# Set the columns to display.

if (@columns && ! $list_as_rows && ! $list_all_columns) {
	$sqlo->set_columns(@columns);

	# Automatically add join clause if columns reference a table
	# and the primary table's schema joins the referenced table.

	ADD_JOIN_TABLE: {
		my $primary_table = $sqlo->primary_table;
		my $pt = $schema->t($primary_table)		or last;
		my $join_clause = $pt->join			or last;

		my $j = new Acme::Base::Schema::Join $primary_table, $join_clause
		  or last;

		my ($tbl,$fld);

		for my $col ($sqlo->dbcolumns) {

			($tbl,$fld) = $col =~ /([^(]*?)\.(.*)/;	# no functions

			# Note: unqualified columns are assumed to be with
			# the primary table which doesn't need to be checked
			# here.

			next unless $tbl;

			next if $sqlo->is_table($tbl);
			next unless Acme::Security::allow('Access',$tbl);

			$sqlo->add_join($j->join_expr($tbl));
		}
	}

	$sql = $sqlo->as_sql;		# redefine
	$list_fields = join ",\n", @columns;	# redefine
}

#
# Make sure key field is selected in SQL for use in links
#

my $col = $table ? "$table.$key" : $key;

my $added_columns = 0;			# flag
if ($col && ! $sqlo->is_column($col)) {
	$sqlo->add_column($col);
	$added_columns++;		# set to true
}

#
# Make sure sort columns are listed the select clause.
#

# Note: normally this is/should not be neccessary. However, to ensure
# our queries return only one record per key value, we construct the
# sql as "select distinct ...". Postgres will allow sorts to be
# performed on a non-listed column only when the "distinct" keyword 
# is not used.

for my $col (@sort_columns, @filter_columns, @filter_desc) {

	# Remove the "desc" indicator of sort columns and
	# remove the "null" option of filter columns.

	($_ = $col) =~ s/ \w+$//;

	$sqlo->add_column($_), $added_columns++ unless $sqlo->is_column($_);
}

$sqlo->add_column($selector), $added_columns++
  if $selector ne '(all)' && $selector ne '(none)' && ! $sqlo->is_column($selector);

#
# Set SQL limits
#

$sqlo->set_limit($limit_abort) if $limit_abort;

#
# Evaluate Search script
#

if ($search) {

	if ((my $sh = $cgi->param('save_history')) && ! $redrawing) {

		#
		# Save search in history list in user options file.
		#

		if ($sh =~ /yes/i && $table) {
			$opt->enter("${table}Search",$search);
		}

		$cgi->delete('save_history');
	}

	if ($id_field || $code_field || $name_field || $last_name_field || $phone_field || $email_field) {

		#
		# Preprocess search script
		#

		if ($first_name_field || $last_name_field) {
			$search = Acme::Searcher::preprocess($search, {
			  RecordSetName   => $record_set_name,
			  IDField	  => $id_field,
			  CodeField	  => $code_field,
			  NameField       => $name_field,
			  FirstNameField  => $first_name_field,
			  LastNameField   => $last_name_field,
			  PhoneField      => $phone_field,
			  EmailField      => $email_field,
			  CodeFieldMatch  => $code_field_match,
			  PhoneFieldMatch => $phone_field_match,
			});
		}
		else {
			$search = Acme::Searcher::preprocess($search, {
			  RecordSetName   => $record_set_name,
			  IDField	  => $id_field,
			  CodeField	  => $code_field,
			  NameField	  => $name_field,
			  PhoneField      => $phone_field,
			  EmailField      => $email_field,
			  CodeFieldMatch  => $code_field_match,
			  PhoneFieldMatch => $phone_field_match,
			});
		}
	}

	#
	# Combine sql with search script.
	#

	my $srch = new Acme::Searcher $search, $cgi;	# parse search script
	$sql = $srch->as_sql($sqlo);	

	#
	# Redefine title if defined in script.
	#

	# Note: Because the Searcher gets values for its title and annotation
	# from the cgi if they are not set in the script, be careful not to
	# reset alterations done above.

	$list_title ||= $srch->title    if $srch->title    && ! $redrawing;
	$annotate   ||= $srch->annotate if $srch->annotate && ! $redrawing;
}
elsif ($table) {
	$sql = $sqlo->as_sql;
}

#
# Set optional sort order
#

$sort = join ',', @sort_columns;	# redefine sort parameter

if ($sort) {

	# Parse sql again. It is more efficient for the user to put an
	# "order by" clause in the search script, but it's not convenient.
	# We could wait until the parsing of sql for column names below,
	# but that would be changing the sql after it has been "cleaned"
	# by security.

	$sqlo = new Acme::SQL $sql;
	$sqlo->sort($sort);
	$sql = $sqlo->as_sql;
}

#
# Set up list
#

$sql = Acme::Security::clean_sql($sql);

our $l = new Acme::Lister (
	sql          => $sql,
	sandbox      => 'list',
	total        => \@total_columns,
	);

if ($l->count == 0) {
	display_message("No records were found.", "$title",undef,0);
	exit 1;
}

#
# Set final default values
#

$sqlo = new Acme::SQL $sql;		 # reparse after clean_sql() expands *
our @field_names = $sqlo->columns;
pop @field_names while $added_columns--; # remove key field added to sql above

if ($list_as_rows
|| (! $table && ! exists $cgi->{'columns'} && ! ($customize eq 'previous' && $list_settings))) {
	@columns = @field_names;	# reassign
	$list_fields = join ",\n", @columns;	# redefine
}

$title = "SQL List" if (! $title && $sql);
$title ||= "List";			# final default

#
# Define global variables for list helper functions
#

our $b = new Acme::HTML::NavButtons print => $Printing;

our $n_rec = 0;				# record number (for odd/even rows)

our $lsn = $list_format_style;		# default CSS "list style" name

#
# Initialize filters and accumulator hashes
#

# Set up a hash with field names as keys which point to a hash with
# field values as a key and the count of records as its value.

# This is complicated by the filter columns possibly having a "null" keyword
# appended to the column name, and needing to change the periods in the
# column name to double underlines __.

our @filter_fields 	= @filter_columns;
s/\./__/g for @filter_fields;

our %is_null_filter = map { s/ null$//i; $_, 1 } grep { / null$/i } @filter_fields;
s/ null$//i for @filter_fields;		# remove null keyword

our @filter_fields_desc	= @filter_desc;
s/\./__/g for @filter_fields_desc;

our %filter      = map { $_, {} } @filter_fields;
our %filter_desc = map { $_, {} } @filter_fields;	# mirrors %filter

# Make lookup hash to find variable name of description corresponding to
# the filter's column name.

my $i=0; our %filter_desc_of = map { $_, $filter_fields_desc[$i++] } @filter_fields;

#
# Initialize group totals
#

###
# Move to Acme::Lister ???
#our %group_totals;			# initialized before group breaks

#
# Initialize limits
#

# Note: No limits are applied if we are just counting records. We might want
# to still apply the $limit_abort limit, or instead of terminating the count,
# use it as a limit to issue a warning alert to the administrators.

my $limits;
if (($limit_abort || $limit_warn) && ! $counting) {
	$limits = {};
	$limits->{$limit_abort} = \&limit_abort_func	if $limit_abort;
	$limits->{$limit_warn}  = \&limit_warn_func	if $limit_warn;
}

#
# Display list
#

display_header();

$l->print(
	before_list  => \&display_list_header,
	after_list   => \&display_list_footer,
	every_row    => \&display_list_item,
	limits       => $limits,
	) if $search || ($sql && ! $table);

display_footer();

exit 0;

#------------------------------------------------------------------------------
#
# Get search parameter
#
#------------------------------------------------------------------------------

sub get_search {

	our $schema = new Acme::Base::Schema;

	my ($table_title, $table_html);
	my ($search_title, $search_fields_help, $query_by_form);

	if ($table) {

		#
		# Save table value in a hidden field 
		#

		$table_title = '';
		$search_title = 'Search';

		$search_fields_help = <<End_Of_HTML;
<a href="/cgi-bin/schema.cgi?table=$table" onClick="new_target(this,'doc')" class="link">Columns</a>
End_Of_HTML

		my $query = $schema->t($table)->query_prog;

		# Append table and report name to query url
		$query .= ($query =~ /\?/ ? '&' : '?') . "table=$table";
		$query .= "&report=$report" if $report;

		$query_by_form = <<End_Of_HTML;
<a href="${query}" class="link">Construct&nbsp;Search</a><br>
End_Of_HTML

		$table_html = <<End_Of_HTML;
      <input type="hidden" name="table" value="$table">
End_Of_HTML
	}
	else {
		#
		# Show drop-down list of all tables
		#

		$table_title = 'Table';
		$search_title = 'Search';
		$query_by_form = '';

		$search_fields_help = <<End_Of_HTML;
<a href="/cgi-bin/schema.cgi" onClick="new_target(this,'doc')" class="link">Columns</a>
End_Of_HTML

		$table_html = <<End_Of_HTML;
      <tr valign="top"> 
        <td width="105" align="right" nowrap><font class="plain">$table_title</font>
        </td>

        <td width="340"> 
          <select name="table" style="width:100%">
End_Of_HTML

		for ($schema->tables) {
			$table_html .= qq[            <option value="$_">$_</option>\n];
		}

		$table_html .= <<End_Of_HTML;
          </select>
	  </td>
	</tr>
End_Of_HTML
	}

	my $search_instructions = <<End_Of_HTML;
<a href="${Help}/search_instructions.htm" onClick="new_target(this,'help')" class="link">Instructions</a><br>
End_Of_HTML

	my $encoded_search = escape_html($search);

	my $search_html = <<End_Of_HTML;
    <table border="0" cellspacing="0" cellpadding="2">
$table_html
      <tr valign="top"> 
        <td width="105" align="right" nowrap><font class="plain">$search_title</font>

	  $query_by_form
          $search_instructions
          $search_fields_help
        </td>

        <td width="340"> 
          <textarea name="search" wrap="virtual" rows="5" style="width:100%">$encoded_search</textarea>
        </td>
      </tr>
    </table>
End_Of_HTML

	my $html = <<End_Of_HTML;
<table border="0" cellspacing="0" cellpadding="2" style="margin-top:2ex">
  <tr valign="top"> 
    <td width="16">
    </td>

    <td width="445" align="right">
    $search_html
    </td>

    <td width="105"> 

      <table border="0" cellspacing="0" cellpadding="1">
        <tr> 
          <td> 
            <button class="control" accesskey="S" onClick="if(this.form.search.value.length==0)this.form.search.value='*';this.form.submit()"><u>S</u>earch</button>
          </td>
        </tr>
        <tr> 
          <td>
            <button class="control" accesskey="C" onClick="self.close()"><u>C</u>ancel</button>
          </td>
        </tr>
        <tr> 
          <td>
          </td>
        </tr>
        <tr>
          <td>
          </td>
        </tr>
      </table>

    </td>

    <td width="16"><img src="$Images/blank.gif"></td>
  </tr>
</table>
End_Of_HTML

	$title = $table
	       ? $schema->t($table)->title . " Search Parameter"
	       : "List a Table";

	print $cgi->header(-type=>'text/html') unless @ARGV;

	Acme::HTML::Parameters->get_parameters(
		window_title => 'Enter Search Parameter',
		title        => "$title",
		method       => 'POST',
		submit       => undef,	# Submit button text
		reset        => undef,	# Reset button text
		exclude      => ['search','table'],
		cgi          => $cgi,
		html         => $html,
		);
}

#------------------------------------------------------------------------------
#
# Get show list of list files.
#
#------------------------------------------------------------------------------

sub get_list_file {

	#
	# Show list of lists.
	#

	our $Base        = Acme::Config->Base;
	our $Browse_Prog = Acme::Config->Browse_Prog;

	#our $List_Prog   = Acme::Config->List_Prog;
	our $List_Prog   = $cgi->url();		# call ourselves again

	# Note: more than one directory can be listed by listing them in @dirs.

	my @dirs;

	if ($list_id && Acme::Security::check('Access', $list_id)) {
		#@dirs = ($list_id,'/');
		@dirs = ($list_id);	# no common directory for all users
	}
	
	#
	# Redirect to the file browser program.
	#

	my $lister = "$List_Prog"
		. ($List_Prog =~ /\?/ ? '&' : '?')
		. 'list=$';

	my $url = $Base
		. (substr($Browse_Prog,0,1) ne '/' ? '/' : '')
		. $Browse_Prog
		. "?lists="
		. (join '&lists=', @dirs)
		. "&link=" . escape($lister)
		. "&title=Lists";

	$url .= "&table=$table"     if $table;
	$url .= "&list_id=$list_id" if $list_id && $list_id ne $table;

	print $cgi->redirect($url);

	exit;
}

#------------------------------------------------------------------------------
#
# Document header and footer functions
#
#------------------------------------------------------------------------------

sub display_header {

	my $aux_script = $t->list_include_html if $t;
	$aux_script ||= '';
	$aux_script &&= "\n$aux_script";

	# Append user option setting for highlighting rows.
	$aux_script .= qq[
<script language="JavaScript">nohilo=true</script>]
	  if $nohilo;

	my $aux_stylesheet = $t->list_stylesheet_html if $table;
	$aux_stylesheet ||= '';
	$aux_stylesheet &&= "\n$aux_stylesheet";

	my $print_stylesheet = '';
	my $onload_cmd = '';

	if ($Printing) {
  		$onload_cmd = 
  qq[ onLoad="window.print();window.setTimeout('window.close()',1500)"];

		$print_stylesheet =
  qq[<link rel="stylesheet" href="$Styles/list_print.css" type="text/css">\n];

		my $aux_print_stylesheet = $t->list_print_stylesheet_html if $table;
		$aux_print_stylesheet  &&= "$aux_print_stylesheet\n";

		$print_stylesheet .= $aux_print_stylesheet;
  	}

	my $title_type	= $report_id
			? 'report'
			: $request =~ /Count/i
			? 'count'
			: 'list';

# Keep IE6 in IE5 compatablilty mode by not specifying url
#<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
#"http://www.w3.org/TR/html4/loose.dtd">
# or
#"http://www.w3.org/TR/1999/REC-html401-19991224/loose.dtd">

print <<End_Of_HTML;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">

<html>
<head>
<title>$title_type $title</title>

<meta http-equiv="Content-Type" content="text/html; iso-8859-1">
<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<link rel="stylesheet" href="$Styles/button_bar.css" type="text/css">
<link rel="stylesheet" href="$Styles/list.css" type="text/css">$aux_stylesheet
$print_stylesheet
<script src="$Scripts_Lib/menu.js" type="text/javascript"></script>
<script src="$Scripts_Lib/param.js" type="text/javascript"></script>
<script src="$Scripts_Lib/list.js" type="text/javascript"></script>$aux_script
</head>
<body$onload_cmd>
End_Of_HTML
print $b->first_buttons;
print common_header($title,$Printing);

my $annotation = $annotate;
$annotation &&= qq[<span class="annotation">$annotation</span>];

print <<End_Of_HTML if $list_title;
<div class="title">$list_title$annotation</div>
End_Of_HTML

print button_bar();

print <<End_Of_HTML;

<!-- Begin List -->
<div class="list">
End_Of_HTML

print qq[<div class="list-rows">\n]  if $list_as_rows;
print qq[<div class="list-table">\n] if $list_as_table;
}

#------------------------------------------------------------------------------

sub display_footer {

print "</div>\n";
print "</div>\n";
print common_footer($title,$Printing);
print $b->last_buttons;

if ($nullkey) {
	die "There are $nullkey records that do not have a key value.\n"
	  . "Every record must have a unique value for the '$key' column.\n"
	  . "The operation of the check selectors on each row is not valid.\n"
	  . "Error";
}

print "</body>\n</html>";
}

#------------------------------------------------------------------------------
#
# Button bar functions
#
#------------------------------------------------------------------------------

sub button_bar {

return '' if $Printing;

my $list_prog = $cgi->url();	# reference to self

#
# Build form of list parameters.
#

my $html = <<End_Of_HTML;
<form	name="list_control"
action="$list_prog"
method="POST"
style="margin:0px"
>
End_Of_HTML

# Pass program arguments on to export routines in hidden fields.

$html .= Acme::HTML::Parameters::save_all_param_except(
					'request',
					'redrawing',
					'save_history',
					'list_title',
					'annotate',
					'list_id',
					'report_id',
					'table',
					'key',
					'limit',
					'email_field',
					'name_field',
					'sql',
					'search',
					'original_search',
					'exclude',
					'include',
					'invert_all',
					'select_all',
					'list_as_rows',
					'list_as_table',
					'list_all_columns',
					'list_column_headers',
					'list_format_style',
					'customize',
					'columns',
					'sort_columns',
					'total_columns',
					'group_columns',
					'filter_columns',
					'filter_desc',
					'available_columns',
					'available_sort_columns',
					'available_total_columns',
					'available_group_columns',
					'available_filter_columns',
					'selector',
					'selector_cond',
					'selector_desc',
					'selector_param1',
					'selector_param2',
					'selector_param3',
					'n_rows',
					'n_checked',
					'n_unchecked');

my $encoded_original_sql    = escape_html($original_sql);
my $encoded_original_search = escape_html($original_search);
my $encoded_include         = escape_html($include);
my $encoded_exclude         = escape_html($exclude);

# Save computed list parameters in hidden fields
$html .= <<End_Of_HTML;
<!-- List control settings -->
<input type="hidden" name="request" value="$request">
<input type="hidden" name="redrawing" value="yes">

<input type="hidden" name="list_id" value="$list_id">
<input type="hidden" name="report_id" value="$report_id">
<input type="hidden" name="table" value="$table">
<input type="hidden" name="key" value="$key">
<input type="hidden" name="limit" value="$limit_abort">

<input type="hidden" name="email_field" value="$email_field">
<input type="hidden" name="name_field" value="$name_field">

<textarea name="sql" wrap="virtual" style="display:none">$encoded_original_sql</textarea>
<textarea name="original_search" wrap="virtual" style="display:none">$encoded_original_search</textarea>

<textarea name="exclude" wrap="virtual" style="display:none">$encoded_exclude</textarea>
<textarea name="include" wrap="virtual" style="display:none">$encoded_include</textarea>

<input type="hidden" name="invert_all" value="$invert_all">
<input type="hidden" name="select_all" value="$select_all">

<!-- List option settings -->
<input type="hidden" name="list_as_rows" value="$list_as_rows">
<input type="hidden" name="list_as_table" value="$list_as_table">
<input type="hidden" name="list_all_columns" value="$list_all_columns">
<input type="hidden" name="list_column_headers" value="$list_column_headers">
<input type="hidden" name="list_format_style" value="$list_format_style">

<!-- Customize settings -->
<input type="hidden" name="customize" value="$customize">

<!-- List tallies -->
<input type="hidden" name="n_rows" value="">
<input type="hidden" name="n_checked" value="">
<input type="hidden" name="n_unchecked" value="">
End_Of_HTML

if (@filter_desc) {
	$html .= qq[\n<!-- Filter Description Columns -->];
	$html .= qq[\n<input type="hidden" name="filter_desc" value="]
	       . (join ",", @filter_desc)
	       . qq[">\n];
}

#
# Construct button bar
#

$html .= <<End_Of_HTML;

<!-- Start Button Bar -->
<div class="button-bar">
<table cellspacing="0" cellpadding="0">
<tr>

<td class="button-bar-menu">
<a id="bb_menu_head"
class="bb-menu-header"
href="javascript:menu('bb_menu','bb_menu_head')"
>&bull;</a>
</td>

End_Of_HTML

$html .= <<End_Of_HTML if $key;
<td><a href="javascript:do_open('$list_id')" title="Open a saved list file.">Open</a></td>
<td><a href="javascript:do_save('$list_id')" title="Save list in a file.">Save</a></td>
<td><a href="javascript:do_compare('$list_id')" title="Compare two list files.">Compare</a></td>
End_Of_HTML

###	if (($table && $t->log)
###	&& Acme::Security::allow('Access',program_of($logall_prog))) {
###		$html .= <<End_Of_HTML;
### <td><a href="javascript:do_log('$logall_prog')" title="Attach a log record to selected list items.">Log</a></td>
###End_Of_HTML
###	}

if (Acme::Security::allow('Access',program_of($export_prog))
&& (!$export_security_code || Acme::Security::allow('Access',$export_security_code))
&& ( (!$table && Acme::Security::allow('Access','SQL_EXPORT'))
  || ($table  && $t->export && Acme::Security::allow('Access',$table))
   ) ) {
	$html .= <<End_Of_HTML;
<td><a href="javascript:do_export('$export_prog')" title="Export list to a file.">Export</a></td>
End_Of_HTML
}
if ($email_field && Acme::Security::allow('Access',program_of($emailall_prog))
&& (!$email_security_code || Acme::Security::allow('Access',$email_security_code))
&& (!$table || $t->email) ) {
	$html .= <<End_Of_HTML;
<td><a href="javascript:do_email('$emailall_prog','$email_field')" title="Send an email message to the names in the list.">Email</a></td>
End_Of_HTML
}

###	if (Acme::Security::allow('Access',program_of($editall_prog))
###	&& ((!$table && $editall_prog)
###	  || ($table && Acme::Security::allow('Update',$table)))) {
###		$html .= <<End_Of_HTML;
### <td><a href="javascript:do_edit('$editall_prog')" title="Edit selected list items.">Edit</a></td>
###End_Of_HTML
###	}

#
# Add custom buttons defined in the schema
#

$html .= $t->list_buttons_html if $t;	

#
# Finish off button bar with selector (check mark) operators.
#

$html .= <<End_Of_HTML if $table || $key;

<td><a href="javascript:invert_all()" title="Change check marks to unchecked and vica versa.">Invert</a></td>
<td><a href="javascript:select_all()" title="Change all check marks to checked.">Select&nbsp;All</a></td>
End_Of_HTML

$html .= <<End_Of_HTML;
</tr>
</table>
</div>
End_Of_HTML

	#
	# Define button bar menu inserts
	#

	# Define menu options' check-mark variables.

	my $check=qq[ <img src="$Images/check.gif" border="0">];

	my $check_list_as_rows     = $list_as_rows     && ! $counting	? $check : '';
	my $check_list_as_table	   = $list_as_table    && ! $counting	? $check : '';
	my $check_list_all_columns = $list_all_columns && ! $counting	? $check : '';
	my $check_list_as_count    = $request =~ /Count/i	? $check : '';

	my $check_none	   = $list_column_headers eq 'none'		   ? $check : '';
	my $check_name	   = $list_column_headers eq 'FieldNames'	   ? $check : '';
	my $check_fullname = $list_column_headers eq 'QualifiedFieldNames' ? $check : '';
	my $check_title	   = $list_column_headers eq 'FieldTitles'	   ? $check : '';

	$check_none = $check_name = $check_fullname = $check_title = ''
	  if $list_as_rows || $counting;

	# Make menu items on Style submenu

	my $list_format_menu     = Acme::Config->List_Format_Menu;
	my $list_format_prefixes = Acme::Config->List_Format_Prefixes;

	my $format_menu_items = '';
	my ($i, $m, $p, $c);
	for ($i = 0; $i < @$list_format_menu; $i++) {

		#
		# Set check mark on menu item if it is selected
		#

		$m = $$list_format_menu[$i];
		$p = $$list_format_prefixes[$i];
		$c = $$list_format_prefixes[$i] eq $list_format_style
		     ? $check
		     : '';
		$format_menu_items .= qq[
<a href="javascript:format_style('$p')">$m$c</a>];
	}

	# Make menu items on Customize submenu

	my $customize_option = $customize eq 'none'
	  ?  qq[<a href="javascript:customize('previous')">previous</a>]
	  :  qq[<a href="javascript:customize('none')">none</a>];

	#
	# Make array of available columns for parameter dialog menus below.
	#

	# Note: This list of available columns is incomplete. It should include
	# those columns in the Master table, all Join tables, and all Choose
	# and Search tables' columns. However, join clauses for the joined 
	# tables needs to be propogated to the SQL.
	# If a table is not being listed, then only the fields in the SQL
	# script are available.

	my @available_columns = available_columns() if $table;
	@available_columns = $sqlo->columns unless @available_columns;

	#
	# Construct button bar menu
	#

	$html .= <<End_Of_HTML;

<!-- Button Bar Menu -->
<div id="bb_menu" class="bb-menu">
<a href="javascript:list_as_rows()">List Rows$check_list_as_rows</a>
<a href="javascript:list_as_table()">List Table$check_list_as_table</a>
<a href="javascript:list_all_columns()">List All Fields$check_list_all_columns</a>
<a href="javascript:list_as_count()">Count$check_list_as_count</a>
<hr> 
<a href="javascript:submenu('bb_menu_customize')" class="bb-submenu-header">Customize</a>
<a href="javascript:submenu('bb_param_search')" class="bb-param-header">Search</a>
<a href="javascript:submit_param()">Redraw</a>
<hr>
<a href="javascript:do_open_report('$list_id')">Open Report</a>
<a href="javascript:do_save_report('$list_id')">Save as Report</a>
<hr>
<a href="javascript:do_print()">Print</a>
<a href="javascript:self.close()">Exit</a>
</div>
<!-- End Button Bar Menu -->

<script language="JavaScript">make_menu('bb_menu','bb_menu_head').SubmenuOffsetX=-4;</script>
End_Of_HTML

	#
	# Define Customize submenu
	#

	$html .= <<End_Of_HTML;

<!-- Button Bar Customize Submenu -->
<div id="bb_menu_customize" class="bb-submenu">
$customize_option
<a href="javascript:submenu('bb_menu_format')" class="bb-submenu-header">Style</a>
<a href="javascript:submenu('bb_menu_headers')" class="bb-submenu-header">Headers</a>
<a href="javascript:submenu('bb_param_title')" class="bb-param-header">Title</a>
<a href="javascript:submenu('bb_param_columns')" class="bb-param-header">Columns</a>
<a href="javascript:submenu('bb_param_sort')" class="bb-param-header">Sort</a>
<a href="javascript:submenu('bb_param_totals')" class="bb-param-header">Totals</a>
End_Of_HTML

	$html .= <<End_Of_HTML	if 0;		## not implemented yet ##
<a href="javascript:submenu('bb_param_groups')" class="bb-param-header">Groups</a>
End_Of_HTML

	$html .= <<End_Of_HTML;
<a href="javascript:submenu('bb_param_filters')" class="bb-param-header">Filters</a>
End_Of_HTML

	$html .= <<End_Of_HTML if $table || $key;
<a href="javascript:submenu('bb_param_selector')" class="bb-param-header">Selector</a>
End_Of_HTML

	$html .= <<End_Of_HTML;
</div>
<!-- End Button Bar Customize Submenu -->
End_Of_HTML

	#
	# Define Search parameter form
	#

	my $encoded_search = escape_html($search);

	$html .= <<End_Of_HTML;

<!-- Begin Search Parameter -->
<div id="bb_param_search" class="bb-param">

<h3>
Enter Search Script
</h3>

<table>
<tr>
<td colspan="3"><textarea name="search" wrap="virtual" cols="60" rows="7" onChange="search_reset()">$encoded_search</textarea>
</td>
</tr>

<tr>
<td width="33%"><a href="${Help}/search_instructions.htm" onClick="new_target(this,'help')" class="link">Instructions</a></td>
<td align="center" width="33%">
<input type="button" value="Redraw" class="submit-button" onClick="submit_param()">
</td>
<td width="33%"></td>
</tr>
</table>
</div>
<!-- End Search Parameter -->
End_Of_HTML

	#
	# Define Format submenu
	#

	$html .= <<End_Of_HTML;

<!-- Button Bar Style Submenu -->
<div id="bb_menu_format" class="bb-submenu">$format_menu_items
</div>
<!-- End Button Bar Style Submenu -->
End_Of_HTML

	#
	# Define Headers submenu
	#

	$html .= <<End_Of_HTML;

<!-- Button Bar Headers Submenu -->
<div id="bb_menu_headers" class="bb-submenu">
<a href="javascript:column_headers('none')">none$check_none</a>
<a href="javascript:column_headers('FieldNames')">Field Names$check_name</a>
<a href="javascript:column_headers('QualifiedFieldNames')">Field Full Names$check_fullname</a>
<a href="javascript:column_headers('FieldTitles')">Field Titles$check_title</a>
</div>
<!-- End Button Bar Headers Submenu -->

End_Of_HTML


	#
	# Define Title parameter form
	#

	my $encoded_list_title = escape_html($list_title);
	my $encoded_annotate = escape_html($annotate);

	$html .= <<End_Of_HTML;

<!-- Begin Title Parameter -->
<div id="bb_param_title" class="bb-param">

<h3>
Enter List Title
</h3>

<table>
<tr>
<th align="right">Title</th>
<td><input type="text" name="list_title" value="$encoded_list_title" size="50"></td>
</tr>

<tr>
<th align="right" valign="top">Annotation</th>
<td><textarea name="annotate" wrap="virtual" cols="50" rows="4">$encoded_annotate</textarea>
</td>
</tr>

<tr>
<td colspan="2" align="center">
<input type="button" value="Redraw" class="submit-button" onClick="submit_param()">
</td>
</tr>
</table>
</div>
<!-- End Title Parameter -->
End_Of_HTML

	#
	# Define Display Columns parameter form
	#

	$html .= <<End_Of_HTML;

<!-- Begin Display Columns Parameter -->
<div id="bb_param_columns" class="bb-param">

<h3>
Choose Display Columns
</h3>

<table>
<tr>
<th width="45%">Available Columns</th>
<th width="5%"></th>
<th width="45%">Display Columns</th>
<th width="5%"></th>
</tr>

<tr>
<td width="45%"><select size="9" width="50" name="available_columns" multiple onDblClick="addSrcToDestList(this,this.form.columns,true,'')">
End_Of_HTML

	# Fill in available fields for display
	$html .= qq[<option value="$_">$_</option>\n] for @available_columns;

	$html .= <<End_Of_HTML;
</select></td>

<td width="5%" align="center"><input type="button" value="&rarr;" class="left-right-button" onClick="addSrcToDestList(this.form.available_columns,this.form.columns)"><br><br><input type="button" value="&larr;" class="left-right-button" onClick="deleteFromDestList(this.form.columns,this.form.available_columns);this.form.list_all_columns.value='no'"></td>

<td width="45%"><select size="9" name="columns" multiple onDblClick="deleteFromDestList(this,this.form.available_columns);this.form.list_all_columns.value='no'">
End_Of_HTML

	# Fill in displayed columns.
	$html .= qq[<option value="$_">$_</option>\n] for @columns;

	$html .= <<End_Of_HTML;
</select></td>

<td width="5%" align="center"><input type="button" value="&uarr;" class="up-down-button" onClick="moveUpList(this.form.columns)"><br><br><input type="button" value="&darr;" class="up-down-button" onClick="moveDownList(this.form.columns)"></td>
</tr>

<tr>
<td colspan="3" align="center">
<input type="button" value="Redraw" class="submit-button" onClick="this.form.list_as_rows.value='no';this.form.list_as_table.value='yes';submit_param()">
</td>
<td>
</td>
</tr>
</table>
</div>
<!-- End Display Columns Parameter -->
End_Of_HTML

	#
	# Define Sort Columns parameter form
	#

	$html .= <<End_Of_HTML;

<!-- Begin Sort Columns Parameter -->
<div id="bb_param_sort" class="bb-param">

<h3>
Choose Sort Columns
</h3>

<table>
<tr>
<th width="45%">Available Columns</th>
<th width="5%"></th>
<th width="45%">Sort by Columns</th>
<th width="5%"></th>
</tr>

<tr>
<td width="45%"><select size="9" name="available_sort_columns" multiple onDblClick="addSrcToDestList(this,this.form.sort_columns,true,'');set_available_group_columns_to_sort_columns()">
End_Of_HTML

	# Fill in available fields for display
	$html .= qq[<option value="$_">$_</option>\n] for @available_columns;

	$html .= <<End_Of_HTML;
</select></td>

<td width="5%" align="center"><input type="button" value="&rarr;" class="left-right-button" onClick="addSrcToDestList(this.form.available_sort_columns,this.form.sort_columns,true,'');set_available_group_columns_to_sort_columns()"><br><br><input type="button" value="&rarr;" class="left-right-button" onClick="addSrcToDestList(this.form.available_sort_columns,this.form.sort_columns,true,'desc');set_available_group_columns_to_sort_columns()"><br><span class="left-right-button-label">desc</span><br><br><input type="button" value="&larr;" class="left-right-button" onClick="deleteFromDestList(this.form.sort_columns,this.form.available_sort_columns);set_available_group_columns_to_sort_columns()"></td>

<td width="45%"><select size="9" name="sort_columns" multiple onDblClick="deleteFromDestList(this,this.form.available_sort_columns);set_available_group_columns_to_sort_columns()">
End_Of_HTML

	# Change numeric sort columns into fully qualified field names.
	s/^(\d+)/$columns[($1-1)]/ for @sort_columns;

	# Fill in sort columns.
	$html .= qq[<option value="$_">$_</option>\n] for @sort_columns;

	$html .= <<End_Of_HTML;
</select></td>

<td width="5%" align="center"><input type="button" value="&uarr;" class="up-down-button" onClick="moveUpList(this.form.sort_columns)"><br><br><input type="button" value="&darr;" class="up-down-button" onClick="moveDownList(this.form.sort_columns)"></td>
</tr>

<tr>
<td colspan="3" align="center">
<input type="button" value="Redraw" class="submit-button" onClick="submit_param()">
</td>
<td>
</td>
</tr>
</table>
</div>
<!-- End Sort Columns Parameter -->
End_Of_HTML

	#
	# Define Total Columns parameter form
	#

	$html .= <<End_Of_HTML;

<!-- Begin Total Columns Parameter -->
<div id="bb_param_totals" class="bb-param">

<h3>
Choose Columns to Total
</h3>

<table>
<tr>
<th width="45%">Displayed Number Columns</th>
<th width="5%"></th>
<th width="45%">Total Columns</th>
<th width="5%"></th>
</tr>

<tr>
<td width="45%"><select size="9" name="available_total_columns" multiple onDblClick="addSrcToDestList(this,this.form.total_columns,true,'')">
End_Of_HTML

	# Fill in available fields for display
	{
		my ($tab,$col,$t,$f);
		for my $c (@columns) {
			($tab,$col) = $c =~ /(.*)\.(.*)/ or ($tab,$col) = ($table,$c);
			$t = $schema->t($tab);
			if ($t) {
				# Allow only number fields
				$f = $t ? $t->f($col) : undef;
				$html .= qq[<option value="$c">$c</option>\n]
				  if $f && $f->is_number;
			}
			else {
				# Assume unknown fields can be totaled
				$html .= qq[<option value="$c">$c</option>\n]
			}
		}
	}

	$html .= <<End_Of_HTML;
</select></td>

<td width="5%" align="center"><input type="button" value="&rarr;" class="left-right-button" onClick="addSrcToDestList(this.form.available_total_columns,this.form.total_columns,true,'')"><br><br><input type="button" value="&larr;" class="left-right-button" onClick="deleteFromDestList(this.form.total_columns,this.form.available_total_columns)"></td>

<td width="45%"><select size="9" name="total_columns" multiple onDblClick="deleteFromDestList(this,this.form.available_total_columns)">
End_Of_HTML

	# Fill in total columns.
	$html .= qq[<option value="$_">$_</option>\n] for @total_columns;

	$html .= <<End_Of_HTML;
</select></td>

<td width="5%" align="center"></td>
</tr>

<tr>
<td colspan="3" align="center">
<input type="button" value="Redraw" class="submit-button" onClick="this.form.list_as_rows.value='no';this.form.list_as_table.value='yes';submit_param()">
</td>
<td>
</td>
</tr>
</table>
</div>
<!-- End Total Columns Parameter -->
End_Of_HTML

	#
	# Define Group Columns parameter form
	#

	$html .= <<End_Of_HTML;

<!-- Begin Group Columns Parameter -->
<div id="bb_param_groups" class="bb-param">

<h3>
Choose Columns to Group Together for Totals
</h3>

<table>
<tr>
<th width="45%">Available Sorted Columns</th>
<th width="5%"></th>
<th width="45%">Group by Columns</th>
<th width="5%"></th>
</tr>

<tr>
<td width="45%"><select size="9" name="available_group_columns" multiple onDblClick="addSrcToDestList(this,this.form.group_columns,true,'')">
End_Of_HTML

	# Fill in available fields for display
	$html .= qq[<option value="$_">$_</option>\n]
	  for map { s/ desc$//i; $_ } grep {$_} @sort_columns;

	$html .= <<End_Of_HTML;
</select></td>

<td width="5%" align="center"><input type="button" value="&rarr;" class="left-right-button" onClick="addSrcToDestList(this.form.available_group_columns,this.form.group_columns,true,'')"><br><br><input type="button" value="&larr;" class="left-right-button" onClick="deleteFromDestList(this.form.group_columns,this.form.available_group_columns)"></td>

<td width="45%"><select size="9" name="group_columns" multiple onDblClick="deleteFromDestList(this,this.form.available_group_columns)">
End_Of_HTML

	# Fill in group columns.
	$html .= qq[<option value="$_">$_</option>\n] for @group_columns;

	$html .= <<End_Of_HTML;
</select></td>

<td align="center"></td>
</tr>

<tr>
<td colspan="3" align="center">
<input type="button" value="Redraw" class="submit-button" onClick="this.form.list_as_rows.value='no';this.form.list_as_table.value='yes';submit_param()">
</td>
<td>
</td>
</tr>
</table>
</div>
<!-- End Group Columns Parameter -->
End_Of_HTML

	#
	# Define Choose Filter Columns parameter form
	#

	$html .= <<End_Of_HTML;

<!-- Begin Filter Columns Parameter -->
<div id="bb_param_filters" class="bb-param">

<h3>
Choose Filter Columns
</h3>

<table>
<tr>
<th width="45%">Displayed Columns</th>
<th width="5%"></th>
<th width="45%">Filter by Columns</th>
<th width="5%"></th>
</tr>

<tr>
<td width="45%"><select size="9" name="available_filter_columns" multiple onDblClick="addSrcToDestList(this,this.form.filter_columns,true,'')">
End_Of_HTML

	# Fill in available fields for display
	$html .= qq[<option value="$_">$_</option>\n] for @columns;

	$html .= <<End_Of_HTML;
</select></td>

<td width="5%" align="center"><input type="button" value="&rarr;" class="left-right-button" onClick="addSrcToDestList(this.form.available_filter_columns,this.form.filter_columns,true,'')"><br><br><input type="button" value="&rarr;" class="left-right-button" onClick="addSrcToDestList(this.form.available_filter_columns,this.form.filter_columns,true,'null')"><br><span class="left-right-button-label">null</span><br><br><input type="button" value="&larr;" class="left-right-button" onClick="deleteFromDestList(this.form.filter_columns,this.form.available_filter_columns);filter_reset()"></td>

<td width="45%"><select size="9" name="filter_columns" multiple onDblClick="deleteFromDestList(this,this.form.available_filter_columns)">
End_Of_HTML

	# Fill in filter columns.
	$html .= qq[<option value="$_">$_</option>\n] for @filter_columns;

	$html .= <<End_Of_HTML;
</select></td>

<td width="5%" align="center"><input type="button" value="&uarr;" class="up-down-button" onClick="moveUpList(this.form.filter_columns);filter_reset()"><br><br><input type="button" value="&darr;" class="up-down-button" onClick="moveDownList(this.form.filter_columns);filter_reset()"></td>
</tr>

<tr>
<td colspan="3" align="center">
<input type="button" value="Redraw" class="submit-button" onClick="submit_param()">
</td>
<td>
</td>
</tr>
</table>
</div>
<!-- End Filter Columns Parameter -->
End_Of_HTML

	#
	# Define Selector parameter form
	#

	# Note: allow only the displayed columns to be chosen for a new
	# selector -- it's easier to understand. However, selector filters
	# on non-displayed columns do work. The selector column is added 
	# to the list of displayed columns if is not one of the displayed
	# columns.

	$html .= <<End_Of_HTML;

<!-- Begin Selector Column Parameter -->
<div id="bb_param_selector" class="bb-param">

<h3>
Define Record Selector
</h3>

<table cellspacing="0">
<tr>
<th colspan="2">Display Columns</th>
</tr>

<tr>
<td colspan="2" align="center"><select size="7" name="selector" onChange="select_cond()" style="width:auto">
End_Of_HTML

	# Fill in available fields for display and for "all" and "none".
	my $selected = '';
	my $selector_found = 0;	# flag
	for ('(none)', '(all)', @columns) {
		$selected = $_ eq $selector ? ' selected' : '';
		$html .= qq[<option value="$_"$selected>$_</option>\n];
		$selector_found = 1 if $selected;
	}

	# Add selector column to the list of display columns.
	if (! $selector_found) {
		$html .= qq[<option value="$selector" selected>$selector</option>\n];
	}

	$html .= <<End_Of_HTML;
</select></td>
</tr>
<tr>
<td colspan="2">
<span id="l_cond_desc">Rows will be selected (<img src="$Images/checked.gif" border="0" style="background-color:white">) if the above column matches the following condition.</span>
<span id="l_none_desc">The check marks for selecting records are not displayed.</span>
<span id="l_all_desc">All the rows are initially selected (<img src="$Images/checked.gif" border="0" style="background-color:white">).</span></td>
</tr>
End_Of_HTML

	#
	# Display Selector condition choices.
	#

	$html .= <<End_Of_HTML;
<tr>
<td width="50%" align="right"><span id="l_selector_cond"><select name="selector_cond" onChange="select_cond()">
End_Of_HTML

	# Fill in available conditions.
	$selected = '';
	for (Acme::Lister::Selector->conditions) {
		$selected = $_ eq $selector_cond ? ' selected' : '';
		$html .= qq[<option value="$_"$selected>]
		  . Acme::Lister::Selector->description($_)
		  . qq[</option>\n];
	}

	my $encoded_param1 = escape_html($selector_param1);
	my $encoded_param2 = escape_html($selector_param2);
	my $encoded_param3 = escape_html($selector_param3);

	$html .= <<End_Of_HTML;
</select></span></td>
<td><span id="l_selector_param1"><input type="text" name="selector_param1" value="$encoded_param1" style="width:100%" title="Enter a value for the condition." onChange="select_reset()"></span></td>
</tr>

<tr>
<td align="right"><span id="l_selector_param2_label">and</span></td>
<td><span id="l_selector_param2"><input type="text" name="selector_param2" value="$encoded_param2" style="width:100%" title="Enter a second value for the condition." onChange="select_reset()"></span></td>
</tr>

<tr>
<td colspan="2"><span id="l_selector_param3"><input type="text" name="selector_param3" value="$encoded_param3" style="width:100%" title="Enter a comma separated list of values." onChange="select_reset()"></span></td>
</tr>

<tr>
<td colspan="2" align="center">
<input type="button" value="Redraw" class="submit-button" onClick="submit_selector()">
</td>
</table>
</div>
<script language="JavaScript">select_cond()</script><!-- initialize -->
<!-- End Selector Column Parameter -->
End_Of_HTML

	#
	# End of button bar
	#

	$html .= <<End_Of_HTML;
</form>
<!-- End Button Bar -->

End_Of_HTML

	return $html;
}

#------------------------------------------------------------------------------
#
# Lister helper functions
#
#------------------------------------------------------------------------------

sub display_list_header {

	if ($counting) {
		#
		# Print counting indicator if counting.
		#

		print "<br>Counting..." if $counting;
		return;
	}

	if ($list_as_table) {

		#
		# Redefine format for table list.
		#

		($format,$footer_format) = make_formats();

		print qq[<table class="${lsn}" id="list_table">\n];

		#
		# Display a button that selects entire contents of table
		# for cut & paste or drag & drop.
		#

		my $select_table_button = qq[<td></td>\n];	# empty

		if ($ENV{HTTP_USER_AGENT} =~ /MSIE/
		&& $ENV{HTTP_USER_AGENT} !~ /Macintosh/
		&& $ENV{HTTP_USER_AGENT} !~ /Opera/) {
			# This only works for Internet Explorer on PC.

			# The following produces an "invisible" button that
			# doesn't get copied when pasting the selection into
			# an Excel spreadsheet.

			$select_table_button =
			  qq[<td]
			. qq[ style="height:8pt]
			. qq[;border-bottom:solid #CCCCCC 2px]
			. qq[;border-right:solid #CCCCCC 2px]
			. qq[;padding:2px]
			. qq[">]
			. qq[<div title="Select and highlight entire list."]
			. qq[ style="border:solid #CCCCCC 2px]
			. qq[;width:10px;height:10px;font-size:3pt"]
			. qq[ onMouseOver="this.style.borderColor='#FF00FF']
			. qq[;this.style.cursor='hand'"]
			. qq[ onMouseOut="this.style.borderColor='#CCCCCC'"]
			. qq[ onClick="select_table('list_table')">]
			. qq[</div></td>\n];
		}
		else {
			# Empty button
			$select_table_button =
			  qq[<td]
			. qq[ style="height:8pt]
			. qq[;border-bottom:solid #CCCCCC 2px]
			. qq[;border-right:solid #CCCCCC 2px]
			. qq[;padding:2px]
			. qq[">]
			. qq[</td>\n];
		}

		#
		# Display arguments as column headers.
		#

		print qq[<tr class="${lsn}-column-header">\n];
		print qq[<td class="${lsn}-item-index"></td>\n];
		print qq[<td class="${lsn}-item-selector"></td>\n]
		  unless $selector eq '(none)';
		print $select_table_button;
		print_column_headings();
		print qq[<td class="${lsn}-item-right"></td>\n];
		print qq[</tr>\n];
	}
	else {	# list_as_rows

		#
		# Display info as a list.
		#

		# Note: We display the list as a table irregardless of the
		# ListAsTable option setting. The main difference is that the
		# table is forced to a 100% fixed width table so that long
		# lists can be displayed as they are transmitted instead of
		# waiting for the browser to receive the entire list before
		# rendering the list.

		$format ||= "%s &nbsp; " x @columns;
		print qq[<table class="${lsn}" style="table-layout:fixed;width:99.75%">\n];
		my $header = $t->list_header if $table;
		$header = $format if $header && $header eq 'yes';
		my $list_header = make_list_header(
					$header,
					$list_fields,
					$table,
					$schema,
					$t);

		if ($list_header) {
			# Surround $list_header with <td> </td> tags if not already.
			if ($list_header !~ m|^\s*<t[hd]|i
			||  $list_header =~ s|^\s*<t([hd])>(.*)</t\1>\s*$|$2|i) {
				$list_header = qq[<th>$list_header</th>\n];
			}

			#
			# Display header
			#

			print qq[<tr class="${lsn}-column-header">\n];
			print qq[<td class="${lsn}-item-index"></td>];
			print qq[<td class="${lsn}-item-selector"></td>]
			  unless $selector eq '(none)';
			print qq[<td class="${lsn}-item-buttons"></td>];
			print $list_header;
			#print qq[<td class="${lsn}-item-right"></td>\n];
			print qq[</tr>\n];
		}
	}
}

#------------------------------------------------------------------------------

sub display_list_item {

	use vars qw(@a);  	# array of values read from list query

	$n_rec++;		# record number

	#
	# Compute the number of checked and unchecked records.
	#

	my $checked = check_me_and_compute_counts();

	#
	# Accumulate totals and filter counts
	#

	no strict qw(refs);
	no warnings qw(uninitialized);
	
	my $val;
	for my $fld (keys %filter) {
		$val = ($$fld && $is_null_filter{$fld}) ? '(not empty)' : $$fld;
		$filter{$fld}->{$val} += 1;
		$filter_desc{$fld}->{$val} ||= ${$filter_desc_of{$fld}};
	}

	# Move to Acme::Lister ???
	#for my $gtot (keys %group_totals) {
	#	$group_totals{$gtot} += $$gtot;
	#}

	#
	# Do not print every record if request is Count.
	#

	return if $counting;

	#
	# Construct url for viewing program
	#

	my $url =
	    $view_prog
	  . ($view_prog =~ /[?]/ ? '&' : '?')
	  . ($table ? qq[table=$table&] : '')
	  . escape($key) . '=' . escape(keyval())
	  if $table;

	#
	# Print record index
	#

	my $ri = sprintf("%5s", $n_rec);
	$ri =~ s/ /&nbsp;/g;

	my $id = escape(check_selector_id());
	my $class = ($n_rec % 2) ? "${lsn}-row-odd" : "${lsn}-row-even";

	print qq[\n<tr class="$class"]
	    . qq[ onMouseOver="hi(this)"]
	    . qq[ onMouseOut="lo(this)"];
	print qq[ onClick="do_click('$id',event)"]		if $id;
	print qq[ onDblClick="do_dblclick('$id','$url',event)"]	if $id && $url;
	print qq[>\n];

	#
	# Print record index
	#

	print qq[<td class="${lsn}-item-index">$ri</td>\n];

	#
	# Print selection icon
	#

	print_check_selector($id,$checked);

	#
	# Print edit icon
	#

	print_edit_link();

	#
	# Print row
	#

	print qq[<td class="${lsn}-item">]	if $format !~ /^<t[dh]/i;
	print sprintf_cmd($format, @a);
	print qq[</td>\n] 			if $format !~ /^<t[dh]/i;
	print qq[<td class="${lsn}-item-right"></td>\n] if $list_as_table;
	print qq[</tr>\n];
}

#------------------------------------------------------------------------------

sub display_list_footer {

	if (! $counting) {
		if ($list_as_table) {

			#
			# Display footer line and totals
			#

			print qq[\n<tr class="${lsn}-column-footer">\n];
			print qq[<td class="${lsn}-item-index">&nbsp;</td>\n];
			print qq[<td class="${lsn}-item-selector"></td>\n]
			  unless $selector eq '(none)';
			print qq[<td class="${lsn}-item-buttons"></td>\n];
			print_column_totals();
			print qq[<td class="${lsn}-item-right"></td>\n];
			print qq[</tr>\n];
		}
		else {	# list_as_rows

			my $footer = $t->list_footer if $table;
			$footer = $format if $footer && $footer eq 'yes';
			my $list_footer = make_list_footer(
						$footer,
						$list_fields,
						$table,
						$schema,
						$t);

			if ($list_footer) {
				# Surround $list_footer with <td> </td> tags if not already.
				if ($list_footer !~ m|^\s*<t[hd]|i) {
					$list_footer = qq[<td>$list_footer</td>\n];
				}

				#
				# Display footer
				#

				print qq[\n<tr class="${lsn}-column-footer">\n];
				print qq[<td class="${lsn}-item-index"></td>];
				print qq[<td class="${lsn}-item-selector"></td>]
				  unless $selector eq '(none)';
				print qq[<td class="${lsn}-item-buttons"></td>];
				print $list_footer;
				#print qq[<td class="${lsn}-item-right"></td>\n];
				print qq[</tr>\n];
			}
		}

		print "</table>\n";
	}

	#
	# Print accumulators
	#

	display_list_summary($n_checked, $n_unchecked)
	  if $table || $key || $counting;

	display_list_filters() if @filter_columns;

	#
	# Assign computed accumulators to JavaScript global variables.
	#

	print <<End_Of_Javascript;
<script language="JavaScript">

// Assign globals
nrows        = $n_rows;

nchecked     = $n_checked;
nunchecked   = $n_unchecked;

include_keys($include);
exclude_keys($exclude);

</script>
End_Of_Javascript
}

#------------------------------------------------------------------------------

sub display_list_summary {
	my ($n_checked, $n_unchecked) = @_;

	#
	# Display summary
	#
	
	# Note: Summary only lists the number of checked and unchecked records.
	# Don't display if the checkmarks are not displayed.

	if ($selector eq '(none)') {
		print "<br><p>$n_rows rows counted in list.</p>" if $counting;
		return;
	}

	# Note: Because the selector checkmarks are background images they
	# usually do not appear on printed pages. Don't display the summary
	# either.

	return if $Printing;

	#
	# Compose selector explanation.
	#

	my $check_msg = qq[<div class="summary-description">\n]
	  . $sel->explanation
	  . qq[\n</div>\n];

	#
	# Print summary
	#

	print <<End_Of_HTML;

<h3>Summary</h3>

<div class="summary">
<form name="list_summary">
<table cellspacing="0" cellpadding="0" border="0">
<tr>
<td>
 <table cellspacing="0" cellpadding="0" border="0">
 <tr>
 <th>Checked</th>
 <td><input type="text" class="number-readonly" readOnly name="n_checked" value="$n_checked"></td>
 </tr>
 <tr>
 <th>Unchecked</th>
 <td><input type="text" class="number-readonly" readOnly name="n_unchecked" value="$n_unchecked"></td>
 </tr>
 <tr>
 <th class="summary-total">Total Rows</th>
 <td class="summary-total"><input type="text" class="number-readonly" readOnly name="n_rows" value="$n_rows"></td>
 </tr>
 </table>
 </td>
 <td valign="top">
 $check_msg
 </td>
 </tr>
 </table>
</form>
</div>
End_Of_HTML
}

#------------------------------------------------------------------------------

sub display_list_filters {

	#
	# Display the filters in a form that can be used to resubmit the list.
	#

	return if $Printing;

	print <<End_Of_HTML;
<!---------------- Filters ----------------->
<form name="filters" onSubmit="return false">
<div class="filter-header">&nbsp;</div>
End_Of_HTML

	#
	# Print filters
	#

	my ($field, $col, $tab,$fld,$val,$cnt,$desc, $t, $total_cnt);

	for $field (@filter_fields) {

		($tab,$fld) = $field =~ /(\w+)__(\w+)/ 
		  or ($tab,$fld) = ($table,$field);

		$t = $schema->t($tab);

		$col = column_name($t,$fld);

		$total_cnt = 0;

		print <<End_Of_HTML;
<h3>Filter on $col</h3>

<div class="filter">
<table cellspacing="0" cellpadding="0" border="0">
<tr>
<td>
 <table cellspacing="0" cellpadding="0" border="0">
End_Of_HTML

		my $n = 0;
		for $val (sort keys %{$filter{$field}}) {

			$n++;
			$cnt = $filter{$field}->{$val} || 0;
			$val = escape_html($val);
			$desc = $filter_desc{$field}->{$val};
			$desc = $desc ? "$desc ($val)" : $val;
			$desc ||= '(empty)';

			$total_cnt += $cnt;

			print <<End_Of_HTML;
 <tr>
  <td>
  &nbsp;$desc
  </td>
  <td align="right" valign="top">
  <input type="text" name="F${n}_$field" value="$cnt" class="number-readonly" size="4" readonly tabindex="$No_Tabindex">
  </td>
  <td align="right" valign="top">
End_Of_HTML
			print <<End_Of_HTML if ! $Printing;
  <input type="checkbox" name="C${n}_$field" value="$val" title="Check to include this category in next query" checked onClick="compute_filter(this)">
End_Of_HTML

			print <<End_Of_HTML;
  </td>
 </tr>
End_Of_HTML
		}

		print <<End_Of_HTML;
 <tr>
  <td>
  &nbsp;Total
  </td>
  <td valign="top" align="center" style="border-top: solid black 1px">
  <input type="text" name="T_$field" value="$total_cnt" class="number-readonly" size="4" readonly tabindex="$No_Tabindex">
  </td>
  <td align="right" valign="top">
  </td>
 </tr>
 </table>
</td>
</tr>

</table>
</div>

End_Of_HTML
	}

	#
	# Print a button to re-submit list with a filter.
	#

	print <<End_Of_HTML if ! $Printing;
<div class="filter-footer">
<input type="button" class="button" value="Redo List With Only Checked Filter Categories" onClick="filter(this.form)">
</div>
End_Of_HTML

	print <<End_Of_HTML;
</form>
End_Of_HTML
}

#------------------------------------------------------------------------------
#
# Support functions
#
#------------------------------------------------------------------------------

sub program_of {
	
	# Extract program name from url

	my $url = shift;
	$url =~ s|.*://||;	# remove protocol
	$url =~ s|.*?/|/|;	# remove host name (till /)
	$url =~ s|\?.*||;	# remove parameters after ?
	return $url;
}

#------------------------------------------------------------------------------

sub fully_qualify {
	my $tbl = shift;	# table name
	return '' unless @_;

	# Each field/column argument can be a comma separated list of fields,
	# all of which have the $tbl table name inserted as part if the name.
	# Example: customers.email,customers.email_acknowledge

	# Return fully qualified arguments concatonated into one comma
	# separated string.

	return	join ",",			# concatonate
		map {/\./ ? $_ : "$tbl.$_"}	# fully qualify name (if needed)
		grep {$_}			# remove empty names
		map { split /\s*,\s*/ } @_;	# split each argument
}

#------------------------------------------------------------------------------

sub available_columns {
	our $table;	# available columns for table and its joined tables

	return () unless $table;

	# Note: this routine also collects the tables referred to in @table
	# for no good reason yet.

	my (@tables,@available_columns);

	push @tables, $table;
	push @available_columns, table_columns($table);

	#
	# Get columns of joined tables.
	#

	my $join_clause = $t->join;
	if ($join_clause) {
		my $j = new Acme::Base::Schema::Join $table, $join_clause;
		for my $join_table ($j->join_tables) {
			push @tables, $join_table;
			push @available_columns, table_columns($join_table);
		}
	}

	#
	# Get fields of master tables.
	#

	my $master_clause = $t->master;
	if ($master_clause) {
		my $m = new Acme::Base::Schema::Master $table, $master_clause;
		for my $master_table ($m->master_tables) {
			push @tables, $master_table;
			push @available_columns, table_columns($master_table);
		}
	}

	#
	# Get fields of schema list join clause.
	#

	my $list_join_clause = $t->list_join;
	if ($list_join_clause) {
		my $j = new Acme::Base::Schema::Join $table, $list_join_clause;
		for my $join_table ($j->join_tables) {
			push @tables, $join_table;
			push @available_columns, table_columns($join_table);
		}
	}

	# Note: @tables holds the names of tables referred to. It is ignored.

	return @available_columns;
}

sub table_columns {
	my $table = shift;

	my @cols;
	my $t = $schema->t($table);
	push @cols, "$table.". $t->id if $t->id;
	push @cols, map { "$table.$_" } $t->fields;
	if ($t->audit) {
		push @cols,
			"$table." . $t->audit_entry,
			"$table." . $t->audit_update,
			"$table." . $t->audit_user;
	}
	return @cols;
}

#------------------------------------------------------------------------------
#
# List Header Utilities
#
#------------------------------------------------------------------------------

sub print_column_headings {
	
	# Construct <td> tags with title attributes set to field descriptions
	# to provide popup descriptions when the user hovers over a column
	# heading.

	my ($f,$head,$desc);
	my ($tab,$fld);

	# Track previous table objects $t to reduce calls to $schema->t().
	my $tt     = $t;
	my $ttable = $table;

	for my $field (@columns) {
		($tab,$fld) = $field =~ /(\w+)\.(\w+)/ 
		  or ($tab,$fld) = ($table,$field);
		
		if ($tab ne $ttable) {
			$tt     = $schema->t($tab);
			$ttable = $tab;
		}

		#
		# Make header for each column
		#

		if (! $list_column_headers || $list_column_headers =~ /none/i) {
			# Display header line with blank headers.
			# (Other controls are in header line).

			$head = '';
		}
		else {
			$head = column_name($tt,$fld);
		}

		#
		# Get field description to use as a pop-up description
		# when user places the mouse over the column header.
		#

		$desc = '';
		if ($tt) {

			if ($f = $tt->f($fld)) {
				$desc = $f->description;
			}
			elsif ($fld eq $tt->audit_entry) {
				$desc = "Date this record was entered."
			}
			elsif ($fld eq $tt->audit_update) {
				$desc = "Time this record was last updated."
			}
			elsif ($fld eq $tt->audit_user) {
				$desc = "User ID of the person who last updated this record."
			}
			elsif ($fld eq $tt->id) {
				$desc = $tt->id_description;
			}
			else  { $desc = ''; }
		}

		$desc = escape_html($desc);

		print qq[<th title="$desc" nowrap>$head</th>\n];
	}
}

#------------------------------------------------------------------------------

sub column_name {
	my ($tt,$fld) = @_;

	#
	# Return the printed name for the column depending upon the user's
	# list headings option.
	#

	if ($list_column_headers =~ /FieldTitles/i) {
		my $head = '';
		if ($tt) {
			# Get titles from schema
			if (my $f = $tt->f($fld)) {
				$head = $f->title;
			}
			elsif ($fld eq $tt->audit_entry
			    || $fld eq $tt->audit_update
			    || $fld eq $tt->audit_user) {
				# Change underlines to spaces and
				# capitalize words.
				$head = $fld;
				$head =~ s/_/ /g;
				$head =~ s/\b(\w+)\b/ucfirst(lc($1))/eg;
			}
			else {
				$head = $tt->id_title || '';
			}
		}
		else {
			# Make titles out of field names
			# Change underlines to spaces and
			# capitalize words.
			$head = $fld;
			$head =~ s/_/ /g;
			$head =~ s/\b(\w+)\b/ucfirst(lc($1))/eg;
			$head =~ s/\bId\b/ID/; # capitalize id
		}
		return $head;
	}

	if ($list_column_headers =~ /QualifiedFieldNames/i) {
		if ($tt) {
			return $tt->name . ".$fld";
		}
		else {
			return $fld;
		}
	}

	if ($list_column_headers =~ /FieldNames/i) {
		return $fld;
	}

	return $fld;
}

#------------------------------------------------------------------------------

sub make_formats {

	# Construct a sprintf() format string for printing rows.
	# Build a format for the footer too. Return both as a two
	# element array. The $footer_format does not have any sprintf()
	# format codes between <td></td> for the columns that are not
	# being totaled. This is because feeding undefs into sprintf
	# formats like "%d" returns a "0" instead of nothing. Therefore
	# we use %s formats for all columns not being totaled.

	my $format = '';	# returned format string
	my $footer_format = '';	# footer format

	# Construct classified <td> tags.

	my ($f,$class);
	my ($tab,$fld);

	# Track previous table objects $t to reduce calls to $schema->t().
	my $tt     = $t;
	my $ttable = $table;

	for my $col (@columns) {
		($tab,$fld) = $col =~ /(\w+)\.(\w+)/ 
		  or ($tab,$fld) = ($table,$col);
		
		if ($tab ne $ttable) {
			$tt     = $schema->t($tab);
			$ttable = $tab;
		}

		my $class = '';
		my $fmt   = '%s';
		if ($f = $tt->f($fld)) {
			$class = qq[ class="number"] if $f->is_number;
			$fmt   = $f->format;
		}
		$format .= qq[<td$class>$fmt</td>\n];

		$footer_format	.= exists $list::_Totals{ "$tab.$fld" }
				 ? qq[<td$class>$fmt</td>\n]
				 : qq[<td$class>%s</td>\n];
	}
	return ($format,$footer_format);
}

#------------------------------------------------------------------------------

sub make_list_header {
	my ($format,$fields,$table,$schema,$t) = @_;

	# Helper function for building a list header from field names.
	# and a sprintf type format string. Used when $list_as_rows
	# is true and list_header was not specified in the schema.

	$format ||= '';
	return $format if $format !~ /\%/;	# not a sprintf() format

	# Merge field titles into format string.

	my ($tab,$fld,$tt,@headers);
	for my $h (split(/\s*,\s*/,$fields)) {
		($tab,$fld) = $h =~ /(\w+)\.(\w+)/ 
		  or ($tab,$fld) = ($table,$h);

		if ($tab eq $table) { $tt = $t; }
		else                { $tt = $schema->t($tab); }

		if ($tt && $tt->{$fld}) {
			push @headers, $tt->{$fld}->title;
		}
		else {
			# Capitalize field name
			# (happens on ID fields)
			$fld =~ s/_/ /g;
			$fld =~ s/\b(\w+)\b/ucfirst(lc($1))/eg;
			push @headers, $fld;
		}
	}

	# Format header and evaluate embedded functions
	return clean_a_tags(sprintf_cmd($format,@headers));
}

sub clean_a_tags {
	my ($line) = @_;

	# Remove <a> tags from an output line.

	$line =~ s|</?a(\s[^>]*)?>||gsi;
	return $line;
}

#------------------------------------------------------------------------------
#
# List Item Utilities
#
#------------------------------------------------------------------------------

sub keyval {
	no strict 'refs';
	return ${$1.'__'.$2} if $key =~ /(.*?)\.(.*)/;
	return ${$table.'__'.$key} if $table;
	return ${$key};
}

#------------------------------------------------------------------------------

sub check_me_and_compute_counts {

	# Note: This routine returns a value of:
	#	1	-- checked
	#	0	-- unchecked

	my $checked;		# return value

	#
	# Compute counts for each record.
	#

	$n_rows++;

	#
	# Determine check setting.
	#

	my $keyval = keyval();

	if ($select_all) {
		if ($invert_all) {
			if ($exclude{$keyval}) {
				$checked = 1
			}
			else {
				$checked = 0
			}
		}
		else {
			if ($exclude{$keyval}) {
				$checked = 0
			}
			else {
				$checked = 1
			}
		}
	}
	else {
		if ($invert_all) {
			if ($include{$keyval}) {
				$checked = 0
			}
			elsif ($exclude{$keyval}) {
				$checked = 1
			}
			else {
				$checked = $sel->is_checked ? 0 : 1;
			}
		}
		else {
			if ($include{$keyval}) {
				$checked = 1
			}
			elsif ($exclude{$keyval}) {
				$checked = 0
			}
			else {
				#
				# Normal case
				#

				# Check selector filter to see what initial
				# check settings should be.

				$checked = $sel->is_checked ? 1 : 0;
			}
		}
	}

	$n_checked++     if $checked;
	$n_unchecked++   if ! $checked;

	return $checked;
}

#------------------------------------------------------------------------------

sub print_check_selector {

	return if $selector eq '(none)';

	if (! ($table || $key)) {
		print qq[<td class="${lsn}-item-selector"></td>\n];
		return;
	}

	my $id = shift || check_selector_id();
	my $checked = shift;

	my $class = $checked ? 'checked' : 'unchecked';

	$id = escape_html($id);

	my $space = ($ENV{HTTP_USER_AGENT} =~ /MSIE/ && $ENV{HTTP_USER_AGENT} !~ /Opera/)
		  ? ''
		  : '&nbsp;';
	print
	    qq[<td class="${lsn}-item-selector">]
	  . qq[<a]
	  . qq[ href="javascript:void 0"]
	  . qq[ id="$id"]
	  . qq[ class="$class"]
	  . qq[ onClick="flip(this)"]
	  . qq[>$space</a>]
	  . qq[</td>\n];
}

sub check_selector_id {

	return unless $table || $key;

	#
	# Check for null key error condition (every key value must be unique)
	#

	my $keyval = keyval();
	$nullkey++ unless defined $keyval;

	#
	# Make id by appending key value to 'l_'.
	#

	return 'l_' . $keyval;
}

#------------------------------------------------------------------------------

sub print_edit_link {

	if (! ($table || ($edit_prog && $key))
	||  ! Acme::Security::allow('Access',program_of($edit_prog))) {
	
		print qq[<td class="${lsn}-item-buttons"></td>\n];
		return;
	}

	my $request = 'edit';
	my $alt_title = ucfirst(lc($request)) . ' ' . ($t ? $t->row_name : '') . '.';

	# The following links work well for lists in the browser, but
	# when copied to other programs, they break because of the 
	# javascript and relative url.
	#my $href = qq[javascript:open_form(]
	#  . qq['$edit_prog]
	#  . ($edit_prog =~ /[?]/ ? '&' : '?')
	#  . qq[request=$request&]
	#  . ($table ? qq[request=table=$table&] : '')
	#  . escape("$key) . '=' . escape(keyval())
	#  . qq[')];
	#
	#print qq[<td><a href="$href" title="$alt_title">$Bullet</a></td>];

	# The following links work when the list is copied to Excel.
	# They embed the full url of the web site.

	my $href = $Base
	  . ($edit_prog =~ m|^/| ? '' : '/')
	  . $edit_prog
	  . ($edit_prog =~ /[?]/ ? '&' : '?')
	  . qq[request=$request&]
	  . ($table ? qq[table=$table&] : '')
	  . escape($key) . '=' . escape(keyval());

	print
	    qq[<td class="${lsn}-item-buttons">]
	  . qq[<a href="$href"]
	  . qq[ title="$alt_title"]
	  . qq[ onClick="window.new_target&&new_target(this,'form')"]
	  . qq[ tabindex="$No_Tabindex"]
	  . qq[>$Bullet</a>]
	  . qq[</td>\n];
}

#------------------------------------------------------------------------------
#
# List Footer Utilities
#
#------------------------------------------------------------------------------

sub print_column_totals {
	
	#
	# Stuff totals accumulated in the %_Totals hash by column name into
	# the @totals array in the correct column position.
	#

	my @totals =  ('') x @columns;
	for (keys %list::_Totals) {
		($col = $_) =~ s/__/\./g;	# convert perl to sql
		$totals[ $l->column_n($col) ] = $list::_Totals{$_};
	}

	print qq[<td class="${lsn}-item">]	if $footer_format !~ /^<t[dh]/i;
	print sprintf_cmd($footer_format, @totals);
	print qq[</td>] 			if $footer_format !~ /^<t[dh]/i;
}

#------------------------------------------------------------------------------

sub make_list_footer {
	my ($format,$fields,$table,$schema,$t) = @_;

	# Helper function for building a list footer from field names.
	# and a sprintf type format string. Used when $list_as_rows
	# is true and list_footer was not specified in the schema.

	return '' unless $format;
	return sprintf_cmd($format,@a);
}

#------------------------------------------------------------------------------
#
# List limit functions
#
#------------------------------------------------------------------------------

# Note: These are copies of the functions in Acme::Lister::Limit module.
# They allow the schema files to specify additional email addresses to 
# send the email alerts to.

sub limit_abort_func {
	my $limit = shift;

	my $to = $t->limit_alert if $t;

	my $subject = qq[Aborting: Limit of $limit for $list_id list has been exceeded.];

	my $msg = <<End_Of_Msg;
The $list_id list has exceeded its limit of $limit records.
End_Of_Msg

	$msg .= ($t && $t->nolimit)
	      ? qq[The list is continuing on as the table's schema file specifies "no limit".]
	      : qq[The list has been terminated.\n];

	Acme::Security::alert($msg, ($to || ''), $subject)
	  unless $t && $t->nolimit_alert;

	$msg .= <<End_Of_Msg;
These limits are usually set by administrators in the site configuration file or in database schema files.
End_Of_Msg

	die $msg unless $t && $t->nolimit;
}

#------------------------------------------------------------------------------

sub limit_warn_func {
	my $limit = shift;

	return if $t && $t->nolimit_alert;

	my $to = $t->limit_alert if $t;

	my $subject = qq[Warning: Limit of $limit for $list_id list has been reached.];

	my $msg = <<End_Of_Msg;
The $list_id list has reached a limit of $limit records and is generating this notice.
The list is continuing to be rendered. This is only a warning.
End_Of_Msg

	Acme::Security::alert($msg, ($to || ''), $subject);
}

__END__
