#!/usr/bin/perl -wT -I../../lib

=head1 NAME

compare.cgi -- Compare 2 lists

=head1 SYNOPSIS

    compare.cgi?list1=mylist1&list2=mylist2&option=23&result=myresultlist
      or to display a form of parameters
    compare.cgi

=head1 DESCRIPTION

The compare.cgi program compares two lists such as customer lists.
This is essentially an interface to the Linux/Unix utility comm. The options
are the same as comm's.

  Flags 1, 2, or 3 suppress printing of the corresponding column(list). Thus
  comm -12 prints only the lines common to the two files; comm -23 prints only
  lines in the first file but not in the second; comm -123 is a no-op (does
  nothing).

=head1 CHANGE LOG

 12/31/00 Initial version.
 6/2/02   Converted to AcmeBase schema.
 7/26/02  Generalized so table not assumed if list_id and key in parameters.
 11/15/02 Added comparison of record columns and text array values.

=head1 AUTHOR

Copyright 2000-2002, Mackley F. Pexton.  All rights reserved.

Address bug reports and comments to: mackleyp@montana.net

=cut

package compare;

require 5.6.0;
use strict;

use Acme::HTTP			qw(escape);
use Acme::HTML;
use Acme::HTML::Parameters;
use Acme::HTML::ListMenus;
use Acme::Base::Schema;
use Acme::Searcher;
use Acme::Db;
use Acme::Config;
use Local::Acme::User		qw(user);
use Text::ParseWords;
use CGI;

use Acme::Security;
Acme::Security::check('Access');

$|=1;		# flush output (synchronize errors with output)

our $cgi = new CGI;

our $Base	= Acme::Config->Base;
our $Root	= Acme::Config->Root;
our $Temp_Dir	= Acme::Config->Temp_Dir;
our $List_Dir	= Acme::Config->List_Space;
our $List_Prog	= Acme::Config->List_Prog;

our $Entire_Record = '(entire record)';		# menu selection / flag

#
# Program Arguments
#

our $table     = $cgi->param('table')     || $cgi->param('lists');

# in lieu of $table
our $list_id   = $cgi->param('list_id')   || $table		|| '';
our $key       = $cgi->param('key')       || '';
our $sql       = $cgi->param('sql')       || '';

our $list_prog = $cgi->param('list_prog') || $List_Prog;

our $list1     = $cgi->param('list1');
our $list2     = $cgi->param('list2');
our $result    = $cgi->param('result')    || 'compare_result.txt';
our $option    = $cgi->param('option')    || '13';	# a /usr/bin/comm option

our @columns   = $cgi->param('columns');	# list.cgi parameter
our @compare_columns = $cgi->param('compare_columns');

our $user = user();

our $default_result = $user . "_compare_list";	# default result file name

our ($schema, $t);

$schema = new Acme::Base::Schema;

if ($table) {

	$t = $schema->t($table) or die "Unknown table '$table'";

	$key       ||= $t->key;
	$list_prog ||= $t->list_prog;
}

unless ($list_id && $list1 && $list2) {

	print $cgi->header(-type=>'text/html') unless @ARGV;

	Acme::HTML::Parameters::get_parameters(
		title        => 'Compare Lists',
		submit       => 'Compare',
		reset        => ' Cancel',
		html         => compare::param_html(),
		script       => compare::param_script(),
		instructions => compare::instructions(),
		);
	exit 0;		# the parameter form will re-call this routine
}

our ($result_file) = "$List_Dir/$list_id/$result" =~ /([^\$\@`'"|;]+)/;	 # untaint

if (-e $result_file && $result ne $default_result) {
	my $overwrite = $cgi->param('overwrite') || 'no';
	unless ($overwrite =~ /yes/i) {
		$cgi->cache('no');
		print $cgi->header(-type=>'text/html', -expires=>'-1d') unless @ARGV;
		print Acme::HTML::Parameters->confirm(
			cgi	=> $cgi,
			title   => 'Compare Lists',
			message => "The file <code>$result</code> already exists.",
			prompt  => "Do you want to overwrite it?",
			answer  => 'overwrite',
			);

		exit 0;
	}
}

# Debug needs to output header as program normally does a redirect to a url.
#print $cgi->cache('no');
#print $cgi->header(-type=>'text/html') unless @ARGV;

if (! @compare_columns && ! (grep { /__/ } $cgi->param())
|| (@compare_columns == 1 && $compare_columns[0] eq $Entire_Record)
) {
	compare_records();
}
else {
	compare_columns();
}

display_result();

exit 0;

#------------------------------------------------------------------------------

sub compare_records {

	($list1)   = $list1	=~ /([^\$\@`'"|;]+)/;		# untaint
	($list2)   = $list2	=~ /([^\$\@`'"|;]+)/;		# untaint
	($option)  = $option	=~ /(\d+)/;			# untaint

	#
	# Make compare file files.
	#

	# Note: Temporary file names are randomly generated. However, 
	# the file for $list1 ends with "1", whereas the second list
	# $list2 file ends with "2".

	my $cmp1 = "$Temp_Dir/" . sprintf("%05d", int(rand 30000)+1) . '1';
	my $cmp2 = "$Temp_Dir/" . sprintf("%05d", int(rand 30000)+1) . '2';

	#
	# Use shell commands to compare files.
	#

	$ENV{PATH} = '/bin:/usr/bin';				# untaint
	delete @ENV{ qw(IFS CDPATH ENV BASH_ENV) };		# untaint

	system(<<EOF) == 0 or die "Cannot compare lists: $?";
umask 077
sort "$List_Dir/$list1" > $cmp1 2>/dev/null
sort "$List_Dir/$list2" > $cmp2 2>/dev/null

umask 022
comm -$option $cmp1 $cmp2 | sort -u > $result_file 2>/dev/null

rm -f $cmp1 $cmp2 >/dev/null 2>&1
EOF
}

#------------------------------------------------------------------------------

sub compare_columns {

	# To compare record fields, 4 files are created. The first two have
	# the contents of the columns with the key value appended, the
	# second two have just the contents of the columns of the input
	# lists. The second two lists with just the column contents are 
	# used to compare. The rows in the resulting list are then searched
	# for in one or the other of the first two lists and the key values
	# in those records are striped and returned as the resulting list.

	($list1)   = $list1	=~ /([^\$\@`'"|;]+)/;		# untaint
	($list2)   = $list2	=~ /([^\$\@`'"|;]+)/;		# untaint
	($option)  = $option	=~ /(\d+)/;			# untaint

	#
	# Make temporary working directory. Its more secure seeing as
	# we are copying actual contents of database.
	#

	# Note: Temporary file names are randomly generated. However, 
	# the file for $list1 ends with "1", whereas the second list
	# $list2 file ends with "2".

	my $work_dir = $Temp_Dir . '/.' . sprintf("%05d", int(rand 30000)+1);

	umask 0077;		# allow search and write to directory
	mkdir $work_dir or die "Cannot make working directory '$work_dir'". $!;

	#
	# Make reference file files (column contents with key).
	#

	my $ref1 = "$work_dir/" . sprintf("%05d", int(rand 30000)+1) . '1';
	my $ref2 = "$work_dir/" . sprintf("%05d", int(rand 30000)+1) . '2';

	#
	# Make compare file files (only column contents).
	#

	my $cmp1 = "$work_dir/" . sprintf("%05d", int(rand 30000)+1) . '1';
	my $cmp2 = "$work_dir/" . sprintf("%05d", int(rand 30000)+1) . '2';

	#
	# Make intermediate result file name.
	#

	my $result = "$work_dir/" . sprintf("%05d", int(rand 30000)+1) . '3';

	#
	# Write files used to compare.
	#

	make_cmp_files($list1,$cmp1,$ref1);
	make_cmp_files($list2,$cmp2,$ref2);

	#
	# Use shell commands to compare files.
	#

	$ENV{PATH} = '/bin:/usr/bin';				# untaint
	delete @ENV{ qw(IFS CDPATH ENV BASH_ENV) };		# untaint

	system(<<EOF) == 0 or die "Cannot compare lists: $!.";
umask 077
sort -o $cmp1 $cmp1 2>/dev/null
sort -o $cmp2 $cmp2 2>/dev/null

comm -$option $cmp1 $cmp2 | sort -u > $result 2>/dev/null

umask 022
case "$option" in
"12")	grep -F -f$result $ref1		# lines common to both lists
	;;
"23")	grep -F -f$result $ref1		# lines in list1 that are not in list2
	;;
"13")	grep -F -f$result $ref2		# lines in list2 that are not in list1
	;;
esac | perl -F'\\\|' -ane 'print pop(\@F), "\n"' | sort -u > $result_file

rm -rf $work_dir >/dev/null 2>&1
EOF
}

sub make_cmp_files {
	my ($list,$cmp,$ref) = @_;

	# Make reference and comparison files.

	umask 077;

	open REF, ">", $ref or die "Cannot open output file '$ref'. $!";
	open CMP, ">", $cmp or die "Cannot open output file '$cmp'. $!";

	my $sql = make_cmp_sql($list);

	my $q = Acme::Db->query($sql);

	my @a;
	while ( $q->next(\@a) ) {
		# Output to reference file.
		print REF (join '|', @a), "\n";

		# Strip key value, then output to compare file
		pop @a;
		print CMP (join '|', @a), "\n";
	}

	close REF;
	close CMP;
}

sub make_cmp_sql {
	my $list = shift;
	
	#
	# Make sql script for selecting fields from table of list file.
	#

	# Search cgi arguments for text array fields and their values.
	# Any field name that matches "table__column" pattern is assumed
	# to be a text array here.

	my ($fld,$val,$delim,$tbl,$col,$t,$f);
	my @array_cols;
	my $array_where = '';
	my $match_op = $schema->dbengine =~ /Postgres/i
		     ? '~'
		     : $schema->dbengine =~ /MySQL/i
		     ? 'regexp'
		     : $schema->dbengine =~ /Informix/i
		     ? 'matches'
		     : '~';

	for $fld ($cgi->param()) {
		($tbl,$col) = $fld =~ /(\w+)__(\w+)/	or next;
		$t = $schema->t($tbl)			or next;
		$f = $t->{$col}				or next;
		$f->array				or next;

		$val = $cgi->param($fld);

		$delim = $f->array_separator;

		# Push field value as a literal sql column.
		push @array_cols, "'$val'";

		# Add where clause to match array value pattern.
		$array_where .= " and $tbl.$col $match_op '$delim$val$delim'";
	}

	my $sql = "select "
	  . (join ", ", @compare_columns, @array_cols, $key)
	  . " from $table"
	  . " where 1 = 1"
	  . $array_where;

	my $search = ";list_join('$list','$table.$key');";

	my $sqlo = new Acme::SQL $sql;
	my $srch = new Acme::Searcher $search;

	return Acme::Security::clean_sql( $srch->as_sql($sqlo) );
}

#------------------------------------------------------------------------------

sub display_result {

	#
	# Use the list.cgi program to display result list.
	#

	# Note: we change our environment and pipe the cgi arguments to
	# the list program. This "preserves" the state of the list.cgi
	# program a little better than using $cgi->redirect() because 
	# the its option settings exceeds the length of urls accepted
	# by web servers.

	#
	# Redefine search script.
	#

	my $search_script =
  qq[Compare Result: $result;list_join("$list_id/$result","$table.$key");];

	$cgi->param('search',$search_script);	# set new search
	
	#
	# Clear arguments we don't want the list program to process (again).
	#

	$cgi->delete('overwrite');		# clear old list title

	$cgi->delete('list1');
	$cgi->delete('list2');
	$cgi->delete('option');
	$cgi->delete('result');
	$cgi->delete('compare_columns');

	# Delete array fields.
	for ($cgi->param()) {
		$cgi->delete($_) if /__/;
	}

	# Set list annotation depending upon compare options.
	(my $list_name1 = $list1) =~ s|[\w/]+?/||;	# strip directory
	(my $list_name2 = $list2) =~ s|[\w/]+?/||;	# strip directory
	if ($option eq '12') {
		$cgi->param('annotate',
		  qq[Items common to both lists "$list_name1" and "$list_name2".]);
	}
	elsif ($option eq '13') {
		$cgi->param('annotate',
		  qq[Items in "$list_name2" that are not in "$list_name1".]);
	}
	elsif ($option eq '23') {
		$cgi->param('annotate',
		  qq[Items in "$list_name1" that are not in "$list_name2".]);
	}
	else {
		$cgi->param('annotate',
		  qq[option='$option']);
	}

	# Clear out list customization parameters.
	$cgi->delete('list_title');
	$cgi->delete('redrawing');	
	$cgi->delete('original_search');	
	$cgi->delete('include');
	$cgi->delete('exclude');
	$cgi->delete('select_all');	
	$cgi->delete('invert_all');	
	$cgi->delete('n_rows');
	$cgi->delete('n_checked');
	$cgi->delete('n_unchecked');

	#
	# Get full program path of list program.
	#

	my $list_prog_path = $Root . (substr($list_prog,0,1) eq '/' ? '' : '/') . "$list_prog";
	my $list_prog_url  = $Base . (substr($list_prog,0,1) eq '/' ? '' : '/') . "$list_prog";

	#
	# Change working directory to directory of list program.
	#

	my ($dir,$prog) = $list_prog_path =~ m|(.*/)(.*)|;
	chdir $dir or die "Cannot change directory to $dir. $!";

	#
	# Get modified cgi program paramters.
	#

	(my $query_string = $cgi->url(-query=>1)) =~ s/.*\?//;

	#
	# Execute list program.
	#

	# Note: we open a pipe to the list program instead of using a redirect
	# url because the list options have become too large to pass in the
	# url.

	$ENV{PATH} = '/bin:/usr/bin';                   # untaint
	delete @ENV{ qw(IFS CDPATH ENV BASH_ENV) };     # untaint

	#
	# Set environment variables so that CGI in list.cgi reads input.
	#

	$ENV{REQUEST_METHOD}='POST';
	$ENV{CONTENT_TYPE} = 'application/x-www-form-urlencoded';
	$ENV{CONTENT_LENGTH}=length $query_string;
	$ENV{SCRIPT_NAME} = $list_prog;
	$ENV{SCRIPT_FILENAME} = $list_prog_path;
	$ENV{HTTP_REFERER} = $list_prog_url;
	$ENV{REQUEST_URI} = $list_prog;

	open LIST, '|-', "./$prog" or die "Cannot open $prog. $!";
	print LIST $query_string;
	close LIST;
}

#------------------------------------------------------------------------------

sub param_html {

	# Note: If $list_id is empty, show a dropdown list of list directories.
	# Usually the $list_id is set to a table's name, so the dropdown list
	# does not appear.

	my $list = $list_id || '';
	my $param_lists = $list_id ? '' : param_lists();

	my $column_list = column_list();
	my $text_array_list = text_array_list();

	my $more_options = '';
	$more_options = <<EOF if $column_list || $text_array_list;

<tr><td></td><td colspan="2"><hr></tr>  <!-- Horz line -->

$column_list
$text_array_list
EOF

  	#
	# Display list of columns to compare.
	#

	return <<EOF;
<div style="margin-top:1ex; text-align:center;">
<table cellspacing="0" cellpadding="4" width="60%">

$param_lists

<tr>
<th align="right">List 1</th>
<td colspan="2" nowrap><input type=text name="list1" value="" size="50">
<input type="button" value=" Browse " onClick="browse('$list',this.form.list1)" class="browse-button"></td>
</tr>

<tr>
<th align="right">List 2</th>
<td colspan="2" nowrap><input type=text name="list2" value="" size="50">
<input type="button" value=" Browse " onClick="browse('$list',this.form.list2)" class="browse-button"></td>
</tr>

<tr>
<th align="right" valign=top>Options</th>
<td><div style="border:solid black 1px; padding:1ex;">
<input type="radio" name="option" value="12" checked> Select items that are in both List 1 and List 2
<br>
<input type="radio" name="option" value="23"> Select items in List 1 that are not in List 2
<br>
<input type="radio" name="option" value="13"> Select items in List 2 that are not in List 1
</div></td>
<td></td>
</tr>

<tr>
<th align="right"><nobr>Result List</nobr></th>
<td colspan="2"><input type=text size=50 name="result" value="$default_result" onFocus="this.select()"></td>
</tr>$more_options
</table>
</div>
EOF
}

#------------------------------------------------------------------------------

sub param_script {

	return <<EOF;
<script language="JavaScript" type="text/javascript">
var browse_args = new Object;		// Saved arguments for browse_return()

function browse(list, field) {

	if (!list) {
		// Get list from form field
		var form = field.form;
		with (form["lists"]) {
			list = options[selectedIndex].text
		}
	}

	// Open browse window.
	var url = '/cgi-bin/browse.cgi?lists='+ escape(list);
	var window_name = "w_" + field.name;

	// Save arguments for return
	browse_args[window_name] = field;

	open_browse(url,window_name);	// in common.js
	return true;
}

function browse_return(window_name,file_name) {
	var field = browse_args[window_name];

	if (!file_name) return false;	// no return value

	field.value = file_name;

	field.focus();
	if (field.select) field.select();

	return true;
}
</script>
EOF
}

#------------------------------------------------------------------------------

sub param_lists {

	my $options = join "\n",
		      map {qq[<option value="$_">$_</option>] }
		      grep { $_ }
		      list_files();

	return '' unless $options;

	return <<EOF;
<tr>
<th align=right>Lists</th>
<td colspan="2">
<select name="lists">
$options
</select>
</td>
</tr>

EOF
}

sub list_files {

	#
	# Return a list of available lists.
	#

	opendir DIR, $List_Dir or return;

	my @files = readdir DIR;
	closedir DIR;
	shift @files; shift @files;	# remove . and ..

	#
	# Filter files for directories
	#

	return grep { -d "$List_Dir/$_" } @files;
}

#------------------------------------------------------------------------------

sub column_list {

	my @col = @columns
		? @columns
		: $table
		? $t->fields
		: $sql
		? Acme::SQL->new($sql)->columns
		: ();

	return '' unless @col;

	unshift @col, $Entire_Record;	# add entire record to column list

	my $column_list = Acme::HTML::ListMenus->List(
				\@col,			# data
				'compare_columns',	# name of control
				undef,			# default value
				undef,			# width (char)
				undef,			# <select> attributes
				undef,			# sprintf() like format
				'multi',		# 'multi' or 'noundef'
				);

	return <<EOF;
<tr>
<th align=right>Columns</th>
<td colspan="2">
$column_list
</td>
</tr>

EOF
}

#------------------------------------------------------------------------------

sub text_array_list {

	my $sqlo = new Acme::SQL $sql if $sql;

	my @col = @columns
		? @columns
		: $table
		? $t->fields
		: $sqlo
		? $sqlo->columns
		: ();

	return '' unless @col;

	my $text_array_list = '';

	# Set default table to first table listed in sql if it is not defined.
	local $table = $table;
	($table) = $sqlo->tables if $sqlo && ! $table;

	my ($c, $tbl, $fld, $f);
	my $prev_tbl = $tbl = $table;
	for $c (@col) {

		($tbl,$fld) = $c =~ /(\w+)\.(\w+)/
		  or
		($tbl,$fld) = ($table, $c);

		$t = $schema->t($tbl) unless $t && $tbl eq $prev_tbl;
		$prev_tbl = $tbl;

		$f = $t->{$fld} or next;
		
		next unless $f->array && $f->choose;

		my $form_field_name = "${tbl}__${fld}";
		my $choose_list = _choose_dropdown($f,$form_field_name);

		my $col_title = $f->title;

		$text_array_list .= <<EOF;
<tr>
<th align=right>$col_title</th>
<td colspan="2">
$choose_list
</td>
</tr>

EOF
	}

	$text_array_list &&= <<EOF . $text_array_list;
<tr>
<td colspan="3">The following columns have multiple values which you can compare.</td>
</tr>

EOF
	return $text_array_list;
}

sub _choose_dropdown {
	my ($f, $form_field_name,$default,$size,$attributes,$format) = @_;

	#
	# This routine is a simplified version of _choose_dropdown() function
	# in Acme::Form::Filler.
	#

	$format ||= $f->choose_format;
	$format =~ s/^(["'])(.*)\1$/$2/s;	# strip quotes if found

	my $select_rows = $f->choose_rows;

	$attributes ||= '';
	$attributes .= qq( size="$select_rows") if $select_rows;

	my $multi = $f->array ? 'multiple' : '';

	#
	# Render dropdown list
	#

	my $tbl = $f->choose_tables;

	if ($tbl && Acme::HTML::ListMenus->can($tbl)) {
		return Acme::HTML::ListMenus->$tbl(
						$form_field_name,
						$default,
						$size,
						$attributes,
						$format,
						$multi);
	}

	if (my $menu = $f->choose_menu) {
		my @menu_opt = Text::ParseWords::parse_line('\s+',0,$menu);
		return Acme::HTML::ListMenus->Menu(\@menu_opt,
						$form_field_name,
						$default,
						$size,
						$attributes,
						$format);
	}

	if (my $list = $f->choose_list) {
		my %list_opt = Text::ParseWords::parse_line('\s+',0,$list);
		return Acme::HTML::ListMenus->List(\%list_opt,
						$form_field_name,
						$default,
						$size,
						$attributes,
						$format);
	}

	my $sql = $f->choose_sql;

	return '' if ! $sql;

	#
	# Check query with security.
	#

	$sql = Acme::Security::clean_sql($sql);

	return Acme::HTML::ListMenus->List(\$sql,
					$form_field_name,
					$default,
					$size,
					$attributes,
					$format,
					$multi);
}

#------------------------------------------------------------------------------

sub instructions {

	return <<EOF;
<p style="margin-top:0">
This procedure compares two previously saved lists.
It compares the items in the first list with the items in the second list
and stores the results in a list with the name you enter below.
The resulting list is displayed when done.
</p>
<p>
You can compare entire records or you can select the record column(s)
to compare.  To select more than one column, hold the <code>Ctrl</code>
key down while clicking on the selections.
</p>
<p>
Some columns hold multiple values (text-arrays in AcmeBase parlance). Those
columns have a separate selection list from which you can choose the value(s)
you want to compare.
</p>
EOF
}
