#!/usr/bin/perl -wT -I../../lib

=head1 NAME

email.cgi -- Send email to a list of people

=head1 SYNOPSIS

    /cgi-bin/list/email.cgi

    Parameters:
	from	-- the return email address
	to	-- email address(es)
	subject	-- email subject line
	body	-- user message to insert at beginning of email
	message	-- programmer message to insert at beginning of email
	file	-- local file name
	request	-- "Send" or "Preview" or "Test"

=head1 DESCRIPTION

This routine sends a file and/or a given message to a list of email addresses.

Email addresses appear like they were mailed personally. 

There is a "request" parameter that can be set to "Send" or "Preview" or "Test".
If it is not set to either, then a parameter/confirmation form is displayed.

There are two messages that can be inserted into the beginning of the email
before sending. They are the C<message> and C<body> parameters. The C<message>
parameter is intended for other program(er)s to use. The C<body> parameter is a
form field that is used for the user's message. If both are specified, the
C<body> parameter (user's message) precedes the C<message> (programmed message)
parameter.

This routine is used as a list processor to send email to addresses extracted
from the database. It stores the list in a special queue directory which is
defined in the site configuration file .../etc/site.conf. It then invokes
the C<.../bin/sendmaillist> program to process the list.  It is executed
immediately (for tests) or in the background.

Email can be mail-merged. The columns passed to the C<.../bin/sendmaillist>
program can be defined in the AcmeBase schema files using the Email_List
property, or if an arbitrary SQL statement is used (a table name is not
defined), the selected columns are used.

Variables in the email template are substituted with the contents of the
current record's column before sending the email. Variables can be 
defined like %var% or they can by shell like variables $var or ${var} but
not \$var or \${var}. The substitution delimiters can be defined in the
site configuration file (.../etc/site.conf).

=head1 NOTE

This procedure requires the name of the table holding the email addresses.
It does not allow the emailing to addresses in an arbitrary SQL.

=head1 CHANGE LOG

 4/25/00 Initial version.
 8/6/02  Use Acme::HTML::Parameters and Preview
 9/8/02  Updated list processing
 11/6/02 Allow direct execution without special email accounts set up.
 3/8/03  Added keep_alive() function to prevent browser from closing the
 	 connection before the email list has been processed.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000-2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

package email;

require 5.6.0;
use strict;

use Acme::HTML			qw(display_message	escape_html);
use Acme::HTML::Parameters;
use Acme::Base::Schema;
use Local::Acme::User		qw(user			user_email);
use Local::Acme::Banners	qw(common_scripts);
use Acme::Config;
use CGI;

use Acme::Security;
Acme::Security::check('Access');

our $cgi = new CGI;
print $cgi->header(-type=>'text/html') unless @ARGV;

# Check for upload errors
die "Email failed.\n" . $cgi->cgi_error if $cgi->cgi_error;

#
# Define globals
#

no warnings 'once';	# Turn off warnings, many globals used in AUTOLOAD.

our $user = user();

our $html = '';

our $Root               = Acme::Config->Root;
our $Help               = Acme::Config->Help;

our $Queue_Dir		= Acme::Config->Email_List_Space;
our $Queue_Path		= "$Queue_Dir/${user}_queue.txt";	# output file
our $Lock_File		= "$Queue_Path.lock";			# lock file

our $No_Tabindex	= Acme::Config->No_Tabindex;

#
# Get CGI parameters
#

our $request	  = $cgi->param('request')	|| $cgi->url_param('request')	|| '';

our $from   	  = $cgi->param('from')         || '';
our $to     	  = $cgi->param('to')           || '';  # not really used
our $cc           = $cgi->param('cc')           || '';	# not really used
our $bcc          = $cgi->param('bcc')          || '';	# not really used
our $subject	  = $cgi->param('subject')      || '';
our $body         = $cgi->param('body')         || '';

our $message	  = $cgi->param('message')      || '';
our $file         = $cgi->param('file');

#
# List processing parameters
#

our $list_id      = $cgi->param('list_id');	# defined if processing a list
our $table        = $cgi->param('table')	|| '';
our $key          = $cgi->param('key')		|| '';
our $sql          = $cgi->param('sql')		|| '';

our @sort_columns = $cgi->param('sort_columns');
our $sort         = $cgi->param('sort')		|| '';
$sort ||= join ", ", @sort_columns if @sort_columns;

our $limit        = $cgi->param('limit')	|| 0;

our $email_field  = $cgi->param('email_field')	|| '';
our $name_field   = $cgi->param('name_field')	|| '';
our $selector     = $cgi->param('selector')	|| '';

#
# Operational flags
#

our $Previewing   = $cgi->url_param('preview')	|| $cgi->param('preview') || '';
$Previewing       = $Previewing =~ /yes/i       ||
		    $request =~ /preview/i      ||
		    ($Previewing !~ /^no$/i && exists $cgi->{'preview'} ? 1 : 0);

our $Testing      = $cgi->url_param('test')	|| $cgi->param('test')	  || '';
$Testing          = $Testing =~ /yes/i          ||
		    $request =~ /test/i         ||
		    ($Testing !~ /^no$/i && exists $cgi->{'test'} ? 1 : 0);

our $Sending     = $cgi->url_param('send')	|| $cgi->param('send')	  || '';
$Sending         = $Sending =~ /yes/i           ||
		   $request =~ /send/i          ||
		   ($Sending !~ /^no$/i && exists $cgi->{'send'} ? 1 : 0);

#
# Check email security
#

our ($schema, $t);

if ($table) {
	$schema = new Acme::Base::Schema;
	$t = $schema->t($table);

	my $email_security_code = $t->email_security_code;
	Acme::Security::check('Access',$email_security_code)
	  if $email_security_code;

	die "Cannot email to addresses in the table '$table'."
	  . " The option has been turned off in the schema tables"
	  unless $t->email;
}

if (! $from) {
	$from = user_email();
}

if (! $Sending && ! $Previewing && ! $Testing) {
	get_parameters();		# display form
	exit;
}

# Combine $body and $message into one $message
if ($message && $body) {
	$message = qq[<p>$body</p>\n<p>&nbsp;</p>\n<p>$message</p>];
}
elsif ($message || $body) {
	$message = qq[<p>$body$message</p>];
}

die "Need a message or a file to send. Error" if (! $file && ! $message);

if ($file) {
	# Client file being uploaded
	upload($file, \$html);

	if (! $html) {
		# Upload of file failed. Try to open a server side file.
		open(HTML,'<', $file) or die "Cannot open file '$file'. $!";
		local $/;
		$html .= <HTML>;	# read entire file
		close(HTML);
	}
}

if ($message) {
	
	if ($html) {

		#
		# Substitute message into html right after <body>.
		#

		$html =~ s|(<body.*?>)|$1\n\n$message\n<hr><p>&nbsp;</p>\n|si
		  or
		$html = "$message\n\n$html";	# prepend to text files
	}
	else {
		$html = <<EOF;
<html>
<body>
$message
</body>
</html>
EOF
	}
}

#
# Do program request
#

if ($Previewing) {
	preview_email();
}
elsif ($Testing) {
	test_email();
}
elsif ($Sending) {
	send_email();
}

exit;

#------------------------------------------------------------------------------
#
# Autoload routines
#
#------------------------------------------------------------------------------

sub define_subs {

our %Subs = (

#----------------------------------------
'send_email' => <<'END_OF_FUNC',

use Acme::HTML::KeepAlive;

sub send_email {

	#
	# Send html to mail server
	#

	$ENV{PATH} = '/bin:/usr/bin';                   # untaint
	delete @ENV{ qw(IFS CDPATH ENV BASH_ENV) };     # untaint

	if (open(SENDMAIL,'|-', "$Root/bin/sendmaillist -nq '$Queue_Path'")) {
		Acme::HTML::KeepAlive::keep_alive(20);	# don't loose connection

		$cc  &&= "\nCC: $cc";
		$bcc &&= "\nBCC: $bcc";

		$html = quoted_printable($html);

		print SENDMAIL <<End_of_Email;
From: $from
To: $to
Subject: $subject$cc$bcc
Mime-Version: 1.0
Content-Type: text/html; charset=US-ASCII
Content-Transfer-Encoding: quoted-printable

$html
End_of_Email

		close(SENDMAIL);

		display_message("The file has been emailed.","Message",'',0);
	}
	else {
		# It is assumed that an error message has already been output.
		display_message("The file has <font color='red'>NOT</font> been emailed.",
				"Message",'',0);
	}
}
END_OF_FUNC

#----------------------------------------
'test_email' => <<'END_OF_FUNC',
sub test_email {

	$subject = "Test: $subject";		# Identify email as a test

	#
	# Send html to mail server
	#

	$ENV{PATH} = '/bin:/usr/bin';                   # untaint
	delete @ENV{ qw(IFS CDPATH ENV BASH_ENV) };     # untaint

	if (open(SENDMAIL,'|-', "$Root/bin/sendmaillist -t '$Queue_Path'")) {
		$cc  &&= "\nCC: $cc";
		$bcc &&= "\nBCC: $bcc";

		$html = quoted_printable($html);

		print SENDMAIL <<End_of_Email;
From: $from
To: $to
Subject: $subject$cc$bcc
Mime-Version: 1.0
Content-Type: text/html; charset=US-ASCII
Content-Transfer-Encoding: quoted-printable

$html
End_of_Email

		close(SENDMAIL);

		display_message(
		  "A test merge file has been emailed back to you using the email address: "
		  . "<nobr>" . escape_html($from) . "</nobr>.",
		  "Message",'',0);
	}
	else {
		# It is assumed that an error message has already been output.
		display_message("The test file has <font color='red'>NOT</font> been emailed.",
				"Message",'',0);
	}
}
END_OF_FUNC

#----------------------------------------
'preview_email' => <<'END_OF_FUNC',
sub preview_email {

	$cc  &&= qq[<tr>\n<th align="right">CC:</th>\n<td align="left">$cc</td>\n</tr>];
	$bcc &&= qq[<tr>\n<th align="right">BCC:</th>\n<td align="left">$bcc</td>\n</tr>];

	my $mail_header = common_scripts() . <<End_of_Email;
<div class="email-header">
<table cellspacing="0" cellpadding="2">
<tr>
<th align="right">From:</th>
<td align="left">$from</td>
</tr>
<tr>
<th align="right">To:</th>
<td align="left">$to</td>
</tr>
$cc$bcc
<tr>
<th align="right">Subject:</th>
<td align="left">$subject</td>
</tr>
</table>
</div>
End_of_Email

	#
	# Substitute mail headers into html right after <body> for display.
	#

	$html =~ s|(<body.*?>)|$1\n$mail_header\n|si
	  or
	$html = "$mail_header\n\n$html";	# prepend to text files
	
	#
	# Print a preview of the text.
	#

	print $html;
}
END_OF_FUNC

#----------------------------------------
'get_parameters' => <<'END_OF_FUNC',
sub get_parameters {

	$to 	 = escape_html($to);
	$from	 = escape_html($from);
	$subject = escape_html($subject);

	my $statistics = '';

	if ($list_id) {
	
		#
		# Check if queue is still in use from a previous mailing.
		#

		die "The email queue file '$Queue_Path' is still being"
		 .  " processed from a previous email.\n"
		 .  "Please wait until the email queue has been sent"
		 .  " before sending another list."
		  if -f $Lock_File;

		#
		# Process list of names
		#

		my $count = save_email_list($Queue_Path, get_email_and_name_fields());

		$statistics = "$count names are selected for email."
		  if $count;
	}

	#
	# Assemble To: field
	#

	my ($email, $name) = get_email_and_name_fields();
	my $email_field_choice = $name ? escape_html("$name <$email>") : $email;

	#
	# Assemble parameters screen
	#

	my $parameters = <<EOF;

<table cellspacing="0" cellpadding="2" border="0" width="85%">

<tr>
<th></th>
<td>$statistics</td>
</tr>

<tr><th align="right" valign="top"></th><td>&nbsp;</td></tr> <!-- blank line -->

<tr>
<th align="right" valign="top" nowrap>From</th>
<td><input type=text name="from" value="$from" size=50 style="width:100%"></td>
</tr>

<tr>
<th align="right" valign="top" nowrap>To</th>
<td>$email_field_choice</td>
</tr>

<tr>
<th align="right" valign="top" nowrap>Subject</th>
<td><input type="text" name="subject" value="$subject" size=50 style="width:100%"></td>
</tr>

<tr><th align="right" valign="top"></th><td>&nbsp;</td></tr> <!-- blank line -->

<tr>
<th align="right" valign="top">File</th>
<td><input type="file" name="file" value="" size="35" maxlength="256" style="width:100%"></td>
</tr>

<tr>
<th align="right" valign="top">Message</th>
<td><textarea name="body" wrap="virtual" rows="10" cols="50" style="width:100%"></textarea></td>
</tr>

<tr>
<th></th>
<td><small>The above message will be inserted into the beginning of the email
before it is sent.</small></td>
</tr>

<tr><th align="right" valign="top"></th><td>&nbsp;</td></tr> <!-- blank line -->

</table>
<input type="button" class="button" value="Preview" onClick="do_preview()">
<input type="button" class="button" value=" Test  " onClick="do_test()">
<input type="button" class="button" value=" Send  " onClick="this.form.submit()">
<input type="button" class="button" value=" Close " onClick="self.close()">
<input type="hidden" name="request" value="send">
EOF

	my $other_button =
qq[<input type="button" class="button" value="Preview" onClick="do_preview()">];

	print Acme::HTML::Parameters->get(
				window_title => 'Send Email',
				title        => 'Send Email to a List of Names',
				html         => $parameters,
				script       => scripts(),
				submit       => undef,
				cgi	     => $cgi,
				instructions_file => "$Root$Help/email_instructions.htm",
				);
}

sub get_email_and_name_fields {
	
	#
	# Return email field and name field
	#

	# Get just one email field and name field from the possible list of
	# fields in the "email_field" and "name_field" CGI parameters.
	# The "separator" field is looked at as the preferred candidate for
	# the email field. The preferred name field is the first field listed
	# in the name_fields parameter.

	my @emails = split /\s*,\s*/, $email_field;
	my @names  = split /\s*,\s*/, $name_field;

	# Note: the email_field CGI parameter is used as a search
	# parameter. Override it with the table's property email_field_list,
	# which is just for lists.

	my $email_field_list = $t->email_field_list_fields;
	my $name_field_list  = $t->name_field_list_fields;

	my @email_choices = split /\s*,\s*/, $email_field_list;
	my @name_choices  = split /\s*,\s*/, $name_field_list;

	if (@emails == 1 && grep { $_ eq $emails[0] } @email_choices) {
		return ($emails[0], $names[0]);
	}

	for (my $i = 0; $i < @email_choices; $i++) {
		
		#
		# Choose selector field if it is in list of email fields
		#

		return ($email_choices[$i], $name_choices[$i] || $name_choices[0] || '')
		  if $selector && $email_choices[$i] eq $selector;
	}

	if (@email_choices > 1) {
		
		#
		# First get which email address to use in list
		#

		get_email_field_choice(\@email_choices,\@name_choices);
		exit;
	}

	return ($email_choices[0], $name_choices[0]);
}

sub get_email_field_choice {
	my ($emails, $names) = @_;

	my $parameters = <<End_Of_HTML;
<div style="margin-left:5ex;text-align:left;">
<table>
End_Of_HTML

	my $checked = ' checked';
	for (@$emails) {
		$parameters .= <<End_Of_HTML;
<tr><td><input type="radio" name="email_field" value="$_"$checked></td><td>$_</td></tr>
End_Of_HTML
		$checked = '';
	}

	$parameters .= <<End_Of_HTML;
</table>
End_Of_HTML

	if (@$names > 1) {
		$parameters .= <<End_Of_HTML;
</div>
<h3>Choose Corresponding Name Field for Email</h3>
<div style="margin-left:5ex; text-align:left;">
<table>
End_Of_HTML

		my $checked = ' checked';
		for (@$names) {
			$parameters .= <<End_Of_HTML;
<tr><td><input type="radio" name="name_field" value="$_"$checked></td><td>$_</td></tr>
End_Of_HTML
			$checked = '';
		}
	}

	$parameters .= <<End_Of_HTML;
</table>
</div>
End_Of_HTML

	print Acme::HTML::Parameters->get(
			cgi	=> $cgi,
			title   => 'First Choose Email Address',
			html    => $parameters,
			submit	=> 'Submit'
			);
}

sub scripts {

	return <<End_Of_JavaScript;
<script language="JavaScript" type="text/javascript">
var wn = Math.floor(Math.random() * 10000);	// window number (to provide semi-uniqueness)

function do_preview() {
	var f = window.document.parameters;

	var saved_target  = f.target;
	var saved_request = f.request.value;

	f.target = "Preview" + wn++;

	f.request.value="preview";

	open_window("",f.target);
	f.submit()

	f.target        = saved_target;
	f.request.value = saved_request;
	return;
}

function do_test() {
	var f = window.document.parameters;

	var saved_target  = f.target;
	var saved_request = f.request.value;

	f.target = "Test" + wn++;

	f.request.value="test";

	open_window("",f.target);
	f.submit()

	f.target        = saved_target;
	f.request.value = saved_request;
	return;
}
</script>
End_Of_JavaScript
}
END_OF_FUNC

#----------------------------------------
'upload' => <<'END_OF_FUNC',
sub upload {
        # Save uploaded file in a scalar

        my $file = shift;               # file handle named same as file name
	my $ref  = shift;		# reference to a variable

        my $bufsize = 5 * 1024;         # 5K buffer size
        my $buffer = '';

        # copy file
        while (read($file,$buffer,$bufsize)) {
		$$ref .= $buffer;
        }
}
END_OF_FUNC

#----------------------------------------
'save_email_list' => <<'END_OF_FUNC',

use Acme::SQL;
use Acme::Db;
use Acme::Lister::Filter;

sub save_email_list {
	my ($path, $email_field, $name_field) = @_;

	#
	# Export list into a mail queue.
	#

	my $count = 0;		# Returned number of records saved.

	our $Separator = Acme::Config->Export_Separator;

	#
	# Set key field for source sql (table) and target temporary table.
	#

	$key = "$table.$key" if $table && $key !~ /\./; # fully qualify key

	(my $temp_key = $key) =~ s/.*\.//;	# strip table name

	#
	# Set up list filter
	#

	my $fltr = new Acme::Lister::Filter;

	my $temp_table = $fltr->temp_table;

	#
	# Create default sql to search the table if needed.
	#

	if ($table && ! $sql) {

		# Note: this sql should match the sql in list.cgi
		# with the exception of the columns selected.

		$sql = $t->email_list_sql;
	}

	my $sqlo = new Acme::SQL $sql;

	$sqlo->add_join("$table join $temp_table on $key = $temp_table.$temp_key");

	#
	# Make format string and header record to use to print queue file.
	#

	my @columns = $sqlo->columns;

	my @header = @columns;			# make header record
	##s/.*\.// for @header;			# strip off table names
	unshift @header, 'name';		# push on the "name" field
	unshift @header, 'email';		# push on the "email" field as 1st

	my $format = join $Separator, map { "%s" } @header;

	#
	# Set columns so that the email address and name to use are the first
	# and second fields in the record.
	#

	# Note: there must be an email field provided above otherwise a blank
	# field is used by sendmaillist.pl causing email not to be sent unless
	# one of the database columns selected is named "email".

	$name_field  ||= "''";
	$email_field ||= "''";

	$name_field  =~ s/\s*,.*//;	# use only first field in list
	$email_field =~ s/\s*,.*//;	# use only first field in list

	$email_field = "$table.$email_field" unless $email_field =~ /\./;
	$name_field  = "$table.$name_field"  unless $name_field =~ /\./;

	unshift @columns, $name_field;
	unshift @columns, $email_field;

	$sqlo->set_columns(@columns);

	$sqlo->sort($sort) if $sort;

	for my $col (split /\s*,\s*/, $sqlo->order_by) {
		($_ = $col) =~ s/ desc$//i;	# remove descending indicator of sort columns
		$sqlo->add_column($_) unless $sqlo->is_column($_);
	}

	$limit ||= $t->limit if $t && $t->limit;
	$sqlo->set_limit($limit) if $limit;

	#
	# Redefine sql with filters and exported columns.
	#

	$sql = $fltr->as_sql . $sqlo->as_sql;

	#
	# Check sql with security.
	#

	$sql = Acme::Security::clean_sql($sql);

	#
	# Open output file before doing query
	#

	open(QUEUE, '>', $path)
	  or die "Cannot open email queue file '$path'";

	#
	# Do query
	#

	my $q = Acme::Db->query($sql);

	#
	# Print header
	#

	no warnings qw(uninitialized);

	local $\ = "\n";	# terminate print lines with LF.

	print QUEUE sprintf($format,@header);	# print header record

	#
	# Print lines
	#

	while (my @a = $q->next) {

		if ($Separator) {

			#
			# For CSV type files, quote values containing the delimiter
			# and double any quotes found inside quotes.
			#

			@a = map { /$Separator/ ? map { s/"/""/g; qq["$_"]} $_ : $_ } @a;
		}

		print QUEUE sprintf($format,@a);

		$count++;

		#
		# Check list limits
		#

		last if $limit && $count > $limit;
	}

	close(QUEUE);

	return $count;
}
END_OF_FUNC

#----------------------------------------
'quoted_printable' => <<'END_OF_FUNC',
sub quoted_printable {

	# Return argument as a quoted-printable MIME type.

# Use following if present on your system
#use MIME::QuotedPrint;
#return encode_qp(shift);

	# The following was stolen from MIME::QuotedPrint.pm by Gisle Aas.

    my $res = shift;
    my $eol = shift;
    $eol = "\n" unless defined($eol) || length($eol);
        $res =~  s/([^ \t\n!"#\$%&'()*+,\-.\/0-9:;<>?\@A-Z[\\\]^_`a-z{|}~])/sprintf("=%02X", ord($1))/eg;  # rule #2,#3
        $res =~ s/([ \t]+)$/
          join('', map { sprintf("=%02X", ord($_)) }
    		   split('', $1)
          )/egm;                        # rule #3 (encode whitespace at eol)

    # rule #5 (lines must be shorter than 76 chars, but we are not allowed
    # to break =XX escapes.  This makes things complicated :-( )
    my $brokenlines = "";
    $brokenlines .= "$1=$eol"
	while $res =~ s/(.*?^[^\n]{73} (?:
		 [^=\n]{2} (?! [^=\n]{0,1} $) # 75 not followed by .?\n
		|[^=\n]    (?! [^=\n]{0,2} $) # 74 not followed by .?.?\n
		|          (?! [^=\n]{0,3} $) # 73 not followed by .?.?.?\n
	    ))//xsm;

    return "$brokenlines$res";
}
END_OF_FUNC
);
}

#------------------------------------------------------------------------------

sub AUTOLOAD {
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;
	return if $method eq 'DESTROY';		# gobble

	our %Subs;
	define_subs() if ! defined %Subs;

	if (exists $Subs{$method}) {
		my $s = $Subs{$method};
		delete $Subs{$method};		# release memory

		if ($s) {
			eval "package $pkg; $s";
			die "${pkg}::$method(): $@" if $@;
			goto &$method;
		}
	}
	die "Unknown method: $AUTOLOAD";
}
