#!/usr/bin/perl -wT -I../../lib

=head1 NAME

export.cgi -- Export a list

=head1 SYNOPSIS

    export.cgi?table=customers&search=.*

=head1 DESCRIPTION

The export.cgi program exports a list of records that has been reviewed by the
user in a list window. The user sees a "Download File" dialog box just like
they would when they download a regular file from the server.

The arguments include those used to select and render a list, e.g. the
table name and search script. They also include extra arguments from the
list window, which include the field name used as a filter and lists of ID
numbers that were specifically included or excluded from the list by the user.

Several types of exports can be defined in the schema files for table exports.
The schema can also define C<beforeExport>, C<onExport>, and C<afterExport>
clauses that execute before, during, and after an export. 

Exports are monitored. A log is kept of all exports. An administrator is
also emailed when an export has occured. All security settings can be set
in the AcmeBase schema files.

=head1 CHANGE LOG

 2/20/02 Initial version.
 8/24/02 Converted to new list.cgi. Added onExport, ... clauses.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

$|=1;

package export;

require 5.6.0;

use Acme::HTML			qw(escape_html		display_message);
use Acme::HTML::Parameters;
use Local::Acme::User		qw(user);
use Acme::Config;
use Text::ParseWords;
use CGI;

use Acme::Security;
Acme::Security::check('Access');

$|=1;		# flush output (synchronize errors with output)

our $user = user();

#
# Parse cgi arguments
#

our $cgi = new CGI;

our $table	 = $cgi->url_param('table')      || $cgi->param('table') || '';
our $key	 = $cgi->url_param('key')        || $cgi->param('key')   || '';
our $title	 = $cgi->param('title')          || '';
our $confirm	 = $cgi->param('confirm')        || '';

our $n_checked	 = $cgi->param('n_checked')      || 0;
our $n_unchecked = $cgi->param('n_unchecked')    || 0;

our $export_file_type = $cgi->url_param('export_file_type') ||
			$cgi->param('export_file_type')     || '';

if (! $n_checked || $n_checked <= 0) {
	print $cgi->header(-type=>'text/html', -expires=>'-1d') unless @ARGV;
	display_message("No records are selected for export.", "$title",undef,0);
	exit 1;
}


#
# Save original table parameter for logs.
#

our $export_table = $table;

#
# Derive table name form key parameter if table not specified.
#

($table) = $key =~ /(\w+)\./ if $key && ! $table;

#
# Check export security
#

our $schema = new Acme::Base::Schema;
our $t;

if ($table) {
	$t = $schema->t($table);

	# Check alternate security codes
	my $sc = $t->export_security_code;
	Acme::Security::check('Access',$sc) if $sc;

	die "Cannot export table '$table'. The option has been turned off in the schema tables"
	  unless $t->export;
}
else {
	Acme::Security::check('Access','SQL_EXPORT');

	die "This site does not allow the export of SQL queries"
	  unless Acme::Config->Can_Export_SQL;
}
	
if (! $confirm) {

	#
	# Get confirmation
	#

	$cgi->cache('no');
	print $cgi->header(-type=>'text/html', -expires=>'-1d') unless @ARGV;

	get_parameters();
	exit;
}

if ('yes' =~ /\Q$confirm\E/i) {

	#
	# Export records
	#

	do_export();
	exit;
}

#
# Export was not confirmed.
#

display_message("No records were exported.");
exit;

#------------------------------------------------------------------------------

sub get_parameters {
	my $added_html = shift || '';

	#
	# Define standard form elements on parameter screen
	#

	my $select_str  = "records are selected for export.";
	my $exclude_str = "records are excluded from the export list.";

	# Note: the values for the drop-down list of export types are file
	# extensions used to propose a download file name for the user.
	# In other words, "csv" is used to propose a file name like
	# "myfilename.csv".

	my $attr = q[onChange="this.form.action=this.form.action.replace(/(.*)\\..*$/,'$1.'+this.value)"];
	my $export_file_type = Acme::HTML::ListMenus->Menu(
					{
					 csv => 'CSV (comma delimited) (*.csv)',
					 txt => 'Text (tab delimited) (*.txt)',
					}, 'export_file_type',undef,undef,$attr);

	#
	# Define optional form elements for parameter screen
	#

	my $export_type_html = '';
	my $instructions = '';
	if ($t) {
		if (my $types = $t->export_types) {
			my %opt = Text::ParseWords::parse_line('[,;\s]+', 0, $types);
			my $export_type = Acme::HTML::ListMenus->Menu(
						\%opt,
						'export_type'
						);

			my $export_type_label = $t->export_type_label || 'Export Type';

			$export_type_html = <<End_Of_HTML;
<tr>
<td class="label"><nobr>$export_type_label</nobr></td>
<td class="field">$export_type</td>
</tr>
<tr><td colspan="2">&nbsp;</td></tr>
End_Of_HTML
		}

		$instructions = $t->export_instructions;
	}

	#
	# Define parameter form
	#

	my $html = <<End_Of_HTML;
<div style="margin:2ex;text-align:center">

<table cellspacing=0 cellpadding=2 border=0 style="margin-bottom:2ex">
<tr>
<th>$n_checked</th>
<td>$select_str</td>
</tr>
<tr>
<th>$n_unchecked</th>
<td>$exclude_str</td>
</tr>
</table>

<table cellspacing=0 cellpadding=2 border=0 style="margin-top:3ex">
$export_type_html
<tr>
<td class="label"><nobr>Export File Type</nobr></td>
<td class="field">$export_file_type</td>
</tr>
<tr>
</table>
<div align="center" style="margin:5ex">
<input type="hidden" name="confirm" value="no">
<input type="submit" value="Begin Export" class="button" onClick="this.form.confirm.value='yes'">
&nbsp;
<input type="reset"  value="Close Window" class="button" onClick="window.close()">
</div>
</div>
$added_html
End_Of_HTML

	#
	# Make URL that will call ourselves again with new parameters.
	# Append proposed file name as final component of path name.
	#
	# Note: for a URL like:
	#
	#   http://www.acmebase.com/cgi-bin/export.cgi/myfilename.txt
	#
	# Apache calls the export.cgi program with "myfilename.txt" in the
	# PATH_INFO environment variable. Browsers downloading a file
	# use the last component of the URL as a candidate name for
	# the downloaded file. By returning a Mime type of text/csv
	# or text/tab-separated-value, the browser will display the
	# "Download File" dialog box to the user using the "myfilename.txt"
	# component of the URL as the proposed name of the downloaded file.
	#

	my $url = $cgi->url(-full=>1, -path=>1);

	# Append candidate file name to current URL. Use the
	# current date as part of the proposed file name.
	my ($d,$m,$y) = (localtime)[3,4,5];
	$url .= '/'
	  . ($table || 'export')
	  . '-'
	  . sprintf("%4d%02d%02d",$y+1900,$m+1,$d)
	  . '.csv';

	$url .= '?' . $ENV{QUERY_STRING}
	  if $ENV{REQUEST_METHOD} eq 'POST' && $ENV{QUERY_STRING};

	#
	# Display parameter form
	#

	print Acme::HTML::Parameters->get(
				cgi	     => $cgi,
				url          => $url,
				title        => $title,
				html         => $html,
				instructions => $instructions,
				window_title => 'Confirm Export',
				submit       => undef,  # no button in header
				);
}

#------------------------------------------------------------------------------

use Acme::SQL;
use Acme::Db;
use Acme::Base::Schema;
use Acme::Base::Schema::Join	qw(parse_join);
use Acme::Lister::Filter;

sub do_export {

	our $sql                 = $cgi->param('sql');
	our $search              = $cgi->param('search');
	our @columns             = $cgi->param('columns');
	our $list_column_headers = $cgi->param('list_column_headers')	|| '';
	our $list_as_rows        = $cgi->param('list_as_rows')		|| '';
	$list_as_rows = '' unless 'yes' =~ /\Q$list_as_rows\E/i;

	our @sort_columns        = $cgi->param('sort_columns');
	our $sort                = $cgi->param('sort')			|| '';
	$sort ||= join ", ", @sort_columns if @sort_columns;

	our $limit               = $cgi->param('limit')		|| 0;
	$limit = $t->limit
	  if $t && $t->limit && (! $limit || $t->limit < $limit);

	die "No columns to export" unless @columns;

	$key = "$table.$key" if $table && $key !~ /\./; # fully qualify key

	(my $temp_key = $key) =~ s/.*\.//;	# strip table name

	#
	# Get beforeExport, onExport, and afterExport statements from schema.
	#

	my ($beforeExport, $onExport, $afterExport);
	if ($t) {
		
		#
		# Event statements defined for export_type override general
		# event statements.
		#

		my $export_type	= $cgi->param('export_type');

		$beforeExport	= ($export_type && $t->property("beforeExport_$export_type"))
				|| $t->beforeExport;

		$onExport	= ($export_type && $t->property("onExport_$export_type"))
				|| $t->onExport;

		$afterExport	= ($export_type && $t->property("afterExport_$export_type"))
				|| $t->afterExport;
	}

	#
	# Setup filter from CGI
	#

	my ($fltr, $temp_table);

	if ($t) {
		$fltr = new Acme::Lister::Filter;

		$temp_table = $fltr->temp_table;
	}

	#
	# Setup object to assemble SQL
	#

	# Note: setting up an object to assemble the sql query allows the
	# sql() function below to be used in beforeExport, onExport, or
	# in afterExport event statements defined in the schema.

	local our $sqls = new export::SQL;

	#
	# Create default sql to search the table if needed.
	#

	if ($table && ! $sql) {

		# Note: this sql should match the default sql in list.cgi.

		$sql = $t->list_sql if $t->defined('list_sql');

		if (! $sql) {

			# Construct query here if not explicity defined in schema.
			# This allows customized list of fields such as the
			# ListAllColumns option.

			my ($join_expr,$join_clause);

			if ($list_as_rows) {
				# list as rows: use defined list_join property.
				($join_expr,$join_clause) = parse_join($table,$t->list_join);
			}
			else {
				# build custom list of table joins
				($join_expr,$join_clause) = parse_join($table,
				  (join " and ", grep {$_} $t->join, $t->master, $t->list_join));
			}

			$sql = 'select'
			  . ' distinct ' . (join ",\n", @columns)
			  . ' from '	 . $join_expr
			  . ' where '	 . $t->list_where . $join_clause
			  . ' order by ' . $t->list_order_by_clause;

		}
	}

	#
	# Change main SQL to select only chosen columns.
	#

	my $sqlo = new Acme::SQL $sql;

	$sqlo->set_columns(@columns);

	#
	# Set sort order if specified.
	#

	$sqlo->sort($sort) if $sort;

	#
	# Make sure sort columns are listed the select clause.
	#

	# Note: normally this is/should not be neccessary. However, to ensure
	# our queries return only one record per key value, we construct the
	# sql as "select distinct ...". Postgres will allow sorts to be
	# performed on a non-listed column, but not when the "distinct" keyword 
	# is used.

	for my $col (split /\s*,\s*/, $sqlo->order_by) {
		($_ = $col) =~ s/ desc$//i;	# remove descending indicator of sort columns
		$sqlo->add_column($_) unless $sqlo->is_column($_);
	}

	#
	# Set limit if defined.
	#

	$sqlo->set_limit($limit) if $limit;

	#
	# Redefine $sql with filters.
	#

	if ($fltr && (my $fltr_sql = $fltr->as_sql)) {

		#
		# Add filter SQL to assembled SQL.
		#

		$sqls->add($fltr_sql);

		#
		# Join temporary table.
		#

		$sqlo->add_join("$table join $temp_table on $key = $temp_table.$temp_key");

		$sql = $sqlo->as_sql;
	}
	elsif ($search) {

		#
		# Combine sql with search script.
		#

		# Without a filter (which requires a table name and key to 
		# be specified), we need to combine a search script with
		# the sql. This happens when an arbitrary SQL statement is
		# entered (no table).

		my $srch = new Acme::Searcher $search, $cgi;	# parse search script

		$sql = $srch->as_sql($sqlo);	

		if ($sort) {

			# Need to parse sql again to override the sort in
			# search script.

			$sqlo = new Acme::SQL;
			$sqlo->sort($sort);
			$sql = $sqlo->as_sql;
		}
	}
	else {
		$sql = $sqlo->as_sql;
	}

	#
	# Add main SQL statement to assembled SQL.
	#

	$sqls->add($sql);

	#
	# Clean combined SQL
	#

	$sql = Acme::Security::clean_sql($sqls->as_sql);

	#
	# Do query
	#

	my $dbh = Acme::Db->connect();	# connect to database

	my $q = $dbh->query($sql);

	$q->bind_columns_to_variables;

	# Retreive row count to trigger execution of query and produce any
	# error messages before presenting the File Download dialog box.

	my $count = $q->count;	

	#
	# Save a log of the export
	#

	log_export();

	log_alert();

	#
	# Make appropriate format and produce HTTP header for file download.
	#

	my (@a, $format, $delimiter);

	if ($export_file_type eq 'csv') {
		$delimiter = ',';
		print "Content-Type: text/csv\n\n";
	}
	elsif ($export_file_type eq 'txt') {

		$delimiter = "\t";
		print "Content-Type: text/tab-separated-value\n\n";
	}
	else {
		$delimiter = "\t";
		print "Content-Type: text/plain\n\n";
	}

	#
	# Make format string for printing rows (%s,%s,%s,...)
	#

	$format = join $delimiter, map { "%s" } @columns;

	#
	# Trigger beforeExport
	#

	my $do;
	if ($beforeExport && ($do = prep_do($beforeExport,$table))) {

		# Create new sql object to separate beforeExport
		# clause from export query.

		$sqls = new export::SQL;	# start new sql statment

		# Note: taint checks are subverted here.
		($do) = $do =~ /^(.*)$/s;       # untaint

		no strict;
		eval $do;
		die "Error evaluating schema's beforeExport property: $@" if $@;

		#
		# Execute sql first before export query if defined.
		#

		if (my $sql = $sqls->as_sql) {
			$dbh->execute($sql)
			  or die "Errors occurred in beforeExport clause";
		}
	}

	#
	# Export file
	#

	no warnings qw(uninitialized);
	local $\ = "\r\n";	# terminate print lines with CR/LF for Windows.

	print sprintf($format,@a) if @a = column_headings();	# print header

	# Create new sql object to accumulate onExport SQL statements.
	$sqls = new export::SQL;	# start new sql statment

	(my $keyvar = $key) =~ s/\W/__/g;	# change periods to __
	my $keyval;

	$count = 0;
	while ($q->next) {

		$count++;
		
		#
		# Check list limits
		#

		last if $limit && $count > $limit;

		#
		# Trigger onExport
		#

		{ no strict qw(refs); $keyval = $$keyvar; }

		if ($onExport && ($do = prep_do($onExport,$table,"$key = '$keyval'"))) {
			# Note: taint checks are subverted here.
			($do) = $do =~ /^(.*)$/s;       # untaint

			no strict;
			eval $do;
			die "Error evaluating schema's onExport property: $@" if $@;
		}

		#
		# Quote values containing the delimiter and double the quotes
		# inside quotes.
		#

		@a = map { /$delimiter/ ? map { s/"/""/g; qq["$_"]} $_ : $_ } $q->rec
		  if $delimiter;

		#
		# Export record
		#

		print sprintf($format,@a);
	}

	#
	# Execute onExport SQL statements accumulated in loop above if defined.
	#

	if (my $sql = $sqls->as_sql) {

		$dbh->execute($sql) or die "Errors occurred in onExport clause";
	}

	#
	# Trigger afterExport
	#

	if ($afterExport && ($do = prep_do($afterExport,$table,"$key in (select $temp_key from $temp_table)"))) {
		# Create new sql object to separate beforeExport
		# clause from export query.

		$sqls = new export::SQL;

		# Note: taint checks are subverted here.
		($do) = $do =~ /^(.*)$/s;       # untaint

		no strict;
		eval $do;
		die "Error evaluating schema's afterExport property: $@" if $@;

		#
		# Execute sql first before export query if defined.
		#

		if (my $sql = $sqls->as_sql) {
			$dbh->execute($sql)
			  or die "Errors occurred in afterExport clause";
		}
	}
}

#------------------------------------------------------------------------------

sub log_export {

	#
	# Write to a log file
	#

	my $log_file = $t->export_log if $t;
	$log_file = Acme::Config->Export_Log
	  if !$log_file || $log_file =~ /^yes$/i;

	return unless $log_file;

	my $file_exists = (-f $log_file);

	open(LOG, '>>', $log_file)
	  or die "Cannot open log file '$log_file'. $!";

	no warnings qw(uninitialized);

	my $time = localtime;
	my $user = user();
	my $host = $ENV{SERVER_NAME};

	my $total = $n_checked + $n_unchecked;
	my $source = $export_table ? "$export_table table" : "SQL:\n$sql";

	print LOG <<End_Of_Log;

$user\@$host: $time: Export to $export_file_type file at $ENV{REMOTE_ADDR}
	$n_checked records of $total queried were exported from $source.
End_Of_Log

	close(LOG);

	chmod 0622, $log_file if ! $file_exists;  # allow others to write to file
}

#------------------------------------------------------------------------------

sub log_alert {

	#
	# Send email alert to administrator.
	#

	our $Webmaster        = Acme::Config->Webmaster;
	our $Webadministrator = Acme::Config->Webadministrator;

	my $email;
	if ($t) {
		$email = $t->export_alert or return;
		return if ! $email && $export_table;	# alert not wanted
	}
	if (! $export_table) {
		$email = Acme::Config->Alert_Export_SQL or return;
		return if $email =~ /^no$/i;
	}
	$email = "$Webadministrator; $Webmaster"	# default email address
	  if $email =~ /^yes$/i;

	#
	# Send log message
	#

	no warnings qw(uninitialized);

	my $time = localtime;
	my $user = user();
	my $host = $ENV{SERVER_NAME};

	my $total = $n_checked + $n_unchecked;
	my $source = $export_table ? "$export_table table" : "SQL:\n$sql";

	delete @ENV{ qw(IFS CDPATH ENV BASH_ENV) };     # untaint
	$ENV{PATH} = '/bin:/usr/bin';                   # untaint

	if (open(SENDMAIL,"| /usr/sbin/sendmail -oi -t")) {
		my $to = $email;
		$subject = "Export notification: $user\@$host is exporting "
		  . "$n_checked "
		  . ($table || 'SQL records.');
		print SENDMAIL <<End_Of_Msg;
From: $Webmaster
To: $to
Subject: $subject
Mime-Version: 1.0
Content-type: text/plain


$user\@$host: $time: Export to $export_file_type file at $ENV{REMOTE_ADDR}
	$n_checked records of $total queried were exported from $source.
End_Of_Msg
		close(SENDMAIL);
	}

}

#------------------------------------------------------------------------------

# Note: the following routines for printing column headers are copied from
# the list.cgi program, except they return an array of headers, not a print line.

sub column_headings {
	
	# Construct <td> tags with title attributes of field descriptions.

	my @headers;		# returned array of column headers

	my ($f,$head,$desc);
	my ($tab,$fld);

	# Track previous table objects $t to reduce calls to $schema->t().
	my $tt     = $t;
	my $ttable = $table;

	for my $field (@columns) {
		($tab,$fld) = $field =~ /(\w+)\.(\w+)/ 
		  or ($tab,$fld) = ($table,$field);
		
		if ($tab ne $ttable) {
			$tt     = $schema->t($tab);
			$ttable = $tab;
		}

		#
		# Make header for each column
		#

		if ($list_column_headers && $list_column_headers !~ /none/i) {
			push @headers, column_name($tt,$fld);
		}
	}

	return @headers;
}

sub column_name {
	my ($tt,$fld) = @_;

	#
	# Return the printed name for the column depending upon the list
	# headings option.
	#

	if ($list_column_headers =~ /FieldTitles/i) {
		my $head = '';
		if ($tt) {
			# Get titles from schema
			if (my $f = $tt->f($fld)) {
				$head = $f->title;
			}
			elsif ($fld eq $tt->audit_entry
			    || $fld eq $tt->audit_update
			    || $fld eq $tt->audit_user) {
				# Change underlines to spaces and
				# capitalize words.
				$head = $fld;
				$head =~ s/_/ /g;
				$head =~ s/\b(\w+)\b/ucfirst(lc($1))/eg;
			}
			else {
				$head = $tt->id_title || '';
			}
		}
		else {
			# Make titles out of field names
			# Change underlines to spaces and
			# capitalize words.
			$head = $fld;
			$head =~ s/_/ /g;
			$head =~ s/\b(\w+)\b/ucfirst(lc($1))/eg;
			$head =~ s/\bId\b/ID/; # capitalize id
		}
		return $head;
	}

	if ($list_column_headers =~ /QualifiedFieldNames/i) {
		if ($tt) {
			return $tt->name . ".$fld";
		}
		else {
			return $fld;
		}
	}

	if ($list_column_headers =~ /FieldNames/i) {
		return $fld;
	}

	return $fld;
}

#-------------------------------------------------------------------------------

sub prep_do {
	my ($do,$table,$where) = @_;

	return unless $do;

	#
	# Convert "do" property shortcuts and SQL into Perl code.
	#

	# Note: The sandbox where the Perl code is executed has a subroutine
	# sql() defined in it which is used to assemble all the sql statements
	# together to be submitted as one transaction to the database.

	my ($first_word) = $do =~ /^\s*(\w+)/;

	KEYWORD: {

		# Keyword SET updates the current database record.
		if ($first_word =~ /^set$/i) {
			$do = qq[sql(q[update $table $do where $where]);] if $where;
			last KEYWORD;
		}

		# SQL keywords are registered to be executed by the database.
		if ($first_word =~ /^(insert|update|delete)$/i) {
			# The $do clause is a full SQL statement.
			$do = "sql(q[$do]);";
			last KEYWORD;
		}

		# Anything else is Perl code and returned intact.
	}
	
	return $do;
}

sub sql {

	#
	# Assemble sql statement from arguments.
	#

	# This is a helper function for the "do" statements beforeExport,
	# onExport, and afterExport Perl code from schema files. It allows
	# the Perl code to add to the SQL statements being assembled.

	our $sqls;

	$sqls->add(@_);
}

#-------------------------------------------------------------------------------
#
# Module to build SQL statement
#
#-------------------------------------------------------------------------------

package export::SQL;

sub new { 
	my $this = shift;
	my $class = ref($this) || $this;

	my $self = {
		sql => '',		  # Accumulated SQL statements.
		@_
		};

	return bless $self, $class;
}

sub add {
	my $self = shift;

	# Separate statements with semi-colons.
	$self->{sql} .= join "\n", (map { /;\s*$/ ? qq[$_] : qq[$_;] } @_), '';
}

sub as_sql {
	my $self = shift;
	return $self->{sql};
}
