#!/usr/bin/perl -wT -I../../lib

=head1 NAME

save_list.cgi -- Save a list file

=head1 SYNOPSIS

    list/save_list.cgi?list_id=customers&file=my_list_file.txt

=head1 DESCRIPTION

The save_list.cgi program saves a list of the database record's key values
in a text file. The lists can later be retreived and viewed, and they can
be compared with each other.

Lists files are different from export files in that only the table's key
value is saved in a list file. The location of the list file determines
what database table the list is for.

Only the records that were "selected" (e.g. "checked") in the list window
are saved in the list file.

=head1 CHANGE LOG

 3/2000   Initial version.
 7/27/02  Generalized for AcmeBase.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000-2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

$|=1;

require 5.6.0;

use Acme::HTML			qw(escape_html		display_message);
use Acme::HTML::Parameters;
use Acme::Base::Schema;
use Acme::Lister::Filter;
use Acme::Config;
use Acme::File			qw(mktemp);
use Local::Acme::User		qw(user);
use CGI;

use Acme::Security;
Acme::Security::check('Access');

$|=1;		# flush output (synchronize errors with output)

our $List_Space	    = Acme::Config->List_Space;
our $List_Extension = Acme::Config->List_Extension;

#
# Parse cgi arguments
#

our $cgi = new CGI;

our $table	= $cgi->url_param('table')   || $cgi->param('table') || '';
our $title	= $cgi->param('title')       || '';

our $list_id	= $cgi->url_param('list_id') || $cgi->param('list_id') || $table || 'SQL';

our $list_name	= $cgi->param('list_name');	# required file name

our $n_checked	= $cgi->param('n_checked')   || 0;
our $n_unchecked= $cgi->param('n_unchecked') || 0;

if (! $n_checked || $n_checked <= 0) {
	$cgi->cache('no');
	print $cgi->header(-type=>'text/html', -expires=>'-1d') unless @ARGV;
	display_message("No records are selected to save.", "$title",undef,0);
	exit 1;
}

our ($schema, $t);
$schema = new Acme::Base::Schema;

if ($table) {
	$t = $schema->t($table);

	# Check alternate security codes
	my $sc = $t->export_security_code;
	Acme::Security::check('Access',$sc) if $sc;
}
	
if (! $list_name) {

	#
	# Get parameters
	#

	$cgi->cache('no');
	print $cgi->header(-type=>'text/html', -expires=>'-1d') unless @ARGV;

	get_parameters();
	exit;
}

#
# Save list
#

print $cgi->header(-type=>'text/html', -expires=>'-1d') unless @ARGV;

do_save_list();
exit;

#------------------------------------------------------------------------------

sub do_save_list {

	#
	# Remove illegal characters from list name
	#

	$list_name =~ s{[/<>|'`]}{ }g;
	($list_name) = $list_name =~ /(.*)/;	# untaint

	($list_id) = $list_id =~ /(\w*)/;	# untaint

	$list_name .= $List_Extension unless $list_name =~ /\./;

	#
	# Append or overwrite list file
	#

	my $append = $cgi->param('append') =~ /yes/i;

	my $unload_file;
	my $target_file = "$List_Space/$list_id/$list_name";
	($target_file) = $target_file =~ /(.*)/;	# untaint

	if ($append) {
		$unload_file = mktemp(user());
	} else {

		#
		# Create file first, then change permissions to world writable
		# so the database engine can write to the file
		#

		if (! -f $target_file) {
			# Create file under current user id (not database id)
			open FILE, ">$target_file" and close FILE;
		}

		$unload_file = $target_file;
	}
	chmod 0666, $target_file;	# allow database engine to write to file


	my $fltr = new Acme::Lister::Filter;

	my $unload_sql = '';
	my $dbengine = $schema->dbengine;
	if ($dbengine =~ /Postgres/i) {
		$unload_sql =
		    qq[copy ] . $fltr->temp_table
		  . qq[ to '] . $unload_file . qq[';];
	}
	elsif ($dbengine =~ /MySQL/i) {
		$unload_sql = 
		    qq[select ] . $fltr->temp_key
		  . qq[ from ] . $fltr->temp_table
		  . qq[ into outfile '] . $unload_file . qq[';];
	}
	elsif ($dbengine =~ /Informix/i) {
		$unload_sql = 
		    qq[unload to '] . $unload_file . qq[']
		  . qq[ select ] . $fltr->temp_key
		  . qq[ from ] . $fltr->temp_table . qq[;];
	}

	#
	# Save file
	#

	Acme::Db->execute($fltr->as_sql . $unload_sql);

	# Append to target file if flag set.
	$action_text = "saved";			# for user message
	if ($append) {
		open IF, '<',  $unload_file;
		open OF, '>>', $target_file;
		while (<IF>) { print OF; }
		close OF;
		close IF;
		unlink "$unload_file";
		
		$action_text = "appended";	# for user message
	}

	#
	# Change protections of target file back to what is desired now
	# that the database engine is done trying to write to it.
	#

	chmod 0644, $target_file;

	#
	# Display results screen
	#

	display_message("The list has been $action_text to $list_name.",
			"Message",
			"save_cleanup()");

}

#------------------------------------------------------------------------------

sub get_parameters {

	my $u = user();
	my $default_name = "${u}_list";		# default list file name

	#
	# Define standard form elements on parameter screen
	#

	my $select_str  = "records are selected to save.";
	my $exclude_str = "records are excluded from the list.";

	#
	# Define parameter form
	#

	my $script = <<End_Of_JavaScript;
<script language="JavaScript">
function save_as(filename) {
	var fld = self.document.parameters.list_name;
	fld.value = filename;
	window.scrollTo(0,0);	// show buttons
	fld.focus();
}
</script>
End_Of_JavaScript

	#
	# Define parameter form
	#

	my $html = <<End_Of_HTML;
<div style="margin:2ex;text-align:center">
<table cellspacing="0" cellpadding="2" border="0">

<tr>
<td align="right"><font size="+1">$n_checked</font></td>
<td>$select_str</td>
</tr>
<tr>
<td align="right"><font size="+1">$n_unchecked</font></td>
<td>$exclude_str</td>
</tr>

<tr>
<th nowrap align="right" width="10%">Enter List Name</th>
<td nowrap><input type="text" name="list_name" value="$default_name" style="width:100%" onFocus="this.select()">
</td>
</tr>
<tr>
<td colspan="2">
You can append to or overwrite an existing list.
<br><input type=radio name=append value="yes" checked> Append (add) to the file
<br><input type=radio name=append value="no"> Overwrite the file
</td>
</tr>

</table>
</div>
End_Of_HTML

	#
	# Get list of list files
	#

	($list_id) = $list_id =~ /(\w*)/;	# untaint
	mkdir "$List_Space/$list_id", 0777 unless -d "$List_Space/$list_id";
	opendir LISTS, "$List_Space/$list_id"
	  or die "Can't open saved lists in $List_Space/$list_id. $!";
	my @list_files = readdir LISTS;
	closedir LISTS;
	shift @list_files; shift @list_files;	# remove . and ..

	#
	# Print listing header
	#

	(my $list_title = $list_id) =~ s/\b(\w+)\b/ucfirst(lc($1))/eg; # capitalize

	$html .= <<End_Of_HTML;

<h2>$list_title Lists</h2>
<ul class="file-list">
End_Of_HTML

	if (scalar(@list_files) == 0) {
		$html .= "(empty)";
	}
	else {
		#
		# Print listing, each line a link to a routine to load the file
		#

		my $encoded_file;
		for (sort { lc($a) cmp lc($b) } @list_files) {

			$encoded_file = escape_html($_);

			next if /^\./;	# skip hidden files

			$html .= "<li>"
			  . "<a href=\"javascript:save_as('$encoded_file')\""
			  . ">$_</a>";
		}
	}
	$html .= <<End_Of_HTML;
</ul>

</div>
End_Of_HTML

	#
	# Display parameter form
	#

	print Acme::HTML::Parameters->get(
				cgi	=> $cgi,
				title   => $title,
				html    => $html,
				script  => $script,
				window_title => 'Save List',
				submit  => '  Save  ',
				);
}
