#!/usr/bin/perl -wT -I../../lib
require 5.6.0;
use strict;

=head1 NAME

save_report.cgi -- Save a report file

=head1 SYNOPSIS

    list/save_report.cgi?list_id=customers&file=my_report_file.txt

=head1 DESCRIPTION

The save_report.cgi program saves the current list settings in a file that
can be retreived later.

When rendering a saved report, if the list_id of the current list matches
the list_id of the saved report, the current list's search string is used
instead of the search embedded in the report file.

=head1 CHANGE LOG

 11/27/02 Initial version.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Acme::HTML			qw(escape_html		display_message);
use Acme::HTML::Parameters;
use Acme::Base::Schema;
use Acme::Config;
use Local::Acme::User		qw(user);
use CGI;

use Acme::Security;
Acme::Security::check('Access');

$|=1;		# flush output (synchronize errors with output)

our $Report_Space     = Acme::Config->Report_Space;
our $Report_Extension = Acme::Config->Report_Extension;
our $Script_Separator = Acme::Config->Script_Separator || ';';

#
# Parse cgi arguments
#

our $cgi = new CGI;

our $table	= $cgi->url_param('table')   || $cgi->param('table')  || '';
our $search	= $cgi->url_param('search')  || $cgi->param('search') || '';
our $title	= $cgi->url_param('title')   || $cgi->param('title')  || '';

our $list_id	= $cgi->url_param('list_id') || $cgi->param('list_id') || $table || 'SQL';

our $report_name= $cgi->param('report_name');	# required file name

if (! $report_name || ! exists $cgi->{'overwrite'}) {

	#
	# Get parameters
	#

	$cgi->cache('no');
	print $cgi->header(-type=>'text/html', -expires=>'-1d') unless @ARGV;

	get_parameters();
	exit;
}

my $filepath = make_target_file_path();

if (-e $filepath) {
	die Acme::Security::alert "You do not have privileges to overwrite $report_name."
	  unless Acme::Security::allow('Update',"$list_id/$report_name");

	my $overwrite = $cgi->param('overwrite') || 'no';
	unless ($overwrite =~ /yes/i) {
		print $cgi->header(-type=>'text/html', -expires=>'-1d') unless @ARGV;
		print Acme::HTML::Parameters->confirm(
			cgi	=> $cgi,
			title   => 'Save Report',
			message => "The file <code>$report_name</code> already exists.",
			prompt  => "Do you want to overwrite it?",
			answer  => 'overwrite',
			);

		exit;
	}
}

#
# Save report
#

print $cgi->header(-type=>'text/html', -expires=>'-1d') unless @ARGV;

do_save_report($filepath);

exit;

#------------------------------------------------------------------------------

sub make_target_file_path {

	#
	# Return untainted path to target file.
	#

	# Remove illegal characters from report name
	(my $file_path = $report_name) =~ s{[/<>|'`";@\$]}{ }g;

	$file_path .= $Report_Extension unless $file_path =~ /\./;

	my ($id) = $list_id =~ /(\w*)/;			# untaint

	$file_path = "$Report_Space/$id/$file_path";

	($file_path) = $file_path =~ /([^\$\@`'"|;]*)/;	# untaint

	return $file_path;
}

#------------------------------------------------------------------------------

sub do_save_report {
	my $target_file = shift;

	#
	# Ensure necessary parameters are set in cgi before saving.
	#

	#
	# Set report title.
	#

	my $list_title = $cgi->param('list_title');
	if ($search && $list_title && $search =~ /^[^;|]+[;|]/) {

		# The title $list_title is attached to the search script unless
		# it already has one. This allows the report search script to
		# compute and set a title to the report but still allow the
		# user to change the title $list_title manually.
		
		# Attach list_title to search script
		$search = $list_title . $Script_Separator . $search;

		$cgi->delete('list_title');
	}

	#
	# Clear out unnecessary parameters in cgi before saving.
	#

	$cgi->delete('overwrite');
	$cgi->delete('redrawing');
	$cgi->delete('original_search');

	# Remove params used for list selector filter.
	$cgi->delete('exclude');
	$cgi->delete('include');
	$cgi->delete('select_all');
	$cgi->delete('invert_all');
	$cgi->delete('n_rows');
	$cgi->delete('n_checked');
	$cgi->delete('n_unchecked');

	#
	# Save current cgi parameters to a report file.
	#

	open my $opt, '>', $target_file or die "Cannot save to file '$target_file'. $!";
	$cgi->save($opt);
	close $opt;
	
	#
	# Display results screen
	#

	display_message("The report has been saved to $report_name.",
			"Message",
			"save_cleanup()");

}

#------------------------------------------------------------------------------

sub get_parameters {

	my $u = user();
	my $default_name = $report_name || "${u}_report";  # default file name

	# Use list_id and a timestamp for the report_id.
	my ($sec,$min,$hour,$day,$mon,$year) = localtime; $mon+=1; $year+=1900;
	my $report_id = $list_id . '-' .
	  sprintf("%4d%02d%02d%02d%02d%02d%02d",$year,$mon,$day,$hour,$min,$sec,(rand 100));


	my $script = <<End_Of_JavaScript;
<script language="JavaScript">
function save_as(filename) {
	var fld = self.document.parameters.report_name;
	fld.value = filename;
	window.scrollTo(0,0);	// show buttons
	fld.focus();
}
</script>
End_Of_JavaScript

	#
	# Define an empty "overwrite" parameter used as a flag whose
	# existance signifies that this form has been output.
	#

	my $overwrite = <<End_Of_HTML;
<input type="hidden" name="overwrite" value="">
End_Of_HTML

	#
	# Extract original search script lines (which were modified if any
	# parameters were entered by the user) and attach them to the current
	# search script's where clause (which could have been modified by the
	# user applying filters).
	#

	my $original_search = $cgi->param('original_search') || $search;

	my $modified_original_search = $search;

	if ($modified_original_search =~ /[;|]/) {

		# Extract script lines from original search script.
		my ($orig_scripts) = $original_search =~ /.*?[;|](.+)[;|].*$/s;

		my $sep = $search =~ /;/ ? ';' : '|';
		my ($cur_title) = $search =~ /(.*?)${sep}/s;
		my ($cur_where) = $search =~ /.*${sep}\s*(.*)/s;

		# Redefine the separator to be the current configured in site.conf.
		$sep = Acme::Config->Script_Separator || ';'; # either ; or |

		# Ensure scripts use same delimiter
		$orig_scripts =~ s/[;|]\n*/$sep/g if $orig_scripts;	

		# Reassemble the search script.
		$modified_original_search = $orig_scripts
					  ? "$cur_title$sep$orig_scripts$sep$cur_where"
					  : $cur_title && $cur_where
					  ? "$cur_title$sep$cur_where"
					  : "$cur_title$cur_where";
	}

	#
	# Define parameter form.
	#

	my $html = <<End_Of_HTML;
<div style="margin:2ex;text-align:center">
<table cellspacing="0" cellpadding="2" border="0">
<tr>
<th nowrap align="right" width="10%">Enter Report Name</th>
<td nowrap><input type="text" name="report_name" value="$default_name" style="width:100%" onFocus="this.select()">
</td>
</tr>
<tr>
<td colspan="2">
<br><input type="radio" name="report_id" value="$list_id" checked> Allow any search on $list_id to be used with this report.
<br><input type="radio" name="report_id" value="$report_id"> Allow only the following search to be used with this report.
</td>
</tr>
<tr>
<td colspan="2"><textarea name="search" rows="5" style="margin-left:4ex; width:100%">$modified_original_search</textarea></td>
</tr>
</table>
</div>
$overwrite
End_Of_HTML

	#
	# Get list of report files
	#

	($list_id) = $list_id =~ /(\w*)/;	# untaint

	# Make directory if it does not exists.
	mkdir "$Report_Space/$list_id", 0777 unless -d "$Report_Space/$list_id";

	my @report_files = get_file_names("$Report_Space/$list_id")
	  if Acme::Security::allow('Update',$list_id);

	if (@report_files) {
		#
		# Print listing header
		#

		(my $report_title = $list_id) =~ s/\b(\w+)\b/ucfirst(lc($1))/eg; # capitalize

		$html .= <<End_Of_HTML;

<h2>$report_title Reports</h2>
<ul class="file-list">
End_Of_HTML

		#
		# Print listing, each line a link to a routine to load the file
		#

		my $encoded_file;
		for (sort { lc($a) cmp lc($b) } @report_files) {

			next if /^\./;	# skip hidden files

			next unless Acme::Security::allow('Access',$_);

			$encoded_file = escape_html("$_");

			$html .= "<li>"
			  . "<a href=\"javascript:save_as('$encoded_file')\""
			  . ">$_</a>";
		}

		#
		# Print listing footer
		#

		$html .= <<End_Of_HTML;
</ul>

</div>
End_Of_HTML
	}

	#
	# Display parameter form
	#

	print Acme::HTML::Parameters->get(
				cgi	=> $cgi,
				title   => $title,
				html    => $html,
				script  => $script,
				window_title => 'Save Report',
				submit  => '  Save  ',
				);
}

#-------------------------------------------------------------------------------

sub get_file_names {
	my $directory = shift;

	#
	# Get list of files
	#

	opendir DIR, $directory or do {
		warn "Cannot find $directory. $!";
		return ();
	};

	my @file_names = readdir DIR;

	closedir DIR;
	shift @file_names; shift @file_names;	# remove . and ..

	return @file_names;
}
