#!/usr/bin/perl -wT -I../../lib

=head1 NAME

orders/view.cgi -- Display orders

=head1 SYNOPSIS

    orders/view.cgi?search=Folly

    orders/view.cgi?order_id=12345&print

=head1 DESCRIPTION

This program displays orders. It displays them in two different ways. One
way is as a simple order in a window that can be saved or printed. The 
other way is as a View screen, with company headers and footers and a
button-bar to manipulate the order.

Putting the C<print> parameter in the url causes the program to NOT display
site's common headers and footers -- only the order suitable for printing --
and display's the browser's print dialog box.

Using the C<email> parameter in the url causes the program to display the
order like the C<print> parameter, however, the browser's print dialog box
is not displayed. The contents of the browser window can then  be put into
an email message.

Using the C<save> parameter in the url causes the program to save the order
in a file. The file name used is made out of the order number and the 
order name. It looks like: C<order 12345 Software Designs>
The location where the file is stored is determined by the $Order_Folder
setting in the site configuration file (.../etc/site.conf). If a file
already exists, a dash and a number 1,2,3... is appended until a file
name is found that does not already exists.

The C<request> parameter can be used in lieu of C<print>, C<email>, or
C<save> by using C<request=print>, C<request=email>, or C<request=save>
respectfully. 

The C<request=insert>parameter can be used to have this program render
the orders for another program. It "inserts" it into the output stream.
In other words, it prints them. But it does not print the <html> header,
the style sheet <link> tags or any headers, just the list of orders as
this program would render them. This is still experimental, and does
not work well yet with the navigation buttons.

=back

=head1 CHANGE LOG

 10/7/00  Initial version.
 08/01/02 Convert to use Acme::Lister.
 08/05/02 Added "acknowledge" parameter to produce a printer/email ready order.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000-2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

package order;

require 5.6.0;
use strict;

use Acme::HTTP			qw(escape);
use Acme::HTML			qw(escape_html	display_message);
use Acme::HTML::NavButtons;
use Acme::Base::Schema;
use Acme::Lister;
use Acme::Searcher;
use Acme::SQL;
use Acme::Money			qw(dollars	commafy);
use Acme::Options;
use Acme::Address;
use Local::Acme::Banners	qw(common_header common_footer);
use Local::Acme::Format;
use Acme::Config;
use CGI;

use Acme::Security;
Acme::Security::check('Access');

$|=1;		# flush output (synchronize errors with output)

our $cgi = new CGI;
die "No arguments\n" unless $cgi->param();

our $Base         = Acme::Config->Base;		# site base url
our $Root         = Acme::Config->Root;		# corresponds to $Base
our $Images       = Acme::Config->Images;	# location of image files
our $Styles       = Acme::Config->Styles;	# location of CSS styles
our $Scripts      = Acme::Config->Scripts;	# location of scripts
our $Scripts_Lib  = Acme::Config->Scripts_Lib;	# location of library scripts
our $Audit_Prog   = Acme::Config->Audit_Prog;	# display audit info

our $Order_CSS    = Acme::Config->Order_CSS;	# location of stylesheet
our $Order_Folder = Acme::Config->Order_Folder;	# folder where orders are saved

#
# Program Arguments
#

our $request = $cgi->param('request')	|| '';	# 'print', 'email', 'insert' or 'save'

our $search  = $cgi->url_param('search')|| $cgi->param('search');
our $title   = $cgi->url_param('title')	|| $cgi->param('title');

our $table   = 'orders';
our $sql     = '';

our $Printing = $cgi->url_param('print')|| $cgi->param('print')	|| '';
$Printing = $Printing =~ /yes/i		||
	    $request =~ /print/i	||
	    exists $cgi->{'print'} ? 1 : 0;

our $Emailing = $cgi->url_param('email')|| $cgi->param('email')	|| '';
$Emailing = $Emailing =~ /yes/i		||
	    $request =~ /email/i	||
	    exists $cgi->{'email'} ? 1 : 0;
$Printing = 1 if $Emailing;	# Emailing is special case of Printing.

our $Saving = $cgi->url_param('save')	|| $cgi->param('save')	|| '';
$Saving   = $Saving =~ /yes/i		||
	    $request =~ /save/i		||
	    exists $cgi->{'save'} ? 1 : 0;
$Printing = 1 if $Saving;	# Saving is special case of Printing.

our $Inserting = $cgi->url_param('insert')|| $cgi->param('insert')	|| '';
$Inserting = $Inserting =~ /yes/i	||
	    $request =~ /insert/i	||
	    exists $cgi->{'insert'} ? 1 : 0;

our $Schema;		# schema pointer
our $t;			# schema table pointer

#
# Get user options
#

our $opt = new Acme::Options;

#
# Search hints
#

our $record_set_name;

our ($id_field,$code_field,$name_field,$phone_field,$email_field,
     $first_name_field,$last_name_field,$code_field_match,$phone_field_match);

#
# Set defaults
#

if ($table) {

	Acme::Security::check('Access',$table);

	$Schema = new Acme::Base::Schema;

	$t = $Schema->t($table)
	  or die "Cannot find a database table named '$table'\n";

	$title ||= $t->title;

	$record_set_name = $t->title unless $record_set_name;

	$id_field          ||= "$table." . $t->id_field;
	$code_field        ||= "$table." . $t->code_field;
	$name_field        ||= "$table." . $t->name_field;
	$phone_field       ||= "$table." . $t->phone_field;
	$email_field       ||= "$table." . $t->email_field;
	$first_name_field  ||= "$table." . $t->first_name_field;
	$last_name_field   ||= "$table." . $t->last_name_field;
	$code_field_match  ||= "$table." . $t->code_field_match;
	$phone_field_match ||= "$table." . $t->phone_field_match;

	if (! $sql) {
		# Create default sql to search the table.
		# Assure key field is part of sql.
		$sql = <<End_Of_SQL;
select	orders.*,
	customers.address1,
	customers.address2,
	customers.city,
	customers.state,
	customers.zip,
	customers.province,
	customers.postal_code,
	customers.country,
	customers.address_locale,
	countries.address_style,
	ship_countries.address_style,
	ship_via_codes.description,
	ship_via_codes.track_url,
	order_status_codes.description,
	cancellation_codes.reason,
	ad_sources.ad_source_name
from (((((((orders join customers on orders.customer_id = customers.customer_id)
      left join countries on customers.country = countries.country)
      left join countries as ship_countries on orders.ship_country = ship_countries.country)
      left join ship_via_codes on orders.ship_via = ship_via_codes.ship_via)
      left join order_status_codes on orders.order_status = order_status_codes.order_status)
      left join cancellation_codes on orders.cancellation_code = cancellation_codes.cancellation_code)
      left join ad_sources on orders.ad_source = ad_sources.ad_source)
End_Of_SQL
	}

	if (! $search) {
		# Create default search parameter from key argument.
		my $key = $t->key;
		$search = $cgi->url_param($key)	|| $cgi->param($key);
	}
}

die "Empty search\n" unless $sql;

#
# Parse SQL script
#

my $sqlo = new Acme::SQL($sql);

#
# Evaluate Search script
#

if ($search) {
	if (my $sh = $cgi->param('save_history')) {
		# Save search in history
		if ($sh =~ /yes/i and $table) {
			$opt->enter("${table}Search",$search);
		}
		$cgi->delete('save_history');
	}

	if ($id_field || $code_field || $name_field || $last_name_field || $phone_field || $email_field) {

		#
		# Preprocess search script
		#

		if ($first_name_field || $last_name_field) {
			$search = Acme::Searcher::preprocess($search, {
			  RecordSetName   => $record_set_name,
			  IDField	  => $id_field,
			  CodeField	  => $code_field,
			  NameField       => $name_field,
			  FirstNameField  => $first_name_field,
			  LastNameField   => $last_name_field,
			  PhoneField      => $phone_field,
			  EmailField      => $email_field,
			  CodeFieldMatch  => $code_field_match,
			  PhoneFieldMatch => $phone_field_match,
			});
		}
		else {
			$search = Acme::Searcher::preprocess($search, {
			  RecordSetName   => $record_set_name,
			  IDField	  => $id_field,
			  CodeField	  => $code_field,
			  NameField	  => $name_field,
			  PhoneField      => $phone_field,
			  EmailField      => $email_field,
			  CodeFieldMatch  => $code_field_match,
			  PhoneFieldMatch => $phone_field_match,
			});
		}
	}

	my $srch = new Acme::Searcher $search, $cgi;

	#
	# Combine search script with SQL.
	#

	$sql = $srch->as_sql($sqlo);	# combine sql with search script
}

#
# Check query with security
#

$sql = Acme::Security::clean_sql($sql);

#
# Set final default values
#

our $b = new Acme::HTML::NavButtons print => $Printing;

#
# Execute query
#

our $l = new Acme::Lister (
	title        => $title,
	sql          => $sql,
	before_row   => \&display_order_header,
	after_row    => \&display_order_footer,
	every_row    => \&display_order,
	sandbox      => 'order',
	);

print $cgi->header(-type=>'text/html') unless @ARGV;

if ($l->count == 0) {
	display_message("No records were found.", "$title",undef,0);
	exit 1;
}

#
# Display order
#

display_page_header() unless $Inserting;

$l->print();

display_page_footer() unless $Inserting;

exit 0;

#------------------------------------------------------------------------------
#
# HTML page header and footer.
#
#------------------------------------------------------------------------------

our $save_file;			# name of file to save order
our $save_fh;			# saved file handle when saving

sub display_page_header {

	no strict 'vars';

	my $page_title = $title;

	my $stylesheet = '';
	my $javascript = '';

	my $print_cmd = '';
	if ($Printing) {

		# Change the title to include the order number and 
		# customer name. This is done so that is shows in the
		# printed page header or footer (depending upon the
		# browser's print settings). It is also done so that
		# if the user saves the page (browser: File->Save As...)
		# the title is used as the default file name.

		$page_title = "order $orders__order_id $orders__ship_name";

		if ($Saving) {

			# Note: two ways to save. If $Order_Folder,
			# then send output directly to a file.
			# otherwise, render the order as printer
			# friendly in a new window and let user do
			# a "Save As..." on that window.

			if ($Order_Folder) {
				
				# Open file for saved order -- always overwrite

				$save_file = "$Order_Folder$page_title.htm";

				# Iterate until a file name is found which does not exist.
				my $n = 0; my $suffix = '';
				while (
				  ($save_file = "$Order_Folder$page_title$suffix.htm")
				  && -e $save_file
				) {
					$suffix = '-' . ++$n;
				}

				open(SAVE,'>',$save_file)
				  or die qq[Cannot open "$save_file". $!];

				$save_fh = select(SAVE);
			}
		}
		elsif (! $Emailing && ! $Saving) {
			$print_cmd = 
  qq[ onLoad="window.print();window.setTimeout('window.close()',1500)"];
		}
  	}
	else {
		#
		# Include styles and javascript for interactive use.
		#

		$page_title ||= "order $orders__ship_name";

		$stylesheet .= <<End_Of_HTML;
<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<link rel="stylesheet" href="$Styles/button_bar.css" type="text/css">
<link rel="stylesheet" href="$Styles/view.css" type="text/css">

<script src="$Scripts_Lib/menu.js" type="text/javascript"></script>
End_Of_HTML
		
		$javascript .= qq[
<script src="$Scripts/order.js" type="text/javascript"></script>];

		if ($opt->AutoRefresh =~ /yes/) {
			$javascript .= qq[
<script language="JavaScript">function auto_refresh() { self.location.reload(true) }</script>];
		}
	}

	# Make stylesheet available to public for reading in email.
	#$stylesheet .= qq[<link rel="stylesheet" href="$Base$Order_CSS" type="text/css">];

	# Include stylesheet into HTML source
	if (open(STYLEFILE, '<', "$Root$Order_CSS")) {
		local $/;		# read entire file

		$stylesheet .= qq[<style type="text/css">\n<!--\n];
		$stylesheet .= <STYLEFILE>;
		$stylesheet .= qq[-->\n</style>\n];

		close(STYLEFILE);
	}

	print <<EOF;
<html>
<head>
<title>$page_title</title>
$stylesheet$javascript
<script src="/scripts/debug_css.js" type="text/javascript"></script>
</head>
<body$print_cmd>
EOF

	print $b->buttons;

	# Do not need to put a title in the header as it is in each order.
	#print common_header($title, $Printing) if ! $Printing;
	print common_header('', $Printing) if ! $Printing;

}

#------------------------------------------------------------------------------

sub display_page_footer {

	# Do not need to put a title in the footer as it is in each order.
	#print common_header($title, $Printing) if ! $Printing;
	print common_footer('',$Printing) if ! $Printing;

	print $b->last_buttons;

	print <<EOF;

</body>
</html>
EOF

	if ($Saving && $Order_Folder) {
		close(SAVE);
		select($save_fh);	# restore STDOUT

		no strict;

		display_message(
  qq[Order <big>$orders__order_id</big> has been saved to "$save_file".]);
	}
}

#------------------------------------------------------------------------------
#
# Order display.
#
#------------------------------------------------------------------------------

sub display_order_header {

	return if $Printing;

	print $b->buttons() if $l->nrec > 1;

	print button_bar();

	#
	# Print status of order (acknowledged, picked, shipped, or invoiced)
	#

	no warnings qw(uninitialized once);
	no strict 'vars';

	my $status = '';
	$status .= qq[<nobr>Status: $order_status_codes__description ($orders__order_status)</nobr>\n]
	  if $orders__order_status;

	$status .= qq[&nbsp;\n]
	  if $status && $cancellation_codes__reason;

	$status .= qq[<nobr><span class="cancel-code">Canceled: $cancellation_codes__reason ($orders__cancellation_code)</span></nobr>\n]
	  if $orders__cancellation_code;

	$status .= qq[<br>\n]
	  if $status
	  && ($orders__invoiced_date
	   || $orders__shipped_date
	   || $orders__picked_date
	   || $orders__acknowledged_date);

	$status .= qq[<nobr>Invoiced $orders__invoiced_date by $orders__invoiced_by.</nobr>\n]
	  if $orders__invoiced_date;
	$status .= qq[<nobr>Shipped $orders__shipped_date by $orders__shipped_by.</nobr>\n]
	  if $orders__shipped_date;
	$status .= qq[<nobr>Picked $orders__picked_date by $orders__picked_by.</nobr>\n]
	  if $orders__picked_date;
	$status .= qq[<nobr>Acknowledged $orders__acknowledged_date by $orders__acknowledged_by.</nobr>\n]
	  if $orders__acknowledged_date;

	$status .= qq[<br>\n] if $status && $orders__ad_source;

	$status .= qq[<nobr>Ad Source: $ad_sources__ad_source_name ($orders__ad_source)</span></nobr>\n]
	  if $orders__ad_source;

	#
	# Surround order with its own display container.
	#

	print <<EOF if $status;

<!-- Order Status and Information -->

<div class="view-status"><table width="96%"><tr>
<td align="left">$status</td>
</tr></table></div>
EOF

	print <<EOF;

<!-- Order -->

<div class="order" id="order_$orders__order_id">
EOF
}

#------------------------------------------------------------------------------

sub display_order_footer {

	return if $Printing;

	print <<EOF;
</div>
EOF

	#
	# Construct audit link
	#

	no warnings qw(uninitialized once);
	no strict 'vars';

	my $audit = '';
	if (Acme::Security::allow('Access',$Audit_Prog)) {
    		$audit = <<EOF;
<font class=field-name>Entered</font> $orders__entry_date
<font class=field-name>Changed</font> $orders__last_update
<font class=field-name>by</font> $orders__last_update_by
<a href="$Audit_Prog?table=orders&order_id=$orders__order_id" onClick="new_target(this,'audit')" title="Display audit list of changes to this record"><img src="$Images/audit.gif" border="0" onMouseOver="this.src='$Images/audit_hilite.gif'" onMouseOut="this.src='$Images/audit.gif'"></a>
EOF
		print <<EOF;

<!-- Order Audit Information -->

<div class="view-status"><table width="96%"><tr>
<td align="right" valign="bottom">$audit</td>
</tr></table></div>
EOF
	}
}

#------------------------------------------------------------------------------

sub display_order {

	#
	# Customer Information
	#

	no strict 'vars';

	display_order_banner();

	display_customer_info();

	display_order_summary();

	display_order_items($orders__order_id);

	display_order_totals_and_comments()
}


#------------------------------------------------------------------------------
#
# Display the sections of the order.
#
#------------------------------------------------------------------------------

no warnings qw(uninitialized once);
no strict qw(vars);

sub display_order_banner {

	# Display the very top of the order page.

	print <<EOF;
<div class="order-banner">
<table width="99%">
<tr>
<td width="33%" align="left"><img src="$Base$Images/local/order_logo.gif"></td>
<td width="33%" align="center"><span id="order-title">Order</span></td>
<td width="33%" align="right"><span class="label">Date</span> <span id="order-date">$orders__order_date</span></td>
</tr>
</table>
</div>

EOF
}

#------------------------------------------------------------------------------

sub display_customer_info {

	#
	# Print customer's shipping and billing addresses.
	#

	no warnings qw(uninitialized once);

        my $address_class = 'address';

	my $addr = new Acme::Address 
				$orders__customer_name,
				'',
				$customers__address1,
				$customers__address2,
				$customers__city,
				$customers__state,
				$customers__zip,
				$customers__province,
				$customers__postal_code,
				$customers__country,
				$customers__address_locale,
				$countries__address_style;

	my $ship_addr = new Acme::Address 
				$orders__ship_name,
				$orders__ship_company,
				$orders__ship_address1,
				$orders__ship_address2,
				$orders__ship_city,
				$orders__ship_state,
				$orders__ship_zip,
				$orders__ship_province,
				$orders__ship_postal_code,
				$orders__ship_country,
				$orders__ship_address_locale,
				$ship_countries__address_style;

	#
	# Constuct links to address maps
	#

	my $map_link = $ship_map_link = '';
	if (! $Printing) {
		$map_link = $addr->map_link
		  if ! $addr->is_empty;
		$ship_map_link = $ship_addr->map_link
		  if ! $ship_addr->is_empty;
	}

	#
	# Print address headers (include map link if not printing)
	#

	local $\="\n";	# append line feed to all print statements

	print qq[<div class="addresses">];
	print qq[<table width=100% cellspacing=0 cellpadding=0>];

	print qq[<tr>];
	##print qq[<td width="5%"></td>];
	if (! $ship_addr->is_empty) {
		print
		    qq[<td width="40%"]
		  . qq[ align="left"]
		  . qq[ class="address-top">]
		  . qq[<span class="address-title">Ship To</span>]
		  . qq[&nbsp;&nbsp;$ship_map_link]
		  . qq[</td>];
	}
	else {
		print qq[<td width="40%">&nbsp;</td>];
	}
	print qq[<td>&nbsp;</td>];
	if (! $addr->is_empty) {
		print
		    qq[<td width="40%"]
		  . qq[ align="left"]
		  . qq[ class="address-top">]
		  . qq[<span class="address-title">Bill To</span>]
		  . qq[&nbsp;&nbsp;$map_link]
		  . qq[</td>];
	}
	else {
		print qq[<td width="40%">&nbsp;</td>];
	}
	##print qq[<td width="5%"></td>];
	print qq[</tr>];

	#
	# Print Addresses
	#

	print qq[<tr>];
	##print qq[<td></td>];
        print qq[<td valign="top">];
	if (! $ship_addr->is_empty) {
		print $ship_addr->as_html;
	}
        print qq[</td>];
        print qq[<td></td>];
        print qq[<td valign="top">];
	if (! $addr->is_empty) {
		print $addr->as_html;
	}
        print qq[</div></td>];
	##print qq[<td></td>];
	print qq[</tr>];

	#
	# Print address footers (lines under addresses)
	#

	print qq[<tr>];
	##print qq[<td></td>];
	if (! $ship_addr->is_empty) {
		print
		    qq[<td width="40%"]
		  . qq[ align="left"]
		  . qq[ class="address-bottom">]
		  . qq[&nbsp;]
		  . qq[</td>];
	}
	else {
		print qq[<td></td>];
	}
	print qq[<td></td>];
	if (! $addr->is_empty) {
		print
		    qq[<td width="40%"]
		  . qq[ align="left"]
		  . qq[ class="address-bottom">]
		  . qq[&nbsp;]
		  . qq[</td>];
	}
	else {
		print qq[<td></td>];
	}
	##print qq[<td></td>];
	print qq[</tr>];
	print qq[</table>];
	print qq[</div>\n];
}

#------------------------------------------------------------------------------

sub display_order_summary {

	# Display summary of terms, order numbers, ...

	print <<EOF;
<div class="order-summary">
<table width="100%" cellspacing="0" cellpadding="0" border="0">
<tr>
<td width="33%"><span class="label">Ship Date</span> $orders__ship_date</td>
<td width="33%"><span class="label">Cancel Date</span> $orders__cancel_date</td>
<td width="33%"><span class="label">Your PO#</span> $orders__pruchase_order</td>
</tr>
<tr>
<td width="33%"><span class="label">Terms</span> $orders__terms</td>
<td width="33%"><span class="label">Buyer</span> $orders__salesrep</td>
<td width="33%"><span class="label">Our PO#</span> $orders__order_id</td>
</tr>
</table>
</div>

EOF
}

#------------------------------------------------------------------------------

sub display_order_totals_and_comments {

	# Display taxes, totals, and comments.

	my $subtotal = dollars($orders__order_items_gross,2);
	my $total    = dollars($orders__total_gross,2);

	my $tracking = '';
	if ($orders__ship_tracking_no) {
		my $link = $orders__ship_tracking_no;
		if (my $url = $ship_via_codes__track_url) {
			$url =~ s/([^\\])\$/$1$orders__ship_tracking_no/g
			  or 
			$url .= $orders__ship_tracking_no;

			$link = qq[<a]
			  . qq[ href="$url" title="Show tracking information."]
			  . qq[ target="_blank"]
			  . qq[>$orders__ship_tracking_no]
			  . qq[</a>];
		}

		$tracking = qq[<nobr><span class="label">Tracking #</span>]
	                  . qq[$link</nobr>]
	}


	my $payment  = 'Pay ' . $orders__payment_type
	  if $orders__payment_type;
	my $pay_suffix = substr($orders__payment_number,-4);
	$payment &&= ' ...' . $pay_suffix
	  if $pay_suffix;

	print <<EOF;
<div class="order-footer">
<table width="100%" cellspacing="0" cellpadding="0" border="0">
<tr>
 <td width="75%" align="left">
 <div class="order-comments">
 <span class="label">Ship via</span> $ship_via_codes__description $tracking
EOF

	print <<EOF if $orders__shipping_instructions;
 <br><span class="label">Shipping Instructions</span>
 <div class="shipping-instructions">
 $orders__shipping_instructions
 </div>
EOF
	print <<EOF;
 </div>
 </td>

 <td width="25%" align="right">
  <div class="order-totals">
  <table width="100%" cellspacing="0" cellpadding="0" border="0">
  <tr>
   <th width="50%">Subtotal</th>
   <td width="50%" align="right"><span class="subtotal">$subtotal</span></td>
  </tr>
EOF

	print <<EOF if $orders__misc_charges_description || ($orders__misc_charges && $orders__misc_charges != 0);
  <tr>
   <th width="50%">$orders__misc_charges_description</th>
   <td width="50%" align="right">$orders__misc_charges</td>
  </tr>
EOF

  	print <<EOF;	# always print shipping charges
  <tr>
   <th width="50%">Shipping</th>
   <td width="50%" align="right">$orders__shipping</td>
  </tr>
EOF

	print <<EOF if $orders__fed_tax;
  <tr>
   <th width="50%">Fed Tax</th>
   <td width="50%" align="right">$orders__fed_tax</td>
  </tr>
EOF

	print <<EOF if $orders__state_tax;
  <tr>
   <th width="50%">State Tax</th>
   <td width="50%" align="right">$orders__state_tax</td>
  </tr>
EOF

	print <<EOF if $orders__city_tax;
  <tr>
   <th width="50%">City Tax</th>
   <td width="50%" align="right">$orders__city_tax</td>
  </tr>
EOF

	print <<EOF if $orders__other_tax;
  <tr>
   <th width="50%">$orders__other_tax_description</th>
   <td width="50%" align="right">$orders__other_tax</td>
  </tr>
EOF

	print <<EOF;
  <tr>
   <th>Total</th>
   <td align="right"><span class="total">$total</span></td>
  </tr>

  <tr>
   <th width="50%">$payment_desc</th>
   <td width="50%" align="right"></td>
  </tr>
  </table>
  </div>

 </td>
</tr>
</table>
</div>

EOF
	print <<EOF if $orders__comments;
<div class="comments"><table><tr><td>
$orders__comments
</td></tr></table></div>
EOF
}

#------------------------------------------------------------------------------
#
# Order items display.
#
#------------------------------------------------------------------------------

use warnings qw(uninitialized once);

sub display_order_items {
	my ($keyval) = @_;

	#
	# Set up parameters from globals for listing order items.
	#

	my $detail_table = 'order_items';
	my $detail_field = 'order_id';

	my $table = 'orders';
	my $field = 'order_id';

	my $order_list = <<End_Of_List;

	order_items.product_id,
	order_items.item,
	order_items.color,
	products.description,
	color_codes.description,
	order_items.sizes,
	order_items.sizes_qty,
	order_items.price,
	order_items.unit,
	order_items.qty,
	order_items.amount
End_Of_List

	my $order_header = <<End_Of_Format;
<th align="left" width="125">Item</th>
<th align="left" width="40%">Description</th>
<th align="right">Price</th>
<th>&nbsp;</th>
<th align="right">Qty</th>
<th align="right">Amount</th>
End_Of_Format

	my $order_format = <<End_Of_Format;
<td valign="top" width="125">%s %s %s</td>
<td valign="top" width="30%"><nobr><big>%s</big></nobr> <nobr>%s</nobr><br><nobr>&mingle([": ",", "],"%s","%s")</nobr></td>
<td valign="top" align="right">%s</td>
<td valign="top">%s</td>
<td valign="top" align="right">%s</td>
<td valign="top" align="right">%s</td>
End_Of_Format

	my $order_sql = <<End_Of_SQL;
select $order_list
from ((order_items left join products on order_items.product_id = products.product_id)
      left join color_codes on order_items.color = color_codes.color)
order by order_items.order_item_id
End_Of_SQL

	#
	# Construct sql for list of detail records.
	#

	my $t  = $Schema->t($table);
	my $f  = $t->f($field);
	my $dt = $Schema->t($detail_table);

	package order_item;

	my $detail      = ($f && $f->detail)		 || $t->detail;

	my $detail_key  = $dt->key;
	my $edit_prog   = $dt->edit_prog;

	my $list_fields = $order_list	||
			  ($f && $f->detail_list_fields) || $t->detail_list_fields;

	my $list_format = $order_format ||
			  ($f && $f->detail_format)	 || $t->detail_format;
	my $list_header = $order_header ||
			  ($f && $f->detail_header)	 || $t->detail_header;

	my $detail_sql  = $order_sql	||
			  ($f && $f->detail_sql)	 || $t->detail_sql;

	#
	# Determine if user has privilages to edit record.
	#

	my $has_edit_priv = Acme::Security::allow('Update',
						  $detail_table,
						  $dt->secure,
						  $dt->restrict);
	#
	# Process SQL query for detail records.
	#

	my $sqlo = new Acme::SQL $detail_sql;

	# Assure detail table key field is in query
	$sqlo->add_column("$detail_table.$detail_key")
	  if ! $sqlo->is_column("$detail_table.$detail_key");

	(my $detail_join_field = $detail) =~ s/.*\.//;

	$sqlo->add_where("$detail_table.$detail_join_field = $keyval");

	$detail_sql = Acme::Security::clean_sqlo($sqlo);

	#
	# Define globals for the list helper functions.
	#

	my $bullet = Acme::Config->Bullet;

	my $row_name = $dt->row_name;

	my $nrec = 0;				# record number

	#
	# Define helper functions to list detail items.
	#

	#----------------------------------------
	my $before_list = sub {

		#
		# Compose list header
		#

		my $header = '';
		if ($list_header) {
			for my $h (grep { s/$detail_table\.// } split(/\s*,\s*/,$list_fields)) {
				if ($dt->{$h}) {
					push @headers, $dt->{$h}->title;
				}
				else {
					# Capitalize field name
					# (happens on ID fields)
					push @headers, main::capitalize($h);
				}
			}

			$header = Local::Acme::Format::sprintf_cmd($list_header,@headers);

			if ($header !~ m|^\s*<th|i
			||  $header =~ s|^\s*<td>(.*)</td>\s*$|$1|i) {
				$header = qq[<th>$list_header</th>];
			}

			#
			# Prepend extra columns for record index and edit bullet.
			#

			# Edit bullet not allowed unless order_items post to
			# orders every time they post, which they don't now.
			#$header = qq[<th>&nbsp;</th>$header]
			#  if $has_edit_priv;

			$header = qq[<th>&nbsp;</th>\n$header];
		}

		return <<End_Of_HTML;
<div class="detail-list">
<table class="detail-list" width="100%">
$header
End_Of_HTML
	};

	#----------------------------------------
	my $after_list = sub {
		return <<End_Of_HTML;
</table>
</div>

End_Of_HTML
	};

	#----------------------------------------
	my $before_row = sub {

		$nrec++;

		my $html = '';	# returned value

		$html .= <<End_Of_HTML;
<tr>
<td class="detail-index">$nrec</td>
End_Of_HTML

		#
		# Create edit bullets for editing order items.
		#

		# Note: Do not do this at this time, if ever...
		return $html;	# Do not put edit bullets on items.

		no strict 'refs';
		my $keyval = escape(${${detail_table}.'__'.${detail_key}});


		my $sep    = $edit_prog =~ /\?/ ? '&' : '?';
		my $url    = "$edit_prog${sep}table=${detail_table}&$detail_key=${keyval}&request=edit";

		# Edit bullet not allowed unless order_items post to
		# orders every time they post, which they don't now.
		$html .= <<End_Of_HTML if $has_edit_priv;
<td class="detail-link"><a href="$url" onClick="new_target(this,'form')" title="Edit $row_name." class="edit-bullet">$bullet</a></td>
End_Of_HTML
		return $html;
	};

	#----------------------------------------
	my $after_row = sub {
		return <<End_Of_HTML;
</tr>
End_Of_HTML
	};

	#----------------------------------------
	my $every_row = sub {
		no warnings qw(uninitialized once);	# for @a in Lister

		my $row = Local::Acme::Format::sprintf_cmd($list_format,@a);

		$row = qq[<td>$row</td>\n]
		  if $row !~ m|^\s*<[tT][dD]|;
		return $row
	};

	#
	# Construct List
	#

	my $l = new Acme::Lister
			sql => $detail_sql,
			sandbox    => 'order_item';

	print
	    &$before_list()
	  . $l->list(	before_row => $before_row,
			after_row  => $after_row,
			every_row  => $every_row,
			)
	 . &$after_list();
}

#------------------------------------------------------------------------------
#
# Button bar functions
#
#------------------------------------------------------------------------------

sub button_bar {

	return '' if $Printing;

	#
	# Construct links for the button bar
	#

	no strict 'refs';

	my $key = $t->key;
	my $keyval = escape(${${table}.'__'.${key}});
	my $sep;

	my $edit_prog = $t->edit_prog;
	$sep = $edit_prog =~ /\?/ ? '&' : '?';
	$edit_prog .= "${sep}table=$table&$key=$keyval&request=edit";

	my $acknowledge_prog = "/cgi-bin/orders/acknowledge.cgi";
	$sep = $acknowledge_prog =~ /\?/ ? '&' : '?';
	$acknowledge_prog .= "${sep}table=$table&$key=$keyval";

	my $pick_prog = "/cgi-bin/orders/pick.cgi";
	$sep = $pick_prog =~ /\?/ ? '&' : '?';
	$pick_prog .= "${sep}table=$table&$key=$keyval";

	my $ship_prog = "/cgi-bin/orders/ship.cgi";
	$sep = $ship_prog =~ /\?/ ? '&' : '?';
	$ship_prog .= "${sep}table=$table&$key=$keyval";

	my $invoice_prog = "/cgi-bin/orders/invoice.cgi";
	$sep = $invoice_prog =~ /\?/ ? '&' : '?';
	$invoice_prog .= "${sep}table=$table&$key=$keyval";

	my $save_prog = $cgi->self_url();
	# Substitute keyval for search parameter
	unless ($save_prog =~ s/\bsearch=[^?&;]*/$key=$keyval/i) {
		# Add key value to url
		$sep = $save_prog =~ /\?/ ? '&' : '?';
		$save_prog .= "${sep}$key=$keyval";
	}
	unless ($save_prog =~ s/\b(request=)[^?&;]*/$1save/i) {
		$sep = $save_prog =~ /\?/ ? '&' : '?';
		$save_prog .= "${sep}save=yes";
	}

	my $print_prog = $cgi->self_url();
	# Substitute keyval for search parameter
	unless ($print_prog =~ s/\bsearch=[^?&;]*/$key=$keyval/i) {
		# Add key value to url
		$sep = $print_prog =~ /\?/ ? '&' : '?';
		$print_prog .= "${sep}$key=$keyval";
	}
	unless ($print_prog =~ s/\b(request=)[^?&;]*/$1print/i) {
		$sep = $print_prog =~ /\?/ ? '&' : '?';
		$print_prog .= "${sep}print=yes";
	}

	my $email_prog = $cgi->self_url();
	# Substitute keyval for search parameter
	unless ($email_prog =~ s/\bsearch=[^?&;]*/$key=$keyval/i) {
		# Add key value to url
		$sep = $email_prog =~ /\?/ ? '&' : '?';
		$email_prog .= "${sep}$key=$keyval";
	}
	unless ($email_prog =~ s/\b(request=)[^?&;]*/$1email/i) {
		$sep = $email_prog =~ /\?/ ? '&' : '?';
		$email_prog .= "${sep}email=yes";
	}

	#
	# Construct button bar
	#


	my $html = '';

	$html .= <<End_Of_HTML;

<!-- Start Button Bar -->
<div class="button-bar">
<table cellspacing="0" cellpadding="0">
<tr>

 <td class="button-bar-menu">
 <a id="bb_menu_head"
  class="bb-menu-header"
  href="javascript:menu('bb_menu','bb_menu_head')"
 >&bull;</a>
 </td>

End_Of_HTML

	if (Acme::Security::allow('Update','orders',main::program_of($edit_prog))) {
		$html .= <<End_Of_HTML;
 <td><a href="$edit_prog" onClick="new_target(this,'form')" title="Edit the order information.">Change</a></td>
End_Of_HTML
	}

	$html .= <<End_Of_HTML;
 <td><a href="javascript:do_save('$save_prog')" title="Save a printer friendly version of the order to a file.">Save</a></td>
 <td><a href="$print_prog" onClick="new_target(this,'print')" title="Render a printer friendly version of the order so it can be printed.">Print</a></td>
 <td><a href="javascript:do_email('$email_prog')" title="Render a printer friendly version of the order so it can be emailed.">Email</a></td>
End_Of_HTML

	if (Acme::Security::allow('Access',main::program_of($acknowledge_prog))) {
		$html .= <<End_Of_HTML;
 <td><a href="javascript:do_acknowledge('$acknowledge_prog')" title="Send acknowledgement that we received this order.">Acknowledge</a></td>
End_Of_HTML
	}

	if (Acme::Security::allow('Access',main::program_of($pick_prog))) {
		$html .= <<End_Of_HTML;
 <td><a href="javascript:do_pick('$pick_prog')" title="Mark which items on this order that can be shipped.">Pick</a></td>
End_Of_HTML
	}

	if (Acme::Security::allow('Access',main::program_of($ship_prog))) {
		$html .= <<End_Of_HTML;
 <td><a href="javascript:do_ship('$ship_prog')" title="Make a shipping manifest from this order.">Ship</a></td>
End_Of_HTML
	}

	if (Acme::Security::allow('Access',main::program_of($invoice_prog))) {
		$html .= <<End_Of_HTML;
 <td><a href="javascript:do_invoice('$invoice_prog')" title="Make an invoice out of the shipment from this order.">Invoice</a></td>
End_Of_HTML
	}

	$html .= <<End_Of_HTML;
</tr>
</table>
</div>
End_Of_HTML

	#
	# Construct button bar menu
	#

	$html .= <<End_Of_HTML;

<!-- Button Bar Menu -->
<div id="bb_menu" class="bb-menu">
<a href="javascript:select_all('order_$orders__order_id')">Select All</a>
<hr>
<a href="$print_prog" onClick="new_target(this,'print');hide_menus()">Print</a></td>
<a href="javascript:self.close()">Exit</a>
</div>
<!-- End Button Bar Menu -->

<script language="JavaScript">make_menu('bb_menu','bb_menu_head');</script>
<!-- End Button Bar -->

End_Of_HTML
	return $html;
}

#------------------------------------------------------------------------------
#
# Support and utility routines
#
#------------------------------------------------------------------------------

package main;

sub program_of {
	
	# Extract program name from url

	my $url = shift;
	$url =~ s|.*://||;	# remove protocol
	$url =~ s|.*?/|/|;	# remove host name (till /)
	$url =~ s|\?.*||;	# remove parameters after ?
	return $url;
}

sub capitalize {
	my $s = shift;
	$s =~ s/_/ /g;		# change _ to spaces
	$s =~ s/\bid\b/ID/;	# uppercase ID
	$s =~ s/\b(\w+)\b/ucfirst(lc($1))/eg; # capitalize
	return $s;
}

