#!/usr/bin/perl -wT -I../lib
require 5.6.0;
use strict;

=head1 NAME

pick.cgi -- Display database information in a selection list.

=head1 SYNOPSIS

    pick.cgi?table=customers

    pick.cgi?table=customers&search=Folly

    pick.cgi?sql=select+f1 ...

    pick.cgi?sql=select+f1...&search=Foo

=head1 DESCRIPTION

The pick.cgi program displays rows of data from a database table or from
the sql script passed as an argument. The output from the script is filtered
by the search parameter if specified.

This program is a simplified version of the C<list.cgi> program. The list
is to be used to select one row for other operations (request=C<Edit>, C<View>,
or C<New>).

A format string may be specified by using the 'format' parameter. It is a
string used by I<Local::Acme::Format>, which is essentially a string to be formatted
by sprintf() and then scanned for functions to execute. The functions need
to be defined in I<Local::Acme::Format> in order to be recognized in the formatted
string.

A list is formated as an HTML table, with each list item placed between <td>
and </td> tags. That allows the format string to also have <td> and </td>
tags embedded in it to format fields into table columns.

=back

=head1 CHANGE LOG

 6/12/01 Initial version. Converted from list.cgi to specialize in just a
         selection type list.
 7/29/01 Changed name to pick.cgi and added pick options in Acme::Base::Schema.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

package pick;

use Acme::HTTP			qw(escape);
use Acme::HTML			qw(escape_html display_message);
use Acme::Base::Schema;
use Acme::Base::Schema::Join	qw(parse_join);
use Acme::SQL;
use Acme::Searcher;
use Acme::Options;
use Acme::Lister;
use Acme::HTML::NavButtons;
use Local::Acme::Banners  	qw(common_header common_footer);
use Local::Acme::Format		qw(sprintf_cmd);
use Acme::Config;
use CGI;

use Acme::Security;
Acme::Security::check('Access');

$|=1;		# flush output (synchronize errors with output)

our $cgi = new CGI;
die "No arguments\n" unless $cgi->param();

our $Styles = Acme::Config->Styles;            # location of CSS styles

#
# Program Arguments
#

our $request	= $cgi->url_param('request')	  || $cgi->param('request');

our $table	= $cgi->url_param('table')	  || $cgi->param('table');
our $form	= $cgi->url_param('form')	  || $cgi->param('form');
our $template	= $cgi->url_param('template')	  || $cgi->param('template');

our $sql	= $cgi->url_param('sql')	  || $cgi->param('sql');
our $search	= $cgi->url_param('search')	  || $cgi->param('search');

our $list	= $cgi->url_param('list')	  || $cgi->param('list');
our $format 	= $cgi->url_param('format')	  || $cgi->param('format');

our $title	= $cgi->url_param('title')	  || $cgi->param('title');
our $list_title = $cgi->url_param('list_title') || $cgi->param('list_title');

our $record_set_name  = $cgi->url_param('record_set_name') || $cgi->param('record_set_name');

our $id_field	= $cgi->url_param('id_field')	  || $cgi->param('id_field');
our $code_field	= $cgi->url_param('code_field') || $cgi->param('code_field');
our $name_field	= $cgi->url_param('name_field') || $cgi->param('name_field');
our $first_name_field = $cgi->url_param('first_name_field') || $cgi->param('first_name_field');
our $last_name_field  = $cgi->url_param('last_name_field')  || $cgi->param('last_name_field');
our $phone_field = $cgi->url_param('phone_field')|| $cgi->param('phone_field');
our $email_field = $cgi->url_param('email_field')|| $cgi->param('email_field');

our $code_field_match = $cgi->url_param('code_field_match')  || $cgi->param('code_field_match');
our $phone_field_match= $cgi->url_param('phone_field_match') || $cgi->param('phone_field_match');

our $limit_abort   = $cgi->url_param('limit_abort')  || $cgi->url_param('limit')    || '';
our $limit_warn    = $cgi->url_param('limit_warn')   || '';
our $limit_alert   = $cgi->url_param('limit_alert')  || '';

our $list_id	= $cgi->url_param('list_id')	  || $cgi->param('list_id');

our $Printing     = $cgi->url_param('print')       || $cgi->param('print') || '';
$Printing         = $Printing =~ /yes/i            ||
		    $request =~ /print/i           ||
		    ($Printing !~ /^no$/i && exists $cgi->{'print'} ? 1 : 0);

#
# Get user options
#

our $opt = new Acme::Options;

#
# Set defaults
#

if ($form || $template) {

	$request ||= 'Edit';	# infer default if not specified

	if (! $table) {
		
		#
		# Get default (primary) table from form.
		#

		require Acme::Form;

		my $f = new Acme::Form form => $form, template => $template;
		$table = $f->primary_table;
	}
}

die "Table name not specified.\n" unless $table;

$list_id ||= 'table';
$request ||= "View";

#
# Save Search history
#

if (my $save_history = $cgi->param('save_history')) {
	if ($save_history =~ /yes/i) {
		my $hist = $form || $table;
		$opt->enter("${hist}Search",$search)
		  if $search && $hist;
	}
	$cgi->delete('save_history');
}

#
# Check request with security
#

Acme::Security::check('Access',$table);

#
# Get list parameters from schema definitions
#

our $schema = new Acme::Base::Schema;

our $t = $schema->t($table)
  or die "Cannot find a database table named '$table'\n";

$title   ||= $t->title;
$list    ||= $t->pick_fields;

$format  ||= $t->pick_format;
$format  =~ s/^\s*(["'])(.*)\1\s*$/$2/;	# remove possible quotes

$list_title = $t->pick_title unless ($list_title || $search);

$record_set_name   ||= $t->title;

$id_field	   ||= fully_qualify($table, $t->id_field);
$code_field	   ||= fully_qualify($table, $t->code_field);
$name_field	   ||= fully_qualify($table, $t->name_field);
$first_name_field  ||= fully_qualify($table, $t->first_name_field);
$last_name_field   ||= fully_qualify($table, $t->last_name_field);
$phone_field       ||= fully_qualify($table, $t->phone_field);
$email_field       ||= fully_qualify($table, $t->email_field);

$code_field_match  ||= $t->code_field_match;
$phone_field_match ||= $t->phone_field_match;

#
# Get limits
#

my $limit = $t->limit_abort || Acme::Config->List_Limit_Abort;
$limit_abort = $limit if ! $limit_abort || $limit < $limit_abort;

$limit_warn ||= $t->limit_warn || Acme::Config->List_Limit_Warn;
$limit_alert .= $t->limit_alert;

#
# Create default sql to search the table if needed.
#

$sql ||= $t->pick_sql if $t->defined('pick_sql');

if (! $sql) {

	# Construct query here if not explicity defined in schema.
	# This allows customized list of fields.

	my ($join_expr,$join_clause) = parse_join($table,$t->pick_join);

	$sql = 'select'
	  . ' distinct ' . $list
	  . ' from '	 . $join_expr
	  . ' where '	 . $t->pick_where . $join_clause
	  . ' order by ' . $t->pick_order_by_clause;

}

#
# Parse SQL statement
#

our $sqlo = new Acme::SQL $sql;

#
# Make sure key field is selected in SQL for use in links
#

## Note: This program uses the table's Select fields, which by definition has
## the key field as the first (invisible) field in Choose lists, so this
## should not be necessary.

my $added_column = 0;		# flag
my $col = $table . '.' . $t->key;
if (! $sqlo->is_column($col)) {
	$sqlo->add_column($col);
	$added_column = 1;	# set to true
}

#
# Set SQL limits
#

$sqlo->set_limit($limit_abort) if $limit_abort;

#
# Evaluate Search script
#

if ($search) {

	if ($id_field || $code_field || $name_field || $last_name_field || $phone_field || $email_field) {

		#
		# Preprocess search script
		#

		if ($first_name_field || $last_name_field) {
			$search = Acme::Searcher::preprocess($search, {
			  RecordSetName   => $record_set_name,
			  IDField	  => $id_field,
			  CodeField	  => $code_field,
			  NameField       => $name_field,
			  FirstNameField  => $first_name_field,
			  LastNameField   => $last_name_field,
			  PhoneField      => $phone_field,
			  EmailField      => $email_field,
			  CodeFieldMatch  => $code_field_match,
			  PhoneFieldMatch => $phone_field_match,
			});
		}
		else {
			$search = Acme::Searcher::preprocess($search, {
			  RecordSetName   => $record_set_name,
			  IDField	  => $id_field,
			  CodeField	  => $code_field,
			  NameField	  => $name_field,
			  PhoneField      => $phone_field,
			  EmailField      => $email_field,
			  CodeFieldMatch  => $code_field_match,
			  PhoneFieldMatch => $phone_field_match,
			});
		}
	}

	#
	# Combine sql with search script.
	#

	my $do_http_header = 1;	# set to true so HTTP header output if needed.

	my $srch = new Acme::Searcher $search, $cgi, $do_http_header;

	$sql = $srch->as_sql($sqlo);	

	#
	# Redefine title if defined in script.
	#

	$list_title ||= "$request " . $srch->title if $srch->title;
}

#
# Set up a selection list.
#

$sql = Acme::Security::clean_sql($sql);

our $l = new Acme::Lister sql => $sql, sandbox => 'pick';

if ($l->count == 1 && ($request =~ /Edit/i || $request =~ /View/i)) {


	#
	# Only one record selected, display edit form or view screen immediately
	#

	my ($key,$keyval);

	no strict 'refs';

	$key = $t->key;
	$keyval = $$key || ${"${table}__${key}"};

	if ($keyval) {
		my $url = Acme::Config->Base;
		$url .= $t->edit_prog		 if $request =~ /Edit/i;
		$url .= $t->view_prog		 if $request =~ /View/i;
		$url .= ($url =~ /[?]/) ? '&' : '?';
		$url .= "$key=" . escape($keyval);
		$url .= "&table=$table";
		$url .= "&form=" . escape($form) if $form;
		$url .= "&template=" . escape($template) if $template;
		$url .= "&request=$request";

		print $cgi->redirect($url);
		exit;
	}
}

#
# Display list
#

print $cgi->header(-type=>'text/html') unless @ARGV;

if ($l->count == 0) {
	display_message("No records were found.", "$title",undef,0);
	exit 1;
}

#
# Set final default values
#

if (! $format) {
	$sqlo = new Acme::SQL $sql;		# reparse after clean_sql() expands *
	my @field_names = $sqlo->columns;
	pop @field_names if $added_column;	# remove key field added to sql above
	$format = "%s &nbsp; " x @field_names;
}

#
# Define global variables for list helper functions
#

our ($alink_prog, $alink_help);			# parts of <a> tag
our $n_rec = 0;					# record number

our $lsn = 'list';				# default CSS "list style" name
$lsn = 'graybar' if $opt->get("${list_id}ListGraybar") =~ /yes/i;

our $b = new Acme::HTML::NavButtons;

#
# Define link (<a> tag) for each row
#

if ($request =~ /View/i) {
	$alink_prog   = $t->view_prog;
	$alink_help   = 'View ' . $t->row_name . '.';
}
else {
	$alink_prog = $t->edit_prog;
	$alink_help = 'Edit ' . $t->row_name . '.';
}

#
# Initialize limits
#

my $limits;
if ($limit_abort || $limit_warn) {
	$limits = {};
	$limits->{$limit_abort} = \&limit_abort_func	if $limit_abort;
	$limits->{$limit_warn}  = \&limit_warn_func	if $limit_warn;
}

display_header();

$l->print(
	before_list => \&display_list_header,
	after_list  => \&display_list_footer,
	every_row   => \&display_list_item,
	limits      => $limits,
	);

display_footer();

exit 0;

#------------------------------------------------------------------------------
#
# Select helper routines
#
#------------------------------------------------------------------------------

sub display_header {

	my $aux_script = $t->pick_include_html;
	$aux_script ||= ''; $aux_script &&= "\n$aux_script";

	my $aux_stylesheet = $t->pick_stylesheet_html;
	$aux_stylesheet ||= ''; $aux_stylesheet  &&= "\n$aux_stylesheet";

	my $print_stylesheet = '';
	my $print_cmd = '';

	if ($Printing) {
  		$print_cmd = 
  qq[ onLoad="window.print();window.setTimeout('window.close()',1500)"];

		$print_stylesheet =
  qq[<link rel="stylesheet" href="$Styles/pick_print.css" type="text/css">\n];

		my $aux_print_stylesheet = $t->pick_print_stylesheet_html if $table;
		$aux_print_stylesheet  &&= "$aux_print_stylesheet\n";

		$print_stylesheet .= $aux_print_stylesheet;
  	}

# Keep IE6 in IE5 compatablilty mode by not specifying url
#<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
#"http://www.w3.org/TR/html4/loose.dtd">

	print <<EOF;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">

<html>
<head>
<title>pick $title</title>

<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<link rel="stylesheet" href="$Styles/pick.css" type="text/css">$aux_stylesheet
$print_stylesheet
<script language="JavaScript" type="text/javascript">
function hi_o(row) {
	// Highlight odd numbered rows.

	// Uncomment next to switch class attributes instead of changing styles.
	//if (row) row.className='${lsn}-row-odd-hilite';

	// Use style settings to implement hilites. This is faster in IE5.
	with (row.style) {
		color='FF00FF';
		backgroundColor='DDDDDD';
		cursor='hand';	// ### change to 'pointer' in IE6
	}
}
function hi_e(row) {
	// Highlight even numbered rows.

	// Uncomment next to switch class attributes instead of changing styles.
	//if (row) row.className='${lsn}-row-even-hilite';

	// Use style settings to implement hilites. This is faster in IE5.
	with (row.style) {
		color = 'FF00FF';
		backgroundColor = 'DDDDDD';
		cursor = 'hand';	// ### change to 'pointer' in IE6
	}
}
function lo_o(row) {
	// Un-highlight odd numbered rows.

	// Uncomment next to switch class attributes instead of changing styles.
	//if (row) className='${lsn}-row-odd';

	// Use style settings to implement hilites. This is faster in IE5.
	with (row.style) {
		color = '000000';
		backgroundColor = 'transparent';
		cursor = 'auto';
	}
}
function lo_e(row) {
	// Un-highlight even numbered rows.

	// Uncomment next to switch class attributes instead of changing styles.
	//if (row) className='${lsn}-row-even';

	// Use style settings to implement hilites. This is faster in IE5.
	with (row.style) {
		color = '000000';
		backgroundColor = 'transparent';
		cursor = 'auto'
	}
}
function go(url) {
	if (url) {
		// Display "Please wait...", use a common.js routine.
		if (navigator.appName != "Netscape") display_message();

		// Go to url
		self.location.replace(url);
	}
}
</script>$aux_script
</head>
<body$print_cmd>
EOF
	print $b->first_buttons;
	print common_header($title);

	print <<EOF if $list_title;
<h3>
$list_title
</h3>
EOF

	print <<EOF;
<div class="list">
EOF
}

#------------------------------------------------------------------------------

sub display_footer {

	print "</div>\n";
	print common_footer($title);
	print $b->last_buttons;
	print "</body>\n</html>";
}

#------------------------------------------------------------------------------

sub display_list_header {

	print qq(<table class="${lsn}" style="table-layout:fixed;width:99%">);

	my $header = $t->pick_header if $table;
	$header = $format if $header eq 'yes';

	my $list_header = pick::make_list_header(
				$header,
				$list,
				$table,
				$schema,
				$t);

	if ($list_header) {
		# Surround $list_header with <td> </td> tags if not already.
		if ($list_header !~ m|^\s*<t[hd]|i
		||  $list_header =~ s|^\s*<t([hd])>(.*)</t\1>\s*$|$2|i) {
			$list_header = qq[<th>$list_header</th>\n];
		}

		#
		# Display arguments as column headers.
		#

		print qq(<tr class="${lsn}-column-header">\n);
		print qq(<td class="${lsn}-item-index"></td>);
		print qq(<td class="${lsn}-item-buttons"></td>);
		print $list_header;
		print qq(</tr>\n);
	}
}

#------------------------------------------------------------------------------

sub display_list_footer {

	my $footer = $t->pick_footer if $table;
	$footer = $format if $footer eq 'yes';

	my $list_footer = pick::make_list_footer(
				$footer,
				$list,
				$table,
				$schema,
				$t);

	if ($list_footer) {
		# Surround $list_footer with <td> </td> tags if not already.
		if ($list_footer !~ m|^\s*<t[hd]|i
		||  $list_footer =~ s|^\s*<t([hd])>(.*)</t\1>\s*$|$2|i) {
			$list_footer = qq[<th>$list_footer</th>\n];
		}

		#
		# Display arguments as column footers.
		#

		print qq(<tr class="${lsn}-column-footer">\n);
		print qq(<td class="${lsn}-item-index"></td>);
		print qq(<td class="${lsn}-item-buttons"></td>);
		print $list_footer;
		print qq(</tr>\n);
	}

	print "</table>\n";
}

#------------------------------------------------------------------------------

sub display_list_item {

	use vars qw(@a);  	# array of values read from list query

	#
	# Make link with current key value
	#

	# First value in @a is the key value
	my $keyval = shift @a;

	my $alink_href = qq[$alink_prog]
		. ($alink_prog =~ /[?]/ ? '&' : '?')
		. qq[request=$request&table=$table&]
		. escape_html($t->key)
		. '='
		. escape_html($keyval);

	#
	# Print row
	#

	$n_rec++;

	if ($n_rec % 2) {
		print qq[<tr class="${lsn}-row-odd"]
		    . qq[ onMouseOver="hi_o(this)"]
		    . qq[ onMouseOut="lo_o(this)"]
		    . qq[ onClick="go('$alink_href')"]
		    . qq[ title="$alink_help">];
	}
	else {
		print qq[<tr class="${lsn}-row-even"]
		    . qq[ onMouseOver="hi_e(this)"]
		    . qq[ onMouseOut="lo_e(this)"]
		    . qq[ onClick="go('$alink_href')"]
		    . qq[ title="$alink_help">];
	}

	#
	# Print record index
	#

	my $ri = sprintf("%4s", $n_rec);
	$ri =~ s/ /&nbsp;/g;

	print qq(<td class="${lsn}-item-index">$ri</td>);

	#
	# Print record selection icon.
	#

	print qq(<td class="${lsn}-item-buttons">);
	print qq(</td>\n);

	#
	# Print record 
	#

	local $^W;		# turn off warnings of undefined variables in @a

	print qq(<td class="${lsn}-item">)	if $format !~ /^<[tT][dD]/;
	print clean_a_tags(sprintf_cmd($format, @a));
	print qq(</td>) 			if $format !~ /^<[tT][dD]/;
	print qq(\n</tr>\n);
}

#------------------------------------------------------------------------------

sub make_list_header {
	my ($format,$fields,$table,$schema,$t) = @_;

	# Helper function for building a list header from field names.

	$format ||= '';
	return $format unless $format =~ /\%|\&\w+\(/;	# not a sprintf() format
							# and no functions

	# Merge field titles into format string.

	my ($tab,$fld,$tt,@headers);
	for my $h (split(/\s*,\s*/,$fields)) {
		($tab,$fld) = $h =~ /(\w+)\.(\w+)/ 
		  or ($tab,$fld) = ($table,$h);

		if ($tab eq $table) { $tt = $t; }
		else                { $tt = $schema->t($tab); }

		if ($tt && $tt->{$fld}) {
			push @headers, $tt->{$fld}->title;
		}
		else {
			# Capitalize field name
			# (happens on ID fields)
			$fld =~ s/_/ /g;
			$fld =~ s/\b(\w+)\b/ucfirst(lc($1))/eg;
			push @headers, $fld;
		}
	}

	# Format header and evaluate embedded functions
	return clean_a_tags(sprintf_cmd($format,@headers));
}

sub make_list_footer {
	my ($format,$fields,$table,$schema,$t) = @_;

	# Helper function for building a list footer.
	# There is no default footer defind at this time.

	return $format || '';
}

sub clean_a_tags {
	my ($line) = @_;

	# Remove <a> tags from an output line.

	$line =~ s|</?a(\s[^>]*)?>||gsi;
	return $line;
}

sub fully_qualify {
	my $tbl = shift;	# table name
	return '' unless @_;

	# Each field/column argument can be a comma separated list of fields,
	# all of which have the $tbl table name inserted as part if the name.
	# Example: customers.email,customers.email_acknowledge

	# Return fully qualified arguments concatonated into one comma
	# separated string.

	return	join ",",			# concatonate
		map {/\./ ? $_ : "$tbl.$_"}	# fully qualify name (if needed)
		grep {$_}			# remove empty names
		map { split /\s*,\s*/ } @_;	# split each argument
}

#------------------------------------------------------------------------------
#
# List limit functions
#
#------------------------------------------------------------------------------

# Note: These are copies of the functions in Acme::Lister::Limit module.
# They allow the schema files to specify additional email addresses to 
# send the email alerts to.

sub limit_abort_func {
	my $limit = shift;

	my $to = $t->limit_alert if $t;

	my $subject = qq[Aborting: Limit of $limit for $list_id list has been exceeded.];

	my $msg = <<End_Of_Msg;
The $list_id list has exceeded its limit of $limit records.
The list has been terminated.
End_Of_Msg

	Acme::Security::alert($msg, ($to || ''), $subject);

	$msg .= <<End_Of_Msg;

These limits are usually set by administrators in the site configuration file
or in database schema files.
End_Of_Msg

	die $msg;
}

#------------------------------------------------------------------------------

sub limit_warn_func {
	my $limit = shift;

	my $to = $t->limit_alert if $t;

	my $subject = qq[Warning: Limit of $limit for $list_id list has been reached.];

	my $msg = <<End_Of_Msg;
The $list_id list has reached a limit of $limit records and is generating this notice.
The list is continuing to be rendered. This is only a warning.
End_Of_Msg

	Acme::Security::alert($msg, ($to || ''), $subject);
}

__END__
