#!/usr/bin/perl -wT -I../lib
require 5.6.0;
use strict;

=head1 NAME

query.cgi -- Query the database

=head1 SYNOPSIS

    #
    # Display a form to use to enter query parameters
    #

    query.cgi?form=customers

    query.cgi?table=customers

    query.cgi?table=customers&report=report_file_name

    query.cgi?template=orders.html&request=query

    #
    # Process the query form and list or count results
    #

    query.cgi?request=List&table=customers&last_name=SMITH


=head1 DESCRIPTION

The query.cgi program assembles CGI arguments into an AcmeBase search script and
submits them to the list.cgi program to list (or count) the results. The CGI arguments
have names like:

    customers__customer_id         -- customers table, customer_id field
    invoices__invoice_date         -- invoices table, invoice_date field

    _private                       -- private variable, discarded
    request                        -- program directive to query.cgi
                                      (if recognized)

The list.cgi program is the actual program used to list the query results.
The I<Acme::Form::Query> module simply assembles CGI arguments into an AcmeBase
search script and then invokes the list.cgi program automatically to process
the script.

If the program is passed the parameter C<request=Query>, then it displays
a form to use to enter query parameters -- just like the C<form.cgi>
program. The following two commands are equivalent:

	/cgi-bin/form.cgi?table=customers&request=Query
	/cgi-bin/query.cgi?table=customers&request=Query

=head1 CHANGE LOG

 6/17/00  Initial version.
 1/17/02  Added list_exclude to exclude rows saved in a list file from a
	  previous query.
 6/18/02  Repackaged and moved most of the functionality to Acme::Form.
 4/9/03   Added report argument this can be used as a report front end.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000,2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Acme::HTML;
use Acme::Form;
use Acme::Config;
use CGI;

use Acme::Security;
Acme::Security::check('Access');

$|=1;		# flush output (synchronize errors with output)

our $cgi = new CGI;

our $request  = $cgi->url_param('request')	||
		$cgi->param('_request')		|| $cgi->param('request') || '';

our $title    = $cgi->url_param('title')	||
		$cgi->param('_title')		|| $cgi->param('title');

our $template = $cgi->url_param('template')	||
		$cgi->param('_template')	|| $cgi->param('template');

our $form     = $cgi->url_param('form')		||
		$cgi->param('_form')		|| $cgi->param('form');

our $table    = $cgi->url_param('table')	||
		$cgi->param('_table')		|| $cgi->param('table');

our $report   = $cgi->url_param('report')	||
		$cgi->param('_report')		|| $cgi->param('report');

our $search   = $cgi->param('search');

# Guess what request should be if not specified
$request ||= 'List' if $search;
$request ||=  $cgi->param > $cgi->url_param ? 'List' : 'Query';

my $f = new Acme::Form (
		request  => $request,
		args     => $cgi,
		title    => $title,
		table    => $table,
		form     => $form,
		template => $template,
		report   => $report,
		);

if ($request =~ /Query/i) {

	$cgi->cache('no');
	print $cgi->header(-type=>'text/html', -expires=>'-1d') unless @ARGV;

	print $f->as_html;	# display query form

	exit 0;
}

my $list_title;

($search,$list_title) = $f->query if ! $search;

my $report_arg = $report ? "&report=$report" : '';

#
# Execute search
#

# The following uses the browser to redirect query to the list.cgi program.
# This introduces a limit on the search parameter from the allowed length of
# a url and it exposes the search on the address bar of the browser. 
# It allows the list program to be customized because the url is changed to
# the list program's.

print $cgi->redirect(Acme::Config->Base . "/cgi-bin/list.cgi?request=$request&table=$table$report_arg&search=$search&list_title=$list_title");

exit;

# The following executes the list program directly, fooling it to read the
# new search arguments from the standard input just like a normal CGI program.
# We need to adjust some environment variables before executing the list program.

$ENV{PATH} = '/bin:/usr/bin';			# untaint
delete @ENV{ qw(IFS CDPATH ENV BASH_ENV) };	# untaint

my $args = "request=$request&table=$table$report_arg&search=$search&list_title=$list_title";

$ENV{REQUEST_METHOD} = 'POST';
$ENV{CONTENT_LENGTH} = length($args);
delete $ENV{CONTENT_TYPE};		# so cgi does not expect boundary strings

my $list_prog = Acme::Config->Root . Acme::Config->List_Prog;
open LIST,"|-",$list_prog 
  or die "Cannot open list program '$list_prog' to perform the query. $!";
print LIST $args;
close LIST;

exit;
