#!/usr/bin/perl -wT -I../lib
require 5.6.0;
use strict;

=head1 NAME

report.cgi -- Choose and display a report

=head1 SYNOPSIS

    report.cgi?report=customers_report


=head1 DESCRIPTION

The report.cgi program lists all the reports which can be executed.

This program is simple front-end to the list.cgi or browse.cgi programs.
It uses the browse.cgi program to list the reports, and list.cgi to 
render them.

Custom reports can be installed in the list of reports simply by creating
a file and entering one line with the real url of the customer report, like:

	report_url=http://www.acmebase.com/cgi-bin/custom_report.cgi?param=1

Report files are simple files written by the CGI module. They consist of
a key, an equals sign, and a value on each line (myvar=value). Blank lines
and comments are not allowed.

Subdirectories to the report directory are assumed to be for database table
reports, so the directory name is checked with security before listing its
contents. Restricting access to a database table in the .../etc/access.conf 
essentially restricts access to its reports.

By convention, if the report_id of the report being opened matches the
list_id of the current list that is rendered, the search parameter of the
current list overrides the search parameter saved with the report. That
means that if the report is to always execute its own search (which can
define parameters to enter), then change the report_id value to be a unique
non-list_id value. Because list.cgi automatically defines missing list_id's
to be table names, change the report_id in the report to a value that will
always be an illegal table name, such as "customers-1".

=head1 NOTE

Report files are simply the list.cgi CGI parameters written to a file. Most
punctuation and non-alphanumeric characters are "escaped" using HTTP encoding.
However, the CGI module is very forgiving when reading in a report file.
Very few characters actually need to be escaped. The semi-colon character is
one that DOES need to be escaped. It needs to be specified as C<%3B>.
Search scripts are also delimited by semi-colons, but they can also be
delimited the old way by using the pipe character (|), which might be easier
to use when hand editing a report file.

=head1 CHANGE LOG

 12/2/02 Initial version.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Acme::HTTP		qw(escape);
use Acme::HTML;
use Acme::Config;
use CGI;

use Acme::Security;
Acme::Security::check('Access');

$|=1;		# flush output (synchronize errors with output)

our $Base        = Acme::Config->Base;
our $Root        = Acme::Config->Root;
our $Report_Dir  = Acme::Config->Report_Space;
our $List_Prog   = Acme::Config->List_Prog;
our $Browse_Prog = Acme::Config->Browse_Prog;

our $cgi = new CGI;

# Debug needs to output header as program normally does a redirect to a url.
#print $cgi->cache('no');
#print $cgi->header(-type=>'text/html') unless @ARGV;

our $report   = $cgi->url_param('report')	|| $cgi->param('report') ||
		$cgi->url_param('reports')	|| $cgi->param('reports');

our $search   = $cgi->param('search');
our $table    = $cgi->param('table')   || '';	# used as flag for using search
our $list_id  = $cgi->param('list_id') || $table || '';

unless ($report) {

	#
	# Show list of reports.
	#

	my @dirs;
	if ($list_id && Acme::Security::allow('Access', $list_id)) {
		@dirs = ($list_id,'/');
	}
	else {
		# Check list of report directories with security.
		@dirs = grep { Acme::Security::allow('Access', $_) } '/', report_dirs();
	}
	
	#
	# Redirect to the file browser.
	#

	my $lister = "$List_Prog"
		. ($List_Prog =~ /\?/ ? '&' : '?')
		. 'report=$';

	my $url = $Base
		. (substr($Browse_Prog,0,1) ne '/' ? '/' : '')
		. $Browse_Prog
		. "?reports="
		. (join '&reports=', @dirs)
		. "&link=" . escape($lister)
		. "&title=Reports";

	$url .= "&search=$search"   if $search;
	$url .= "&table=$table"     if $table;
	$url .= "&list_id=$list_id" if $list_id && $list_id ne $table;

	print $cgi->redirect($url);

	exit;
}

#
# Redirect to the lister.
#

my $url = $Base
	. (substr($List_Prog,0,1) ne '/' ? '/' : '')
	. $List_Prog
	. "?report=$report";

$url .= "&search=$search"   if $search;
$url .= "&table=$table"     if $table;
$url .= "&list_id=$list_id" if $list_id && $list_id ne $table;

#########
###print $cgi->redirect($url);
###exit;
#########

# Note: we change our environment and pipe the cgi arguments to
# the list program. This "preserves" the state of the list.cgi
# program a little better than using $cgi->redirect() because 
# the list option settings exceeds the length of urls accepted
# by web servers.

#
# Get full program path of list program.
#

my $list_prog = $List_Prog;
my $list_prog_path = $Root . (substr($list_prog,0,1) eq '/' ? '' : '/') . "$list_prog";
my $list_prog_url  = $Base . (substr($list_prog,0,1) eq '/' ? '' : '/') . "$list_prog";

#
# Change working directory to directory of list program.
#

my ($dir,$prog) = $list_prog_path =~ m|(.*/)(.*)|;
chdir $dir or die "Cannot change directory to $dir. $!";

#
# Get modified cgi program paramters.
#

#(my $query_string = $cgi->url(-query=>1)) =~ s/.*\?//;
#my ($query_string) = $url =~ /.*?\?(.*)/;
my $query_string = $url;

#
# Execute list program.
#

$ENV{PATH} = '/bin:/usr/bin';                   # untaint
delete @ENV{ qw(IFS CDPATH ENV BASH_ENV) };     # untaint

#
# Set environment variables so that CGI in list.cgi reads input.
#

$ENV{REQUEST_METHOD}='POST';
$ENV{CONTENT_TYPE} = 'application/x-www-form-urlencoded';
$ENV{CONTENT_LENGTH}=length $query_string;
$ENV{SCRIPT_NAME} = $list_prog;
$ENV{SCRIPT_FILENAME} = $list_prog_path;
$ENV{HTTP_REFERER} = $list_prog_url;
$ENV{REQUEST_URI} = $list_prog;

open LIST, '|-', "./$prog" or die "Cannot open $prog. $!";
print LIST $query_string;
close LIST;

exit;

#------------------------------------------------------------------------------

sub report_dirs {

	opendir DIR, $Report_Dir or return;

	my @dirs = readdir DIR;
	closedir DIR;
	shift @dirs; shift @dirs;	# remove . and ..

	#
	# Filter files for directories
	#

	return grep { -d "$Report_Dir/$_" } @dirs;
}
__END__
