#!/usr/bin/perl -wT -I../lib

require 5.6.0;
use strict;

=head1 NAME

save.cgi -- Save form information to database

=head1 SYNOPSIS

    save.cgi?form=customers

    save.cgi?table=customers

    save.cgi?template=orders.html&request=new

=head1 DESCRIPTION

The save.cgi program saves CGI arguments to the database. The CGI arguments
have names like:

    customers__customer_id         -- customers table, customer_id field
    invoices__invoice_date         -- invoices table, invoice_date field
    N3_invoice_items__description  -- third record of invoice_items table,
                                      description field

    _private                       -- private variable, discarded
    request                        -- program directive to save.cgi
                                      (if recognized)

Acme::Base::Schema is used to determine default table relationships and validation
parameters.

Some forms are used to enter "detail" records of another form -- the "master"
form.  If a form being saved has a master form (determined by the _master
cgi parameter), then the just saved record is formatted and returned in the
C<detail_result> field.  The C<detail_result> is enclosed in a hidden C<<div>>
tag and not displayed in the confirmation screen but is instead pasted into
the master form.

=head1 CHANGE LOG

 5/26/00  Initial version.
 3/16/02  Major update to add audit table processing. Also added a reconciliation
          screen if someone changed a record while it was being edited.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000,2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Acme::HTML;
use Acme::Form;
use CGI;

use Acme::Security;
Acme::Security::check('Access');

$|=1;		# flush output (synchronize errors with output)

our $cgi = new CGI;
$cgi->cache('no');
print $cgi->header(-type=>'text/html', -expires=>'-1d') unless @ARGV;

our $request  = $cgi->url_param('request')	||
		$cgi->param('_request')		|| $cgi->param('request') || '';

our $title    = $cgi->url_param('title')	||
		$cgi->param('_title')		|| $cgi->param('title');

our $template = $cgi->url_param('template')	||
		$cgi->param('_template')	|| $cgi->param('template');

our $form     = $cgi->url_param('form')		||
		$cgi->param('_form')		|| $cgi->param('form');

our $table    = $cgi->url_param('table')	||
		$cgi->param('_table')		|| $cgi->param('table');

my $f = new Acme::Form (
		request  => $request,
		args     => $cgi,
		title    => $title,
		table    => $table,
		form     => $form,
		template => $template,
		);

if ($f->save) {

	# Keep the result screen open when posting new records to the database.
	# The ID number of the record is often used by the user for their 
	# paper trail. Close the message screen automatically if the record
	# was being edited or if the new record is a detail record added to
	# another (master) table, as those ID numbers tend to be trivial. 

	my $auto_close = 0	# seconds to wait before closing window
	  if $request =~ /new/i && $f->has_id_key() && ! $cgi->param('_master');

	Acme::HTML::display_message(
		$f->result,
		'Message',
		'form_save_return(window.name)',  # JavaScript return function
		$auto_close	# if undefined, user's setting is used.
		);
}
else {
	Acme::HTML::display_message(
		'There were no differences between your form values' .
		  ' and the existing database values. Nothing was saved.',
		'Message',
		'form_save_return(window.name)',  # JavaScript return function
		0		# keep message window open until user closes it
		);
}

exit;
