#!/usr/bin/perl -wT -I../lib

=head1 NAME

    Search - Search an item from the database

=head1 SYNOPSIS

    cgi-bin/search.cgi

    Arguments to the  program are:
	request	-- set to 'Search' to initiate search,
		   otherwise a dialog is displayed to enter search parameters
	sql	-- sql script (required)
	search	-- search script for Acme::Searcher
	format	-- display format of list
	rows	-- number of lines in scrolling list box
	title	-- title displayed over list box
	autoselect	-- if set to anything and only one record found,
			   return selection without user confirmation.

    Optional arguments for the script preprocessor if search script specified;
	id_field	 -- id field
	code_field	 -- code field
	name_field	 -- name field
	first_name_field -- first name field
	last_name_field	 -- last name field
	phone_field	 -- phone field 
	email_field	 -- email field 

	save_history	 -- saves search in specified history list

=head1 DESCRIPTION

The C<search.cgi> program displays a simple scrolling list box so that the user
can select a record. This package was intended to be a tool in presenting to the
user a pop-up look-up window for an entry form. 

When the user selects the item, the calling window's JavaScript function
_search_return_WINDOWNAME() is executed, where WINDOWNAME is the name of the
search window. It reads the variables defined in the search window's 
"result" object and closes the window when done.

The _search_return_WINDOWNAME() function must be aware that if no records were
found in the search, return variables are not defined.

It is the responsibility of the search window to close itself. Until
January, 2002 it was the responsibility of the calling window to close the
search look-up window, but that was changed because if the calling window
was refreshed by the user, it looses the window handle it needs to close
the search window.

Variables defined in the search window depend upon the column names in the
sql script specified in the C<sql> argument.

A search script as defined by I<Acme::Searcher> can also be included in the
arguments. It is combined with the sql script before executing.

If a search script is specified and one or more of C<id_field>, C<code_name>, C<name_field>,
C<first_name_field>, C<last_name_field>, C<phone_field>, or C<email_field> is specified,
the script is preprocessed (see Search::preprocess).

Setting the C<save_history> argument to the name of a history list saves
the search script in the history list.

Setting autoselect to "yes" automatically selects a record without user
confirmation if the record was the only record found when searching.

The display of the information in the list box in the search window is 
determined by the formatting string defined in the C<format> argument.
It is used by sprintf().
A format string may be specified by using the 'format' parameter. It is a
string used by I<Local::Acme::Format>, which is essentially a string to be formatted
by sprintf() and then scanned for functions to execute. The functions need
to be defined in I<Local::Acme::Format> in order to be recognized in the formatted
string.

Unlike the formatting in list.cgi or pick.cgi, the format here cannot contain
any HTML codes as the text itself is displayed in a HTML <select> tag.

= head2 Example

Here are example JavaScript functions placed in the calling window. Note
that the name of the new pop-up window used to display the search is used
to determine the window handle variable (_search_win_ prefix), the function
to use to initiate the search (_search_ prefix), and the function to call
after the user has made their selection (_search_return_ prefix):

	<script language=JavaScript type="text/javascript">
	var _search_win_supplier;	// search window handle
					// window name w. '_search_win_' prefix
	function _search_supplier (options) {	// window name with '_search_'
		var window_name = "supplier";	// suffix to _search_win_,
						// _search_, and _return_search_
		var my_search = escape("supplier_type matches " +
		  self.document.myform.supplier_type.value);

		var table = "suppliers";

		_search_win_supplier=window.open("",window_name,
		  "width=875,height=500,resizable");

		_search_win_supplier.document.write("<br>Searching ...");

		// display scrolling list box
		_search_win_supplier.location.href = "/cgi-bin/search.cgi" +
		  "?search=" + my_search +
		  "&table=" + table + "&format=" + format + "&rows=12" +
		  (options ? "&" + options : "");
	}

	function _search_return_supplier () {
		var f = self.document.myform;
		if (_search_win_supplier && _search_win_supplier.result) {
			var r = _search_win_supplier.result;

			// assign selection values to fields
			f.id.value    = r.supplier_id.value;
			f.name.value  = r.supplier_name.value;
			f.phone.value = r.phone.value;
		} else {
			// no records were found
			f.id.focus();
		}
	}

	function _search_error_return(window_name) {
		eval("_search_return_"+window_name+"()");
	}
	</script>


=head1 CHANGE LOG

 1/6/99   Initial version.
 4/28/00  Generalize program to use Acme::Schema definitions.
 5/5/01   Convert to Acme::Base::Schema.
 6/14/01  Convert to Acme::SQL
 6/26/01  Changed result object to be like a form object with field values.
 11/6/01  Added phones and email fields search preprocess

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1999-2000, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

package search;

use Acme::HTTP		qw(escape		unescape);
use Acme::HTML		qw(display_message	escape_html);
use Acme::Db;
use Acme::SQL;
use Acme::Searcher;
use Acme::Options;
use Acme::Base::Schema;
use Acme::HTML::ListMenus;
use Local::Acme::Format	qw(sprintf_cmd);
use Acme::Config;
use CGI;

use Acme::Security;
Acme::Security::check('Access');

$Images      = Acme::Config->Images;		# location of image files
$Styles      = Acme::Config->Styles;		# location of css stylesheets
$Scripts     = Acme::Config->Scripts;		# location of JavaScript scripts
$Scripts_Lib = Acme::Config->Scripts_Lib;	# location of JavaScript library
$Help        = Acme::Config->Help;		# location of help files

$Rows = 11;	# default number of rows displayed in selection list

$|=1;		# flush output (synchronize errors with output)

$cgi = new CGI;
die "No arguments\n" unless $cgi->param();
print $cgi->header(-type=>'text/html') unless @ARGV;

#
# Program Arguments
#

local $request = $cgi->param('request') || '';

if ($request =~ /Search/i) {
	do_search();
}
else {
	display_search_dialog();
}
exit;

#------------------------------------------------------------------------------

sub do_search {

	local $sql		 = $cgi->param('sql')		   || '';
	local $search		 = $cgi->param('search')	   || '';
	local $title		 = $cgi->param('title')		   || '';
	local $format		 = $cgi->param('format')	   || '';
	local $rows		 = $cgi->param('rows')		   || '';
	local $autoselect	 = $cgi->param('autoselect')	   || 'yes';

	local $id_field		 = $cgi->param('id_field')	   || '';
	local $code_field	 = $cgi->param('code_field')	   || '';
	local $name_field	 = $cgi->param('name_field')	   || '';
	local $first_name_field  = $cgi->param('first_name_field') || '';
	local $last_name_field	 = $cgi->param('last_name_field')  || '';
	local $phone_field	 = $cgi->param('phone_field')	   || '';
	local $email_field	 = $cgi->param('email_field')	   || '';

	local $code_field_match  = $cgi->param('code_field_match') || '';
	local $phone_field_match = $cgi->param('phone_field_match')|| '';

	local $save_history_list = $cgi->param('save_history')	   || '';

	#
	# Parse SQL statement
	#

	local $sqlo = new Acme::SQL $sql;

	local @sql_columns = $sqlo->columns;

	s/\W/__/g foreach @sql_columns;		# change table.col to table__col

	#
	# Set defaults
	#

	$title  ||= "Search List Item";
	$rows   ||= 7;
	$format ||= "%s " x @sql_columns;

	#
	# Evaluate Search script
	#

	if ($search) {
		if ($save_history_list) {
			my $opt = new Acme::Options;
			$opt->enter($save_history_list,$search);
		}

		if ($id_field || $code_field || $name_field || $last_name_field || $phone_field || $email_field) {

			#
			# Preprocess search script
			#

			if ($first_name_field || $last_name_field) {
				$search = Acme::Searcher::preprocess($search, {
				  RecordSetName   => $record_set_name,
				  IDField	  => $id_field,
				  CodeField	  => $code_field,
				  NameField       => $name_field,
				  FirstNameField  => $first_name_field,
				  LastNameField   => $last_name_field,
				  PhoneField      => $phone_field,
				  EmailField      => $email_field,
				  CodeFieldMatch  => $code_field_match,
				  PhoneFieldMatch => $phone_field_match,
				});
			}
			else {
				$search = Acme::Searcher::preprocess($search, {
				  RecordSetName   => $record_set_name,
				  IDField	  => $id_field,
				  CodeField	  => $code_field,
				  NameField	  => $name_field,
				  PhoneField      => $phone_field,
				  EmailField      => $email_field,
				  CodeFieldMatch  => $code_field_match,
				  PhoneFieldMatch => $phone_field_match,
				});
			}
		}

		#
		# Combine sql with search script.
		#

		my $srch = new Acme::Searcher $search, $cgi;
		$sql = $srch->as_sql($sqlo);

		#
		# Redefine title if specified in search script.
		#

		$title = $srch->title if $srch->title;
	}

	#
	# Check query with security
	#

	$sql = Acme::Security::clean_sql($sql);

	#
	# Execute query
	#

	$q = Acme::Db->query($sql);

	if ($q->count() == 0) {
		display_message("No records were found.",
		  "Search","_search_error_return(window.name)",0);
		exit 1;
	}

	$q->bind_columns_to_variables();

	#
	# Define JavaScript statements used to track list values and to
	# assign variables used for return values. The variable names come from
	# the field names in the SQL statement.
	#
	# Note that database text is "escaped" by the server before producing
	# the JavaScript variable assignment text, and unescaped before
	# assigning the return variable on the client side.
	#

	my $js_array_defs =
	  qq(// Define arrays to hold list values for each list item.\n)
	  . join "\n", map { "var a_${_} = new Array;" } @sql_columns;

	my $js_result_defs =
	  qq(// Define result object to hold selected search item values.\n)
	  . qq(var result = new Object;\n)
	  . join "\n", map { qq(result["${_}"] = new Object;) } @sql_columns;

	my $js_return_var_assign =
	  qq(// Assign return variables\n\t\t)
	  . join "\n\t\t", map { qq(result["${_}"].value = unescape(a_${_}[i]);); } @sql_columns;

	#
	# Display Search Window
	#

# Keep IE6 in IE5 compatablilty mode by not specifying url
#<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
#"http://www.w3.org/TR/html4/loose.dtd">

	print <<End_Of_HTML;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">

<html>
<head>
<title>$title</title>

<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<link rel="stylesheet" href="$Styles/form.css" type="text/css">

<script src="$Scripts/common.js" type="text/javascript"></script>
<script src="$Scripts_Lib/form_events.js" type="text/javascript"></script>

<script language="JavaScript" type="text/javascript">

$js_array_defs

$js_result_defs

function search_return() {
	var i = self.document.search_form.search_options.selectedIndex;
	if (i >= 0) {
		$js_return_var_assign
	}

	// Return to calling window
	var w = window.opener;
	if (!w || w.closed) self.close();	// user closed caller already

	// Construct return function name
	var return_func = "_search_return_" + window.name;
	if (eval("typeof w." + return_func) == "undefined") {
		self.close();
		return;	
	}

	// Normal return. At this point hide and let the calling window
	// read the return variables set according to the user's selection.
	self.blur();				// hide window
	w.focus();				// bring to top calling window
	eval("w." + return_func + "()");	// return to calling window func

	self.close();
}

function enterKey_search_form() {
	search_return();
}

// Define arrays of list values
End_Of_HTML

	#
	# Output array assignments (like: a_field[1] = $field;)
	# Also build list box html for output later. Don't display it now
	# to assure all the JavaScript variables are defined before selections
	# are displayed to the user.
	#

	my $list_box_def = <<End_Of_HTML;

<!-- List box of database selections -->
<select name="search_options" size="$rows" onDblClick="search_return()">
End_Of_HTML

	my $d;
	my $count = 0;
	while ( my @f = $q->next ) {

		local $^W;	# turn off -w processing

		# Build list box definition.
		# Option value is the first sql field value.
		$d = sprintf_cmd($format,@f[1..$#f]);
		$list_box_def .= qq( <option)
		  . ($count == 0 ? " selected" : "")
		  . ">$d"
		  . "</option>\n";

		# Print array assignments
		print qq(a_${_}[$count] = "), escape("$$_"), qq(";\n)
		  foreach (@sql_columns);

		$count++;
	}
	$list_box_def .= "</select>";

	if ($count == 0) {
		# No records found.
		# Substitute error message for normal list box definition.

		$list_box_def = <<End_Of_HTML;
<p style="font-size:large; margin-top:24pt">
No records were found.
</p>
End_Of_HTML
	}

	if ($count == 1 and $autoselect =~ /yes/i) {
		# Only one record found.
		# Substitute for normal list box definition a script to
		# return immediately.

		print "\nself.blur();		// hide window\n";

		$list_box_def .= <<End_Of_HTML;
<script language="JavaScript" type="text/javascript">search_return()</script>
End_Of_HTML

	}
	else {
		# Resize message window to display search results.
		print "\nwindow.resizeTo(850,450)";
	}

	#
	# Print rest of search window html
	#

	print <<End_Of_HTML;
</script>

</head>
<body onLoad="with(self.document.search_form){if(search_options) search_options.focus()}">

<form	name="search_form"
	onSubmit="search_return();"
	style="margin-top:5pt; margin-bottom:1pt"
>

<table width=100% border=1><tr><td>
<table cellpadding=0 cellspacing=0 border=0 width=100%>
<tr>
 <td>
 <table cellspacing=0 cellpadding=2 style="margin:6pt 0pt; background-color:white; width:100%; height:50">
 <tr valign=middle>
 <td width=9>
  <a href="javascript:console()" tabindex=-1><img
     src="$Images/local/company_icon.gif"
     name='console'
     alt='Console'
     border=0
  ></a>
 </td>
 <td>
  <span class="title">$title</span>
 </td>
 </tr>
 </table>
</td>
<td align=right>
 <table cellspacing=0 cellpadding=2 style="margin:6pt 0pt; background-color:white; width:100%; height:50"><tr><td align=right>
<input type="button" value="  Select  " onClick="search_return()" class=button>
<input type="button" value="  Cancel  " onClick="self.close()" class=button>
 </td></tr></table>
</td>
</tr>
</table>
</td></tr></table>

<div style="text-align:center; margin-top:15pt; margin-bottom:1pt">
$list_box_def
</div>

</form>
</body>
</html>
End_Of_HTML

}

#------------------------------------------------------------------------------

sub display_search_dialog {

	# The form displayed by this routine re-calls this script when
	# submitted.

	local $sql   = $cgi->param('sql');
	local $table = $cgi->param('table');

	die "Empty search\n" unless $sql || $table; # no sql or table defined

	local $search = $cgi->param('search')			    || '';
	local $title  = $cgi->param('title')			    || '';
	local $format = $cgi->param('format')			    || '';
	local $autosearch = $cgi->param('autosearch')		    || '';

	# Following parameters are usually defined here, but they can
	# be optionally predefined.
	local $rows		  = $cgi->param('rows')		    || '';
	local $autoselect	  = $cgi->param('autoselect')	    || '';
	local $save_history	  = $cgi->param('save_history')	    || '';
	local $id_field		  = $cgi->param('id_field')	    || '';
	local $code_field	  = $cgi->param('code_field')	    || '';
	local $name_field	  = $cgi->param('name_field')	    || '';
	local $first_name_field	  = $cgi->param('first_name_field') || '';
	local $last_name_field	  = $cgi->param('last_name_field')  || '';
	local $phone_field	  = $cgi->param('phone_field')	    || '';
	local $email_field	  = $cgi->param('email_field')	    || '';

	local $table_title	  = 'Search List';	# default
	local $search_fields_help = '';			# initialize
	local $search_history	  = '';			# initialize

	local $opt = new Acme::Options;

	local $row_name;

	my ($schema, $t);

	if ($table) {
		$schema = new Acme::Base::Schema;
		$t = $schema->t($table) or die("Unknown table '$table'");

		$row_name = $t->row_name;

		$search ||= $opt->get("${row_name}Search");
		$title  ||= "Search for $row_name";

		$sql    ||= $t->select_sql;

		$format ||= $t->select_format;
		$format =~ s/^\s*(["'])(.*)\1\s*$/$2/;	# remove possible quotes

		$rows   ||= $t->select_rows || $Rows;

		$autoselect   ||= "yes";
		$save_history ||= "${table}Search";

		$id_field    = fq_fields($id_field    || $t->id_field,    $table)	|| '';
		$code_field  = fq_fields($code_field  || $t->code_field,  $table)	|| '';
		$name_field  = fq_fields($name_field  || $t->name_field,  $table)	|| '';
		$phone_field = fq_fields($phone_field || $t->phone_field, $table)	|| '';
		$email_field = fq_fields($email_field || $t->email_field, $table)	|| '';
		$first_name_field = fq_fields($first_name_field || $t->first_name_field,$table) || '';
		$last_name_field  = fq_fields($last_name_field  || $t->last_name_field,$table) || '';
	}

	if ($autosearch =~ /yes/i) {

		# Execute search immediately

		$cgi->param('search',$search);
		$cgi->param('sql',$sql);
		$cgi->param('title',$title);
		$cgi->param('format',$format);
		$cgi->param('rows',$rows);
		$cgi->param('autoselect',$autoselect);
		$cgi->param('save_history',$save_history);
		$cgi->param('id_field',$id_field);
		$cgi->param('code_field',$code_field);
		$cgi->param('name_field',$name_field);
		$cgi->param('first_name_field',$first_name_field);
		$cgi->param('last_name_field',$last_name_field);
		$cgi->param('phone_field',$phone_field);
		$cgi->param('email_field',$email_field);

		return do_search();
	}

	if ($table) {
		$table_title = $t->title;

		$search_fields_help = <<End_Of_HTML;
      <br><a href="/cgi-bin/schema.cgi?table=$table" onClick="new_target(this,'doc')" class="link">Fields</a>
End_Of_HTML

		#
		# Prebuild search history list control
		#

		my @a = $opt->list("${table}Search");

		my $attr = 'class="history" style="width:100%" onChange="if (this.value) self.document.forms[0].search.value = this.value;self.document.forms[0].search.focus()" tabindex="-1"';

		my $search_history_options = Acme::HTML::ListMenus->Menu(\@a,'search_history','',40,$attr,'','noundef');

		$search_history = <<End_Of_HTML;
<form style="margin:0px">
<table border="0" cellspacing="0" cellpadding="2">
  <tr valign="top"> 
    <td width=16><img src=$Images/transparent.gif></td>
    <td width=105 align=right nowrap>
      <font style="font-size:x-small; color:#666666">History</font> 
    </td>
    <td width=340> 
    $search_history_options
    </td>
    <td width=105><img src=$Images/transparent.gif></td>
    <td width=16><img src=$Images/transparent.gif></td>
  </tr>
</table>
</form>
End_Of_HTML

		$search ||= $a[0] || '';	# initialize to last search
	}

	#
	# Display Search Window
	#

	$sql = escape_html($sql);	# prepare for being an <input> value
	$format = escape_html($format);	# prepare for being an <input> value

	print <<End_Of_HTML;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">

<html>
<head>
<title>$title</title>

<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<link rel="stylesheet" href="$Styles/console.css" type="text/css">

<script src="$Scripts/common.js" type="text/javascript"></script>
<script src="$Scripts_Lib/form_events.js" type="text/javascript"></script>

<script language="JavaScript" type="text/javascript">
self.focus();		// assure window is displayed
var SearchEnter = "@{[$opt->SearchEnter]}";

function start_search() {
        // Display selection
        self.document.forms[0].submit();
	return true;
}
function enterKey_search_form() {
	start_search();
}
window.resizeTo(600,300);
</script>

</head>
<body onLoad="with(self.document.forms[0].search){focus();select()}">
<span class="title">$title</span>
<div style="text-align:center; margin-top:5pt; margin-bottom:1pt">

<!-- Start Search Control -->
<div id="search_control" style="display:block; text-align:center; margin-top:20pt">

<form name="search_form" method="post" action="/cgi-bin/search.cgi" style="margin:0px">

<table border="0" cellspacing="0" cellpadding="2">
  <tr valign="top"> 
    <td width=16>
    </td>

    <td width=105 align=right nowrap><font class="plain">$table_title</font>

      <a href="${Help}/search_instructions.htm" onClick="new_target(this,'help')" class="link">Instructions</a>
      $search_fields_help
    </td>

    <td width=340> 
      <textarea name="search" rows="5" style="width:100%">$search</textarea>
    </td>

    <td width=105> 

      <table border="0" cellspacing="0" cellpadding="1">
        <tr> 
          <td> 
            <button class="control" accesskey="S" onClick='start_search()'><u>S</u>earch</button>
          </td>
        </tr>
        <tr> 
          <td>
            <button class="control" accesskey="C" onClick='self.close()'><u>C</u>ancel</button>
          </td>
        </tr>
        <tr> 
          <td>
          </td>
        </tr>
        <tr>
          <td>
          </td>
        </tr>
      </table>

    </td>

    <td width=16><img src='$Images/blank.gif'></td>
  </tr>
</table>

<input type="hidden" name="request" value="Search">
<input type="hidden" name="sql" value="$sql">
<input type="hidden" name="title" value="$title">
<input type="hidden" name="format" value="$format">
<input type="hidden" name="rows" value="$rows">

<input type="hidden" name="autoselect" value="$autoselect">
<input type="hidden" name="save_history" value="$save_history">
<input type="hidden" name="id_field" value="$id_field">
<input type="hidden" name="code_field" value="$code_field">
<input type="hidden" name="name_field" value="$name_field">
<input type="hidden" name="first_name_field" value="$first_name_field">
<input type="hidden" name="last_name_field" value="$last_name_field">
<input type="hidden" name="phone_field" value="$phone_field">
<input type="hidden" name="email_field" value="$email_field">
</form>
$search_history
</div>

</div>
</body>
</html>
End_Of_HTML
}

sub fq_fields {
	my ($list,$table) = @_;
	# Return a fully qualified list of fields.
	return join ",", map { /\./ ? $_ : "$table.$_" } split(/\s*,\s*/, $list);
}
__END__
