#!/usr/bin/perl -wT -I../lib

=head1 NAME

searches.cgi -- Display canned searches forms

=head1 SYNOPSIS

    searches.cgi?template=customer_searches.html

=head1 DESCRIPTION

The searches.cgi program is used to fill in default values and drop down
lists of database information in forms used to enter parameters for canned
searches.

Only <input> fields are substituted, and only if their field name parses
to a table/field that is listed in the database schema.  Fields that are
not named like "table__field" are ignored.

=head1 CHANGE LOG

 8/5/00   Initial version.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

$|=1;

use Acme::HTML				qw(unescape_html);
use Acme::HTML::ListMenus;
use Acme::Base::Schema;
use Acme::Config;
use Local::Acme::Form::Extensions	qw(tag_command);
use CGI;

use Acme::Security;
Acme::Security::check('Access');

#
# Begin processing form
#

$|=1;		# flush output (synchronize errors with output)

$cgi = new CGI;
$cgi->cache('no');
print $cgi->header(-type=>'text/html', -expires=>'-1d') unless @ARGV;

#
# Parse cgi arguments
#

$Template_file = $cgi->url_param('template')	|| $cgi->param('template');
$Template_file = $cgi->url_param('form')	|| $cgi->param('form')
  unless $Template_file; 		# alternative template keyword
die "No template file specified for search.\n"
  unless $Template_file; 		

#
# Prepend directory of template to template file specification
#

$Template_file = Acme::Config->Templates
  . ($Template_file !~ m|^/| ? '/' : '')
  . $Template_file;

#
# Read form template
#

open(TEMPLATE, "<$Template_file")
  or die "Cannot open form template '$Template_file' $!.\n";
local $/;
$Template = <TEMPLATE>;
close(TEMPLATE);

#
# Setup database schema
#

$Schema = new Acme::Base::Schema;

#
# Substitute commands
#

$Template =~ s/(<!--+\$+\s*.*?\s*--+>)/tag_command($1)/egs;

#
# Substitute <input> tags.
#

$Template =~ s{(<input\s.*?>)}
	      {my_input_tag_substitution() || tag_input($1)}egsi;

print $Template;
exit;

#-------------------------------------------------------------------------------

sub my_input_tag_substitution {
	my $tag = shift;

	#####
	# Put custom substitutions here
	#####
	return;
}

#-------------------------------------------------------------------------------

sub tag_input {
	my $tag = shift;

	my ($tag_name) = $tag =~ m/name\s*=\s*"(.*?)"/s;

	return $tag unless $tag_name;		# ignore tags w/o names

	# Parse table and field names from form field name
	($table_name = $tag_name) =~ s/^([A-Z]_|\d+_)(?=[^_])//;# strip prefix
	my ($table, $field) = $table_name =~ /(\w*?)__(.*)/;

	return $tag unless $table;
	return $tag unless $Schema;

	my $t = $Schema->t($table);
	return $tag unless $t;

	my $f = $t->f($field);		# can be undefined if field = ID
	return $tag unless $f;

	my ($tag_value)	= $tag =~ m/value\s*=\s*"(.*?)"/s;
	my ($tag_size)	= $tag =~ m/size\s*=\s*"(\d*?)"/;
	my ($tag_type)	= $tag =~ m/type\s*=\s*"(.*?)"/s;
	($tag_type)	= $tag =~ m/type=(.*?)\s/ unless $tag_type;	#safety
	my ($tabindex)	= $tag =~ m/(tabindex\s*=\s*["']?[-]?\d+["']?)/;

	my ($title)	= $tag =~ m/(title\s*=\s*".*?")/s;
	$title		= '' unless $title;	# to silence -w warnings
	$tabindex	= '' unless $tabindex;	# to silence -w warnings
	my $attr	= "$tabindex $title";

	my $form_tag_value = $tag_value;	# save for setting checkboxes

	#
	# Substitute dropdown list or menu for input field
	#

	my $default = unescape_html($tag_value);

	my $format = $f->choose_format;
	$format =~ s/^(["'])(.*)\1$/$2/s;	# strip quotes if found

	my $attributes = $attr;
	if (my $select_rows = $f->choose_rows) {
		$attributes .= qq( size="$select_rows");
	}

	my $size = 0;		# width of list/menu if not equal to 0

	my $multi = $f->array ? 'multiple' : '';

	my $tbl = $f->choose_tables;

	if ($tbl && Acme::HTML::ListMenus->can($tbl)) {
		return Acme::HTML::ListMenus->$tbl(
						$tag_name,
						$default,
						$size,
						$attributes,
						$format,
						$multi);
	}

	if (my $menu = $f->choose_menu) {
		my @menu_opt = Text::ParseWords::parse_line('\s+',0,$menu);
		return Acme::HTML::ListMenus->Menu(\@menu_opt,
						$tag_name,
						$default,
						$size,
						$attributes,
						$format);
	}

	if (my $list = $f->choose_list) {
		my %list_opt = Text::ParseWords::parse_line('\s+',0,$list);
		return Acme::HTML::ListMenus->List(\%list_opt,
						$tag_name,
						$default,
						$size,
						$attributes,
						$format);
	}

	if (my $sql = $f->choose_sql) {
		$sql = Acme::Security::clean_sql($sql);
		return Acme::HTML::ListMenus->List(\$sql,
						$tag_name,
						$default,
						$size,
						$attributes,
						$format,
						$multi);
	}

	#
	# Substitute values into <input> fields.
	#

	TAG_TYPE: {
		if ($tag_type =~ /text/i
		or  $tag_type =~ /file/i
		or  $tag_type =~ /hidden/i) {
			# Substitue current value into field.
			$tag =~ s/(value\s*=\s*")[^"]*?(")/$1.$tag_value.$2/se;
			last TAG_TYPE;
		}

		if ($tag_type =~ /checkbox/i
		 or $tag_type =~ /radio/i) {
			# Adjust "checked" settings
			$tag =~ s/ checked\b//;		# remove checked setting
			if ($form_tag_value eq $tag_value) {
				$tag =~ s/>/ checked>/s;
			}
			last TAG_TYPE;
		}
	}

	return $tag;
}
__END__
