#!/usr/bin/perl -wT -I../lib
require 5.6.0;
use strict;

=head1 NAME

simplelist.cgi -- Display database information in a list format.

=head1 SYNOPSIS

    simplelist.cgi?table=customers

    simplelist.cgi?table=customers&search=Folly

    simplelist.cgi?sql=select+f1 ...

    simplelist.cgi?sql=select+f1...&search=Foo

=head1 DESCRIPTION

The simplelist.cgi program displays rows of data from a database table or from
the sql script passed as an argument. The output from the script is filtered
by the search parameter if specified.

There are several options affecting the display of a list. The 'request'
parameter determines if the list is to be used for list operations
(request=List or Count) or if the list is to be used to select one row for
other operations (request=Edit, View, or New).

Each record listed has an icon displayed at the beginning of the record. The
icon is a link to select the record. The link is dependent upon 'request'
argument -- there is one for selecting the record for new entries, one for
editing the record. In the future, there will be one to select/deselect the
row from a list operation (like export), and there will be one for deleting
the record (maybe).

A format string may be specified by using the 'format' parameter. It is a
string used by I<Local::Acme::Format>, which is essentially a string to be formatted
by sprintf() and then scanned for functions to execute. The functions need
to be defined in I<Local::Acme::Format> in order to be recognized in the formatted
string.

A list is formatted as an HTML table, with each list item placed between <td>
and </td> tags. That allows the format string to also have <td> and </td>
tags embedded in it to format fields into table columns.

User options are used to determine the display of the list. A list ID is used
to determine specific list options. In the future, if a table name is
specified and a list ID is not, then the table name is used as a list ID.
Currently, only 'table' and 'sql' list IDs are used.

User options have the list ID prepended to list options. For example, if
the list ID is 'table', then the options tableListAsTable and
tableListColumnHeaders are used as user options.

If the user option ${list_id}ListAsTable is set to 'yes' the list is displayed
as a table. Tables can have headers of field names depending upon the
${list_id}ListColumnHeaders setting. (see Acme::Options)

=back

=head1 CHANGE LOG

 5/04/00 Initial version.
 9/09/00 Introduced list_id to use to identify user options on the list display.
 6/09/01 Converted to Acme::Base::Schema and Acme::SQL.
 6/12/01 Converted to Acme::Lister. Divided into list.cgi and pick.cgi,
         with this version specializing in list operations.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000,2001 Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

package list;

use Acme::HTTP			qw(escape);
use Acme::HTML			qw(escape_html display_message);
use Acme::Base::Schema;
use Acme::Base::Schema::Join	qw(parse_join);
use Acme::SQL;
use Acme::Searcher;
use Acme::Lister;
use Acme::Options;
use Acme::HTML::NavButtons;
use Local::Acme::Banners  	qw(common_header common_footer);
use Local::Acme::Format		qw(sprintf_cmd);
use Acme::Config;
use CGI;

use Acme::Security;
Acme::Security::check('Access');

$|=1;		# flush output (synchronize errors with output)

our $cgi = new CGI;
die "No arguments\n" unless $cgi->param();

our $Styles      = Acme::Config->Styles;	# location of CSS styles
our $Bullet      = Acme::Config->Bullet;	# edit bullet

#
# Program Arguments
#

our $request	  = $cgi->url_param('request')	   || $cgi->param('request') || '';

our $table	  = $cgi->url_param('table')	   || $cgi->param('table');

our $sql	  = $cgi->url_param('sql')	   || $cgi->param('sql');
our $search	  = $cgi->url_param('search')	   || $cgi->param('search');
our $sort	  = $cgi->url_param('sort')	   || $cgi->param('sort');

our $list	  = $cgi->url_param('list')	   || $cgi->param('list');
our $format 	  = $cgi->url_param('format')	   || $cgi->param('format');

our $title	  = $cgi->url_param('title')	   || $cgi->param('title')	|| '';
our $list_title   = $cgi->url_param('list_title')  || $cgi->param('list_title');

our $record_set_name  = $cgi->url_param('record_set_name') || $cgi->param('record_set_name');

our $id_field	  = $cgi->url_param('id_field')	   || $cgi->param('id_field')     ||
		    $cgi->url_param('id_fields')   || $cgi->param('id_fields');
our $code_field	  = $cgi->url_param('code_field')  || $cgi->param('code_field')   ||
		    $cgi->url_param('code_fields') || $cgi->param('code_fields');
our $name_field	  = $cgi->url_param('name_field')  || $cgi->param('name_field')   ||
		    $cgi->url_param('name_fields') || $cgi->param('name_fields');
our $phone_field  = $cgi->url_param('phone_field') || $cgi->param('phone_field')  ||
		    $cgi->url_param('phone_fields')|| $cgi->param('phone_fields');
our $email_field  = $cgi->url_param('email_field') || $cgi->param('email_field')  ||
		    $cgi->url_param('email_fields')|| $cgi->param('email_fields');
our $first_name_field = $cgi->url_param('first_name_field')  || $cgi->param('first_name_field');
our $last_name_field  = $cgi->url_param('last_name_field')   || $cgi->param('last_name_field');

our $code_field_match = $cgi->url_param('code_field_match')  || $cgi->param('code_field_match');
our $phone_field_match= $cgi->url_param('phone_field_match') || $cgi->param('phone_field_match');

our $limit_abort   = $cgi->url_param('limit_abort')  || $cgi->url_param('limit')    || '';
our $limit_warn    = $cgi->url_param('limit_warn')   || '';
our $limit_alert   = $cgi->url_param('limit_alert')  || '';

our $list_id	  = $cgi->url_param('list_id')	   || $cgi->param('list_id');

our $Printing     = $cgi->url_param('print')       || $cgi->param('print') || '';
$Printing         = $Printing =~ /yes/i            ||
		    $request =~ /print/i           ||
		    ($Printing !~ /^no$/i && exists $cgi->{'print'} ? 1 : 0);

$request ||= "List";

our $counting = $request =~ /Count/i;	# flag

print $cgi->header(-type=>'text/html') unless @ARGV;

#
# Get user option settings
#

our $opt = new Acme::Options;

#
# Construct SQL from table schema
#

our ($schema, $t);  		# schema/table properties pointer

if ($table) {
	
	Acme::Security::check('Access',$table);

	$schema = new Acme::Base::Schema;

	$t = $schema->t($table)
	  or die "Cannot find a database table named '$table'\n";

	#
	# Get list_id to use from table name or use default for all tables.
	#

	$list_id ||= $opt->get("${table}ListID");
	$list_id ||= "list";

	#
	# Get user options settings
	#

	if ($opt->get("${list_id}ListAllFields") =~ /yes/i
	&&  $opt->get("${list_id}ListAsTable")   =~ /yes/i) {
		$list = join ", ", map { "$table.$_" } $t->fields;
		if (my $id = $t->id) { $list = "$table.$id, $list" }
		$format = '';	# define as empty to force auto-columns below
	}

	#
	# Set defaults from schema properties
	#

	$list  ||= $t->list_fields;

	if (! defined $format) {
		$format = $t->list_format;
		$format =~ s/^\s*(["'])(.*)\1\s*$/$2/;	# remove possible quotes
	}

	$title             ||= $t->title;
	$list_title = $t->list_title unless ($list_title || $search);

	$record_set_name   ||= $t->title;

	$id_field	   ||= $t->id_field;
	$code_field	   ||= $t->code_field;
	$name_field	   ||= $t->name_field;
	$first_name_field  ||= $t->first_name_field;
	$last_name_field   ||= $t->last_name_field;
	$phone_field       ||= $t->phone_field;
	$email_field       ||= $t->email_field;

	$code_field_match  ||= $t->code_field_match;
	$phone_field_match ||= $t->phone_field_match;

	#
	# Get limits
	#

	my $limit = $t->limit_abort || Acme::Config->List_Limit_Abort;
	$limit_abort = $limit if ! $limit_abort || $limit < $limit_abort;

	$limit_warn ||= $t->limit_warn || Acme::Config->List_Limit_Warn;
	$limit_alert .= $t->limit_alert;

	#
	# Create default sql to search the table if needed.
	#

	$sql ||= $t->list_sql if $t->defined('list_sql');

	if (! $sql) {

		# Construct query here if not explicity defined in schema.
		# This allows customized list of fields such as the
		# ListAllFields user option above.

		my ($join_expr,$join_clause) = parse_join($table,$t->list_join);
		$sql =
		    'select '	 . $list
		  . ' from '	 . $join_expr
		  . ' where '	 . $t->list_where . $join_clause
		  . ' order by ' . $t->list_order_by_clause;

	}
}

die "Empty search\n" unless $sql;

#
# Parse SQL statement
#

our $sqlo = new Acme::SQL $sql;

my $added_column = 0;		# flag
if ($table) {

	#
	# Make sure key field is selected in SQL for use in links
	#

	my $col = $table . '.' . $t->key;
	if (! $sqlo->is_column($col)) {
		$sqlo->add_column($col);
		$added_column = 1;	# set to true
	}
}

#
# Set SQL limits
#

$sqlo->set_limit($limit_abort) if $limit_abort;

#
# Evaluate Search script
#

if ($search) {

	if (my $sh = $cgi->param('save_history')) {

		#
		# Save search in history
		#

		if ($sh =~ /yes/i && $table) {
			$opt->enter("${table}Search",$search);
		}

		$cgi->delete('save_history');
	}

	if ($id_field || $code_field || $name_field || $last_name_field || $phone_field || $email_field) {

		#
		# Preprocess search script
		#

		if ($first_name_field || $last_name_field) {
			$search = Acme::Searcher::preprocess($search, {
			  RecordSetName   => $record_set_name,
			  IDField	  => $id_field,
			  CodeField	  => $code_field,
			  NameField       => $name_field,
			  FirstNameField  => $first_name_field,
			  LastNameField   => $last_name_field,
			  PhoneField      => $phone_field,
			  EmailField      => $email_field,
			  CodeFieldMatch  => $code_field_match,
			  PhoneFieldMatch => $phone_field_match,
			});
		}
		else {
			$search = Acme::Searcher::preprocess($search, {
			  RecordSetName  => $record_set_name,
			  IDField	 => $id_field,
			  CodeField	 => $code_field,
			  NameField	 => $name_field,
			  PhoneField     => $phone_field,
			  EmailField     => $email_field,
			});
		}
	}

	#
	# Combine sql with search script.
	#

	my $srch = new Acme::Searcher $search, $cgi;	# parse search script
	$sql = $srch->as_sql($sqlo);	

	#
	# Redefine title if defined in script.
	#

	$list_title ||= "$request " . $srch->title if $srch->title;
}
else {
	$sql = $sqlo->as_sql() if $table;
}

#
# Set optional sort order
#

if ($sort) {
	# Parse sql again. It is more efficient for the user to put an
	# "order by" clause in the search script, but it's not convenient.
	# We could wait until the parsing of sql for column names below,
	# but that would be changing the sql after it has been "cleaned"
	# by security.

	$sqlo = new Acme::SQL $sql;
	$sqlo->sort($sort);
	$sql = $sqlo->as_sql();
}

#
# Set defaults if only a sql query was specified in the cgi.
#

if ($sql && ! $search && ! $table) {

	if (my $sh = $cgi->param('save_history')) {

		#
		# Save sql query in history
		#

		$list_id ||= 'sql';
		if ($sh =~ /yes/i) {
			$opt->enter("${list_id}Query",$sql);
		}

		$cgi->delete('save_history');
	}

	$list_title ||= $sql;

	#
	# Get limits
	#

	my $limit = Acme::Config->SQL_Limit_Abort || Acme::Config->List_Limit_Abort;
	$limit_abort = $limit if ! $limit_abort || $limit < $limit_abort;

	$limit_warn ||= Acme::Config->SQL_Limit_Warn;
}

#
# Initialize limits
#

# Note: No limits are applied if we are just counting records. We might want
# to still apply the $limit_abort limit, or instead of terminating the count,
# use it as a limit to issue a warning alert to the administrators.

my $limits;
if (($limit_abort || $limit_warn) && ! $counting) {
	$limits = {};
	$limits->{$limit_abort} = \&limit_abort_func	if $limit_abort;
	$limits->{$limit_warn}  = \&limit_warn_func	if $limit_warn;
}

#
# Set up list
#

$sql = Acme::Security::clean_sql($sql);

our $l = new Acme::Lister (
	sql          => $sql,
	before_list  => \&display_list_header,
	after_list   => \&display_list_footer,
	every_row    => \&display_list_item,
	limits       => $limits,
	sandbox      => 'list',
	);

if ($l->count == 0) {
	display_message("No records were found.", "$title",undef,0);
	exit 1;
}

#
# Set final default values
#

$sqlo = new Acme::SQL $sql;		# reparse after clean_sql() expands *
my @field_names = $sqlo->columns;
pop @field_names if $added_column;	# remove key field added to sql above

$title = "SQL List" if (! $title && $sql);
$title ||= "List";			# final default

#
# Define global variables for list helper functions
#

our $lsn = 'list';			# default CSS "list style" name
$lsn = 'greenbar' if $opt->get("${list_id}ListGreenbar") =~ /yes/i;

our $b = new Acme::HTML::NavButtons print => $Printing;
our $n_rec = 0;				# record number

our $list_as_table = ($request eq 'List' && $opt->get("${list_id}ListAsTable") =~ /yes/i);

#
# Display list
#

display_header();

$l->print();

display_footer();

exit 0;

#------------------------------------------------------------------------------
#
# Page header and footer functions
#
#------------------------------------------------------------------------------

sub display_header {

	my $aux_script = $t->list_include_html if $table;
	$aux_script ||= '';
	$aux_script &&= "\n$aux_script";

	my $aux_stylesheet = $t->list_stylesheet_html if $table;
	$aux_stylesheet ||= '';
	$aux_stylesheet  &&= "\n$aux_stylesheet";

	my $print_stylesheet = '';

	my $onload_cmd = qq[ onLoad="self.focus()"];

	if ($Printing) {
  		$onload_cmd = 
  qq[ onLoad="window.print();window.setTimeout('window.close()',1500)"];

		$print_stylesheet =
  qq[<link rel="stylesheet" href="$Styles/list_print.css" type="text/css">\n];

		my $aux_print_stylesheet = $t->list_print_stylesheet_html if $table;
		$aux_print_stylesheet  &&= "$aux_print_stylesheet\n";

		$print_stylesheet .= $aux_print_stylesheet;
  	}

# Keep IE6 in IE5 compatablilty mode by not specifying url
#<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
#"http://www.w3.org/TR/html4/loose.dtd">

        print <<EOF;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">

<html>
<head>
<title>list $title</title>

<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<link rel="stylesheet" href="$Styles/list.css" type="text/css">$aux_stylesheet
$print_stylesheet
<script language="JavaScript" type="text/javascript">
function select_table() {
	if (!document.all) return;
	if (!document.all.list_table) return;
	var r = document.body.createTextRange();
	r.moveToElementText(document.all.list_table);
	r.select();
}
</script>$aux_script
</head>
<body$onload_cmd>
EOF
	print $b->first_buttons;
	print common_header($title,$Printing);

	print <<EOF if $list_title;
<h3>
$list_title
</h3>
EOF

	print <<EOF;
<div class="list">
EOF
}

#------------------------------------------------------------------------------

sub display_footer {

	print "</div>\n";
	print common_footer($title,$Printing);
	print $b->last_buttons;
	print "</body>\n</html>";
}

#------------------------------------------------------------------------------
#
# List helper functions
#
#------------------------------------------------------------------------------

sub display_list_header {

	if ($list_as_table) {
		#
		# Redefine format for table list.
		#

		$format = join qq(</td><td class="${lsn}-item">), ("%s") x @field_names;
		print qq(<table class="${lsn}" id="list_table">\n);

		#
		# Display a button that selects entire contents of table
		# for cut & paste or drag & drop.
		#

		my $select_table_button = qq(<td></td>\n);	# empty

		if ($ENV{HTTP_USER_AGENT} =~ /MSIE/ && $ENV{HTTP_USER_AGENT} !~ /Opera/) {
			# This only works for Internet Explorer.

			# The following uses an image for the button
			#our $Images = Acme::Config->Images;	# location of image files
			#$select_table_button =
			#  qq(<td>)
			#. qq(<a href="javascript:select_table()")
			#. qq( title="Select and highlight entire list.">)
			#. qq(<img src="$Images/select_table.gif")
			#. qq( onMouseOver="hilite(this)")
			#. qq( onMouseOut="unhilite(this)")
			#. qq( border="0">)
			#. qq(</a>)
			#. qq(</td>\n);

			# The following produces an "invisible" button that
			# doesn't get copied when pasting the selection into
			# an Excel spreadsheet.
			#$select_table_button =
			#  qq(<td title="Select and highlight entire list.")
			#. qq( style="border:solid #CCCCCC 2px;height:8pt")
			#. qq( onMouseOver="this.style.borderColor='#FF00FF';)
			#. qq(this.style.cursor='hand'")
			#. qq( onMouseOut="this.style.borderColor='#CCCCCC'")
			#. qq( onClick="select_table()">)
			#. qq(</td>\n);

			# The following also produces an "invisible" button that
			# doesn't get copied when pasting the selection into
			# an Excel spreadsheet.
			$select_table_button =
			  qq(<td)
			. qq( style="height:8pt)
			. qq(;border-bottom:solid #CCCCCC 2px)
			. qq(;border-right:solid #CCCCCC 2px)
			. qq(;padding:2px)
			. qq(">)
			. qq(<div title="Select and highlight entire list.")
			. qq( style="border:solid #CCCCCC 2px)
			. qq(;width:10px;height:10px;font-size:3pt")
			. qq( onMouseOver="this.style.borderColor='#FF00FF')
			. qq(;this.style.cursor='hand'")
			. qq( onMouseOut="this.style.borderColor='#CCCCCC'")
			. qq( onClick="select_table()">)
			. qq(</div></td>\n);
		}
		else {
			# Empty button
			$select_table_button =
			  qq(<td)
			. qq( style="height:8pt)
			. qq(;border-bottom:solid #CCCCCC 2px)
			. qq(;border-right:solid #CCCCCC 2px)
			. qq(;padding:2px)
			. qq(">)
			. qq(</td>\n);
		}

		#
		# Display arguments as column headers.
		#

		print qq(<tr class="${lsn}-column-header">\n);
		print qq(<td></td>\n);
		print $select_table_button;
		print_column_headings();
		print qq(</tr>\n);
	}
	else {
		#
		# Display info as a list.
		#

		# Note: We display the list as a table irregardless of the
		# ListAsTable option setting. The main difference is that the
		# table is forced to a 100% fixed width table so that long
		# lists can be displayed as they are transmitted instead of
		# waiting for the browser to receive the entire list before
		# rendering the list.

		$format ||= "%s &nbsp; " x @field_names;
		print qq(<table class="${lsn}" style="table-layout:fixed;width:100%">\n);
		my $header = $t->list_header if $table;
		$header = $format if $header eq 'yes';
		my $list_header = make_list_header(
					$header,
					$list,
					$table,
					$schema,
					$t);

		if ($list_header) {
			# Surround $list_header with <td> </td> tags if not already.
			if ($list_header !~ m|^\s*<t[hd]|i
			||  $list_header =~ s|^\s*<t([hd])>(.*)</t\1>\s*$|$2|i) {
				$list_header = qq[<th>$list_header</th>\n];
			}

			#
			# Display arguments as column headers.
			#

			print qq(<tr class="${lsn}-column-header">\n);
			print qq(<td class="${lsn}-item-index"></td>);
			print qq(<td class="${lsn}-item-buttons"></td>);
			print $list_header;
			print qq(</tr>\n);
		}
	}
}

#------------------------------------------------------------------------------

sub display_list_footer {

	print "</table>\n";
	print "<p>".$l->count." records selected.</p>\n";
}

#------------------------------------------------------------------------------

sub display_list_item {

	use vars qw(@a);  	# array of values read from list query

	$n_rec++;

	#
	# Do not print every record if request is Count.
	#

	return if $counting;

	#
	# Print record index
	#

	my $ri = sprintf("%4s", $n_rec);
	$ri =~ s/ /&nbsp;/g;

	if ($n_rec % 2) {
		print qq(<tr class="${lsn}-row-odd">\n<td class="${lsn}-item-index">$ri</td>);
	}
	else {
		print qq(<tr class="${lsn}-row-even">\n<td class="${lsn}-item-index">$ri</td>);
	}

	print qq(<td class="${lsn}-item-buttons">);

	#
	# Print selection icon
	#

	print_rec_selector();

	#
	# Print row
	#

	local $^W;	# turn off warnings of undefined variables in @a

	print qq(</td>\n);
	print qq(<td class="${lsn}-item">)	if $format !~ /^<[tT][dD]/;
	print sprintf_cmd($format, @a);
	print qq(</td>) 			if $format !~ /^<[tT][dD]/;
	print qq(\n</tr>\n);
}

#------------------------------------------------------------------------------
#
# Support functions
#
#------------------------------------------------------------------------------

sub print_column_headings {
	
	# Construct <td> tags with title attributes of field descriptions.
	my @h = column_headings();
	my ($f,$desc);
	for my $fld (@field_names) {
		$fld =~ s/.*\.//;	# remove table qualifier
		$f = $t->f($fld);
		if ($f) { $desc = $f->description; }
		else    { $desc = $t->id_description; }

		$desc = escape_html($desc);

		print qq(<th title="$desc" nowrap>);
		print shift @h;
		print qq(</th>\n);
	}
}

sub column_headings {

	my $colhead = $opt->get("${list_id}ListColumnHeaders");
	my @h = @field_names;

	COLUMN_HEADER: {
		if (! $colhead || $colhead =~ /none/i) {
			# Display header line with blank headers.
			# (Other controls are in header line).

			@h = ('') x @h;		# initialize all to empty string
			last COLUMN_HEADER;
		}

		if ($colhead =~ /FieldTitles/i) {
			if ($table) {
				# Get titles from schema
				my $f;
				for (@h) {
					s/.*\.//;	# remove table qualifier
					$f = $t->f($_);
					if ($f) { $_ = $f->title; }
					else    { $_ = $t->id_title; }
				}
			}
			else {
				# Make titles out of field names
				for (@h) {
					s/.*\.//; # remove table name
					s/_/ /g;  # change underscores to spaces
					s/\b(\w+)\b/ucfirst(lc($1))/eg; # capitalize words
					s/\bId\b/ID/; # capitalize id
				}
			}
			last COLUMN_HEADER;
		}

		if ($colhead =~ /QualifiedFieldNames/i) {
			if ($table) {
				for (@h) { $_ = "$table.$_" unless /\./; }
			}
			last COLUMN_HEADER;
		}

		if ($colhead =~ /FieldNames/i) {
			for (@h) { s/.*\.//; }	# strip table name
			last COLUMN_HEADER;
		}
	}

	return @h;
}

sub make_list_header {
	my ($format,$fields,$table,$schema,$t) = @_;

	# Helper function for building a list header from field names.

	$format ||= '';
	return $format if $format !~ /\%/;	# not a sprintf() format

	# Merge field titles into format string.

	my ($tab,$fld,$tt,@headers);
	for my $h (split(/\s*,\s*/,$fields)) {
		($tab,$fld) = $h =~ /(\w+)\.(\w+)/ 
		  or ($tab,$fld) = ($table,$h);

		if ($tab eq $table) { $tt = $t; }
		else                { $tt = $schema->t($tab); }

		if ($tt && $tt->{$fld}) {
			push @headers, $tt->{$fld}->title;
		}
		else {
			# Capitalize field name
			# (happens on ID fields)
			$fld =~ s/_/ /g;
			$fld =~ s/\b(\w+)\b/ucfirst(lc($1))/eg;
			push @headers, $fld;
		}
	}

	# Format header and evaluate embedded functions
	return sprintf_cmd($format,@headers);
}

#------------------------------------------------------------------------------

sub print_rec_selector {

	return unless $table;

	no strict 'refs';

	my $request = 'edit';
	my $alt_title = ucfirst(lc($request)) . ' ' . $t->row_name . '.';
	my $edit_prog = $t->edit_prog;

	# The following links work well for lists in the browser, but
	# when copied to other programs, they break because of the 
	# javascript and relative url.
	#my $href = qq[javascript:open_form(]
	#  . qq['$edit_prog]
	#  . ($edit_prog =~ /[?]/ ? '&' : '?')
	#  . qq[request=$request&table=$table&]
	#  .  escape_html($t->key)
	#  . '='
	#  . escape_html(${$table.'__'.$t->key})
	#  . qq[')];
	#
	#print qq[<a href="$href" title="$alt_title">$Bullet</a>];

	# The following links work when the list is copied to Excel.
	# They embed the full url of the web site.
	my $base = Acme::Config->Base;

	my $wname = ($table || 'list') . $$ . $n_rec;

	my $href = $base
	  . ($edit_prog =~ m|^/| ? '' : '/')
	  . $edit_prog
	  . ($edit_prog =~ /[?]/ ? '&' : '?')
	  . qq[request=$request&table=$table&]
	  .  escape_html($t->key)
	  . '='
	  . escape_html(${$table.'__'.$t->key});

	print
	    qq[<a href="$href"]
	  . qq[ title="$alt_title"]
	  . qq[ target="$wname"]
	  . qq[ onClick="window.open_form&&open_form('','$wname')"]
	  . qq[>$Bullet</a>];
}

#------------------------------------------------------------------------------
#
# List limit functions
#
#------------------------------------------------------------------------------

# Note: These are copies of the functions in Acme::Lister::Limit module.
# They allow the schema files to specify additional email addresses to 
# send the email alerts to.

sub limit_abort_func {
	my $limit = shift;

	my $to = $t->limit_alert if $t;

	my $subject = qq[Aborting: Limit of $limit for $list_id list has been exceeded.];

	my $msg = <<End_Of_Msg;
The $list_id list has exceeded its limit of $limit records.
The list has been terminated.
End_Of_Msg

	Acme::Security::alert($msg, ($to || ''), $subject);

	$msg .= <<End_Of_Msg;
These limits are usually set by administrators in the site configuration file or in database schema files.
End_Of_Msg

	die $msg;
}

#------------------------------------------------------------------------------

sub limit_warn_func {
	my $limit = shift;

	my $to = $t->limit_alert if $t;

	my $subject = qq[Warning: Limit of $limit for $list_id list has been reached.];

	my $msg = <<End_Of_Msg;
The $list_id list has reached a limit of $limit records and is generating this notice.
The list is continuing to be rendered. This is only a warning.
End_Of_Msg

	Acme::Security::alert($msg, ($to || ''), $subject);
}

__END__
