#!/usr/bin/perl -wT -I../../lib

=head1 NAME

suppliers/view.cgi -- View Supplier Information

=head1 SYNOPSIS

    /cgi-bin/suppliers/view.cgi?search=Folly

=head1 DESCRIPTION

This routine displays all the information about a supplier - their address,
phones, pos, receivings, and log history.

If the supplier_id is not specified, the "search" argument is looked for.

The search argument is checked to see if it is a number in which case,
it assigns it to the supplier_id argument and displays the supplier
information like normal.

=head1 CHANGE LOG

 4/5/99   Initial version.
 6/1/00   Broke out section headers into Acme::HTML::SectionControl library.
 8/13/02  Converted to CGI.
 9/02/02  Changed button bar menu, save settings, and optional display of sections.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000-2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

# Note: this is a copy of the customers/view.cgi program with following substitutions.
#
#	s/customers/supplier/g
#	s/Customer/Supplier/g
#	s/ship_/pay_/g
#	s/order/po/g	# manually check each substitution
#	s/invoice/receiving/g
#

package supplier;

use Acme::HTTP			qw(escape);
use Acme::HTML			qw(escape_html		display_message);
use Acme::HTML::HideMoreLess	qw(start_hiding		stop_hiding);
use Acme::HTML::SectionControl;
use Acme::HTML::NavButtons;
use Acme::HTML::Parameters;
use Acme::Db;
use Acme::SQL;
use Acme::Searcher;
use Acme::Base::Schema;
use Acme::Options;
use Acme::Address;
use Acme::TextArray		qw(split_array);
use Acme::Money			qw(dollars		commafy);
use Acme::Date			qw(meddate		timetodate);
use Local::Acme::Banners	qw(common_header	common_footer);
use Local::Acme::User		qw(user_folder);
use Acme::Config;
use Time::Local;
use CGI;

use Acme::Security;
Acme::Security::check('Access');

require "./view/pos.pl";		# for displaying pos info
require "./view/receivings.pl";		# for displaying invoice info

our $Email_Send_Domain = Acme::Config->Email_Send_Domain;
our $Images            = Acme::Config->Images;		# image files
our $Styles            = Acme::Config->Styles;		# css stylesheets
our $Scripts_Lib       = Acme::Config->Scripts_Lib;	# JavaScript library
our $Audit_Prog        = Acme::Config->Audit_Prog;	# display audit info

our $MaxLogDisplay     = 5;	# maximum log messages displayed before hiding

$\="\n";	# append newline to all print statements
$|=1;		# flush output (synchronize errors with output)

our $cgi = new CGI;
die "No arguments\n" unless $cgi->param();

#
# Program Arguments
#

our $request = $cgi->url_param('request')	|| $cgi->param('request') || '';

our $table = 'suppliers';	# hardwire
our $sql   = '';		# defined below

our $search           = $cgi->url_param('search')	|| $cgi->param('search');
our $title	      = $cgi->url_param('title')	|| $cgi->param('title');

our $id_field	      = $cgi->url_param('id_field')	|| $cgi->param('id_field');
our $code_field	      = $cgi->url_param('code_field')	|| $cgi->param('code_field');
our $name_field	      = $cgi->url_param('name_field')	|| $cgi->param('name_field');
our $first_name_field = $cgi->url_param('first_name_field')  || $cgi->param('first_name_field');
our $last_name_field  = $cgi->url_param('last_name_field')   || $cgi->param('last_name_field');
our $phone_field      = $cgi->url_param('phone_field')	|| $cgi->param('phone_field');
our $email_field      = $cgi->url_param('email_field')	|| $cgi->param('email_field');

our $code_field_match = $cgi->url_param('code_field_match')  || $cgi->param('code_field_match');
our $phone_field_match= $cgi->url_param('phone_field_match') || $cgi->param('phone_field_match');

our $Printing = $cgi->url_param('print')	|| $cgi->param('print')	|| '';
$Printing = $Printing =~ /yes/i		||
	    $request =~ /print/i	||
	    ($Printing !~ /^no$/i && exists $cgi->{'print'} ? 1 : 0);

#
# Get the view ID to use for view settings and options.
# The default value is usually the table's name.
#

our $view_id	   = $cgi->param('view_id')	|| $table || 'view';

#
# Get user options
#

my $view_settings_file = $cgi->param('view_settings')     ||
			 user_folder() . "/${view_id}ViewSettings";
our $view_settings = new Acme::Options file => $view_settings_file;

our $opt = new Acme::Options;

#
# Set view options
#

# View settings come from 1) CGI, 2) per view settings, or 3) user option settings.

# First initialize flags which tell us if this is the first time the view
# is rendered. The "repeat" parameter is not intended to be one of the
# view options, but is entered in the view_control form with a constant "yes"
# telling us whe the user has selected the "Redraw" option. This helps us
# know when to reset the options back to "nothing" upon user request.

# Not used at this time
our $redrawing           = $cgi->param('redrawing') || '';
$redrawing               = '' unless 'yes' =~ /\Q$redrawing\E/i;

# view_comments
our $view_comments       = $cgi->param('view_comments')		|| '';
$view_comments         ||= $view_settings->get("ViewComments") if $view_settings;
$view_comments         ||= $opt->get("${view_id}ViewComments");
$view_comments         ||= 'yes';		# Default setting
$view_comments           = '' unless 'yes' =~ /\Q$view_comments\E/i;

# view_pos
our $view_pos         = $cgi->param('view_pos')		|| '';
$view_pos           ||= $view_settings->get("ViewPOs") if $view_settings;
$view_pos           ||= $opt->get("${view_id}ViewPOs");
$view_pos           ||= 'yes';		# Default setting
$view_pos             = '' unless 'yes' =~ /\Q$view_pos\E/i;

# view_receivings
our $view_receivings       = $cgi->param('view_receivings')		|| '';
$view_receivings         ||= $view_settings->get("ViewReceivings") if $view_settings;
$view_receivings         ||= $opt->get("${view_id}ViewReceivings");
$view_receivings         ||= 'yes';		# Default setting
$view_receivings           = '' unless 'yes' =~ /\Q$view_receivings\E/i;

# view_past_receivings
our $view_past_receivings  = $cgi->param('view_past_receivings')	|| '';
$view_past_receivings    ||= $view_settings->get("ViewPastReceivings") if $view_settings;
$view_past_receivings    ||= $opt->get("${view_id}ViewPastReceivings");
$view_past_receivings    ||= 'no';		# Default setting
$view_past_receivings      = '' unless 'yes' =~ /\Q$view_past_receivings\E/i;

#
# Save view settings
#

if ($view_id ne 'view') {
	# Only save the non-standard (non-"view") user settings.
	# Standard settings come from the main user option settings.

	$view_settings->set('ViewComments',$view_comments || 'no');
	$view_settings->set('ViewPOs',$view_pos || 'no');
	$view_settings->set('ViewReceivings',$view_receivings || 'no');
	$view_settings->set('ViewPastReceivings',$view_past_receivings || 'no');

	$view_settings->save;
}

#
# Define global variables
#

our ($schema, $t);	# schema table and field pointers
our $b;			# navigation bar button object
our @fields;		# database field names, as in sql select ... from

#
# Set defaults
#

if ($table) {

	Acme::Security::check('Access',$table);

	$schema = new Acme::Base::Schema;

	$t = $schema->t($table)
	  or die "Cannot find a database table named '$table'\n";

	$title             ||= $t->title;

	$record_set_name   ||= $t->title;

	$id_field	   ||= "$table.".$t->id_field;
	$code_field	   ||= "$table.".$t->code_field;
	$name_field	   ||= "$table.".$t->name_field;
	$first_name_field  ||= "$table.".$t->first_name_field;
	$last_name_field   ||= "$table.".$t->last_name_field;
	$phone_field       ||= "$table.".$t->phone_field;
	$email_field       ||= "$table.".$t->email_field;

	$code_field_match  ||= $t->code_field_match;
	$phone_field_match ||= $t->phone_field_match;


	if (! $sql) {
		# Create default sql to search the table.
		# Assure key field is part of sql.

		$sql = <<End_of_Sql;
		select
			suppliers.*,
			supplier_info.in_office_comments,
			salesreps.salesrep_name,
			salesreps.terminate_date,
                        supplier_types.description as supplier_type_desc,
                        cust_status_codes.description as status_description,
                        address_usage_codes.usage,
                        address_usage_codes.description as address_usage_desc,
                        address_usage_codes.notes as address_usage_notes,
                        pay_usage_codes.usage as pay_usage,
                        pay_usage_codes.description as pay_usage_desc,
                        pay_usage_codes.notes as pay_usage_notes,
                        countries.address_style,
                        pay_countries.address_style as pay_address_style,
                        ad_sources.ad_source_name as ad_source_name,
                        last_ad_sources.ad_source_name as original_ad_source_name
                from ((((((((((suppliers
		       left join supplier_info
                       on suppliers.supplier_id = supplier_info.supplier_id)
		       left join address_usage_codes
                       on suppliers.address_usage = address_usage_codes.address_usage)
		       left join address_usage_codes as pay_usage_codes
                       on suppliers.pay_address_usage = pay_usage_codes.address_usage)
                       left join cust_status_codes
                       on suppliers.supplier_status = cust_status_codes.supplier_status)
                       left join salesreps
                       on suppliers.salesrep = salesreps.salesrep)
		       left join supplier_types
                       on suppliers.supplier_type = supplier_types.supplier_type)
                       left join countries
                       on suppliers.country = countries.country)
		       left join countries as pay_countries
                       on suppliers.pay_country = pay_countries.country)
                       left join ad_sources
                       on suppliers.ad_source = ad_sources.ad_source)
		       left join ad_sources as last_ad_sources
		       on suppliers.original_ad_source = last_ad_sources.ad_source)
                where 1 = 1
End_of_Sql

	}

	if (! $search) {
		# Create default search parameter from key argument.
		my $key = $t->key;
		$search = $cgi->url_param($key)	|| $cgi->param($key);
	}
}

die "Empty search\n" unless $sql;

#
# Parse SQL script
#

my $sqlo = new Acme::SQL($sql);

#
# Evaluate Search script
#

if ($search) {
	if ((my $sh = $cgi->param('save_history')) && ! $redrawing) {

		#
		# Save search in history list in user options file.
		#

		if ($sh =~ /yes/i and $table) {
			$opt->enter("${table}Search",$search);
		}

		$cgi->delete('save_history');
	}

	if ($id_field || $code_field || $name_field || $last_name_field || $phone_field || $email_field) {

		#
		# Preprocess search script
		#

		if ($first_name_field || $last_name_field) {
			$search = Acme::Searcher::preprocess($search, {
			  RecordSetName   => $record_set_name,
			  IDField	  => $id_field,
			  CodeField	  => $code_field,
			  NameField       => $name_field,
			  FirstNameField  => $first_name_field,
			  LastNameField   => $last_name_field,
			  PhoneField      => $phone_field,
			  EmailField      => $email_field,
			  CodeFieldMatch  => $code_field_match,
			  PhoneFieldMatch => $phone_field_match,
			});
		}
		else {
			$search = Acme::Searcher::preprocess($search, {
			  RecordSetName   => $record_set_name,
			  IDField	  => $id_field,
			  CodeField	  => $code_field,
			  NameField	  => $name_field,
			  PhoneField      => $phone_field,
			  EmailField      => $email_field,
			  CodeFieldMatch  => $code_field_match,
			  PhoneFieldMatch => $phone_field_match,
			});
		}
	}

	my $srch = new Acme::Searcher $search, $cgi;

	#
	# Combine search script with SQL.
	#

	$sql = $srch->as_sql($sqlo);	# combine sql with search script
}

#
# Check query with security
#

$sql = Acme::Security::clean_sql($sql);

#
# Execute query
#

my $q = Acme::Db->query($sql);

$q->bind_columns_to_variables();	# reference fields by variables

print $cgi->header(-type=>'text/html') unless @ARGV;

if ($q->count == 0) {
	display_message("No records were found.", "$title",undef,0);
	exit 1;
}

#
# Set final default values
#

@fields = $sqlo->columns;

$title = "SQL View" if (! $title && $sql);
$title ||= "View";			# final default

#
# Display view
#

$b = new Acme::HTML::NavButtons print => $Printing;

display_page_header("view $suppliers__supplier_name ($suppliers__supplier_id)");

while ($q->next()) {
	display_supplier();
}

display_page_footer();

exit 0;

#------------------------------------------------------------------------------

sub display_page_header {
	$title = shift;

	my $auto_refresh = '';
	if ($opt->AutoRefresh =~ /yes/) {
		$auto_refresh = <<EOF;
function auto_refresh() { self.location.reload(true) }
EOF
	}

	my $print_cmd = $Printing
		      ? qq[ onLoad="window.print();window.setTimeout('window.close()',1500)"]
		      : '';

	print <<EOF;
<html>
<head>
<title>$title</title>

<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<link rel="stylesheet" href="$Styles/button_bar.css" type="text/css">
<link rel="stylesheet" href="$Styles/view.css" type="text/css">

<script src="$Scripts_Lib/menu.js" type="text/javascript"></script>
<script language="JavaScript" type="text/javascript">
$auto_refresh
	// Cache control images
	var control_image        = new Image;
	var control_image_hilite = new Image;
	control_image.src        = "$Images/control.gif"
	control_image_hilite.src = "$Images/control_hilite.gif"

	var control_open_image        = new Image;
	var control_open_image_hilite = new Image;
	control_open_image.src        = "$Images/control_open.gif"
	control_open_image_hilite.src = "$Images/control_open_hilite.gif"

	var small_control_image        = new Image;
	var small_control_image_hilite = new Image;
	small_control_image.src        = "$Images/small_control.gif"
	small_control_image_hilite.src = "$Images/small_control_hilite.gif"

	var small_control_open_image        = new Image;
	var small_control_open_image_hilite = new Image;
	small_control_open_image.src        = "$Images/small_control_open.gif"
	small_control_open_image_hilite.src = "$Images/small_control_open_hilite.gif"

function show_all_sections() {
	if (window.show_pos_control)           show_pos_control();
	if (window.show_receivings_control)         show_receivings_control();
	if (window.show_receivings_history_control) show_receivings_history_control();
}

function hide_all_sections() {
	if (window.hide_pos_control)           hide_pos_control();
	if (window.hide_receivings_control)         hide_receivings_control();
	if (window.hide_receivings_history_control) hide_receivings_history_control();
}

function toggle(field) {
	if (! field) return;
	if (field.value == "yes") {
		field.value = "no";
	}
	else {
		field.value = "yes";
	}
}

function submit_param() {
	var f = document.view_control;
	// Nothing to prepare at this time.
	f.submit();
}

</script>

</head>

<body$print_cmd>
EOF
}

#------------------------------------------------------------------------------

sub display_page_footer {

	print "</body></html>";
}

#------------------------------------------------------------------------------

sub display_supplier_header {

	print $b->buttons();

	print common_header("Supplier Information", $Printing);

	print <<EOF;
<div class=header>
<table cellspacing=0 cellpadding=0 width=99% class="supplier-header">
<tr>
 <td align=left valign=baseline>
 <font class=header-id>ID #$suppliers__supplier_id</font> &nbsp; &nbsp;
 </td>

 <td align=left valign=baseline>
 <font class=header-title>$suppliers__supplier_name</font>
 </td>

 <td align=right valign=baseline>
 <font class=header-desc>@{[ucfirst(lc($suppliers__supplier_status))]} @{[escape_html($supplier_type_desc)]}</font>
 </td>
</tr>
</table>
</div>
EOF

}

#------------------------------------------------------------------------------

sub display_supplier_footer {

	print common_footer("Supplier Information", $Printing);

	print $b->last_buttons();
}

#------------------------------------------------------------------------------

sub display_supplier {

	display_supplier_header();

	display_supplier_buttons()
	  if ! $Printing;

	display_supplier_info();

	display_supplier_comments($suppliers__supplier_id)
	  if $view_comments && ! $Printing;

	display_pos($suppliers__supplier_id)
	  if $view_pos && ! $Printing;

	display_receivings($suppliers__supplier_id)
	  if $view_receivings && ! $Printing;

	display_supplier_footer();
}

#------------------------------------------------------------------------------

sub display_supplier_buttons {

	local $\="";	# don't append newline to all print statements

	#
	# Build form of view parameters.
	#

	print <<End_Of_HTML;
<form	name=view_control
 	method=$ENV{REQUEST_METHOD}
	style="margin:0px"
>
End_Of_HTML

	#
	# Pass view arguments on to subordinate programs in hidden fields.
	#

	print Acme::HTML::Parameters::save_cgi_param_except(
							'request',
							'redrawing',
							'save_history',
							'view_comments',
							'view_pos',
							'view_receivings',
							'view_past_receivings',
							);

	#
	# Save view options in hidden fields
	#

	print <<End_Of_HTML;
<!-- List control settings -->
<input type="hidden" name="request" value="View">
<input type="hidden" name="redrawing" value="yes">

<input type="hidden" name="view_comments" value="$view_comments">
<input type="hidden" name="view_pos" value="$view_pos">
<input type="hidden" name="view_receivings" value="$view_receivings">
<input type="hidden" name="view_past_receivings" value="$view_past_receivings">
End_Of_HTML

	#
	# Construct the url to print this view.
	#

	my $print_prog = $t->print_prog;

	if (! $print_prog) {
		$print_prog = $cgi->self_url();
		unless ($print_prog =~ s/\b(request=)[^?&;]*/$1print/i) {
			$sep = $print_prog =~ /\?/ ? '&' : '?';
			$print_prog .= "${sep}print=yes";
		}
	}

	#
	# Substitute keyval for search parameter in print url
	#

	my $keyval = $suppliers__supplier_id;
	my $key = 'suppliers.supplier_id';
	unless ($print_prog =~ s/\bsearch=[^?&;]*/$key=$keyval/i) {
		# Add key value to url
		$sep = $print_prog =~ /\?/ ? '&' : '?';
		$print_prog .= "${sep}$key=$keyval";
	}

	#
	# Begin button bar display
	#

	print <<End_Of_HTML;

<!-- Start Button Bar -->
<div class="button-bar">
<table cellspacing="0" cellpadding="0">
<tr>
End_Of_HTML

	print <<End_Of_HTML;

<!-- Start Button Bar -->
<div class="button-bar">
<table cellspacing="0" cellpadding="0">
<tr>

 <td class="button-bar-menu">
 <a id="bb_menu_head"
  class="bb-menu-header"
  href="javascript:menu('bb_menu','bb_menu_head')"
 >&bull;</a>
 </td>

End_Of_HTML

	#
	# Check each button with security to see if it should be displayed.
	#

	my $edit_prog = "/cgi-bin/form.cgi?table=suppliers&supplier_id=$suppliers__supplier_id&request=Edit"; 

	if (Acme::Security::allow('Update','suppliers')) {
		print <<End_Of_HTML;
 <td><a href="$edit_prog" onClick="new_target(this,'form')" title="Edit the supplier information.">Change</a></td>
End_Of_HTML
	}

	my $log_prog = "/cgi-bin/form.cgi?table=supplier_log&supplier_id=$suppliers__supplier_id&request=New";
	if (Acme::Security::allow('Insert','supplier_log')) {
		print <<End_Of_HTML;
 <td><a href="$log_prog" onClick="new_target(this,'form')" title="Enter a comment.">Comment</a></td>
End_Of_HTML
	}

	my $po_prog = "/cgi-bin/form.cgi?table=purchase_orders&supplier_id=$suppliers__supplier_id&request=New";

	if (Acme::Security::allow('Insert','purchase_orders')) {
		print <<End_Of_HTML;
 <td><a href="$po_prog" onClick="new_target(this,'form')" title="Enter a purchase order.">New Purchase Order</a></td>
End_Of_HTML
	}

	my $invoice_prog = "/cgi-bin/form.cgi?table=receivings&supplier_id=$suppliers__supplier_id&request=New";

	if (Acme::Security::allow('Insert','receivings')) {
		print <<End_Of_HTML;
 <td><a href="$invoice_prog" onClick="new_target(this,'form')" title="Enter an invoice.">New Receiving</a></td>
End_Of_HTML
	}

	print <<End_Of_HTML;
</tr>
</table>
</div>
End_Of_HTML

	#
	# Construct button bar menu
	#

	print <<End_Of_HTML;

<!-- Button Bar Menu -->
<div id="bb_menu" class="bb-menu">
<a
 href="javascript:submenu('bb_param_sections')"
 class="bb-param-header"
>Sections</a>
<hr>
<a href="$print_prog" onClick="new_target(this,'print')">Print</a></td>
<a href="javascript:self.close()">Exit</a>
</div>
<!-- End Button Bar Menu -->
End_Of_HTML

	#
	# Output the hidden submenus and parameter menus.
	#

	my $view_comments_checked	= $view_comments	? ' checked' : '';
	my $view_pos_checked		= $view_pos		? ' checked' : '';
	my $view_receivings_checked	= $view_receivings	? ' checked' : '';
	my $view_past_receivings_checked	= $view_past_receivings	? ' checked' : '';

	print <<End_Of_HTML;
<div id="bb_param_sections" class="bb-param">

<h3>
Choose Display Sections
</h3>

<table>
<tr>
<th width="90%">&nbsp;</th>
</tr>

<tr>
<td><nobr><input type="checkbox" onClick="toggle(this.form.view_comments)"$view_comments_checked> Comments</nobr></td>
</tr>

<tr>
<td><nobr><input type="checkbox" onClick="toggle(this.form.view_pos)"$view_pos_checked> POs</nobr></td>
</tr>

<tr>
<td><nobr><input type="checkbox" onClick="toggle(this.form.view_receivings)"$view_receivings_checked> Receivings</nobr></td>
</tr>

<tr>
<td><nobr><input type="checkbox" onClick="toggle(this.form.view_past_receivings)"$view_past_receivings_checked> Past Receivings</nobr></td>
</tr>

<tr>
<td align="center">
<input type="button" value="Redraw" class="submit-button" onClick="submit_param()">
</td>
</tr>
</table>
</div>

</form>
<script language="JavaScript">var bb_menu = new Menu('bb_menu','bb_menu_head');</script>
<!-- End Button Bar -->
End_Of_HTML
}

#------------------------------------------------------------------------------

sub display_supplier_info {

	local $\="\n";	# append line feed to all print statements

	no warnings qw(uninitialized once);

	our $addr = new Acme::Address 
				$suppliers__name,
				$suppliers__company,
				$suppliers__address1,
				$suppliers__address2,
				$suppliers__city,
				$suppliers__state,
				$suppliers__zip,
				$suppliers__province,
				$suppliers__postal_code,
				$suppliers__country,
				$suppliers__address_locale,
				$countries__address_style,
				'',	# phones
				'',	# email
				$suppliers__address_usage,
				$address_usage_codes__usage;

	our $pay_addr = new Acme::Address 
				$suppliers__pay_name,
				$suppliers__pay_company,
				$suppliers__pay_address1,
				$suppliers__pay_address2,
				$suppliers__pay_city,
				$suppliers__pay_state,
				$suppliers__pay_zip,
				$suppliers__pay_province,
				$suppliers__pay_postal_code,
				$suppliers__pay_country,
				$suppliers__pay_address_locale,
				$pay_countries__address_style,
				'',	# phones
				'',	# email
				$suppliers__pay_address_usage,
				$pay_usage;

	#
	# Supplier Information
	#

	print qq[<table width="100%" cellspacing="2" cellpadding="0" style="margin-top:2pt; margin-bottom:6pt">];

	print qq[<tr>];
	print qq[<td width="10%">&nbsp;</td>];

	if (! $addr->is_empty) {
		my $map_link = $Printing ? '' : $addr->map_link('Map');

		print <<EOF;
<td width=40% align=left class="address-top">
<font class=field-name>Billing Address</font>&nbsp&nbsp;
$map_link
</td>
EOF
	}
	else {
		# Header without map link
		print <<EOF;
<td width=38% align=left class="address-top">
<font class=field-name>Billing Address</font>
</td>
EOF
	}

	print qq[<td width="10%"></td>];

	if (! $pay_addr->is_empty) {
		my $map_link = $Printing ? '' : $pay_addr->map_link('Map');

		print <<EOF;
<td width=38% align=left class="address-top">
<font class=field-name>Shipping Address</font>&nbsp&nbsp;
$map_link
</td>
EOF
	}
	else {
		# No header
		print <<EOF;
<td width=40% align=left>&nbsp;</td>
EOF
	}

	print qq[<td width="4%">&nbsp;</td>];
	print qq[</tr>];

	print qq[<tr>];
	print qq[<td></td>];
	print qq[<td>];
	print $addr->as_html;
	print qq[</td>];

        print qq[<td></td>];

        print qq[<td>];
	print $pay_addr->as_html if ! $pay_addr->is_empty;
	print qq[</td>];
	print qq[<td></td>];
	print qq[</tr>];

	print qq[<tr>];
	print qq[<td></td>];
	print qq[<td align=left valign=baseline class="address-bottom">&nbsp;</td>];
	print qq[<td></td>];
	if (! $pay_addr->is_empty) {
		print qq[<td align=left valign=baseline class="address-bottom">&nbsp;</td>];
	}
	else {
		print qq[<td>&nbsp;</td>];
	}
	print qq[<td></td>];
	print qq[</tr>];

        $original_ad_source_name &&= escape_html($original_ad_source_name);
        $ad_source_name &&= escape_html($ad_source_name);

        print qq[<tr>];
        print qq[<th align=right valign=baseline>Phone</th>];
        print qq[<td align=left valign=baseline>$suppliers__phone</td>];
        print qq[<th align=right valign=baseline>Source</th>];
        print qq[<td align=left valign=baseline>],
          ($suppliers__original_ad_source && $suppliers__original_ad_source ne $suppliers__ad_source)
          ? qq[<span title="$ad_source_name">$suppliers__ad_source</span>]
	    . ($suppliers__ad_source && $suppliers__ad_source_date ? ", " : "")
	    . qq[<span title="The date when the ad source code was last changed.">]
	    . qq[$suppliers__ad_source_date</span>]
	    . qq[ <nobr><font class=field-name>Original</font>]
	    . qq[ <span title="$original_ad_source_name">$suppliers__original_ad_source</span>]
	    . qq[</nobr>]
          : qq[<span title="$ad_source_name">$suppliers__ad_source</span>]
	    . ($suppliers__ad_source && $suppliers__ad_source_date ? ", " : "")
	    . qq[<span title="The date when the ad source code was last changed.">]
	    . qq[$suppliers__ad_source_date</span>],
          $suppliers__referral ? ", " : "",
          " ",
          ($suppliers__referral_id)
	  ? qq[<a href="/cgi-bin/customers/view.cgi?supplier_id=$suppliers__referral_id"]
	    . qq[ onClick="new_target(this,'view')"]
	    . qq[ title="View supplier record.">$suppliers__referral</a>]
	  : $suppliers__referral,
          qq[</td>];
        print qq[</tr>];

	print qq[<tr>];
	print qq[<th align=right valign=baseline>],
	  ($suppliers__company_phone ? "Business" : ''),
	  qq[</th>];
	print qq[<td align=left valign=baseline>$suppliers__company_phone</td>];
	print qq[<th align=right valign=baseline>Type</th>];
        print qq[<td align=left valign=baseline><span title="], escape_html($supplier_type_desc), qq[">$suppliers__supplier_type</span>];
        if ($suppliers__supplier_status) {
                print qq[ &nbsp;<font class=field-name>Status</font> ],
                  qq[<span title="], escape_html($status_description), qq[">$suppliers__supplier_status</span>],
                  ($suppliers__supplier_status_date ? ", " : ""),
                  qq[<span title="The date when the status code was last changed.">$suppliers__supplier_status_date</span>];
        }
        print qq[</td></tr>];

	print qq[<tr>];
	print qq[<th align=right valign=baseline>], $suppliers__fax ? "Fax" : "", qq[</th>];
	print qq[<td align=left valign=baseline>$suppliers__fax</td>];

        if ($suppliers__salesrep) {
                if (! $suppliers__terminate_date) {
                        # Make email link to current salesrep
                        my $s = "Supplier: $suppliers__supplier_name ($suppliers__supplier_id)";

                        print qq[<th align=right valign=baseline>Salesrep</th>];
                        print qq[<td align=left>];
                        print qq[<a href="mailto:$salesrep_name<$suppliers__salesrep\@$Email_Send_Domain>?subject=$s" title="Email $salesrep_name">$suppliers__salesrep</a
>];
                        print qq[</td>];
                }
                else {
                        # Salesrep is not with company, don't make email link
                        print qq[<th align=right valign=baseline>Salesrep</th>];
                        print qq[<td align=left>];
                        print qq[<span title="$salesrep_name">$suppliers__salesrep</spa
n>];
                        print qq[</td>];
                }
        }
        else {
                print qq[<th align=right valign=baseline></th>];
                print qq[<td align=left>];
                print qq[</td>];
        }
        print qq[</tr>];

	if ($suppliers__email) {
		print qq[<tr>];
		print qq[<th align=right>Email</th>];
		if ($suppliers__email !~ /</) {
			# Compose email address with supplier name
			my $ename = escape_html($suppliers__supplier_name);
			print qq[<td colspan=3>],
			  qq[<a href="mailto:$ename<$suppliers__email>" title="Send email">$suppliers__email</a>],
			  qq[</td>];
		}
		else {
			print qq[<td colspan=3>],
			  qq[<a href="mailto:$suppliers__email">$suppliers__email</a>],
			  qq[</td>];
		}
		print qq[</tr>];
	}

	print qq[<tr>];
	print qq[<th align=right valign=baseline>Entered</th>];
	print qq[<td align=left colspan=3> $suppliers__entry_date],
	  qq[ <font class=field-name>Changed</font> $suppliers__last_update],
	  qq[ <font class=field-name>by</font> $suppliers__last_update_by];
	print _make_audit_link('suppliers', 'supplier_id', $suppliers__supplier_id)
	  if ! $Printing;
	print qq[</td>];
	print qq[</tr>];

	print qq[</table>];
}

#------------------------------------------------------------------------------

sub display_supplier_comments {

	my $display_info = "yes";  # by default, expand and display comments

	display_supplier_comments_control_header($display_info);

	print <<EOF;

<!-- Supplier Comments -->

<p>
EOF

	local $\="\n";	# append newline to all print statements

	print "<p class=comment style='margin-top:0'>$suppliers__comments"
	  if $suppliers__comments;

	if ($in_office_comments) {
		# change CR/LF to <br>
		$in_office_comments =~ s/&#013;/<br>/g;
		print "<p class=comment>\n<b>In Office Comments:</b><br>",
		  $in_office_comments, "</p>";
	}

	if (! $addr->is_empty || ! $pay_addr->is_empty) {
		print "<p class=comment><b>Address Restrictions:</b><br>";
		print "$address_usage_desc $address_usage_notes"
		  if $address_usage_desc || $address_usage_notes;

		if (! $pay_addr->is_empty) {
			print "<br>Shipping address: ",
			  "$pay_usage_desc $pay_usage_notes"
			  if $pay_usage_desc || $pay_usage_notes;
		}
		print "</p>";
	}

	if ($suppliers__mailing_lists) {
		print 
		  "<p class=comment>",
		  "<b>Mailing Lists: </b>",
		  (join ", ", split_array($suppliers__mailing_lists)),
		  "</p>";
	}

	my $n = display_supplier_log($suppliers__supplier_id);

	print "(none)"
	  unless ($n || $suppliers__comments
	    || $address_usage_desc || $address_usage_notes
	    || $in_office_comments || $suppliers__mailing_lists);

	print <<EOF;
<!-- End Supplier Comments -->
EOF

	display_supplier_comments_control_trailer();
}

#------------------------------------------------------------------------------

sub display_supplier_comments_control_header {

	my $display_info = shift;

	my ($collapsed_display, $expanded_display);

	if ($display_info =~ /yes/i) {
		$collapsed_display="none";
		$expanded_display="block";
	}
	else {
		$collapsed_display="block";
		$expanded_display="none";
	}

	print <<EOF;

<!--  Begin Supplier Comments Control Header -->
<script language="JavaScript" type="text/javascript">
function show_supplier${suppliers__supplier_id}_comments_control() {
        self.document.getElementById('supplier${suppliers__supplier_id}_comments_control').style.display = 'block';	
        self.document.getElementById('supplier${suppliers__supplier_id}_comments_control_open').style.display = 'none';
}
function hide_supplier${suppliers__supplier_id}_comments_control() {
        self.document.getElementById('supplier${suppliers__supplier_id}_comments_control').style.display = "none";
        self.document.getElementById('supplier${suppliers__supplier_id}_comments_control_open').style.display = "block";
}
</script>

<div id="supplier${suppliers__supplier_id}_comments_control_open" style="display:${collapsed_display}; text-align:center; margin-top:3pt">

<table border="0" cellspacing="0" cellpadding="2" width=100%>
  <tr valign="top"> 
    <td width=16><a
         href="javascript:show_supplier${suppliers__supplier_id}_comments_control()"><img
         src='$Images/control.gif'
         alt='Show Supplier Comments'
	 name='supplier${suppliers__supplier_id}_comments_control_image'
         border=0
         onMouseOver="hilite(this)" onMouseOut="unhilite(this)"
        ></a>
    </td>

    <td width=105 align=left nowrap><a
         href="javascript:show_supplier${suppliers__supplier_id}_comments_control()"
         title='Show Supplier Comments'
         onMouseOver="hilite(self.document.images['supplier${suppliers__supplier_id}_comments_control_image'])"
	 onMouseOut="unhilite(self.document.images['supplier${suppliers__supplier_id}_comments_control_image'])"
	 class="plain"
        >Comments</a>
    </td>

    <td>
    </td>

    <td width=16><img src='$Images/transparent.gif'></td>
  </tr>
</table>
</div>

<div id="supplier${suppliers__supplier_id}_comments_control" style="display:${expanded_display}; text-align:center; margin-top:3pt; padding-bottom:12pt">

<table border="0" cellspacing="0" cellpadding="2" width=100%>
  <tr valign="top"> 
    <td width=16><a
         href="javascript:hide_supplier${suppliers__supplier_id}_comments_control()"><img
         src='$Images/control_open.gif'
         alt='Hide Supplier Comments'
         border=0
	 name='supplier${suppliers__supplier_id}_comments_control_open_image'
         onMouseOver="hilite(this)" onMouseOut="unhilite(this)"
        ></a>
    </td>

    <td width=105 align=left nowrap><a
         href="javascript:hide_supplier${suppliers__supplier_id}_comments_control()"
         title='Hide Supplier Comments'
         onMouseOver="hilite(self.document.images['supplier${suppliers__supplier_id}_comments_control_open_image'])"
	 onMouseOut="unhilite(self.document.images['supplier${suppliers__supplier_id}_comments_control_open_image'])"
	 class="plain"
        >Comments</a>
    </td>

    <td width=100%><a
         href="javascript:hide_supplier${suppliers__supplier_id}_comments_control()"><img
         src='$Images/transparent.gif'
         alt='Hide Supplier Comments'
	 width=100%
	 height=100%
         border=0
         onMouseOver="hilite(self.document.images['supplier${suppliers__supplier_id}_comments_control_open_image'])"
	 onMouseOut="unhilite(self.document.images['supplier${suppliers__supplier_id}_comments_control_open_image'])"
        ></a>
    </td>

    <td width=16><img src='$Images/transparent.gif'></td>
  </tr>

  <tr>
    <td width=16><img src='$Images/transparent.gif'></td>

    <td colspan=2>
      <hr>
<!--  End Supplier Comments Control Header -->
EOF
}

#------------------------------------------------------------------------------

sub display_supplier_comments_control_trailer {

	print <<EOF;
<!--  Begin Supplier Comments Control Trailer -->
    </td>

    <td width=16><img src='$Images/transparent.gif'></td>
  </tr>
</table>
</div>
<!--  End Supplier Comments Control Trailer -->

EOF
}

#------------------------------------------------------------------------------

sub display_supplier_log {

	my $cid = shift;

	no warnings qw(uninitialized once);

	local $\ = "\n";	# auto-append newlines to print statements

	my $qlog = Acme::Db->query(<<End_of_Sql);
		select
			supplier_id,
			log_id,
			log_type,
			log,
			callback_date,
			user_id,
			message,
			last_update,
			last_update_by,
			respond_date,
			respond_by
		from supplier_log
		where supplier_id = $cid
		order by log_type, last_update desc, log_id desc
End_of_Sql

	local our $count;
	local our $total_count = 0;

	$qlog->bind_columns_to_variables;

	$qlog->before_query(sub{ print "<div class=log>\n" });
	$qlog->after_query('print "</div>\n"');

	$qlog->before_group(

		'log_type' => q(
$count = 0;	# reset group count for hiding long lists of comments.
print "<div class=log>\n<p class=comments>\n<b>Log:</b> ",($log_type || ''),"<br>";
		));

	$qlog->after_group(
		'log_type' => q(
if ($count > $MaxLogDisplay) {
	my $count_on_page = 11;
	my $show_less_button = ($count - $MaxLogDisplay) > $count_on_page;
	print stop_hiding("remaining_log_".($log_type || ''), $show_less_button);
}
print "</div>\n";
$total_count += $count;
		));

	while ( $qlog->next() ) {
		if ($count == $MaxLogDisplay) {
			#
			# Hide the rest of log messages with a button to unhide
			#

			print start_hiding("remaining_log_${log_type}");
		}

		$count++;

		print "<p class=log>";
		print qq[<a href="/cgi-bin/form.cgi?request=edit&table=supplier_log&log_id=$log_id" onClick="new_target(this,'log')" title="Change log comment."> &bull; </a>],
		  meddate(timetodate($last_update)) . " $last_update_by",
		  ":<br>";

		print " $log " if ($log);
		print "</p>";

		#
		# Print callback message 
		#

		if ($callback_date && ! $respond_date) {

			my ($m,$d,$y) = split /\W+/, $callback_date;
			($y,$m,$d)= split /\W+/, $callback_date if $m > 12;
			my $callbacktime = timelocal(0,0,0,$d+0,$m-1,$y-1900);
			my $todaytime = timelocal(0,0,0,(localtime)[3,4,5]);

			print "<p class=callback>";

			print '<font style="color:red">missed</font> '
			  if ($callbacktime < $todaytime);

			print 
			  "callback",
			  " scheduled ";

			print meddate($callback_date);

			print " for $user_id" if ($user_id ne $last_update_by);

			print ": ";

			# Acknowledge tag is the same for all tables.
			my $wid = 'w' . $count;
			print
			  qq[<span style="font-size:7pt">(],
			  qq[<a href="/cgi-bin/suppliers/log/respond.pl?log_id=$log_id" onClick="new_target(this,'response')" title="Enter a response to a callback.">respond</a>],
			  qq[ | ],
			  qq[<a href="/cgi-bin/callbacks/acknowledge.cgi?table=supplier_log&log_id=$log_id" onClick="open_message('Acknowledging callback.','','','$wid')" target="$wid" title="Acknowledge callback without entering a response.">acknowledge</a>],
			  qq[)</span>\n];


			print "<br>$message";
			print "</p>";
		}
	}

	return $total_count;
}

#-------------------------------------------------------------------------------

sub _make_audit_link {
	my ($table, $key, $keyval) = @_;

	return '' unless Acme::Security::allow('Access',$Audit_Prog);

    	return <<EOF;

<a href="$Audit_Prog?table=$table&$key=$keyval" onClick="new_target(this,'audit')" title="Display audit list of changes to this record"><img src="$Images/audit.gif" border="0" onMouseOver="this.src='$Images/audit_hilite.gif'" onMouseOut="this.src='$Images/audit.gif'"></a>
EOF
}

