#!/usr/bin/perl -wT -I../lib

=head1 NAME

view.cgi -- Display database information.

=head1 SYNOPSIS

    view.cgi?table=customers&search=Folly

=head1 DESCRIPTION

The view.cgi program displays a row of data from a database table or sql query. 

This is a quick hack to display database records. Usually, the view program
is a custom program to view all relevant information. Until a custom program
is created, this is a placeholder.

=back

=head1 CHANGE LOG

 8/4/00   Initial version.
 6/14/01  Convert to Acme::SQL.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

package view;

use Acme::HTTP			qw(escape);
use Acme::HTML			qw(escape_html display_message);
use Acme::Db;
use Acme::SQL;
use Acme::Searcher;
use Acme::Base::Schema;
use Acme::HTML::NavButtons;
use Acme::Options;
use Local::Acme::Banners	qw(common_header common_footer);
use Acme::Config;
use CGI;

use Acme::Security;
Acme::Security::check('Access');

$|=1;		# flush output (synchronize errors with output)

$cgi = new CGI;
die "No arguments\n" unless $cgi->param();

$Styles = Acme::Config->Styles;		# location of css stylesheets

#
# Program Arguments
#

$request = $cgi->url_param('request')	|| $cgi->param('request') || '';

$table	 = $cgi->url_param('table')	|| $cgi->param('table');
$sql	 = $cgi->url_param('sql')	|| $cgi->param('sql');
$search  = $cgi->url_param('search')	|| $cgi->param('search');
$title	 = $cgi->url_param('title')	|| $cgi->param('title');

$id_field	  = $cgi->url_param('id_field')	   || $cgi->param('id_field');
$code_field	  = $cgi->url_param('code_field')  || $cgi->param('code_field');
$name_field	  = $cgi->url_param('name_field')  || $cgi->param('name_field');
$first_name_field = $cgi->url_param('first_name_field')  || $cgi->param('first_name_field');
$last_name_field  = $cgi->url_param('last_name_field')   || $cgi->param('last_name_field');
$phone_field      = $cgi->url_param('phone_field') || $cgi->param('phone_field');
$email_field      = $cgi->url_param('email_field') || $cgi->param('email_field');

$code_field_match = $cgi->url_param('code_field_match')  || $cgi->param('code_field_match');
$phone_field_match= $cgi->url_param('phone_field_match') || $cgi->param('phone_field_match');

$Printing = $cgi->url_param('print')	|| $cgi->param('print')	|| '';
$Printing = $Printing =~ /yes/i		||
	    $request =~ /print/i	||
	    ($Printing !~ /^no$/i && exists $cgi->{'print'} ? 1 : 0);

use vars qw($t $f);		# schema table and field pointers
use vars qw($b);		# navigation bar button object
use vars qw(@fields @values);	# database field names and their values

#
# Get user options
#

$opt = new Acme::Options;

#
# Set defaults
#

if ($table) {

	Acme::Security::check('Access',$table);

	$schema = new Acme::Base::Schema;

	$t = $schema->t($table)
	  or die "Cannot find a database table named '$table'\n";

	$fields = join ", ", $t->fields;
	$fields = $t->id . ", " . $fields if $t->id;

	$title             ||= $t->title;

	$record_set_name   ||= $t->title;

	$id_field	   ||= $t->id_field;
	$code_field	   ||= $t->code_field;
	$name_field	   ||= $t->name_field;
	$first_name_field  ||= $t->first_name_field;
	$last_name_field   ||= $t->last_name_field;
	$phone_field       ||= $t->phone_field;
	$email_field       ||= $t->email_field;

	$code_field_match  ||= $t->code_field_match;
	$phone_field_match ||= $t->phone_field_match;


	if (! $sql) {
		# Create default sql to search the table.
		# Assure key field is part of sql.
		$sql = "select $fields"
		     . " from " . $table
		     . " where 1 = 1";
	}

	if (! $search) {
		# Create default search parameter from key argument.
		my $key = $t->key;
		$search = $cgi->url_param($key) || $cgi->param($key);
	}
}

die "Empty search\n" unless $sql;

#
# Parse SQL script
#

my $sqlo = new Acme::SQL($sql);

#
# Evaluate Search script
#

if ($search) {
	if (my $sh = $cgi->param('save_history')) {
		# Save search in history
		if ($sh =~ /yes/i and $table) {
			$opt->enter("${table}Search",$search);
		}
		$cgi->delete('save_history');
	}

	if ($id_field || $code_field || $name_field || $last_name_field || $phone_field || $email_field) {

		#
		# Preprocess search script
		#

		if ($first_name_field || $last_name_field) {
			$search = Acme::Searcher::preprocess($search, {
			  RecordSetName   => $record_set_name,
			  IDField	  => $id_field,
			  CodeField	  => $code_field,
			  NameField       => $name_field,
			  FirstNameField  => $first_name_field,
			  LastNameField   => $last_name_field,
			  PhoneField      => $phone_field,
			  EmailField      => $email_field,
			  CodeFieldMatch  => $code_field_match,
			  PhoneFieldMatch => $phone_field_match,
			});
		}
		else {
			$search = Acme::Searcher::preprocess($search, {
			  RecordSetName   => $record_set_name,
			  IDField	  => $id_field,
			  CodeField	  => $code_field,
			  NameField	  => $name_field,
			  PhoneField      => $phone_field,
			  EmailField      => $email_field,
			  CodeFieldMatch  => $code_field_match,
			  PhoneFieldMatch => $phone_field_match,
			});
		}
	}

	my $srch = new Acme::Searcher $search, $cgi;

	#
	# Combine search script with SQL.
	#

	$sql = $srch->as_sql($sqlo);	# combine sql with search script
}

#
# Check query with security
#

$sql = Acme::Security::clean_sql($sql);

#
# Execute query
#

my $q = Acme::Db->query($sql);

$q->bind_columns_to_variables();	# reference fields by variables

print $cgi->header(-type=>'text/html') unless @ARGV;

if ($q->count == 0) {
	display_message("No records were found.", "$title",undef,0);
	exit 1;
}

#
# Set final default values
#

@fields = $sqlo->columns;

$title = "SQL View" if (! $title && $sql);
$title ||= "View";			# final default

#
# Display view
#

$b = new Acme::HTML::NavButtons print => $Printing;

display_header();

if (@values = $q->next()) {
	display_item();
}

display_footer();

exit 0;

#------------------------------------------------------------------------------

sub display_header {

	use vars qw($title $opt);

	my $auto_refresh = '';
	if ($opt->AutoRefresh =~ /yes/) {
		$auto_refresh = <<EOF;
<script language="JavaScript" type="text/javascript">
function auto_refresh() { self.location.reload(true) }
</script>
EOF
	}


	my $aux_script = $t->view_include_html if $table;
	$aux_script ||= '';
	$aux_script &&= "\n$aux_script";

	my $aux_stylesheet = $t->view_stylesheet_html if $table;
	$aux_stylesheet ||= '';
	$aux_stylesheet &&= "\n$aux_stylesheet";

	my $print_stylesheet = $print_cmd = '';
	if ($Printing) {
  		$print_cmd = 
  qq[ onLoad="window.print();window.setTimeout('window.close()',1500)"];

		$print_stylesheet =
  qq[<link rel="stylesheet" href="$Styles/view_print.css" type="text/css">\n];

		my $aux_print_stylesheet = $t->view_print_stylesheet_html if $table;
		$aux_print_stylesheet  &&= "$aux_print_stylesheet\n";

		$print_stylesheet .= $aux_print_stylesheet;
  	}

# Keep IE6 in IE5 compatablilty mode by not specifying url
#<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
#"http://www.w3.org/TR/html4/loose.dtd">

        print <<EOF;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">

<html>
<head>
<title>view $title</title>

<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<link rel="stylesheet" href="$Styles/button_bar.css" type="text/css">
<link rel="stylesheet" href="$Styles/view.css" type="text/css">$aux_stylesheet$aux_script
$print_stylesheet
$auto_refresh
</head>
<body$print_cmd>
EOF
	print $b->first_buttons;
	print common_header($title,$Printing);

	print <<EOF;
<div class="view">
<table>
EOF
}

#------------------------------------------------------------------------------

sub display_footer {
	print "</table>\n";
	print "</div>\n";
	print common_footer($title,$Printing);
	print $b->last_buttons;
	print "</body>\n</html>";
}

#------------------------------------------------------------------------------

sub display_item {

	local $^W;	# turn off -w warning on using undefined variables

	my $field_name;
	for (my $i=0; $i < @values; $i++) {
		$f = $t->f($fields[$i]);	# get field schema properties

		if (defined $f) {
			next if $f->private;
			$field_name = $f->title;
		}
		else {
			# ID key fields are undefined.
			$field_name = capitalize($fields[$i]);
		}

		print qq(<tr>\n);

		print qq(  <td class="field">).$field_name.qq(</td>\n);
		print qq(  <td class="value">).$values[$i].qq(</td>\n);

		print qq(</tr>\n);
	}
}

sub capitalize {
	my $s = shift;
	$s =~ s/_/ /g;		# change _ to spaces
	$s =~ s/\bid\b/ID/;	# uppercase ID
	$s =~ s/\b(\w+)\b/ucfirst(lc($1))/eg; # capitalize
	return $s;
}

__END__
