/******************************************************************************

Calendar -- Display a calendar of buttons

These routines display calendars and act on the date selected by the user --
usually by filling in a form field with the selected date.

The displayCalendar() is the main interface routine. It can be used
to display a pop-up calendar, or it can be used as a front-end to a cgi
program. The callback argument is a string which can be a function name
call (like 'func(arg)') or a URL (like 'http://www.company.com') or a form
field name. The callback function gets called with the added arguments of
the calendar object, the selected date (as a string), and a boolean which
is true if the selected date is an active date. If the callback is a URL,
it is called with the extra parameter 'date=mm/dd/yyyy' appended. If the
callback is the name of a form field, its value is set to the selected date,
and if the field has an onChange event handler defined, it is executed.

There are a multitude of options to displayCalendar(), all are optional.
Usually, only the first couple of arguments are used in pop-up calendars.
For example:
  
 <input type="text" name="fld">
 <input type="button"
  onClick="displayCalendar(1, this.form.fld, this.form.fld.value, 'today')">

This displays a one month calendar, filling in the 'fld' field with the
selected date, the selected date displayed on the calendar being the date
currently in the 'fld' field, and only the dates on or after today can be
selected as a return date by the user.

Options to displayCalendar() can be specified with key/value pairs if the
first argument is an Object. For example, the above call to displayCalendar
could be written like:

 displayCalendar( { 'n'           : 1,
                    'callback'    : this.form.fld,
		    'select_date' : this.form.fld.value,
		    'active_date' : 'today' } )

The hiliteDate() routine highlights dates after the calendar has been rendered.

The appendDate() routine fills in date cells with HTML elements. It is used
to display the pop-up menus in the date cells on full screen calendars. It
only works works if the calendar is rendered in the main window (not a
lookup window).

The showMenu() and hideMenu() routines assist displayCalendar() to implement
pop-up menus over the calendar. They do not work if the calendar is rendered
in a different browser window (like a lookup window).

Calendars are named to keep them separate. Calendar names correspond to
browser window names. Null or missing names use default names or the first
name in the Calendars object. Although several calendars can be rendered at
once, only one calendar per window frame is allowed because each calendar's
JavaScript that controls the date buttons uses global variables.

Calendars have a navigation button bar. On it is a menu of options which
allows the calendar to be resized and allows the calendar to be printed.
If the calendar is in a lookup window it is simply printed without the
navigation button bar. If the calendar is in the primary window, the
page is reloaded with the '&print=yes' parameter added to the url.

The navigation bar can be overwritten by specifying a 'toolbar' argument
or additional menu items can be added to the menu by using the 'menu'
function argument.

There are different types of calendars. Currently, they are:

	'full', 'list', 0, 1, 3, 6, or 12.

The calendar types 1, 3, 6, or 12 show little calendars with pop-up menus
listing the day's contents. They are shown with 1, 3, 6, or 12 calendars on
a page.

The calendar type 0 is the same as 'full' which stands for "fullscreen".
It displays one month's calendar on a page and the contents of the dates can
be seen without the pop-up menus.

The 'list' type calender is really not a calendar at all. Only the navigation
bar is displayed. The server is responsible for printing a list of items.

The 'full' and 'list' calendar types need a cooperating server side program.
They are not available on lookup windows.

On 2/9/2003 this was hacked to work on Netscape 6+.

*******************************************************************************/

/*
 * Calendar v2.1
 *
 * Copyright 2001-2003, Mackley F. Pexton.  All rights reserved.
 */

function displayCalendar(n_months,	/* number of months to display        */
			 callback,	/* form field name or function name   */
			 select_date,	/* date to be highlighted on calendar */
			 active_date,	/* before which buttons are inactive  */
			 start_date,	/* beginning month/year of calendar   */
			 win,		/* window to use or null              */
			 cal_name,	/* name of calendar                   */
			 autoclick,	/* seconds before hover becomes click */
			 unclickfunc,	/* function invoked to un-autoclick   */
			 stylesheet,	/* url of css stylesheet for calendar */
			 hilites,	/* array or object of date strings    */
			 cal_type,	/* 'full' for full window or 'list'   */
			 printing,	/* true if calendar is for print      */
			 menu,      	/* menu html added to toolbar         */
			 toolbar,	/* substitute navigation toolbar      */
			 button_help,	/* day button pop-up help text on full*/
			 day_menu_id,	/* id of HTML tag to pop-up over date */
			 id_prefix      /* menu id prefix for each pop-up     */
			) {

	// Arguments can be key/value pairs if the first argument is an object.
	if (typeof arguments[0] == 'object') {
		var args = arguments[0];
		//                 -- Program --       -- Alternate --
		//                 --  Names  --       --   Names   --
		n_months     = args['n_months']     || args['n'];
		callback     = args['callback']     || args['click'];
		select_date  = args['select_date'];
		active_date  = args['active_date'];
		start_date   = args['start_date'];
		win          = args['win']          || args['window'];
		cal_name     = args['cal_name']     || args['name'];
		autoclick    = args['autoclick'];
		unclickfunc  = args['unclick'];
		stylesheet   = args['stylesheet'];
		hilites      = args['hilites'];
		cal_type     = args['cal_type']     || args['type'];
		printing     = args['print']        || args['printing'];
		menu         = args['menu'];
		toolbar      = args['toolbar'];
		button_help  = args['button_help']  || args['button_title'];
		day_menu_id  = args['day_menu_id'];
		id_prefix    = args['id_prefix']    || args['day_menu_id_prefix'];
	}

	// Set defaults for missing parameters.
	if (n_months == null) n_months = 1;	// number of months
	if (n_months == 0) { cal_type = 'full'; n_months = 1; }
	if (!select_date && typeof callback == 'object' && callback.value) {
		select_date = callback.value;
	}
	// Note: dumb Netscape 6.2 needs dates in 01/01/2002 format, not 01-01-2002 (!!)
	if (!start_date) start_date = select_date;
	if (!start_date) start_date = new Date();
	else		 start_date = new Date(start_date.replace(/-/g,"/"));

	if (!cal_name) {			// calendar name
		if (win == self) {
			if (self.name) {
				cal_name = self.name;
			} else {
				// assign name to this window
				cal_name = 'self' + n_months;
				self.name = cal_name;
			}
		} else {
			// Pop-up window

			// Different pop-up window for each field.
			cal_name = 'cal' + Math.floor(Math.random() * 10000);

			// All fields use same pop-up window.
			//cal_name = 'cal'  + n_months;
		}
	}
	if (!win) {
		var settings = "resizable";
		var w,h;
	
		// Following dimensions are for a lookup window without
		// browser buttons, menu, scrollbar, or status line.
		if (n_months > 6) {       w=840; h=840; }
		else if (n_months > 3)  { w=840, h=520; }
		else if (n_months > 1)  { w=840, h=300; }
		else if (n_months == 1) { w=280, h=275; }

		settings += ",width="+w+",height="+h;

		// Center new window
		var w_width  = window.document.body.offsetWidth;
		var w_height = window.document.body.offsetHeight;
		var top  = window.screenTop  + (w_height/2 - h);
		var left = window.screenLeft + (w_width/2  - w/2);
		settings += ',top='+top+',left='+left;

		var win = window.open("", cal_name, settings);
		win.focus();
	}

	// Determine starting month/year of calendar
	start_month = start_date.getMonth();
	if (n_months > 6) {
		start_month = 0;			// beginning of year
		n_months = 12 * Math.floor((n_months+11) / 12);
	} else if (n_months > 3) {
		start_month -= start_month % 6;		// beginning of half
		n_months = 6 * Math.floor((n_months+5) / 6);
	} else if (n_months > 1) {
		start_month -= start_month % 3;		// beginning of quarter
		n_months = 3 * Math.floor((n_months+2) / 3);
	}
	start_month++;		// between 1 and 12
	start_year = start_date.getFullYear();

	Calendars.newCalendar(cal_name, n_months, start_month, start_year, win);
	if (callback)	 Calendars.setCallback(cal_name, callback);
	else		 Calendars.setCallback(cal_name, '');
	if (hilites)	 Calendars.setHilites(cal_name, hilites);
	if (autoclick)	 Calendars.setAutoClick(cal_name, autoclick);
	if (unclickfunc) Calendars.setAutoUnClick(cal_name, unclickfunc);
	if (stylesheet)	 Calendars.setAuxiliaryStylesheet(cal_name, stylesheet);
	if (active_date) Calendars.setActiveDate(cal_name, active_date);
	if (cal_type)    Calendars.setType(cal_name, cal_type);
	if (button_help) Calendars.setDayButtonHelp(cal_name, button_help);
	if (day_menu_id) Calendars.setDayMenuId(cal_name, day_menu_id);
	if (id_prefix)   Calendars.setDayMenuIdPrefix(cal_name, id_prefix);
	if (toolbar)     Calendars.setNavigationHTML(cal_name, toolbar);
	if (menu)        Calendars.setMenuHTML(cal_name, menu);
	if (printing)    Calendars.setPrinting(cal_name, printing);

	Calendars.display(cal_name);
	if (select_date) Calendars.hiliteDate(cal_name,'-select',select_date);
		
}

//-----------------------------------------------------------------------------

//
// Highlight dates on the calendar.
//

function hiliteDate() {
	// Convenience wrapper to Calendars.hiliteDate().

	// usage: hiliteDate(cal_name, hilite_code, date ...);
	// The calendar name cal_name is an optional argument.
	// The hilite_code can be a background color like '#33AA99'
	// or it can be a CSS class suffix like '-holiday'. It could
	// be a *.gif, *.jpg, *.jpeg, or *.png image file, but that
	// does not work in IE5.5.

	var n = 0;
	var cal_name = arguments[n];
	if (cal_name == null || Calendars[cal_name]) n++;
	var hilite_code = arguments[n++];

	if (cal_name == null || !Calendars[cal_name]) {
		// Use the last entered calendar name.
		cal_name = Calendars.last_calendar;
	}

	for (; n < arguments.length; n++) {
		Calendars.hiliteDate(cal_name, hilite_code, arguments[n]);
	}
}

//-----------------------------------------------------------------------------

//
// Fill in dates on the calendar.
//

function appendDate() {
	// Convenience wrapper to Calendars.appendDate().

	// This fills in dates on fullscreen calendars.
	// usage: appendDate(cal_name, element, date ...);
	// The calendar name cal_name is an optional argument.

	var n = 0;
	var cal_name = arguments[n];
	if (cal_name == null || Calendars[cal_name]) n++;
	var e = arguments[n++];

	if (cal_name == null || !Calendars[cal_name]) {
		// Use the last entered calendar name.
		cal_name = Calendars.last_calendar;
	}

	for (; n < arguments.length; n++) {
		Calendars.appendDate(cal_name, e, arguments[n]);
	}
}

//-----------------------------------------------------------------------------

//
// Popup display of menu when user hovers over a date on the calendar.
//

function showMenu(calendar, date, active) {
	// Display a menu when hovering over a date.
	var w = calendar.win;
	if (!w) w = self;
	var e = w.document.getElementById('cal_'+date)
	if (!e) return;
	var m;
	var menu_id = calendar.getDayMenuIdPrefix();
	if (menu_id) menu_id += date;
	if (!menu_id && active) menu_id = calendar.getDayMenuId() || 'day_menu';
	if (menu_id) m = w.document.getElementById(menu_id);
	if (!m) m = w.document.getElementById('day_' + date);
	if (!m) return;
	return do_showmenu(w,e,m);
}

function hideMenu(calendar, date) {
	// Hide the menu after hovering over a date.
	var w = calendar.win;
	if (!w) w = self;
	var e = w.document.getElementById('cal_'+date)
	if (!e) return;
	var m;
	var menu_id = calendar.getDayMenuIdPrefix();
	if (menu_id) menu_id += date;
	if (!menu_id) menu_id = calendar.getDayMenuId() || 'day_menu';
	if (menu_id) m = w.document.getElementById(menu_id);
	if (!m) m = w.document.getElementById('day_' + date);
	if (!m) return;
	return do_hidemenu(w,e,m);
}

//-----------------------------------------------------------------------------

function do_showmenu(w,e,m) {
	// Compute coordinates
	if (window.event.pageX || window.event.pageY) {
		m.style.left = window.event.pageX;
		m.style.top  = window.event.pageY;
	} else {
		m.style.left = document.body.scrollLeft + event.clientX;
		m.style.top  = document.body.scrollTop  + event.clientY;
	}

	// Display menu
	m.style.display = "block";
	m.style.visibility = "visible";

	// Recompute coordinates if display off the page
	if (window.event) {
		var rightedge  = document.body.clientWidth  - event.clientX;
		var bottomedge = document.body.clientHeight - event.clientY;
		var leftedge = document.body.scrollLeft + event.clientX - m.offsetWidth;
		var topedge  = document.body.scrollTop  + event.clientY - m.offsetHeight;
		if (rightedge < m.offsetWidth)
			m.style.left = (leftedge < document.body.scrollLeft) ? document.body.scrollLeft + 3 : leftedge;
		if (bottomedge < m.offsetHeight)
			m.style.top  = (topedge  < document.body.scrollTop) ? document.body.scrollTop + 3 : topedge;
		event.cancelBubble = true;
	}
}

function do_hidemenu(w,e,m) {
	m.style.display = "none";
	m.style.visibility = "hidden";
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

//
// Calendars is a global variable that points to a CalendarDisplays array of
// named CalendarDisplay objects, each of which contains a Calendar object
// used to render a calendar.
//

var Calendars = new CalendarDisplays();	// list of calendars by name

function CalendarDisplays() {
	// Object constructor
	// Arguments are optional.
	for (var i = 0; i < arguments.length; i++) {
		this[arguments[i]] = new CalendarDisplay();
	}
}

CalendarDisplays.prototype.newCalendar = function(
		cal_name, n_months, start_month, start_year, win) {
	this.setCalendar(cal_name, n_months, start_month, start_year, win);
	this[cal_name].hilites = new Object;	// Clear old hilites
	this.last_calendar = cal_name;
}

CalendarDisplays.prototype.display = function(cal_name,cmd) {
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;
	c.display(cmd);

	// Redo highlights
	if (!this[cal_name].hilites) return;

	var date, hilite_code;
	for (date in this[cal_name].hilites) {
		hilite_code = this[cal_name].hilites[date];
		c.hiliteDate(date, hilite_code);
	}
}

CalendarDisplays.prototype.redisplay = function(cal_name,cmd) {
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;

	if (c.win != self) {
		if (cmd != null) this.resize(cal_name,cmd);
		this.display(cal_name,cmd);
	} else {
		this.reload(cal_name,cmd);
	}
}

CalendarDisplays.prototype.reload = function(cal_name,cmd) {
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;
	if (cmd == null) cmd = c.Type;

	// Compose url with calendar command
	// Like: 'calendar=n_months,start_date,end_date'

	var newurl = self.location.href;

	// Remove possible resize command from url.
	newurl = newurl.replace(/([?&;])resize=[^&;]*[&;]?/,'$1');

	// Remove possible calendar command from url.
	newurl = newurl.replace(/([?&;])calendar=[^&;]*[&;]?/,'$1');

	// Add calendar command to url
	var sep = '?';
	if (newurl.match(/[?]/)) sep = '&';
	if (newurl.match(/[?&;]$/)) sep = '';

	newurl += sep+'calendar='+cmd+','+c.getStartDate()+','+c.getEndDate();
	// Add resize flag to url
	if (c.Type == cmd) {
		newurl +=  '&resize=no';
	}

	// Redisplay page
	self.location.href = newurl;
}

CalendarDisplays.prototype.resize = function(cal_name,cmd) {
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;

	if (cmd.match(/^\d+$/)) {
		var n_months = cmd;

		// Following dimensions are for a lookup window without
		// browser buttons, menu, scrollbar, or status line.
		// Note: resizeTo() uses different coordinates than
		// window.open().
		if (n_months > 6) {       w=844; h=920; }
		else if (n_months > 3)  { w=844, h=530; }
		else if (n_months > 1)  { w=844, h=340; }
		else if (n_months == 1) { w=292, h=318; }
		c.win.resizeTo(w,h);
	}
}

CalendarDisplays.prototype.print = function(cal_name) {
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;

	if (c.win != self) {
		c.print();
	} else {
		var cmd = c.Type;

		// Compose url with calendar command
		// Like: 'calendar=n_months,start_date,end_date'

		var newurl = self.location.href;

		// Remove possible calendar command at end of url.
		newurl = newurl.replace(/\Wcalendar=.*$/,'');
		var sep = newurl.match(/\?/) ? '&' : '?';

		// Add print cgi parameter
		newurl += sep+'print=yes';
		sep = '&';

		// Add calendar command
		newurl += sep+'calendar='+cmd+','+c.getStartDate()+','+c.getEndDate();

		// Set window to small size
		var settings = 'width=80,height=80,top=100,left=100,resizable';

		// Display new window for print job
		var w = window.open('','',settings);

		var title, message = title = "Printing...";
		var d = w.document;
		d.open();
		d.writeln('<html>');
		d.writeln('<head>');
		d.writeln('<title>' + title + '</title>');
		d.writeln('<style>.message{font-family:"Comic Sans MS" serif; font-size:8pt; padding:12pt 3pt</style>');
		d.writeln('</head>');
		d.writeln('<body>');
		d.writeln('<div class="message">');
		d.writeln('<p>' + message + '</p>');
		d.writeln('</div>');
		d.writeln('</body>');
		d.writeln('</html>');
		d.close();

		// Display calendar formatted for printer
		w.location.href = newurl;
	}
}

CalendarDisplays.prototype.setCalendar = function(
		cal_name, n_months, start_month, start_year, win) {
	if (!this[cal_name]) this[cal_name] = new CalendarDisplay();
	this[cal_name].calendar = new Calendar(
			n_months, start_month, start_year, win)
}

CalendarDisplays.prototype.setCallback = function(cal_name,callback) {
	if (!this[cal_name]) this[cal_name] = new CalendarDisplay();
	this[cal_name].callback = callback;
}

CalendarDisplays.prototype.setHilites = function(cal_name,hilites) {
	if (!this[cal_name]) this[cal_name] = new CalendarDisplay();
	var h = this[cal_name].hilites;
	var date;
	if (typeof hilites == 'array') {
		for (date in hilites) {
			h[date] = '-select';	// set to default highlight
		}
	} else if(typeof hilites == 'object') {
		for (date in hilites) {
			h[date] = hilites[date];// set to specified highlight
		}
	}
}

CalendarDisplays.prototype.setType = function(cal_name, cal_type) {
	if (!this[cal_name]) return;
	return this[cal_name].calendar.setType(cal_type);
}

CalendarDisplays.prototype.setStylesheet = function(cal_name,stylesheet) {
	if (!this[cal_name]) return;
	return this[cal_name].calendar.setStylesheet(stylesheet);
}

CalendarDisplays.prototype.setAuxiliaryStylesheet = function(cal_name,stylesheet) {
	if (!this[cal_name]) return;
	return this[cal_name].calendar.setAuxiliaryStylesheet(stylesheet);
}

CalendarDisplays.prototype.setActiveDate = function(cal_name,date) {
	if (!this[cal_name]) return;
	if (date == null || date == 'today') {
		var d = new Date();
		date = (d.getMonth()+1) + '-' + d.getDate() + '-' + d.getFullYear();
	}
	return this[cal_name].calendar.setActiveDate(date);
}

CalendarDisplays.prototype.setDayMenuIdPrefix = function(cal_name,str) {
	if (!this[cal_name]) return;
	return this[cal_name].calendar.setDayMenuIdPrefix(str);
}

CalendarDisplays.prototype.setDayMenuId = function(cal_name,str) {
	if (!this[cal_name]) return;
	return this[cal_name].calendar.setDayMenuId(str);
}

CalendarDisplays.prototype.setMenuHTML = function(cal_name,html) {
	if (!this[cal_name]) return;
	return this[cal_name].calendar.setMenuHTML(html);
}

CalendarDisplays.prototype.setNavigationHTML = function(cal_name,html) {
	if (!this[cal_name]) return;
	return this[cal_name].calendar.setNavigationHTML(html);
}

CalendarDisplays.prototype.setDayButtonHelp = function(cal_name, button_help) {
	if (!this[cal_name]) return;
	return this[cal_name].calendar.setDayButtonHelp(button_help);
}

CalendarDisplays.prototype.setPrinting = function(cal_name,printing) {
	if (!this[cal_name]) return;
	return this[cal_name].calendar.setPrinting(printing);
}

CalendarDisplays.prototype.setAutoClick = function(cal_name,n) {
	if (!this[cal_name]) return;
	return this[cal_name].calendar.setAutoClick(n);
}

CalendarDisplays.prototype.setAutoUnClick = function(cal_name,unclickfunc) {
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;

	if (eval('typeof('+unclickfunc.replace(/\(.*\)/,'')+')') == 'function') {
		// Register unclickfunc function
		this[cal_name].uncallback = unclickfunc;
		c.AutoUnClick = true;
	}
}

CalendarDisplays.prototype.getCalendar = function(cal_name) {
	if (!this[cal_name]) return;
	return this[cal_name].calendar;
}

CalendarDisplays.prototype.getCallback = function(cal_name) {
	if (!this[cal_name]) return;
	return this[cal_name].callback;
}

CalendarDisplays.prototype.getHilites = function(cal_name) {
	if (!this[cal_name]) return;
	return this[cal_name].hilites;
}

//-------------------------------------------------------------------------

CalendarDisplays.prototype.nextCalendar = function(cal_name) {
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;
	c.addStartMonth(c.n_months);
	this.redisplay(cal_name);
}

CalendarDisplays.prototype.nextYearCalendar = function(cal_name) {
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;
	c.addStartYear(1);
	this.redisplay(cal_name);
}

CalendarDisplays.prototype.prevCalendar = function(cal_name) {
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;
	c.addStartMonth(-c.n_months);
	this.redisplay(cal_name);
}

CalendarDisplays.prototype.prevYearCalendar = function(cal_name) {
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;
	c.addStartYear(-1);
	this.redisplay(cal_name);
}

CalendarDisplays.prototype.setYearCalendar = function(cal_name,year) {
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;
	c.setStartYear(year);
	this.redisplay(cal_name);
}

CalendarDisplays.prototype.setMonthCalendar = function(cal_name,month) {
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;
	if (month < 1 || month > 12) month = 1;
	c.setStartMonth(month-1);
	this.redisplay(cal_name);
}

CalendarDisplays.prototype.setQuarterCalendar = function(cal_name,quarter) {
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;
	var month = 10;
	if (quarter < 4) month = 7;
	if (quarter < 3) month = 4;
	if (quarter < 2) month = 1;
	c.setStartMonth(month-1);
	this.redisplay(cal_name);
}

CalendarDisplays.prototype.setHalfCalendar = function(cal_name,half) {
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;
	var month = 7;
	if (half < 2) month = 1;
	c.setStartMonth(month-1);
	this.redisplay(cal_name);
}

CalendarDisplays.prototype.setStyleCalendar = function(cal_name,style,date) {
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;
	var d, m;
	if (date) { d = new Date(date.replace(/-/g,"/")) }
	else      { d = c.start_date }
	m = d.getMonth();

	switch (style) {
	case '1':
		c.setStartMonth(m);
		c.setNMonths(1);
		break;
	case '3':
		c.setStartMonth(m - (m % 3));
		c.setNMonths(3);
		break;
	case '6':
		c.setStartMonth(m - (m % 6));
		c.setNMonths(6);
		break;
	case '12':
		c.setStartMonth(m - (m % 12));
		c.setNMonths(12);
		break;
	case '0':
	case 'full':
		c.setStartMonth(m);
		c.setNMonths(1);	// display only on full screen month
		break;
	case 'list':
		c.setStartMonth(m);
		c.setNMonths(1);	// list only one month at a time
		break;
	}
	this.redisplay(cal_name,style);
}

// hiliteDate() can be used after the calendar is displayed to change the
// css class name used to describe a date's display.
// This routine accepts more than one date and has a side effect of
// reactivating inactive buttons (unlike Calendar.prototype.hiliteDate()).

CalendarDisplays.prototype.hiliteDate = function(cal_name, hilite_code, date) {
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;

	var n, date;
	for (n=2; n < arguments.length; n++) {
		date = arguments[n];
		if (!c.isActive(date)) c.activateDate(date);
		c.hiliteDate(date,hilite_code);
		// save highlight for calendar redisplay
		this[cal_name].hilites[date] = hilite_code;
	}
}

// appendDate() can be used after the calendar is displayed to add
// content to date cells. This should only be used for full screen calendars.
// This routine accepts more than one date.

CalendarDisplays.prototype.appendDate = function(cal_name, e, date) {
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;

	var n, date;
	for (n=2; n < arguments.length; n++) {
		date = arguments[n];
		c.appendDate(date,e);
	}
}

CalendarDisplays.prototype.returnCalendar = function(cal_name, command, arg, date) {
	// This is the standard callback routine executed when a
	// user clicks a date or toolbar button.

	switch (command.toLowerCase()) {
	case 'select':
		Calendars.doSelect(cal_name,arg);
		break;
	case 'unselect':
		Calendars.doUnSelect(cal_name,arg);
		break;
	case 'next':
		Calendars.nextCalendar(cal_name);
		break;
	case 'nextyear':
		Calendars.nextYearCalendar(cal_name);
		break;
	case 'prev':
		Calendars.prevCalendar(cal_name);
		break;
	case 'prevyear':
		Calendars.prevYearCalendar(cal_name);
		break;
	case 'setmonth':
		Calendars.setMonthCalendar(cal_name,arg);
		break;
	case 'setquarter':
		Calendars.setQuarterCalendar(cal_name,arg);
		break;
	case 'sethalf':
		Calendars.setHalfCalendar(cal_name,arg);
		break;
	case 'setyear':
		Calendars.setYearCalendar(cal_name,arg);
		break;
	case 'setstyle':
		Calendars.setStyleCalendar(cal_name,arg,date);
		break;
	case 'print':
		Calendars.print(cal_name,arg);
		break;
	}
}

CalendarDisplays.prototype.doSelect = function(cal_name, date) {
	
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;

	var callback  = this[cal_name].callback;
	var active = this[cal_name].isActive(date);

	// Blank strings (not null) re-executes current url with date
	if (callback == '' && active) {
		var sep = (self.location.toString().search(/\?/) >= 0) ? '&' : '?';

		// Append date specification to url
		callback = self.location + sep + 'date=' + date;

		// Goto new url
		c.win.location.href = callback;
		return;
	}

	if (!callback) return;	// null strings are no-ops

	if (eval('typeof(self.'+callback.replace(/\(.*\)/,'')+')') == 'function') {
		// Callback is a function name.
		// Three arguments are appended to the callback function
		// arguments: the calendar object, selected date as a string,
		// and a boolian flag which is true unless the active date
		// was specified and the selected date is less than the 
		// active date.
		//
		// Note: The function must close the calendar window when done.
		// It is also responsible for making buttons inactive if before
		// the active date.

		callback = callback.replace(/\((.*)\)/,'');
		var callargs = (RegExp.$1.length > 0) ? ($RegExp.$1 + ',') : '';
		return eval(callback + '(' + callargs
			+ 'c'
			+ ', "' + date + '"'
			+ ', active'
			+ ')'
			);
	}
	
	if (!active) return;	// button is inactive

	if (typeof callback == 'string') {
		// Check if callback is a url.
		if (callback.search(/^\w+:/) >= 0) {
			var sep = (callback.search(/\?/) >= 0) ? '&' : '?';

			// Append date specification to url
			callback += sep + 'date=' + date;

			// Goto new url
			window.open(callback,'');
			self.close();
		}

		// Look for a form field with the name of the callback.
		var f,t,frm;
		if (callback.search(/(.*)\.(.*)/) >= 0) {
			// Parse form name from string.
			frm = RegExp.$1;
			fld = RegExp.$2;
		} else {
			// Search for field name in all forms.
			fld = callback;
			for (var i = 0; i < self.document.forms.length; i++) {
				frm = self.document.forms[i].name;
				if (frm[fld]) break;
			}
		}
		f = self.document[frm][fld];
		if (f) {
			t = f.type;
			if (t == 'text' || t == 'textarea' || t == 'hidden') {
				f.value = date;

				// Do field's onChange event.
				if (! this.doOnChange(f)) {
					c.win.close();
					return false;
				}

				// Close calendar window
				c.win.close();

				return true;
			}
		}
	}
	else if (typeof callback == 'object') {
		// object is assumed to be a form field object
		callback['value'] = date;

		// Do field's onChange event.
		if (! this.doOnChange(callback)) {
			c.win.close();
			return false;
		}

		// Close calendar window
		c.win.close();

		return true;
	}
}

CalendarDisplays.prototype.doUnSelect = function(cal_name, date) {
	
	if (!this[cal_name]) return;
	var c = this[cal_name].calendar;

	var uncallback = this[cal_name].uncallback;
	if (uncallback != null && eval('typeof(self.'+uncallback.replace(/\(.*\)/,'')+')') == 'function') {
		var active = this[cal_name].isActive(date);

		// Callback is a function name
		uncallback = uncallback.replace(/\((.*)\)/,'');
		var callargs = (RegExp.$1.length > 0) ? ($RegExp.$1 + ',') : '';
		return eval(uncallback + '(' + callargs
			+ 'c'
			+ ', "' + date + '"'
			+ ', active'
			+ ')'
			);
	}
}

CalendarDisplays.prototype.doOnChange = function(obj) {
	if (obj.onchange) {
		// execute the onChange event function

		/*
		 * Note: this should be called like Javascript calls
		 * an event handler. But I haven't figured that out yet
		 * so I just substitute 'obj' for any reference
		 * to 'this'. IE uses the function name 'anonymous()'
		 * but Netscape uses 'onchange()'. Netscape needs a
		 * event parameter passed to the handler, according
		 * to its syntax.
		 */

		var func = obj.onchange.toString();
		func = func.replace(/\bthis\b/g,"obj");
		func.match(/(\w+) (\w+)/);
		var func_call = RegExp.$2;

		if ( eval(func + ";" + func_call + "()") == false) {
			return false;
		}
	}
	return true;
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

// Define individual object of Calendars container.

function CalendarDisplay() {
	// Object constructor.
	this.calendar = null;		// Calendar object
	this.callback = null;		// callback function or field name
	this.uncallback = null;		// function called when date unselected
	this.active_date = null;	// buttons before this date are inactive
	this.hilites  = new Object;	// collection of date, highlight pairs
}

CalendarDisplay.prototype.isActive = function(date) {
	return this.calendar.isActive(date);
}

//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------

// Calendar object is to display a calendar

function Calendar(n_months, month, year, win) {
	// Calendar object constructor.
	// month is between 1 and 12

	// Check type of calendar (fullscreen, list, or normal)
	if (n_months == null) n_months = 1;
	if (n_months == 0 || n_months == 'full') {
		n_months = 1;
		this.setType('full');
	} else if (n_months == 'list') {
		this.setType('list');
	} else {
		this.setType(n_months);
	}

	// Properties
	this.n_months	= n_months	|| 1;
	this.win	= win		|| self;

	// Instance variables
	// Set today to midnight today.
	this.today = new Date();
	this.today.setHours(0);
	this.today.setMinutes(0);
	this.today.setSeconds(0);
	this.today.setMilliseconds(0);

	// Set default arguments
	if (!year)  year  = this.today.getFullYear();
	if (!month) month = this.today.getMonth() + 1;

	this.start_date = new Date(year,month-1);
	this.date = new Date(this.start_date); // current date being processed

	this.activated   = new Object;		// list of reactivated buttons
	this.deactivated = new Object;		// list of deactivated buttons
}

/*
 * Class constants
 */

Calendar.prototype.DaysInMonth = [31,28,31,30,31,30,31,31,30,31,30,31];
Calendar.prototype.MonthNames  = [
		"January",	"February",	"March",
		"April",	"May",		"June",
		"July",		"August",	"September",
		"October",	"November",	"December"
		];

Calendar.prototype.Columns = 3;			// number of columns in calendar
Calendar.prototype.WeekColumns = 7;		// number of columns in week

/*
 * Class default values
 */

// Define the default stylesheet.
// Note: stylesheet settings are used only on lookup windows, not primary windows.
Calendar.prototype.Stylesheet = '/styles/calendar.css';
Calendar.prototype.AuxiliaryStylesheet = null;	/* ancillary stylesheet */

// Define the function the calendar is to invoke when a date button 
// or any of the navigation buttons are clicked.
// The function arguments are the window name (used as a name of a calendar),
// a command, and the date of the button if the date button was clicked.
Calendar.prototype.CallbackFunction = 'Calendars.returnCalendar';

// Define number of seconds before a mouse click is simulated when the user
// hovers over a date button. Set to 0 to disable.
Calendar.prototype.AutoClick = 0;

// Define whether callback should be executed if date becomes out of focus.
Calendar.prototype.AutoUnClick = false;

// Define the date before which buttons are not highlighted. The buttons are
// still active if clicked by the user allowing the callback function to 
// determine if they should be executed.
Calendar.prototype.ActiveDate = null;

// Setting NavigationHTML overrides the default HTML generated programically
// for the button bar of calendar navigation buttons.
Calendar.prototype.NavigationHTML = null;

/*
 * Calendar methods to change settings
 */

Calendar.prototype.setToday = function(s) {
	// Set today's date (which has its own display class)
	if (!s) return;
	this.today.setTime(Date.parse(s));
	// Set to midnight for date comparisons
	this.today.setHours(0);
	this.today.setMinutes(0);
	this.today.setSeconds(0);
	this.today.setMilliseconds(0);
}

Calendar.prototype.setType = function(s) {
	// Set calendar type. Currently: 'full', 'list', or 0, 1, 3, 6, or 12.
	var p = this.Type;
	this.Type = s;
	return p;
}

Calendar.prototype.setStylesheet = function(s) {
	// Set CSS style sheet used by calendars.
	var p = this.Stylesheet;
	this.Stylesheet = s;
	return p;
}

Calendar.prototype.setAuxiliaryStylesheet = function(s) {
	// Set auxiliary CSS style sheet used by calendars.
	var p = this.AuxiliaryStylesheet;
	this.AuxiliaryStylesheet = s;
	return p;
}

Calendar.prototype.setCallbackFunction = function(s) {
	// Set function used to accept calendar requests and results.
	var p = this.CallbackFunction;
	this.CallbackFunction = s;
	return p;
}

Calendar.prototype.setAutoClick = function(n) {
	// Set number of seconds before a date button is automatically clicked.
	var p = this.AutoClick;
	this.AutoClick = n;
	return p;
}

Calendar.prototype.setAutoUnClick = function(bool) {
	// Set number of seconds before a date button is automatically clicked.
	var p = this.AutoUnClick;
	this.AutoUnClick = (bool) ? true : false;
	return p;
}

Calendar.prototype.setDayMenuIdPrefix = function(s) {
	// Set menu id prefix to use as a pop-up menu for each day.
	// The prefix has the date (mm-dd-yyyy) appended to make the DayMenuId.
	var p = this.DayMenuIdPrefix;
	this.DayMenuIdPrefix = s;
	return p;
}

Calendar.prototype.getDayMenuIdPrefix = function() {
	// Get menu id prefix to use as a pop-up menu for each day.
	return this.DayMenuIdPrefix;
}

Calendar.prototype.setDayMenuId = function(s) {
	// Set menu id to use as a pop-up menu for each day.
	var p = this.DayMenuId;
	this.DayMenuId = s;
	return p;
}

Calendar.prototype.getDayMenuId = function() {
	// Get menu id to use as a pop-up menu for each day.
	return this.DayMenuId;
}

Calendar.prototype.setMenuHTML = function(s) {
	// Set menu to display on default navigation toolbar
	var p = this.MenuHTML;
	this.MenuHTML = s;
	return p;
}

Calendar.prototype.setNavigationHTML = function(s) {
	// Set navigation toolbar
	var p = this.NavigationHTML;
	this.NavigationHTML = s;
	return p;
}

Calendar.prototype.setDayButtonHelp = function(s) {
	// Set pop-up help text when hovering over a button.
	var p = this.DayButtonHelp;
	this.DayButtonHelp = s;
	return p;
}

Calendar.prototype.setActiveDate = function(s) {
	// Set date before which date buttons are not highlighted.
	var p = this.ActiveDate;
	if (s) this.ActiveDate = new Date(s.replace(/-/g,"/"));	// enable
	else   this.ActiveDate = null;				// disable
	return p;
}

Calendar.prototype.setPrinting = function(printing) {
	var p = this.Printing;
	if (printing) { this.Printing = true; }
	else          { this.Printing = false;}
	return p;
}

Calendar.prototype.isPrinting = function() {
	if (this.Printing) return true;
	return false;
}

Calendar.prototype.isActive = function(date) {
	if (this.deactivated[date] && this.deactivated[date] != 'undefined') return false;
	if (this.activated[date] && this.activated[date] != 'undefined') return true;

	if (!this.ActiveDate) return true;	// default is yes

	var testdate = new Date(date.replace(/-/g,"/"));
	return (testdate.valueOf() >= this.ActiveDate.valueOf());
}

Calendar.prototype.activateDate = function(date) {
	// Re-activate date by entering it in the activation object.
	if (this.deactivated[date]
	&& this.deactivated[date] != 'undefined')
		delete this.deactivated[date];
	this.activated[date] = true;

	// Adjust global lists in calendar window
	if (this.win.deactivated_dates[date]
	&& this.win.deactivated_dates[date] != 'undefined')
		delete this.win.deactivated_dates[date];
	this.win.activated_dates[date] = true;
}

Calendar.prototype.deactivateDate = function(date) {
	// De-activate date by entering it in the deactivation object.
	if (this.activated[date]
	&& this.activated[date] != 'undefined')
		delete this.activated[date];
	this.deactivated[date] = true;

	// Adjust global lists in calendar window
	if (this.win.activated_dates[date]
	&& this.win.activated_dates[date] != 'undefined')
		delete this.win.activated_dates[date];
	this.win.deactivated_dates[date] = true;
}

/*
 * Methods to assist creating url calendar commands.
 */

Calendar.prototype.getStartDate = function() {
	var mm   = this.start_date.getMonth() + 1;
	var dd   = this.start_date.getDate();
	var yyyy = this.start_date.getFullYear();

	if (mm < 10) mm = '0' + mm;
	if (dd < 10) dd = '0' + dd;

	return mm+'-'+dd+'-'+yyyy;
}

Calendar.prototype.getEndDate = function() {
	// Return last date of calendar.
	var mm   = this.start_date.getMonth(); 
	var yyyy = this.start_date.getFullYear();
	var n = this.n_months - 1;

	mm += (n % 12);
	yyyy += Math.floor(n / 12);
	var dd = this.getDaysInMonth(mm,yyyy);
	mm++;
	if (mm < 10) mm = '0' + mm;

	return mm+'-'+dd+'-'+yyyy;
}

Calendar.prototype.getNMonths = function() {
	return this.n_months;
}

Calendar.prototype.setNMonths = function(n) {
	var p = this.n_months;
	this.n_months = n;
	return p;
}

/*
 * Calendar methods for years
 */

Calendar.prototype.getStartYear = function() {
	return this.start_date.getFullYear();
}

Calendar.prototype.setStartYear = function(y) {
	return this.start_date.setFullYear(y);
}

Calendar.prototype.addStartYear = function(n) {
	return this.setStartYear(this.start_date.getFullYear() + n);
}

/*
 * Calendar methods for months
 */

Calendar.prototype.getStartMonth = function() {
	// Return start month (0-11)
	return this.start_date.getMonth();
}

Calendar.prototype.setStartMonth = function(m) {
	// Set current month (0-11), return previous date
	return this.start_date.setMonth(m);
}

Calendar.prototype.addStartMonth = function(n) {
	return this.setStartMonth(this.start_date.getMonth() + n);
}

Calendar.prototype.getMonthName = function(d) {
	if (!d) d = this.date;
	return this.MonthNames[d.getMonth()];
}

Calendar.prototype.addMonths = function(n) {
	// Add number of months to start_date (not current date month).
	this.date.setTime(this.start_date.getTime());
	return this.date.setMonth(this.date.getMonth() + n);
}

/*
 * Calendar methods for dates
 */

Calendar.prototype.addDate = function(n) {
	if (!n) n = 1;
	return this.date.setDate(this.date.getDate() + n);
}

Calendar.prototype.isWeekday = function(wday) {
	if (!wday) wday = this.date.getDay();
	if (wday == 0 || wday == 6) return false;	// not on Sun or Sat
	return true;
}

Calendar.prototype.isWeekend = function(wday) {
	return ! this.isWeekday(wday);
}

Calendar.prototype.isLeapYear = function(y) {
	if (!y) y = this.date.getFullYear();
	return ((y % 4) == 0) && ( ((y % 100) != 0) || ((y % 400) == 0) );
}

Calendar.prototype.getDaysInMonth = function(m, y) {
	// m between 0-11
	if (!m) m = this.date.getMonth();
	if (!y) y = this.date.getFullYear();
	if (m != 1 || ! this.isLeapYear(y)) return this.DaysInMonth[m];
	return 29;
}

/*
 * Calendar display methods
 */

Calendar.prototype.display = function(cmd) {
	
	this.date.setTime(this.start_date.getTime());

	if (this.win != self) {
		// This is a lookup window, write HTML header.
		this.win.document.open();

		this.writeln('<html>');
		this.writeln('<head><title>Calendar</title>');
		if (this.Stylesheet) {
			this.writeln('<link rel="stylesheet"'
			  + ' href="'+this.Stylesheet+'"'
			  + ' type="text/css">');
		}
		if (this.AuxiliaryStylesheet) {
			this.writeln('<link rel="stylesheet"'
			  + ' href="'+this.AuxiliaryStylesheet+'"'
			  + ' type="text/css">');
		}
		this.writeln(this.getJavaScript());
		this.writeln('</head>');

		this.writeln('<body onClick="do_unselect()">');
	} else {
		// This is the primary window, HTML header has already been output.
		this.writeln(this.getJavaScript());
		if (! document.body.onclick) document.body.onclick = do_unselect;
	}

	var y = this.date.getFullYear();
	this.write('<div class="calendar" id="calendar'+y+'">');
	if (this.n_months > 1) {
		this.write('<span class="calendar-title">Year '+y+'</span>');
	}

	this.writeln(this.getNavigationHTML());

	if (this.Type == 'list') {
		this.writeln('</div>');
		return;
	}

	this.writeln('<center><table class="calendar" cellspacing="0" cellpadding="0">');

	// Start row of calendar
	this.writeln('<tr class="quarter">');
	var m, month_title;
	if (this.n_months > this.Columns) {
		// Beginning blank months to keep display even.
		var beginning_m = this.start_date.getMonth() % this.Columns;
		for (m = 0; m < beginning_m; m++) {
			this.writeln('<td class="month" align="center"> </td>');
		}
	}

	// Get the complete calendar code for each month.
	for (m = 0; m < this.n_months; m++) {
		this.addMonths(m);
		month_title = this.getMonthName();
		if (this.n_months == 1) month_title += '/'+y

		this.writeln('<td class="month" align="center">');
		this.writeln(this.getMonthHTML(month_title));
		this.writeln('</td>');

		if ((m % this.Columns) == (this.Columns - 1)) {
			this.writeln('</tr>');
			this.writeln('<tr class="quarter">');
		}
	}

	// Finish out row of calendar with blank months.
	if (this.n_months > 1) {
		var remaining_m = (this.start_date.getMonth() + this.n_months) % this.Columns;
		if (remaining_m > 0) {
			for (m = 0; m < remaining_m; m++) {
				this.writeln('<td class="month"> </td>');
			}
		}
	}
	this.writeln('</tr>');

	this.writeln('</table></center></div>');

	if (this.win != self) {
		this.writeln('</body>');
		this.writeln('</html>');
		this.win.document.close();
	}
}

Calendar.prototype.print = function() {
	// Hide navigation button bar
	var e = this.win.document.getElementById("navigation_toolbar");
	e.style.display = 'none';
	e.style.visibility = 'hidden';

	// Print calendar window
	this.win.print();

	// Show navigation button bar
	e.style.display = 'block';
	e.style.visibility = 'visible';
}
	
// hiliteDate() can be used after the calendar is displayed to change
// a date's appearance.  The hilite_code can be a background color like
// '#33AA99' or it can be a CSS class suffix like '-holiday'. It could
// be a *.gif, *.jpg, *.jpeg, or *.png image file, but that does not work
// in IE5.5.

Calendar.prototype.hiliteDate = function(date, hilite_code) {
	var w = this.win;
	if (!w) w = self;

	var d = new Date(date.replace(/-/g,"/"));
	if (isNaN(d)) return;

	var mm   = d.getMonth() + 1;
	var dd   = d.getDate();
	var yyyy = d.getFullYear();

	if (mm < 10) mm = '0' + mm;
	if (dd < 10) dd = '0' + dd;

	var date_string = mm+'-'+dd+'-'+yyyy;

	var e = w.document.getElementById('cal_'+date_string)
	  
	this.hiliteElement(e, hilite_code);

	// Highlight calendar cell if full screen display.
	// Only style codes are used, colors and pix are for buttons only.
	if (e && hilite_code.match(/^-/) && this.Type == 'full') {
		this.hiliteElement(e.parentNode, hilite_code);
	}
}

Calendar.prototype.hiliteElement = function(e, hilite_code) {
	if (!e) return;

	if (hilite_code.match(/^#[0-9aAbBcCdDeEfF]{6}/)) {
		// hilite_code is a background color
		e.style.backgroundColor = hilite_code;
	}
	else if (hilite_code.match(/\w\.(gif)|(jpg)|(jpeg)|(png)$/i)) {
		// hilite_code is a background image
		//## This does not work yet in IE5.5
		//##e.style.backgroundImage = hilite_code;
	}
	else if (hilite_code.match(/^-/)) {
		// hilite_code is a css class suffix
		// Remove any suffix after 'day' in existing class name
		var cl = e.className.replace(/day.*/,'day');

		// Highlight day button by appending hilite_code argument to 
		// existing class name.
		e.className = cl + hilite_code;
	}
	else {
		// assume hilite_code is a wordy background color
alert("unknown hilite code: '"+hilite_code+"'"); //###
		//###e.style.backgroundColor = hilite_code;
	}
}

// appendDate() can be used after the calendar is displayed to add
// content to date cells. This should only be used for full screen calendars.

Calendar.prototype.appendDate = function(date, e) {
	if (typeof e != 'object') return;

	var w = this.win;
	if (!w) w = self;

	// Note: dumb Netscape 6.2 needs dates in mm/dd/yyyy format (not mm-dd-yyyy) !!
	var d = new Date(date.replace(/-/g,"/"));
	if (isNaN(d)) return;

	var mm   = d.getMonth() + 1;
	var dd   = d.getDate();
	var yyyy = d.getFullYear();

	if (mm < 10) mm = '0' + mm;
	if (dd < 10) dd = '0' + dd;

	var date_string = mm+'-'+dd+'-'+yyyy;

	var cell = w.document.getElementById('cal_'+date_string)
	  
	if (cell) {
		var newe = cell.parentNode.appendChild(e);
		newe.style.display = "block";
		newe.style.visibility = "visible";
	}
	else {
		// no cell to attach to, hide entry
		e.style.display = "none";
		e.style.visibility = "hidden";
	}
}

Calendar.prototype.getJavaScript = function() {
	// Return the JavaScript functions needed for the calendar

	/*
	 * The following Javascript supports elements which have the event
	 * handlers defined as:
	 *
	 *   onMouseOver="hilite_date(this)" onMouseOut="unhilite_date(this)"
	 *   onClick="do_callback(this)"
	 *
	 * Note: for cross browser support hack, above handlers are now defined
	 * as following and the event is saved in window.event global if non IE.
	 *
	 *   onMouseOver="hilite_date(this,event)"
	 *   onMouseOut="unhilite_date(this,event)"
	 *   onClick="do_callback(this,event)"
	 */

	// Note: the list of dates that have been deactivated/reactivated must
	// reside in global variables in the calendar window.
	var d;
	var activate_dates_cmd = '';
	var deactivate_dates_cmd = '';
	for (d in this.activated)
		activate_dates_cmd += 'activated_dates["'+d+'"] = true;\n';

	for (d in this.deactivated)
		deactivate_dates_cmd += 'deactivated_dates["'+d+'"] = true;\n';

	// Note: IE5 considers the next line to be line number 194 in errors.

	return ''
	  + '<script language="JavaScript" type="text/javascript">\n'
	  + 'var cal=' + (this.win == self ? 'self' : 'self.opener') + ';\n'
	  + 'var hilited_date = new Object;\n'
	  + 'var selected_date = new Object;\n'
	  + 'var activated_dates = new Object;\n'
	  + 'var deactivated_dates = new Object;\n'
	  + 'var save_events = document.all ? false : true;\n'
	  + '//------------------------------------\n'
	  + 'function save_event(evt) {\n'
	  + '  if (save_events && evt) {\n'
	  + '    event = { "pageX"   : evt.pageX,   "pageY"   : evt.pageY,\n'
	  + '              "clientX" : evt.clientX, "clientY" : evt.clientY };\n'
	  + '  }\n'
	  + '}\n'
	  + '//------------------------------------\n'
	  + 'function hilite_date(e,evt) {\n'
	  + '  if(!is_active_date(e.id.substr(4))) return true;\n'
	  + ((this.Type != 'full')
	    ? '  if(hilited_date.is_hilited && hilited_date.element == e) return true;\n'
	     +'  if(selected_date.is_selected) {\n'
	     +'    if(selected_date.element == e) return true;\n'
	     +'    if(selected_date.autoclicked) do_unselect();\n'
	     +'  }\n'
	    : '    if(selected_date.autoclicked) do_unselect();\n')
	  + '  hilited_date.element = e;\n'
	  + '  hilited_date.id = e.id;\n'
	  + '  hilited_date.is_hilited = true;\n'
	  + '  // hilite by changing css class\n'
	  + '  //e.className += "-hilite";\n'
	  + '  // hilite by changing object style properties (faster?)\n'
	  + '  with (e.style) {borderColor="black";fontWeight="bold"}\n'
	  + ((this.Type == 'full')
	    ? '  e.parentNode.parentNode.style.borderColor="black";\n'
	    : '  save_event(evt);schedule_autoclick()\n')
	  + '  return true;\n'
	  + '}\n'
	  + '//------------------------------------\n'
	  + 'function unhilite_date(e,evt) {\n'
	  + '  unschedule_autoclick();\n'
	  + '  if (hilited_date.is_hilited\n'
	  + '  && (!selected_date.is_selected || selected_date.element != hilited_date.element)) {\n'
	  + '    // unhighlight by changing css class\n'
	  + '    //hilited_date.element.className = hilited_date.origClassName;\n'
	  + '    // unhighlight by changing object style properties (faster?)\n'
	  + '    with (hilited_date.element.style) {borderColor="#CCCCCC";fontWeight=""}\n'
	  + ((this.Type == 'full')
	    ? '  hilited_date.element.parentNode.parentNode.style.borderColor="#CCCCCC";\n'
	    : '')
	  + '    hilited_date.is_hilited = false;\n'
	  + '  }\n'
	  + '}\n'
	  + '//------------------------------------\n'
	  + 'function is_active_date(date) {\n'
	  + 'if(deactivated_dates[date] && deactivated_dates[date] != "undefined") return false;\n'
	  + 'if(activated_dates[date] && activated_dates[date] != "undefined") return true;\n'
	  + (this.ActiveDate
	    ? ('  var d = new Date(date.replace(/-/g,"/"));\n  return (d.valueOf() >= '+this.ActiveDate.valueOf()+');\n')
	    : '  return true;')
	  + '}\n'
	  + '//------------------------------------\n'
	  + 'var timeout_id;\n'
	  + 'function schedule_autoclick() {\n'
	  + (this.AutoClick
	    ? ('  timeout_id = setTimeout("do_autoclick()",'+(this.AutoClick*1000)+');\n')
	    :  '  return;\n')
	  + '}\n'
	  + '//------------------------------------\n'
	  + 'function unschedule_autoclick() {\n'
	  + (this.AutoClick
	    ? ('  clearTimeout(timeout_id);\n')
	    :  '  return;\n')
	  + '}\n'
	  + '//------------------------------------\n'
	  + 'function do_autoclick() {\n'
	  + '  if(!selected_date.is_selected && hilited_date.is_hilited) {\n'
	  + '    selected_date.autoclicked = true;\n'
	  + '    if (save_events) {\n'
	  + '      do_callback(hilited_date.element,window.event); // Call click handler\n'
	  + '    } else {\n'
 	  + '      hilited_date.element.click();	// Simulate click event\n'
	  + '    }\n'
	  + '  }\n'
	  + '}\n'
	  + '//------------------------------------\n'
	  + 'function do_select(e) {\n'
	  + '  unschedule_autoclick();\n'
	  + '  selected_date.element = e;\n'
	  + '  selected_date.id = e.id;\n'
	  + '  selected_date.is_selected = true;\n'
	  + '  selected_date.last_id = e.id;\n'
//##	  + '  selected_date.origClassName = (hilited_date.is_hilited) ? hilited_date.origClassName : e.className;\n'
	  + '  cal.'+this.CallbackFunction+'(self.name, "select", e.id.replace(\/cal_\/,""));\n'
	  + '}\n'
	  + '//------------------------------------\n'
	  + 'function do_unselect() {\n'
	  + '  unschedule_autoclick();\n'
	  + '  if(selected_date.is_selected) {\n'
	  + '    selected_date.is_selected = false;\n'
	  + '    selected_date.autoclicked = false;\n'
//##	  + '    selected_date.element.className = selected_date.origClassName;\n'
	  + '    // unhighlight by changing object style properties (faster?)\n'
	  + '    with (selected_date.element.style) {borderColor="#CCCCCC";fontWeight=""}\n'
	  + ((this.Type == 'full')
	    ? '    selected_date.element.parentNode.parentNode.style.borderColor="#CCCCCC";\n'
	    : '')
	  + '    if(selected_date.element == hilited_date.element) hilited_date.is_hilited = false;\n'
	  + '    cal.'+this.CallbackFunction+'(self.name, "unselect", selected_date.id.replace(\/cal_\/,""));\n'
	  + '  }\n'
	  + '}\n'
	  + '//------------------------------------\n'
	  + 'function do_choose(e) {\n'
	  + '  // do select without a callback\n'
	  + '  // can get called on same click event as do_select()\n'
	  + '  unschedule_autoclick();\n'
	  + '  do_unselect();\n'
	  + '  hilite_date(e);\n'
	  + '}\n'
	  + '//------------------------------------\n'
	  + 'function do_request(cmd,arg) {\n'
	  + '  unschedule_autoclick();\n'
	  + '  var d;\n'
	  + '  if(selected_date.last_id) {\n'
	  + '    d = selected_date.last_id.replace(/cal_/,"");\n'
	  + '    do_unselect();\n'
	  + '  }\n'
	  + '  cal.'+this.CallbackFunction+'(self.name, cmd, arg, d);\n'
	  + '}\n'
	  + '//------------------------------------\n'
	  + 'function do_callback(e,evt) {\n'
	  + '  unschedule_autoclick();\n'
	  + '  if(selected_date.is_selected) do_unselect();\n'
	  + ((this.Type == 'full')
	    ? '  hilite_date(e);\n'
	    : '')
	  + '  if(!selected_date.autoclicked) save_event(evt);\n'
	  + '  do_select(e);\n'
	  + '}\n'
	  + '//------------------------------------\n'
	  + deactivate_dates_cmd
	  + activate_dates_cmd
	  + '//------------------------------------\n'
	  + 'function calkeypress(evt) {\n'
	  + '  var unicode = evt.charCode ? evt.charCode : evt.keyCode;\n'
	  + '  switch (unicode) {\n'
	  + '  case 27:	// escape key\n'
	  + '     window.close();\n'
	  + '    break;\n'
	  + '  }\n'
	  + '}\n'
	  + 'if(window.document.onkeypress == null)\n'
	  + '  window.document.onkeypress = calkeypress;\n'
	  + '//------------------------------------\n'
	  + '// Navigation menu scripts            \n'
	  + '//------------------------------------\n'
	  + 'var nav_menu = new Object;\n'
	  + 'nav_menu.is_hilited = false;\n'
	  + '//------------------------------------\n'
	  + 'function show_nav_menu() {\n'
	  + '  clearTimeout(nav_menu.nav_hide_timer);\n'
	  + '  if(nav_menu.is_hilited) return;\n'
	  + '  if(window.event && ! save_events) {\n'
	  + '    var m  = document.getElementById("calendar_menu");\n'
	  + '    m.style.left=document.body.scrollLeft +'
	  + ' event.clientX -'
	  + ' event.offsetX - 3;\n'
	  + '    if(event.srcElement.clientHeight != null) {\n'
	  + '      m.style.top=document.body.scrollTop+event.clientY -'
	  + ' event.offsetY +'
	  + ' event.srcElement.clientHeight - 2;\n'
	  + '    }\n'
	  + '  }\n'
	  + '  do_show_nav_menu();\n'
	  + '}\n'
	  + '//------------------------------------\n'
	  + 'function hide_nav_menu() {\n'
	  + '  if(!nav_menu.is_hilited) return;\n'
	  + '  nav_menu.nav_hide_timer = setTimeout("do_hide_nav_menu()",100);\n'
	  + '}\n'
	  + '//------------------------------------\n'
	  + 'function do_show_nav_menu() {\n'
	  + '  clearTimeout(nav_menu.nav_hide_timer);\n'
	  + '  if(selected_date.is_selected) do_unselect();\n'
	  + '  if(nav_menu.is_hilited) return;\n'
	  + '  var m  = document.getElementById("calendar_menu");\n'
	  + '  var mh = document.getElementById("calendar_menu_head");\n'
	  + '  m.style.display="block";\n'
	  + '  m.style.visibility="visible";\n'
	  + '  mh.style.backgroundColor="#FFFFDC";\n'	// light yellow
	  + '  nav_menu.is_hilited = true;\n'
	  + '}\n'
	  + '//------------------------------------\n'
	  + 'function do_hide_nav_menu() {\n'
	  + '  if(!nav_menu.is_hilited) return;\n'
	  + '  clearTimeout(nav_menu.nav_hide_timer);\n'
	  + '  var m  = document.getElementById("calendar_menu");\n'
	  + '  var mh = document.getElementById("calendar_menu_head");\n'
	  + '  m.style.display="none";\n'
	  + '  m.style.visibility="hidden";\n'
	  + '  mh.style.backgroundColor="transparent";\n'
	  + '  nav_menu.is_hilited = false;\n'
	  + '}\n'
	  + '</script>\n';
}

Calendar.prototype.getNavigationHTML = function() {

	// Don't render navigation buttons on print jobs.
	if (this.isPrinting()) return '';

	// Check if navigation bar previously defined by caller.
	if (this.NavigationHTML) return this.NavigationHTML;

	var html = '';
	html += '<form name="navigation"'
	  + ' class="navigation"\n'
	  + ' action="javascript:do_request(\'setyear\',document.navigation.year.value)"'
	  + '>\n';
	html += '<div id="navigation_toolbar" class="navigation">\n';
	html += '<table>\n';
	html += '<tr>\n';

	html += '<td width="20" class="navigation-menu">';
	html += ''
	  + '<a id="calendar_menu_head"'
	  + ' href="javascript:do_show_nav_menu()"'
	  + ' onMouseOver="show_nav_menu()"'
	  + ' onMouseOut="hide_nav_menu()"'
	  + '>'
	  + '&bull;'		// Bullet
	  + '</a>\n';
	html += '</td>\n';

	/*
	 * Left section
	 */

	html += '<td class="navigation-left">';

	html += ''
	  + '<a href="javascript:do_request(\'prevyear\')" title="Previous Year">&lt;&lt;<\/a>';

	if (this.n_months < 12) html += ''
	  + '<a href="javascript:do_request(\'prev\')" title="Previous Calendar">&lt;<\/a>';

	html += '</td>\n';

	/*
	 * Center section
	 */

	html += '<td class="navigation-center">';

	if (this.n_months < 3) {
		// Select months
		html += '<select name="month" class="navigation"'
		  + ' onChange="do_request(\'setmonth\',this[this.selectedIndex].value)">\n'
		  + ' <option value="1">January</option>\n'
		  + ' <option value="2">February</option>\n'
		  + ' <option value="3">March</option>\n'
		  + ' <option value="4">April</option>\n'
		  + ' <option value="5">May</option>\n'
		  + ' <option value="6">June</option>\n'
		  + ' <option value="7">July</option>\n'
		  + ' <option value="8">August</option>\n'
		  + ' <option value="9">September</option>\n'
		  + ' <option value="10">October</option>\n'
		  + ' <option value="11">November</option>\n'
		  + ' <option value="12">December</option>\n'
		  + '</select>\n'
		  + '<script language="JavaScript" type="text/javascript">'
		  + 'var fld=document.navigation.month;\n'
		  + 'fld.selectedIndex=-1;\n'
		  + 'for(var i=0; i<fld.length; i++){\n' 
		  + '  if (fld[i].value == "'+(this.date.getMonth()+1)+'"){fld[i].selected=true;break;}\n'
		  + '}\n' 
		  + '</script>\n';
	}
	else if (this.n_months < 6) {
		// Select quarters
		var month = this.date.getMonth();
		var quarter = 4;
		if (month < 9) quarter = 3;
		if (month < 6) quarter = 2;
		if (month < 3) quarter = 1;
		html += '<select name="quarter" class="navigation"'
		  + ' onChange="do_request(\'setquarter\',this[this.selectedIndex].value)">\n'
		  + ' <option value="1">1st Quarter</option>\n'
		  + ' <option value="2">2nd Quarter</option>\n'
		  + ' <option value="3">3rd Quarter</option>\n'
		  + ' <option value="4">4th Quarter</option>\n'
		  + '</select>\n'
		  + '<script language="JavaScript" type="text/javascript">'
		  + 'var fld=document.navigation.quarter;\n'
		  + 'fld.selectedIndex=-1;\n'
		  + 'for(var i=0; i<fld.length; i++){\n' 
		  + '  if (fld[i].value == '+quarter+'){fld[i].selected=true;break;}\n'
		  + '}\n' 
		  + '</script>\n';
	}
	else if (this.n_months < 12) {
		// Select half years
		var month = this.date.getMonth();
		var half = 2;
		if (month < 6) half = 1;
		html += '<select name="half" class="navigation"'
		  + ' onChange="do_request(\'sethalf\',this[this.selectedIndex].value)">\n'
		  + ' <option value="1">1st Half</option>\n'
		  + ' <option value="2">2nd Half</option>\n'
		  + '</select>\n'
		  + '<script language="JavaScript" type="text/javascript">'
		  + 'var fld=document.navigation.half;\n'
		  + 'fld.selectedIndex=-1;\n'
		  + 'for(var i=0; i<fld.length; i++){\n' 
		  + '  if (fld[i].value == '+half+'){fld[i].selected=true;break;}\n'
		  + '}\n' 
		  + '</script>\n';
	}

	html += ''
	  + '<input type="text" name="year"'
	  + ' value="' + this.date.getFullYear() + '"'
	  + ' class="navigation"'
	  + ' size="5"'
	  + '><a href="javascript:do_request(\'setyear\','
	  + 'document.navigation.year.value)"'
	  + ' id="go-link">Go</a>';
	  
	html += '</td>\n';

	/*
	 * Right section
	 */

	html += '<td class="navigation-right">';

	if (this.n_months < 12) html += ''
	  + '<a href="javascript:do_request(\'next\')" title="Next Calendar">&gt;<\/a>';

	html += ''
	  + '<a href="javascript:do_request(\'nextyear\')" title="Next Year">&gt;&gt;<\/a>';

	html += '</td>\n';

	html += '<td class="navigation-spacer" width="20">';
	html += '</td>\n';

	html += '</tr></table>\n</div>\n</form>\n';

	/*
	 * Calendar menu
	 */

	html += ''
	  + '<div id="calendar_menu" class="nav-menu"'
	  + ' onMouseOver="show_nav_menu()" onMouseOut="hide_nav_menu()">\n'
	  + '<a href="javascript:do_request(\'setstyle\',\'12\')">12 months</a>\n'
	  + '<a href="javascript:do_request(\'setstyle\',\'6\')">6 months</a>\n'
	  + '<a href="javascript:do_request(\'setstyle\',\'3\')">3 months</a>\n'
	  + '<a href="javascript:do_request(\'setstyle\',\'1\')">1 month</a>\n'
	  + this.getMenuHTML()
	  + '<hr>\n'
	  + '<a href="javascript:do_request(\'print\')">Print</a>\n'
	  + '<a href="javascript:self.close()">Exit</a>\n'
	  + '</div>\n';

	return html;
}

Calendar.prototype.getMenuHTML = function() {
	if (this.MenuHTML) return this.MenuHTML;
	return '';
}

Calendar.prototype.getMonthHTML = function(title) {
	var m = '';	// index for month class name
	if (this.n_months > 1) m = this.date.getMonth() + 1;
	return  '<div class="month'+m+'"><center>'
		+ (title ? '<span class="month-title">'+title+'</span>' : '')
		+ '<table class="month" cellspacing="0" cellpadding="0">\n'
		+ this.getWeeksHeaderHTML()
		+ this.getWeeksHTML()
		+ '</table>'
		+ '</center></div>';
}

Calendar.prototype.getWeeksHeaderHTML = function() {
	return '<tr class="week-header">\n'
	+ '<td class="day-header"><span class="day-title">Sun</span></td>\n'
	+ '<td class="day-header"><span class="day-title">Mon</span></td>\n'
	+ '<td class="day-header"><span class="day-title">Tue</span></td>\n'
	+ '<td class="day-header"><span class="day-title">Wed</span></td>\n'
	+ '<td class="day-header"><span class="day-title">Thu</span></td>\n'
	+ '<td class="day-header"><span class="day-title">Fri</span></td>\n'
	+ '<td class="day-header"><span class="day-title">Sat</span></td>\n'
	+ '</tr>\n';
}

Calendar.prototype.getWeeksHTML = function() {
	// Return HTML for all weeks in month.
	this.date.setDate(1);	// safety

	var i;
	var first_day=this.date.getDay();
	var last_date=this.getDaysInMonth();
	var html = '';

	// Write first week up to first day of month
	html += '<tr class="week">\n';
	if (first_day > 0) {
		// Write first week up to first day.
		this.addDate(-first_day);	// set calendar back for display
		for (i=0; i < first_day; i++) {
			html += '<td class="day-blank">'
			  + this.getBlankDateHTML()
			  + '</td>';
			this.addDate(1);
		}
	}

	// Write calendar
	for (i = 0; i < last_date; i++) {
		html += '<td class="day">'
		  + this.getDateHTML()
		  + '</td>\n';
		this.addDate(1);

		if ( ((first_day+i) % this.WeekColumns) == (this.WeekColumns-1) ) {
			html += "</tr>\n";
			html += '<tr class="week">\n';
		}
	}

	// Finish out week with blank days
	var remaining_days = this.WeekColumns - ((first_day+last_date) % this.WeekColumns);
	if (remaining_days > 0 && remaining_days < this.WeekColumns) {
		for (i = 0; i < remaining_days; i++) {
			html += '<td class="day-blank">'
			  + this.getBlankDateHTML()
			  + '</td>\n';
			this.addDate(1);
		}
		this.addDate(-remaining_days);	// reset date to 1st of next mo.
	}
	html += "</tr>\n";

	return html;
}

Calendar.prototype.getDateHTML = function() {
	var m   = this.date.getMonth() + 1;
	var d   = this.date.getDate();
	var yyyy = this.date.getFullYear();

	var mm = (m < 10) ? '0' + m : m;
	var dd = (d < 10) ? '0' + d : d;

	if (this.Type == 'full') {
		// Surround date button with info box
		return '<div class="' + this.getDayClass() + '"'
		  + ' onMouseOver="hilite_date(document.getElementById('
		  + '\'cal_'+mm+'-'+dd+'-'+yyyy+'\'))"'
		  + ' onMouseOut="unhilite_date(document.getElementById('
		  + '\'cal_'+mm+'-'+dd+'-'+yyyy+'\'))"'
		  + ' onClick="do_choose(document.getElementById('
		  + '\'cal_'+mm+'-'+dd+'-'+yyyy+'\'));event.cancelBubble=true"'
		  + '>'
		  + '<button id="cal_'+mm+'-'+dd+'-'+yyyy+'"'
		  + ' class="' + this.getDayClass() + '"'
		  + (this.DayButtonHelp
		    ? ' title="'+this.DayButtonHelp+'"'
		    : '')
		  + ' onMouseOver="schedule_autoclick()"'
		  + ' onMouseOut="unschedule_autoclick()"'
		  + ' onClick="do_callback(this)"'
		  + '><span class="button-title">' + d + '</span></button>'
		  + '</div>';
	}
	else {
		return '<button id="cal_'+mm+'-'+dd+'-'+yyyy+'"'
		  + ' class="' + this.getDayClass() + '"'
		  + ((this.DayButtonHelp && this.Type == 'list')
		    ? ' title="'+this.DayButtonHelp+'"'
		    : '')
		  + ' onMouseOver="hilite_date(this,event)"'
		  + ' onMouseOut="unhilite_date(this)"'
		  + ' onClick="do_callback(this,event)"'
		  + '><span class="button-title">' + d + '</span></button>';
	}
}

Calendar.prototype.getBlankDateHTML = function() {
	if (this.n_months > 1) return ' ';
	return '<div class="day-' + this.getBlankDayClass() + '">'
	  + '<div class="' + this.getBlankDayClass() + '">' + this.date.getDate() + '</div>'
	  + '</div>';
}

Calendar.prototype.getDayClass = function() {
	// Return class for days in calendar
	if (this.date.getTime() == this.today.getTime()) {
		if (this.isWeekday()) return "week-today";
		return "weekend-today";
	} else {
		if (this.isWeekday()) return "week-day";
		return "weekend-day";
	}
}

Calendar.prototype.getBlankDayClass = function() {
	// Return class for blank days in calendar
	if (this.isWeekday(this.date.getDay())) return "week-day-blank";
	return "weekend-day-blank";
}

/*
 * Utility methods
 */

Calendar.prototype.write = function(s) {
	this.win.document.write(s);
}

Calendar.prototype.writeln = function(s) {
	this.win.document.writeln(s);
}
