/*
 * Form Save 
 *
 * The functions below are for saving a form to the server. The form_is_setup()
 * function tracks whether forms have been completely downloaded from the
 * server. The form_is_dirty() tracks the forms that have been modified so
 * that form_cancel() can ask the user to whether to save the changes on the
 * form by using the confirm() and ask() functions.
 *
 * The form_verify() function looks for special images with ID attributes 
 * flagging which fields are required and makes sure those fields are filled
 * in before saving the form.
 *
 * The form_save() function is the primary function that should be called
 * just before submitting the form. It sets up a data structure that the
 * form_save_return() function uses to display the saved results on the
 * current form.
 *
 * Three functions form_reinitialize(), form_compute(), and form_events_setup()
 * are refered to below but are assumed to be defined elsewhere.
 *
 * Two form fields are used below: _request and _master. If the _request field
 * matches 'New', then the form stays open after it has been saved and the
 * form_reinitialize() function is called. This behavior can be forced by 
 * setting a global JavaScript variable "Keep_Open" to true. Normally, forms
 * that are for editing existing records of information exit when they are
 * saved.
 *
 * The form field _master is used to construct function names that are
 * called to post results to the master form and to return to the master
 * form when this (detail) form closes. The _master field holds the field
 * name (consisting of the corresponding database table name and column name)
 * of the field in the master form controlling the detail form -- the link
 * field. 
 */

// Define form to use by default if one is not specified below.
var Default_Form = 0;

var Forms_Setup = new Object;

function form_is_setup(f, cmd) {
	if (!f) return false;

	if (arguments.length > 1) {
		if (cmd) Forms_Setup[f.name] = true;
		else     Forms_Setup[f.name] = false;
	}

	if (Forms_Setup[f.name]) return true;
	else                     return false;
}

var Forms_Dirty = new Object;

function form_is_dirty(f, cmd) {
	if (!f) return false;

	if (arguments.length > 1) {
		if (cmd) Forms_Dirty[f.name] = true;
		else     Forms_Dirty[f.name] = false;
	}

	if (Forms_Dirty[f.name]) return true;
	else                     return false;
}

// Define persistant storage of arguments passed between form_save() and
// the form_save_return() functions. The arguments for each invocation are
// stored by the form's target name (the same as the dialog window's name).
var Form_Save_Args = new Object;

function form_save_setup_args(f) {

	// Define arguments object for form_save() and form_save_return().

	var obj =  {
		'save_and_continue': void 0,	// function(dialog_window)
		'post_to_master':    void 0,	// function(form,dialog_window)
		'return_to_master':  void 0,	// function called w/o arguments
		'window':            void 0,	// dialog window
		'form':              void 0};	// form object

	// Set routines to call to post results back to master form
	if (f && f._master != null && ! is_empty(f._master.value)) {
		obj['post_to_master']   = '_detail_post_'   + f._master.value;
		obj['return_to_master'] = '_detail_return_' + f._master.value;
	}

	return Form_Save_Args[f.target] = obj;
}

function form_save(f,msg,title,return_func) {
	if (typeof f != 'object') f = self.document.forms[f];
	if (f == null) { f = self.document.forms[Default_Form] }

	if (! form_is_setup(f)) {
		alert('This form has not been fully setup yet.\n' +
			'Wait until the form has been completely loaded before\n' +
			'saving. You might have to refresh your screen.');
		return false;
	}

	if (! f.target) {
		// Forms must have a target name. Use form name as a pattern.
		f.target = 'FS_' + f.name;
	}

	if (Form_Save_Args[f.target]) {
		// A form is already saving a record. (should seldom happen)

		if (f._request && f._request.value.match(/new/i)) {
			// Save only one form at a time as it messes up
			// re-initialization of forms entering new records.

			var confirm_msg =
			  'This form is already being saved.\n' +
			  'Please check your other windows before continuing.\n\n' +
			  'Continue saving this form anyway?';

			if (! confirm(confirm_msg)) return false;
		}
	}

	// Compute fields if a form_compute() function defined.
	if (self.form_compute && ! form_compute(f)) return false;

	// Make sure required fields are entered.
	if (! form_verify(f)) return false;

	// Security requires the argument form_name (a hidden input field)
	// to be set to the form name being saved.
	if (f._form_name && f._form_name.value) f._form_name.value = f.name;

	// Construct save message and title if not defined.
	if (msg   == null) msg   = 'Saving ' + f.name + '.';
	if (title == null) title = 'Save';

	// Construct save object of arguments for form_save_return().
	var args = form_save_setup_args(f);

	// Set function to return to if specified in arguments.
	if (return_func != null) args['save_and_continue'] = return_func;

	// Save the form object which is being submitted.
	args['form'] = f;

	// Open new dialog window for saving the form.
	args['window'] = new_message(msg, title, '', f.target);

	return true;
}

function form_save_return(dialog_window_name) {

	// This routine gets called from the dialog window that form_save()
	// opened to convey messages to the user. The name of the window
	// that form_save() opened (the dialog window) is used to reference
	// the arguments that form_save() saved for this routine to use.

	var args = Form_Save_Args[dialog_window_name];

	if (args['save_and_continue']) {
		var continue_func = args['save_and_continue'];
		var dialog_window = args['window'];
		Form_Save_Args[dialog_window_name] = null;	// clear
		if (continue_func) {
			continue_func(dialog_window);
		}
		return;
	}
		
	var mw = window.opener;	// get the master window that opened this form

	// Windows with names beginning with "Reconcile_" are reconciliation
	// windows that use to be message windows. Cascade window closures to
	// the original form window.

	if (window.name.match(/^Reconcile_/)) {
		window.name = window.name.replace(/^Reconcile_/,'');
		if (mw && ! mw.closed && mw.form_save_return) {
			mw.form_save_return(window.name);
		}
	}

	var f = args['form'];

	if (args['post_to_master'] && mw && ! mw.closed) {
		var post_func     = args['post_to_master'];
		var dialog_window = args['window'];
		if (eval('mw.'+post_func)) {
			eval('mw.'+post_func+'(f,dialog_window)');
		}
	}

	// We check page title for the first word of 'new' to detect posting
	// new master forms that had detail records entered while still entering
	// the master form. That caused the master form to be posted first before
	// entering the detail record so that the master key can be entered with
	// the detail records. However, that action causes the mode of the form
	// to change from "new" to "edit", so we check the page title to see if
	// that is still "new".

	var is_new_form = Boolean((f._request && f._request.value.match(/new/i)) || self.document.title.match(/^new /));

	var do_close_form = (typeof Keep_Open == 'boolean' && ! Keep_Open);

	if (is_new_form && ! do_close_form) {
		IDNode.clear();			// reset ID field
		NewNodes.clear();		// clear added html

		if (self.form_reinitialize) form_reinitialize(f);

		if (self.form_events_setup) form_events_setup();

		Form_Save_Args[dialog_window_name] = null;	// clear
		return;
	}
	else {
		if (self.Keep_Open) {
			// do nothing
			Form_Save_Args[dialog_window_name] = null;	// clear
			return;
		}
		else {
			if (args['return_to_master'] && mw && ! mw.closed) {
				var return_func = args['return_to_master'];
				if (eval('mw.'+return_func)) {
					eval('mw.'+return_func+'()');
				}
			}

			if (mw
			&& (typeof mw.auto_refresh == 'function'
			||  typeof mw.auto_refresh == 'object')) {	// for IE5.5
				// This does not work well if using POST method
				// to call up this window instead of GET. When
				// auto refresh is requested, the browser
				// displays an alert saying "The page cannot be
				// refreshed without resending the information.
				// blah blah blah"
				mw.auto_refresh();
			}

			Form_Save_Args[dialog_window_name] = null;	// clear

			self.close();
			return;
		}
	}
}

function form_verify(f) {
	if (f == null) { f = self.document.forms[Default_Form] }

	// Make sure required fields are entered.

	// The presence of an image which has an ID of "require_" with
	// the <input> field's name appended signals that the corresponding
	// <input> field is required. The "alt" parameter of the image
	// contains the field label used to formulate the alert message.
	// If the "alt" parameter is missing, the form field name is used
	// in the message.

	var n, l;
	for (var i=0; i < document.images.length; i++) {
		n = document.images[i].id;
		if (! n) continue;
		if (! (n = n.replace(/^require_/,''))) continue;
		if (f.elements[n] == null) continue;
		if (! is_empty(f.elements[n].value)) continue;

		l = document.images[i].alt;

		if (is_empty(l)) {
			// Make label from <input> field name
			l = n.replace(/^[A-Z\d]\d*_/,'');// strip special char
			l = l.replace(/.*__/,'');	// strip table name
			l = l.replace(/_/g,' ');	// change _ to spaces
			var l1 = l.substr(0,1);		// Capitalize
			l = l1.toUpperCase() + l.substr(1);
		}

		// This field is required and empty
		if (! confirm(
		  'The '+l+' field is empty and it should not be.\n\n' +
		  'Continue saving anyway?')) {
			f[n].focus();
			return false;
		}
	}

	return true;
}

function form_cancel(f) {
	if (f == null) { f = self.document.forms[Default_Form] }

	// This is executed when 'Cancel' button clicked or 'Esc' pressed.

	if (Form_Save_Args[f.target]) {
		// Cancel if in the middle of saving the form.
		// Something has gone wrong.
		self.close();
		return;
	}

	if (form_is_dirty(f)) {
		ask('Do you want to save your changes?',
		  ['<u>Y</u>es','<u>N</u>o','Cancel'],
		  ["self.form_save('"+f.name+"')&&self.document.forms['"+f.name+"'].submit()",
		   'self.close()',
		   ''],
		  ['Y','N',''],
		  'Save Changes?');
	}
	else {
		self.close();
	}
}

function ask(question, buttons, functions, keys, title, name, options, quiet) {

	// Print the given question with buttons appended.

	var ask_window_width  = 475;	// pixels
	var ask_window_height = 250;	// pixels

	// This was intended to be general purpose but the function definitions
	// in the functions array matching the buttons in the button array are
	// tricky because they are executed in the message window and apply
	// to the calling window. References to 'window' and 'self' in the 
	// functions array are replaced by references to the calling window.
	// Be careful using " quotes in functions[] definitions.

	if (! question) question = '';
	if (! title)    title    = 'Question';
	if (! name)     name     = 'Question';

	var settings = 'resizable,scrollbars' +
	  ',width='  + ask_window_width +
	  ',height=' + ask_window_height;
	settings = options ? options : settings;

	// Center new window
	var width  = window.document.body.offsetWidth;
	var height = window.document.body.offsetHeight;
	var top  = window.screenTop  + (height/2 - ask_window_height);
	var left = window.screenLeft + (width/2  - ask_window_width/2);
	settings += ',top='+top+',left='+left;

	var w = window.open('', name, settings);
	w.focus();

	var d = w.document;
	d.open();
	d.writeln('<html>');
	d.writeln('<head>');
	d.writeln('<title>' + title + '</title>');
	d.writeln('<link rel="stylesheet" href="/styles/message.css" type="text/css">');
	d.writeln('<script language="JavaScript" type="text/javascript">');
	d.writeln('window.document.onkeypress = do_keypress;');
	d.writeln('function do_keypress() {');
	d.writeln('switch (window.event.keyCode) {');
	d.writeln('case 27: window.close(); return true;');

	var k;
	for (var i = 0; i < keys.length; i++) {
		k = keys[i].charCodeAt(0);
		if (k >= 0) {
			d.writeln(
			  'case ' + (k+32) + ': ' +
			  ask_button_func(functions[i],'self.close();return true;'));
		}
	}

	d.writeln('} }');
	d.writeln('</script>');
	if (! quiet) d.writeln('<bgsound src="/sounds/chord.wav"></bgsound>');
	d.writeln('</head>');
	d.writeln('<body onLoad="self.focus();f=self.document.forms.length-1;self.document.forms[f].elements[0].focus()">');
	d.writeln('<table width="100%" border="1" cellspacing="0" cellpadding="10" align="center"><tr><td>');
	d.writeln('<div class="message">');
	d.writeln('<p>' + question + '</p>');
	d.writeln('<form>');
	d.writeln('<div class="button">');

	for (var i = 0; i < buttons.length; i++) {
		d.write(
		  ' <button type="button"' +
		  ' onClick="' + ask_button_func(functions[i],'self.close()') + '"');
		if (keys[i]) d.write(' accesskey="' + keys[i] + '"');
		d.write(' class="button"');
		d.writeln('>' + buttons[i] + '</button>');
	}

	d.writeln('</div>');
	d.writeln('</form>');
	d.writeln('</div>');
	d.writeln('</td></tr></table>');
	d.writeln('</body>');
	d.writeln('</html>');
	d.close();

	return w;
}

function ask_button_func(func,return_func) {
	// Alter button function.
	// This is a helper function for ask().
	if (func) {
		func = func.replace(/window\./g,'window.opener.');
		func = func.replace(/opener\.opener/g,'opener');
		func = func.replace(/self\./g,'window.opener.');
		func = func.replace(/"/g, '\\"');// escape " characters
		func += ';';
	}
	func += return_func;
	return func;
}

function is_empty(data) {
	var s = '' + data
	// All spaces is also considered to be empty.
	if (s.match(/^\s*$/)) return true;
	return false;
}

/*
 * NewNodes object
 *
 * The nodeTracker methods are used to track DOM nodes added to the document so
 * that they may be cleared before displaying the form for the next entry.
 */

var NewNodes = new nodeTracker;

function nodeTracker() {
	
	// Two parallel arrays are used to hold the added node object and its
	// parent node.

	this.nodes = new Array;
	this.parent_nodes = new Array;
}

nodeTracker.prototype.track = function(parent_node, new_node) {
	
	// Track the detail nodes
	this.parent_nodes[this.parent_nodes.length] = parent_node;
	this.nodes[this.nodes.length] = new_node;
}

nodeTracker.prototype.append = function(parent_node, new_node) {
	
	// Track the added nodes
	this.track(parent_node, new_node);
	parent_node.appendChild(new_node);
}

nodeTracker.prototype.replace = function(parent_node, new_node, old_node) {

	// Replace old node with a new one.
	if (parent_node && new_node) {
		for (var i = 0; i < this.parent_nodes.length; i++) {
			if (this.parent_nodes[i] == parent_node && this.nodes[i] == old_node) {
				this.parent_nodes[i].replaceChild(new_node,old_node);
				this.nodes[i] = new_node;
				return;
			}
		}

		// Did not find previously appended node.
		// Check document, Forms in Edit mode do not register
		// previously added records when displaying the form.

		parent_node.replaceChild(new_node,old_node);
	}
}

nodeTracker.prototype.clear = function() {

	// Remove nodes in reverse order, last node added is removed first.

	for (var i = this.parent_nodes.length - 1; i >= 0; i--) {
		if (this.parent_nodes[i] && this.nodes[i]) {
			this.parent_nodes[i].removeChild(this.nodes[i]);
		}
	}

	this.parent_nodes.length = this.nodes.length = 0;	// clear arrays
}

/*
 * IDNode object
 *
 * The idNodeTracker methods are used to track the DOM objects displaying
 * a record's ID number so that it may be reset to "Undefined" before
 * displaying the form for the next entry.
 */

var IDNode = new idNodeTracker;

function idNodeTracker() {
	
	// The replaced ID node is saved so that it can be reused when
	// the ID node is "cleared".

	this.node        = null;
	this.parent_node = null;
	this.saved_node  = null;
}

idNodeTracker.prototype.replace = function(parent_node, new_node, old_node) {

	if (parent_node) {
		parent_node.replaceChild(new_node,old_node);

		this.node        = new_node;
		this.parent_node = parent_node;
		this.saved_node  = old_node;
	}
}

idNodeTracker.prototype.clear = function() {

	if (this.parent_node && this.saved_node && this.node) {
		this.parent_node.replaceChild(this.saved_node,this.node);
	}

	this.node        = null;
	this.parent_node = null;
	this.saved_node  = null;
}
