/*
 * Parameter Form Utility Functions
 *
 * The functions below are for manipulating items in <select> lists.
 * They move selections from one <select> list to another or up and down
 * within a <select> list.
 *
 *	addSrcToDestList(srcList,destList,unique,suffix)
 *	deleteFromDestList(destList,srcList)
 *
 *	moveUpList(list)
 *	moveDownList(list)
 *
 *	selectList(list_id)
 *
 * Note: if using parameter screens, where the options in the left <select>
 * list are moved to the right <select> list, be sure to select all the items
 * in the <select> lists before submitting the form, otherwise the server
 * program won't see them. The selectList(list_id) function is for that purpose.
 */

/* ------------------------------------------------------------------------ */

/*
 * Move selected items in a <select> list up or down.
 *
 * All selected items in the list are moved up or down. If the selected
 * items are contiguous, they all move as a block.
 */

function moveUpList(list) {

	// Move all selected items in list up.

	// Note: Option objects are not moved within the list. Attributes
	// of the existing Option objects are changed.

	if (! list || ! list.options || list.options.length == 0)  return;

	var saved = new Object;

	var i,j,k;
        for (i = list.options.length - 1; i >= 0; i--) {

                if (list.options[i].selected) {

			// Find next unselected item.
			for (j = i-1; j >= 0; j--) {
				if (!list.options[j].selected) break;
			}

			if (j >= 0) {
				// Save current selection.
				saved.text            = list.options[j].text;
				saved.value           = list.options[j].value;
				saved.selected        = list.options[j].selected;
				saved.defaultSelected = list.options[j].defaultSelected;

				// Add the item above selection.
				for (k = j; k < i; k++) {
					list.options[k].text            = list.options[k+1].text;
					list.options[k].value           = list.options[k+1].value;
					list.options[k].selected        = list.options[k+1].selected;
					list.options[k].defaultSelected = list.options[k+1].defaultSelected;
				}

				list.options[i].text            = saved.text;
				list.options[i].value           = saved.value;
				list.options[i].selected        = saved.selected;
				list.options[i].defaultSelected = saved.defaultSelected;

				i = j;
			}
                }
        }
}

function moveDownList(list) {

	// Move all selected items in list down.

	// Note: Option objects are not moved within the list. Attributes
	// of the existing Option objects are changed.

	if (! list || ! list.options || list.options.length == 0)  return;

	var saved = new Object;

	var i,j,k;
        for (i = 0; i < list.options.length; i++) {

                if (list.options[i].selected) {

			// Find next unselected item.
			for (j = i+1; j < list.options.length; j++) {
				if (! list.options[j].selected) break;
			}

			if (j < list.options.length) {
				// Save current selection.
				saved.text            = list.options[j].text;
				saved.value           = list.options[j].value;
				saved.selected        = list.options[j].selected;
				saved.defaultSelected = list.options[j].defaultSelected;

				// Add the item above selection.
				for (k = j; k > i; k--) {
					list.options[k].text            = list.options[k-1].text;
					list.options[k].value           = list.options[k-1].value;
					list.options[k].selected        = list.options[k-1].selected;
					list.options[k].defaultSelected = list.options[k-1].defaultSelected;
				}

				list.options[i].text            = saved.text;
				list.options[i].value           = saved.value;
				list.options[i].selected        = saved.selected;
				list.options[i].defaultSelected = saved.defaultSelected;

				i = j;
			}
                }
        }
}

/* ------------------------------------------------------------------------ */

/*
 * Move selected items in a <select> list to another <select> list.
 *
 * All selected items in the list are moved to the destination list. A flag
 * "unique" can be set to true to avoid copying items that are already in
 * the destination list.
 *
 * A suffix can be added to the source text before adding it to the destination
 * list. It is stripped from the text before testing for uniqueness. The suffix
 * argument was added as a hack to accommodate adding "Sort Ascending" or
 * "Sort Descending" fields to a destination list.
 *
 */

/* The back and forth button scripts below are from:
 *	Pankaj Mittal (pankajm@writeme.com)
 *	http://www.fortunecity.com/lavendar/lavender/21
 * which were listed on:
 *	The JavaScript Source!!
 *	http://javascript.internet.com,
 * a wonderful place of ideas.
 */

function addSrcToDestList(srcList,destList,unique,suffix) {

	// Add the selected items from the source to destination list

	var found;	// flag
	var src_text, dest_text;	

	var len = destList.length;
	for(var i = 0; i < srcList.length; i++) {
		if ((srcList.options[i] != null) && (srcList.options[i].selected)) {
			found = false;
			if (unique) {
				//Check if this value already exist in the destList or not.
				//If not then add it, otherwise do not add it.
				src_text = srcList.options[i].text;

				for(var count = 0; count < len; count++) {
					if (destList.options[count] != null) {
						// Also match pattern like "column desc".
						dest_text = destList.options[count].text;
						if (suffix != null) dest_text = dest_text.replace(/(.*) [^ ]*$/,"$1");

						if (src_text == dest_text
						|| (suffix != null && (src_text == destList.options[count].text)) ) {
							found = true;
							break;
						}
					}
				}
			}

			if (!found) {
				srcList.options[i].selected = false; 
				src_text = srcList.options[i].text;
				if (suffix) src_text += " " + suffix;
				destList.options[len] = new Option(src_text,src_text); 
				destList.options[len].selected = true; 
				len++;
			 }
		}
	}
}

function deleteFromDestList(destList,srcList) {

	// Delete selected items in a <select> list.

	var opt = new Object();

	var len = destList.options.length;
	for(var i = (len-1); i >= 0; i--) {
		if ((destList.options[i] != null) && (destList.options[i].selected == true)) {
			if (srcList) opt[destList.options[i].value] = true;

			// To accommodate list items with an option word,
			// (e.g. sort desc) strip off last word of the option
			// value and set that value to true too.
			// Comment this out if you are using valid multi-word values.
			if (srcList) opt[destList.options[i].value.replace(/([^\s]+)\s+.+/,"$1")] = true;
			destList.options[i] = null;
		}
	}

	if (srcList && srcList.options) {
		// Select items in source list that were deleted.
		for (i = 0; i < srcList.options.length; i++) {
			if (opt[srcList.options[i].value]) {
				srcList.options[i].selected = true;
			}
			else {
				srcList.options[i].selected = false;
			}
		}
	}
}

/* ------------------------------------------------------------------------ */

function selectList(sourceList) {

	// Select all the items in a <select> list before submitting the form.

	for(var i = 0; i < sourceList.options.length; i++) {
		if (sourceList.options[i] != null) {
			sourceList.options[i].selected = true;
		}
	}
	return true;
}
