package Acme::Address;

=head1 NAME

Acme::Address - Address processing and formatting methods

=head1 SYNOPSIS

    use Acme::Address;

    my $a = new Acme::Address;

    if (! $a->is_empty) {
    	print $a->as_html;
    }

=head1 DESCRIPTION

This module is a collection of methods to parse an address and format them.

An C<Acme::Address> object can be constructed three ways. The first is by
passing a list of fields to new(). If the first argument is a hash, its
components are used for the address. Finally, and this is not implemented
yet, if there is only one scalar argument to new(), it is taken as an address
string and is parsed into the components.

The is_empty() method returns true if the address is considered to be empty.

The as_html() method formats an address.

The map_link() method returns a link that will display a map of the address's
location. Currently, MapQuest is the mapping service used.

=head1 CHANGE LOG

 8/25/02  Initial version.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1997-2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

use Acme::HTTP		qw(escape);

#
# Build hash of CSS classes that are recognized to display different type 
# of addresses.
#

our %css_classes = (
	      # Code		    CSS class
		ok 		=> 'address',
		limited		=> 'address-limited',
		private		=> 'address-private',
		secret		=> 'address-confidential',
		company		=> 'address-company',
		bad		=> 'address-bad',
		returned	=> 'address-returned',
		);

#
# Define the address components that are recognized.
#

# Note: The order of this list determines the order of the arguments to new()
# if passing it a list # of components.

our @components = (
		'name',
		'company',
		'address1',
		'address2',
		'city',
		'state',
		'zip',
		'province',
		'postal_code',
		'country',
		'locale',	      	# 'A' or 'I' for American or International
		'international_style',	# 'pre' or 'post' placement of postal_code
		'phones',
		'email',
		'usage_class',
		'usage'			# 'Y', 'N', or 'C' if conditional
		);

#------------------------------------------------------------------------------

sub new {
	my $proto = shift;	# caller object (should be class/package name)
	my $class = ref($proto) || $proto;

	my $self = { };
	bless $self, $class;

	if (ref $_[0] eq 'HASH') {
		# Copy option hash into $self.
		%$self = %{$_[0]};
	}
	elsif (@_ > 1) {
		# Change arguments @_ into component values of $self.
		%$self = map { $_, shift || '' } @components;
	}
	else {
		# Parse components from address
		$self->{address} = shift;
		$self->parse;
	}

	$self->{usage} ||= 'Y';		# Set default usage code

	return $self;
}

#------------------------------------------------------------------------------

sub parse {
	my $self = shift;

	########
	# TODO #
	########

	my @lines = split /\n/s, $self->{address};
	$self->{name} = $lines[0];

	return 0;	# parse failed
}

#------------------------------------------------------------------------------

sub is_empty {
	my $self = shift;

	return 0 if $self->{address1}
		 || $self->{address2}
		 || $self->{state}
		 || $self->{city};
	return 1;
}

#------------------------------------------------------------------------------

sub as_html {
	my $self = shift;

	my $address_class = $css_classes{ $self->{usage_class} };

	if (! $address_class) {
		# Look to usage codes to determine the display class
		$address_class = 'address-private' if $self->{usage} eq 'N';
		$address_class = 'address-limited' if $self->{usage} eq 'C';
		$address_class = 'address'	   if $self->{usage} eq 'Y';
	}

	# Set default values
	if (! $self->{locale}) {
		$self->{locale} = 'A';
		$self->{locale} = 'I'
		  if $self->{province} || $self->{postal_code};
	}
	$self->{international_style} ||= 'pre';	# only English Commonwealth is 'post'

	# Compose lines of address
	my @lines;
	push @lines, qq[<span class="address-name">$self->{name}</span>]
	  if $self->{name};
	push @lines, qq[<span class="address-name">$self->{company}</span>]
	  if $self->{company};
	push @lines, $self->{address1} if $self->{address1};
	push @lines, $self->{address2} if $self->{address2};

	if (uc($self->{locale}) eq 'I') {
		# International Addresses
		my $line = '';
		if ($self->{international_style} eq 'post') {
			$line .= $self->{province}
			  if $self->{province};
			$line .= qq[, &nbsp; ]
			  if $self->{province} && $self->{postal_code};
			$line .= qq[<span class="address-postal-code">$self->{postal_code}</span>]
			  if $self->{postal_code};
			push @lines, $line     if $line;
		}
		else {
			$line .= qq[<span class="address-postal-code">$self->{postal_code}</span>]
			  if $self->{postal_code};
			$line .= qq[ &nbsp; ]
			  if $self->{province} && $self->{postal_code};
			$line .= $self->{province}
			  if $self->{province};
			push @lines, $line    if $line;
		}
		push @lines, $self->{country} if $self->{country};
	}
	else {
		# American Addresses
		my $line = '';
		$line .= "$self->{city}, "
		  if $self->{city};
		$line .= "$self->{state} &nbsp; "
		  if $self->{state};
		$line .= qq[<span class="address-postal-code">$self->{zip}</span>]
		  if $self->{zip};
		push @lines, $line if $line;
	}

	push @lines, qq[<span class="address-phones">$self->{phones}</span>]
	  if $self->{phones};
	push @lines, qq[<span class="address-email">$self->{email}</span>]
	  if $self->{email};

	my $addr_sym = '';
	$addr_sym = qq[<div class="address-symbol">$self->{usage_class}</div>]
          if $self->{usage} ne 'Y';

	return 
	    qq[<div class="$address_class">\n]
	  . $addr_sym
	  . (join "<br>\n", @lines)
	  . qq[</div>\n];
}

#------------------------------------------------------------------------------

sub map_link {
	my ($self,$text) = @_;

	#
	# Create a link to a map of the given address.
	#

	$text ||= 'Map';	# default link text

	my $class = "small";	# CSS class of link
	my $title = "View a map showing address location.";	# button help

	my $n  = escape($self->{name});
	my $a  = escape($self->{address2} ? $self->{address2} : $self->{address1});
	my $c  = escape($self->{city});
	my $st = escape($self->{state});
	my $z  = escape($self->{zip});

	#
	# Create link to a mapping service.
	#

	# Use Yahoo maps.
	# Only American addresses accommodated.
	#my $href = "http://maps.yahoo.com/py/maps.py?name=$n&addr=$a&city=$c&state=$st&zip=$z";

	# Use MapQuest maps.
	# Only American addresses accommodated.
	my $href = "http://www.mapquest.com/maps/map.adp?size=big&address=$a&city=$c&state=$st&zip=$z&zoom=6";

	return
	    qq[<a]
	  . qq[ href="$href"]
	  . qq[ target="_blank"]
	  . qq[ class="$class"]
	  . qq[ title="$title"]
	  . qq[>$text</a>];
}

1;
