package Acme::Base;

=head1 NAME

Acme::Base - AcmeBase top level object

=head1 SYNOPSIS

    use Acme::Base;

    #
    # Schema interface
    #

    my $ab = new Acme::Base;

    my $schema = $ab->schema;
      -or-
    my $schema = Acme::Base->schema;
      -or-
    my $schema = new Acme::Base::Schema;

    #
    # SQL interface
    #

    my $sqlo = $ab->sql($sql_statement);

    #
    # Search interface
    #

    my $script = $ab->preprocess_search($user_search, {
	RecordSetName	   => "Customers",
	IDField		   => "customer_id",
	IDFieldName	   => "ID",
	CodeField	   => "user",
	CodeFieldName	   => "User",
	NameField	   => "name",
	NameFieldName      => "Name",
	PhoneField	   => "phone",
	PhoneFieldName     => "Phone",
	EmailField	   => "email",
	EmailFieldName     => "Email",
	FirstNameField	   => "first_name",
	FirstNameFieldName => "First Name",
	LastNameField	   => "last_name",
	LastNameFieldName  => "Last Name",
	});

    my $srch = $ab->searcher($script);

    my $combined_sql = $srch->as_sql($sqlo);

    #
    # Database interface
    #

    $q = $ab->query $script;
    @rec = $q->next;
      -or-
    $q = Acme::Base->query $script;
    @rec = $q->next;
      -or-
    $dbh = $ab->connect($connect_string, $user, $passwd);
    $dbh->query $script;
    @rec = $dbh->next;

    @first_rec = $ab->get $query;
      -or-
    @first_rec = Acme::Base->get $query;
      -or-
    $ab->get $query, \%result or die;

    #
    # Execute non-query SQL
    #

    $status = $ab->execute $script;
	or
    $dbh = $ab->connect($connect_string, $user, $passwd);
    $status = $dbh->execute $script;
	or
    $status = Acme::Base->execute $script;

    #
    # Utility methods
    #

    $s = $ab->escape_sql($s);

    ($join_expr, $join_clause) = $ab->parse_join($table, $join_clause);

=head1 DESCRIPTION

The C<Acme::Base> object is an umbrella for C<Acme::Base::Schema>,
C<Acme::Base::SQL>, and C<Acme::Base::Db>. C<Acme::Base::Schema> is
an object of database table and field descriptions. C<Acme::Base::SQL>
is an object holding a parsed SQL statement, and C<Acme::Base::Db> is a
database engine interface object. Each of these enclosed object modules
were developed independently, but the have been grouped together here
because of their inter-dependencies. The SQL standard is, in my opinion,
an overly complex standard that is inconsistently implemented by the
different database engines. Successful parsing of SQL statements cannot occur
without the knowledge of which database engine the SQL is written for. The
C<Acme::Base::Schema> object describes the database tables and fields that
C<Acme::Base::Db> methods access.

=head1 NOTE

This module is under construction. It's use has not been evaluated nor has
it been spread throughout the other modules.

=head1 CHANGE LOG

 8/17/02  Initial version.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1997-2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

# This object adds the following properties to the Acme::Base::Db object:
#	schema
#	sqlo
#	searcher


#------------------------------------------------------------------------------

sub new {
	my $proto = shift;	# caller object (should be class/package name)
	my $class = ref($proto) || $proto;

	require Acme::Base::Db;		# dynamically load database files

	our @ISA = ('Acme::Base::Db');

	my $self = new Acme::Base::Db @_;

	return bless $self, $class;
}

sub schema {
	my $self = shift;

	# Return the Acme::Base::Schema object.

	require Acme::Base::Schema;

	if (ref($self)) {
		# Object invocant
		return $self->{schema} if $self->{schema};
		return $self->{schema} = new Acme::Base::Schema;
	}
	else {
		# Class invocant
		return new Acme::Base::Schema;
	}
}

sub parse_join {
	my $self = shift;

	# Interface to Acme::Base::Schema::Join::parse_join.

	require Acme::Base::Schema::Join;

	return Acme::Base::Schema::Join::parse_join(@_);
}

sub sql {
	my $self = shift;
	my $sql  = shift;

	# Return object of the parsed sql statement.

	require Acme::Base::SQL;

	if (ref($self)) {
		# Object invocant
		return $self->{sqlo} = new Acme::Base::SQL $sql, dbengine=>$self->{dbengine}, @_;
	}
	else {
		# Class invocant
		return new Acme::Base::SQL $sql, @_;
	}
}

sub preprocess_search {
	my $self = shift;
	my $user_input  = shift;
	my $opt = shift || {};

	# Return preprocessed search script

	$opt->{dbengine} ||= $self->{dbengine};
	require Acme::Base::Searcher;
	return Acme::Base::Searcher::preprocess($user_input,$opt);
}

sub searcher {
	my $self = shift;
	my $script  = shift;
	my $opt = shift || {};

	# Return object of the parsed sql statement.

	require Acme::Base::Searcher;

	if (ref($self)) {
		# Object invocant
		$opt->{dbengine} ||= $self->{dbengine};
		return $self->{searcher} = new Acme::Base::Searcher $script, $opt, @_;
	}
	else {
		# Class invocant
		return new Acme::Base::Searcher $script, @_;
	}
}

sub db {
	my $self = shift;

	# Return database object.

	if (ref($self)) {
		# Object invocant
		return $self;
	}
	else {
		# Class invocant
		require Acme::Base::Db;		# dynamically load database files
		return new Acme::Base::Db @_;
	}
}

sub query {
	my $self = shift;

	# Issue a query to the database.

	if (ref($self)) {
		# Object invocant
		 return $self->SUPER::query(@_)
	}
	else {
		# Class invocant
		require Acme::Base::Db;		# dynamically load database files
		return Acme::Base::Db->query(@_);
	}
}

sub get {
	my $self = shift;

	# Get a single record from database.

	if (ref($self)) {
		# Object invocant
		 return $self->SUPER::get(@_)
	}
	else {
		# Class invocant
		require Acme::Base::Db;		# dynamically load database files
		return Acme::Base::Db->get(@_);
	}
}

sub execute {
	my $self = shift;

	# Execute SQL statement.

	if (ref($self)) {
		# Object invocant
		 return $self->SUPER::execute(@_)
	}
	else {
		# Class invocant
		require Acme::Base::Db;		# dynamically load database files
		return Acme::Base::Db->execute(@_);
	}
}

1;
