package Acme::Base::Object;

=head1 NAME

Acme::Base::Object - Object objects

=head1 SYNOPSIS

    use Acme::Base::Object;

    use Acme::Example::Defaults;
    my $obj = new Acme::Base::Object
        '.property_defaults'	=> 'Acme::Example::Defaults',	# module name
        '.property_filter'	=> \&yesno,			# code ref
        '.type'			=> 'Table';			# text string

    for (@mycomponents) { $obj->add($_) }	# add component to obj

    $obj->set(myproperty=>$value);		# set object property

    my $component = $obj->{component_name};	# get component value
    my $parameter = $obj->{.parent};		# get parameter setting
    my $property  = $obj->myproperty;		# get property setting

    # Process components in order they were defined (e.g. add()'ed).
    for $component ($obj->components) {
        print "$component=".$obj->{$component};
    }

    # Process properties in alphabetical order.
    for $property ($obj->properties) {
        print "$property=".$obj->property($property);
          -or-
        print "$property=",$obj->$property;
    }

=head1 DESCRIPTION

The I<Acme::Base::Object> object contains components and parameters, and one
parameter points to the object's properties. Three keywords ...components
...parameters ...properties.

The I<Acme::Base::Object> object contains components (normal names) and
parameters (names beginning with periods), and the parameter named
'.properties' points to a I<Acme::Base::Object::Properties> object.

The I<Acme::Base::Object::Properties> object is used to hold object properties.
Property names are one word, all lowercase. Default values for properties are
supplied by methods in a module specified by the '.property_defaults' parameter
when the object is constructed. The system automatically adds another parameter
'.got_property_defaults_module' to the object as a performance enhancing flag.

Every I<Acme::Base::Object> object has a I<Acme::Base::Object::Properties> object,
and every properties object has a property named 'object' which points
to the parent object. It also has another property named 'components',
an array, which lists the components added with the add() method in the
order they were added.

The difference between parameters and properties is the parameters
hold information about the object such as links to other nodes, type,
and so forth, whereas properties hold descriptive information pertinent
to the data in the object. Parameters are static, properties can be
determined dynamically. Parameters are referred to as hash components 
whereas properties are retrieved by method calls.

Standard parameters of every object are:
    .properties -- reference to a I<Acme::Base::Object::Properties> object

Standard properties of every object are:
    components  -- list of components added to parent object Acme::Base::Object
    object      -- I<Acme::Base::Object::Properties>' parent object

An optional property '.property_filter' contains a reference to a function that
is used to filter the propterty setting before setting it. An example of this
use is the yesno() function which is used to set a property to "true" if it
exists. Normally, a property would be "false" because it would have no value.
The yesno() function provides a value (yes) if the property is defined but does
not have a defined value.

=head1 CHANGE LOG

 7/18/00  Initial version.
 4/28/01  Refined usage && terminology.
 8/8/01   Added n() method returning number of components.
 11/14/01 Added '.got_property_defaults_module' parameter flag for speed.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000, 2001 Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

#------------------------------------------------------------------------------

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my $self = { @_ };
	$self->{'.properties'} = new Acme::Base::Object::Properties
					components => [],
					object => $self;

	return bless $self, $class;
}

sub add {
	my $self = shift;

	# Add component(s).
	my ($k, $v);
	while ($k = shift) {
		$v = shift;
		$self->{$k} = $v;

		# Track the order components are added in the 'components'
		# property array.
		push @{$self->{'.properties'}->{components}}, $k;
	}
}

sub n {
	my $self = shift;

	# Number of components.
	return scalar @{$self->{'.properties'}->{components}};
}

sub set {
	my $self = shift;

	# Set property.
	$self->{'.properties'}->set(@_);
}

sub get {
	my $self = shift;

	# Get property setting.
	return $self->{'.properties'}->get(@_);
}

sub defined {
	my $self = shift;
	my $prop = shift;
	# Return true if property has been explicitly set.
	return $self->{'.properties'}->defined($prop);
}

sub properties {
	my $self = shift;

	# List properties in object.
	return sort keys %{$self->{'.properties'}};
}

sub DESTROY {}

sub AUTOLOAD {
	my $self = shift;
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;

	# Get parameter setting if defined.
	# Uncomment the following you want to refer to parameter setting
	# (not property setting) by method name.
	#return $self->{".$method"}
	#  if defined $self->{".$method"};

	# Get property setting.
	return $self->{'.properties'}->$method(@_)
	  if defined $self->{'.properties'};

	return;
}

#------------------------------------------------------------------------------

package Acme::Base::Object::Properties;

# Property names are lowercase.
# Default values and translations are used if a property does not exist.

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my $self = {};
	set($self,@_) if @_;
	return bless $self, $class;
}

sub set {
	my $self = shift;

	# Set new properties.
	my ($k, $v);
	while ($k = shift) {
		$v = shift;
		($k,$v) = Acme::Base::Object::Properties::Private::mung($self,$k,$v);
		$self->{$k} = $v;
	}
}

sub append {
	my $self = shift;

	# Add new properties.
	# Append properties if they already exist.
	my ($k, $v);
	while ($k = shift) {
		$v = shift;
		($k,$v) = Acme::Base::Object::Properties::Private::mung($self,$k,$v);
		if (not exists $self->{$k}) {
			$self->{$k} = $v;
		}
		elsif (! ref $self->{$k}) {
			# Add a newline and value to existing property.
			$self->{$k} .= "\n$v";
		}
		elsif (ref $self->{$k} eq 'SCALAR') {
			# Add a newline and value to existing property.
			${$self->{$k}} .= "\n$v";
		}
		elsif (ref $self->{$k} eq 'ARRAY') {
			# Add to end of array
			push @{$self->{$k}}, $v;
		}
		elsif (ref $self->{$k} eq 'HASH') {
			# Add value to hash keys
			$self->{$k}->{$v} = 1;
		}
	}
}

sub defined {
	my ($self,$k) = @_;

	# Return true if property has been explicitly set.
	return exists $self->{$k};
}

sub get {
	my ($self,$k) = @_;
	$k = Acme::Base::Object::Properties::Private::mung($self, $k);
	return $self->{$k} if defined $self->{$k};
	return Acme::Base::Object::Properties::Private::mash($self, $k);
}

sub DESTROY {}

sub AUTOLOAD {
	my $self = shift;
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;

	return $self->get($method);
}

#------------------------------------------------------------------------------

package Acme::Base::Object::Properties::Private;

# Private routines for Acme::Base::Object::Properties.
# Stored here so they don't pollute its namespace.

sub mung {
	my ($obj, $k, $v) = @_;

	#
	# Translate key/value pairs into format stored in
	# Acme::Base::Object::Property object.
	# Settings are mung'ed into internal storage format here.
	#

	$k = lc($k);		# force property name to lowercase

	return $k unless wantarray;

	#
	# Perform translations of property values before setting them.
	#

	# The object can contain a property named '.property_filter'
	# which identifies a routine to use to filter property settings
	# before setting them.

	if (defined $obj->{object}
	&&  defined $obj->{object}->{'.property_filter'}) {
		my $filter = $obj->{object}->{'.property_filter'};
		$v = &$filter($k,$v);
	}

	return ($k, $v);
}

sub mash {
	my ($obj, $k) = @_;

	# Substitute property's key value into a value used by programs.
	# Default values are used for undefined keys.
	# The object contains a parameter '.property_defaults' that holds the
	# module name used to compute default values.
	# Settings are mash'ed into a format usable by the program here.

	my @result;
	if (defined $obj->{object}->{'.property_defaults'}) {
		my $module = $obj->{object}->{'.property_defaults'};

		if (! $obj->{'.got_property_defaults_module'}) {
			eval "require $module";
			die $@ if $@;
			$obj->{'.got_property_defaults_module'} = 1;	# set flag
		}

		my @result = $module->$k($obj);
		if (wantarray)	{ return @result }
		else		{ return $result[0] }
	}
	# No default value for key.
	return;
}

1;
