package Acme::Base::Schema;

=head1 NAME

Acme::Base::Schema - Perl module interface database field definitions

=head1 SYNOPSIS

    use Acme::Base::Schema;

    my $schema = new Acme::Base::Schema;             # use default database
      -or-
    my $schema = new Acme::Base::Schema 'table1', 'table2', ...;
      -or-
    my $schema = new Acme::Base::Schema 'database';  # specify database
      -or-
    my $schema = new Acme::Base::Schema 'database' 'table1', 'table2', ...;
      -or-
    my $schema = new Acme::Base::Schema { dbname => 'database', file => 'filename' }
    
    # Field properties
    my $fieldtype = $schema->{$tablename}->{$fieldname}->type;
      -or-
    my $field_type = $schema->t($tablename)->f($fieldname)->p('type');
    
    # Table properties
    my $tabledesc = $schema->{$tablename}->description;

    @fields = $schema->{$tablename}->fields;

    @indexes = $schema->{$tablename}->indexes;
    
    # List table's fields that can be accessed by the user.
    # This is the same list that is returned by $schema->{$tablename}->fields.
    @fields = $schema->{$tablename}->fields_secure;

    # List table's fields that can be 'inserted', 'updated', or 'deleted'
    @fields = $schema->${tablename}->fields_secure('Insert,Update,Delete');

    # List all the table's fields.
    @fields = $schema->{$tablename}->fields_all;

    # Schema properties
    my $dbname = $schema->dbname;

    # Test if a field node is a display field.
    if ($schema->{tablename}->{fieldname}->is_display_field) { ... }

    # Preferred way to access Table properties
    # This allows dynamic loading of table information when initially accessed.
    my $t = $schema->t($tablename);
    my $tabledesc = $t->description;

=head1 DESCRIPTION

Acme::Base::Schema provides a program interface to database information,
the table names, field names, types, descriptions, and so on.

Creating a new schema object reads the database schema description file
(schema.txt) and parses it. The schema description file can be specified in
the argument hash { file => 'filename' }, the first argument to new().

The schema description file is a structured file of key value pairs.  Lines are
white-space delimited. They contain two fields: the first word is a key name,
also referred to as an object name or property name. Everything else on the
line is the value.  Beginning white-space on a line is ignored.

"Table", "Index", "Field" and "Display" are key words (case insensitive).
They identify the current table, index, field, or display field that the
following description lines apply to.

A description line is continued on the next line if the line is indented.
Tabs in indented lines are equivalent to 8 spaces.  Lines can also be
continued onto the next by ending the line with a backslash (\).

Comments begin with # and extend to the end of the line. Empty lines are ignored.

Here documents beginning with <<WORD are read verbatum until a line containing
only WORD is read.

Garbage is either processed as a weird property or ignored.

Very few error messages are given. However, debug output can be turned on
by setting

	$Acme::Base::Schema::Debug = 1;

The caching of the schema files can be checked by setting

	$Acme::Base::Schema::Debug_Cache = 1;

Note: Tables are referred to by name, fields are referred to by name, but
the name of indexes is the string of column names to be indexed. For
example, in the schema.txt file:

        Table CUSTOMERS
        Field FIRST_NAME
        Type  varchar(50)
        Field LAST_NAME
        Type  varchar(50)

        Index LAST_NAME, FIRST_NAME
        Type  unique

the field LAST_NAME Type properties are obtained by:

        $schema->{CUSTOMERS}->{LAST_NAME}->type

whereas the index's properties are obtained by:

        $schema->{CUSTOMERS}->{'LAST_NAME, FIRST_NAME'}->type;

=head2 Methods, Objects and Properties

Table definitions are loaded dynamically when first used. Therefore,
the first time a table component is accessed, it should use the table()
or t() method:

        $schema->t('CUSTOMERS');

The schema object reference is stored as a table parameter.  Likewise,
the table object is stored as a field parameter.  Therefore, to refer to a
schema property 'dbname' from a table object $t, use:

        $t->schema->dbname;

or to find the schema object from a field object $f, use:

        $schema = $f->schema;

A Schema object contains a collection of Table objects, each of which
contains a collection of Field objects. The Schema, Table, and Field objects
each have a set of properties. Properties are determined dynamically if not
explicitly specified.

             ----> Schema ----> Table ----> Field
                     |            |           |
                     V            V           V
                 Properties   Properties  Properties

Properties are dynamic so they are referenced by using object methods.
Table and Field objects can also be referenced by methods. The table objects
are dynamically built from the schema files only when needed (or if they are
specified in the new() constructor), so Table object names need to be specified
by a method name the first time they are used (unless they were specified in
the new() constructor).
In other words, both property settings and object components can be specified 
by using method names. Where confusion arises, Table object names take
precedence over Schema Property names. The number of Schema Properties are
few and Table objects need to be dynamically loaded. However, Table Property
names take precedence over Field object names. There are many more Table
Property names and it is perceived that when writing code, it is Table Property
names that are intended.

To reduce confusion, the t(), f(), and p() methods have been introduced.
The t() method specifies a table name, f() specifies a field name, and
p() specifies a property name. They are intended for use in following the
object chain from left to right (---->) like:

    $s->t($tbl)->f($fld) or $s->t($tbl)->p($prop)

Usually the p() method is used only with the schema object because property
names take precedence over table or field names. However, table names take
precedence over schema property names because table definitions are usually
dynamically loaded the first time they are referenced.

To provide easy access to related objects, all objects have the methods
schema(), table() and field(). The methods return references to the
controlling object, themselves, or (if a name is specified) a subordinate
object. (The t() and f() methods are actually synonyms for table() and
field() in the Schema and Table objects.) They are intended to follow the
object chain from right to left (<----) like

    $f->schema->t($tbl)

which could also be written as

    $f->schema->table($tbl)

The property() and p() methods are synonyms for the get() method inherited
from Acme::Base::Object.

Display fields masquerade as normal fields. To tell the difference between
the two, both the Display and Field objects have a method is_display_field()
which returns true or false.

=head2 Reserved Names

The method names for Schema, Table, and Field objects models the same
namespace as SQL field and table names. Therefore, the following names are
added to the normal SQL reserved names:

	schema
	table and tables
	index and indexes
        field and fields
	display and displays
	elem  and elems
	property
	t and f and p

	new
	add
	add_dir
	add_table
	add_index
	add_field
	add_display
	add_elem
        tables_secure
	tables_all
        fields_secure
	fields_all
	debug

	n
	set
	get
	defined
	properties

	select
	menu
	list

The I<Acme::Base::Object> methods 'n', 'set', 'get', 'defined', and 'properties'
are also reserved as Schema, Table, and Field objects belong to the
Acme::Base::Object class.

The way the Choose and Search schema options are implemented, table names
are recognized as table names only if they do not match the sub options
of 'select', 'menu',  or 'list', so they are reserved table names too.

=head1 CHANGE LOG

 3/18/00  Initial version.
 4/20/00  Cached schema hash to avoid re-reading schema file over and over.
 4/28/00  Added $table->fields_secure() method to filter list of fields 
          according to user's privileges.
 4/30/00  Added master schema file with other schema files loaded dynamically.
 5/29/00  Broke out default value processing to Acme::Base::Schema::TableDefault
          and Acme::Base::Schema::FieldDefault modules
 3/17/01  Added $table->fields_all() and made $table->fields() same
          as $table->fields_secure().
 4/6/01   Change $table->fields() to exclude private fields. Added extra flag
          argument to fields_secure() to force including private fields.
 4/26/01  Major rewrite to use Acme::Base::Object.
 5/2/01   Added Elem objects for HTML and form directives.
 5/6/01   Added add_dir() so a directory can be used instead of a master
          schema file. Also added Display fields.
 6/13/01  Added ability to pre-load table definitions if specified in new().
 7/28/01  Changed arguments so first one can be a hash of options specifying
          dbname and file name.
 8/23/01  Added 'here' document processing in schema files.
 3/26/02  Added hooks, hacks and cache control for Acme::Form::Schema files
          which are essentially "views" of schema files for building forms.
 9/29/02  Added tables_secure(), tables_all(); tables() now is tables_secure().

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000,2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

use Acme::Base::Object;
use Acme::Config;

#------------------------------------------------------------------------------

our @ISA = qw(Acme::Base::Object);

our $Debug=0;
sub debug { print "<br>  ", @_, "\n" if $Debug }

our $Debug_Cache=0;
sub debug_cache { print "<br> --- ", @_, "\n" if $Debug_Cache }

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my ($dbname, $dbengine, $filename, $cachename);
	if (defined $_[0]) {
		if (ref $_[0] eq 'HASH') {
			# Decode options hash
			my $opt = shift;

			$dbname    = $opt->{dbname} || Acme::Config->DBNAME;
			$filename  = $opt->{file}   || Acme::Config->DBSCHEMA($_[0]);
			$cachename = $opt->{file}   || $dbname;
		}
		elsif ($filename = Acme::Config->Databases->{$_[0]}) {
			$dbname = shift; # first argument is name of database
			$cachename = $dbname;
		}
	}

	$dbname    ||= Acme::Config->DBNAME;	# use default db name
	$filename  ||= Acme::Config->DBSCHEMA($dbname) || 'Schema.txt';
	$cachename ||= $dbname;

	$dbengine = Acme::Config->DBENGINE($dbname);

	debug("\nSchema: dbname='$dbname', dbengine='$dbengine', filename='$filename'");

	my $self = Acme::Base::Schema::Cache->get($cachename);
	$self->{cachename}= $cachename;	# store cache name for DESTROY

	bless $self, $class;

	debug_cache("new Acme::Base::Schema: "
	  . "dbname='$dbname', filename='$filename', cachename='$cachename'"
	  . ", caller=".(join ", ", (caller)));

	if ($self->{'.last_checked'} > 0) {
		# Schema file already in cache.

		# Return if last time checked was less than 60 seconds
		# (Refresh_Interval) or if file has not been modified
		# since last read.

		if (time < ($self->{'.last_checked'} + Acme::Config->Refresh_Interval)) {
			# Dynamically add table definitions for remaining args.
			for my $tbl (@_) { $self->table($tbl) }

			debug_cache("new Acme::Base::Schema:"
			  . " RETURNING cached self (not yet time to check)");

			return $self;
		}

		$self->{'.last_checked'} = time;

		if ((stat $filename)[9] < $self->{'.timestamp'}) {
			# Dynamically add table definitions for remaining args.
			for my $tbl (@_) { $self->table($tbl) }

			debug_cache("new Acme::Base::Schema:"
			  . " RETURNING cached self (file not changed)");

			return $self;
		}

		# Re-initialize schema object and re-read schema file.
		$self = new Acme::Base::Schema::Cache $dbname;
		bless $self, $class;
	}

	# Set default directory where to look for schema definition files.
	# Use the same directory as where the master schema file was found.
	my $filedir;
	if (-d $filename) {
		# Master schema file is a directory, not a file.
		$filedir = $filename;
		$filename = '';
	}
	else {
		# Strip directory from master schema file name.
		$filename =~ s|(.*)/||;
		$filedir = $1 || '.';
	}

	$self->set('dbname'   => $dbname,	# database name
		   'dbengine' => $dbengine,	# database engine type
		   'filename' => $filename,	# main schema def file
		   'filedir'  => $filedir,	# location of schemas
		   'tables'   => [],		# list of tables defined
		   'files'    => {},		# files already read
		   );

	if ($filename)	{ $self->add($filename)    }
	else		{ $self->add_dir($filedir) }

	# Dynamically add table definitions for remaining arguments.
	for my $tbl (@_) { $self->table($tbl) }

	# Set flag when file last checked
	$self->{'.last_checked'} = time;

	debug_cache("new Acme::Base::Schema: RETURNING (Re)INITIALIZED self");

	return $self;
}

sub add {
	my $self = shift;
	my $class = ref($self) or return $self->new(@_);
	my ($filename) = @_;

	return $self if $self->files->{$filename};	# already read

	#
	# Read/Process schema description file
	#

	my ($prefix, $data, $key, $value, $here);
	my $curr_table   = '';	# current table name
	my $curr_field   = '';	# current field name
	my $curr_index   = '';	# current index name
	my $curr_display = '';	# current display field name
	my $curr_elem    = '';	# current elem command
	my $curr_prop    = '';	# current property name
	my $curr_indent  = 0;	# current indentation
	my ($table_file);

	my $f = $filename;
	$f = $self->filedir .'/' . $filename
	  if $filename !~ m|^/|;

	debug_cache("reading schema file '$f'");

	open(SCHEMA, "<$f") or die "Cannot open schema file '$filename'",join ", ", caller(1)," | ",caller(2), " | ", caller(3);

	# Set the "file read" flag.
	$self->files->{$filename} = 1;

	debug("\n\nReading schema file: $filename\n\n");

	local $/ = "\n";		# set input line separator.
	while (<SCHEMA>) {		# read line
		chomp;

		last if /^\s*__END__\b/;

		# Ignore comment lines.
		next if /^\s*#/;	

		# Continue to nextline if last chararacter on the line is \.
		if (s/\\$//) { $_ .= <SCHEMA>; redo }

		# Strip comments from end of line.
		s/([^\\])#.*/$1/;	

		# Unescape escaped # comment delimiter.
		s/[\\]#/#/g;

		# Strip trailing white space.
		s/\s+$//;		

		# Continue to nextline if last chararacter on the line is \.
		# This test is repeated after a comment at the end is stripped.
		if (s/\\$//) { $_ .= <SCHEMA>; redo }

		# Strip white space at the beginning of the line.
		($prefix, $data) = m/^(\s*)(.*)/;

		# Ignore empty lines.
		next unless $data;	

		# Expand tabs to 8 spaces before counting spaces of indentation
		1 while $prefix =~ s/\t+/' 'x(length($&)*8-length($`)%8)/e;

		# Split line into key value pair.
		($key, $value) = $data =~ m/([^\s]+)\s*(.*)/;

		# Strip possible quotes from value.
		$value =~ s/^(["'])(.*)\1$/$2/;

		#
		# Check if value comes from a here document (<<WORD)
		#

		if ($value =~ /^<<(.*)/) {
			$here = $1;
			$value = '';
			# Read file until a line with only contents of $here.
			while (my $v = <SCHEMA>) {
				chomp $v;
				last if $v eq $here;
				$value .= "$v\n";
			}
			chomp $value;
		}

		debug("\n>>data='$data', key='$key', value='$value' ");

		#
		# Check for keywords
		#

		if ($key =~ /^table$/i) {

			#
			# Keyword 'table'
			#

			($curr_table, $table_file) = split /\s+/, $value;
			$curr_field = $curr_index = $curr_display = $curr_elem = $curr_prop = '';
			$curr_indent = length($prefix);

			# Strip possible quotes from table_file.
			$table_file =~ s/^(["'])(.*)\1$/$2/ if $table_file;

			debug("keyword 'table': setting curr_table='$curr_table', curr_indent='$curr_indent'");

			$self->add_table(name => $curr_table, file => $table_file)
			  unless $self->{$curr_table};

			next;
		}

		# We can ignore all data lines until a Table keyword, or we can
		# attach all settings to the Schema object by commenting out the
		# next line.
		#next unless $curr_table;	# ignore all until table spec

		if ($key =~ /^index$/i) {

			#
			# Keyword 'index'
			#

			$curr_index = $value;
			$curr_field = $curr_display = $curr_elem = $curr_prop = '';
			$curr_indent = length($prefix);

			debug("keyword 'index': curr_table='$curr_table', index='$curr_index'");

			$self->{$curr_table}->add_index(value => $curr_index);
			next;
		}

		if ($key =~ /^field$/i) {

			#
			# Keyword 'field'
			#

			$curr_field = $value;
			$curr_index = $curr_display = $curr_elem = $curr_prop = '';
			$curr_indent = length($prefix);

			debug("keyword 'field': setting curr_field='$curr_field', curr_indent='$curr_indent'");

			$self->{$curr_table}->add_field(name => $curr_field)
			  unless $self->{$curr_table}->{$curr_field};
			next;
		}

		if ($key =~ /^display$/i) {

			#
			# Keyword 'display'
			#

			$curr_display = $value;
			$curr_field = $curr_index = $curr_elem = $curr_prop = '';
			$curr_indent = length($prefix);

			debug("keyword 'display': curr_table='$curr_table', display='$curr_display'");

			$self->{$curr_table}->add_display(value => $curr_display);
			next;
		}

		if ($data =~ /^</) {

			#
			# HTML text
			#

			#
			# Check if this is the beginning of a here document (<<WORD)
			#

			if ($data =~ /^<<(.*)/) {
				$here = $1;
				$data = '';
				# Read file until a line with only contents of $here.
				while (my $d = <SCHEMA>) {
					chomp $d;
					last if $d eq $here;
					$data .= "$d\n";
				}
				chomp $data;
			}

			$curr_field = $curr_display = $curr_index = '';
			$curr_indent = length($prefix);

			# Note that html $data will not display correctly if
			# debugging in a browser window. It is html code.

			debug("keyword '<': curr_table='$curr_table', elem='$data'");

			# There is not a name or optional property settings
			# for HTML text. The element name is created when
			# added and used as the current property name so that
			# the next line is appended if indented.

			$curr_prop = $curr_elem = $self->{$curr_table}->add_elem(
								type  => 'html',
								value => $data
								);
			next;
		}

		#
		# Check if property setting is continued onto this line
		#
		# Note: as the schema files become more and more extensive,
		# I debate moving this section before the above section which
		# checks for keywords, effectivily requiring keywords to
		# be in the far left column in order to be recognized, but
		# ignoring keywords which might appear in properties continued
		# on following lines. (mp 9/20/01)
		#

		if ($curr_prop and length($prefix) > $curr_indent) {
			if ($curr_field) {
				# Append field property.
				debug("  append field '$curr_table.$curr_field' property '$curr_prop' = '$data'");
				$self->{$curr_table}->{$curr_field}->append($curr_prop => $data);
			}
			elsif ($curr_display) {
				# Append display property.
				debug("  append display '$curr_display' property '$curr_prop' = '$data'");
				$self->{$curr_table}->display->{$curr_display}->append($curr_prop => $data);
			}
			elsif ($curr_elem) {
				# Append elem property.
				debug("  append elem '$curr_elem' property '$curr_prop' = '$data'");
				$self->{$curr_table}->elem->{$curr_elem}->append($data);
			}
			elsif ($curr_index) {
				# Append index property.
				debug("  append index '$curr_index' property '$curr_prop' = '$data'");
				$self->{$curr_table}->index->{$curr_index}->append($curr_prop => $data);
			}
			else {
				# Append table property.
				debug("  append table '$curr_table' property '$curr_prop' = '$data'");
				$self->{$curr_table}->append($curr_prop => $data);
			}

			next;
		}

		#
		# Set property
		#

		if ($curr_field) {
			# Set field property
			debug("set field '$curr_table.$curr_field' property '$key' = '$value'");
			$self->{$curr_table}->{$curr_field}->set($key => $value);
		}
		elsif ($curr_display) {
			# Set display property
			debug("set display '$curr_display' property '$key' = '$value'");
			$self->{$curr_table}->display->{$curr_display}->set($key => $value);
		}
		elsif ($curr_index) {
			# Set index property
			debug("set index '$curr_index' property '$key' = '$value'");
			$self->{$curr_table}->index->{$curr_index}->set($key => $value);
		}
		elsif ($curr_table) {
			# Set table property
			debug("set table '$curr_table' property '$key' = '$value'");
			$self->{$curr_table}->set($key => $value);
		}
		else {
			# Set schema property
			debug("set schema property '$key' = '$value'");
			$self->{lc($key)} = $value;
		}

		$curr_indent = length($prefix);
		$curr_prop = $key;
	}

	close(SCHEMA);

	return $self;		# return object as a success value
}

sub add_dir {
	my $self = shift;
	my ($filedir) = @_;

	debug_cache("adding directory '$filedir'");

	# Construct list of tables from directory contents.
	opendir DIR, $filedir
	  or die "Can't get directory list of '$filedir'. $!";
	my @files = readdir DIR;
	closedir DIR;

	my ($file, $table);
	for $file (sort @files) {
		next if $file =~ /^\./;		  # ignore hidden files
		next unless -f "$filedir/$file";  # ignore directories, ...
		($table = $file) =~ s/\.[^.]*$//; # trim file extension
		$self->add_table(name => $table, file => $file);
	}

	return $self;		# return object as a success value
}

sub add_table {
	my $self = shift;

	# Add a Table to a Schema

	my %arg = @_;
	my $name = $arg{name};

	debug_cache("adding table '$name'");

	$self->{$name} = Acme::Base::Schema::Table->new(
						name	=> $name,
						file	=> $arg{file},
						schema	=> $self
						);

	# Track order of table definitions.
	push @{$self->{'.properties'}->{tables}}, $name;

	return $self->{$name};
}

sub tables { return tables_secure(@_) }

sub tables_all { my $self = shift; return @{$self->{'.properties'}->{tables}} }

sub tables_secure {
	my ($self, $priv) = @_;

	# List tables that can be accessed by the user.
	# $priv is optional. Its codes are 'Access', 'Insert', 'Update',
	# or 'Delete'. Separate multiple codes with any delimiter if more than
	# one privilege is to checked at once.

	require Acme::Security;

	$priv ||= 'Access';	# default value

	my @tables;		# returned list of table names

	my ($tbl, $sc);
	for $tbl ( @{$self->{'.properties'}->{tables}} ) {
		if ($sc = $self->{$tbl}->security_code) {
			next unless Acme::Security::allow($priv,$sc);
		}
		push @tables, $tbl;
	}
	return @tables;
}

sub t {
	my ($self,$name) = @_;

	return unless defined $name && $self->{$name};

	# Dynamically load table definition if needed
	if (my $f = $self->{$name}->file) {
		$self->add($f) if ! $self->files->{$f};
	}

	return $self->{$name};
}

sub table { return t(@_) }

sub f {
	my ($self,$name) = @_;
	# Must be a fully qualified name here.
	my ($table,$field) = $name =~ /(.+?)\.(.*)/;	# 'table.field'
	($table,$field) = $name =~ /(.+?)__(.*)/	# 'table__field'
	  unless $table;
	return unless $table;
	my $t = $self->table($table) or return;
	return $t->f($field);
}

sub field { return f(@_) }

sub p {
	my $self = shift;
	my $name = shift;
	return $self->{'.properties'}->$name(@_);
}

sub property { return p(@_) }

sub schema {
	my $self = shift;
	# Return Schema object

	# Note: This object is usually the "schema" object. However,
	# Acme::Base::Schema cannot be sub-classed by using the @ISA array
	# because the data is cached -- especially true when running under
	# mod-perl. Acme::Form::Schema sub-classes this class by delegating.
	# It has it's own object of both the form's schema and the database's
	# schema. It need's to be the top-level schema object, so we check
	# to see if this object has a ".parent" component and whether the
	# parent has a "schema" method.

	return $self->{'.parent'}->schema
	  if $self->{'.parent'} && $self->{'.parent'}->can('schema');

	return $self;
}

sub DESTROY {
	my $self = shift;
	no warnings qw(uninitialized);
	if (($self->{nocache})
	||  ($self->{cache} && $self->{cache} =~ /^no$/i)) {
		Acme::Base::Schema::Cache::destroy($self,$self->{cachename});
	}
}

sub AUTOLOAD {
	my $self = shift;
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;

	return $self->t($method) || $self->p($method,@_);
}

#------------------------------------------------------------------------------

package Acme::Base::Schema::Table;

our @ISA = qw(Acme::Base::Object);

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my %arg = @_;

	my $self = new Acme::Base::Object(
		'.property_defaults' => 'Acme::Base::Schema::TableDefault',
		'.property_filter'   => \&Acme::Base::Schema::Private::mung_filter,
		'.parent'	     => $arg{schema},
		);
	$self->set(
		'name'    => $arg{name},
		'fields'  => [],
		'indexes' => [],
		'index'   => {},
		'displays'=> [],
		'display' => {},
		'elems'   => [],
		'elem'    => {},
		'file'    => $arg{file},
		);

	return bless $self, $class;
}

sub add_index {
	my $self = shift;

	# Add an index to a Table's list of indexes kept in its properties.

	my %arg = @_;
	my $value = $arg{value};
	my $name = $value;

	$self->index->{$name} = Acme::Base::Schema::Index->new(
							name => $name,
							table => $self
							);

	# Track order of index definitions.
	push @{$self->{'.properties'}->{indexes}}, $name;

	return $self->{$name};
}

sub indexes {
	my $self = shift;
	return @{$self->{'.properties'}->{indexes}};
}

sub add_display {
	my $self = shift;

	# Add an display to a Table's list of displays kept in its properties.

	my %arg = @_;
	my $value = $arg{value};
	my $name = $value;

	$self->display->{$name} = Acme::Base::Schema::Display->new(
							name => $name,
							table => $self
							);

	# Track order of index definitions.
	push @{$self->{'.properties'}->{displays}}, $name;

	# Enter into elems array too.
	$self->add_elem(type => 'display', name => $name);

	return $self->{$name};
}

sub displays {
	my ($self, $priv, $include_private) = @_;

	# List fields that can be accessed by the user.
	# $priv is optional. Its codes are 'Access', 'Insert', 'Update',
	# or 'Delete'. Separate multiple codes with any delimiter if more than
	# one privilege is to checked at once.
	# The $include_private, if true, forces the list to include fields
	# marked as private.

	$priv ||= 'Access';	# default value

	if (my $sc = $self->security_code) {

		# To optimize, we check field security codes only if the
		# table has a security code.

		require Acme::Security;

		Acme::Security::check('Access',$sc);

		my @displays;	# list of fields to list their properties

		# filter list of display fields to list
		for my $fld ( @{$self->{'.properties'}->{displays}} ) {
			if ($self->display->{$fld}->private) {
				next unless $include_private;
			}
			if ($sc = $self->display->{$fld}->security_code) {
				next unless Acme::Security::allow($priv,$sc);
			}
			push @displays, $fld;
		}
		return @displays;
	}
	else {
		return @{$self->{'.properties'}->{displays}}
		  if $include_private;

		# Exclude private displays
		return grep { ! $self->display->{$_}->private } @{$self->{'.properties'}->{displays}};
	}
}

sub add_field {
	my $self = shift;

	# Add a Field to a Table

	my %arg = @_;
	my $name = $arg{name};

	$self->{$name} = Acme::Base::Schema::Field->new(
						name	=> $name,
						table	=> $self
						);

	# Track order of field definitions.
	push @{$self->{'.properties'}->{fields}}, $name;

	# Enter into elems array too.
	$self->add_elem(type => 'field', name => $name);

	return $self->{$name};
}

sub fields { return fields_secure(@_) }

sub fields_all { my $self = shift; return @{$self->{'.properties'}->{fields}} }

sub fields_secure {
	my ($self, $priv, $include_private) = @_;

	# List fields that can be accessed by the user.
	# $priv is optional. Its codes are 'Access', 'Insert', 'Update',
	# or 'Delete'. Separate multiple codes with any delimiter if more than
	# one privilege is to checked at once.
	# The $include_private forces the list to include fields marked as
	# private.

	# Note: the returned list of fields does NOT include the audit fields.

	$priv ||= 'Access';	# default value

	if (my $sc = $self->security_code) {

		# To optimize, we check field security codes only if the
		# table has a security code.

		require Acme::Security;

		Acme::Security::check('Access',$sc);

		my @fields;	# list of fields to list their properties

		# filter list of fields to list
		for my $fld ( @{$self->{'.properties'}->{fields}} ) {
			if ($self->{$fld}->private) {
				next unless $include_private;
			}
			if ($sc = $self->{$fld}->security_code) {
				next unless Acme::Security::allow($priv,$sc);
			}
			push @fields, $fld;
		}
		return @fields;
	}
	else {
		return @{$self->{'.properties'}->{fields}}
		  if $include_private;

		# Exclude private fields.
		return grep { ! $self->{$_}->private } @{$self->{'.properties'}->{fields}};
	}
}

sub fields_audit {
	my $self = shift;

	# Audit fields are those the user can access which are not
	# excluded in the schema Audit_exclude property.

	my %exclude = map { $_,1 } split /\s*,\s*/, $self->{'.properties'}->fields_exclude;
	return grep { ! $exclude{$_} } $self->fields_secure(@_);
}


sub f {
	my ($self,$k) = @_;
	return $self->{$k};
}

sub field { return f(@_) }

sub add_elem {
	my $self = shift;

	# Add an elem to a Table's list of elems kept in its properties.

	my %arg = @_;

	my $name  = $arg{name} || 'elem' . (@{$self->{'.properties'}->{elems}} || '0');
	my $type  = $arg{type} || 'field';
	my $value = $arg{value};

	$self->elem->{$name} = Acme::Base::Schema::Elem->new(
							type  => $type,
							name  => $name,
							value => $value
							);

	# Track order of elem definitions.
	push @{$self->{'.properties'}->{elems}}, $name;

	return $name;
}

sub elems {
	my $self = shift;
	return @{$self->{'.properties'}->{elems}};
}

sub p {
	my $self = shift;
	my $name = shift;
	return $self->{'.properties'}->$name(@_);
}

sub property { return p(@_) }

sub table { return $_[0] }

sub schema {
	my $self = shift;
	# Return Schema object
	return $self->{'.parent'}->schema;
}

sub DESTROY {}

sub AUTOLOAD {
	my $self = shift;
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;

	my $prop = $self->p($method,@_);
	return $prop if defined $prop;
	return $self->{$method} if $self->{$method};
	return;
}

#------------------------------------------------------------------------------

package Acme::Base::Schema::Index;

our @ISA = qw(Acme::Base::Object);

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my %arg = @_;

	my $self = new Acme::Base::Object(
		'.property_filter'   => \&Acme::Base::Schema::Private::mung_filter,
		'.parent'	     => $arg{table},
		);
	$self->set(
		'name'   => $arg{name},
		);

	return bless $self, $class;
}

#------------------------------------------------------------------------------

package Acme::Base::Schema::Field;

our @ISA = qw(Acme::Base::Object);

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my %arg = @_;

	my $self = new Acme::Base::Object(
		'.property_defaults' => 'Acme::Base::Schema::FieldDefault',
		'.property_filter'   => \&Acme::Base::Schema::Private::mung_filter,
		'.parent'	     => $arg{table},
		);
	$self->set(
		'name'   => $arg{name},
		);

	return bless $self, $class;
}

sub schema {
	my $self = shift;
	# Return Schema object
	return $self->{'.parent'}->schema;
}

sub table {
	my $self = shift;
	# Return Table object
	return $self->{'.parent'};
}

sub field { return $_[0] }

sub p {
	my $self = shift;
	my $name = shift;
	return $self->{'.properties'}->$name(@_);
}

sub property { return p(@_) }

sub is_display_field { 0 }

#------------------------------------------------------------------------------

package Acme::Base::Schema::Display;

our @ISA = qw(Acme::Base::Object);

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my %arg = @_;

	die "Bad name for a display-only field in the schema."
	  . " The field name ". $arg{name}
	  . " must begin with a capital letter like: D_\n"
	  unless $arg{name} =~ /^(\d+_)*[A-Z]_/;

	my $self = new Acme::Base::Object(
		'.property_defaults' => 'Acme::Base::Schema::FieldDefault',
		'.property_filter'   => \&Acme::Base::Schema::Private::mung_filter,
		'.parent'	     => $arg{table},
		);
	$self->set(
		'name'   => $arg{name},
		);

	return bless $self, $class;
}

sub schema {
	my $self = shift;
	# Return Schema object
	return $self->{'.parent'}->schema;
}

sub table {
	my $self = shift;
	# Return Table object
	return $self->{'.parent'};
}

sub field { return $_[0] }

sub p {
	my $self = shift;
	my $name = shift;
	return $self->{'.properties'}->$name(@_);
}

sub property { return p(@_) }

sub is_display_field { 1 }

#------------------------------------------------------------------------------

package Acme::Base::Schema::Elem;

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my $self = {
		name	=> '',
		value	=> '',
		type	=> '',
		@_
		};

	return bless $self, $class;
}

sub add {
	my $self = shift;
	return $self->{value} = shift;
}

sub append {
	my $self = shift;
	return $self->{value} .= "\n".shift;
}

sub DESTROY {}

sub AUTOLOAD {
	my $self = shift;
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;

	return $self->{$method} if $self->{$method};
	return;
}

#------------------------------------------------------------------------------

package Acme::Base::Schema::Private;

#
# Private routines for Acme::Base::Schema.
# Stored here so they don't pollute namespace.
#

# The following properties customize their specified values by the subroutines
# listed in the %mung_filter hash. The subroutines are passed one argument,
# the value they are suppose to modify.

our %mung_filter = (
	'type'			=> sub { lc shift },	# force to lowercase
	'label'			=> \&yesno,
	'audit'			=> \&yesno,
	'secure'		=> \&yesno,
	'required'		=> \&yesno,
	'readonly'		=> \&yesno,
	'hidden'		=> \&yesno,
	'private'		=> \&yesno,
	'combo'			=> \&yesno,
	'array'			=> \&yesno,
	'query_all'		=> \&yesno,
	'choose_query_all'	=> \&yesno,
	'search_autoselect'	=> \&yesno,
	'search_autosearch'	=> \&yesno,
	'search_history_save'	=> \&yesno,
	'search_header'		=> \&yesno,
	'search_footer'		=> \&yesno,
	'pick_header'		=> \&yesno,
	'pick_footer'		=> \&yesno,
	'list_header'		=> \&yesno,
	'list_footer'		=> \&yesno,
	'form_list_header'	=> \&yesno,
	'form_list_footer'	=> \&yesno,
	'detail_header'		=> \&yesno,
	'detail_footer'		=> \&yesno,
	'email'			=> \&yesno,
	'nocalendar'		=> \&yesno,
	'nocolorbutton'		=> \&yesno,
	'noreadonly'		=> \&yesno,
	'nolabel'		=> \&yesno,
	'noid_label'		=> \&yesno,
	'noedit_list'		=> \&yesno,
	'noemail'		=> \&yesno,
	'noexport'		=> \&yesno,
	'noexport_log'		=> \&yesno,
	'nolimit'		=> \&yesno,
	'nolimit_alert'		=> \&yesno,
	'export'		=> \&yesno,
	'export_log'		=> \&yesno,
	'export_alert'		=> \&yesno,
	'is_text'		=> \&yesno,
	'is_number'		=> \&yesno,
	'is_date'		=> \&yesno,
	'is_time'		=> \&yesno,
	);

sub mung_filter {
	my ($k,$v) = @_;

	# Custom translation of key/value pairs for Acme::Base::Object::Property's

	return ($k, &{ $mung_filter{$k} }($v)) if $mung_filter{$k};
	return ($k, $v);
}

sub yesno {
	my $v = shift;

	return 'yes' unless $v;		# default is yes if no value
	return '' if 'no' =~ /\Q$v\E/i;	# false value
	return $v;			# original (true) value if not 'no'
}

#------------------------------------------------------------------------------

package Acme::Base::Schema::Cache;

#
# The schema file is read into memory once and cached here.
#

use Acme::Base::Object;

our %cache;

sub new {
	my ($this, $cachename) = @_;

	# Return a reference to a hash in persistant (global) memory.

	$cachename ||= 'default';
	$cache{$cachename} = Acme::Base::Object->new(
					'.timestamp'    => time,
					'.last_checked' => 0,
					);
	return $cache{$cachename};
}

sub get {
	my ($this, $cachename) = @_;

	# Return a reference to a hash in persistant (global) memory.

	$cachename ||= 'default';
	return $cache{$cachename} if defined $cache{$cachename};
	return $this->new($cachename);
}

sub destroy {
	my ($this, $cachename) = @_;

	# Return a hash from persistant (global) memory.

	$cachename ||= 'default';
	undef $cache{$cachename} if defined $cache{$cachename};
}

1;
