package Acme::Base::Schema::FieldDefault;

=head1 NAME

Acme::Base::Schema::FieldDefault - Compute default properties for schema

=head1 SYNOPSIS

    require Acme::Base::Schema::FieldDefault;

    my $select_property = Acme::Base::Schema::FieldDefault->select;
    my $key = Acme::Base::Schema::FieldDefault->key;

=head1 DESCRIPTION

These routines are suppose to provide intelligent default values for
schema properties. They are called by I<Acme::Base::Schema> and are not 
intended for general use.

=head1 CHANGE LOG

 5/29/00  Initial version.
 5/3/01   Converted to use Acme::Base::Schema and Acme::Base::Object.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000,2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

# Note: The $pobj parameters passed to the routines below is a
# Acme::Base::Object::Properties object. To get at the parent Object object,
# use $pobj->{object} or $pobj->object.

# Note: 9/13/00
# In a timing experiment making a template to enter new data, the folowing
# autoloaded routines produced the form in 3.5 sec on a 200 Mhz PII Pentium.
# Using the standard SelfLoader module, the time was 4.5 sec, and without
# using regular sub definitions, the time was 6 seconds.


require 5.6.0;
use strict;

use Acme::Config;

#------------------------------------------------------------------------------
#
# Autoload routines
#
#------------------------------------------------------------------------------

our %Subs = (

#
# The name of the following routines correspond to field property names in
# the schema.
#

#----------------------------------------
'type' => <<'END_OF_FUNC',
sub type {
	my ($class,$pobj) = @_;
	die "The field type for field '"
	  . $pobj->object->name
	  . "' of table '"
	  . $pobj->object->table->name
	  . "' was not specified in the schema. It is required.\n";
}
END_OF_FUNC


#----------------------------------------
'description' => <<'END_OF_FUNC',
sub description {
	my ($class,$pobj) = @_;
	# Allow Desc as an abbreviation to Description
	return $pobj->{desc} || '';
}
END_OF_FUNC

#----------------------------------------
'class' => <<'END_OF_FUNC',
sub class {
	my ($class,$pobj) = @_;
	# Choose from text, date, time, money, number, decimal, id
	my $t = $pobj->type;
	if ($t =~ /char/)    { return 'text'	}
	if ($t =~ /float/)   { return 'decimal' }
	if ($t =~ /decimal/) { return 'decimal' }
	if ($t =~ /serial/)  { return 'id'	}
	if ($t =~ /int/)     { return 'number'	}
	if ($t =~ /money/)   { return 'money'	}
	if ($t =~ /date/)    { return 'date'	}
	if ($t =~ /timespan/){ return 'text'	}
	if ($t =~ /time/)    { return 'time'	}
	if ($t =~ /bool/)    { return 'text'	}
	return 'text';
}
END_OF_FUNC

#----------------------------------------
'length' => <<'END_OF_FUNC',
sub length {
	my ($class,$pobj) = @_;
	my $t = $pobj->type;
	if ($t =~ /char/)  {
		my ($n) = $t =~ /(\d+)/;
		return $n;
	}
	if ($t =~ /smallint/)  { return 6; }
	if ($t =~ /integer/)  { return 11; }
	if ($t =~ /serial/)  { return 11; }
	if ($t =~ /int/)  {
		my ($n) = $t =~ /(\d+)/ || "0";
		if ($n <= 2) { return 6; }
		if ($n <= 4) { return 11; }
		return 22;
	}
	if ($t =~ /date/)  {
		return 10;
	}
	if ($t =~ /time/)  {
		return 7;
	}
	if ($t =~ /timestamp/)  {
		return 19;
	}
	if ($t =~ /float/)  {
		return 16;
	}
	if ($t =~ /decimal/)  {
		my ($n) = $t =~ /(\d+)/;
		return $n+2;
	}
	return 10;	# arbitrary number (chosen for input field size)
}
END_OF_FUNC

#----------------------------------------
'decimal' => <<'END_OF_FUNC',
sub decimal {
	my ($class,$pobj) = @_;
	my $c = $pobj->class;
	if ($c =~ /money/) { return 2; }
	my $t = $pobj->type;
	if ($t =~ /float/) { return -1; }
	if ($t =~ /decimal/) {
		# Extract number of decimals from Type statement.
		my ($n) = $t =~ /\d+,(\d+)/;
		return $n || "0";
	}
	return "0";
}
END_OF_FUNC

#----------------------------------------
'title' => <<'END_OF_FUNC',
sub title {
	my ($class,$pobj) = @_;

	# Capitalize name
	if (my $t = $pobj->{name}) {
		$t =~ s/^[A-Z]_//;	# remove display only field name prefix
		$t =~ s/_/ /g;
		$t =~ s/\bid\b/ID/;	# uppercase ID
		$t =~ s/\b(\w+)\b/ucfirst(lc($1))/eg;	# capitalize
		return $t;
	}
}
END_OF_FUNC

#----------------------------------------
'required' => <<'END_OF_FUNC',
sub required {
	my ($class,$pobj) = @_;

	# Accommodate misspelling Required as Require
	return $pobj->{require} || '';
}
END_OF_FUNC

#----------------------------------------
'label' => <<'END_OF_FUNC',
sub label {
	my ($class,$pobj) = @_;

	return '' if $pobj->nolabel;
	return $pobj->title if ! $pobj->{object}->is_display_field;
	return '';
}
END_OF_FUNC

#----------------------------------------
'format' => <<'END_OF_FUNC',
sub format {
	my ($class,$pobj) = @_;

	# Return format codes for sprintf().
	my $ftype = $pobj->{type};

	return '$%-0.2f' if $ftype =~ /money/ || $pobj->class =~ /money/;
	return "%-0.$2f" if $ftype =~ /decimal\((\d+),(\d+)\)/;
	return "%s"      if ($ftype =~ /int/ && $pobj->class =~ /id/i) || $ftype =~ /serial/;
	return "%0d"     if $ftype =~ /int/;
	return '%0f'     if grep { $ftype =~ /$_/ } 'float','decimal';
	return '%s';
}
END_OF_FUNC

#----------------------------------------
'default_static' => <<'END_OF_FUNC',
sub default_static {
	my ($class,$pobj) = @_;

##
## Make all defaults dynamic, i.e. fill in during form filler phase.
##
return '';
##

	if (my $d = $pobj->{default} || '') {
		# Return string if quoted.
		# Note: calling program need to be aware of possible quotes
		# surrounding value if default_static is to be a 
		# user setting in the schema.
		return $d if $d =~ m/^(['"])(.*)\1$/;

		return $d unless
		  $d =~ /^today$/i ||
		  $d =~ /^timestamp$/i ||
		  $d =~ /^now$/i ||
		  $d =~ /^tomorrow$/i ||
		  $d =~ /^yesterday$/i ||
		  $d =~ /^week$/i ||
		  $d =~ /^user$/i ||
		  $d =~ /\w+\(.*\)/;
	}
	return '';
}
END_OF_FUNC

#----------------------------------------
'default_dynamic' => <<'END_OF_FUNC',
sub default_dynamic {
	my ($class,$pobj) = @_;

	use Acme::Date qw(today timestamp dayadd);

	my $d = $pobj->{default};
	if (defined $d) {
##
## All defaults are filled in here during form filler phase.
##
##		# don't process quoted strings
##		return '' if $d =~ /^(['"]).*\1$/;
##
## Return string if quoted.
##
return $d if $d =~ m/^(['"])(.*)\1$/;
##

		return 'require Acme::Date; today()'
		  if $d =~ /^today$/i;
		return 'require Acme::Date; timestamp()'
		  if $d =~ /^timestamp$/i;
		return 'require Acme::Date; timestamp()'
		  if $d =~ /^now$/i;
		return 'require Acme::Date; scalar dayadd(1)'
		  if $d =~ /^tomorrow$/i;
		return 'require Acme::Date; scalar dayadd(-1)'
		  if $d =~ /^yesterday$/i;
		return 'require Acme::Date; scalar dayadd(7)'
		  if $d =~ /^week$/i;
		return 'require Local::Acme::User; user()'
		  if $d =~ /^user$/i;

		# return string if anything looks like a function call
		return $d if $d =~ /\w+\(.*\)/;
##
## Return anything because all defaults are filled in during form filler phase.
## Quote the string as dynamic defaults are eval'ed.
##
return "'$d'" if $d;
##
	}
	return '';
}
END_OF_FUNC

#----------------------------------------
'default_value' => <<'END_OF_FUNC',
sub default_value {
	my ($class,$pobj) = @_;

	# Return default value, either default_static or default_dynamic.

##	return default_static(@_) || default_dynamic(@_);
	return default_dynamic(@_);
}
END_OF_FUNC

#----------------------------------------
'valid' => <<'END_OF_FUNC',
sub valid {
	my ($class,$pobj) = @_;

	my $type = $pobj->type;
	if ($type =~ /smallint/i or $type =~ /int2/i) {
		return "this.value > -32768 && this.value < 32768";
	}
}
END_OF_FUNC

#----------------------------------------
'choose' => <<'END_OF_FUNC',
sub choose {
	my ($class,$pobj) = @_;
	return	$pobj->{choose_table}  ||
		$pobj->{choose_tables} ||
		$pobj->{choose_sql}    ||
		$pobj->{choose_menu}   ||
		$pobj->{choose_list}   ||
		'';
}
END_OF_FUNC

#----------------------------------------
'choose_sql' => <<'END_OF_FUNC',
use Acme::Base::Schema::Join	qw(parse_join);
sub choose_sql {
	my ($class,$pobj) = @_;

	my $choose = $pobj->{choose} || '';
	return $choose if $choose =~ m/^select\b/i;

	# Construct query from choose table
	my $table  = $pobj->choose_table                or return '';
	my $t      = $pobj->{object}->schema->t($table) or return ''; 

	my ($join_expr, $join_clause) = parse_join($table,$pobj->choose_join);

	my $fields = $pobj->choose_fields;

	my $where  = fully_qualify($t,$pobj->choose_where);

	my $order_by = $pobj->choose_order_by_clause;

	return 'select '    . $fields
	     . ' from '     . $join_expr
	     . ' where '    . $where . $join_clause
	     . ' order by ' . $order_by;
}
END_OF_FUNC

#----------------------------------------
'choose_query_sql' => <<'END_OF_FUNC',
use Acme::Base::Schema::Join	qw(parse_join);
sub choose_query_sql {
	my ($class,$pobj) = @_;

	# A version of choose_sql() for Query forms.

	my $choose = $pobj->{choose} || '';
	return $choose if $choose =~ m/^select\b/i;

	# Construct query from choose table
	my $table  = $pobj->choose_table                or return '';
	my $t      = $pobj->{object}->schema->t($table) or return ''; 

	my ($join_expr, $join_clause) = parse_join($table,$pobj->choose_join);

	my $fields = $pobj->choose_fields;

	my $where  = fully_qualify($t,$pobj->choose_query_where);

	my $order_by = $pobj->choose_order_by_clause;

	return 'select '    . $fields
	     . ' from '     . $join_expr
	     . ' where '    . $where . $join_clause
	     . ' order by ' . $order_by;
}
END_OF_FUNC

#----------------------------------------
'choose_tables' => <<'END_OF_FUNC',
sub choose_tables {
	my ($class,$pobj) = @_;
	my $choose = $pobj->{choose} or return '';
	return ''	# unknown
	  if ($choose =~ /^select\b/i
	  || $choose =~ /^menu\b/i
	  || $choose =~ /^list\b/i);

	# Return all tables in a comma separated list
	return $choose;
}
END_OF_FUNC

#----------------------------------------
'choose_table' => <<'END_OF_FUNC',
sub choose_table {
	my ($class,$pobj) = @_;
	my $choose = $pobj->choose_tables or return '';

	# Return first table found in list if more than one listed.
	my ($table) = $choose =~ /(\w+)/;
	return $table || '';
}
END_OF_FUNC

#----------------------------------------
'choose_fields' => <<'END_OF_FUNC',
sub choose_fields {
	my ($class,$pobj) = @_;
	# Get fields from target table properties
	my $t = get_choose_target($pobj) or return '';
	return $t->select_fields;
}
END_OF_FUNC

#----------------------------------------
'choose_join' => <<'END_OF_FUNC',
sub choose_join {
	my ($class,$pobj) = @_;
	# Get join clause from target table properties
	my $t = get_choose_target($pobj) or return '';
	return $t->select_join;
}
END_OF_FUNC

#----------------------------------------
'choose_join' => <<'END_OF_FUNC',
sub choose_join {
	my ($class,$pobj) = @_;

	my @tables = split /\W+/, $pobj->choose_tables;
	return '' unless @tables;

	my $table = shift @tables;

	if (! @tables) {
		# No extra tables listed in Search clause.
		# Use target table's select_join property

		my $t = get_choose_target($pobj) or return '';
		return $t->select_join;
	}

	my $join_clause = '';
	my $t = $pobj->{object}->schema->t($table) or return ''; 

	if (my $clause = $t->join) {

		# Build join structure to obtain join clauses of other
		# tables. Only join those tables listed in the 'choose'
		# clause.

		require Acme::Base::Schema::Join;

		my $j = new Acme::Base::Schema::Join $table, $clause;
		my @join_tables = $j->join_tables;

		for my $table (@tables) {
			if (grep { $table eq $_ } @join_tables) {
				$join_clause .= "\n and " if $join_clause;
				$join_clause .= $j->join_clause($table);
			}
		}
	}

	return $join_clause;
}
END_OF_FUNC

#----------------------------------------
'choose_where' => <<'END_OF_FUNC',
sub choose_where {
	my ($class,$pobj) = @_;
	# Get where from target table properties
	my $t = get_choose_target($pobj) or return '';
	return $t->select_where;
}
END_OF_FUNC

#----------------------------------------
'choose_query_where' => <<'END_OF_FUNC',
sub choose_query_where {
	my ($class,$pobj) = @_;
	return "1 = 1" if $pobj->choose_query_all;
	# Get where from target table properties
	my $t = get_choose_target($pobj) or return '';
	return $t->query_where;
}
END_OF_FUNC

#----------------------------------------
'choose_query_all' => <<'END_OF_FUNC',
sub choose_query_all {
	my ($class,$pobj) = @_;
	# Get where from target table properties
	my $t = get_choose_target($pobj) or return '';
	return $t->query_all;
}
END_OF_FUNC

#----------------------------------------
'choose_order' => <<'END_OF_FUNC',
sub choose_order {
	my ($class,$pobj) = @_;
	# Get order from target table properties
	my $t = get_choose_target($pobj) or return '';
	return $t->select_order;
}
END_OF_FUNC

#----------------------------------------
'choose_order_by_clause' => <<'END_OF_FUNC',
sub choose_order_by_clause {
	my ($class,$pobj) = @_;

	return fully_qualify($pobj->{object}->table,$pobj->{choose_order})
	  if $pobj->{choose_order};

	my $t = get_choose_target($pobj) or return '';
	return $t->select_order_by_clause;
}
END_OF_FUNC

#----------------------------------------
'choose_menu' => <<'END_OF_FUNC',
sub choose_menu {
	my ($class,$pobj) = @_;
	my $choose = $pobj->{choose} or return '';
	$choose =~ s/^menu\s+//i and return $choose;
	return '';
}
END_OF_FUNC

#----------------------------------------
'choose_list' => <<'END_OF_FUNC',
sub choose_list {
	my ($class,$pobj) = @_;
	# 'Choose list' parameter
	my $choose = $pobj->{choose} or return '';
	$choose =~ s/^list\s+//i and return $choose;
	return '';
}
END_OF_FUNC

#----------------------------------------
'choose_format' => <<'END_OF_FUNC',
sub choose_format {
	my ($class,$pobj) = @_;
	# Get format from target table properties
	my $t = get_choose_target($pobj) or return '';
	return $t->select_format;
}
END_OF_FUNC

#----------------------------------------
'choose_rows' => <<'END_OF_FUNC',
sub choose_rows {
	my ($class,$pobj) = @_;
	# Get rows from target table properties
	my $t = get_choose_target($pobj) or return '';
	return $t->select_rows;
}
END_OF_FUNC

#----------------------------------------
'choose_size' => <<'END_OF_FUNC',
sub choose_size {
	my ($class,$pobj) = @_;
	# Length of choose option character strings which determines
	# the width of dropdown list on screen.
	return $pobj->{choose_length} if $pobj->{choose_length};
	# Get rows from target table properties
	my $t = get_choose_target($pobj) or return '';
	return $t->select_size;
}
END_OF_FUNC

#----------------------------------------
'choose_length' => <<'END_OF_FUNC',
sub choose_length {
	my ($class,$pobj) = @_;
	# Synonym for choose_size
	return $pobj->choose_size || 0;
}
END_OF_FUNC

#----------------------------------------
'search' => <<'END_OF_FUNC',
sub search {
	my ($class,$pobj) = @_;
	return	$pobj->{search_table}  ||
		$pobj->{search_tables} ||
		$pobj->{search_sql}    ||
		'';
}
END_OF_FUNC

#----------------------------------------
'search_sql' => <<'END_OF_FUNC',
use Acme::Base::Schema::Join	qw(parse_join);
sub search_sql {
	my ($class,$pobj) = @_;

	my $search = $pobj->{search} || '';
	return $search if $search =~ m/^select\b/i;

	# Construct query from search table
	my $table  = $pobj->search_table                or return '';
	my $t      = $pobj->{object}->schema->t($table) or return ''; 

	my ($join_expr, $join_clause) = parse_join($table,$pobj->search_join);

	my $fields = $pobj->search_fields;

	my $where  = fully_qualify($t,$pobj->search_where);

	my $order_by = $pobj->search_order_by_clause;

	return 'select '    . $fields
	     . ' from '     . $join_expr
	     . ' where '    . $where . $join_clause
	     . ' order by ' . $order_by;
}
END_OF_FUNC

#----------------------------------------
'search_tables' => <<'END_OF_FUNC',
sub search_tables {
	my ($class,$pobj) = @_;
	my $search = $pobj->{search} or return '';
	return '' if $search =~ /^select\b/i;
	# Return all tables in list
	return $search;
}
END_OF_FUNC

#----------------------------------------
'search_table' => <<'END_OF_FUNC',
sub search_table {
	my ($class,$pobj) = @_;
	my $search = $pobj->search_tables or return '';

	# Return first table found in list if more than one listed.
	my ($table) = $search =~ /(\w+)/;
	return $table || '';
}
END_OF_FUNC

#----------------------------------------
'search_list' => <<'END_OF_FUNC',
sub search_list {
	my ($class,$pobj) = @_;
	# Get list clause from target table properties.
	my $t = get_search_target($pobj) or return '';
	return $t->select;
}
END_OF_FUNC

#----------------------------------------
'search_fields' => <<'END_OF_FUNC',
sub search_fields {
	my ($class,$pobj) = @_;
	# Get fully qualified list of fields.
	my $t = get_search_target($pobj) or return '';
	return fully_qualify($t,$pobj->search_list);
}
END_OF_FUNC

#----------------------------------------
'search_join' => <<'END_OF_FUNC',
sub search_join {
	my ($class,$pobj) = @_;
	my $search = $pobj->{search} or return '';
	return '' if $search =~ /^select\b/i;

	my @tables = split /\W+/, $pobj->search_tables;
	return '' unless @tables;

	my $table = shift @tables;

	if (! @tables) {
		# No extra tables listed in Search clause.
		# Use target table's select_join property

		my $t = get_search_target($pobj) or return '';
		return $t->select_join;
	}

	my $join_clause = '';
	my $t = $pobj->{object}->schema->t($table) or return ''; 

	if (my $clause = $t->join) {

		# Build join structure to obtain join clauses of other
		# tables. Only join those tables listed in the 'search'
		# clause.

		require Acme::Base::Schema::Join;

		my $j = new Acme::Base::Schema::Join $table, $clause;
		my @join_tables = $j->join_tables;

		for my $table (@tables) {
			if (grep { $table eq $_ } @join_tables) {
				$join_clause .= "\n and " if $join_clause;
				$join_clause .= $j->join_clause($table);
			}
		}
	}

	return $join_clause;
}
END_OF_FUNC

#----------------------------------------
'search_where' => <<'END_OF_FUNC',
sub search_where {
	my ($class,$pobj) = @_;
	# Get where clause from target table properties.
	my $t = get_search_target($pobj) or return '';
	return $t->select_where;
}
END_OF_FUNC

#----------------------------------------
'search_order' => <<'END_OF_FUNC',
sub search_order {
	my ($class,$pobj) = @_;
	# Get order by clause from target table properties.
	my $t = get_search_target($pobj) or return '';
	return $t->select_order;
}
END_OF_FUNC

#----------------------------------------
'search_order_by_clause' => <<'END_OF_FUNC',
sub search_order_by_clause {
	my ($class,$pobj) = @_;

	return fully_qualify($pobj->{object}->table,$pobj->{search_order})
	  if $pobj->{search_order};

	my $t = get_search_target($pobj) or return '';
	return $t->select_order_by_clause;
}
END_OF_FUNC

#----------------------------------------
'search_title' => <<'END_OF_FUNC',
sub search_title {
	my ($class,$pobj) = @_;
	# Get title from target table properties.
	my $t = get_search_target($pobj) or return '';
	return $t->title;
}
END_OF_FUNC

#----------------------------------------
'search_format' => <<'END_OF_FUNC',
sub search_format {
	my ($class,$pobj) = @_;
	# Get format clause from target table properties.
	my $t = get_search_target($pobj) or return '';
	return $t->select_format;
}
END_OF_FUNC

#----------------------------------------
'search_history' => <<'END_OF_FUNC',
sub search_history {
	my ($class,$pobj) = @_;
	# Return name constructed from table name.
	my $table = $pobj->search_table or return '';
	return "${table}Search";
}
END_OF_FUNC

#----------------------------------------
'search_id' => <<'END_OF_FUNC',
sub search_id {
	my ($class,$pobj) = @_;
	# Get list clause from target table properties.
	my $t = get_search_target($pobj) or return '';
	return $t->name.'.'.$t->id_field;
}
END_OF_FUNC

#----------------------------------------
'search_name' => <<'END_OF_FUNC',
sub search_name {
	my ($class,$pobj) = @_;
	# Get list clause from target table properties.
	my $t = get_search_target($pobj) or return '';
	return $t->name_field;
}
END_OF_FUNC

#----------------------------------------
'search_last_name' => <<'END_OF_FUNC',
sub search_last_name {
	my ($class,$pobj) = @_;
	# Get list clause from target table properties.
	my $t = get_search_target($pobj) or return '';
	return $t->last_name_field;
}
END_OF_FUNC

#----------------------------------------
'search_first_name' => <<'END_OF_FUNC',
sub search_first_name {
	my ($class,$pobj) = @_;
	# Get list clause from target table properties.
	my $t = get_search_target($pobj) or return '';
	return $t->first_name_field;
}
END_OF_FUNC

#----------------------------------------
'detail' => <<'END_OF_FUNC',
use Acme::SQL;
sub detail {
	my ($class,$pobj) = @_;
	return $pobj->{detail_sql} || '';
}
END_OF_FUNC

#----------------------------------------
'detail_field' => <<'END_OF_FUNC',
use Acme::SQL;
sub detail_field {
	my ($class,$pobj) = @_;
	my $detail = $pobj->{detail} || '';
	return $detail if $detail !~ m/^select\b/i;

	my $detail_sql = $pobj->{detail_sql} or return '';

	# Derive detail table.field from first field detail_sql.
	my $sqlo = new Acme::SQL $detail_sql;
	my ($field) = $sqlo->dbcolumns;		# get first field in query
	return $field if $field =~ /\./;	# field is fully qualified
	my ($table) = $sqlo->dbtables;		# get first table in query
	return "$table.$field";
}
END_OF_FUNC

#----------------------------------------
'detail_table' => <<'END_OF_FUNC',
sub detail_table {
	# Helper function
	my ($class,$pobj) = @_;
	my $detail = $pobj->detail_field or return '';
	my ($table) = $detail =~ /(\w+)/;
	return $table || '';
}
END_OF_FUNC

#----------------------------------------
'detail_sql' => <<'END_OF_FUNC',
use Acme::Base::Schema::Join	qw(parse_join);
sub detail_sql {
	my ($class,$pobj) = @_;

	my $detail = $pobj->{detail} || '';
	return $detail if $detail =~ m/^select\b/i;

	my $table  = $pobj->detail_table;
	my $t      = $pobj->{object}->schema->t($table) or return ''; 

	my ($join_expr, $join_clause) = parse_join($table,$pobj->detail_join);

	my $fields = $pobj->detail_list_fields;

	my $where  = fully_qualify($t,$pobj->detail_where);

	my $order_by = $pobj->detail_order_by_clause;

	return 'select '    . $fields
	     . ' from '     . $join_expr
	     . ' where '    . $where . $join_clause
	     . ' order by ' . $order_by;
}
END_OF_FUNC

#----------------------------------------
'detail_list' => <<'END_OF_FUNC',
sub detail_list {
	my ($class,$pobj) = @_;
	return $pobj->detail_list_fields;
}
END_OF_FUNC

#----------------------------------------
'detail_list_fields' => <<'END_OF_FUNC',
sub detail_list_fields {
	my ($class,$pobj) = @_;
	if ($pobj->{detail_list}) {
		my $table = $pobj->detail_table or return;
		my $t = $pobj->{object}->schema->t($table); 
		return fully_qualify($t,$pobj->{detail_list});
	}
	else {
		# Get list clause from target table properties.
		my $t = get_detail_target($pobj) or return '';
		return $t->form_list_fields;
	}
}
END_OF_FUNC

#----------------------------------------
'detail_join' => <<'END_OF_FUNC',
sub detail_join {
	my ($class,$pobj) = @_;
	# Get list clause from target table properties.
	my $t = get_detail_target($pobj) or return '';
	return $t->form_list_join;
}
END_OF_FUNC

#----------------------------------------
'detail_where' => <<'END_OF_FUNC',
sub detail_where {
	my ($class,$pobj) = @_;
	my $t = get_detail_target($pobj) or return '';
	return $t->form_list_where;
}
END_OF_FUNC

#----------------------------------------
'detail_order' => <<'END_OF_FUNC',
sub detail_order {
	my ($class,$pobj) = @_;
	# Get list clause from target table properties.
	my $t = get_detail_target($pobj) or return '';
	return $t->form_list_order;
}
END_OF_FUNC

#----------------------------------------
'detail_order_by_clause' => <<'END_OF_FUNC',
sub detail_order_by_clause {
	my ($class,$pobj) = @_;

	return fully_qualify($pobj->{object}->table,$pobj->{detail_order})
	  if $pobj->{detail_order};

	my $t = get_detail_target($pobj) or return '';
	return $t->form_list_order_by_clause;
}
END_OF_FUNC

#----------------------------------------
'detail_format' => <<'END_OF_FUNC',
sub detail_format {
	my ($class,$pobj) = @_;
	# Get list clause from target table properties.
	my $t = get_detail_target($pobj) or return '';
	return $t->form_list_format;
}
END_OF_FUNC

#----------------------------------------
'detail_header' => <<'END_OF_FUNC',
sub detail_header {
	my ($class,$pobj) = @_;
	# Get list clause from target table properties.
	my $t = get_detail_target($pobj) or return '';
	return $t->form_list_header || '';
}
END_OF_FUNC

#----------------------------------------
'detail_footer' => <<'END_OF_FUNC',
sub detail_footer {
	my ($class,$pobj) = @_;
	# Get list clause from target table properties.
	my $t = get_detail_target($pobj) or return '';
	return $t->form_list_footer || '';
}
END_OF_FUNC

#----------------------------------------
'detail_total' => <<'END_OF_FUNC',
sub detail_total {
	my ($class,$pobj) = @_;
	return $pobj->{detail_totals} if $pobj->{detail_totals};
	my $t = get_detail_target($pobj) or return '';
	return $t->form_list_total || '';
}
END_OF_FUNC

#----------------------------------------
'detail_total_fields' => <<'END_OF_FUNC',
sub detail_total_fields {
	my ($class,$pobj) = @_;
	# Return list of table's fully qualified column names
	my $n = $pobj->{object}->table->name;
	my @f;		# returned totals
	for (split /,\s*/, $pobj->detail_total) {
		if (/\./) {
			push @f, $_;	# field already fully qualified
		}
		else {
			push @f, "$n.$_";
		}
	}
	return join ', ', @f if @f;
	return '';
}
END_OF_FUNC

#----------------------------------------
'detail_title' => <<'END_OF_FUNC',
sub detail_title {
	my ($class,$pobj) = @_;
	# Get list clause from target table properties.
	my $t = get_detail_target($pobj) or return '';
	return $t->form_list_title;
}
END_OF_FUNC

#----------------------------------------
'input' => <<'END_OF_FUNC',
sub input {
	my ($class,$pobj) = @_;
	my $input = $pobj->{input_table} ||
		    $pobj->{input_sql}   ||
		    $pobj->{input_array};
	return $input || '';
}
END_OF_FUNC

#----------------------------------------
'input_table' => <<'END_OF_FUNC',
sub input_table {
	my ($class,$pobj) = @_;
	my $input = $pobj->{input} or return '';

	return
	  if ($input =~ /^select/	# is a query, not a table name
	  or $input =~ /^(\W).*\1$/);	# is an array, string has delimiters

	my ($table,$field) = $input =~ /(.*)\.(.*)/;
	return $table || $input;
}
END_OF_FUNC

#----------------------------------------
'input_sql' => <<'END_OF_FUNC',
sub input_sql {
	my ($class,$pobj) = @_;
	my $input = $pobj->{input} or return '';
	return $input if $input =~ m/^select\b/i;

	# Construct query from input table
	if ($pobj->input_table) {
		my $t = get_input_target($pobj);
		my $fields = $t->select_fields;
		$fields =~ s/.*?, //;	# remove first (key) field
		return 'select '    . $fields            .
		       ' from '     . $t->name           .  
		       ' where ( '  . $t->select_where   .
		       ' ) '        . $t->restrict_where .
		       ' order by ' . $t->select_order_by_clause;
	}
	return '';
}
END_OF_FUNC

#----------------------------------------
'input_array' => <<'END_OF_FUNC',
sub input_array {
	my ($class,$pobj) = @_;
	my $input = $pobj->{input} or return '';
	return $input if $input =~ m/^(\W).*\1$/;	# string has delimiters
	return '';
}
END_OF_FUNC

#----------------------------------------
'input_length' => <<'END_OF_FUNC',
sub input_length {
	my ($class,$pobj) = @_;
	# Return default size of input array fields
	return Acme::Config->Input_Array_Size;
}
END_OF_FUNC

#----------------------------------------
'input_decimal' => <<'END_OF_FUNC',
sub input_decimal {
	my ($class,$pobj) = @_;
	my $t = $pobj->type;
	if ($t =~ /money/)   { return 2; }
	if ($t =~ /decimal/) {
		my ($n) = $t =~ /\d+,(\d+)/;
		return $n || "0";
	}
	return "0";
}
END_OF_FUNC

#----------------------------------------
'private' => <<'END_OF_FUNC',
sub private {
	my ($class,$pobj) = @_;
	# Return synonym for private.
	return $pobj->{not_used};
}
END_OF_FUNC

#----------------------------------------
'not_used' => <<'END_OF_FUNC',
sub not_used {
	my ($class,$pobj) = @_;
	# Return synonym for not_used.
	return $pobj->{private};
}
END_OF_FUNC

#----------------------------------------
'secure' => <<'END_OF_FUNC',
sub secure {
	my ($class,$pobj) = @_;
	return $pobj->security_code;
}
END_OF_FUNC

#----------------------------------------
'security_code' => <<'END_OF_FUNC',
sub security_code {
	my ($class,$pobj) = @_;
	return $pobj->{security_codes} if $pobj->{security_codes};
	return '' if ! $pobj->{secure};

	# Return fully qualified field name as a default security code.
	return $pobj->{object}->table->name  . '.' . $pobj->{name}
	  if $pobj->{secure} eq 'yes';

	return $pobj->{secure};
}
END_OF_FUNC

#----------------------------------------
'array_separator' => <<'END_OF_FUNC',
sub array_separator {
	return Acme::Config->Array_Separator;
}
END_OF_FUNC

#-------------------------------------------------------------------------------
'is_text' => <<'END_OF_FUNC',
sub is_text {
	my ($class,$pobj) = @_;
	my $ftype = $pobj->{type};
	return grep { $ftype =~ /$_/ } 'char','varchar','text';
}
END_OF_FUNC

#-------------------------------------------------------------------------------
'is_number' => <<'END_OF_FUNC',
sub is_number {
	my ($class,$pobj) = @_;
	my $ftype = $pobj->{type};
	return 1 if grep { $ftype =~ /$_/ } 'float','decimal','money';
	return 0 if $ftype =~ /int/ && $pobj->class =~ /id/i;
	return 1 if $ftype =~ /int/;
	return 0;
}
END_OF_FUNC

#-------------------------------------------------------------------------------
'is_date' => <<'END_OF_FUNC',
sub is_date {
	my ($class,$pobj) = @_;
	my $ftype = $pobj->{type};
	return grep { $ftype =~ /$_/ } 'date';
}
END_OF_FUNC

#-------------------------------------------------------------------------------
'is_time' => <<'END_OF_FUNC',
sub is_time {
	my ($class,$pobj) = @_;
	my $ftype = $pobj->{type};
	return grep { $ftype =~ /$_/ } 'time';
}
END_OF_FUNC
);

#------------------------------------------------------------------------------
#
# Utility routines
#
#------------------------------------------------------------------------------

sub get_choose_target {
	my ($pobj) = @_;

	# Return a pointer to the table structure identified in argument.
	# This is a helper function for finding the target table of a field's
	# choose statement.

	my $choose = $pobj->{choose_table} || $pobj->{choose} || '';

	return
	  if ($choose =~ /^select/i
	  or $choose =~ /^menu/i
	  or $choose =~ /^list/i);

	my ($table,$field) = $choose =~ /(.*)\.(.*)/;
	$table ||= $choose;

	return $pobj->{object}->schema->t($table);
}

sub get_search_target {
	my ($pobj) = @_;

	# Return a pointer to the table structure identified in argument.
	# This is a helper function for finding the target table of a field's
	# search statement.

	my $search = $pobj->{search} or return;;

	return if $search =~ /^select/i;

	my ($table) = $search =~ /(\w+)/;

	return $pobj->{object}->schema->t($table);
}

sub get_detail_target {
	my ($pobj) = @_;

	# Return a pointer to the table structure identified in argument.
	# This is a helper function for finding the target table of a field's
	# detail statement.

	my $detail = $pobj->{detail} or return;

	my ($table) = $detail =~ /(\w+)/;

	return $pobj->{object}->schema->t($table);
}

sub get_input_target {
	my ($pobj) = @_;

	# Return a pointer to the table structure identified in argument.
	# This is a helper function for finding the target table of a field's
	# input statement.

	my $input = $pobj->{input_table} || $pobj->{input} || '';

	return if $input =~ /^select/i;

	my ($table,$field) = $input =~ /(.*)\.(.*)/;
	$table ||= $input;

	return $pobj->{object}->schema->t($table);
}

sub fully_qualify {
	my ($t, $string) = @_;

	#
	# Return string qualifying unqualified field names found in the table
	# object pointed to by $t with the table's name. 
	#

	my $table = $t->name;

	my $fq = sub {
		my $f = shift;
		if (defined $t->{$f} || $f eq $t->id) {
			return "$table.$f";	# return fully qualified name
		}
		return $f;
	};

	# Change non-quoted fields in string.
	$string =~ s/(?<!['"])([\w.]+)(?!['"])/&$fq($1)/eg;

	return $string;
}

#
# Fixed table of default values
#

our %Default_Value = (
	search_autosearch	=> 'yes',
	search_autoselect	=> 'yes',
	search_history_save	=> 'yes',
	);

#------------------------------------------------------------------------------

sub AUTOLOAD {
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;
	return if $method eq 'DESTROY';		# gobble

	my $s;
	if (exists $Default_Value{$method}) {
		return $Default_Value{$method};
	}
	elsif (exists $Subs{$method}) {
		$s = $Subs{$method};
		delete $Subs{$method};		# release memory

		if ($s) {
			eval "package $pkg; $s";
			die "${pkg}::$method(): $@" if $@;
			goto &$method;
		}
	}
	return '';				# unknown parameter/method
}

1;
