package Acme::Base::Schema::Join;

=head1 NAME

Acme::Base::Schema::Join - Join object for Acme::Base::Schema

=head1 SYNOPSIS

    use Acme::Base::Schema::Join;

    my $join_clause = $t->join;		# Table's join string from schema

    my $j = new Acme::Base::Schema::Join $table, $join_clause;

    for my $tab ($j->join_tables) {
	    push @join_fields, $j->fields($tab);
    }

    my $other_field = $j->join_field($table_name);
    my @tables      = $j->join_tables();

    my @joins       = j->joins();

    my $expr        = $j->join_expr($table);		# SQL92
    my $str         = $j->join_clause($table);	# Informix

    my @allfields   = $j->fields();

    my $myfield     = $j->field();

    # Convenience function
    ($join_expr, $join_clause) = parse_join($table, $join_clause);

=head1 DESCRIPTION

Acme::Base::Schema::Join provides an object interface to the parsed tokens of the
join clause specified in the schema.

The parse_join() function takes a join clause from the schema and returns
a two element array, the first element is a join expression that is to be
inserted into the from clause. If no tables are joined, the primary table
is returned. The second element is a join clause that is to be inserted
into the where clause. It is blank for SQL92 databases such as Postgres.

If the database engine is Informix, the join expression returned is a comma
separated list of tables joined, and the second element is the join clause.

For the Informix database engine, where outer joins can be specified with
simple text operators, both elements are character strings that are
formatted to be inserted directly into a SQL query. The tables string begins
with a comma, and the join clause begins with " and (...)".

For the Postgres database and SQL92 engines, where outer joins have to be
expressed as a join expression and combining join clauses with simple text
operators does not work, only the table name or join expression is 
returned. No commas or " and (...)" strings are added to the output.

=head1 CHANGE LOG

 5/11/00  Initial version.
 10/18/01 Added parse_join() convenience function.
 11/26/01 Added join_expr() method to support SQL92 and Acme::SQL::Join.

=head1 TODO

Allow joins on multiple fields. Currently, only one field can join to one
field in the joined table. Change join_field() to join_fields() and change
field() to fields().

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.000;
require Exporter;

@ISA = qw(Exporter);
@EXPORT_OK = qw(parse_join);

use Acme::Config;

#-------------------------------------------------------------------------------
#
# Library functions
#
#-------------------------------------------------------------------------------

sub parse_join {
	my ($table, $join_clause) = @_;

	return parse_join_informix($table,$join_clause)
	  if Acme::Config->DBENGINE =~ /Informix/i;

	# Return a join expression as defined in SQL92.

	if ($join_clause) {
		my $j = new Acme::Base::Schema::Join $table, $join_clause;
		my $join_expr = $j->join_expr($table);
		return ($join_expr,'');
	}
	else {
		return ($table,'');
	}
}

sub parse_join_informix {
	my ($table, $join_clause) = @_;

	# Informix uses a join expression like "table, outer table2"
	# and places the conditions in the where clause.

	my $join_tables = '';

	$join_clause ||= '';

	if ($join_clause) {

		my $j = new Acme::Base::Schema::Join $table, $join_clause;

		$join_tables = join ", outer ", $table, $j->join_tables;
		$join_clause = "\n and ($join_clause)\n";
	}

	return ($join_tables,$join_clause);
}

#-------------------------------------------------------------------------------
#
# Object methods
#
#-------------------------------------------------------------------------------

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my ($table, $join) = @_;

	return unless $join;
	die "Table was not specified in arguments." unless $table;

	my $self = {};

	my ($fld, $join_tab, $join_fld, $join_clause);

	for my $clause (split /\s+[aA][nN][dD]\s+/, $join) {
		$join_clause = $clause;
		$clause =~ s/$table\.//;	# remove table spec before parse
		($fld, $join_tab, $join_fld) = $clause =~ /(\w+)\s*=\s*(\w+)\.(\w+)/ or
		  ($join_tab, $join_fld, $fld) = $clause =~ /(\w+)\.(\w+)\s*=\s*(\w+)/;
		if ($join_tab) {
			$self->{$fld} ||= [];
			push @{$self->{$fld}}, { join_table => $join_tab,
						 join_field => $join_fld,
						 join_clause => $join_clause,
						};
		}
	}

	return bless $self, $class;
}

sub joins {
	my $self = shift;

	# Return an array of hashes of all join conditions.
	my @a;
	
	for my $fld (keys %$self) {
		for (@{$self->{$fld}}) {
			my $h = { 'field' => $fld,
				  'join_table' => $_->{join_table},
				  'join_field' => $_->{join_field},
				  'join_clause' => $_->{join_clause},
				};
			push @a, $h;
		}
	}
	return @a;
}

sub join_tables {
	my $self = shift;

	# List tables joined in the join_clause.
	my @a;
	for my $fld (keys %$self) {
		for (@{$self->{$fld}}) {
			push @a, $_->{join_table};
		}
	}
	return @a;
}
 
sub join_field {
	my ($self,$field,$join_table) = @_;

	# Return the field name in the other joined table that joins
	# this table's $field. 

	my $n = 0;
	for (@{$self->{$field}}) {
		return $self->{$field}->[$n]->{join_field}
		  if $self->{$field}->[$n]->{join_table} eq $join_table;
		$n++; 
	}
	return;

}

sub join_expr {
	my ($self, $table) = @_;

	# Return the join expression for the specified table.

	require Acme::SQL::Join;
	my $j = new Acme::SQL::Join;

	my $join_expr = '';
	for my $jtbl ($self->join_tables) {
		$j->add_join("$table left join $jtbl on "
		  . $self->join_clause($jtbl));
	}
	return $j->as_string;
}

sub join_clause {
	my ($self,$table,$field) = @_;

	# Return the specific join clause for a table field.
	# The field is optional. If not specified, the first join clause
	# found containing the table is returned.

	my @a;
	if ($field) {
		for (@{$self->{$field}}) {
			push @a, $_->{join_clause}
			  if $_->{join_table} eq $table;
		}

		return join " and ", @a;
	}
	else {
		# Search all joined fields for first table matching $table
		for $field (keys %$self) {
			for (@{$self->{$field}}) {
				push @a, $_->{join_clause}
				  if $_->{join_table} eq $table;
			}
		}

		return join " and ", @a;
	}
}

sub fields {
        my $self = shift;

        # Return an array of fields of all joins.
        return keys %$self;
}

sub field {
	my ($self,$join_table,$join_field) = @_;


	# Return the field name in this table that joins with the joined table.
	# Return the first one found for table if $join_field not specified.

	# Search all join fields
	for my $fld (keys %$self) {
		for (@{$self->{$fld}}) {
			if ($_->{join_table} eq $join_table) {
				if ($join_field) {
					return $fld
					  if $_->{join_field} eq $join_field;
				}
				else {
					return $fld
				}
			}
		}
	}

	return;
}

1; 
