package Acme::Base::Schema::TableDefault;

=head1 NAME

Acme::Base::Schema::TableDefault - Compute default properties for schema

=head1 SYNOPSIS

    require Acme::Base::Schema::TableDefault;

    my $select_property = Acme::Base::Schema::TableDefault->select;
    my $key = Acme::Base::Schema::TableDefault->key;

=head1 DESCRIPTION

These routines are suppose to provide intelligent default values for
schema properties. They are called by I<Acme::Base::Schema> and are not 
intended for general use.

=head1 CHANGE LOG

 5/29/00  Initial version.
 5/3/01   Converted to use Acme::Base::Schema and Acme::Base::Object.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000,2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

# Note: The $pobj parameters passed to the routines below is a
# Acme::Base::Object::Properties object. To get at the parent Object object,
# use $pobj->{object} or $pobj->object.

# Note: 9/13/00
# In a timing experiment making a template to enter new data, the folowing
# autoloaded routines produced the form in 3.5 sec on a 200 Mhz PII Pentium.
# Using the standard SelfLoader module, the time was 4.5 sec, and without
# using regular sub definitions, the time was 6 seconds. Good thing they
# make faster computers.


use Acme::Config;

#------------------------------------------------------------------------------
#
# Autoload routines
#
#------------------------------------------------------------------------------

our %Subs = (

#
# The name of the following routines correspond to table property names in the
# schema.
#

#----------------------------------------
'key' => <<'END_OF_FUNC',
sub key {
	my ($class,$pobj) = @_;
	# Return ID number as the key field.
	return $pobj->{id} if defined $pobj->{id}; 
	
	# This system requires a unique key in order to add and edit records.

	# Return first listed field.
	return $pobj->{fields}->[0];
}
END_OF_FUNC

#----------------------------------------
'description' => <<'END_OF_FUNC',
sub description {
	my ($class,$pobj) = @_;
	# Allow Desc as an abbreviation to Description
	return $pobj->{desc} || '';
}
END_OF_FUNC

#----------------------------------------
'title' => <<'END_OF_FUNC',
sub title {
	my ($class,$pobj) = @_;
	# Capitalize name
	my $title = $pobj->{name};
	$title =~ s/_/ /g;
	$title =~ s/\b(\w+)\b/ucfirst(lc($1))/eg;	# captitalize
	return $title;
}
END_OF_FUNC

#----------------------------------------
'id_title' => <<'END_OF_FUNC',
sub id_title {
	my ($class,$pobj) = @_;
	# ID fields don't have their own field structures.
	return '' unless defined $pobj->{id};

	# Turn name into title.
	my $title = $pobj->{id};
	$title =~ s/_/ /g; 			# change underscores to spaces
	$title =~ s/\b(\w+)\b/ucfirst(lc($1))/eg; # capitalize
	$title =~ s/\bId\b/ID/;			# capitalize id
	return $title;
}
END_OF_FUNC

#----------------------------------------
'id_label' => <<'END_OF_FUNC',
sub id_label {
	my ($class,$pobj) = @_;

	return '' if $pobj->noid_label;
	return $pobj->id_title;
	return '';
}
END_OF_FUNC

#----------------------------------------
'id_description' => <<'END_OF_FUNC',
sub id_description {
	my ($class,$pobj) = @_;

	return ucfirst(lc($pobj->row_name)) . ' ID number.';
}
END_OF_FUNC

#----------------------------------------
'class' => <<'END_OF_FUNC',
sub class {
	my ($class,$pobj) = @_;

	# Table classes are advisory. They are used primarily to sort and
	# group tables of information together. They can also be used to 
	# set up default values for other field setting while designing
	# new tables.

	# Infer table class from my field naming conventions.
	my $tabname = $pobj->{name};
	if ($pobj->id) {
		return 'detail' if defined $pobj->{master};
		return 'master' if defined $pobj->{detail};

		my $t = $pobj->{schema}->{$tabname};	# table object
		return 'log' if $t->{log};

		if ($pobj->name_field) {
			return 'people'
			  if defined $t->{terminate_date}
			  || defined $t->{stop_date};
			return 'master';
		}
	}
	elsif (defined $pobj->{key}) {
		return 'join' if defined $pobj->{join};

		my $t = $pobj->{schema}->{$tabname};	# table object
		return 'log' if $t->{log};

		my $f = $t->{$pobj->{key}};		# key field
		return 'join' if (defined $f && ($f->search || $f->choose));

		return 'code' if $t->{retired};
		return 'code';		# default
	}

	# Audit tables are not listed in the schema, so a table class 
	# cannot be 'audit' here.

	return 'work';	# default
}
END_OF_FUNC

#----------------------------------------
'row_name' => <<'END_OF_FUNC',
sub row_name {
	my ($class,$pobj) = @_;
	# Make singular name out of table name.
	my $n = $pobj->{name};
	$n =~ s/_/ /g;	# change underlines to spaces
	$n =~ s/ies$/y/ and return $n;
	$n =~ s/s$// and return $n;
	return $n;
}
END_OF_FUNC

#----------------------------------------
'row_title' => <<'END_OF_FUNC',
sub row_title {
	my ($class,$pobj) = @_;
	my $title = ucfirst(lc($pobj->row_name));
	$title =~ s/ ([a-z])/' '.uc($1)/ge;	# capitalize words
	return $title;
}
END_OF_FUNC

#----------------------------------------
'id_field' => <<'END_OF_FUNC',
sub id_field {
	my ($class,$pobj) = @_;
	return $pobj->{id_fields} if $pobj->{id_fields}; # synonym
	# Use key field name only if it is numeric
	my $id = $pobj->id;
	return $id if $id;

	my $tbl = $pobj->name;
	my $fld = $pobj->key; 
	my $cls = $pobj->object->schema->t($tbl)->f($fld)->class;
	return $fld if $cls eq 'number';
	return '';
}
END_OF_FUNC

#----------------------------------------
'code_field' => <<'END_OF_FUNC',
sub code_field {
	my ($class,$pobj) = @_;
	return $pobj->{code_fields} if $pobj->{code_fields}; # synonym
	# Use key field name only if it is not an ID
	return '' if $pobj->id;	# unknown
	return $pobj->key;	# use key field
}
END_OF_FUNC

#----------------------------------------
'name_field' => <<'END_OF_FUNC',
sub name_field {
	my ($class,$pobj) = @_;
	return $pobj->{name_fields} if $pobj->{name_fields}; # synonym
	# Use key name unless it is numeric
	my $fld = $pobj->key; 
	my $id = $pobj->id;
	if ($fld ne $id) {
		my $cls = $pobj->{object}->{$fld}->class;
		return $fld if $cls eq 'text';
	}

	return $pobj->{last_name_field} if $pobj->{last_name_field};

	# Finding the first field with the pattern 'name' in it.
	my $first_text_field;
	for ($pobj->{object}->fields) {
		next if $pobj->{object}->{$_}->class ne 'text';
		$first_text_field ||= $_;
		return $_ if /name/;
	}

	return $first_text_field || '';
}
END_OF_FUNC

#----------------------------------------
'email_field' => <<'END_OF_FUNC',
sub email_field {
	my ($class,$pobj) = @_;
	return $pobj->{email_fields} || '';	# synonym
}
END_OF_FUNC

#----------------------------------------
'phone_field' => <<'END_OF_FUNC',
sub phone_field {
	my ($class,$pobj) = @_;
	return $pobj->{phone_fields} || '';	# synonym
}
END_OF_FUNC

#----------------------------------------
'last_name_field' => <<'END_OF_FUNC',
sub last_name_field {
	my ($class,$pobj) = @_;
	return $pobj->name_field || '';
}
END_OF_FUNC

#----------------------------------------
'email_field_list_fields' => <<'END_OF_FUNC',
sub email_field_list_fields {
	my ($class,$pobj) = @_;
	# Return list of fields that hold email addresses
	# that can be emailed to in lists.
	my $t = $pobj->{object};
	return fully_qualify($t,($pobj->{email_field_list} || $pobj->email_field || ''));
}
END_OF_FUNC

#----------------------------------------
'name_field_list_fields' => <<'END_OF_FUNC',
sub name_field_list_fields {
	my ($class,$pobj) = @_;
	# Return list of name fields. Corresponds to email_field_list_fields().
	my $t = $pobj->{object};
	return fully_qualify($t,($pobj->{name_field_list} || $pobj->name_field || ''));
}
END_OF_FUNC

#----------------------------------------
'select' => <<'END_OF_FUNC',
sub select {
	my ($class,$pobj) = @_;

	# Return list of table's column names
	# The first field must be the key field.

	my $code_field = $pobj->code_field || $pobj->key;

	# Restrict list to only the first 6 fields of table.
	return join ', ', $code_field, $code_field, ($pobj->{object}->fields)[0..5]
	  if $code_field;

	return join ', ', $pobj->{id}, $pobj->{id}, ($pobj->{object}->fields)[0..5]
	  if $pobj->{id};

	return join ', ', $pobj->{key}, ($pobj->{object}->fields)[0..5]
	  if $pobj->{object}->{key};

	return join ', ', $pobj->{fields}->[0], ($pobj->{object}->fields)[0..5];
}
END_OF_FUNC

#----------------------------------------
'select_sql' => <<'END_OF_FUNC',
use Acme::Base::Schema::Join	qw(parse_join);
sub select_sql {
	my ($class,$pobj) = @_;
	my $select = $pobj->select or return '';
	return $select if $select =~ m/^select\b/i;

	# Construct query from select table
	if ($pobj->select) {
		my $t = $pobj->{object};
		my ($join_expr, $join_clause) = parse_join($t->name, $pobj->select_join);
		my $where    = fully_qualify($t,$pobj->select_where);

		# Note: using "select distinct ..." is safer but causes errors
		# in Postgres if a column in the "order by" clause is not also
		# in the columns selected.

		return 
		    'select '    . $pobj->select_fields .		"\n"
		  . ' from '     . $join_expr .                         "\n"
		  . ' where '    . $where  . $join_clause .		"\n"
		  . ' order by ' . $pobj->select_order_by_clause .	"\n";
	}
	return '';
}
END_OF_FUNC

#----------------------------------------
'select_fields' => <<'END_OF_FUNC',
sub select_fields {
	my ($class,$pobj) = @_;
	# Return list of table's fully qualified column names
	my $n = $pobj->name;
	my @f;		# returned fields
	for (split /,\s*/, $pobj->select) {
		if (/\./) {
			push @f, $_;	# field already fully qualified
		}
		else {
			push @f, "$n.$_";
		}
	}
	return join ', ', @f;
}
END_OF_FUNC

#----------------------------------------
'select_join' => <<'END_OF_FUNC',
sub select_join {
	my ($class,$pobj) = @_;
	return '';
}
END_OF_FUNC

#----------------------------------------
'select_where' => <<'END_OF_FUNC',
sub select_where {
	my ($class,$pobj) = @_;
	# Default where clause is to select all
	return '1 = 1';
}
END_OF_FUNC

#----------------------------------------
'select_order' => <<'END_OF_FUNC',
sub select_order {
	my ($class,$pobj) = @_;
	return '1';
}
END_OF_FUNC

#----------------------------------------
'select_order_by_clause' => <<'END_OF_FUNC',
sub select_order_by_clause {
	my ($class,$pobj) = @_;

	return fully_qualify($pobj->{object},$pobj->select_order) || 1;
}
END_OF_FUNC

#----------------------------------------
'select_format' => <<'END_OF_FUNC',
sub select_format {
	my ($class,$pobj) = @_;
	if (defined $pobj->{select}) {
		my @f = split /\s*,\s*/, $pobj->{select};
		if (@f > 2) {
			return '%s -- ' . (join ', ', map {'%s'} @f[2..$#f]);
		} else {
			return '%s';
		}
	}
	return '';
}
END_OF_FUNC

#----------------------------------------
'select_rows' => <<'END_OF_FUNC',
sub select_rows {
	my ($class,$pobj) = @_;
	return '0';			# default is dropdown list style
}
END_OF_FUNC

#----------------------------------------
'select_size' => <<'END_OF_FUNC',
sub select_size {
	my ($class,$pobj) = @_;
	# Length of select option character strings which determines
	# the width of dropdown list on screen.
	return $pobj->{select_length} if $pobj->{select_length};
	return 0;	# arbitrary length
}
END_OF_FUNC

#----------------------------------------
'select_length' => <<'END_OF_FUNC',
sub select_length {
	my ($class,$pobj) = @_;
	# Synonym for select_size
	return $pobj->select_size || 0;
}
END_OF_FUNC

#----------------------------------------
'query_where' => <<'END_OF_FUNC',
sub query_where {
	my ($class,$pobj) = @_;
	return '1 = 1' if $pobj->query_all;
	return $pobj->select_where;
}
END_OF_FUNC

#----------------------------------------
'list' => <<'END_OF_FUNC',
sub list {
	my ($class,$pobj) = @_;
	# Return list of table's column names
	if (my $s = $pobj->{select}) {
		$s =~ s/^.*?,\s*//;	# remove value field from list
		return $s;
	}
	my $n = $pobj->name;
	return "$n." . (join ", $n.", $pobj->{id}, $pobj->{object}->fields)
	  if $pobj->{id};
	return "$n." . (join ", $n.", $pobj->{object}->fields);
}
END_OF_FUNC

#----------------------------------------
'list_sql' => <<'END_OF_FUNC',
use Acme::Base::Schema::Join	qw(parse_join);
sub list_sql {
	my ($class,$pobj) = @_;
	my $list = $pobj->list or return '';
	return $list if $list =~ m/^select\b/i;

	# Construct query from list table
	my $t = $pobj->{object};
	my ($join_expr, $join_clause) = parse_join($t->name, $pobj->list_join);

	# Note: returned query does not include $t->restrict clauses as it
	# is assumed that it will be combined with a search script and then
	# cleaned by Acme::Security::clean_sql(), and it is at that time that
	# the restrict clauses will be added.

	my $where    = fully_qualify($t,$pobj->list_where);

	return 'select'
	  . ' distinct ' . $pobj->list_fields .			"\n"
	  . ' from '     . $join_expr .		                "\n"
	  . ' where '    . $where  . $join_clause .		"\n"
	  . ' order by ' . $pobj->list_order_by_clause .	"\n";
}
END_OF_FUNC

#----------------------------------------
'list_fields' => <<'END_OF_FUNC',
sub list_fields {
	my ($class,$pobj) = @_;
	# Return list of table's fully qualified column names
	my $n = $pobj->name;
	my @f;		# returned fields
	for (split /,\s*/, $pobj->list) {
		if (/\./) {
			push @f, $_;	# field already fully qualified
		}
		else {
			push @f, "$n.$_";
		}
	}
	return join ', ', @f;
}
END_OF_FUNC

#----------------------------------------
'list_join' => <<'END_OF_FUNC',
sub list_join {
	my ($class,$pobj) = @_;
	# Default join clause
	# I debated returning $t->join, but decided not to.
	return '';
}
END_OF_FUNC

#----------------------------------------
'list_where' => <<'END_OF_FUNC',
sub list_where {
	my ($class,$pobj) = @_;
	# Default where clause is to select all
	return '1 = 1';
}
END_OF_FUNC

#----------------------------------------
'list_order' => <<'END_OF_FUNC',
sub list_order {
	my ($class,$pobj) = @_;
	return $pobj->{list_order_by} || $pobj->{select_order} || '1';
}
END_OF_FUNC

#----------------------------------------
'list_order_by_clause' => <<'END_OF_FUNC',
sub list_order_by_clause {
	my ($class,$pobj) = @_;

	return fully_qualify($pobj->{object},$pobj->list_order) || 1;
}
END_OF_FUNC

#----------------------------------------
'list_format' => <<'END_OF_FUNC',
sub list_format {
	my ($class,$pobj) = @_;

	# Build list format from individual field formats.
	my $fmt = '';
	my ($tbl,$fld,$fp);
	for (split(/,\s*/,$pobj->list_fields)) {
		($tbl,$fld) = /(.*)\.(.*)/;
		$fp = $pobj->object->schema->t($tbl)->f($fld);
		$fmt .= ($fmt ? ', ' : '') . $fp->format if $fp;
	}
	return $fmt;
}
END_OF_FUNC

#----------------------------------------
'list_header' => <<'END_OF_FUNC',
sub list_header {
	my ($class,$pobj) = @_;
	return '';
}
END_OF_FUNC

#----------------------------------------
'list_footer' => <<'END_OF_FUNC',
sub list_footer {
	my ($class,$pobj) = @_;
	return '';
}
END_OF_FUNC

#----------------------------------------
'list_title' => <<'END_OF_FUNC',
sub list_title {
	my ($class,$pobj) = @_;
	# Capitalize name
	my $title = $pobj->{name};
	$title =~ s/_/ /g;
	$title =~ s/\b(\w+)\b/ucfirst(lc($1))/eg;
	return $title . " List";
}
END_OF_FUNC

#----------------------------------------
'list_total' => <<'END_OF_FUNC',
sub list_total {
	my ($class,$pobj) = @_;
	return $pobj->{list_totals} || '';
}
END_OF_FUNC

#----------------------------------------
'list_total_fields' => <<'END_OF_FUNC',
sub list_total_fields {
	my ($class,$pobj) = @_;
	# Return list of table's fully qualified column names
	my $n = $pobj->name;
	my @f;		# returned totals
	for (split /,\s*/, $pobj->list_total) {
		if (/\./) {
			push @f, $_;	# field already fully qualified
		}
		else {
			push @f, "$n.$_";
		}
	}
	return join ', ', @f if @f;
	return '';
}
END_OF_FUNC

#----------------------------------------
'list_group' => <<'END_OF_FUNC',
sub list_group {
	my ($class,$pobj) = @_;
	return $pobj->{list_groups} || '';
}
END_OF_FUNC

#----------------------------------------
'list_group_fields' => <<'END_OF_FUNC',
sub list_group_fields {
	my ($class,$pobj) = @_;
	# Return list of table's fully qualified column names
	my $n = $pobj->name;
	my @f;		# returned fields
	for (split /,\s*/, $pobj->list_group) {
		if (/\./) {
			push @f, $_;	# field already fully qualified
		}
		else {
			push @f, "$n.$_";
		}
	}
	return join ', ', @f if @f;
	return '';
}
END_OF_FUNC

#----------------------------------------
'list_filter' => <<'END_OF_FUNC',
sub list_filter {
	my ($class,$pobj) = @_;
	return $pobj->{list_filters} || '';
}
END_OF_FUNC

#----------------------------------------
'list_filter_fields' => <<'END_OF_FUNC',
sub list_filter_fields {
	my ($class,$pobj) = @_;
	# Return list of table's fully qualified column names
	my $n = $pobj->name;
	my @f;		# returned fields
	for (split /,\s*/, $pobj->list_filter) {
		if (/\./) {
			push @f, $_;	# field already fully qualified
		}
		else {
			push @f, "$n.$_";
		}
	}
	return join ', ', @f if @f;
	return '';
}
END_OF_FUNC

#----------------------------------------
'list_selector' => <<'END_OF_FUNC',
sub list_selector {
	my ($class,$pobj) = @_;
	return '(all)';
}
END_OF_FUNC

#----------------------------------------
'list_selector_cond' => <<'END_OF_FUNC',
sub list_selector_cond {
	my ($class,$pobj) = @_;
	return 'not null';
}
END_OF_FUNC

#----------------------------------------
'list_selector_description' => <<'END_OF_FUNC',
sub list_selector_description {
	my ($class,$pobj) = @_;
	# Allow list_selector_desc as an abbreviation to description.
	return $pobj->{list_selector_desc} || '';
}
END_OF_FUNC

#----------------------------------------
'list_selector_param1' => <<'END_OF_FUNC',
sub list_selector_param1 {
	my ($class,$pobj) = @_;
	# Allow list_selector_param as an abbreviation to param1.
	return $pobj->{list_selector_param} || '';
}
END_OF_FUNC

#----------------------------------------
'list_buttons_html' => <<'END_OF_FUNC',
sub list_buttons_html {
	my ($class,$pobj) = @_;

	# At this time, there are no other list_buttons options
	# to assimilate into HTML. The schema must contain the
	# full HTML code for the buttons, like:
	#   <td><a href="myurl.com" title="Popup help text">My Button</a></td>

	return '';
}
END_OF_FUNC

#----------------------------------------
'pick' => <<'END_OF_FUNC',
sub pick {
	my ($class,$pobj) = @_;
	return $pobj->object->key . ', ' . $pobj->list;
}
END_OF_FUNC

#----------------------------------------
'pick_sql' => <<'END_OF_FUNC',
use Acme::Base::Schema::Join	qw(parse_join);
sub pick_sql {
	my ($class,$pobj) = @_;
	my $pick = $pobj->pick or return '';
	return $pick if $pick =~ m/^select\b/i;

	# Construct query from list table
	my $t = $pobj->{object};
	my ($join_expr, $join_clause) = parse_join($t->name, $pobj->pick_join);

	# Note: returned query does not include $t->restrict clauses as it
	# is assumed that it will be combined with a search script and then
	# cleaned by Acme::Security::clean_sql(), and it is at that time that
	# the restrict clauses will be added.

	my $where    = fully_qualify($t,$pobj->pick_where);

	return 'select'
	  . ' distinct ' . $pobj->pick_fields .			"\n"
	  . ' from '     . $join_expr .				"\n"
	  . ' where '    . $where  . $join_clause .		"\n"
	  . ' order by ' . $pobj->pick_order_by_clause .	"\n";
}
END_OF_FUNC

#----------------------------------------
'pick_fields' => <<'END_OF_FUNC',
sub pick_fields {
	my ($class,$pobj) = @_;
	# Return pick of table's fully qualified column names
	# Like select_fields, the first field is the key field.
	my $n = $pobj->name;
	my @f;		# returned fields
	for (split /,\s*/, $pobj->pick) {
		if (/\./) {
			push @f, $_;	# field already fully qualified
		}
		else {
			push @f, "$n.$_";
		}
	}
	return join ', ', @f;
}
END_OF_FUNC

#----------------------------------------
'pick_join' => <<'END_OF_FUNC',
sub pick_join {
	my ($class,$pobj) = @_;
	return $pobj->list_join;
}
END_OF_FUNC

#----------------------------------------
'pick_where' => <<'END_OF_FUNC',
sub pick_where {
	my ($class,$pobj) = @_;
	# Default where clause is to select all
	return '1 = 1';
}
END_OF_FUNC

#----------------------------------------
'pick_order' => <<'END_OF_FUNC',
sub pick_order {
	my ($class,$pobj) = @_;
	return $pobj->list_order;
}
END_OF_FUNC

#----------------------------------------
'pick_order_by_clause' => <<'END_OF_FUNC',
sub pick_order_by_clause {
	my ($class,$pobj) = @_;

	return fully_qualify($pobj->{object},$pobj->pick_order) || 1;
}
END_OF_FUNC

#----------------------------------------
'pick_format' => <<'END_OF_FUNC',
sub pick_format {
	my ($class,$pobj) = @_;
	return $pobj->list_format;
}
END_OF_FUNC

#----------------------------------------
'pick_header' => <<'END_OF_FUNC',
sub pick_header {
	my ($class,$pobj) = @_;
	return $pobj->list_header if ! $pobj->{pick};
	return '';
}
END_OF_FUNC

#----------------------------------------
'pick_footer' => <<'END_OF_FUNC',
sub pick_footer {
	my ($class,$pobj) = @_;
	return $pobj->list_footer if ! $pobj->{pick};
	return '';
}
END_OF_FUNC

#----------------------------------------
'pick_total' => <<'END_OF_FUNC',
sub pick_total {
	my ($class,$pobj) = @_;
	return $pobj->{pick_totals} if $pobj->{pick_totals};
	return $pobj->list_total || '';
}
END_OF_FUNC

#----------------------------------------
'pick_total_fields' => <<'END_OF_FUNC',
sub pick_total_fields {
	my ($class,$pobj) = @_;
	# Return list of table's fully qualified column names
	my $n = $pobj->name;
	my @f;		# returned totals
	for (split /,\s*/, $pobj->pick_total) {
		if (/\./) {
			push @f, $_;	# field already fully qualified
		}
		else {
			push @f, "$n.$_";
		}
	}
	return join ', ', @f if @f;
	return '';
}
END_OF_FUNC

#----------------------------------------
'pick_title' => <<'END_OF_FUNC',
sub pick_title {
	my ($class,$pobj) = @_;
	return $pobj->list_title;
}
END_OF_FUNC

#----------------------------------------
'form_list' => <<'END_OF_FUNC',
sub form_list {
	my ($class,$pobj) = @_;
	# Return list of table's column names
	if (my $l = $pobj->list) {
		return $l;
	}
	my $n = $pobj->name;
	return "$n." . (join ", $n.", $pobj->{id}, $pobj->{object}->fields)
	  if $pobj->{id};
	return "$n." . (join ", $n.", $pobj->{object}->fields);
}
END_OF_FUNC

#----------------------------------------
'form_list_sql' => <<'END_OF_FUNC',
use Acme::Base::Schema::Join	qw(parse_join);
sub form_list_sql {
	my ($class,$pobj) = @_;

	my $list = $pobj->form_list or return '';
	return $list if $list =~ m/^select\b/i;

	# Construct query from list table
	my $t = $pobj->{object};
	my ($join_expr, $join_clause) = parse_join($t->name, $pobj->form_list_join);

	my $where    = fully_qualify($t,$pobj->form_list_where);

	return 'select' 
	  . ' distinct ' . $pobj->form_list_fields .		"\n"
	  . ' from '     . $join_expr .				"\n"
	  . ' where '    . $where . $join_clause .		"\n"
	  . ' order by ' . $pobj->form_list_order_by_clause .	"\n";
}
END_OF_FUNC

#----------------------------------------
'form_list_fields' => <<'END_OF_FUNC',
sub form_list_fields {
	my ($class,$pobj) = @_;
	# Return list of table's fully qualified column names
	my $n = $pobj->name;
	my @f;		# returned fields
	for (split /,\s*/, $pobj->form_list) {
		if (/\./) {
			push @f, $_;	# field already fully qualified
		}
		else {
			push @f, "$n.$_";
		}
	}
	return join ', ', @f;
}
END_OF_FUNC

#----------------------------------------
'form_list_join' => <<'END_OF_FUNC',
sub form_list_join {
	my ($class,$pobj) = @_;
	return $pobj->list_join;
}
END_OF_FUNC

#----------------------------------------
'form_list_where' => <<'END_OF_FUNC',
sub form_list_where {
	my ($class,$pobj) = @_;
	# Default where clause is to select all
	return '1 = 1';
}
END_OF_FUNC

#----------------------------------------
'form_list_order' => <<'END_OF_FUNC',
sub form_list_order {
	my ($class,$pobj) = @_;
	return $pobj->list_order || '1';
}
END_OF_FUNC

#----------------------------------------
'form_list_order_by_clause' => <<'END_OF_FUNC',
sub form_list_order_by_clause {
	my ($class,$pobj) = @_;

	return fully_qualify($pobj->{object},$pobj->form_list_order) || 1;
}
END_OF_FUNC

#----------------------------------------
'form_list_format' => <<'END_OF_FUNC',
sub form_list_format {
	my ($class,$pobj) = @_;
	return $pobj->list_format;
}
END_OF_FUNC

#----------------------------------------
'form_list_header' => <<'END_OF_FUNC',
sub form_list_header {
	my ($class,$pobj) = @_;
	# Don't use list_header if form_list parameter was specified.
	return $pobj->list_header if ! $pobj->{form_list};
	return '';
}
END_OF_FUNC

#----------------------------------------
'form_list_footer' => <<'END_OF_FUNC',
sub form_list_footer {
	my ($class,$pobj) = @_;
	# Don't use list_footer if form_list parameter was specified.
	return $pobj->list_footer if ! $pobj->{form_list};
	return '';
}
END_OF_FUNC

#----------------------------------------
'form_list_total' => <<'END_OF_FUNC',
sub form_list_total {
	my ($class,$pobj) = @_;
	return $pobj->{form_list_totals} if $pobj->{form_list_totals};
	return $pobj->list_total || '';
}
END_OF_FUNC

#----------------------------------------
'form_list_total_fields' => <<'END_OF_FUNC',
sub form_list_total_fields {
	my ($class,$pobj) = @_;
	# Return list of table's fully qualified column names
	my $n = $pobj->name;
	my @f;		# returned totals
	for (split /,\s*/, $pobj->form_list_total) {
		if (/\./) {
			push @f, $_;	# field already fully qualified
		}
		else {
			push @f, "$n.$_";
		}
	}
	return join ', ', @f if @f;
	return '';
}
END_OF_FUNC

#----------------------------------------
'form_list_title' => <<'END_OF_FUNC',
sub form_list_title {
	my ($class,$pobj) = @_;
	return $pobj->list_title;
}
END_OF_FUNC

#----------------------------------------
'detail' => <<'END_OF_FUNC',
use Acme::SQL;
sub detail {
	my ($class,$pobj) = @_;
	return $pobj->{detail_sql} || '';
}
END_OF_FUNC

#----------------------------------------
'detail_field' => <<'END_OF_FUNC',
use Acme::SQL;
sub detail_field {
	my ($class,$pobj) = @_;
	my $detail = $pobj->{detail} || '';
	return $detail if $detail !~ m/^select\b/i;

	my $detail_sql = $pobj->{detail_sql} or return '';

	# Derive detail table.field from first field detail_sql.
	my $sqlo = new Acme::SQL $detail_sql;
	my ($field) = $sqlo->dbcolumns;		# get first field in query
	return $field if $field =~ /\./;	# field is fully qualified
	my ($table) = $sqlo->dbtables;		# get first table in query
	return "$table.$field";
}
END_OF_FUNC

#----------------------------------------
'detail_table' => <<'END_OF_FUNC',
sub detail_table {
	# Helper function
	my ($class,$pobj) = @_;
	my $detail = $pobj->detail_field or return;
	my ($table) = $detail =~ /(\w+)/;
	return $table || '';
}
END_OF_FUNC

#----------------------------------------
'detail_sql' => <<'END_OF_FUNC',
use Acme::Base::Schema::Join	qw(parse_join);
sub detail_sql {
	my ($class,$pobj) = @_;

	my $detail = $pobj->{detail} || '';
	return $detail if $detail =~ m/^select\b/i;

	my $table  = $pobj->detail_table;
	my $t      = $pobj->{object}->schema->t($table) or return ''; 

	my ($join_expr, $join_clause) = parse_join($table,$pobj->detail_join);

	my $fields = $pobj->detail_list_fields;

	my $where  = fully_qualify($t,$pobj->detail_where);

	my $order_by = fully_qualify($t,$pobj->detail_order);

	return 'select'
	     . ' distinct ' . $fields
	     . ' from '     . $join_expr
	     . ' where '    . $where . $join_clause
	     . ' order by ' . $order_by;
}
END_OF_FUNC

#----------------------------------------
'detail_join' => <<'END_OF_FUNC',
sub detail_join {
	my ($class,$pobj) = @_;
	# Get list clause from target table properties.
	my $t = get_detail_target($pobj) or return '';
	return $t->form_list_join;
}
END_OF_FUNC

#----------------------------------------
'detail_list' => <<'END_OF_FUNC',
sub detail_list {
	my ($class,$pobj) = @_;
	return $pobj->detail_list_fields;
}
END_OF_FUNC

#----------------------------------------
'detail_list_fields' => <<'END_OF_FUNC',
sub detail_list_fields {
	my ($class,$pobj) = @_;
	if ($pobj->{detail_list}) {
		my $table = $pobj->detail_table or return;
		my $t = $pobj->{object}->schema->t($table); 
		return fully_qualify($t,$pobj->{detail_list});
	}
	else {
		# Get list clause from target table properties.
		my $t = get_detail_target($pobj) or return '';
		return $t->form_list_fields;
	}
}
END_OF_FUNC

#----------------------------------------
'detail_where' => <<'END_OF_FUNC',
sub detail_where {
	my ($class,$pobj) = @_;
	my $t = get_detail_target($pobj) or return '';
	return $t->form_list_where;
}
END_OF_FUNC

#----------------------------------------
'detail_order' => <<'END_OF_FUNC',
sub detail_order {
	my ($class,$pobj) = @_;
	# Get list clause from target table properties.
	my $t = get_detail_target($pobj) or return '';
	return $t->form_list_order;
}
END_OF_FUNC

#----------------------------------------
'detail_order_by_clause' => <<'END_OF_FUNC',
sub detail_order_by_clause {
	my ($class,$pobj) = @_;

	return fully_qualify($pobj->{object},$pobj->detail_order) || 1;
}
END_OF_FUNC

#----------------------------------------
'detail_format' => <<'END_OF_FUNC',
sub detail_format {
	my ($class,$pobj) = @_;
	# Get list clause from target table properties.
	my $t = get_detail_target($pobj) or return '';
	return $t->form_list_format;
}
END_OF_FUNC

#----------------------------------------
'detail_header' => <<'END_OF_FUNC',
sub detail_header {
	my ($class,$pobj) = @_;
	# Get list clause from target table properties.
	my $t = get_detail_target($pobj) or return '';
	return $t->form_list_header || '';
}
END_OF_FUNC

#----------------------------------------
'detail_footer' => <<'END_OF_FUNC',
sub detail_footer {
	my ($class,$pobj) = @_;
	# Get list clause from target table properties.
	my $t = get_detail_target($pobj) or return '';
	return $t->form_list_footer || '';
}
END_OF_FUNC

#----------------------------------------
'detail_total' => <<'END_OF_FUNC',
sub detail_total {
	my ($class,$pobj) = @_;
	return $pobj->{detail_totals} if $pobj->{detail_totals};
	my $t = get_detail_target($pobj) or return '';
	return $t->form_list_total || '';
}
END_OF_FUNC

#----------------------------------------
'detail_total_fields' => <<'END_OF_FUNC',
sub detail_total_fields {
	my ($class,$pobj) = @_;
	# Return list of table's fully qualified column names
	my $t = get_detail_target($pobj) or return '';
	my $n = $t->name;
	my @f;		# returned totals
	for (split /,\s*/, $pobj->detail_total) {
		if (/\./) {
			push @f, $_;	# field already fully qualified
		}
		else {
			push @f, "$n.$_";
		}
	}
	return join ', ', @f if @f;
	return '';
}
END_OF_FUNC

#----------------------------------------
'detail_title' => <<'END_OF_FUNC',
sub detail_title {
	my ($class,$pobj) = @_;
	# Get list clause from target table properties.
	my $t = get_detail_target($pobj) or return '';
	return $t->form_list_title;
}
END_OF_FUNC

#----------------------------------------
'email_list' => <<'END_OF_FUNC',
sub email_list {
	my ($class,$pobj) = @_;
	# Return list of table's column names
	my $n = $pobj->name;
	return "$n." . (join ", $n.", $pobj->{id}, $pobj->{object}->fields)
	  if $pobj->{id};
	return "$n." . (join ", $n.", $pobj->{object}->fields);
}
END_OF_FUNC

#----------------------------------------
'email_list_fields' => <<'END_OF_FUNC',
sub email_list_fields {
	my ($class,$pobj) = @_;
	# Return list of table's fully qualified column names
	my $n = $pobj->name;
	my @f;		# returned fields
	for (split /,\s*/, $pobj->email_list) {
		if (/\./) {
			push @f, $_;	# field already fully qualified
		}
		else {
			push @f, "$n.$_";
		}
	}
	return join ', ', @f;
}
END_OF_FUNC

#----------------------------------------
'email_list_sql' => <<'END_OF_FUNC',
use Acme::Base::Schema::Join	qw(parse_join);
sub email_list_sql {
	my ($class,$pobj) = @_;

	my $list = $pobj->email_list or return '';
	return $list if $list =~ m/^select\b/i;

	# Construct query from list table
	# Use standard list clauses except for columns, which can be set.
	my $t = $pobj->{object};
	my ($join_expr, $join_clause) = parse_join($t->name, $pobj->list_join);

	my $where    = fully_qualify($t,$pobj->list_where);

	return 'select' 
	  . ' distinct ' . $pobj->email_list_fields .		"\n"
	  . ' from '     . $join_expr .				"\n"
	  . ' where '    . $where . $join_clause .		"\n"
	  . ' order by ' . $pobj->list_order_by_clause .	"\n";
}
END_OF_FUNC

#----------------------------------------
'stylesheet' => <<'END_OF_FUNC',
sub stylesheet {
	# Synonym for Stylesheet
	my ($class,$pobj) = @_;
	return $pobj->{form_stylesheet} || $pobj->{stylesheets};
}
END_OF_FUNC

#----------------------------------------
'stylesheets' => <<'END_OF_FUNC',
sub stylesheets {
	# Synonym for Stylesheet
	my ($class,$pobj) = @_;
	return $pobj->{form_stylesheet} || $pobj->{stylesheet};
}
END_OF_FUNC

#----------------------------------------
'form_stylesheet_html' => <<'END_OF_FUNC',
sub form_stylesheet_html {
	my ($class,$pobj) = @_;
	# Internal routine
	return parse_stylesheet(
		$pobj->{form_stylesheet}  || $pobj->{form_stylesheets} ||
		$pobj->{stylesheet}	  || $pobj->{stylesheets}
		);
}
END_OF_FUNC

#----------------------------------------
'list_stylesheet_html' => <<'END_OF_FUNC',
sub list_stylesheet_html {
	my ($class,$pobj) = @_;
	# Internal routine
	return parse_stylesheet(
		$pobj->{list_stylesheet}  || $pobj->{list_stylesheets} ||
		$pobj->{stylesheet}	  || $pobj->{stylesheets}
		);
}
END_OF_FUNC

#----------------------------------------
'pick_stylesheet_html' => <<'END_OF_FUNC',
sub pick_stylesheet_html {
	my ($class,$pobj) = @_;
	# Internal routine
	return parse_stylesheet(
		$pobj->{pick_stylesheet}  || $pobj->{pick_stylesheets} ||
		$pobj->{stylesheet}	  || $pobj->{stylesheets}
		);
}
END_OF_FUNC

#----------------------------------------
'view_stylesheet_html' => <<'END_OF_FUNC',
sub view_stylesheet_html {
	my ($class,$pobj) = @_;
	# Internal routine
	return parse_stylesheet(
		$pobj->{view_stylesheet}  || $pobj->{view_stylesheets} ||
		$pobj->{stylesheet}	  || $pobj->{stylesheets}
		);
}
END_OF_FUNC

#----------------------------------------
'print_stylesheet' => <<'END_OF_FUNC',
sub print_stylesheet {
	# Synonym for Stylesheet
	my ($class,$pobj) = @_;
	return $pobj->{form_print_stylesheet} || $pobj->{print_stylesheets};
}
END_OF_FUNC

#----------------------------------------
'print_stylesheets' => <<'END_OF_FUNC',
sub print_stylesheets {
	# Synonym for Stylesheet
	my ($class,$pobj) = @_;
	return $pobj->{form_print_stylesheet} || $pobj->{print_stylesheet};
}
END_OF_FUNC

#----------------------------------------
'form_print_stylesheet_html' => <<'END_OF_FUNC',
sub form_print_stylesheet_html {
	my ($class,$pobj) = @_;
	# Internal routine
	return parse_stylesheet(
		$pobj->{form_print_stylesheet} || $pobj->{form_print_stylesheets} ||
		$pobj->{print_stylesheet}      || $pobj->{print_stylesheets}
		);
}
END_OF_FUNC

#----------------------------------------
'list_print_stylesheet_html' => <<'END_OF_FUNC',
sub list_print_stylesheet_html {
	my ($class,$pobj) = @_;
	# Internal routine
	return parse_stylesheet(
		$pobj->{list_print_stylesheet} || $pobj->{list_print_stylesheets} ||
		$pobj->{print_stylesheet}      || $pobj->{print_stylesheets}
		);
}
END_OF_FUNC

#----------------------------------------
'view_print_stylesheet_html' => <<'END_OF_FUNC',
sub view_print_stylesheet_html {
	my ($class,$pobj) = @_;
	# Internal routine
	return parse_stylesheet(
		$pobj->{view_print_stylesheet} || $pobj->{view_print_stylesheets} ||
		$pobj->{print_stylesheet}      || $pobj->{print_stylesheets}
		);
	}
END_OF_FUNC

#----------------------------------------
'include_html' => <<'END_OF_FUNC',
sub include_html {
	# Synonym for include
	my ($class,$pobj) = @_;
	return parse_include($pobj->{include} || $pobj->{includes});
}
END_OF_FUNC

#----------------------------------------
'form_include_html' => <<'END_OF_FUNC',
sub form_include_html {
	my ($class,$pobj) = @_;
	# Internal routine
	return parse_include(
		$pobj->{form_include} || $pobj->{form_includes} ||
		$pobj->{include}      || $pobj->{includes}
		);
}
END_OF_FUNC

#----------------------------------------
'query_include_html' => <<'END_OF_FUNC',
sub query_include_html {
	my ($class,$pobj) = @_;
	# Internal routine
	return parse_include($pobj->{query_include} || $pobj->{query_includes});
}
END_OF_FUNC

#----------------------------------------
'list_include_html' => <<'END_OF_FUNC',
sub list_include_html {
	my ($class,$pobj) = @_;
	# Internal routine
	return parse_include($pobj->{list_include} || $pobj->{list_includes});
}
END_OF_FUNC

#----------------------------------------
'pick_include_html' => <<'END_OF_FUNC',
sub pick_include_html {
	my ($class,$pobj) = @_;
	# Internal routine
	return parse_include($pobj->{pick_include} || $pobj->{pick_includes});
}
END_OF_FUNC

#----------------------------------------
'view_include_html' => <<'END_OF_FUNC',
sub view_include_html {
	my ($class,$pobj) = @_;
	# Internal routine
	return parse_include($pobj->{view_include} || $pobj->{view_includes});
}
END_OF_FUNC

#----------------------------------------
'limit_abort' => <<'END_OF_FUNC',
sub limit_abort {
	my ($class,$pobj) = @_;
	# Internal routine
	return $pobj->{limit} if defined $pobj->{limit};
	return Acme::Config->List_Limit_Abort;
}
END_OF_FUNC

#----------------------------------------
'limit_warn' => <<'END_OF_FUNC',
sub limit_warn {
	my ($class,$pobj) = @_;
	return Acme::Config->List_Limit_Warn;
}
END_OF_FUNC

#----------------------------------------
'export' => <<'END_OF_FUNC',
sub export {
	my ($class,$pobj) = @_;
	return '' if $pobj->{noexport} && $pobj->{noexport} !~ /^no$/i;
	return 'yes';	# Default
}
END_OF_FUNC

#----------------------------------------
'noexport' => <<'END_OF_FUNC',
sub noexport {
	my ($class,$pobj) = @_;
	return 1 if $pobj->{export} && $pobj->{export} =~ /^no$/i;
	return '';	# Default is no, can export
}
END_OF_FUNC

#----------------------------------------
'export_log' => <<'END_OF_FUNC',
sub export_log {
	my ($class,$pobj) = @_;
	return '' if $pobj->{noexport_log} && $pobj->{noexport_log} !~ /^no$/i;
	return Acme::Config->Export_Log;	# Default
}
END_OF_FUNC

#----------------------------------------
'noexport_log' => <<'END_OF_FUNC',
sub noexport_log {
	my ($class,$pobj) = @_;
	return 1 if $pobj->{export_log} && $pobj->{export_log} =~ /^no$/i;
	return '';	# Default is no, log export
}
END_OF_FUNC

#----------------------------------------
'email' => <<'END_OF_FUNC',
sub email {
	my ($class,$pobj) = @_;
	return '' if $pobj->{noemail} && $pobj->{noemail} !~ /^no$/i;
	return 'yes' if $pobj->email_field_list_fields;	# Default
	return '';
}
END_OF_FUNC

#----------------------------------------
'secure' => <<'END_OF_FUNC',
sub secure {
	my ($class,$pobj) = @_;
	return $pobj->security_code;
}
END_OF_FUNC

#----------------------------------------
'security_code' => <<'END_OF_FUNC',
sub security_code {
	my ($class,$pobj) = @_;
	return $pobj->{security_codes} if $pobj->{security_codes};
	return '' if ! $pobj->{secure};

	# return table name as a default security code (id)
	return $pobj->{name} if $pobj->{secure} eq 'yes';

	return $pobj->{secure};
}
END_OF_FUNC

#----------------------------------------
'restrict_codes' => <<'END_OF_FUNC',
sub restrict_codes {
	my ($class,$pobj) = @_;
	# Return list of restriction codes applying to user
	return $pobj->restrict || '';
}
END_OF_FUNC

#----------------------------------------
'restrict_where' => <<'END_OF_FUNC',
sub restrict_where {
	my ($class,$pobj) = @_;

	# Add restrictions

	use Acme::Security;

	my $where = '';
	my $r;
	for (grep { Acme::Security::deny('A',$_) } split /\W+/,$pobj->restrict_codes) {
		$r = eval "\$pobj->restrict_$_";
		$where .= ' and ' . $r if $r;
	}
	return $where;
}
END_OF_FUNC

#----------------------------------------
'restrict_where_insert' => <<'END_OF_FUNC',
sub restrict_where_insert {
	my ($class,$pobj) = @_;

	# Add restrictions

	use Acme::Security;

	my $where = '';
	my $r;
	for (grep { Acme::Security::deny('I',$_) } split /\W+/,$pobj->restrict_codes) {
		$r = eval "\$pobj->restrict_$_";
		$where .= ' and ' . $r if $r;
	}
	return $where;
}
END_OF_FUNC

#----------------------------------------
'restrict_where_update' => <<'END_OF_FUNC',
sub restrict_where_update {
	my ($class,$pobj) = @_;

	# Add restrictions

	use Acme::Security;

	my $where = '';
	my $r;
	for (grep { Acme::Security::deny('U',$_) } split /\W+/,$pobj->restrict_codes) {
		$r = eval "\$pobj->restrict_$_";
		$where .= ' and ' . $r if $r;
	}
	return $where;
}
END_OF_FUNC

#----------------------------------------
'restrict_where_delete' => <<'END_OF_FUNC',
sub restrict_where_delete {
	my ($class,$pobj) = @_;

	# Add restrictions

	use Acme::Security;

	my $where = '';
	my $r;
	for (grep { Acme::Security::deny('D',$_) } split /\W+/,$pobj->restrict_codes) {
		$r = eval "\$pobj->restrict_$_";
		$where .= ' and ' . $r if $r;
	}
	return $where;
}
END_OF_FUNC

#----------------------------------------
'export_secure' => <<'END_OF_FUNC',
sub export_secure {
	my ($class,$pobj) = @_;
	return $pobj->export_security_code;
}
END_OF_FUNC

#----------------------------------------
'export_security_code' => <<'END_OF_FUNC',
sub export_security_code {
	my ($class,$pobj) = @_;
	return $pobj->{export_security_codes} if $pobj->{export_security_codes};
	return '' if ! $pobj->{export_secure};

	# return table security codes as a default
	return $pobj->security_codes if $pobj->{export_secure} eq 'yes';

	return $pobj->{export_secure};
}
END_OF_FUNC

#----------------------------------------
'email_secure' => <<'END_OF_FUNC',
sub email_secure {
	my ($class,$pobj) = @_;
	return $pobj->email_security_code;
}
END_OF_FUNC

#----------------------------------------
'email_security_code' => <<'END_OF_FUNC',
sub email_security_code {
	my ($class,$pobj) = @_;
	return $pobj->{email_security_codes} if $pobj->{email_security_codes};
	return '' if ! $pobj->{email_secure};

	# return table security codes as a default
	return $pobj->security_codes if $pobj->{email_secure} eq 'yes';

	return $pobj->{email_secure};
}
END_OF_FUNC

#----------------------------------------
'view_prog' => <<'END_OF_FUNC',
sub view_prog {
	return Acme::Config->View_Prog;
}
END_OF_FUNC

#----------------------------------------
'edit_prog' => <<'END_OF_FUNC',
sub edit_prog {
	return Acme::Config->Edit_Prog;
}
END_OF_FUNC

#----------------------------------------
'list_prog' => <<'END_OF_FUNC',
sub list_prog {
	return Acme::Config->List_Prog;
}
END_OF_FUNC

#----------------------------------------
'count_prog' => <<'END_OF_FUNC',
sub count_prog {
	return Acme::Config->Count_Prog;
}
END_OF_FUNC

#----------------------------------------
'export_prog' => <<'END_OF_FUNC',
sub export_prog {
	return Acme::Config->Export_Prog;
}
END_OF_FUNC
#----------------------------------------
'emailall_prog' => <<'END_OF_FUNC',
sub emailall_prog {
	return Acme::Config->Emailall_Prog;
}
END_OF_FUNC

#----------------------------------------
'logall_prog' => <<'END_OF_FUNC',
sub logall_prog {
	return Acme::Config->Logall_Prog;
}
END_OF_FUNC

#----------------------------------------
'editall_prog' => <<'END_OF_FUNC',
sub editall_prog {
	return Acme::Config->Editall_Prog;
}
END_OF_FUNC

#----------------------------------------
'saveall_prog' => <<'END_OF_FUNC',
sub saveall_prog {
	return Acme::Config->Saveall_Prog;
}
END_OF_FUNC

#----------------------------------------
'new_prog' => <<'END_OF_FUNC',
sub new_prog {
	return Acme::Config->New_Prog;
}
END_OF_FUNC

#----------------------------------------
'import_prog' => <<'END_OF_FUNC',
sub import_prog {
	return Acme::Config->Import_Prog;
}
END_OF_FUNC

#----------------------------------------
'query_prog' => <<'END_OF_FUNC',
sub query_prog {
	return Acme::Config->Query_Prog;
}
END_OF_FUNC

#----------------------------------------
'save_prog' => <<'END_OF_FUNC',
sub save_prog {
	return Acme::Config->Save_Prog;
}
END_OF_FUNC

#----------------------------------------
'print_prog' => <<'END_OF_FUNC',
sub print_prog {
	my ($class,$pobj) = @_;
	my $print_prog = Acme::Config->Print_Prog;
	return $print_prog if $print_prog;
	return '';
}
END_OF_FUNC

#----------------------------------------
'audit_table' => <<'END_OF_FUNC',
sub audit_table {
	my ($class,$pobj) = @_;
	return Acme::Config->audit_table_prefix . $pobj->{name}
	  if $pobj->{name};
	return '';
}
END_OF_FUNC
);

sub audit_entry	{ Acme::Config->audit_entry  || 'entry_date'    }
sub audit_update{ Acme::Config->audit_update || 'last_update'   }
sub audit_user	{ Acme::Config->audit_user   || 'last_update_by'}

sub audit_entry_type	{ Acme::Config->audit_entry_type  || 'date'       }
sub audit_update_type	{ Acme::Config->audit_update_type || 'datetime'  }
sub audit_user_type	{ Acme::Config->audit_user_type   || 'varchar(14)'}

#------------------------------------------------------------------------------
#
# Utility routines
#
#------------------------------------------------------------------------------

sub get_detail_target {
	my ($pobj) = @_;

	# Return a pointer to the table structure identified in argument.
	# This is a helper function for finding the target table of a field's
	# detail statement.

	my $detail = $pobj->{detail} or return;

	my ($table) = $detail =~ /(\w+)/;

	return $pobj->{object}->schema->t($table);
}

sub parse_stylesheet {
	my ($stylesheet) = @_;

	return '' unless $stylesheet;

	# Stylesheet is HTML code
	return $stylesheet if $stylesheet =~ /^\s*</;

	# Parse stylesheet filenames
	require Text::ParseWords;
	my $Styles = Acme::Config->Styles;

	my $stylesheets = '';
	for (Text::ParseWords::parse_line('[\s,;]+',1,$stylesheet)) {
		$stylesheets .= qq[<link rel="stylesheet" href="$Styles/$_" type="text/css">\n];
	}
	return $stylesheets;
}

sub parse_include {
	my ($include) = @_;

	return '' unless $include;

	# Script is HTML code
	return $include if $include =~ /^\s*</;

	# Parse script filenames
	require Text::ParseWords;

	my $Scripts     = Acme::Config->Scripts;
	my $Scripts_Lib = Acme::Config->Scripts_Lib;
	my $Root        = Acme::Config->Root;

	my $include_scripts = '';
	for (Text::ParseWords::parse_line('[\s,;]+',1,$include)) {
		if (-r "$Root/$Scripts/$_") {
			$include_scripts .=
		  qq[<script src="$Scripts/$_" type="text/javascript"></script>\n];
		}
		elsif (-r "$Root/$Scripts_Lib/$_") {
			$include_scripts .=
		  qq[<script src="$Scripts_Lib/$_" type="text/javascript"></script>\n];
		}
		else {
			die "Cannot find the script file $_";
		}
	}
	return $include_scripts;
}

sub fully_qualify {
	my ($t, $string) = @_;

	#
	# Return string qualifying unqualified field names found in the table
	# object pointed to by $t with the table's name. 
	#

	my $table = $t->name;

	my $fq = sub {
		my $f = shift;
		if (defined $t->{$f} || $f eq $t->id) {
			return "$table.$f";	# return fully qualified name
		}
		return $f;
	};

	# Change non-quoted, non-numeric fields in string.
	$string =~ s/(?<!['"])([A-Za-z_][\w.]*)(?!['"])/&$fq($1)/eg;

	return $string;
}

#------------------------------------------------------------------------------

sub AUTOLOAD {
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;
	return if $method eq 'DESTROY';		# gobble
	if (exists $Subs{$method}) {
		my $s = $Subs{$method};
		delete $Subs{$method};		# release memory

		if ($s) {
			eval "package $pkg; $s";
			die "${pkg}::$method(): $@" if $@;
			goto &$method;
		}
	}
	return '';				# unknown method
}

1;
