package Acme::Cache;

=head1 NAME

Acme::Cache - Return a reference to a cache of a file.

=head1 SYNOPSIS

    use Acme::Cache;

    sub do_something {
	my $struct = Acme::Cache->get
    		file        => '/etc/passwd',
		constructor => \&user_names_passwd;
	...
    }

    -or-

    $Cache = new Acme::Cache
    		file        => '/etc/passwd',
		constructor => \&user_names_passwd;

    sub do_something {
	my $struct = $Cache->get;
	...
    }

    $Cache->destroy;	# remove cached data

=head1 DESCRIPTION

This module manages a cache of a file. A cache is initialized in a global
variable and "got" (using the get() method) when needed. Periodically,
according to a configuration variable C<$Refresh_Interval>, the file is checked
to see if it has been modified since the last time the file was processed by
the specified constructor routine.

The constructor routine called with the arguments listed in the C<args> 
array followed by the file name.

To cache an object that is built by the constructor function, specify
the C<class> parameter. The first time the constructor is called, the
value of the C<class> parameter is passed to the constructor as its
first argument. After the first invocation, the value of the C<data>
argument -- the results of the constructor function -- is used as the
first argument.

If a constructor routine is not specified, then the file is processed by
Perl's C<do> function. This can be used for configuration files.

The get() method will automatically call new() with its arguments if
it is called with a class ('Acme::Cache') instead of an object.

The destroy() method removes the cached data.

If the program is running under Doug MacEachern's mod_perl, then the
file is first tested to see if it has any errors before re-reading
the file and processing it with C<do>.

=head1 EXAMPLE

    # An example caching an object

    use Acme::Cache;

    $Cache = new Acme::Cache
    		file        => '/data/datafile',
		constructor => \&add,
		class       => __PACKAGE__;

    package Acme::Other;

    sub new {
        my $class = shift;
    	
	my $self = $Cache->get;

	bless $self, $class;
    }

    sub add {		# Constructor
    	my $self = shift;

	#
	# Define object to cache if needed.
	#

	unless (ref($self)) {
		my $class = $self;
		$self = {};
		bless $self, $class;
	}

	#
	# Other processing...
	#

	$self->{param} = $value;

	#
	# Return object that is to be cached.
	#

	return $self;
    }

=head1 CHANGE LOG

 12/6/01  Initial version.
 6/5/02   Added %Caches master cache and added class parameter.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright(c) 2001,  Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Acme::Config;

#
# Define reposiory for cache objects if they are autocreated in get()
#

our %Caches;

# Turn on special checking for Doug MacEachern's modperl
# Stolen from CGI.pm

our $MOD_PERL = 0;
$MOD_PERL = $ENV{'GATEWAY_INTERFACE'} =~ /^CGI-Perl\//
  if exists $ENV{'GATEWAY_INTERFACE'};

#------------------------------------------------------------------------------

sub new { 
        my $this = shift;
        my $class = ref($this) || $this;

	my $self = {
		timestamp    	=> 0,		# time file last checked
		next_check   	=> 0,		# next check if file changed
		refresh 	=> 0,		# seconds between checks
		constructor	=> \&_do,	# function to build cache
		class           => '',		# optional class of retured object
		args		=> [],		# arguments passed to constructor
		file		=> undef,	# file's path name to check
		data		=> undef,	# constructor's return value
		@_
	};

	return bless $self, $class;
}

sub get {
	my $self = shift;

	if (! ref($self)) {
		my $class = $self;
		my %opt = @_;
		$self = $Caches{$opt{file}} || new Acme::Cache @_;

		$Caches{$self->{file}} ||= $self;	# store in master cache
	}

	my $file = $self->{file};
	my $func = $self->{constructor} || \&_do;
	my $time = time;

	# Return cached data if it is not time to check file yet.
	return $self->{data} || $self
	  if $time < $self->{next_check};
	
	# Set time to check file next.
	$self->{next_check} = $time +
	  ($self->{refresh} || Acme::Config->Refresh_Interval || 60);

	# Return cached data if file not changed since last read.
	return $self->{data} || $self
	  if $file && $self->{timestamp}
	  && (stat $file)[9] < $self->{timestamp};

	# Set flag when file last checked
	$self->{timestamp} = $time;

	if ($self->{class}) {
		# Return object.
		# Use previously constructed object or use the class parameter.
		$self->{data} ||= $self->{class};
		return ($self->{data} = &$func($self->{data},@{$self->{args}},$file)) || $self;
	}
	else {
		# Return new data in cache.
		return ($self->{data} = &$func(@{$self->{args}},$file)) || $self;
	}
}

sub destroy {
	my ($this, $file) = @_;

	undef $Caches{$file} if defined $Caches{$file};
}

#------------------------------------------------------------------------------
#
# Private routines
#
#------------------------------------------------------------------------------

sub _do {
	my $file = shift;

	# Private function. The default constructor if none is specified.
	# "Do" a file in the calling routine's package.

	# If operating under mod_perl, test the file first before
	# do'ing it.
	if ($MOD_PERL) {
		return unless _do_test($file);
	}
	
	my ($call_pkg) = caller 1;	# get package of get()'s caller
	eval qq[
		package $call_pkg;
		do "$file" or die "Cannot read the file '$file'";
		];
}

sub _do_test {
	my $file = shift;

	#
	# Check file syntax.
	# (is this neccessary? or desired?)
	#
	
	package test;
	do $file and return 1;		# successful test

	# Print to the web error log only.
	warn "Error in configuration file: $file: $!";

	return 0;			# keep same configuration settings
}

1;
