package Acme::Calendar::DayMenu;

=head1 NAME

    DayMenu - Build menus for calendar days

=head1 SYNOPSIS

    use Acme::Calendar::DayMenu;

    my $cdm = new Acme::Calendar::DayMenu;
      or
    my $cdm = new Acme::Calendar::DayMenu $begin_date;
      or
    my $cdm = new Acme::Calendar::DayMenu $begin_date,$empty_menu_re;

    $cdm->fill_menus($begin_date,$end_date,$menu_text);
    $cdm->print_menus($begin_date,$end_date);
      or
    $cdm->fill_menus($begin_date,$end_date,$menu_text,$sort_key,\&code, ...);
    $cdm->print_menus($begin_date,$end_date);
      or
    $cdm->print_menus_till($till_date);
    $cdm->print_menus($end_date);

    $cdm->set_menus_header($day_menu_header);
    $cdm->set_menus_footer($day_menu_footer);
      or
    $cdm->set_menus_header(&make_day_menu_header);
    $cdm->set_menus_footer(&make_day_menu_footer);

    my $menu = $cdm->day_menu($date);	# usually used internally


=head1 DESCRIPTION

This module provides server-side support for calendars created by the
/scripts/lib/calendars.js JavaScript routines. They are used to fill up and
print menus between date ranges that can be used as pop-up menus when the
user clicks on a date.

The fill_menus() method fills in menu items for dates in a date range.
The print_menus() method prints the menus from the first date through the
last date specified. If only one argument given, it prints from the last
day printed by print_menus_till() to the specified end date.
The print_menus_till() method prints the day menus up till the specified date.

Printing a menu deletes it from memory.

An optional argument to fill_menus() is a $sort_key which is used to sort
the daily menus before printing. The default sort order is the order which
the menu items were entered.

You need to specify a beginning date of the calendar when constructing it.
You may specify a regular expression which, if the menu matches the
regular expression, the menu will NOT be output. Empty menus don't have
any menu items nor do they have any links to add new menu items.

The set_menus_header()/set_menus_footer() methods define a header/footer to
print at the top/bottom of each menu. The argument can be a scalar or it can
be a function reference. If it is a function reference, it is called with the
current date as an argument before/after each day's menu. Other arguments 
passed are the collected codes assigned to the date (see below).

In producing menus from the database I found that I would want to alter the
display of the date buttons depending upon the menu selections. That is 
done in two parts. The first part is to collect codes depending upon the
menu (database) item, and then classify the date depending upon the codes
saved and display it using a different css classification, background color
or background image.

A list of codes can be specified as fill_menu() arguments, or a reference
to a function can be specified which will be called for each date (with the
date as an argument) and the returned codes will be associated with the date.
The codes registered for each date are returned as arguments to the functions
specified as header/footer functions.

=head1 NOTE

Dates must adhere to the format mm-dd-yyyy at this point to simplify 
program development.

These routines uses Acme::Date::dayadd() which limits dates to be within 
the years of 1970 to 2038.

=head1 CHANGE LOG

 10/17/00 Initial version.
 8/19/01  Added sort key to sort each day's entries before printing.
 4/6/02   Moved dates_between() to Acme::Date.
 4/8/02   Hacked in empty_re regular expression to recognize and
          suppress empty menus -- those menus w/o items or menu
	  links to add new items.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Acme::Date	qw(dates_between);
use Time::Local;

#------------------------------------------------------------------------------

package Acme::Calendar::DayMenuItems;

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my $self = { @_ };

	return bless $self, $class;
}

sub add {
	my $self = shift;
	my ($item, $sort_key) = @_;
	$self->{items} ||= [];
	push @{$self->{items}}, $item;

	$self->{sort_keys} ||= {};
	$sort_key ||= $#{$self->{items}};
	$self->{sort_keys}->{$#{$self->{items}}} = $sort_key;
}

sub items {
	my $self = shift;
	# Return menu items for specified date.
	#return @{$self->{items}};

	# Return sorted menu items for specified date.
	# Default order is order items were added.
	my @sorted_items;
	my $sk = $self->{sort_keys};
	for (sort { $sk->{$b} cmp $sk->{$a} } keys %$sk) {
		push @sorted_items, $self->{items}[$_];
	}
	return @sorted_items;
}

sub add_code {
	my $self = shift;
	$self->{codes} ||= [];
	push @{$self->{codes}}, @_;
}

sub codes {
	my $self = shift;
	# Return list of codes specified for date.
	return @{$self->{codes}};
}

#------------------------------------------------------------------------------

package Acme::Calendar::DayMenu;

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my $self = { 
		'.begin_date'	=> shift,
		'.header'	=> '',
		'.footer'	=> '',
		'.empty_re'	=> shift,
		};

	return bless $self, $class;
}

sub fill_menus {
	my $self = shift;
	my ($begin_date, $end_date, $menu_text, $sort_key) = @_;

	my ($date, $code);
	for $date (dates_between($begin_date, $end_date)) {

		$self->{$date} ||= new Acme::Calendar::DayMenuItems;
		$self->{$date}->add($menu_text,$sort_key);

		for $code (@_) {
			if ((ref $code) =~ /CODE/) {
				$self->{$date}->add_code(&{$code}($date));
			}
			else {
				$self->{$date}->add_code($code);
			}
		}
	}
}

sub set_menus_header {
	my ($self, $day_menu_header) = @_;
	$self->{'.header'} = $day_menu_header;
}

sub set_menus_footer {
	my ($self, $day_menu_footer) = @_;
	$self->{'.footer'} = $day_menu_footer;
}

sub print_menus {
	my ($self, $begin_date, $end_date) = @_;

	if ($begin_date && ! $end_date) {
		# Only one argument causes printing from the last date
		# print_menus_till() printed to, through the argument date.
		$end_date = $begin_date;
		$begin_date = $self->{'.begin_date'};
	}
	my $date;
	my $empty_menu = $self->{'.empty_re'};
	for $date (dates_between($begin_date, $end_date)) {
		for ($self->day_menu($date)) {
			print $_ unless ($empty_menu && /$empty_menu/);
		}
		delete $self->{$date};
	}
}

sub print_menus_till {
	my ($self,$end_date) = @_;
	return unless $self->{'.begin_date'};
	my @dates = dates_between($self->{'.begin_date'}, $end_date);
	$self->{'.begin_date'} = pop @dates; # end_date becomes next begin_date
	my $date;
	my $empty_menu = $self->{'.empty_re'};
	for $date (@dates) {
		for ($self->day_menu($date)) {
			print $_ unless ($empty_menu && /$empty_menu/);
		}
		delete $self->{$date};
	}
}

sub day_menu {
	my ($self,$date) = @_;

	my @codes;
	if ($self->{$date})	{ @codes = $self->{$date}->codes() }
	else			{ @codes = () }

	my $menu = '';
	if ($self->{'.header'}) {
		if ((ref $self->{'.header'}) =~ /CODE/) {
			$menu .= &{$self->{'.header'}}($date,@codes);
		}
		else {
			$menu .= $self->{'.header'};
		}
	}
	
	if ($self->{$date}) {
		$menu .= join "\n", $self->{$date}->items();
	}

	if ($self->{'.footer'}) {
		if ((ref $self->{'.footer'}) =~ /CODE/) {
			$menu .= &{$self->{'.footer'}}($date,@codes);
		}
		else {
			$menu .= $self->{'.footer'};
		}
	}
	return $menu;
}

1;
