package Acme::Config;

=head1 NAME

Acme::Config - Site configuration settings

=head1 SYNOPSIS

    use Acme::Config;


=head1 DESCRIPTION

This module sets up initial conditions for the Acme system at this installation.
It defines default values for common system variables, database settings,
and user settings. It is included in most programs.

Most settings are read from the site configuration file identified in the
SITE_CONF environment variable which is usually defined to be the file
C<../etc/site.conf>.

Settings are accessed as object methods (e.g. Acme::Config->DBNAME) so that
the settings may be determined dynamically.

Site specific settings that are to be determined programically should be
returned by functions defined in the I<Local::Acme::Config> module.

The configuration file is read once and cached.

When running under mod_perl, the configuration file is checked periodically
to see if it has changed, and if so, it is re-read. Note that commenting
out a setting in the configuration file does not "unset" the setting if was
defined previously. To unset a variable when this software is running under
mod_perl, you need to restart the web server.

=head1 CHANGE LOG

 3/10/00  Initial version.
 10/16/01 Changed to monitor and use a configuration file.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000,2001 Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

use Acme::Cache;
use Local::Acme::Config;

# read SITE_CONF environment variable once and untaint
our $_Config = $ENV{SITE_CONF} and ($_Config) = $_Config =~ /(.*)/;

$_Config ||= "../etc/site.conf"    if -r "../etc/site.conf";
$_Config ||= "../../etc/site.conf" if -r "../../etc/site.conf";
die "Cannot find configuration file '$_Config'" if ! $_Config;

#
# Refresh_Interval defines when caches are checked to see if they need
# refreshing. Refresh_Interval() must be defined here to avoid infinite
# recursion when initially reading the configuration file, but it still
# allows the $Refresh_Interval to be configured in the site configuration
# file.
#

our $Refresh_Interval = 60;		# initial default value
sub Refresh_Interval { return $Refresh_Interval }

#
# Construct cache and read in configuration file
#

our $_Cache = new Acme::Cache file => $_Config;

#------------------------------------------------------------------------------

sub AUTOLOAD {
        my $class = shift;

	$_Cache->get();		# refresh cache if necessary

        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;

	# Return global variable if set
	no strict 'refs';
	return $$method if defined $$method;
	if (defined %$method) {
		return %$method if wantarray;
		return \%$method;
	}
	if (defined @$method) {
		return @$method if wantarray;
		return \@$method;
	}
        return Local::Acme::Config->$method();
}

1;
