package Acme::DailyPics;

=head1 NAME

Acme::DailyPics - Daily Pictures

=head1 SYNOPSIS

    use Acme::DailyPics qw (pic_of_day, next_pic, prev_pic);

    $pic = pic_of_day();

    $pic = next_pic($pic);

    $pic = prev_pic($pic);

=head1 DESCRIPTION

Pic_of_day() method returns a picture of the day. It is chooses
the picture at random from the picture files located in
$DocumentRoot/images/pic_of_day.

Picture files are identified if they have the sufix '.gif', '.jpg', '.jpeg',
or '.png'. All other files are ignored in the directory.

Next_pic() returns the file name of the picture file next in line from 
a list of sorted file names from the directory $DocumentRoot/images/pic_of_day.

Prev_pic() returns the file name previous to the specified file in the
sorted list.

=head1 CHANGE LOG

 11/13/1998 Initial version.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1998, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut


use strict;

require 5.6.0;
require Exporter;

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(pic_of_day next_pic prev_pic);

use Acme::Config;

our $DocumentRoot = Acme::Config->DocumentRoot;# location of web files
our $Default_Pic  = Acme::Config->Default_Pic; # default if something is wrong

#------------------------------------------------------------------------------

sub pic_of_day {
	# pick a random image 

	my $dir = Acme::Config->User_Pic_Dir;	# location of daily pictures
	return $Default_Pic unless $DocumentRoot && $dir;

	opendir PICDIR, "$DocumentRoot/$dir" or return $Default_Pic;
	# Allow only graphics files to be listed.
	my @pic_files = grep { /(gif|jpg|jpeg|png)$/ } readdir PICDIR;
	closedir PICDIR;

	my $n = rand scalar(@pic_files);
	my $p = $pic_files[$n];
	return "$dir/$p" if ($p);
	return "";
}

#------------------------------------------------------------------------------

sub next_pic {
	my $curr_pic = shift;

	my $dir = Acme::Config->User_Pic_Dir;	# location of daily pictures
	return $Default_Pic unless $DocumentRoot && $dir;

	opendir PICDIR, "$DocumentRoot/$dir" or return $Default_Pic;
	# Allow only graphics files to be listed.
	my @pic_files = sort { lc($a) cmp lc($b) }
			grep { /(gif|jpg|jpeg|png)$/ }
			readdir PICDIR;
	closedir PICDIR;

	# set to first pic in list if pic has never been specified
	return "$dir/$pic_files[0]" if $curr_pic eq $Default_Pic;

	my $pic;
	for (my $n = 0; $n <= $#pic_files; $n++) {
		if ($curr_pic eq "$dir/$pic_files[$n]") {
			$n++;	# return next picture in list
			$n = 0 if $n > $#pic_files;  # wrap to first if at end
			return "$dir/$pic_files[$n]";
		}
	}

	# matching file not found, start at beginning of list
	return "$dir/$pic_files[0]";
}

#------------------------------------------------------------------------------

sub prev_pic {
	my $curr_pic = shift;

	my $dir = Acme::Config->User_Pic_Dir;	# location of daily pictures
	return $Default_Pic unless $DocumentRoot && $dir;

	opendir PICDIR, "$DocumentRoot/$dir" or return $Default_Pic;
	# Cull hidden files, *.txt files, and readme files.
	my @pic_files = sort { lc($a) cmp lc($b) }
			grep { /(gif|jpg|jpeg|png)$/ }
			readdir PICDIR;
	closedir PICDIR;

	# set to last pic in list if pic has never been specified
	return "$dir/$pic_files[$#pic_files]" if $curr_pic eq $Default_Pic;

	for (my $n = 0; $n <= $#pic_files; $n++) {
		if ($curr_pic eq "$dir/$pic_files[$n]") {
			$n--;	# return next picture in list
			$n = $#pic_files if $n < 0;  # wrap to last if at first
			return "$dir/$pic_files[$n]";
		}
	}

	# matching file not found, start at end of list
	return "$dir/$pic_files[$#pic_files]";
}

1;
