package Acme::Date;

=head1 NAME

Acme::Date - Date routines for the Acme system

=head1 SYNOPSIS

    use Acme::Date qw(today timestamp timetodate weekday 
		dayadd daydiff dates_between
		mdate meddate ldate longdate
		short_date_range long_date_range
		strtodate datetostr strtotime timetostr);

    $today = &today;

    $timestamp = &timestamp;

    $date = timetodate($time);

    ($m,$d,$y) = strtodate("1/1/2001");
    $str = datetostr($m,$d,$y);

    $time = strtotime("2001-01-01 12:01.01");	# no timezones yet
    $str = timetostr($time);

    $today_next_month = dayadd(30);
    $date = dayadd(7, "1/1/98");

    $days = daydiff($bgndate, $enddate);

    $day_of_week = weekday($date);

    print ldate($date);				# July 10, 1998

    print longdate($date);			# Wednesday July 10, 1998

    print mdate($date);				# Jul 10, 1998

    print meddate($date);			# Wed Jul 10, 1998

    print short_date_range($date1,$date2);	# 7/10-28

    print long_date_range($date1,$date2);	# July 10 - 28, 2002

    print "$_\n" for dates_between('1/1/2002', '1/31/2002');

=head1 DESCRIPTION

The today() function returns todays date as a string like "mm/dd/yyyy". It
is used quite often in database queries as a date value.

The timestamp() function returns today's time and date as a string like
"yyyy-mm-dd hh:mm:ss". It also is used in database queries.

The timetodate() function converts a time string into a date recognized by
other routines in this library (mm/dd/yyyy). It can convert internal time
numbers, a stamp in the '2000-80-12 06:30:20' format, or a string like
'Aug 12 06:30:20 2000 MDT'.

The strtotime() function converts a time string like timetodate() 
above into a Unix time (number of seconds since 1970), which timetostr()
converts back. NOTE: timezones are not correctly parsed at this time. 
These routines are quick and dirty routines made so the Lister::Selector
filter can compare time values.

The dayadd() function returns the list ($mm, $dd, $yyyy) with $mm between
0 and 11, and $yyyy between 1970 and 2038. If return value is a scalar,
the string "$mm/$dd/$yyyy" is returned with $mm between 1 and 12.

The daydiff() function returns the number of days between the two given
dates. It is like dayadd() which relies on the localtime() system function
which limits the dates to be between 1970 and 2038.

The weekday() function returns the day of week of the specified date, where
0 is Sunday and 6 is for Saturday.

The mdate() function returns a date formatted as a "medium" date, like
"Wed Jul 10, 1998". The specified date is a string formatted like
"mm/dd/yy" or "mm/dd/yyyy" or "mmddyy".

The ldate() function returns a date formatted as a "long" date, like
"July 10, 1998". The day of the week is not included. The specified
date is a string formatted like "mm/dd/yy" or "mm/dd/yyyy" or "mmddyy".

The longdate() returns a date formatted as a "long" date, like
"Wednesday July 10, 1998". The specified date is a string formatted like
"mm/dd/yy" or "mm/dd/yyyy" or "mmddyy".

The short_date_range() function returns a date range with a minimal number
of digits.  The year is not printed, and if the month of the second date
is the same as the month of the first date, the month of the second date
is suppressed.  For example, if only one date of '7/10/1998' is passed, 7/10
is returned.  If a second date of '7/28/1998' is passed, 7/10-28 is returned.

The long_date_range() function returns a date range with a minimal number
of words. It is similar in spirit to the short_date_range() function 
except full words are used for the date. For example, the date range
above would be returned as: July 10 - 28, 2002.

The dates_between() function produces a list of dates between two dates.
Dates are in the format of 'mm-dd-yyyy' or whatever the date seperator is
defined as in I<Acme::Config> (...etc/site.conf).

=head1 CHANGE LOG

 11/14/98 Initial version.
 1/15/01  Added short_date_range().
 4/6/02   Moved dates_between() here from Acme::Calendar::DayMenu.
 11/5/02  Added long_date_range(), strtodate() and datetostr().
 12/17/02 Added strtotime() and timetostr(). (quick & dirty w/o timezones)

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1998-2001 Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.000;
require Exporter;

@ISA = qw(Exporter);
@EXPORT_OK = qw(today timestamp timetodate weekday 
		dayadd daydiff dates_between
		mdate meddate ldate longdate
		short_date_range long_date_range
		strtodate datetostr strtotime timetostr);

use Acme::Config;

our $DS = Acme::Config->Date_Separator;
our $DO = Acme::Config->Date_Order;

our %Days = (			# long day names
	Sun => 'Sunday',
	Mon => 'Monday',
	Tue => 'Tuesday',
	Wed => 'Wednesday',
	Thu => 'Thursday',
	Fri => 'Friday',
	Sat => 'Saturday',
	);

our %Months = (			# long month names
	Jan => 'January',
	Feb => 'February',
	Mar => 'March',
	Apr => 'April',
	May => 'May',
	Jun => 'June',
	Jul => 'July',
	Aug => 'August',
	Sep => 'September',
	Oct => 'October',
	Nov => 'November',
	Dec => 'December',
	);

our %MonthNum = (		# month numbers
	Jan => 1,
	Feb => 2,
	Mar => 3,
	Apr => 4,
	May => 5,
	Jun => 6,
	Jul => 7,
	Aug => 8,
	Sep => 9,
	Oct => 10,
	Nov => 11,
	Dec => 12,
	);

#------------------------------------------------------------------------------

sub today {
	my ($d,$m,$y)=(localtime)[3,4,5];
	return datetostr(++$m,$d,$y+1900);
}

#------------------------------------------------------------------------------

sub timestamp {
	my $ds = Acme::Config->Date_Separator || '-';
	my $ts = Acme::Config->Time_Separator || ':';
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime;
	$mon++; $year+=1900;
	return sprintf("%d$ds%02d$ds%02d %02d$ts%02d$ts%02d",$year,$mon,$mday,$hour,$min,$sec);
}

#------------------------------------------------------------------------------

sub timetodate {
	my $time = shift;
	my $ts = Acme::Config->Time_Separator || ':';

	if ($time =~ /[A-Za-z]/) {
		# Alphanumaric type: Sat Aug 12 06:15:23 2000 MDT
		$time =~ s/ \d\d$ts\d\d$ts\d\d//;	# remove time stamp
		$time =~ s/ [A-Z][A-Z][A-Z]$//;		# remove locale (MDT)

		my ($day,$mon,$d,$y) = (split(/\W+/, $time));
		my $m = $MonthNum{$mon};

		return $time unless $m;		# date does not parse
		return datetostr($m,$d,$y);
	}
	elsif ($time =~ s/^(\d+\W\d+\W\d+) \d+:\d+:\d+$/$1/) {
		# Numeric type: 2000-08-12 06:15:23
		my ($y,$m,$d) = $time =~ /(\d+)\W(\d+)\W(\d+)/;
		return datetostr($m,$d,$y);
	}
	elsif ($time =~ /^\d+$/) {
		# Internal time stamp
		my ($d, $m, $y) = (localtime($time))[3,4,5];
		return datetostr(++$m,$d,$y+1900);
	}
}

#------------------------------------------------------------------------------

sub strtodate {
	my $str = shift or return;

	# Return an array of month (1-12), day, year parsed from string.

	# Note: The date order $DO should be called dynamically
	# for mod_perl, however for Safe processing in Acme::Searcher,
	# using the dayadd() function, the Acme::Config method cannot
	# be called here.

	#my $do = Acme::Config->Date_Order     || 'mdy';
	my $do = $DO;

	if ($str =~ /^[A-Za-z]/) {
		# Alphanumaric type: Sat Aug 12 06:15:23 2000 MDT
		my ($day,$mon,$d,$hr,$min,$sec,$y,$tz) = (split(/\W+/, $str));
		my $m = $MonthNum{$mon} or return;
		return (--$m,$d,$y);
	}

	$str =~ s|(\d\d)(\d\d)(\d\d\d*)|$1/$2/$3|;	# allow for 010198 dates

	$str =~ m|^(\d+)\W(\d+)\W(\d+)| or return;

	# Return order: mdy
	my ($m,$d,$y);
	($m,$d,$y) = ($2,$3,$1) if $do eq 'ymd' || $1 > 12;
	($m,$d,$y) = ($2,$1,$3) if $do eq 'dmy';
	($m,$d,$y) = ($1,$2,$3);

	# Adjust years to current millenium if less than 100.
	$y += 2000 if ($y  < 30);
	$y += 1900 if ($y  >= 30 && $y <= 100);

	return ($m,$d,$y);

}

#------------------------------------------------------------------------------

sub datetostr {
	my ($m, $d, $y) = @_;

	return '' unless $m && $d && $y;

	# Return date string from arguments month (1 based), day, and year.

	# Note: The date separator $DS should be called dynamically
	# for mod_perl, however for Safe processing in Acme::Searcher,
	# using the dayadd() function, the Acme::Config method cannot
	# be called here.

	#my $ds = Acme::Config->Date_Separator || '/';
	#my $do = Acme::Config->Date_Order     || 'mdy';
	my $ds = $DS;
	my $do = $DO;

	return sprintf("%02d$ds%02d$ds%02d",$y,$m,$d) if $do eq 'ymd';
	return sprintf("%02d$ds%02d$ds%02d",$d,$m,$y) if $do eq 'dmy';
	return sprintf("%02d$ds%02d$ds%02d",$m,$d,$y);
}

#------------------------------------------------------------------------------

use Time::Local;

sub strtotime {
	my $str = shift;
	my $ts = Acme::Config->Time_Separator || ':';

	if ($str =~ /^[A-Za-z]/) {
		# Alphanumaric type: Sat Aug 12 06:15:23 2000 MDT
		$str =~ s/ [A-Z][A-Z][A-Z]$//;		# remove locale (MDT)

		my ($day,$mon,$d,$hr,$min,$sec,$y) = (split(/\W+/, $str));
		my $m = $MonthNum{$mon};

		return unless $m;		# date does not parse
		return timelocal($sec, $min, $hr, $d, --$m, $y);
	}
	elsif ($str =~ s/^(\d+\W\d+\W\d+( \d+:\d+:\d+)?)/$1/) {
		# Numeric type: 2000-08-12 06:15:23
		my ($date,$time);
		($date,$time) = $str =~ /(.*)\s+(.*)/
		or ($date,$time) = ($str,'');

		my ($m,$d,$y) = strtodate($date);
		$m--;	# range from 0 to 11

		my ($hr,$min,$sec) = (split(/\W+/, $time));

		return timelocal($sec, $min, $hr, $d, $m, $y);
	}
	elsif ($str =~ /^\d+$/) {
		# Date string mmddyy
		my ($m,$d,$y) = strtodate($str);
		$m--;	# range from 0 to 11

		return timelocal(0, 0, 0, $d, $m, $y);

		# Internal time stamp
		#return $str;
	}
}

#------------------------------------------------------------------------------

sub timetostr {
	my $time = shift;
	my $ds = Acme::Config->Date_Separator || '-';
	my $ts = Acme::Config->Time_Separator || ':';
	my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime $time;
	$mon++; $year+=1900;
	return sprintf("%d$ds%02d$ds%02d %02d$ts%02d$ts%02d",$year,$mon,$mday,$hour,$min,$sec);
}

#------------------------------------------------------------------------------

use Time::Local;
require "bigint.pl";

sub dayadd {

	# Add days to date
	#	dayadd(3);		# add 3 days to today
	#	dayadd(5,"1/31/98");
	# return value is the list ($mon, $day, $year) with $mon between 0 and
	# 11, and $year between 1900 and 2038.
	# If return value is a scalar, the string "$mon/$day/$year" is returned
	# with $mon between 1 and 12.

	my $days      = shift || 0;
	my $startdate = shift || today();

	my ($mon, $day, $year) = strtodate($startdate);
	$mon -= 1;	# range from 0 to 11
	$year -= 1900 if ($year > 1900);

	my $starttime = timelocal(0, 0, 0, $day, $mon, $year);
	my $daysec = 60 * 60 * 24;

        my $newtime;
	{
		local $^W;	# turn off -w processing
		$newtime = main::badd($starttime,main::bmul($days, $daysec));
	}

        my ($nday, $nmon, $nyear) = (localtime($newtime))[3,4,5];
        if ($day == $nday && $mon == $nmon && $year == $nyear) {
                # Beware of the daylight savings bug.
                # Hack till better Date modules available.
                $newtime = main::badd($newtime,60 * 60);	# add an hour
                ($nday, $nmon, $nyear) = (localtime($newtime))[3,4,5];
        }

        $nyear += 1900;
	return ($nmon, $nday, $nyear) if wantarray;

	$nmon += 1;     # range from 1 to 12
	return datetostr($nmon, $nday, $nyear);
}

#------------------------------------------------------------------------------

sub daydiff {
	# Return number of days between 2 dates.
	# Start and end dates must be specified as "mm/dd/yy" or "mm-dd-yy".

	my $startdate = shift || today();
	my $stopdate  = shift || today();

	my ($startmon, $startday, $startyear) = strtodate($startdate);
	my ($stopmon,  $stopday,  $stopyear)  = strtodate($stopdate);

	return	_calday($stopmon,$stopday,$stopyear) -
		_calday($startmon,$startday,$startyear);
}

sub _calday {
	# Calculate a calendar-day number.
	# Helper function for daydiff.

	# This calculates a calendar-day number as described by Mark Crispin
	# in the IMAP documentation file /usr/share/doc/imap-*/calendar.txt.

	# To find the number of days between two days, calculate a pair of calendar
	# days with the formula (again, all divisions are integral), using new style
	# month/year this time:
	#                         m
	#                     m + -
	#                         8             y    y     y
	#  d + 30 * (m - 1) + ----- + y * 365 + - - --- + --- - ld
	#                       2               4   100   400
	# 
	# where:
	#  d   := day of month (1..31)
	#  m   := month in new style (January = 1..December = 12)
	#  y   := year in new style
	#  ld  := leap day correction factor:
	# 	0 for January and February in non-leap years
	# 	1 for January and February in leap years
	# 	2 for all other months in all years        
	# 
	#      In C code, the leap day correction factor is calculated as:
	#   (m < 3) ? !(y % 4) && ((y % 100) || !(y % 400)) : 2
	# 

	my ($m,$d,$y) = @_;

	use integer;	# integer arithmatic

	my $ld = ($m < 3) ? !($y % 4) && (($y % 100) || !($y % 400)) : 2;

	return	$d + 30 * ($m - 1) + ( ($m + $m/8)/2 ) +
		$y * 365 + $y/4 - $y/100 + $y/400 - $ld;
}

#------------------------------------------------------------------------------

sub weekday {
	# Return the day of the week (0-6) where 0 is Sunday.
	# This routine relies on localtime() and has its restrictions.

	my $date = shift || today();

	my ($m,$d,$y) = strtodate($date);
	$m--;

	my $time = timelocal(0, 0, 0, $d, $m, $y);

	return (localtime($time))[6];
}

#------------------------------------------------------------------------------

sub ldate {
	# Return date formatted as a "long" date like "July 10, 1998".

	# The specified date is a string formatted like mm/dd/yy or mm/dd/yyyy
	# or mmddyy.

	my $date = shift || today();	# specified as "m/d/y"

	my ($m,$d,$y) = strtodate($date);
	return $date unless $m && $d && $y;	# date does not parse
	$m--;

	my $t = timelocal(0,0,0,$d,$m,$y);

	my ($mo, $md, $yr) = (split(/\s+/,scalar(localtime($t))))[1,2,4];
	
	return "$Months{$mo} $md, $yr";
}

#------------------------------------------------------------------------------

sub longdate {
	# Return date formatted as a "long" date like "Wednesday July 10, 1998".

	# The specified date is a string formatted like mm/dd/yy or mm/dd/yyyy
	# or mmddyy.

	my $date = shift || today();	# specified as "m/d/y"

	my ($m,$d,$y) = strtodate($date);
	return $date unless $m && $d && $y;	# date does not parse
	$m--;

	my $t = timelocal(0,0,0,$d,$m,$y);

	my ($wd, $mo, $md, $yr) = (split(/\s+/,scalar(localtime($t))))[0,1,2,4];
	
	return "$Days{$wd} $Months{$mo} $md, $yr";
}

#------------------------------------------------------------------------------

sub mdate {
	# Return date formatted as a "medium" date, like "Jul 10, 1998".

	# The specified date is a string formatted like mm/dd/yy or mm/dd/yyyy
	# or mmddyy.

	my $date = shift || today();	# specified as "m/d/y"

	my ($m,$d,$y) = strtodate($date);
	return $date unless $m && $d && $y;	# date does not parse
	$m--;

	my $t = timelocal(0,0,0,$d,$m,$y);

	my ($mo, $md, $yr) = (split(/\s+/,scalar(localtime($t))))[1,2,4];
	
	return "$mo $md, $yr";
}

#------------------------------------------------------------------------------

sub meddate {
	# Return date formatted as a "medium" date, like "Wed Jul 10, 1998".

	# The specified date is a string formatted like mm/dd/yy or mm/dd/yyyy
	# or mmddyy or mm-dd-yyyy.

	my $date = shift || today();	# specified as "m/d/y"

	my ($m,$d,$y) = strtodate($date);
	return $date unless $m && $d && $y;	# date does not parse
	$m--;

	my $t = timelocal(0,0,0,$d,$m,$y);

	my ($wd, $mo, $md, $yr) = (split(/\s+/,scalar(localtime($t))))[0,1,2,4];
	
	return "$wd $mo $md, $yr";
}

#------------------------------------------------------------------------------

sub short_date_range {
        my ($from_date, $to_date) = @_;

        # Return a date range using the minimal number of digits.
        # The year is not printed on either date.
	# Example: 7/10-28 for July 10 through July 28.

	return '' unless $from_date;

        my ($m1,$d1,$y1);
        my ($m2,$d2,$y2);

        # Extract only month/day from dates
	($m1,$d1,$y1) = strtodate($from_date);
        if ($to_date && $from_date ne $to_date) {
		($m2,$d2,$y2) = strtodate($to_date);
                if ($m1 == $m2) {
                        return sprintf("%d/%d-%d",$m1,$d1,$d2);
                }
                else {
                        return sprintf("%d/%d-%d/%d",$m1,$d1,$m2,$d2);
                }
        }
        else {
                return sprintf("%d/%d",$m1,$d1);
        }
}

#------------------------------------------------------------------------------

sub long_date_range {
        my ($from_date, $to_date) = @_;

        # Return a date range using the minimal number of words.
	# Example: July 10 - 28, 2002

        return '' unless $from_date;

        my ($m1,$d1,$y1);
        my ($m2,$d2,$y2);

        # Extract only month/day from dates
	($m1,$d1,$y1) = strtodate($from_date);
        $m1 = $Months[$m1];
        if ($to_date && $from_date ne $to_date) {
                ($m2,$d2,$y2) = strtodate($to_date);
                $m2 = $Months[$m2];
                if ($m1 eq $m2) {
                        return sprintf("%s %d-%d, %d",$m1,$d1,$d2,$y1);
                }
                else {
                        if ($y1 == $y2) {
                                return sprintf("%s %d - %s %d, %d",$m1,$d1,$m2,$
d2,$y2);
                        }
                        else {
                                return sprintf("%s %d, %d - %s %d, %d",$m1,$d1,$
y1,$m2,$d2,$y2);
                        }
                }
        }
        else {
                return sprintf("%s %d, %d",$m1,$d1,$y1);
        }
}

#------------------------------------------------------------------------------

sub dates_between {
	my ($begin_date, $end_date) = @_;

	# Produce a list of dates between $begin_date and $end_date.
	# Dates are in the format of 'mm-dd-yyyy' or whatever the 
	# date seperator is defined as in Acme::Config (...etc/site.conf).
	# The $end_date better be in that format and be later than
	# $begin_date (or else ...).

	my $ds = Acme::Config->Date_Separator || '/';

	my @dates;			# returned dates
	my ($mm,$dd,$yyyy);		# parts of dates
	my $runaway_limit = 10000;	# hack, should check for valid $end_date

	$end_date ||= $begin_date;

	# Parse and validate end_date.
	($mm,$dd,$yyyy) = strtodate($end_date);
	return unless $mm && $dd && $yyyy;

	my $end_time = &timelocal(0, 0, 0, $dd, ($mm-1), $yyyy);

	# Reformat end_date for comparisons.
	$end_date = datetostr($mm,$dd,$yyyy);

	# Parse and validate begin_date.
	($mm,$dd,$yyyy) = strtodate($begin_date);
	return unless $mm && $dd && $yyyy;

	my $begin_time = &timelocal(0, 0, 0, $dd, ($mm-1), $yyyy);

	return if $end_time < $begin_time;	# no dates between

	my $date;
	do { {
		next unless ($mm >=1 && $mm <= 12 && $dd >= 1 && $dd <= 31);

		$date = datetostr($mm,$dd,$yyyy);
		push @dates, $date;

		($mm,$dd,$yyyy) = dayadd(1,$date); $mm++;

		die "Runaway loop: begin_date=$begin_date, end_date=$end_date"
		  . " current value=$mm$DS$dd$DS$yyyy"
		  if (! $runaway_limit--);

	} } until ($date eq $end_date);
	
	return @dates;
}

1;
