package Acme::Db;

=head1 NAME

Acme::Db - Perl module interface to database engine(s)

=head1 SYNOPSIS

    use Acme::Db;

    #
    # Query database
    #

    $q = Acme::Db->query $script;
    @rec = $q->next;
	or
    $q = Acme::Db->query $script;
    $status = $q->next(\%col);
    $val = $col{'column_name'};
	or
    $q = Acme::Db->query $script;
    $q->bind_columns_to_variables;
    $status = $q->next;
    $val = $column_name;
	or
    $dbh = Acme::Db->connect($connect_string, $user, $passwd);
    $dbh->query $script;
    @rec = $dbh->next;

    $n = $q->count;

    @first_rec = Acme::Db->get $query;
        or
    Acme::Db->get $query, \%result or die;

    #
    # Execute non-query SQL
    #

    $status = Acme::Db->execute $script;
	or
    $dbh = Acme::Db->connect($connect_string, $user, $passwd);
    $status = $dbh->execute $script;

    #
    # Utility methods
    #

    $s = Acme::Db->escape_sql($s);

    $name = $q->column_name($n);
    $n    = $q->column_n($name);

    @a    = $q->rec;

    #
    # Procedure to add new record with database assigned sequential number
    #

    use Acme::Db;

    $dbh = Acme::Db->connect($connect_string, $user, $passwd);

    $r = $dbh->new_record('mytable','id');
    $id = $r->keyval;

    $status = $dbh->execute << End_of_SQL
	update mytable
	set col1='new setting'
	where id = $id
    End_of_SQL

    $r->commit;


    #
    # Procedure to list items with headers and footers
    #

    use Acme::Db;

    my $q = Acme::Db->query(<<End_Of_SQL);
 select customer_name,
        invoice_no,
        invoice_item,
        invoice_item_amount
 from customers, invoices, invoice_items
 where customers.customer_id = invoices.customer_id
   and invoices.invoice_no = invoice_items.invoice_no
 order by customer_name, invoice_no, invoice_item
 End_Of_SQL

    my $before_customer_sub = sub {
    	# routine executed before group of customer_name
	$count = 0;
	$total_amount = 0;
	print "customer $customer_name:\n";
    };

    my $after_customer_sub = sub {
    	# routine executed after group of customer_name
	print "customer $customer_name total is: $total_amount"
    };

    $use_short_names = 1;
    $q->bind_columns_to_variables('package_name',$use_short_names);

    $q->set_nulls_to_empty;

    $q->before_group('customer_name'=> $before_customer_sub,
                     'invoice_no'   => '$item_count = 0');

    $q->after_group('customer_name' => $after_customer_sub,
                    'invoice_no'    => '$item_count = 0');
                    'invoice_item'  => '$total_amount += $invoice_item_amount');

    $q->before_query(sub{print "Beginning of query\n"});

    $q->after_query('Total for all customers is: $total_amount');

    while ($q->next) {
    	print "$invoice_no  $invoice_item   $invoice_item_amount\n";
    }


=head1 DESCRIPTION

The I<Acme::Db> module provides an interface to a SQL database engine.  All the
AcmeBase programs use these routines to access the database.  Originally,
these routines provided database independence, but now the DBI Perl module
by Tim Bunce and company is used. These routines are a superset of the
DBI library. They utilize a look-ahead buffer to implement "begin group"
and "after group" events used in lists and reports. The also automatically
connect to a (site configurable) default database.

The connect() method sets up the object variable and connects to the
database. This is optional as the query() and get() methods automatically
connect to a default database. The connect string passed to connect() can
be either the name of the database as defined in I<Acme::Config> or it 
can be an actual connect string to be passed literally to the database engine.
An exception is raised and the program terminates if connect() fails.

The query() method submits a SQL query to the database engine and sets up an
object variable which can be used to retrieve the resulting rows using the
next() method. Multiple SQL statements separated by semi-colons can be used,
as long as the last statement is a SELECT statement. Unlike the execute()
method, the multiple SQL statements are NOT grouped together as one database
transaction.

Next() returns a list of the next record's field contents. It can also stuff
an array or hash if a reference to it is passed as an argument. If called in
a SCALAR context, it returns the number of elements in the list.

The get() method executes the query and then retrieves the first 
record.

The count() method returns a number representing the number of records
in the query. The number is 0, 1, or one greater than the current record
number, unless the last record has been read, in which case the count()
returns the actual count. The I<Acme::Db> module uses a read-ahead buffer to
implement count()'s functionality. The count() method can be called
immediately after the query() method executes.

The count() is never more than the current record plus 1. Even though
it is not a valid count, it does allow you to determine if the query should
be displayed as a list (count greater than 1), or whether the query should
be displayed as a form (count equals 1), or whether the display should
even be attempted (count equals 0). It also allows you to determine if
you are processing the last record in the selection by:

    if ($current_record_no == $q->count()) { #process last record };

The execute() method executes a SQL script. The return value is 1 or 0
depending upon success or failure. Multiple SQL statements are 
executed as one transaction.

The escape_sql() method is a helper method to encode strings so that
they pass SQL syntax. It can be used to construct UPDATE or INSERT
statements for execute(). It can be invoked as a function or as a
method.

The new_record() method inserts a record into the table and returns
an object that holds the key value of the inserted record. The object
needs to be `committed' with the commit() method, otherwise the record
is deleted from the database. It is similar to a transaction: you get a
record, fill it full of data (update it), and then commit it. 

If a key value is not supplied to new_record(), it uses a database generated
sequential number for the key value. The new_record() method was written
to provide an interface to the different ways the different database engines
maintain serial data fields.

The bind_columns_to_variables() binds the SQL field names to perl variables.
References to variables with the same name as the SQL columns retrieves
the current value of the column read from the database. The column names
are bound to variables in the named package or the calling package if a
package name is not specified. The variables are fully qualified SQL field
names unless the routine is passed a non-zero flag to use short names.
Column names are converted into valid Perl variables by substituting double
underlines (__) for any invalid character such as the period. For example,
C<table.field> becomes C<table__field>.

The rec() method returns the array of column values read from the
database. This is the same array that the variables are bound to by
bind_columns_to_variables() above. This allows a record to be read from the
database by next(), variable changed by using bound variable names, and then
printed by using an array reference.

The before_query() and after_query() methods registers instructions or a
subroutine reference to invoke just before the first record or just after
the last record is read.

The before_group() and after_group() methods registers instructions or a
subroutine reference to invoke whenever values in the named columns change
from one record to the next. It is assumed that the query has been sorted
on the same columns that have routines registered. If the query does not
select any records, the methods are not invoked. Columns can be identified
by name or number.

B<Note>: The bind_columns_to_variables() method must be run before the
before_group() or after_group() methods. Otherwise, it appears as if the
before_group() and after_group() routines are only executed before the query
and after the query.

B<Note>: The before_group/after_group and before_query/after_query statements
or functions are eval'ed and any exceptions that are thrown are ignored. If you
want to terminate the list, you must call die() in the before/after statement,
and then in the main loop processing the query, you need to have the statement:

	die $@ if $@;

The set_nulls_to_empty() method sets a flag so that every undefined (null) value
read from the database is set to the empty string. It is useful for avoiding
warnings when using Perl's -w flag. As of Perl 5.6, you can instead use the
Perl pragma:

	no warnings qw(uninitialized);

Setting the null values to the empty string thwarts the before_group() and
after_group() comparisons at this time (8/2002).

The column_name() and column_n() methods returns the corresponding column
name given a number, or the column number given the column name respectfully.
The column_n() method is quicker than column_name() -- it is easier to find
the number given the column name than it is to find the name given a number.

=head2 LOGS

A log file is kept listing every sql statement executed by whom and when. It
is used for debug and security. It can be turned off by setting the global
$Log_File to an empty string.

An error log file is kept listing every sql statement error. The filename is
stored in the global $Error_Log_File. It can be the same as the $Log_File,
but it should exist and not be set to the empty string.

=head2 NOTES

These routines are intended to provide a database independent interface
upon which to implement a major application. The strategy is to use
a minimal subset of "capable" SQL databases. Some required capabilities are;

    * sub-selects
    * outer joins
    * date / time
    * select into temp table
    * copy to text file
    * copy from text file
    * regular expression matching
    * transactions (at least non-complaining)

Another capability needed, and one these routines provide if not supported
by the database system, is the ability to process multiple statements as
one (and ignore empty statements). Even queries can have multiple SQL
statements leading up to the final query.

The following database inconsistencies are not accommodated yet and requires
SQL code through out the system to be modified by hand:

    * regular expression matching. Postgress uses operators ~ and ~*.
    * select into temp table

Although there are interface methods load() and unload(), they are autonomous
and their commands cannot be embedded in a string of multiple SQL commands
where they are needed. Therefore, the load() and unload() SQL equivalents
must be hand edited by hand too. (see Acme::Searcher).

=head1 CHANGE LOG

 12/20/97 Implemented on Informix SE.
 11/16/98 Added look-ahead buffer and count() method.
 07/02/99 Added escape_sql() method.
 01/14/00 Converted to Postgres.
 01/24/00 Converted to DBI/MySQL. Eliminated first() and report() methods.
 02/20/00 Converted to Postgress again. Added log_transaction() and error().
 02/29/00 Added new_record() and Acme::Db::Record class.
 04/08/00 Added before_group(), after_group(), load() and unload() methods.
 08/12/00 Added before_query() and after_query().
 04/03/01 Added $shortnames flag to bind_columns_to_variables() method.
 06/08/01 Removed bug in new() and connect() so that Db object can be inherited.
          Also changed bind_columns_to_variables() to accept a package name.
 08/20/01 Added set_nulls_to_empty() to set a flag so undefined fields are
          set to the empty string (useful for circumventing Perl's -w warnings).
 09/30/01 Enhanced get() to stuff hash or array in arguments like next().
 07/26/02 Added column_name() and column_n() methods to get column's name
          their number, which is the order they were defined in the sql.
 08/04/02 Added nrec() method which similar to count(), returns current 
          record number.
 09/09/02 Added rec() method to return the current database record as an array.

=head1 TODO

    * Generalize Acme::Db::Site load() method to load specific columns.
    * Allow multiple Acme::Db::Site files for different database systems.
    * Make logging routines into an object, ala Apache::Log.
    * Cleanup process of allowing execute_sql() to be called while a
      query() is begin processed.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1997-2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Carp;

use Acme::SQL;
use Acme::Config;				# local settings
use Local::Acme::User	qw(user user_passwd);	# for connect and log messages
require Acme::Db::Site;				# site specific routines

use Sys::Hostname;

$Log_File	= Acme::Config->Log_File;	# optional security log
$Error_Log_File	= Acme::Config->Error_Log_File || '/var/tmp/Dberror.log';

$Webmaster = Acme::Config->Webmaster;

$Separator = Acme::Config->Unload_Separator || "\t";# record field delimiter
$Null_Literal = Acme::Config->Null_Literal || 'null';

#------------------------------------------------------------------------------

sub new {
	my $proto = shift;	# caller object (should be class/package name)
	my $class = ref($proto) || $proto;

	my $self = {
		sqlref => undef,	# reference to SQL query script
		is_connected => 0,	# true if connected to database
		is_setup => 0,		# true if query set up to read results
		is_first_record => 0,	# true if first record read from query
		is_cursor_open => 0,	# true if query cursor is open
		no_undefs => 0,		# if true, set undefs to empty string
		dbh => undef,		# data base handle (connect)
		sth => undef,		# data base statement handle (result)
		bufp => 0,		# look-ahead buffer pointer
		buf => undef,		# look-ahead buffer (see Acme::Db::Site)
		recp => 0,		# record pointer (current record number)
		rec => [],		# current record columns
		fields => {},		# record fields / column numbers hash
		before_group_triggers => [], # order of before group execution
		after_group_triggers  => [], # order of after group execution
		before_group_start => undef, # index of beginning triggers to do
		after_group_start  => undef, # index of after triggers to do
		before_group_stmts => {},    # group column/subroutine pairs
		after_group_stmts  => {},    # group column/subroutine pairs
		before_query => undef,	# subroutine or eval string
		after_query  => undef,	# subroutine or eval string
		start_time   => time,   # start time of sql
		@_,			# calling arguments
		};

	return bless $self, $class;
}

#------------------------------------------------------------------------------

sub connect {
	my ($self, $connect_string, $user, $passwd, @args) = @_;

	$self = new($self) unless ref($self);

	#
	# Check if $connect_string is actually a database name in the 
	# configuration, and if so, use those settings as a connect string.
	#

	my $dbname = $connect_string || Acme::Config->DBNAME;

	if (Acme::Config->Databases->{$dbname}) {

		#
		# Connect using values in configuration file.
		# Construct a DSN type of connect string:
		#     "dbname=acme;host=192.168.1.1;port=54321"
		#

		my $host          = Acme::Config->DBHOST($dbname);
		my $port          = Acme::Config->DBPORT($dbname);
		$connect_string   = Acme::Config->DBCONNECT($dbname);
		$connect_string  .= $host ? ";host=$host" : '';
		$connect_string  .= $port ? ";port=$port" : '';
		$user           ||= Acme::Config->DBUSER($dbname);
		$passwd         ||= Acme::Config->DBPASS($dbname);
	}
	else {
		#
		# Connect using program arguments.
		#

		$connect_string = Acme::Config->DBCONNECT;
		$user   = Acme::Config->DBUSER   || user();
		$passwd = Acme::Config->DBPASS   || user_passwd();
	}

	Acme::Db::Site::connect($self,$connect_string, $user, $passwd, @args)
	  or croak "$self->{error}\nCannot connect to database.";

	return $self;
}

#------------------------------------------------------------------------------

sub query {
	my $self = shift;
	my $class = ref($self);
	unless ($class) {
		# $self is not a reference to an object.
		$class = $self;
		$self =  new $class;
	}
	$self->connect() unless $self->{is_connected};

	my $sql = shift;
	$self->{sqlref} = \$sql;

	Acme::Db::Site::prepare_query($self)
	  or croak "$self->{error}\nErrors found in query script";

	return $self;
}

#------------------------------------------------------------------------------

sub next {
	my $self = shift;
	my $ptr = shift;
	my ($pkg) = caller;

	if (! $self->{is_setup}) { $self->_setup(@_) or return; }

	if ($self->{is_first_record}) {
		$self->{is_first_record} = 0;
		$self->_stuff_rec($ptr) if $ptr;
		$self->_do_before_query($pkg);
		$self->_do_before_group_triggers($pkg,'all');
	}
	else {
		#
		# Return nothing if last record has been read.
		#

		if ($self->{recp} > $self->{bufp}) {
			return;
		}

		#
		# Last record processing.
		#

		if ($self->{recp} == $self->{bufp}) {
			$self->_do_after_group_triggers($pkg,'all');
			$self->_do_after_query($pkg);
			return;
		}

		#
		# Determine before and after group routines to trigger
		#

		$self->_get_after_group_triggers();
		$self->_get_before_group_triggers();

		$self->_do_after_group_triggers($pkg);

		#
		# Get next record.
		#

		Acme::Db::Site::parse_rec($self);
		Acme::Db::Site::fetch_rec($self);	# read ahead

		$self->_stuff_rec($ptr) if $ptr;

		$self->_do_before_group_triggers($pkg);
	}

	#
	# Return array of parsed fields.
	#

	return @{$self->{rec}};
}

#------------------------------------------------------------------------------

sub get {
	my ($self, $sql, $ptr,@param) = @_;
	my $class = ref($self);
	unless ($class) {
		# $self is not a reference to an object.
		$class = $self;
		$self  = new $class;
	}
	$self->connect() unless $self->{is_connected};

	$self->{sqlref} = \$sql;

	Acme::Db::Site::prepare_query($self)
	  or croak "$self->{error}\nErrors found in query script";

	Acme::Db::Site::execute_query($self,@param)
	  or croak "$self->{error}\nErrors found in query script";

	Acme::Db::Site::open_cursor($self)
	  or croak "$self->{error}\nCannot open query results";

	Acme::Db::Site::fetch_rec($self) and Acme::Db::Site::parse_rec($self);

	Acme::Db::Site::close_cursor($self);

	$self->_stuff_rec($ptr) if $ptr;

	return @{$self->{rec}};
}

#------------------------------------------------------------------------------

sub count {
	my $self = shift;

	if (! $self->{is_setup}) { $self->_setup(@_) or return 0; }

	# The count returned is simply the file pointer, i.e. the number
	# of records read. Because this system uses a read-ahead buffer,
	# the file pointer is always 1 greater than the current record
	# number (until the last record, then it is equal).

	return $self->{bufp};
}

#------------------------------------------------------------------------------

sub nrec {
	my $self = shift;

	if (! $self->{is_setup}) { $self->_setup(@_) or return 0; }

	# The record number returned is simply the file pointer.

	return $self->{recp};
}

#------------------------------------------------------------------------------

sub rec {
	my $self = shift;

	if (! $self->{is_setup}) { $self->_setup(@_) or return; }

	# The record number returned is simply the file pointer.

	return @{$self->{rec}};
}

#------------------------------------------------------------------------------

sub execute {
	my $self = shift;
	my $class = ref($self);
	unless ($class) {
		# $self is not a reference to an object.
		$class = $self;
		$self =  new $class;
	}
	$self->connect() unless $self->{is_connected};

	# Old way, join arguments into one sql
	#my $sql = join "\n", @_;	

	# DBI way, SQL is first argument and others are sql script arguments
	my $sql = shift;

	# Save old sqlref (a pending query?) before assigning new ref.
	my $saveref = $self->{sqlref};
	$self->{sqlref} = \$sql;

	if (Acme::Db::Site::execute_sql($self,@_)) {
		$self->{sqlref} = $saveref;
		return 1;
	}
	else {
		croak "$self->{error}\nErrors found in SQL script";
	}
}

#------------------------------------------------------------------------------

sub load {
	my ($self,$file,$insert_clause,$delimiter) = @_;
	my $class = ref($self);
	unless ($class) {
		# $self is not a reference to an object.
		$class = $self;
		$self =  new $class;
	}
	$self->connect() unless $self->{is_connected};

	$delimiter ||= $Separator;
	Acme::Db::Site::load($self,$file,$insert_clause,$delimiter)
	  or croak "$self->{error}\nErrors found in SQL script";
	
	return 1;
}

#------------------------------------------------------------------------------

sub unload {
	my ($self,$file,$select_clause,$delimiter) = @_;
	my $class = ref($self);
	unless ($class) {
		# $self is not a reference to an object.
		$class = $self;
		$self =  new $class;
	}
	$self->connect() unless $self->{is_connected};

	$delimiter = $Separator unless $delimiter;

	Acme::Db::Site::unload($self,$file,$select_clause,$delimiter)
	  or croak "$self->{error}\nErrors found in SQL script";

	return 1;
}

#------------------------------------------------------------------------------

sub new_record {
	my ($self,$table,$key,$keyval) = @_;
	my $class = ref($self);
	unless ($class) {
		# $self is not a reference to an object.
		$class = $self;
		$self =  new $class;
	}
	$self->connect() unless $self->{is_connected};

	require Acme::Db::Record;
	return new Acme::Db::Record $self, $table, $key, $keyval;
}

#------------------------------------------------------------------------------

sub escape_sql {
	# This routine can be used both as a function or a method.
	shift if ref($_[0]) || $_[0] eq __PACKAGE__;

	return Acme::Db::Site::escape(@_);
}

#------------------------------------------------------------------------------

sub set_nulls_to_empty {
	my $self = shift;
	$self->{no_undefs} = 1;		# set flag
}

#------------------------------------------------------------------------------

sub bind_columns_to_variables {
	my ($self,$pkg,$shortnames) = @_;
	($pkg) = caller unless $pkg;

	#
	# Build hash of field name / column number
	#

	if (! %{$self->{fields}}) {
		my $n = 0;
		my $sqlo = new Acme::SQL ${$self->{sqlref}};
		%{$self->{fields}} = map { $_, $n++ } $sqlo->columns;
	}

	#
	# Tie each field name variable to it's appropriate
	# element in the record columns array rec.
	#

	my ($k, $v);
	for $k (keys %{$self->{fields}}) {
		$v = $self->{fields}->{$k};	# get array index of field $k

		# Make sure column name is a valid Perl variable name.
		# Perl variables begin with an alpha [A-Za-z] or underline [_]
		# and the rest of the characters must be an alpha [A-Za-z],
		# digit [0-9], or underline [_]. Prepend an underline
		# to the column name if first character is not an alpha, and
		# substitute double underlines for every non-valid variable
		# name character in the column name.

		if ($shortnames) {
			# Strip table name off of field name
			$k =~ s/.*\W//;
		}
		else {
			# Change non-characters to double underlines.
			$k =~ s/\W/__/g;
		}
		$k = "_$k" unless $k =~ /^[A-Za-z]/;

		tie ${"${pkg}::$k"}, Acme::Db::Var, \$self->{rec}[$v];
	}
}

#------------------------------------------------------------------------------

sub column_name {
	my ($self,$n) = @_;
	my $flds = $self->{fields};
	for (keys %{$flds}) {
		return $_ if $flds->{$_} == $n;
	}
	return '';
}

#------------------------------------------------------------------------------

sub column_n {
	my ($self,$field) = @_;
	return $self->{fields}->{$field};
}

#------------------------------------------------------------------------------

sub before_query {
	my ($self, $p) = @_;
	$self->{before_query} = $p;
}

#------------------------------------------------------------------------------

sub after_query {
	my ($self, $p) = @_;
	$self->{after_query} = $p;
}

#------------------------------------------------------------------------------

sub before_group {
	my $self = shift;

	while (my $k = shift) {
		push @{$self->{before_group_triggers}}, $k;
		$self->{before_group_stmts}->{$k} = shift;
	}
}

#------------------------------------------------------------------------------

sub after_group {
	my $self = shift;

	while (my $k = shift) {
		push @{$self->{after_group_triggers}}, $k;
		$self->{after_group_stmts}->{$k} = shift;
	}
}

#------------------------------------------------------------------------------

sub log_transaction {
	my ($self, $log_message, $sql) = @_;

	#
	# Note: the option $sql argument is logged as the SQL statement
	# executed if given, otherwise, the object's sql $sql->{sqlref}
	# statement is logged.
	#

	my $elapsed_time = time - $self->{start_time};

	local $\;	# no automatic newlines after print

	# Record query in log file (for security)
	if ($Log_File) {
		my $user = user();
		open LOG, ">>$Log_File" or die "$!. Cannot open log file '$Log_File'";
		print LOG "[>>>>>>> ", $user, ": ", scalar localtime, "  $0\n";
		print LOG $sql || ${$self->{sqlref}};	# log sql statement
		print LOG "\n", $log_message;
		print LOG "\n<<<<<<<] ", $user, ": ", $elapsed_time, " sec. \n";
		close LOG;
	}

	return;
}

#------------------------------------------------------------------------------

sub error {
	# This routine can be used both as a class or object method
	my ($self, $error_message, $sql) = @_;

	# Record query and error in log file (for debug)

	#
	# Note: the optional $sql argument is logged as the SQL statement
	# executed if given, otherwise, the object's sql $sql->{sqlref}
	# statement is logged.
	#

	my $elapsed_time = '';
	if (ref $self) {
		$elapsed_time = time - $self->{start_time};
		$elapsed_time = "$elapsed_time sec. ";

		$sql ||= ${$self->{sqlref}};
	}

	my $user = user();
	my $host = eval { hostname() } || $ENV{SERVER_NAME} || '';
	($host) = $host =~ /([\w.-]+)/;	# untaint

	chomp $error_message;		# remove trailing newlines

	# Save last error message
	$self->{error} = $error_message if ref $self;

	#
	# Compose error log message
	#

	my $errlog = "[>>>>>>>> Error: $user: ".(scalar localtime)."  $0\n";
	$errlog .= $sql if $sql;	# record sql statement
	$errlog .= "\n$error_message";	# record error message
	$errlog .= "\n<<<<<<<<] Error: $user: $elapsed_time\n";

	#
	# Save error message in log file
	#

	if (open ERRORLOG, ">>$Error_Log_File") {
		print ERRORLOG $errlog;
		close ERRORLOG;
	}

	#
	# Send error message to Webmaster
	#

	$ENV{PATH} = '/bin:/usr/bin';			# untaint
	delete @ENV{ qw(IFS CDPATH ENV BASH_ENV) };	# untaint

	if ($Webmaster && open(SENDMAIL,"| /usr/sbin/sendmail -oi -t")) {
		my $to = "$Webmaster";
		(my $subject = $0) =~ s|.*cgi-bin/||;
		$subject = "Database error: $user\@$host using $subject";

		print SENDMAIL <<End_of_Msg;
From: $Webmaster
To: $to
Subject: $subject
Mime-Version: 1.0
Content-type: text/plain

End_of_Msg
		print SENDMAIL $errlog;
		close(SENDMAIL);
	}

	return;
}

#------------------------------------------------------------------------------
#
# Private methods
#
#------------------------------------------------------------------------------

sub _setup {
	my $self = shift;

	unless ($self->{is_setup}) {
		Acme::Db::Site::execute_query($self,@_)
		  or croak "$self->{error}\nErrors found in query script";
		Acme::Db::Site::open_cursor($self)
		  or croak "$self->{error}\nCannot open query results";
		$self->_setup_buffers or return;	# empty query
	}
	return 1;
}

sub _setup_buffers {
        my ($self) = @_;

        # Prime read-ahead buffers.
        if (Acme::Db::Site::fetch_rec($self)) {
                Acme::Db::Site::parse_rec($self);
                Acme::Db::Site::fetch_rec($self);
		$self->{is_first_record} = 1;
		return $self->{is_setup} = 1;	# success
        }

	return;					# empty query
}

sub _get_before_group_triggers {
	my $self = shift;

	return unless @{$self->{before_group_triggers}};

	my $trigger = 0;
	my ($field, $idx);
	for $field (@{$self->{before_group_triggers}}) {
		if ($field =~ /^\d+$/) {
			$idx = $field;	# group is identified by number
		}
		else {
			$idx = $self->{fields}->{$field};
		}

		#
		# Note: undefined (null) values are considered equal
		# here so records are grouped together.
		#

		if (defined $self->{buf}->[$idx]) {
			if (not defined $self->{rec}->[$idx]
			or $self->{buf}->[$idx] ne $self->{rec}->[$idx]) {
				return $self->{before_group_start} = $trigger
			}
		}
		elsif (defined $self->{rec}->[$idx]) {
			return $self->{before_group_start} = $trigger
		}
		$trigger++
	}
	return $self->{before_group_start} = undef;
}

sub _get_after_group_triggers {
	my ($self) = @_;

	return unless @{$self->{after_group_triggers}};

	my ($field, $idx);
	my $trigger = $#{$self->{after_group_triggers}};
	for $field (reverse @{$self->{after_group_triggers}}) {
		if ($field =~ /^\d+$/) {
			$idx = $field;	# group is identified by number
		}
		else {
			$idx = $self->{fields}->{$field};
		}

		#
		# Note: undefined (null) values are considered equal
		# here so records are grouped together.
		#

		if (defined $self->{buf}->[$idx]) {
			if (not defined $self->{rec}->[$idx]
			or $self->{buf}->[$idx] ne $self->{rec}->[$idx]) {
				return $self->{after_group_start} = $trigger
			}
		}
		elsif (defined $self->{rec}->[$idx]) {
			return $self->{after_group_start} = $trigger
		}
		$trigger--
	}
	return $self->{after_group_start} = undef;
}

sub _do_before_group_triggers {
	my ($self,$pkg,$all) = @_;

	return unless defined $self->{before_group_start} || $all;

	my ($field, $trigger, $s);
	my $start = $all ? 0 : $self->{before_group_start};
	for $trigger ($start .. $#{$self->{before_group_triggers}}) {
		$field = $self->{before_group_triggers}->[$trigger];
		$s = $self->{before_group_stmts}->{$field} || '';
		$self->_do($pkg,$s);
	}
}

sub _do_after_group_triggers {
	my ($self,$pkg,$all) = @_;

	return unless defined $self->{after_group_start} || $all;

	my ($field, $trigger, $s);
	my $start = $all ? $#{$self->{after_group_triggers}} : $self->{after_group_start};
	for $trigger (0 .. $start) {
		$field = $self->{after_group_triggers}->[$trigger];
		$s = $self->{after_group_stmts}->{$field} || '';
		$self->_do($pkg,$s);
	}
}

sub _do_before_query {
	my ($self,$pkg) = @_;

	return unless defined $self->{before_query};

	my $s = $self->{before_query};
	$self->_do($pkg,$s);
}

sub _do_after_query {
	my ($self,$pkg) = @_;

	return unless defined $self->{after_query};

	my $s = $self->{after_query};
	$self->_do($pkg,$s);
}

sub _do {
	my $self = shift;
	my $pkg  = shift;
	my $s    = shift;

	# Execute string in pkg namespace.
	# If string is a function, it is called with the remaining arguments.

	# Note: this should probably be promoted to a public method as it is
	# used in Acme::Lister.

	return					if (! $s);
	return eval "package $pkg; $s"		if (! ref $s);
	return eval "package $pkg; $$s"		if (ref $s) =~ /SCALAR/;
	return eval "package $pkg; &\$s(\@_)"	if (ref $s) =~ /CODE/;
}

sub _stuff_rec {
	my ($self, $ptr) = @_;

	# Store database record into array/hash/scalar pointed to by $ptr.
	if ($ptr) {
		if (ref $ptr eq "ARRAY") {
			@{$ptr} = @{$self->{rec}};
		}
		elsif (ref $ptr eq "HASH") {
			unless (%{$self->{fields}}) {
				my $n = 0;
				my $sqlo = new Acme::SQL ${$self->{sqlref}};
				for ($sqlo->columns) {
					${$self->{fields}}{$_} = $n++;
				}
			}

			for (keys %{$self->{fields}}) {
				$$ptr{$_} = ${$self->{rec}}[ ${$self->{fields}}{$_} ];
			}
		}
		elsif (ref $ptr eq "SCALAR") {
			$$ptr = join $Separator, @{$self->{rec}};
		}
	}
}

sub DESTROY {
	my ($self) = @_;

	# Clean up DB connection
	Acme::Db::Site::close_cursor($self) if $self->{is_cursor_open};
	Acme::Db::Site::disconnect($self);
}

#-----------------------------------------------------------------------------
#
# Database bound variable routines
#
#------------------------------------------------------------------------------

package Acme::Db::Var;

sub TIESCALAR {
	my ($class, $ref) = @_;

	# Associate sql field name with perl variable.
	bless $ref, $class;
}

sub FETCH {
	my ($self) = @_;

	# Reference points to the appropriate array element in $query->{rec}.
	return $$self;
}

sub STORE {
	my ($self,$val) = @_;

	# Store value into memory copy of database field.
	# Note that this does not update the database records as
	# updating a database through a cursor is not supported.

	return $$self = $val;
}

sub DESTROY {}		# gobble

1;
