package Acme::Db::Record;

=head1 NAME

Acme::Db::Record - Perl module to control adding new database records.

=head1 SYNOPSIS

	require Acme::Db::Record;

	$r = new Acme::Db::Record $dbh, 'table' 'key';
	  or
	$r = new Acme::Db::Record $dbh, 'table' 'key', $keyval;

	print "New key is ", $r->keyval;

	$r->commit;

=head1 DESCRIPTION

The I<Acme::Db::Record> module provides an interface/methodology for
adding new records to the database tables. It is automatically included
in the I<Acme::Db> package as needed.

There are problems adding records to databases in a standard SQL syntax.
Primarily, a lot of user information needs to be referenced by a 
database generated number, and retreiving that number from the
most recently added record is done differently with every database
software.

Another problem with using standard SQL INSERT statements in a system
where much of the SQL Data Manipulation Language is dynamically 
generated or assembled is that the syntax (and restrictions) are
different than the UPDATE statements. This module provides a methodology
where the INSERT can be very simple, and the user information is
entered into the database using the same logic to change the database
information using UPDATE.

Adding new records to the database is performed as a transaction.
First a new record is created (INSERT). At this point, the database
has a new record that is empty except for database generated values
such as defaults values, nulls, and serialized numbers. You then
access the new record for updates and queries using the key value
stored in the object variable.

When done you must commit the transaction, otherwise the new record
will be deleted automatically.

A key value can be specified for tables with non-sequential keys. If
a key value is specified for a table with a sequential key, this package
does not generate a sequence number. On some systems, the next
sequential number is automatically adjusted to be the specified number
plus 1 (if specified number is greater than the next sequential number),
but not for Postgres.

=head1 CHANGE LOG

 2/28/2000  Initial version.
 4/13/02    Updated to 'use strict'.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1997-2000, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

#------------------------------------------------------------------------------

sub new {
	my $this  = shift;
	my $class = ref($this) || $this;

	my (
		$dbh,		# Acme::Db object
		$table,		# database table name
		$key,		# key field
		$keyval,	# optional value for key field
	) = @_;

	my $self = {
		dbh          => $dbh,	# database handle
		table        => $table,	# database table name
		key          => $key,	# key field of table
		keyval       => undef,	# key field value returned from database
		is_pending   => 0,	# true if record inserted
		is_committed => 0,	# true if record inserted and committed
	};

	if ($self->{keyval} = Acme::Db::Site::insert($dbh,$table,$key,$keyval)) {
		$self->{is_pending} = 1;
		return bless $self, $class;
	}
	return;		# error
}

#------------------------------------------------------------------------------

sub commit {
	my $self = shift;

	return $self->{is_committed} = 1;
}

#------------------------------------------------------------------------------

sub DESTROY {
	my $self = shift;

	return if ! $self->{is_pending} || $self->{is_committed};

	# Remove inserted record
	# Note: This does not necessarily reset the key index used.
	$self->{dbh}->execute(<<EOF);
delete from $self->{table} where $self->{key} = '$self->{keyval}';
EOF
	return;
}

#------------------------------------------------------------------------------

sub AUTOLOAD {
        my $self = shift;
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;

	# Return object parameter if same name as method name.
        return $self->{$method} if exists $self->{$method};
        return;			# unknown parameter/method
}

1;
