package Acme::Db::Schema;

=head1 NAME

Acme::Db::Schema - Perl module to control adding new database records.

=head1 SYNOPSIS

	use Acme::Db::Schema qw(ddl);

	@tables = ('customers','customer_info');
	print_ddl(@tables);

	print ddl('customers');		# print one table

=head1 DESCRIPTION

The ddl() function returns SQL Data Definition Language statements
(e.g. create table...) which can be used to create database tables and indexes.
Table, field, and index definitions come from the I<Acme::Base::Schema> module.

The print_ddl() prints the Data Definition Language statements to STDOUT.
It accepts a list of table names, unlike ddl() which produces statements
for only one table.

=head1 CHANGE LOG

 3/24/01  Initial version
 5/20/02  Removed creation of "rules" used to update audit tables.
          The Acme::Form::Saver module now does the audit tracking.
 12/27/02 Added table name validation rules.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1997-2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

require Exporter;

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(ddl print_ddl);

use Acme::Base::Schema;
use Acme::Base::Schema::Master;
use Acme::Config;
use Acme::Security;

our %Max_Table_Length = (
	Postgres => 31,
	MySQL    => 64,
	Informix => 18,	# old Informix 5.0
	);

#-------------------------------------------------------------------------------

sub ddl($) {
	my $tab = shift;
	my $script = '';		# returned script

	my $s = new Acme::Base::Schema $tab;

	my $t = $s->t($tab)
	  or die "\nThere is not a table with a name of $tab in the schema.\n";

	if (my $sc = $t->security_code) {
		die Acme::Security::alert("\nAccess denied to $tab")
		  unless Acme::Security::check($sc,'Access');
	}

	#
	# Check validity of table name.
	#

	# AcmeBase requires that audited table names be 3 characters shorter
	# than the maximum column names. Columns with timing information are
	# automatically added to the audit history table that has the prefix
	# "_y_" added to the table name.

	if (length($tab) + 3 > $Max_Table_Length{$s->dbengine}) {
		die qq[The table name "$tab" is too long.]
		  . qq[ Audited tables in ] . $s->dbengine
		  . qq[ cannot have names longer than ]
		  . $Max_Table_Length{$s->dbengine} - 3
		  . qq[ characters.\n];
	}

	if ((my $ch) = $tab =~ /(\W)/) {
		die qq[Invalid character "$ch" in the table name.]
		  . qq[ Table names can only have alpha-numeric characters.];
	}

	{
		# Create table
		$script .= "\n\ncreate table $tab (";

		my $comma='';
		my $id;
		if ($id = $t->id) {
			# automatically define ID  key field
			$script .= "$comma\n  $id serial";
			$comma=',';
		}
		for ($t->fields_all) {
			next if $_ eq $id;	# already defined above
			$script .= "$comma\n  $_ ".$t->f($_)->type;
			$comma=",";
		}
		if ($t->audit) {
			$script .= "$comma\n  ".$t->audit_entry .' '.$t->audit_entry_type;
			$script .= "$comma\n  ".$t->audit_update.' '.$t->audit_update_type;
			$script .= "$comma\n  ".$t->audit_user  .' '.$t->audit_user_type;
		}
		$script .= "\n);\n";
	}

	{
		# Create indexes
		my $n = 0;		# index number
		my ($i, $keyidx, $idx);

		# Make key index
		if ($keyidx = $t->key) {
			if ($s->dbengine =~ /Postgres/i
			&&  $keyidx eq $t->id) {
				# Postgress creates its own index on
				# serial fields automatically.
				;	# do nothing
			}
			else {
				# Name of key field index ends in '_0'.
				$script .= "\ncreate ";
				$script .= "unique ";
				$script .= "index " . $tab . "_" . $n;
				$script .= " on $tab ($keyidx);";
			}
		}

		# Make indexes for search fields
		if ($idx = $t->code_field and $idx ne $keyidx) {
			for (split /\s*,\s*/, $idx) {
				$script .= "\ncreate ";
				$script .= "index ".$tab."_".++$n;
				$script .= " on $tab ($_);";
			}
		}
		if ($idx = $t->name_field and $idx ne $keyidx) {
			my $firstnamefield = $t->first_name_field;
			my $lastnamefield = $t->last_name_field;
			if ($firstnamefield and
			   (not $lastnamefield or $idx eq $lastnamefield)) {
				$idx .= ",$firstnamefield";
			}

			for (split /\s*,\s*/, $idx) {
				$script .= "\ncreate ";
				$script .= "index ".$tab."_".++$n;
				$script .= " on $tab ($_);";
			}
		}
		if ($idx = $t->last_name_field and $idx ne $t->name_field) {
			my $firstnamefield = $t->first_name_field;
			$idx .= $firstnamefield ? ",$firstnamefield" : '';

			for (split /\s*,\s*/, $idx) {
				$script .= "\ncreate ";
				$script .= "index ".$tab."_".++$n;
				$script .= " on $tab ($_);";
			}
		}
		if ($idx = $t->email_field and $idx ne $keyidx) {
			for (split /\s*,\s*/, $idx) {
				$script .= "\ncreate ";
				$script .= "index ".$tab."_".++$n;
				$script .= " on $tab ($_);";
			}
		}
		if ($idx = $t->phone_field and $idx ne $keyidx) {
			for (split /\s*,\s*/, $idx) {
				$script .= "\ncreate ";
				$script .= "index ".$tab."_".++$n;
				$script .= " on $tab ($_);";
			}
		}

		# Make indexes for Master table joins
		if (my $m = new Acme::Base::Schema::Master $tab, $t->master) {
			my $numfld = $t->number_field;
			for my $fld ($m->fields) {
				$idx = $fld . ($numfld ? ",$numfld" : '');
				$script .= "\ncreate ";
				$script .= "index ".$tab."_".++$n;
				$script .= " on $tab ($idx);";
			}
		}

		# Make audit indexes
		if ($t->audit) {
			for $idx ($t->audit_entry, $t->audit_update) {
				# Create indexes for audit info
				$script .= "\ncreate ";
				$script .= "index ".$tab."_".++$n;
				$script .= " on $tab ($idx);";
			}
		}

		# Make indexes specified in schema
		for $idx ($t->indexes) {
			$i = $t->index->{$idx};

			$script .= "\ncreate ";
			$script .= "unique " if ($i->type || '') eq 'unique';
			$script .= "index ".$tab."_".++$n;
			$script .= " on $tab ($idx);";
		}
	}

	if ($s->t($tab)->audit) {
		# Create audit

		$script .= "\n\ncreate table ".$t->audit_table." (";

		my $comma='';
		my $id;
		if ($id = $t->id) {
			# automatically define ID  key field
			$script .= "$comma\n  $id integer";	# serial equiv.
			$comma=',';
		}
		my @exclude = split /,/, $t->audit_exclude;
		for my $f ($t->fields_all) {
			next if $f eq $id;	# already defined above
			next if grep { $f eq $_ } @exclude;
			$script .= "$comma\n  $f ".$t->f($f)->type;
			$comma=",";
		}
		$script .= "$comma\n  ".$t->audit_entry .' '.$t->audit_entry_type;
		$script .= "$comma\n  ".$t->audit_update.' '.$t->audit_update_type;
		$script .= "$comma\n  ".$t->audit_user  .' '.$t->audit_user_type;
		$script .= "$comma\n  ".Acme::Config->audit_type.' char(1)';
		$script .= "$comma\n  ".Acme::Config->audit_userstamp.' '
			.Acme::Config->audit_user_type;
		$script .= "$comma\n  ".Acme::Config->audit_timestamp.' '
			.Acme::Config->audit_update_type;
		$script .= "\n);\n";

		# Create audit index
		if (my $idx = $t->key) {
			# Name of key field index ends in '_0'.
			$script .= "\ncreate index "
			  . $t->audit_table . '_0'
			  . " on ".$t->audit_table." ($idx);";
		}

	}
	$script .= "\n";

	return $script;
}

#-------------------------------------------------------------------------------

sub print_ddl(@) {
	if (@_) {
		for (@_) {
			print ddl($_);
		}
	}
	else {
		# list all tables if none specified

		my $s = new Acme::Base::Schema;
		for ($s->tables) {
			print ddl($_);
		}
	}
}

1;
