package Acme::Db::Site;

=head1 NAME

Acme::Db::Site - Site specific routines to Acme Base Company database

=head1 SYNOPSIS

    connect() 		# connect to database
    disconnect() 	# disconnect from database
    escape() 		# quote text for SQL statements
    prepare_query() 	# prepare SQL query script
    execute_query() 	# execute prepared query
    open_cursor() 	# open query cursor
    close_cursor() 	# close query cursor
    fetch_rec()		# get next record from query
    parse_rec()		# parse database record and make it the current record
    execute_sql() 	# execute non-query SQL script(s)
    insert() 		# insert a new record into a database table
    load()		# load file into table
    unload()		# unload table into a file

=head1 DESCRIPTION

These routines are the interface between the Acme system and the database
engine. They also time the transactions, track number of rows (tuples)
changed, and compose error messages if needed.

This version uses the Perl DBI interface. However, the insert(), load(), and
unload() routines use SQL statements specific to the PostgreSQL database.
The execute_sql() therefore uses a Postgres driver-specific attribute for the
log files.

=head1 TODO

    Make into a switchboard to the proper database interface routines as per
    the database engine settings in Acme::Base::Schema.

=head1 CHANGE LOG

 1/1/00   Break out interface routines into this module
 4/28/00  Remove requirement for parse_field_names(). Now done in
          Acme::Db::Parse (which has been changed to Acme::SQL).
 8/20/01  Added no_undefs flag to set undefined (null) values to empty string.
 11/18/01 Quickly converted to use DBI. Still dependent upon Postgres.
  5/1/02  Modified execute_sql() to execute sql as one transaction.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1997-2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

#
# Note: The $db_obj objects used below are actually belong to Acme::Db.
# Database scripts and results are stored in the Acme::Db object.
#

use Acme::Config;		# Acme configuration 
use DBI;
use Text::ParseWords;		# used to split multiple SQL statements

#------------------------------------------------------------------------------

sub connect {		# connect to database
	my ($db_obj, $connect_string, $user, $password, $attr) = @_;

	$attr ||= {};

	# Set flags telling DBI that we do our own error checking.
	$attr->{PrintError} = 0;	# don't report errors via warn()
	$attr->{RaiseError} = 0;	# don't report errors via die()

	if ($db_obj->{dbh} = DBI->connect($connect_string,$user,$password,$attr)) {
		# Set flag to true.
		$db_obj->{is_connected} = 1;
		return $db_obj;
	}
	else {
		$db_obj->error($DBI::errstr);
		return;
	}
}

#------------------------------------------------------------------------------

sub disconnect {	# disconnect from database
	my ($db_obj) = @_;

	return unless $db_obj->{dbh};

	$db_obj->{sth}->finish() if $db_obj->{sth};

	if ($db_obj->{dbh}->disconnect) {
		# Set flag to false.
		$db_obj->{is_connected} = 0;
		return $db_obj;		# true
	}
	else {
		$db_obj->error($DBI::errstr);
		return;
	}
}

#------------------------------------------------------------------------------

sub escape {		# quote text for SQL statements
	my $s = shift;

	# Return string with non-ascii characters encoded into HTML entity
	# characters, such as $#010; for the line-feed character.
	# Normal ascii characters are left untouched.
	#
	# Single quotes are doubled as per SQL92 standard.
	# NOTE: This assumes the programmer uses the single quotes to
	# delimit strings (I use to use double quotes).
	#
	# Note: we cannot use DBI->quote method as it requires a database
	# handle which this package does not require. It also surrounds
	# the returned string with quotes, which this method does not,
	# and it returns the string NULL for undefined variables. This 
	# returns an empty string.

	return '' unless defined($s);

## Not needed for Postgres, they even allow embedded newlines.
##	$s =~ s/([^\w !"#\$%&'()*+,.\/:;<=>?@[\\\]^`{|}~-])/sprintf("&#%03d;",ord($1))/egs;

	$s =~ s/'/''/gs;	# Double string quotes

## I do not know their limit yet...
##	$s = substr($s,0,255);	# maximum string length

	return $s;
}

#------------------------------------------------------------------------------

sub prepare_query {	# prepare SQL query script
	my ($db_obj,@param) = @_;

	# Note: Postgres does not use parameters @param in SQL like DBI.

	# Note: if the database has no prepare query process, execute the query
	# here instead of execute_query(). Two reasons, the first is that
	# the statement handle $db_obj->{sth} needs to be defined here for
	# parse_field_names() if a database routine is used, and second,
	# error messages will point to the place in the code where the 
	# query is defined, not the line where the Acme::Db->next() method
	# is called.

	if (! ${$db_obj->{sqlref}}) {
		$db_obj->error('Empty SQL statement.');
		return;
	}

	${$db_obj->{sqlref}} =~ s/\r//gs;	# Postgres complains about CRs

	# Note: DBI can only process one statement at a time.
	# We need to split possible compound SQL statements.

	my $count = 0;				# number of statements processed
	${$db_obj->{sqlref}} =~ s/;\s*$//;	# trim trailing semi-colons
	my @s = Text::ParseWords::parse_line(';',1,${$db_obj->{sqlref}});
	for (@s) {
		if (! defined $_ || /^\s*$/) {
			$db_obj->error('Empty SQL statement.');
			return;
		}

		# Execute and time sql
		$db_obj->{start_time} = time;

		if ($count < $#s) {
			if (! $db_obj->{dbh}->do($_)) {
				# Record query and error in log file (for debug)
				$db_obj->error($DBI::errstr,$_);
				return;		# failure
			}
		}
		else {
			# On last clause in multi-statement query.
			if ($db_obj->{sth} = $db_obj->{dbh}->prepare($_)) {
				# Record transaction (for security)
				# Transaction also gets logged when executed.
				#$db_obj->log_transaction("Prepared QUERY",$_);
			}
			else {
				# Record query and error in log file (for debug)
				$db_obj->error($DBI::errstr||"No DBI error string.",$_);
				return;		# failure
			}
		}
		$count++;
	}

	return $count;			# success
}

#------------------------------------------------------------------------------

sub execute_query {	# execute prepared query
	my $db_obj = shift;

	if ($db_obj->{sth}->execute(@_)) {
		return 1;		# success
	}
	else {
		# Record query and error in log file (for debug)
		$db_obj->error($DBI::errstr);
		return;			# failure
	}
}

#------------------------------------------------------------------------------

sub open_cursor {	# open query cursor
	my $db_obj = shift;

	close_cursor($db_obj) if $db_obj->{is_cursor_open};

	# Postgres does not require a cursor to be explicitly opened

	# Initialize query structure.
	# Allocate array to use as buffer to read Postgres database.
	$db_obj->{buf} = [];

	# Reset pointers.
	$self->{bufp} = 0;	# current record read from database into buffer
	$self->{recp} = 0;	# current record parsed from buffer

	return $db_obj->{is_cursor_open} = 1;	 # Set flag to true.
}

#------------------------------------------------------------------------------

sub close_cursor {	# close query cursor
	my ($db_obj) = @_;

	# Record transaction (for security)
	$db_obj->log_transaction('QUERY ' . $db_obj->{recp});

	# Deallocate buffer (to be anal) (and for DBI and Apache::Registry)
	$db_obj->{buf} = undef;

	$db_obj->{is_cursor_open} = 0;	# Unsetup: Set flag to false

	return 1;
}

#------------------------------------------------------------------------------

sub fetch_rec {		# get next record from query
	my ($db_obj) = @_;

	# Note: Pre 4/8/00 this routine could use a raw buffer and let 
	# parse_rec() fill the array of values. However, now with filters
	# and before_group() and after_group() processing, this
	# routine should always fill an array pointed to by $db_obj->{buf}.

	# Note: DBI does not want to use my array reference, so redefine
	# $db_obj->{buf} to be theirs.

	if (my $sth = $db_obj->{sth}) {
		if ($db_obj->{buf} = $sth->fetchrow_arrayref()) {

			# increment pointer (record number)
			$db_obj->{bufp}++;

			return 1;	# success
		}
	}
	return;		# failure
}

#------------------------------------------------------------------------------

sub parse_rec {		# parse database record and make it the current record
	my ($db_obj) = @_;

	#
	# Parse look ahead buffer and place in the "current" record array.
	#
	# We don't really parse the values returned from the Postgres database.
	# Just copy the look-ahead buffer to the current record array.
	#

	my $buf = $db_obj->{buf};
	my $rec = $db_obj->{rec};
	my $n = 0;

	if (@$buf) {
		# Copy look-ahead buffer to record array.

		## Note: @$rec = @$buf; does not work.
		## Apparently, individual array elements are reallocated so
		## that the tied columns point to the wrong place.

		for ( @$buf ) { $$rec[$n++] = $_ }	# stuff the array

		# set undefined (null) values to empty string if wanted
		if ($db_obj->{no_undefs}) {
			for (@$rec) {
				$_ = '' if ! defined $_;
			}
		}

		$db_obj->{recp}++;	# increment current record pointer
	}
	else {
		# Clear $db_obj->{rec} array.

		## Note: this also does not work: @$rec = (undef) x @$rec;
		## for the same reasons as above.

		for ( @$rec ) { $_ = undef }
	}

	return 1;			# success
}

#------------------------------------------------------------------------------

sub execute_sql {	# execute non-query SQL script
	my ($db_obj,@param) = @_;

	${$db_obj->{sqlref}} =~ s/\r//gs;	# Postgres complains about CRs

	# Note: DBI can only process one statement at a time.
	# We need to split possible compound SQL statements.

	my $parse_done = 0;	# flag to handle parse_line() bug

	for my $sql (Text::ParseWords::parse_line(';',1,${$db_obj->{sqlref}})) {
		if (!$parse_done) {
			# Begin transaction
			$db_obj->{dbh}->{AutoCommit} = 0;	# begin work

			# Set flag for parse_line() bug below
			$parse_done = 1;
		}

		next if ! $sql || $sql =~ /^\s*$/;	# ignore empty clauses

		# Execute and time sql
		$db_obj->{start_time} = time;

		if (($db_obj->{sth} = $db_obj->{dbh}->prepare($sql))
		&& (my $count = $db_obj->{sth}->execute(@param)) ) {

			# Record transaction (for security)
			$db_obj->log_transaction("$db_obj->{sth}->{pg_cmd_status} $count",$sql);
		}
		else {
			$db_obj->{dbh}->rollback();	  # undo changes
			$db_obj->{dbh}->{AutoCommit} = 1; # return to default

			# Record query and error in log file (for debug)
			$db_obj->error("Transaction not completed.\n".$DBI::errstr,$sql);
			return;		# failure
		}
	}

	# The parse_line() function has a bug (?) where it gobbles a string like
	# "something'else''here;" when parsing on a ';' and there is a
	# mis-matched pair of quotes.

	if (! $parse_done && ${$db_obj->{sqlref}}) {
		# Record query and error in log file (for debug)
		$db_obj->error(
	"SQL statement did not parse. (Possibly because of unmatched quotes?)");
		return;		# failure
	}

	$db_obj->{dbh}->commit();		# commit transaction
	$db_obj->{dbh}->{AutoCommit} = 1;	# return to default state

	return 1;	# success
}

#------------------------------------------------------------------------------

sub insert {		# insert a new record into a database table
	my ($db_obj, $table, $key, $keyval) = @_;

	#
	# Insert a minimal record into database table.
	#
	# If $keyval is not specified, the key is assumed to be a
	# serial key and the next serial number for the key field
	# is generated.
	#
	# Return the value of $key for the record inserted.
	#

	my $sth;
	if ($keyval) {
		my $sql = "insert into $table ($key) values ('$keyval')";
		if ((my $count = $db_obj->{dbh}->do($sql)) > 0) {
			# Record transaction (for security)
			$db_obj->log_transaction("Insert $count",$sql);
			return $keyval;
		}
		else {
			# Record query and error in log file (for debug)
			$db_obj->error($DBI::errstr,$sql);
			return;		# failure
		}
	}
	else {
		# Generate next sequential number for key

		# Postgres keeps serial number in a special sequence table which
		# is named after the table and key field. The sequence name cannot
		# be greater than 31 characters (PostgreSQL v7.1).

		my $seqname;
		if (length($table) > 13) {
			if (length($key) > 13) {
				$seqname = substr($table,0,13)
				  . '_'
				  . substr($key,0,13)
				  . '_seq';
			}
			else {
				$seqname = substr($table,0,26-length($key))
				  . '_'
				  . $key
				  . '_seq';
			}
		}
		else {
			if (length($key) > 13) {
				$seqname = $table
				  . '_'
				  . substr($key,0,26-length($table))
				  . '_seq';
			}
			else {
				$seqname = $table . '_' . $key . '_seq';
			}
		}

		$sql = "insert into $table ($key) values (nextval('$seqname'))";

		my $sth = $db_obj->{dbh}->prepare($sql);
		if (my $count = $sth->execute()) {
			# Record transaction (for security)
			$db_obj->log_transaction("Insert $count",$sql);

			# Get Postgres object ID number of the inserted record
			# and use that to read the database generated ID number
			# from the record just inserted.

			my $oid = $sth->{pg_oid_status};
			my $result_sql = "select $key from $table where oid = $oid";
			my $sth = $db_obj->{dbh}->prepare($result_sql);

			if ($sth && $sth->execute()) {
				($keyval) = $sth->fetchrow_array;
				return $keyval;
			}
			else {
				# Record query and error in log file (for debug)
				$db_obj->error($DBI::errstr,$result_sql);
				return;
			}
		}
		else {
			# Record query and error in log file (for debug)
			$db_obj->error($DBI::errstr,$sql);
			return;
		}
	}
}

#------------------------------------------------------------------------------

sub load {		# load file into a database table
	my ($db_obj,$file,$insert_clause,$delimeter) = @_;

	#
	# Note: This is not fully implemented for Postgres. Only entire
	# tables are loaded. The column specifications in the $insert_clause
	# is ignored. The procedure for Postgres 6.5.3 should be:
	#
	#	use Acme::Base::Schema;
	#	$field_list = fields listed in $insert_clause
	#	lookup database field types for the fields in $field_list;
	#	create temp table with only fields listed in $field_list;
	#	copy $temp_table from '$file' using delimiters '$delimeter';
	#	$insert_clause .= "select fields listed $field_list" .
	#			  "from $temp_table";
	#	
	#

	my ($table) = $insert_clause =~ m/insert into ([^(\s]+)/i;

	# Execute and time sql
	$db_obj->{start_time} = time;

	my $sql = "copy $table from '$file' using delimiters '$delimeter';";

	if (my $count = $db_obj->{dbh}->do($sql)) {
		# Record transaction (for security)
		$db_obj->log_transaction("Load $count",$sql);
		return $count;	# success
	}
	else {
		# Record query and error in log file (for debug)
		$db_obj->error($DBI::errstr,$sql);
		return;
	}
}

#------------------------------------------------------------------------------

sub unload {		# unload database table to a file
	my ($db_obj,$file,$select_clause,$delimeter) = @_;

	my $temp_table = "t_unl$$";

	# Adjust sql query to create a temporary table with
	# only the fields listed in the select clause. The
	# temporary table is the table actually copied to the file.
	
	my $sql = $select_clause;

        # Postgres 'into temp table' clause is before 'from' (of last statement)
        $sql =~ s/(.*;.*?)(from)\b/$1\ninto temp $temp_table\n$2/si
         or $sql =~ s/\b(from)\b/\ninto temp $temp_table\n$1/si;

	($sql = $select_clause) =~ s/(.*)(\bfrom\b)/$1\ninto temp $temp_table\n$2/s;

	$sql .= ";\n" unless $sql =~ /;\s*$/;
	$sql .= "\ncopy $temp_table to '$file' using delimiters '$delimeter';";
	$sql .= "\ndrop table $temp_table;";

	for (Text::ParseWords::parse_line(';',1,$sql)) {
		next unless defined $_;
		next if /^\s*$/;		# ignore empty clauses 

		# Execute and time sql
		$db_obj->{start_time} = time;

		if (my $count = $db_obj->{dbh}->do($sql)) {
			# Record transaction (for security)
			$db_obj->log_transaction("Unload $count",$_);
			return $count;	# success
		}
		else {
			# Record query and error in log file (for debug)
			$db_obj->error($DBI::errstr,$_);
			return;
		}
	}

	return 1;	# success
}

1;
