package Acme::Form;

=head1 NAME

Acme::Form - Form processing methods

=head1 SYNOPSIS

    use Acme::Form;

    $f = new Acme::Form (
    	template => $Template,		# this or form or table option required
    	form     => $Form,		# form file name
    	request  => $Request,		# "New", "Edit", or "Query"
    	args     => %arg,		# optional hash of arguments if no cgi
    	cgi      => $cgi,		# optional CGI object if no args
    	sql      => $sql,		# optional sql to execute
    	table    => $table,		# optional table to construct template
	extend   => {
		    'My Command' => sub { return "This is my command output"; }
		    },
	show_all_audits => 0,		# if true, all tables' audit info shown
    	);

    print $f->as_html;

    # To save a form
    my $nrecords = $f->save;
     -or-
    if ($f->save) { print $f->result }

    # To process a query form
    ($search,$list_title) = $f->query;

    # To build a form template (not filled in)
    print FILE $f->template;


=head1 DESCRIPTION

Acme::Form creates forms and merges database information into form fields.

One of the named parameters C<template>, C<form>, or C<table> is required.

The C<template> parameter is the form file name. For security reasons, template
files are stored in a special directory specified by the Templates variable
defined in the site configuration file. The directory should be outside the
web site's document root hiearchy. However, storing template files outside
the document root directory complicates the construction and alteration of
forms, so an alternate directory AltTemplates can be specified in the site
configuration file which will be checked first before the Templates directory.

The C<form> parameter is a schema definition file stored in another special
directory specified by $Forms defined in Acme::Config.

The C<form> parameter first looks to see if a form specification file exists,
and if not, looks for a template file with the same name. If a template file
does not exist, then a form for a database table with the form's name is built.

The C<table> parameter is the name of a table to lookup in the schema
Acme::Base::Schema. The C<form> definition file is the same type of a file as
the schema definition file. It allows more than one form to be defined for
a table.

If the C<sql> paramter is not given, the database schema is used to construct
a default query.

The reasons for specifying the sql parameter is to query joining tables,
fill in display-only fields, or to provide default values for form fields.

The C<extend> parameter points to a hash of regular expressions and functions.
As the form is filled, the HTML/XML comment strings C<<!--$ ... -->> is looked
for (its the $ character that's important) and substituted with the output of
the function associated with the regular expression that matches the C<...>
part of the comment string.

Specifying the C<table> parameter saves time as the routines do not have to
infer what the primary table is by searching the form.

Function arguments are obtained from CGI parameters if they are not specified.
Therefore, a complete program for displaying a form can be:

    use Acme::HTML;	# Required for HTML programs errors.
    use Acme::Form;	# Form filling methods.
    use CGI;
    
    use Acme::Security;
    Acme::Security::check('Access');
    
    $cgi = new CGI;
    $cgi->cache('no');
    print $cgi->header(-type=>'text/html', -expires=>'-1d');

    print Acme::Form->new(table=>'customers')->as_html;

    exit;

Files that have multiple <form> tags needs to be chunked up and fed to 
this module one form at a time.

The primary_table() method looks for the C<_table> input field containing the
name of the form's primary table, and if not found, looks for the first input
field which has a name in the form of C<table__field>, and returns the table
name parsed from the field's name.

The save() method posts the CGI arguments to the database if they exist in
the form and if they are different from the form values. Security privileges
are checked and audit trails are kept if specified in the schema.

The query() method assembles a search script (see Acme::Searcher) out of the
CGI parameters from submitted from a query form. The search script can be
passed to the C<list.cgi> program to search the database and list the results.

The has_id_key() method returns true (1) if the key of the primary table of
the form is an ID key. It is a convenience method that is used by the Form
Saver. ID keys are generated by the database and when adding new records to
the database, the screen confirming the success of adding the record should
stay open so the user can copy the ID number for their paper records.

Errors raise exceptions and usually causes the program to die.

=head1 CHANGE LOG

 5/5/00  Initial version.
 9/15/00 Automatically reads templates from form files and checks security.
 5/2/01  Rewrite. Remove parsing of form commands/directives.
 7/14/01 Added Print request to fill a form and print the result.
 7/28/01 Redefined form parameter to be a schema type file.
 10/2/01 Delayed form processing to the as_html() method. Added save() method.
 2/16/02 Allow command extensions to be defined in new() method.
 3/17/02 Generalized form parameter to look for templates and schema tables.
 6/20/02 Merged query form processing into fill_form() and query() methods.
 4/9/03  Added report argument to pass on to query forms.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000,2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

use Acme::Config;
use CGI;

use Acme::Security;

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my $self = { 

		# User options default values
		form       => '',	# name of a "schema like" file
		template   => '',	# template html
		table      => undef,	# primary table
		request    => 'New',	# default mode (New or Edit)
		report     => undef,	# optional report for query
		sql        => '',	# optional sql query
		args       => undef,	# optional cgi like arguments
 
		# Results
		html       => undef,	# html form with filled values
		result     => '',	# succesful result message of save
		detail_result => '',	# formatted record to paste into
					# master form 
		n_save     => 0,	# number of SQL save statements

		# Working fields
		schema     => undef,	# pointer to Acme::Form::Schema
		tables     => undef,	# array of tables defined
		fields     => undef,	# array of fields defined
		has_id_key => undef,	# flag

		# Extensions
		extend    => {},	# extensions keyed on reg exp

		# Arguments
		@_
		};

	$self->{args}   ||= $self->{cgi} || new CGI;
	$self->{args}     = new CGI $self->{args} if ref($self->{args}) eq 'HASH';
	my $args          = $self->{args};
	$self->{table}  ||= $args->url_param('table')  || $args->param('table');
	$self->{report} ||= $args->url_param('report') || $args->param('report');

	#
	# Check Security
	#

	my $request      = $self->{request}  || 'Edit';

	for ( ($self->{template},$self->{form}, $self->{table}) ) {
		next unless $_;
		Acme::Security::check("Access",$_);
		Acme::Security::check("Insert",$_) if $request =~ /new/i;
		Acme::Security::check("Update",$_) if $request =~ /edit/i;
	}

	return bless $self, $class;
}

sub as_html {
	my $self = shift;

	$self->_get_form_html;

	#
	# Setup runtime inheritance of Form Filler methods
	#

	require Acme::Form::Filler;
	unshift our @ISA, 'Acme::Form::Filler' unless $self->can('fill');

	return $self->SUPER::fill();
}

sub as_template {
	my $self = shift;

	$self->_get_form_html;

	return $self->{html};
}

sub query {
	my $self = shift;

	#
	# Setup runtime inheritance of Form Query methods
	#

	require Acme::Form::Query;
	unshift our @ISA, 'Acme::Form::Query' unless grep { $_ eq 'Acme::Form::Query' } @ISA;

	return $self->SUPER::query();
}

sub save {
	my $self = shift;

	$self->_get_form_html;

	#
	# Setup runtime inheritance of Form Saver methods
	#

	require Acme::Form::Saver;
	unshift our @ISA, 'Acme::Form::Saver' unless $self->can('save_form');

	return $self->SUPER::save();
}

sub result {
	my $self = shift;

	#
	# Return result message of last save operation
	#

	# Note: the detail_result is a formatted record of the detail form
	# information used to paste into the master form.

	return $self->{result} . $self->{detail_result};
}

sub primary_table {
	my $self = shift;

	#
	# Return the table name parsed from the first field name found in form.
	#

	if (my $fref = $self->{html}) {
		my $table = '';

		# Look for special input field named '_table'.
		if ($fref =~ /<\s*input\s+.*?name\s*=\s*(["'])_table\1.*?value\s*=\s*(['"])(.*?)\2/si ) {
			$table = $3;
			return $table;
		}

		# Look for input field name like "table__field".
		while ($fref =~ /name\s*=\s*(["'])(.*?)\1/sgi ) {
			$table = $2;
			$table =~ s/__.*// or next;
			$table =~ s/^N\d+_//;	# strip possible sequence #
			return $table;
		}

		# Look for special input field named 'table'.
		if ($fref =~ /<\s*input\s+.*?name\s*=\s*(["'])table\1.*?value\s*=\s*(['"])(.*?)\2/si ) {
			$table = $3;
			return $table;
		}

		# Look for name spec. without quotes in tag attribute.
		while ($fref =~ /name\s*=\s*(\w*?)/gi ) {
			$table = $1;
			$table =~ s/__.*// or next;
			$table =~ s/^N\d+_//;	# strip possible sequence #
			return $table;
		}
	}
	return '';
}

sub has_id_key {
	my $self = shift;
	return $self->{has_id_key} || 0;
}

#-------------------------------------------------------------------------------

sub _get_form_html {
	my $self = shift;

	#
	# Get form template html
	#

	my $template     = $self->{template} || '';
	my $form         = $self->{form}     || '';
	my $table        = $self->{table}    || '';
	my $title        = $self->{title}    || '';
	my $request      = $self->{request}  || '';
	my $report       = $self->{report}   || '';

	if ($template) {

		#
		# Read form template
		#

		(Acme::Config->AltTemplates && open(TEMPLATE, "<"
		  . Acme::Config->AltTemplates
		  . (($template =~ m|^/|) ? '' : '/')
		  . "$template") )
		  or open(TEMPLATE, "<"
		  . Acme::Config->Templates
		  . (($template =~ m|^/|) ? '' : '/')
		  . "$template")
		  or die "Cannot open form template '$template'. $!.\n";

		local $/;
		$self->{html} = <TEMPLATE>;	# read entire file
		close(TEMPLATE);

		$self->{table} ||= $self->primary_table;
	}
	elsif ($form) {
		my $form_file = Acme::Config->Forms
		  . (($form =~ m|^/|) ? '' : '/')
		  . "$form";

		# Append suffix of ".txt" if cannot read a form file
		$form_file = "$form_file.txt"
		  if ! -r $form_file && -r "$form_file.txt";

		if (-r $form_file) {

			#
			# Build form from form specification file
			#

			require Acme::Form::Builder;
			require Acme::Form::Schema;	# subclass of Acme::Db::Schema.

			$self->{schema} = new Acme::Form::Schema { form => $form_file };

			$self->{html} = Acme::Form::Builder::make_form(
						form            => $form_file,
						table           => $table,
						title           => $title,
						request         => $request,
						report          => $report,
						schema          => $self->{schema},
						show_all_audits => $self->{show_all_audits},
						);
		}
		elsif (
		  (Acme::Config->AltTemplates && open(TEMPLATE, "<"
		  . Acme::Config->AltTemplates
		  . (($form =~ m|^/|) ? '' : '/')
		  . "$form") )
		  or open(TEMPLATE, "<"
		  . Acme::Config->Templates
		  . (($form =~ m|^/|) ? '' : '/')
		  . "$form")
		) {

			#
			# Form specification is a template file name
			#

			local $/;
			$self->{html} = <TEMPLATE>;	# read entire file
			close(TEMPLATE);

			$self->{table} ||= $self->primary_table;

			# Readjust parameters
			$self->{template} = $self->{form};
			$self->{form} = '';
		}
		else {
			#
			# Build form from database schema table
			#

			require Acme::Form::Builder;

			# Switch arguments for Acme::Form::Filler.
			$self->{form}  = undef;
			$self->{table} = $form;
			$self->{args}->delete('form') if $self->{args}->param('form');

			$self->{html} = Acme::Form::Builder::make_form(
						table           => $form,
						title           => $title,
						request         => $request,
						report          => $report,
						show_all_audits => $self->{show_all_audits},
						);
		}
	}
	elsif ($table) {

		#
		# Create form template from database schema.
		#

		require Acme::Form::Builder;

		$self->{html} = Acme::Form::Builder::make_form(
					table           => $table,
					title           => $title,
					request         => $request,
					report          => $report,
					show_all_audits => $self->{show_all_audits},
					);
	}

	die "Form template cannot be found.\n" unless ($self->{html});
}

1;
