package Acme::Form::Assign;

=head1 NAME

Acme::Form::Assign - Translate AcmeBase assignment statements into JavaScript.

=head1 SYNOPSIS

    use Acme::Form::Assign;

    my $asgn = new Acme::Form::Assign
		form	     => $form,	# form name to use for target fields
		result	     => $result,# object name for result (form) fields
		search_table => $table,	# table name to search to produce the
					# result. Optional if not tracking
					# columns used in assignments.
		t	     => $t,	# AcmeBase schema object for the table
					# whose form fields are being filled.
					# Not the table being searched.
		context	     => $c;	# Acme::Form::Context object which 
					# controls translation


    $javascript .= $asgn->assign_options( $f->search_assign_options );
    $javascript .= $asgn->assign_fields(  $f->search_assign         );
    $javascript .= $asgn->assign_display( $f->search_assign         );
    $javascript .= $asgn->assign_input(   $f->search_assign_input   );


=head1 DESCRIPTION

The I<Acme::Form::Assign> object stores the arguments to the assignment statements
translators. The translators are used to translate assignment statements in
AcmeBase schema files into the actual target and resultant form/objects.

The assign_fields() method translates simple form field assignments. The
assign_display() method processes the same statements as assign_fields() but
produces assignments for only the display-only fields. It is used to fill forms
for editing, as the contents of the other database fields on a form should be
the actual contents of the database record and not by values computed after a
form is displayed.

The assign_options() and assign_input() methods alters the available selections
in a dropdown list or alters which input fields are displayed when entering a
text array. In other words, instead of altering the contents of form fields,
they restrict the options available to the user to a subset of the original
set of options. Both assign_options() and assign_input() use text arrays
(;one;two;three;) to alter the options available to the user.

For example, a dropdown list of color codes on an order form is set up to
contain all the colors codes available for all order items in inventory.
After searching for and choosing an order item, the dropdown list of color
codes can be restricted to be just the color codes applicable to the order
item.

An input array is an AcmeBase concept used to enter the components of a
text array.

For example, an input array can be setup to enter all sizes available for all
order items in the inventory. Similar to assign_options(), after searching
for and choosing an order item, assign_input() restricts the input fields
to showing only those used for entering the sizes applicable to the order item.

=head1 CHANGE LOG

 10/22/01 Initial version (hack).

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

use Text::ParseWords;

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my $self = {

		#
		# All the following settings must be in the arguments.
		#

		form	     => undef,	# form name to use for target fields
		result	     => undef,	# object name for result (form) fields

		search_table => undef,	# table name to search to produce the
					# result. Used to track columns.
		t	     => undef,	# AcmeBase schema object for the table
					# whose form fields are being filled.
					# Not the table being searched.
		context	     => undef,	# Acme::Form::Context object which 
					# controls translation

		@_,			# redefine above from function arguments

		#
		# The SQL columns needed to fulfill the assignments are
		# collected as the statements are translated.
		#

		columns	     => [],

		};

	$self->{targets} ||= {
		S => $self->{result},		# used in search_assign clauses
		C => $self->{result},		# used in choose_assign clauses
		M => $self->{result},		# used in master_assign clauses
		D => $self->{result},		# used in detail_assign clauses
		};

	return bless $self, $class;
}

sub columns {
	my $self = shift;

	# Return SQL columns needed for assignments.
	return @{$self->{columns}};
}

sub assign_fields {
	my $self = shift;
	my ($assign_fields) = @_;

	return $self->_interpret_assignments(
			$self->{context}->translate(
				$assign_fields,
				form   => $self->{form},
				target => $self->{targets},
				),
			\&_lvalue,
			\&_rvalue,
			0,
			);
}

sub assign_display {
	my $self = shift;
	my ($assign_display) = @_;

	return $self->_interpret_assignments(
			$self->{context}->translate(
				$assign_display,
				form   => $self->{form},
				target => $self->{targets},
				),
			\&_lvalue,
			\&_rvalue,
			1,	# display only flag
			);
}

sub assign_options {
	my $self = shift;
	my ($assign_options) = @_;

	#
	# Translate and interpret search assignment clauses.
	#

	return $self->_interpret_assignments(
			$self->{context}->translate(
				$assign_options,
				form   => $self->{form},
				target => $self->{targets},
				),
			\&_lvalue_options,
			\&_rvalue,
			);
}

sub assign_input {
	my $self = shift;
	my ($assign_input) = @_;

	return $self->_interpret_assignments(
			$self->{context}->translate(
				$assign_input,
				form   => $self->{form},
				target => $self->{targets},
				),
			\&_lvalue_input,
			\&_rvalue,
			);
}

#-------------------------------------------------------------------------------
# Helper functions
#-------------------------------------------------------------------------------

sub _interpret_assignments {
	my $self = shift;
	my ($clause, $lvalue_func, $rvalue_func, $displayonly) = @_;
	return '' unless $clause;

	# Parse assignment statements calling the _rvalue() function with the
	# right side of the equation and then calling the _lvalue() function
	# with the left side and right side of the equation.
	# The _rvalue() function should return the (modified) right side of the
	# assignment equation, whereas the _lvalue() function should return the
	# entire assignment statement.

	my $assignments = '';

	my ($lvalue, $rvalue);
	for my $stmt (Text::ParseWords::parse_line('[;\n]+', 1, $clause)) {
		next unless $stmt;
		($lvalue, $rvalue) = $stmt =~ /\s*([\w.]+)\s*=\s*(.+)/
		  or ($lvalue,$rvalue) = ('',$stmt);

		# ignore all but display fields
		next if $displayonly && $lvalue !~ /\bD_/;

		# Set up globals used in functions to translate left and
		# right side fo the assignments.

		local our $form         = $self->{form};
		local our $result       = $self->{result};
		local our $search_table = $self->{search_table} || '';
		local our $t            = $self->{t};
		local our $schema       = $t->schema;
		local our $columns      = $self->{columns};

		$rvalue = &$rvalue_func($rvalue);

		$assignments .= &$lvalue_func($lvalue, $rvalue) . ";\n";
	};

	return $assignments;
}

sub _rvalue {
	# Assignment's right side value

	my ($rvalue) = @_;

	use vars qw($form $result $search_table $columns $schema);

	# Accept an rvalue of 'table.field' format
	# for backwards compatability.
	if ($rvalue =~ /^(\w+)\.(\w+)\s*$/) {
		if ($1 ne $form && $1 ne $result) {
			$rvalue =~ s/\./__/;
			# Prepend form spec to rvalue
			$rvalue = "$result.$rvalue.value";
		}
	}

	# Track columns needed to add to sql query.

	my ($tbl, $fld);
	while ($rvalue =~ /\b$result\.(.*?)\.value\b/g) {
		($tbl, $fld) = $1 =~ /(.+?)__(.+)/
		  or ($tbl, $fld) = ($search_table, $1);

		push @$columns, "$tbl.$fld";
	}

	# Look for radio field values and sub in function to
	# determine the radio button value.
	#
	# Note: the following is superfluous if the browser allows
	# the retreiving of radio button values as other input fields,
	# like:
	#   var myval = myform.myradio_field.value;

	# Don't check if it appears that the programmer is already
	# using the radio_value() or checkbox_value() functions in
	# the assignment statements.

	return $rvalue if $rvalue =~ /(radio_value|checkbox_value)/;

	#
	# Define helper function for substituting variables in assignment
	#

	my $var = sub {
		# substitute function radio_value(fld) or checkbox_value(fld)
		# for fld.value
		my $val = shift;

		# strip '.value' component off of field name.
		(my $check_val = $val) =~ s/\.value$//;

		(my $field_name = $check_val) =~ s/^.*\.//;

		($tbl, $fld) = $field_name =~ /(.+?)__(.+)/
		  or ($tbl, $fld) = ($search_table, $field_name);

		my $t = $schema->t($tbl) or return $val;
		my $f = $t->{$fld};
		$f ||= $t->display->{$field_name} if $field_name =~ /^D_/;
		return $val unless $f;

		return "radio_value($check_val)" if $f->radio;
		return "checkbox_value($check_val)" if $f->checkbox;
		return $val;	# no changes
	};

	$rvalue =~ s/\b([\w.]+\.value)\b/&$var($1)/ge;

	return $rvalue;
}

sub _lvalue {
	# Assignment's left side value for input fields

	my ($lvalue, $rvalue) = @_;

	# Return complete assignment statement for
	# an lvalue and rvalue.

	use vars qw($form $t);

	return $rvalue unless $lvalue;

	my ($field_name, $tbl, $fld);
	if (($field_name) = $lvalue =~ /\b$form\.(.*?)\.value\b/) {
		($tbl, $fld) = $field_name =~ /(.+?)__(.+)/
		  or ($tbl, $fld) = ($t->name, $1);

		my $f = $t->{$fld};
		$f ||= $t->display->{$field_name} if $field_name =~ /^D_/;

		if ($f && $f->array && $f->choose) {
			# Multiple assignments to a field array.

			# Assure $lvalue is a field ref.
			$lvalue =~ s/\.value$//;

			return "assign_multiple($lvalue, $rvalue)"
		}

		if ($f && $f->radio) {
			# Set radio field

			# Assure $lvalue is a field ref.
			$lvalue =~ s/\.value$//;

			return "radio_set($lvalue, $rvalue)"
		}

		if ($f && $f->checkbox) {
			# Set checkbox field

			# Assure $lvalue is a field ref.
			$lvalue =~ s/\.value$//;

			return "checkbox_set($lvalue, $rvalue)"
		}
	}

	return "$lvalue = $rvalue";
};

sub _lvalue_options {
	# Assignment's left side value for assigning options to dropdown lists

	my ($lvalue, $rvalue) = @_;

	# Return complete assignment statement for
	# an lvalue and rvalue.

	# Assure $lvalue is a field ref.
	$lvalue =~ s/\.value$//;

	return $rvalue unless $lvalue;
	return "assign_options($lvalue, $rvalue)";
};

sub _lvalue_input {
	# Assignment's left side value for assigning values to input arrays.

	my ($lvalue, $rvalue) = @_;

	# Return complete assignment statement for
	# an lvalue and rvalue.

	# Assure $lvalue is a field ref.
	$lvalue =~ s/\.value$//;

	return $rvalue unless $lvalue;
	return "assign_input($lvalue, $rvalue)";
};

1;
