package Acme::Form::Builder;

=head1 NAME

Acme::Form::Builder - Make form template from database schema

=head1 SYNOPSIS

    use Acme::Form::Builder;

    my $template = Acme::Form::Builder::make_form(table => 'table_name');
      or
    my $template = Acme::Form::Builder::make_form(
        table => 'table_name',
	title => $title
	);
      or
    my $template = new Acme::Form::Builder
    	table => 'table_name',
	name  => 'form_name',
	title => 'Title for Form,
	type  => 'form';		# or 'query' or 'editlist'


=head1 DESCRIPTION

I<Acme::Form::Builder> produces a form template from the information in the
schema definition file used by I<Acme::Base::Schema>.

There are currently two types of templates: 'form' and 'query'. The future
will probably bring extra form types of 'new' and 'edit' instead of the
generic 'form' type. There will probably be an 'editlist' type too, to 
produce a form that can be used to update a whole list of records.

The type 'form' is intended to be used to enter and update database information.

The template type of 'query' is used to create a form to use as a
"Query by Form" form. The database fields are plain text fields without
any data translation or verification. The fields are intended to be filled
with search parameters that the C<query.cgi> program assembles into a SQL
statement that can be used by C<list.cgi>.

=head1 CHANGE LOG

 4/28/00  Initial version.
 6/17/00  Added query form type.
 4/17/01  Added context and translation of form fields in schema scripts.
 5/2/01   Updated to Acme::Base::Schema.
 3/22/02  Updated to Perl 5.6 and "use strict".

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000,2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

use Acme::HTML			qw(escape_html);
use Acme::HTML::HideMoreLess	qw(start_hiding		stop_hiding);
use Acme::Form::Schema;
use Acme::Base::Schema::Join;
use Acme::SQL;
use Acme::Form::Context;
use Acme::Config;
use Local::Acme::User		qw(user);
use Local::Acme::Form::Banners	qw(common_header	common_footer
				   common_form_header	common_form_footer
				   common_query_header	common_query_footer);
use Text::ParseWords;

our $Base	     = Acme::Config->Base;
our $Root	     = Acme::Config->Root;
our $Scripts	     = Acme::Config->Scripts;
our $Scripts_Lib     = Acme::Config->Scripts_Lib;
our $Styles	     = Acme::Config->Styles;
our $Images	     = Acme::Config->Images;
our $Help            = Acme::Config->Help;
our $No_Tabindex     = Acme::Config->No_Tabindex;
our $Show_All_Audits = Acme::Config->Show_All_Audits;
our $Default_Form    = Acme::Config->Default_Form;
$Default_Form = "'$Default_Form'"	# Enclose in quotes if non-numeric
  if $Default_Form !~ /^\d+$/;

our $Field_Size = 50;		# Average input field size (advisory)

our $Tabindex_Interval = 10;	# Each field's tabindex value in incremented
				# by this value. It allows a designer to 
				# insert another field into the tabbing order.

our $Max_Input_Array = 90;	# Input arrays generate a new <input> tag for
				# value in an array which are inserted between
				# other <input> tags. $tabindex is incremented
				# by this amount for fields which are for
				# database array fields with the Input property.

our $Field_Prefix = 'f_';	# Fields and their attached graphics and buttons
				# are contained in a <span> with an ID of this
				# prefix prepended to field name. This constant
				# must match those defined in form_validate.js.

our $Label_Prefix = 'l_';	# Field labels use an HTML element ID of this
				# prefix prepended to field name. This constant
				# must match those defined in form_validate.js.

# Global variables
our $Schema;
our $Context;

our ($compute_scripts, $form_init_scripts, $init_scripts, $include_scripts,
     $combo_script, $calendar_script, $color_script, $print_script,
     $include_stylesheets, $audit_trail, $detail_placed, $tabindex);

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	# Return template HTML code (not an object).
	return make_form(@_);
}

sub make_form {

	# Build hash of function options with default values and argument values.
	my $opt = {
		table		=> undef,	# required table name
		title		=> '',		# optional form title
		type		=> '',		# 'form' or 'query'
		@_
		};

	my $tmplt;	# constructed form template returned to caller

	my $request	= $opt->{request} || '';	# New or Edit or ...

	my $form	= $opt->{form}    || '';

	my $table	= $opt->{table}   || '';

	my $report	= $opt->{report}  || '';	# Only used for Query

	my $form_name	= $opt->{name}    || $table	|| "form$$";
	my $form_type	= $opt->{type}    || ($request =~ /Query/i ? 'query' : 'form');

	my $show_all_audits = $opt->{show_all_audits} || $Show_All_Audits;

	local $Schema   = $opt->{schema}  || new Acme::Form::Schema { form => $form };

	# First table listed in form specification file is assumed to
	# be the primary table to build the form for.

	($table) = $Schema->tables if ! $table && $form;

	my $t = $Schema->t($table);

	die "Table '$table' is not defined in the database schema"
	  unless $t;

	#
	# Define global variables to hold scripts. They are inserted into
	# $tmplt's header section after rest of form has been processed.
	#

	# Note: initialization scripts are grouped and performed in the
	# order of:
	#	1) form initialization scripts from the system (e.g. buttons),
	# 	2) table initialization scripts in the schema files,
	# 	3) the field initialization scripts specified for each field
	# 	   in the schema files.

	local $compute_scripts		= '';	# entered in form_compute()
	local $form_init_scripts	= '';	# entered in fields_initialize()
	local $init_scripts		= '';	# entered in fields_initialize()

	local $include_stylesheets	= '';	# optional css stylesheets
	local $include_scripts		= '';	# optional script libraries

	local $audit_trail		= '';	# subordinate table audit info

	# Define script for combo buttons.
	local $combo_script = '';

	# Define script for calendar buttons.
	local $calendar_script = '';	# calendar library

	# Define script for color buttons.
	local $color_script = '';	# colorwheel library

	# Define script for print button.
	local $print_script = '';	# custom script for each form

	# Define flag for table detail placement.
	local $detail_placed = 0;	# clear

	# Initialize tabindex counter.
	local $tabindex = 0;

	# Initialize current context for form field name translation.
	local $Context = new Acme::Form::Context table=>$table;

	#
	# Compose template header.
	#

	my $title = $opt->{title} || $t->title;

	#
	# Evaluate do_builder property's Perl script.
	# Note: The do_builder property subverts Perl's
	# taint checks as the schema (and form) files
	# are assumed to be protected files.
	#
	
	no strict;
	my $do = join ';', $t->do, $t->do_builder;
	($do) = $do =~ /^(.*)$/s;	# untaint
	eval $do;
	die "Error evaluating schema's Do property. $@" if $@;
	use strict;

	#
	# Start building a form.
	#

	($form_type) = $form_type =~ /(\w*)/;	# untaint
	if ($form_type eq 'query') {
		$tmplt .= _make_header_query($form_name,$t,$table,$title,$report);
	}
	else {
		$tmplt .= _make_header_form($form_name,$t,$table,$title);
	}

	#
	# Compose main table form.
	#

	$tmplt .= qq[<table class="form">\n];

	$tmplt .= _make_table_id_form($form_name,$form_type,$t);

	$tmplt .= _make_table_form($form_name,$form_type,$request,$t);

	#
	# Compose joined table forms.
	#

	if (my $join_clause = $t->join) {
		my $j = new Acme::Base::Schema::Join $table, $join_clause;
		my ($fld, $join_tab, $join_fld);
		for ($j->joins) {
			$fld = $_->{field};
			$join_tab = $_->{join_table};
			$join_fld = $_->{join_field};

			if (my $jt = $Schema->t($join_tab)) {

				#
				# Evaluate do_builder property's Perl script.
				# Note: The do_builder property subverts Perl's
				# taint checks as the schema (and form) files
				# are assumed to be protected files.
				#
				
				no strict;
				my $do = join ';', $jt->do, $jt->do_builder;
				($do) = $do =~ /^(.*)$/s;	# untaint
				eval $do;
				die "Error evaluating schema's Do property. $@" if $@;
				use strict;

				#
				# Make form fields for joined table.
				#

				$Context->set(table => $jt->name);

				local $detail_placed = 0;	# clear flag

				$tmplt .= _make_table_form(
							$form_name,
							$form_type,
							$request,
							$jt,
							$join_fld);

				if (! $detail_placed && $form_type ne 'query'
				&& (my $detail = $jt->detail_field)) {
					my $field_name = $Context->db_to_form($jt->key);
					$tmplt .= _table_detail(
								$detail,
								$form_name,
								$field_name,
								$jt->name,
								$jt->key,
								$jt,
								undef,
								$request);
				}

				# Store audit info to display all at once 
				# with the primary table audit info.

				$audit_trail .= _make_table_audit(
								$form_name,
								$form_type,
								$jt);
			}
		}
	}

	if (! $detail_placed && $form_type ne 'query'
	&& (my $detail = $t->detail_field)) {
		my $field_name = $Context->db_to_form($t->key);
		$tmplt .= _table_detail(
					$detail,
					$form_name,
					$field_name,
					$table,
					$t->key,
					$t,
					undef,
					$request);
	}

	#
	# Compose footer
	#

	if ($t->audit) {
		$tmplt .= <<End_Of_HTML;

<tr>
  <td colspan="2">

  <div class="audit">
  <table class="audit">
End_Of_HTML

		$tmplt .= _make_table_audit(
					$form_name,
					$form_type,
					$t);

		# Append audit info for all tables on the form.
		$tmplt .= $audit_trail if $show_all_audits;

		$tmplt .= <<End_Of_HTML;
  </table>
  </div>

  </td>
</tr>
End_Of_HTML
	}

	$tmplt .= qq[</table>\n];

	if ($form_type eq 'query') {
		$tmplt .= _make_footer_query($form_name,$t,$table);
		$tmplt .= common_query_footer($title);
	}
	else {
		$tmplt .= common_form_footer($title);

		# Attach a hidden field of a private field name that identifies
		# the name of the current form. It is used by the Saver to 
		# save the correct form on a multi-form html page.

		$tmplt .= qq[<input type="hidden" name="_form_name" value="$form_name">\n];
	}

	$tmplt .= qq[</form>\n];

	$tmplt .= common_footer($title);

	$tmplt .= <<End_Of_HTML;
</body>
</html>
End_Of_HTML

	#
	# Insert scripts into header scripts.
	#

	if ($form_init_scripts || $init_scripts) {
		# Combine initialization scripts.
		$init_scripts =~ s/^\s+/\n/;	# clean up extraneous linefeeds
		$form_init_scripts .= $init_scripts;

		# Final formatting for pretty printing.
		$form_init_scripts =~ s/^\s+/\n/;
		$form_init_scripts =~ s/\s+$/\n/;
		$form_init_scripts =~ s/\n/\n\t/g;

		$tmplt =~ s|(// Set fields to their default values here.)|$1$form_init_scripts|
	}

	if ($compute_scripts) {
		# Final formatting for pretty printing.
		$compute_scripts =~ s/^\s+/\n/;
		$compute_scripts =~ s/\n/\n\t/g;

		$tmplt =~ s|(// Compute field values here.)|$1$compute_scripts|
	}

	#
	# Insert 'Include' scripts libraries in header.
	#

	my $include = "$color_script$calendar_script$include_scripts$combo_script$print_script";
	$tmplt =~ s|<!-- Include Scripts -->|$include|;

	#
	# Insert 'Stylesheet' CSS libraries in header.
	#

	$tmplt =~ s|<!-- Include Stylesheets -->|$include_stylesheets|;

	return $tmplt;
}

#-------------------------------------------------------------------------------
#
# Private routines
#
#------------------------------------------------------------------------------

sub _make_header_form {
	my ($form_name,$t,$table,$title) = @_;

	my $save_msg   = 'Saving ' . $t->row_name . '.';
	my $save_title = 'Save '   . $t->title;

	my $save_prog = $t->save_prog;
	$save_prog .= ($save_prog =~ /\?/ ? '&' : '?');
	$save_prog .= "table=$table";

# Keep IE6 in IE5 compatablilty mode by not specifying url
# It not only changes CSS spacing here, but it also causes the
# Search button to not focus the popup window.
#<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
#"http://www.w3.org/TR/html4/loose.dtd">

	my $tmplt = <<End_Of_HTML;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">

<html>
<head>
<meta http-equiv="Content-Type" content="text/html; iso-8859-1">
<meta http-equiv="pragma" content="no-cache">
<meta name="generator" content="AcmeBase 0.5">

<title>$title</title>

<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<link rel="stylesheet" href="$Styles/form.css" type="text/css">
<!-- Include Stylesheets -->
<script src="$Scripts_Lib/form_events.js" type="text/javascript"></script>
<script src="$Scripts_Lib/form_validate.js" type="text/javascript"></script>
<script src="$Scripts_Lib/form_save.js" type="text/javascript"></script>
<!-- Include Scripts -->
<script language="JavaScript" type="text/javascript">

function form_initialize(f) {
	if (f == null) { f = self.document.forms[$Default_Form] }

	form_is_setup(f,true);		// set flag

	fields_initialize(f);

	form_compute(f);		// initialize display-only fields, ...

	form_is_dirty(f,false);		// clear flag

	// Set keyboard focus to first field on form.
	if (f._request && f._request.value.match(/new/i)) {
		focus_first_blank_field(f); // goto first blank field on form
	} else {
		focus_first_field(f);	// goto first field on form
	}
}

function form_reinitialize(f) {
	if (f == null) { f = self.document.forms[$Default_Form] }

	f.reset();			// this clears all fields

	fields_initialize(f);		// initialize fields with default values

	form_compute(f);		// initialize display-only fields, ...

	form_is_dirty(f,false);		// clear flag

	// Set focus to first field on form.
	//focus_first_field(f);		// goto first field on form
	focus_first_blank_field(f);	// goto first blank field on form
}

function fields_initialize(f) {
	if (f == null) { f = self.document.forms[$Default_Form] }

	// Set fields to their default values here.
}

function form_compute(f) {
	// This is executed after every input and just before saving the form.
	if (f == null) { f = self.document.forms[$Default_Form] }

	// Compute field values here.

	form_is_dirty(f,true);		// set flag
	return true;
}
</script>
</head>

<body onLoad="form_initialize(self.document.forms['$form_name'])">

End_Of_HTML

	$tmplt .= common_header($title);

	$tmplt .= <<End_Of_HTML;
<form   name="$form_name"
	action="$save_prog"
	enctype="multipart/form-data"
	method="POST"
	onSubmit="return form_save(this,'$save_msg','$save_title')"
>
End_Of_HTML

	$tmplt .= common_form_header($title);

	return $tmplt;
}

#------------------------------------------------------------------------------

sub _make_header_query {
	my ($form_name,$t,$table,$title,$report) = @_;

	#
	# Note: this header is different from forms:
	#	* no field validation occurs,
	#	* computations do not occur,
	#	* query.cgi is called instead of save.cgi,
	#	* target of form is a list window, not a message window,
	#	* search instructions button is included.
	#

	my $query_prog = $t->query_prog;
	$query_prog .= ($query_prog =~ /\?/ ? '&' : '?');
	$query_prog .= "table=$table";
	$query_prog .= "&report=$report" if $report;

	$tabindex += $Tabindex_Interval;	# adjust global
	my $tabindex_str = qq[ tabindex="$tabindex"];
	$tabindex += $Tabindex_Interval;	# adjust global
	my $tabindex_str1 = qq[ tabindex="$tabindex"];

# Keep IE6 in IE5 compatablilty mode by not specifying url
#<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
#"http://www.w3.org/TR/html4/loose.dtd">

	my $tmplt = <<End_Of_HTML;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">

<html>
<head>
<meta http-equiv="Content-Type" content="text/html; iso-8859-1">
<meta name="generator" content="AcmeBase 0.5">

<title>$title</title>

<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<link rel="stylesheet" href="$Styles/form.css" type="text/css">
<!-- Include Stylesheets -->
<script src="$Scripts_Lib/form_events.js" type="text/javascript"></script>
<script src="$Scripts_Lib/form_validate.js" type="text/javascript"></script>
<script src="$Scripts_Lib/form_query.js" type="text/javascript"></script>
<!-- Include Scripts -->

</head>

<body>

End_Of_HTML

	$tmplt .= common_header($title);

	my $window_name = "list_$$" . int rand 1000;

	my $browse_button = '';

	$tmplt .= <<End_Of_HTML;
<form   name="$form_name"
	action="$query_prog"
	enctype="multipart/form-data"
	method="POST"
	target="$window_name"
	onSubmit="open_list('','$window_name'); return true"
>
End_Of_HTML

	$tmplt .= common_query_header($title);

	$tmplt .= <<End_Of_HTML;
<h2>
Construct Search
</h2>

<!-- Instructions ------------------------------------------------------------>

End_Of_HTML


	my $instr_file = "$Root/$Help/query_instructions.htm";

	if (open INSTR, '<', $instr_file) {

		#
		# Read instructions file and construct button to show them.
		#

		local $/;	# read entire file
		my $instructions = <INSTR>;
		close INSTR;

		#
		# Extract instructions between <body> and </body> tags.
		#

		$instructions =~ s|^.*?<body.*?>||si;
		$instructions =~ s|</body.*||si;

		#
		# Build button
		#

		my $instr = qq[<div align="left" class="instructions">\n]
		  . start_hiding("instruction_body",
			more_label	 => "Instructions",
			more_description => "Show instructions.",
			less_label	 => "Instructions",
			less_description => "Hide instructions.",
			);

		#
		# Print instructions
		#

		$instr .= <<EOF;
<div style="border:solid black 1px;padding:2ex">
$instructions
</div>
EOF

		$instr .= stop_hiding("instruction_body") . qq[</div>\n];

		$tmplt .= <<EOF;
<div style="margin:1ex 0ex">
$instr
</div>
EOF
	}

	return $tmplt;
}

#------------------------------------------------------------------------------

sub _make_footer_query {
	my ($form_name,$t,$table) = @_;

	#
	# Note: this footer is different from forms:
	#	* extra fields for a title and list are added.
	#

	$tabindex += $Tabindex_Interval;	# adjust global
	my $tabindex_str = qq[ tabindex="$tabindex"];
	$tabindex += $Tabindex_Interval;	# adjust global
	my $tabindex_str1 = qq[ tabindex="$tabindex"];

	my $browse_button = _browse_button(
					"lists=$table",
					$form_name,
					"list_exclude",
					);

	my $tmplt = <<End_Of_HTML;
<hr width="90%">

<table class="form">
<tr>
<td class="label">Title</td>
<td class="field"><input type=text name="list_title" value="" size=60$tabindex_str>
<br><small>You can put a title on the list produced by your search.</small>
</td>
</tr>

<tr>
<td class="label"><nobr>Exclude List</nobr></td>
<td class="field"><input type=text name="list_exclude" value="" size=50$tabindex_str1>$browse_button
<br><small>You can exclude records of a previous query that has been saved in a list file.</small>
</td>
</tr>
</table>
End_Of_HTML

	return $tmplt;
}

#------------------------------------------------------------------------------

sub _make_table_id_form {
	my ($form_name, $form_type, $t) = @_;

	# Table ID fields are treated differently from all others.

	my $tmplt = '';		# returned template fragment

	my $table = $t->name;

	if (my $id = $t->id) {
		my $field_name = "${table}__${id}";

		# Make title out of ID field name.
		my $field_label = $t->id_label;
		my $field_label_str = ($field_label =~ / /)
		  ? "<nobr>$field_label</nobr>"
		  : $field_label;
		my $field_label_style = $t->id_label_style;
		$field_label_style &&= qq[ style="$field_label_style"];

		# Enclose field label with an identifiable tag.
		$field_label_str = qq[<span id="$Label_Prefix$field_name"$field_label_style>]
		  . $field_label_str
		  . qq[</span>];

		my $field_style = $t->id_style;
		$field_style &&= qq[ style="$field_style"];

		my $bgn_fld = qq[<span id="$Field_Prefix$field_name"$field_style>];
		my $end_fld = qq[</span>];

		my $type = 'hidden';
		my $query_button = '';

		if ($form_type eq 'query') {
			$type = 'text';
			$query_button = _query_button($form_name, $field_name, undef);
		}

		$tabindex += $Tabindex_Interval;
		my $tabindex_str = qq[ tabindex="$tabindex"];

		$tmplt .= <<End_Of_HTML;
<tr>
  <td class="id-label">$field_label_str</td>
  <td class="id-field">$bgn_fld<input type="$type" name="$field_name" value="" size="11"$tabindex_str>$query_button$end_fld</td>
</tr>
End_Of_HTML
	}

	return $tmplt;
}

#------------------------------------------------------------------------------

sub _make_table_form {
	my ($form_name,$form_type,$request,$t,$join_fld) = @_;

	my $tmplt = '';		# returned template fragment

	my $table = $t->name;
	my $key   = $t->key;

	my ($elem, $field);
	my ($id, $field_label, $field_name, $type, $maxlength, $size, $decimal);
	my ($class, $readonly, $required, $hidden);
	my ($radio, $checkbox, $combo);
	my ($description, $default);
	my ($translate, $validate, $valid_error, $on_change, $on_focus);
	my ($choose, $search, $same, $detail, $browse);

	my ($required_symbol);
	my ($bgn_nobr, $end_nobr);
	my ($bgn_fld, $end_fld);
	my ($field_label_str);
	my ($field_label_style);
	my ($include, $stylesheet);
	my ($init_str, $translate_str, $check_str, $validate_str, $compute_str);
	my ($on_focus_str, $on_change_str);
	my ($readonly_str, $maxlength_str, $tabindex_str, $title_str);
	my ($class_str, $style_str);
	my ($combo_button, $combo_prefix, $combo_input);
	my ($search_button, $same_button, $browse_button, $query_button);
	my ($calendar_button, $color_button, $detail_button);
	my ($display_field, $display_label);
	my ($container_class, $container_style);
	my ($prev_value);

	# Define flags for rudimentory form control.
	my ($in_newline, $in_break, $in_insert, $first_line);

	# Define stack and flags for begin/hide/end blocks.
	my @blocktypes;	# 'b' block; 'i' insert; 'h' hidden;

	my @hiding;	# flags for hiding begin/end blocks
	my $hiding;	# flag set to true if any @hiding flags set
	my $is_hiding = sub { for (@hiding) { return $_ if $_ } return 0 };

	# Build hash of form fields the user can access.
	my %form_field = map {$_,1} $t->fields, $t->displays;

	#
	# Add initialize scripts to global.
	# Translate form variables (%name/$name) into actual form names/values.
	#

	# Retreive assignment scripts for master table.
	$form_init_scripts .= _master_assign($t) if $t->master;

	# Add initialize scripts global.
	if ($init_str = $t->initialize) {
		# Translate form variables (%name/$name) into actual
		# form names/values.
		$init_scripts .= "\n" . $Context->translate(
						$init_str,
						form=>'f',
						);
	}

	# Add include scripts global.
	if ($form_type eq 'query') {
		$include_scripts .= $t->query_include_html;
	}
	else {
		$include_scripts .= $t->form_include_html;
	}

	# Add include stylesheets global.
	$include_stylesheets .= $t->form_stylesheet_html;

	#
	# Main loop through fields and elements of schema for a table.
	#

	$id = $join_fld || $t->id;

	my ($f, $elemtype);
	$first_line = 1;
	$end_fld = '</span>';

	for $elem ($t->elems) {
		$elemtype = $t->elem->{$elem}->type || '';
		if ($elemtype eq 'html') {

			#
			# Check for form control directives.
			#

			for my $v ($t->elem->{$elem}->value) {
				if ($v =~ s/^<!--%\s*End(\s+(\S+))?\s*-->//i) {
					# End block

					redo if !@blocktypes;	# ingore superflous ends

					my $blocktype = pop @blocktypes;

					# Check if ending a hiding directive.
					if ($blocktype eq 'h') {
						pop @hiding;
						$hiding = &$is_hiding;
						redo;
					}

					if (! $hiding) {
						$tmplt .= qq[$end_fld</td>\n</tr>\n]
						  unless $first_line || $in_break;
						$tmplt .= qq[</table>\n]
						  unless $in_break;
						$tmplt .= qq[</div>\n];

						# Terminate insert blocks.
						$tmplt .= qq[</td>\n</tr>\n</table>\n]
						  if $blocktype eq 'i';

						$in_break = 1;
						$in_insert = $in_newline = 0;
					}
					redo;
				}
				if ($v =~ s/^<!--%\s*Begin\s+(\S+)(.*?)?\s*-->//si) {
					# Begin block

					if ($in_insert) {
						push @blocktypes, 'i';
						redo if $hiding;
					}
					else {
						push @blocktypes, 'b';
						redo if $hiding;

						$tmplt .= qq[$end_fld</td>\n</tr>\n]
						  unless $first_line || $in_break;
						$tmplt .= qq[</table>\n]
						  unless $in_break;
					}

					$tmplt .= qq[<div class="$1"$2>\n];

					$in_break = 1;
					$in_insert = $in_newline = 0;
					redo;
				}
				if ($v =~ s/^<!--%\s*Hide\s+(.*?)?\s*-->//si) {
					# Hide

					my $h = 0;
					if ($1) {
						for (split /\W+/, ($1 || '')) {

							# Start hiding if one of the
							# request codes on the Hide
							# line matches the current
							# Request code (which usually
							# comes from the cgi arguments)

							if ($_ =~ /$request/io) {
								$h = 1;	# start hiding
								last;
							}
						}
					}
					else {
						# Unconditionally start hiding.
						$h = 1;
					}

					if (! $hiding && $h) {
						$tmplt .= qq[$end_fld</td>\n</tr>\n]
						  unless $first_line || $in_break;
						$tmplt .= qq[</table>\n]
						  unless $in_break;

						$in_break = 1;
						$in_insert = $in_newline = 0;
					}

					push @hiding, $h;
					$hiding = &$is_hiding;	# set flag

					push @blocktypes, 'h';
					redo;
				}

				next if $hiding;

				if ($v =~ s/^<!--%\s*BR\s*-->//i) {
					# BR -- break

					$tmplt .= qq[$end_fld</td>\n</tr>\n]
					  unless $first_line || $in_break;
					$tmplt .= qq[</table>\n\n]
					  unless $in_break;
					$in_break = 1;
					$in_insert = $in_newline = 0;
					redo;
				}
				if ($v =~ s/^<!--%\s*NL\s*-->//i) {
					# NL -- newline

					if ($in_break) {
						$tmplt .= qq[\n\n<table class="form">\n];
						$in_break = 0;
						$first_line = 1;
					}
					$tmplt .= qq[$end_fld</td>\n</tr>\n]
					  unless $first_line;
					$tmplt .= "\n<tr>\n  <td colspan=2>";
					$in_newline = 1;
					$in_break = $in_insert = 0;
					$first_line = 0;
					redo;
				}
				if ($v =~ s/^<!--%\s*IN\s*-->//i) {
					# IN -- Insert same column on next line 

					if ($in_break) {
						$tmplt .= qq[\n\n<table class="form">\n];
						$in_break = 0;
						$first_line = 1;
					}

					$tmplt .= qq[$end_fld</td>\n</tr>\n]
					  unless $first_line;
					$tmplt .= "\n<tr>\n  <td></td>\n  <td>";
					$first_line = 0;
					$in_insert = 1;
					redo;
				}
				if ($v =~ s/^<!--%\s*DO\s+(.*)\s*-->/$1/iee) {
					# DO -- Do commands defined in do_builder property
					redo;
				}
				if ($v =~ s/^<!--%\s*Print(\s+(.*?)\s*)?-->//si) {
					# PRINT -- Insert print button

					my $button_title = $2;

					if ($in_break) {
						$tmplt .= qq[\n\n<table class="form">\n];
						$in_break = 0;
						$first_line = 1;
					}

					$tmplt .= qq[$end_fld</td>\n</tr>\n]
					  unless $first_line;
					$tmplt .= "\n<tr>\n  <td colspan=2>";
					$first_line = 0;
					if ($form_type ne 'query') {
						$tmplt .= _print_button(
								$form_name,
								$button_title,
								$table,
								$t);
					}
					redo;
				}
				if ($v =~ s/^<!--%\s*Detail(\s+(.*?))?-->//i) {
					# DETAIL -- Insert table detail

					my $button_attr = $2;

					my $detail = $t->detail_field;
					redo unless $detail;

					if ($in_break) {
						$tmplt .= qq[\n\n<table class="form">\n];
						$in_break = 0;
						$first_line = 1;
					}

					$tmplt .= qq[$end_fld</td>\n</tr>\n]
					  unless $first_line;
					$field_name = $Context->db_to_form($key);
					if ($form_type ne 'query') {
						$tmplt .= _table_detail(
								$detail,
								$form_name,
								$field_name,
								$table,
								$key,
								$t,
								$button_attr,
								$request);
						$tmplt .= "</table>\n\n";

						$in_break = 1;
						$detail_placed = 1;
					}
					redo;
				}
				$tmplt .= "$v";		# append tag
			}
		}

		next if $hiding;
		next if ! $form_field{$elem};
		next if ! ($elemtype eq 'field' || $elemtype eq 'display');
		next if $elem eq $id;	# already defined above

		$field = $elem;

		if ($elemtype eq 'display') {
			$f = $t->display->{$field};

			# Display fields are by default readonly unless
			# explicitly set in schema.

			$f->set(readonly => 'yes')
			  if ! $f->defined('readonly') && ! $f->noreadonly;
		}
		else {
			$f = undef;
			$f = $t->{$field} if $elem;
		}

		next if ! $f || $f->private;

		$hidden = $f->hidden;

		#
		# Reset form formating if needed.
		#

		if ($in_break && ! $hidden) {
			$tmplt .= qq[\n\n<table class="form">\n];
			$in_break = 0;
			$first_line = 1;
		}

		#
		# Terminate previous line.
		#

		$tmplt .= qq[$end_fld</td>\n</tr>\n]
		  unless $first_line || $hidden;

		$first_line = $in_newline = $in_insert = 0;

		#
		# Get input field properties.
		#

		$description = escape_html($f->description);
		$maxlength   = $f->length;
		$size        = $maxlength > $Field_Size ? $Field_Size : $maxlength;
		$decimal     = $f->decimal;
		$class       = $f->class;
		$radio       = $f->radio;
		$checkbox    = $f->checkbox;
		$type        = $radio		? 'radio'
		  	     : $checkbox	? 'checkbox'
			     : $class eq 'file' ? 'file'
			     : $hidden		? 'hidden'
			     :			  'text';
		$readonly    = ($form_type ne 'query' && $f->readonly) ||
			       ($form_type eq 'query' && $f->is_display_field)
			     ? 1 : 0;
		$required    = $f->required || $field eq $key;
		$translate   = $Context->translate($f->translate);
		$validate    = $Context->translate($f->valid);
		$valid_error = $f->valid_error;
		$on_focus    = $Context->translate($f->onfocus);
		$on_change   = $Context->translate($f->onchange);
		$choose      = $f->choose;
		$combo       = $f->combo;
		$search      = $f->search;
		$same        = $f->same;
		$browse      = $f->browse;
		$detail      = $f->detail_field;


		#
		# Construct form field name from database table field name.
		#

		if ($elemtype eq 'display') {
			$field_name = $Context->db_to_form($field);
			$field_label = $f->label || '';
			$field_label = escape_html($f->title)
			  if $field_label eq 'yes';
		}
		else {
			# Make a fully qualified field name (with table name).
			$field_name = $Context->db_to_form("${table}.${field}");

			if ($field_label = $f->label) {
				$field_label = escape_html($f->title)
				  if $field_label eq 'yes';
			}
			else {
				if (! defined($field_label)) {
					$field_label = escape_html($f->title);
				}
			}
		}
		$field_label_str = ($field_label =~ / /)
		  ? "<nobr>$field_label</nobr>"
		  : $field_label;

		if ($field_label_str) {
			# Enclose field label with an identifiable tag.
			$field_label_style = $f->label_style;
			$field_label_style &&= qq[ style="$field_label_style"];
			$field_label_str   =
  qq[<span id="$Label_Prefix$field_name"$field_label_style>$field_label_str</span>];
		}

		#
		# All fields and their attendant buttons are enclosed within
		# a <span> tag so that the entire field can be hidden/shown.
		# The <span> container has its own properties for CSS class
		# or style.
		#

		$container_class = $f->container_class || '';
		$container_class = qq["$container_class"]
		  if $container_class && $container_class !~ /^(['"]).*\1$/;
		$container_class &&= " class=$container_class";

		$container_style = $f->container_style || '';
		$container_style = qq["$container_style"]
		  if $container_style && $container_style !~ /^(['"]).*\1$/;
		$container_style &&= " style=$container_style";

		$bgn_fld = qq[<span id="$Field_Prefix$field_name"$container_class$container_style>];

		#
		# Compute tabbing index.
		#

		if ($readonly) {
			$tabindex_str = qq[ tabindex="$No_Tabindex"];
		}
		else {
			$tabindex += $Tabindex_Interval;
			$tabindex_str = qq[ tabindex="$tabindex"];

			# Bump tabindex if field is an input array to
			# allow for extra <input> fields generated at runtime.
			$tabindex += $Max_Input_Array if $f->input;
		}

		#
		# Set default value.
		#

		$default = '';
		if ($form_type ne 'query') {
			$default = $f->default_static;
			# Remove possible quotes.
			$default =~ s/^\s*(["'])(.*)\1\s*$/$2/;
		}

		#
		# Set up options/extras for field input.
		#

		# The title parameter produces a pop-up "tool tip" in IE5.
		$title_str = $description ? qq[ title="$description"] : '';

		#
		# Set readonly fields to display and operate differently.
		#

		if ($readonly) {
			$readonly_str = qq[ readonly];
			$class_str    = qq[ class="${class}-readonly"];
		}
		else {
			$readonly_str = qq[];
			$class_str    = qq[ class="$class"];
		}

		#
		# Get individual field css style settings.
		#

		$style_str = '';
		if (! $hidden) {
			$style_str = $f->style || '';
			$style_str &&= qq[ style="$style_str"];
		}

		#
		# Get maximum size of input field.
		#

		$maxlength_str = '';
		if ($form_type ne 'query' && ! ($readonly || $hidden)) {
			$maxlength_str = qq[ maxlength="$maxlength"];
		}

		#
		# Accumulate field initialization scripts.
		# They are appended to table initialization scripts.
		#

		if ($init_str = $f->initialize) {
			# Translate form variables (%name/$name) into actual
			# form names/values.
			$init_scripts .= "\n" . $Context->translate(
							$init_str,
							form=>'f',
							);
		}

		#
		# Accumulate field compute scripts.
		# They are appended to table compute scripts.
		#

		if ($compute_str = $f->compute) {
			# Translate form variables (%name/$name) into actual
			# form names/values.
			$compute_scripts .= "\n\n" . $Context->translate(
							$compute_str,
							form=>'f',
							);
		}

		#
		# Compose onChange clause.
		#

		$on_change_str = '';
		if ($form_type ne 'query' && ! ($readonly || $hidden)) {
		
			#
			# Set up form field validation routines.
			#

			$translate_str = '';
			if ($translate) {
				$translate_str = $translate;
				$translate_str = escape_html($translate_str);
				$translate_str = qq[translate_input(this,'$translate_str')];
			}

			$check_str = '';
			if (! $choose) {
				$check_str = qq[check_input(this,'$class',$maxlength,$decimal)];
			}

			$validate_str = '';
			if ($validate) {
				$validate_str = $validate;
				$validate_str = escape_html($validate_str);
				$validate_str = "validate_input(this,($validate_str)";
				$valid_error = escape_html($valid_error);
				$validate_str .= ",'$valid_error'" if $valid_error;
				$validate_str .= ")";
			}

			$compute_str = '';
			if ($on_change) {
				$compute_str = $on_change;
				$compute_str =~ s/\s+$//;	# trim
				$compute_str .= ';'		#{terminate
				  unless $compute_str =~ /[};]$/;
				$compute_str = escape_html($compute_str);
			}

			#
			# Assemble onChange attribute string.
			#

			if ($type eq 'text' || $type eq 'file') {

				# Assemble onChange clause differently if
				# compute_str is defined to allow it to have
				# it's own 'if' statements in AcmeBase files
				#
				# No compute_str:
				#   if($translate_str
				#   && $check_str
				#   && $validate_str
				#   && form_compute()) return true;
				#   this.focus();return false;	// error
				#
				# With compute_str:
				#   if($translate_str
				#   && $check_str
				#   && $validate_str){
				#     $compute_str;
				#     if($form_compute) return true;
				#   }
				#   this.focus();return false;	// error
				#
				# WARNING: it is important that the pattern
				# 'form_compute.*{return true}' not change
				# as it is used in the Form Filler to modify
				# the field's onChange clause.
				# See Choose_assign.

				if ($compute_str) {
					$on_change_str =
					  join " && ", grep { $_ }
					  $translate_str, $check_str, $validate_str;

					$compute_str .=
					  qq[if(form_compute(this.form)){return true}];
					if ($on_change_str) {
						$on_change_str =
			  qq[if($on_change_str){$compute_str}this.focus();return false];
					}
					else {
						$on_change_str = $compute_str;
					}
				}
				else {
					$on_change_str =
					  join " && ", grep { $_ }
					  $translate_str, $check_str, $validate_str,
					  qq[form_compute(this.form)];

					$on_change_str =
			  qq[if($on_change_str){return true} this.focus();return false];
				}

				$on_change_str = qq[ onChange="$on_change_str"];
			}
			elsif ($type eq 'radio' || $type eq 'checkbox') {
				$on_change_str = $compute_str if $compute_str;
				
				$on_change_str .= qq[if(form_compute(this.form)){return true} return false;];

				$on_change_str = qq[ onClick="$on_change_str"];
			}
		}
		elsif ($form_type eq 'query' && ! $hidden && $f->is_date()) {
			$on_change_str = qq[ onChange="check_input(this,'date_range')"];
		}

		#
		# Compose onFocus clause.
		#

		$on_focus_str = '';
		if ($form_type ne 'query' && ! ($readonly || $hidden)) {

			# Form action default is to select contents of the
			# form field upon focus.

			$on_focus_str = $on_focus if $on_focus;

			# To make a single click in field select entire
			# contents, uncomment the next lines.
			#$on_focus_str =~ s/;\s*$//;
			#$on_focus_str .= ';' if $on_focus_str;
                        #$on_focus_str .= qq[this.select()];

			if ($search && $form_type ne 'query') {
				# Save current field value before searching
				# so empty searches can restore old value.
				$on_focus_str = "_search_saved_$field_name = this.value;"
				  . $on_focus_str;
			}

			$on_focus_str = escape_html($on_focus_str);
			$on_focus_str &&= qq[ onFocus="$on_focus_str"];
		}

		#
		# Define Combo field and button.
		#

		$combo_button = $combo_prefix = $combo_input = '';
		if ( ($form_type ne 'query' && $combo && ! ($readonly || $hidden))
		or   ($form_type eq 'query' && ($choose || $f->array)) ) {

			# Note: Netscape requires that <select> boxes to be hidden
			# be wrapped in <div> tags, not <span> tags.

			$combo_prefix =
			    qq[<!-- Begin Combo button $field_name -->]
			  . qq[<div id="_combo_$field_name"]
			  . qq[ style="display:inline">];

			$combo_input =
			    qq[</div>]
			  . qq[<span id="_combo_C_$field_name"]
			  . qq[ style="display:none">]
			  . qq[<input type="text"]
			  . qq[ name="C_$field_name"]
			  . qq[ value="$default"]
			  . qq[ size="$size"]
			  . qq[$maxlength_str$class_str$style_str]
			  . qq[$on_focus_str$on_change_str$title_str]
			  . qq[$readonly tabindex="$No_Tabindex">]
			  . ($form_type eq 'query'
			    ? _query_button($form_name,"C_$field_name",$f)
			    : '')
			  . qq[</span>];

			$combo_button = _combo_button(
						$form_name,
						$field_name,
						$table,
						$tabindex);

		}

		#
		# Define Parameter button.
		#

		# This applies only to query forms.

		$query_button = '';
		if ($form_type eq 'query' && ! $choose && ! $f->is_display_field) {

			$query_button = _query_button(
						$form_name,
						$field_name,
						$f);
		}

		#
		# Define Search button.
		#

		$search_button = '';
		if ($search && $form_type ne 'query' && ! ($readonly || $hidden)) {

			$search_button = _search_button(
						$search,
						$form_name,
						$field_name,
						$table,
						$field,
						$t,
						$f,
						\$on_change_str);
		}

		#
		# Define Same button.
		#

		$same_button = '';
		if ($same && $form_type ne 'query' && ! $hidden) {

			$same_button = _same_button(
						$same,
						$form_name,
						$field_name,
						$table,
						$field,
						$t,
						$f);
		}

		#
		# Define calendar button.
		#

		$calendar_button = '';
		if ($class eq 'date' && ! $f->nocalendar
		&& (($form_type ne 'query' && ! ($readonly || $hidden))
		|| ($form_type eq 'query' && ! $hidden))) {
			$calendar_button = _calendar_button(
						$form_name,
						$field_name);
		}

		#
		# Define color button.
		#

		$color_button = '';
		if ($class eq 'color' && ! $f->nocolorbutton
		&& $form_type ne 'query' && ! ($readonly || $hidden)) {
			$color_button = _color_button(
						$form_name,
						$field_name,
						\$on_change_str);
		}

		#
		# Define Browse button.
		#

		$browse_button = '';
		if ($browse && $form_type ne 'query' && ! ($readonly || $hidden)) {

			$browse_button = _browse_button(
						$browse,
						$form_name,
						$field_name,
						$table,
						$field,
						$t,
						$f);
		}

		#
		# Define Detail button.
		#

		$detail_button = '';
		if ($detail && $form_type ne 'query') {

			$detail_button = _detail_button(
						$detail,
						$form_name,
						$field_name,
						$table,
						$field,
						$t,
						$f,
						undef,
						$request);
		}

		#
		# Define Required image/flag.
		#

		# Define an image with a name corresponding to the form field
		# name to be used in JavaScript to find which fields are
		# required.

		$required_symbol = '';
		if ($required && $form_type ne 'query') {
			my $required_img_id = "require_$field_name";
			$required_symbol =
  qq[<img src="$Images/required.gif" alt="$field_label" id="$required_img_id" class="required" title="This field entry is required.">];
		}
		else {
			if ($search_button
			|| $same_button
			|| $calendar_button
			|| $color_button
			|| $browse_button
			|| $detail_button) {
				# Set to a transparent spacer image.
				$required_symbol = 
  qq[<img src="$Images/transparent.gif" border="0" width="7" height="13">];
			}
	  	}

		#
		# Define field for previous value of key.
		#

		# This allows the non-ID key fields to be changed (if the
		# database allows it).

		$prev_value = '';
		if ($field eq $key) {
			$prev_value =
		  qq[<input type="hidden" name="P_${table}__${key}" value="">];
		}

		#
		# Set word wrap codes if buttons are defined to be on same line.
		#

		$bgn_nobr = $end_nobr = '';
		if ($search_button
		|| $same_button
		|| $calendar_button
		|| $color_button
		|| $browse_button
		|| $detail_button
		|| $combo_button
		|| $query_button
		) {
			$bgn_nobr = '<nobr>';
			$end_nobr = '</nobr>';
		}

		#
		# Compose input fields.
		#

		if ($type eq 'text' || $type eq 'file') {

			# Note: display fields are set to readonly above if
			# they do not accept input. Display fields that do
			# accept input are rendered just like a normal
			# input fields.

			if ($readonly) {
				$display_field = 'display-field';
				$display_label = 'display-label';
			}
			else {
				$display_field = 'field';
				$display_label = 'label';
			}

			my $rows = $f->rows || 0;

			# Note: input arrays are only checked for on <input> tags
			# in the form filler.

			if ($maxlength < 100 || $rows == 1 || $type eq 'file'
			|| ($f->array && $f->input)) {	# input arrays made only for <input>
				# Standard input field.
				$tmplt .= qq[
<tr>
  <td class="$display_label">$field_label_str</td>
  <td class="$display_field">$bgn_fld$bgn_nobr$combo_prefix<input type="$type" name="$field_name" value="$default" size="$size" $maxlength_str$class_str$style_str$on_focus_str$on_change_str$readonly_str$tabindex_str$title_str>$combo_input$required_symbol$combo_button$search_button$same_button$calendar_button$color_button$browse_button$detail_button$query_button$prev_value$end_nobr];
			}
			else {
				if (! $rows) {
					# Long text fields use <textarea> fields.
					# Try to size <textarea> intelligently.
					$rows = int(($maxlength+$Field_Size)/$Field_Size);
					if ($rows > 9) { $rows = 6 }
					elsif ($rows > 7) { $rows = 5 }
					elsif ($rows > 5) { $rows = 4 }
					elsif ($rows > 3) { $rows = 3 }
				}

				$tmplt .= qq[
<tr>
  <td class="$display_label">$field_label_str</td>
  <td class="$display_field">$bgn_fld$bgn_nobr$combo_prefix<textarea name="$field_name" cols="$Field_Size" rows="$rows"$class_str$style_str$on_focus_str$on_change_str$readonly_str$tabindex_str$title_str>$default</textarea>$combo_input$required_symbol$combo_button$search_button$same_button$browse_button$detail_button$query_button$prev_value$end_nobr];
			}
		}
		elsif ($type eq 'hidden') {
			if ($size < 255) {
				# Standard hidden input field.
				$tmplt .= qq[<input type="$type" name="$field_name" value="$default" $tabindex_str>$detail_button$prev_value];
			}
			else {
				# Hide long fields in <textarea> controls.
				$tmplt .= qq[
<textarea name="$field_name" wrap="virtual" tabindex="$No_Tabindex" style="display:none">$default</textarea>$detail_button];
			}
		}
		elsif ($type eq 'radio') {
			$tmplt .= <<End_Of_HTML;
<tr>
  <td class="label">$field_label_str</td>
  <td class="field">$bgn_fld<table class="radio-group">
    <tr>
End_Of_HTML
			my ($rval,$rlab);
			# Use array in lieu of hash to maintain definition order.
			my @radios = Text::ParseWords::parse_line('\s+',0,$radio);
			my $i = 0;
			while ($i < @radios) {
				$rval = $radios[$i++];
				$rlab = $radios[$i++];


				$tmplt .= <<End_Of_HTML;
      <td class="radio-field"><input type="$type" name="$field_name" value="$rval"$class_str$style_str$on_change_str$tabindex_str$title_str></td>
      <td class="radio-label">$rlab</td>
End_Of_HTML
			}
			$tmplt .= <<End_Of_HTML;
    </tr>
    </table>
    $required_symbol$same_button$detail_button$prev_value
End_Of_HTML
		}
		elsif ($type eq 'checkbox') {		
			$tmplt .= qq[
<tr>
  <td class="label">$field_label_str</td>
  <td class="field">$bgn_fld$bgn_nobr<input type="$type" name="$field_name" value="$checkbox"$class_str$style_str$on_change_str$tabindex_str$title_str>$required_symbol$same_button$detail_button$end_nobr];
		}
	}

	#
	# Assure form table tag is left ready to add a new row.
	#

	for my $blocktype (@blocktypes) {
		# Check if ending a hiding directive.
		if ($blocktype eq 'h') {
			pop @hiding;
			$hiding = &$is_hiding;
			next;
		}

		if (! $hiding) {
			$tmplt .= qq[$end_fld</td>\n</tr>\n]
			  unless $first_line || $in_break;
			$tmplt .= qq[</table>\n]
			  unless $in_break;
			$tmplt .= qq[</div>\n];

			# Terminate insert blocks.
			$tmplt .= qq[</td>\n</tr>\n</table>\n]
			  if $blocktype eq 'i';

			$in_break = 1;
			$in_insert = $in_newline = 0;
		}
	}

	if ($in_break) {
		$tmplt .= qq[\n\n<table class="form">\n];
		$in_break = 0;
		$first_line = 1;
	}

	$tmplt .= "$end_fld</td>\n</tr>\n" unless $first_line;

	#
	# Add compute scripts global.
	#

	if ($compute_str = $t->compute) {
		# Translate form variables (%name/$name) into actual
		# form names/values.
		$compute_scripts .= "\n\n" . $Context->translate(
						$compute_str,
						form=>'f',
						);
	}

	return $tmplt;
}

#------------------------------------------------------------------------------

sub _make_table_audit {
	my ($form_name,$form_type,$t) = @_;

	my $tmplt = '';		# returned template fragment

	#
	# Append audit fields.
	#

	my $table = $t->name;
	my $title = $t->row_title;

	my $readonly_str = $form_type ne 'query' ? ' readonly="yes"' : '';

	my $tabindex_str = '';
	if ($form_type eq 'query') {
		$tabindex += $Tabindex_Interval;
		$tabindex_str = qq[ tabindex="$tabindex"];
	}

	my ($query_button1,$query_button2,$query_button3) = ('') x 3;

	if ($t->audit) {
		my $audit_entry  = $t->name . "__" . $t->audit_entry;
		my $audit_update = $t->name . "__" . $t->audit_update;
		my $audit_user   = $t->name . "__" . $t->audit_user;

		if ($form_type eq 'query') {
			$query_button1 = _query_button($form_name, $audit_entry,  'date');
			$query_button2 = _query_button($form_name, $audit_update, 'time');
			$query_button3 = _query_button($form_name, $audit_user,   'text');
		}

		$tmplt .= <<End_Of_HTML;
  <tr>
    <td class="audit-label">$title Entered</td>
    <td class="audit-field"><input type="text" name="$audit_entry" value="" size="10" class="audit"$readonly_str title="This record was entered on this date."$tabindex_str>$query_button1</td>
    <td class="audit-label">Changed</td>
    <td class="audit-field"><input type="text" name="$audit_update" value="" size="10" class="audit"$readonly_str title="This record was changed on this date."$tabindex_str>$query_button2</td>
    <td class="audit-label">by</td>
    <td class="audit-field"><input type="text" name="$audit_user" value="" size="10" class="audit"$readonly_str title="This record was last changed by this user."$tabindex_str>$query_button3</td>
  </tr>
End_Of_HTML
	}

	return $tmplt;
}

#------------------------------------------------------------------------------
#
# Autoload routines
#
#------------------------------------------------------------------------------

our %Subs = (

#-------------------------------------------------------------------------------
'_combo_button' => <<'END_OF_FUNC',
sub _combo_button {
	my ($form_name,$field_name,$table,$tabindex) = @_;

	#
	# Compute tab order.
	#

	#$tabindex += $Tabindex_Interval;
	#my $tabindex_str = qq[ tabindex="$tabindex"];
	my $tabindex_str = qq[ tabindex="$No_Tabindex"];

	$combo_script ||= <<End_Of_JavaScript;
<script language="JavaScript" type="text/javascript">

/*
 * Combo button support
 */

combo_swap.swapped = {};

function combo_swap(fld) {
	// Swap input field with dropdown list and vice versa.
	if (! fld) return;

	var f = fld.form;
	var field_name = fld.name;
	field_name = field_name.replace(/^C_/,'');

	var field  = f.elements[field_name];
	var cfield = f.elements['C_'+field_name];
	var span  = document.getElementById('_combo_'+field_name);
	var cspan = document.getElementById('_combo_C_'+field_name);

	// Swap field names and tabIndex
	var tmpName  = field.name;
	var tmpIndex = field.tabIndex;

	field.name      = cfield.name;
	field.tabIndex  = cfield.tabIndex;

	cfield.name     = tmpName;
	cfield.tabIndex = tmpIndex;

	// Swap display
	if (! combo_swap.swapped[field_name]) {
		// Display input field.
		span.style.display  = 'none';
		cspan.style.display = 'inline';

		cfield.select();
		cfield.focus();

		combo_swap.swapped[field_name] = true;
	}
	else {
		// Display dropdown list.
		span.style.display  = 'inline';
		cspan.style.display = 'none';

		field.focus();

		combo_swap.swapped[field_name] = false;
	}
}
</script>
End_Of_JavaScript

	# For cross-browser implementation, where <a> tag's style needs to be
	# display:block in order to specify the height, use a transparent image for
	# forground so buttons don't get rendered when printed.
	my $img = $ENV{HTTP_USER_AGENT} =~ /MSIE/
		? ''
		: qq[<img src="$Images/transparent.gif" width="22" height="20" border="0">];

	return qq[<a
 href="javascript:combo_swap(self.document.$form_name.$field_name)"
 title="Rotate list and entry field.\nUse to enter an item that is not in the list." 
 class="combo"
>$img</a>];
}
END_OF_FUNC

#-------------------------------------------------------------------------------
'_search_button' => <<'END_OF_FUNC',
use Acme::Form::Assign;
sub _search_button {
	my ($search, $form_name, $field_name, $table, $field, $t, $f, $on_change_ref) = @_;

	# Define window name from field name.
	my $window_name = "$field_name"; 

	my $search_table = $f->search_table  || '';
	my $title        = $f->search_title  || '';
	my $sql          = $f->search_sql    || '';
	my $rows         = $f->search_rows   || '';
	my $format       = $f->search_format || '';
	$format =~ s/^\s*(["'])(.*)\1\s*$/$2/;	# remove quotes

	# Get button title.
	my $search_button_title = $f->search_button_title || 'Search';

	# Compute tab order.
	#$tabindex += $Tabindex_Interval;
	#my $tabindex_str = qq[ tabindex="$tabindex"];
	my $tabindex_str = qq[ tabindex="$No_Tabindex"];

	# Set flag if sql uses fully qualified field names.
	my $sql_uses_fq_names = $sql =~ /\b\w+\.\w+\b/;

	# Get field to go to after a successful search.
	my $focus_next_field = $f->search_next || '';
	if ($focus_next_field) {
		$focus_next_field = ($focus_next_field =~ /^[%\$]/)
		  ? $Context->translate($focus_next_field, form=>'f')
		  : 'f.'.$Context->db_to_form($focus_next_field);
		$focus_next_field .=
qq[.focus();\n\t\tif ($focus_next_field.select) $focus_next_field.select();\n];
	}
	else {
		$focus_next_field = 
qq[f.${field_name}.select();\n\t\tf.${field_name}.focus();\n];
	}

	# Search parameters identify the fields to search if
	# the field contents are numeric (ID) or alpha (name).

	my $search_parameters = '';
	$search_parameters .= "&id_field=".$f->search_id
	  if $f->search_id;
	$search_parameters .= "&name_field=".$f->search_name
	  if $f->search_name;
	$search_parameters .= "&first_name_field=".$f->search_first_name
	  if $f->search_first_name;
	$search_parameters .= "&last_name_field=".$f->search_last_name
	  if $f->search_last_name;
	$search_parameters .= "&autoselect=no"
	  unless $f->search_autoselect;

	# Redefine form field's on_change_str.
	if ($f->search_autosearch) {
		$$on_change_ref =
qq[ onChange="if (this.value.length) return _search_${window_name}('autosearch=yes')"];
	}
	else {
		$$on_change_ref = "";
	}

	# Parse $sql and add needed columns to query.
	my $sqlo = new Acme::SQL $sql;

	#
	# Define assignments.
	#

	my $asgn = new Acme::Form::Assign
		form	     => 'f',	# form name to use for target fields
		result	     => 'r',	# object name for result (form) fields
		search_table => $search_table,# table name to search
					# to produce the result
		t	     => $t,	# AcmeBase schema object for the table
					# whose form fields are being filled.
					# Not the table being searched.
		context	     => $Context;# Acme::Form::Context object which 
					# controls translation

	# Define initial assignment statement for assigning
	# a value to the search field.
	# Assign key value of search table to current form field
	# if the search is of a database table.

	my ($assign_tab, $assign_fld);
	if ($assign_tab = $f->search_table) {
		# Use table's key field as value to assign to the search field.
		my $st = $Schema->t($assign_tab);
		$assign_fld = $st->key;
	}
	else {
		# Use first column of the query as value to assign
		# to the search field.
		my ($col) = $sqlo->columns;
		($assign_tab, $assign_fld) = $col =~ /(\w+?)\.(\w+)/;
		if (! $assign_tab
		&& $search =~ /^[sS][eE][lL][eE][cC][tT]\b/) {
			# Return first table extracted from
			# SQL's first 'from' clause.
			($assign_tab) =
			  $search =~ /.+?\bfrom\s+(\w+)/;
			$assign_fld = $col;
		}
	}
	my $keyfld = "$table.$field";
	$keyfld = "$field" if $f->is_display_field;
	my $search_assign_key = "\$$keyfld = \$S.$assign_tab.$assign_fld\n";

	my $search_assign = $f->search_assign;
	$search_assign  &&= "$search_assign;";
	$search_assign   .= $search_assign_key;

	my $assignments = '';
	$assignments .= $asgn->assign_options( $f->search_assign_options );
	$assignments .= $asgn->assign_fields(  $search_assign            );
	$assignments .= $asgn->assign_input(   $f->search_assign_input   );

	$assignments =~ s/^/\t\t/mg;	# pretty-print

	# Note: make sure the fields in the search_assign strings below are
	# in search_sql. Often the search_sql is derived from the target
	# table's select property, which normally only selects the fields
	# which are displayed in the list box.

	# Make hash of columns to eliminate duplicates.
	my %columns = map { $_, 1 } $asgn->columns;

	for (keys %columns) {
		$sqlo->add_column($_) if ! $sqlo->is_column($_);
	}

	# Redefine $sql with the extra columns needed for assignments.
	$sql = $sqlo->as_sql;

	# Assure $sql does not have characters which breaks JavaScript below.
	$sql =~ s/[\t\n]+/ /g;	# remove newlines and tabs
	$sql =~ s/'/\\'/g;	# escape ' quotes
	$title =~ s/\n+/ /g;	# remove newlines
	$title =~ s/'/\\'/g;	# escape ' quotes
	$format =~ s/\n+/ /g;	# remove newlines
	$format =~ s/'/\\'/g;	# escape ' quotes

	# Define Search button support scripts.
	my $scripts = <<End_Of_Script;

/*
 * Search button support for $field_name
 */

var _search_win_${window_name};
var _search_saved_${field_name} = '';
var _search_active_${field_name} = false;	// flag if already searching

function _search_${window_name}(search_options) {

	if (_search_active_${field_name}) {
		return false;
	}
	else {
		// To avoid Search button's onClick event from simultaneously
		// triggering the field's onChange event, set flag and a timer
		// to clear the flag after 1/2 second.

		_search_active_${field_name} = true;
		setTimeout('_search_active_${field_name} = false',500);
	}

	// Open search window.
	var url = '/cgi-bin/search.cgi' +
	  '?search=' + escape(self.document.forms['$form_name'].$field_name.value) + 
	  '&title=' + escape('Search $title') +
	  '&table=$search_table' +
	  '&sql=' + escape('$sql') +
	  '&format=' + escape('$format') +
	  '&rows=$rows' +
	  '$search_parameters';

	if (search_options) url = url + '&' + search_options;

	_search_win_${window_name} = new_message(
					'Searching...',
					'Search $title',
					url,
					'${window_name}');
	return true;
}
function _search_return_${window_name}() {
	// Return from search window opened in _search_${window_name}().
	var f = self.document.forms['$form_name'] || self.document.forms[$Default_Form];
	if (_search_win_${window_name} && _search_win_${window_name}.result) {
		var r = _search_win_${window_name}.result;
		// Assign selection values to fields.
$assignments
		// Recompute form.
		if (self.form_compute) form_compute(f);

		// Go to next input field.
		$focus_next_field
	} else {
		// No records were found.
		f.${field_name}.value = _search_saved_${field_name};
		if (f.${field_name}.select) f.${field_name}.select();
		f.${field_name}.focus();
	}
}
function _search_error_return(window_name) {
	eval('_search_return_'+window_name+'()');
}
End_Of_Script

	#
	# Construct button for searching other tables.
	#

	return <<End_Of_HTML;
<!-- Begin Search button $field_name -->
<input type="button" value="$search_button_title" onClick="_search_${window_name}()" class="search-button"$tabindex_str>
<script language="JavaScript" type="text/javascript">
$scripts
</script>
<!-- End Search button $field_name -->
End_Of_HTML
}
END_OF_FUNC

#-------------------------------------------------------------------------------
'_same_button' => <<'END_OF_FUNC',
use Acme::Form::Assign;
sub _same_button {
	my ($same, $form_name, $field_name, $table, $field, $t, $f) = @_;

	# Get field to go to after a successful execution.
	my $focus_next_field = $f->same_next || '';
	if ($focus_next_field) {
		$focus_next_field = ($focus_next_field =~ /^[%\$]/)
		  ? $Context->translate($focus_next_field, form=>'f')
		  : 'f.'.$Context->db_to_form($focus_next_field);
		$focus_next_field .=
qq[.focus();\n\tif ($focus_next_field.select) $focus_next_field.select();\n];
	}

	# Get button title.
	my $same_button_title = $f->same_button_title || 'Same';

	#
	# Compute tab order.
	#

	#$tabindex += $Tabindex_Interval;
	#my $tabindex_str = qq[ tabindex="$tabindex"];
	my $tabindex_str = qq[ tabindex="$No_Tabindex"];

	#
	# Define assignments.
	#

	my $asgn = new Acme::Form::Assign
		form	     => 'f',
		result	     => 'f',	# both form and result on same form
		t	     => $t,
		context	     => $Context;

	my $assignments = $asgn->assign_fields($same);

	$assignments =~ s/^/\t/mg;	# pretty-print

	#
	# Define JavaScript assignments.
	#

	# Keep following scripts with button
	my $scripts = <<End_Of_Script;

/*
 * Same button support for $field_name
 */

function _same_${field_name}() {
	var f = self.document.forms['$form_name'] || self.document.forms[$Default_Form];
$assignments
	// Recompute form.
	if (self.form_compute) form_compute(f);

	// Go to next input field.
	$focus_next_field
}
End_Of_Script

	#
	# Assemble HTML and JavaScript.
	#

	return <<End_Of_HTML;
<!-- Begin Same button $field_name -->
<input type="button" value="$same_button_title" onClick="_same_${field_name}()" class="same-button"$tabindex_str>
<script language="JavaScript" type="text/javascript">
$scripts
</script>
<!-- End Same button $field_name -->
End_Of_HTML
}
END_OF_FUNC

#-------------------------------------------------------------------------------
'_calendar_button' => <<'END_OF_FUNC',
sub _calendar_button {
	my ($form_name, $field_name) = @_;

	#
	# Define scripts that get inserted into the form header.
	#

	$calendar_script ||= qq[<script src="$Scripts_Lib/calendar.js" type="text/javascript"></script>\n];

	#
	# Compute tab order.
	#

	#$tabindex += $Tabindex_Interval;
	#my $tabindex_str = qq[\n tabindex="$tabindex"];
	my $tabindex_str = qq[\n tabindex="$No_Tabindex"];

	#
	# Return calendar button image link.
	#

	# For cross-browser implementation, where <a> tag's style needs to be
	# display:block in order to specify the height, so instead use a
	# transparent image for forground so buttons don't get rendered when
	# printed.

	my $img = $ENV{HTTP_USER_AGENT} =~ /MSIE/
		? ''
		: qq[<img src="$Images/transparent.gif" width="28" height="24" border="0">];

	return qq[<a
 href="javascript:displayCalendar(1,'$field_name',self.document.$form_name.$field_name.value)"
 title="Display a calendar."$tabindex_str
 class=calendar
>$img</a>];
}
END_OF_FUNC

#-------------------------------------------------------------------------------
'_color_button' => <<'END_OF_FUNC',
sub _color_button {
	my ($form_name, $field_name, $on_change_ref) = @_;

	#
	# Define scripts that get inserted into the form header.
	#

	$color_script ||= qq[<script src="$Scripts_Lib/colorwheel.js" type="text/javascript"></script>];

	#
	# Define initialization scripts.
	#

	$form_init_scripts .= <<End_Of_JavaScript;

if (f.${field_name}.value.match(/^#?([0-9aAbBcCdDeEfF]{6})/)) {
	document.getElementById('button_${field_name}').style.backgroundColor = '#' + RegExp.\$1;
}
else {
	document.getElementById('button_${field_name}').style.backgroundColor = '#CCCCCC';
}
End_Of_JavaScript

	#
	# Redefine form field's on_change_str. The previous value is preserved.
	#

	my ($on_change) = $$on_change_ref =~ / onChange="(.*)"/;
	$$on_change_ref =
qq[ onChange="if(check_input(this,'color')){document.getElementById('button_${field_name}').style.backgroundColor = this.value;$on_change;return true}else{return false}"];

	#
	# Compute tab order.
	#

	#$tabindex += $Tabindex_Interval;
	#my $tabindex_str = qq[\n tabindex="$tabindex"];
	my $tabindex_str = qq[\n tabindex="$No_Tabindex"];

	#
	# Return color button.
	#

	return qq[<input type="button" 
 id="button_${field_name}"
 class="color-button"
 title="Select a color."
 onClick="colorwheel(this.form.${field_name},this)"$tabindex_str
>];
}
END_OF_FUNC

#-------------------------------------------------------------------------------
'_browse_button' => <<'END_OF_FUNC',
sub _browse_button {
	my ($browse, $form_name, $field_name, $table, $field, $t, $f) = @_;

	# Define window name from field name.
	my $window_name = "$field_name"; 

	# Get button title.
	my $browse_button_title = $f->browse_button_title if $f;
	$browse_button_title ||= 'Browse';

	# Compute tab order.
	#$tabindex += $Tabindex_Interval;
	#my $tabindex_str = qq[ tabindex="$tabindex"];
	my $tabindex_str = qq[ tabindex="$No_Tabindex"];

	# Get field to go to after a successful search.
	my $focus_next_field = $f->browse_next if $f;
	if ($focus_next_field) {
		$focus_next_field = ($focus_next_field =~ /^[%\$]/)
		  ? $Context->translate($focus_next_field, form=>'f')
		  : 'f.'.$Context->db_to_form($focus_next_field);
		$focus_next_field .=
qq[.focus();\n\tif ($focus_next_field.select) $focus_next_field.select();\n];
	}
	else {
		$focus_next_field = 
qq[f.${field_name}.select();\n\t\tf.${field_name}.focus();\n];
	}


	# Define Browse button support scripts.
	my $scripts = <<End_Of_Script;

/*
 * Browse button support for $field_name
 */

var _browse_win_${window_name};

function _browse_${window_name}(options) {
	// Open browse window.
	var url = '/cgi-bin/browse.cgi?$browse';

	_browse_win_${window_name} = new_browse(
					url,
					'${window_name}',
					options);
	return true;
}
function _browse_return_${window_name}() {
	// Return from browse window opened in _browse_${window_name}().
	var f = self.document.forms['$form_name'] || self.document.forms[$Default_Form];
	if (_browse_win_${window_name} && _browse_win_${window_name}.result) {
		var r = _browse_win_${window_name}.result;

		// Assign selection values to fields.
		if (f.${field_name} && r.value) {
			f.${field_name}.value = r.value;
		}

		// Recompute form.
		if (self.form_compute) form_compute(f);

		// Go to next input field.
		$focus_next_field
	} else {
		// No records were found.
		f.${field_name}.focus();
	}
}
End_Of_Script

	#
	# Construct button for browsing other tables.
	#

	return <<End_Of_HTML;
<!-- Begin Browse button $field_name -->
<input type="button" value="$browse_button_title" onClick="_browse_${window_name}()" class="browse-button"$tabindex_str>
<script language="JavaScript" type="text/javascript">
$scripts
</script>
<!-- End Search button $field_name -->
End_Of_HTML
}
END_OF_FUNC

#-------------------------------------------------------------------------------
'_print_button' => <<'END_OF_FUNC',
sub _print_button {
	my ($form_name, $button_title, $table, $t) = @_;

	#
	# Define Print button.
	#

	# Note: this is like the detail button where a horizontal slice through
	# a form with the left side use for a "print title" and the right side
	# is the button with a "button title".

	my $print_title = '';			# not used yet

	$button_title ||= 'Print';

	my @attr;
	my $attributes = '';
	if ($button_title =~ /['"]/) {

		#
		# Quotes are used in print button title. 
		# Parse the line on spaces and use only the first item as the
		# title and use the rest as attributes to the <input> button tag.
		#

		($button_title,@attr) = Text::ParseWords::parse_line('\s+',0,$button_title);
	}
	$attributes = join ' ', @attr  if @attr;
	$attributes = " $attributes\n" if $attributes;

	#
	# Define class prefix for print-header, print-title, print-button.
	#

	my $class="print";

	#
	# Get key field.
	#
	
	my $key = $t->key;

	#
	# Get program to use to print records.
	#

	my $print_url = $t->print_prog;

	if (! $print_url) {
		# Use edit_prog with "&print=yes" parameter appended 
		$print_url = $t->edit_prog;
		$print_url = $print_url . ($print_url =~ /\?/ ? '&' : '?') . 'print=yes';
	}

	$print_url .= ($print_url =~ /\?/ ? '&' : '?') . "table=$table&$key=";

	#
	# Define message text for saving current record (if needed).
	#

	my $save_msg   = 'Saving ' . $t->row_name . '.';
	my $save_title = 'Saving ' . $t->title;

	#
	# Compute tab order.
	#

	$tabindex += $Tabindex_Interval;
	my $tabindex_str = qq[ tabindex="$tabindex"];
	#my $tabindex_str = qq[ tabindex="$No_Tabindex"];

	#
	# Keep following scripts with button.
	#

	$print_script ||= <<End_Of_Script;
<script language="JavaScript" type="text/javascript">

/*
 * Print button
 */

function form_print_${form_name}(options) {
	var f = self.document.forms['$form_name'] || self.document.forms[$Default_Form];

	if (f.${table}__${key}.value == '') {
		if (! form_is_dirty(f)) {
			var q = 'This form must be saved first before printing.\\n' +
			  ' Do you want to save and continue?';
			if (!confirm(q)) return;
		}

		form_save(f,'$save_msg','$save_title',form_print_post_${form_name}) && f.submit();
		return;		// after saving, this routine gets run again
	}

	// Open print window.
	open_print('$print_url' + f.${table}__${key}.value,"",options);
	return true;
}

function form_print_post_${form_name}(message_window) {
	// Post form key value.

	var f = self.document.forms['$form_name'] || self.document.forms[$Default_Form];

	if (message_window) {
		var m = message_window.document.forms['result'];
		if (f.${table}__${key}.value == '' && m && m.keyval) {
			var keyval = m.keyval.value;

			if (keyval != '') {
				// Assign new found key value
				f.${table}__${key}.value = keyval;

				// Display key value on form
				var node = document.getElementById('$Field_Prefix${table}__${key}');
				if (node) node = node.firstChild;	// hidden <input> tag
				if (node) node = node.nextSibling;	// ID <span> tag
				if (node && node.tagName == 'SPAN') {
					IDNode.replace(node,document.createTextNode(keyval),node.firstChild);
				}

				// Change form from New mode to Edit mode
				f._request.value = 'Edit';
			}
		}

		form_is_dirty(f,false);

		// Re-execute print function
		form_print_${form_name}();
	}
}
</script>
End_Of_Script

	#
	# Construct button to print records.
	#

	return <<End_Of_HTML;
<!-- Begin Print button -->
<div class="$class-header">
<table width="100%" align="center" border="0" cellpadding="0" cellspacing="0">
<tr>
  <td align="left">
<div class="$class-title">
$print_title
</div></td>
  
  <td align="right">
<div class="$class-button">
<input type="button"
 value="$button_title"
 onClick="form_print_${form_name}()"
 class="$class-button"$tabindex_str
$attributes>
</div></td>
</tr>
</table>
</div>
<!-- End Print button -->
End_Of_HTML
}
END_OF_FUNC

#-------------------------------------------------------------------------------
'_query_button' => <<'END_OF_FUNC',
sub _query_button {
	my ($form_name,$field_name,$f) = @_;

	#
	# Make a button that calls up a parameter screen for queries.
	#

	my $query_type;
	if (! defined($f)) {
		# Assume undefined signifies an ID field.
		$query_type = 'number';
	}
	elsif (! ref($f)) {
		# \$f is not an object, assume it is a query_type spec.
		$query_type = $f;
	}
	else {
		$query_type = 
		    $f->is_text()   ? 'text'
		  : $f->is_number() ? 'number'
		  : $f->is_time()   ? 'time'
		  : $f->is_date()   ? 'date'
		  : '';
	}

	return '' unless $query_type;

	my $tabindex_str = qq[ tabindex="$No_Tabindex"];

	# For cross-browser implementation, where <a> tag's style needs to be
	# display:block in order to specify the height, so instead use a
	# transparent image for forground so buttons don't get rendered when
	# printed.

	my $img = $ENV{HTTP_USER_AGENT} =~ /MSIE/
		? ''
		: qq[<img src="$Images/transparent.gif" width="22" height="20" border="0">];

	return qq[<a
 href="javascript:${query_type}_query(document.$form_name.$field_name)"
 title="Display a small form to enter $query_type parameters for this field."$tabindex_str
 class="param"
>$img</a>];
}
END_OF_FUNC

#-------------------------------------------------------------------------------
'_master_assign' => <<'END_OF_FUNC',
use Acme::Form::Assign;
sub _master_assign {
	my ($t) = @_;

	#
	# Define assignments.
	#

	my $assignments = '';
	my $assign_scripts = '';# scripts actually assigning master values 

	my $master_assign         = $t->master_assign;
	my $master_assign_options = $t->master_assign_options;
	my $master_assign_input   = $t->master_assign_input;

	if ($master_assign || $master_assign_options || $master_assign_input) {

		my $asgn = new Acme::Form::Assign
			form	     => 'f',	# synonym for current form
			result	     => 'm',	# synonym for search results
			t	     => $t,	# table object of current form
			context	     => $Context;

		$assignments .= $asgn->assign_options( $master_assign_options );
		$assignments .= $asgn->assign_fields(  $master_assign         );
		$assignments .= $asgn->assign_input(   $master_assign_input   );

		chop $assignments;
		$assignments =~ s/^/\t\t/mg;	# pretty-print

		return <<End_Of_JavaScript;

// Initialize form fields from master window.
var m, master_field;
if (f._master && (master_field = f._master.value)) {
	var w = window.opener;
	if (w && ! w.closed) {
		// Find form in master window that contains master_field.
		for (var i = 0; i < w.document.forms.length; i++) {
			if (w.document.forms[i].elements[master_field]) {
				m = w.document.forms[i];
				break;
			}
		}
	}
	if (m) {
		// Assign default values to fields from master form here.
$assignments
	}
}
End_Of_JavaScript
	}

	return '';
}
END_OF_FUNC

#-------------------------------------------------------------------------------
'_detail_button' => <<'END_OF_FUNC',
sub _detail_button {
	my ($detail,$form_name,$field_name,$table,$field,$t,$f,$button_attr,$request) = @_;

	#
	# Define Detail button.
	#

	my $obj = $t;
	my $class = 'detail';		# css style for details after table
	if ($f) {
		$obj = $f;		# use field object for detail settings
		$class = 'detail-field';# css style for detail after fields
	}

	#
	# Define detail name from field name.
	#

	my ($detail_table,$detail_field) = $detail =~ /(\w+)\.(\w+)/;

	my $detail_join = "$detail = $table.$field";

	#
	# Get field to go to after a successful execution.
	#

	my $focus_next_field = $obj->detail_next
	  if ! $request || $request =~ /new/i;
	$focus_next_field ||= '';
	if ($focus_next_field) {
		$focus_next_field = ($focus_next_field =~ /^[%\$]/)
		  ? $Context->translate($focus_next_field, form=>'f')
		  : 'f.'.$Context->db_to_form($focus_next_field);
		$focus_next_field .=
qq[.focus();\n\tif ($focus_next_field.select) $focus_next_field.select();\n];
	}


	#
	# Get programs to use to create new detail records and to edit them.
	#

	my $dt = $Schema->t($detail_table);
	my $new_prog  = $dt->new_prog;
	my $edit_prog = $dt->edit_prog;

	#
	# Get key field to use to check if master record saved.
	#
	
	my $key = $t->key;

	#
	# Get titles
	#

	my $row_name     = $dt->row_name;
	my $button_title = $obj->detail_button_title || 'New ' . $dt->row_title;
	my $detail_title = $obj->detail_title        || $dt->form_list_title;

	#
	# Button options given on Form Directive command are like the Print
	# button where the first argument is the button title and the other
	# items are attributes to the <input> button tag.
	#

	my $attributes = '';
	if ($button_attr) {
		if ($button_attr =~ /['"]/) {

			#
			# Quotes are used in button attributes.
			# Parse the line on spaces and use only the first item as the
			# title and use the rest as attributes to the <input> button tag.
			#

			(my @attr) = Text::ParseWords::parse_line('\s+',1,$button_attr);

			# The first item of the attributes is the button title only if
			# it begins with a quote character. Otherwise, it is interpreted
			# as an attribute key word.

			if ($button_attr =~ /^['"]/) {
				$button_title = shift @attr;
				$button_title =~ s/^(['"])(.*)\1$/$2/;	# remove quotes
			}

			$attributes = join ' ', @attr  if @attr;
			$attributes = " $attributes\n" if $attributes;
		}
		else {
			# No quotes are used.
			# Interpret the entire text as a button title.

			$button_title = $button_attr;
		}
	}

	#
	# Compute tab order
	#

	$tabindex += $Tabindex_Interval;
	#my $tabindex_str = qq[ tabindex="$tabindex"];
	my $tabindex_str = qq[ tabindex="$No_Tabindex"];

	my $save_msg   = 'Saving ' . $t->row_name . '.';
	my $save_title = 'Saving ' . $t->title;

	# Keep following scripts with button
	my $scripts = <<End_Of_Script;

/*
 * Detail button support for $field_name
 */

var _detail_win_${field_name};

function _detail_new_${field_name}(options) {
	var f = self.document.forms['$form_name'] || self.document.forms[$Default_Form];

	if (f.${table}__${key}.value == '') {
		if (! form_is_dirty(f)) {
			var q = 'This form must be saved first before adding' +
			  ' a $row_name.\\n' +
			  ' Do you want to save and continue?';
			if (!confirm(q)) return;
		}

		MessageWindowAutoCloseTime = -1;	// force autoclose of message window
		form_save(f,'$save_msg','$save_title',_detail_post_master_${field_name}) && f.submit();
		return;		// after saving, this routine gets run again
	}

	if (_detail_win_${field_name} == (void 0) || _detail_win_${field_name}.closed) {
		// Open detail window.
		_detail_win_${field_name} = new_form(
		  '$new_prog?table=$detail_table&request=new&master=$field_name',
		  '${field_name}',
		  options);
	}
	else {
		// Show opened detail window.
		_detail_win_${field_name}.focus();
	}
	return true;
}

function _detail_open_${field_name}(url,id) {

	if (! id) id = '0';

	// Open detail window.
	open_form(url+id,'detail_'+id+'_${field_name}');
}

function _detail_post_master_${field_name}(message_window) {
	// Post master form key value so that detail records can be added.

	var f = self.document.forms['$form_name'] || self.document.forms[$Default_Form];

	MessageWindowAutoCloseTime = null;	// disable control over message window closing

	if (message_window) {
		var m = message_window.document.forms['result'];
		if (m && m.keyval) {
			var keyval = m.keyval.value;

			// Assign new found key value
			f.${table}__${key}.value = keyval;

			// Reset form start time to time record was saved.
			f._start.value = m._timestamp.value;

			// Display key value on form
			var node = document.getElementById('$Field_Prefix${table}__${key}');
			if (node) node = node.firstChild;	// hidden <input> tag
			if (node) node = node.nextSibling;	// ID <span> tag
			if (node && node.tagName == 'SPAN') {
				IDNode.replace(node,document.createTextNode(keyval),node.firstChild);
			}

			// Change form from New mode to Edit mode
			f._request.value = 'Edit';

			// Set flag to keep form open like New forms do
			// ### Note: Must close form to force redisplay of original
			// form without the previous record's detail rows.
			//self.Keep_Open = true;

			// Reset dirty flag
			form_is_dirty(f,false);

			// Re-execute new detail function
			_detail_new_${field_name}();
		}
	}
}

function _detail_post_${field_name}(d, message_window) {
	// Post changes in detail form to this form.

	var f = self.document.forms['$form_name'] || self.document.forms[$Default_Form];
	var m = message_window.document.forms['result'];

	// There are forms in three windows open at this time, this form f
	// (the master form which contains the detail record list), the detail
	// form d (the information just entered), and the message window form m
	// (the dialog window) conveying the detail record "posted" message
	// and the information that was just posted.

	// Display detail on current form
	if (m) _detail_display_${field_name}(m.keyval.value,message_window);

	// Set this form's fields to detail form field values.
	if (self._detail_assign_${field_name}) _detail_assign_${field_name}(f,d);

	// Recompute form.
	form_compute(f);
}

function _detail_display_${field_name}(keyval,message_window) {
	// Enter detail item into detail list

	// Get detail <table> in form f where information is to be entered.
	var ftbody = document.getElementById('_detail_table_${field_name}');
	if (! ftbody) return;

	// Import detail record listing from message window.
	// This is complicated by table elements <table>,<tr>,or <td> which
	// cannot be cloned across windows, but their contents can (in IE5).
	// Clone the detail blank row example, and then fill in the <td>
	// elements with the corresponding <td> elements in the message
	// window. This is how the server generates a formated detail row
	// for display in the current master form.

	var msgrow = message_window.document.getElementById('_detail_${field_name}_'+keyval);

	// Clone blank example row in form f table.
	var blankrow = document.getElementById('_detail_blank_${field_name}');
	var newrow = blankrow.cloneNode(true);
	newrow.setAttribute('id','_detail_${field_name}_'+keyval);
	newrow.style.display = '';	// remove 'display:none' directive

	// Copy child nodes of <td> elements in message_window to <td> elements
	// in this window.
	var node;
	for (var i = 0; i < newrow.childNodes.length; i++) {
		node = msgrow.childNodes[i];
		if (node && node.nodeType == 1) {
			// Transfer only Element nodes, not text nodes.
			// Use MS IE5 and NS6 innerHTML method
			newrow.childNodes[i].innerHTML = node.innerHTML;
		}
	}

	// Replace edited row with saved information, or insert a new row.
	var frow = document.getElementById('_detail_${field_name}_'+keyval);
	if (frow) {
		NewNodes.replace(ftbody,newrow,frow);
	} else {
		NewNodes.append(ftbody,newrow);
	}
}

function _detail_return_${field_name}() {
	// Return from detail window opened in _detail_${field_name}().
	var f = self.document.forms['$form_name'] || self.document.forms[$Default_Form];

	// The detail window closes itself, reset flag.
	_detail_win_${field_name} = void 0;
	// Set keyboard focus to next input field here.
	$focus_next_field
}
End_Of_Script

	#
	# Construct button for adding new detail records.
	#

	return <<End_Of_HTML;
<!-- Begin Detail button $field_name -->
<div class="$class-header">
<table width="100%" align="center" border="0" cellpadding="0" cellspacing="0">
<tr>
  <td align="left">
<div class="$class-title">
$detail_title
</div></td>
  
  <td align="right">
<div class="$class-button">
<input type="button"
 value="$button_title"
 onClick="_detail_new_${field_name}()"
 class="$class-button"$tabindex_str
$attributes>
<script language="JavaScript" type="text/javascript">
$scripts
</script></div></td>
</tr>
</table>
</div>

  </td>
</tr>
</table>
<table class="form">
<tr>
  <td colspan=2>

<div id="_detail_$field_name" class="$class-list">
<!--\$ Detail $field_name $detail_join -->
</div>
<!-- End Detail button $field_name -->
End_Of_HTML
}
END_OF_FUNC

#-------------------------------------------------------------------------------
'_table_detail' => <<'END_OF_FUNC',
sub _table_detail {
	my ($detail,$form_name,$field_name,$table,$field,$t,$button_attr,$request)=@_;

	# Make detail button and detail list for inclusion after a table's form.

	my $button = _detail_button(
				$detail,
				$form_name,
				$field_name,
				$table,
				$field,
				$t,
				undef,
				$button_attr,
				$request);

	return <<End_Of_HTML;
<tr>
  <td colspan=2>
  $button
  </td>
</tr>
End_Of_HTML
}
END_OF_FUNC
);

#------------------------------------------------------------------------------

sub AUTOLOAD {
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;
	return if $method eq 'DESTROY';		# gobble
	if (exists $Subs{$method}) {
		my $s = $Subs{$method};
		delete $Subs{$method};		# release memory

		if ($s) {
			eval "package $pkg; $s";
			die "${pkg}::$method(): $@" if $@;
			goto &$method;
		}
	}
	return '';				# unknown method
}

1;
