package Acme::Form::Context;

=head1 NAME

Acme::Form::Context - Track context of form fields in relation to database rows.

=head1 SYNOPSIS

    use Acme::Form::Context;

    $Context = new Acme::Form::Context form=>"form_name";
       or
    $Context = new Acme::Form::Context
                   form=>"form_name",
		   table=>"table_name";

    my $field_name = $Context->translate($string);
      or
    my $field_name = $Context->translate($string, tr_value => 1);
      or
    my $field_name = $Context->translate($string, form => 'myform');
      or
    my $field_name = $Context->translate(
    					$string,
					target => {
						FORM1 => 'myform',
						FORM2 => 'otherform',
						frm   => 'frm_form'
						}
					);

    $Context->push('form_field_name');		# push new context on stack
    $Context->pop();				# pop context

    $Context->next();				# increments current row number

    $Context->set(tr_value => 1);		# manually set an option
    $Context->get(tr_value);			# manually set an option

    my $rownum = $Context->row;

    my $form_var = $Context->db_to_form('table.column');
    my $db_var   = $Context->form_to_db('3_table__column');

Utility routine interface:

    use Acme::Form::Context qw(db_to_form form_to_db);

    my $form_var = db_to_form('table.column', $row, $context);
    my $db_var   = form_to_db('3_table__column');

=head1 DESCRIPTION

The I<Acme::Form::Context> object maintains the current context of a form so
database variables can be correctly translated into corresponding form field
names.

Form field names -- the names of form <input> and <textarea> tags -- contain
context information so that they can be correctly saved to the proper database
tables. Forms can (are intended to) hold several database records, like a
spreadsheet. Form fields can also be related to parent records, such as
when records of a detail table are on the same form as the master table.

Contexts are push()'ed and pop()'ed to/from the context. Both push() and pop()
return the current context string. A context is essentially a form field name
like 'table__column'. It could also have a database row number in the field
name like '4_table__column'.

More than one row of database information can be listed on a form. Row numbers
are not embedded in the form field names for the first row, but the second row
on have the row number prepended to the form field name like '5_table_column'.
The second row number is 1, the third is 2, .... 

The next() method increments the current row number of the context and returns
that number.

Components of the context object are available through method with their same
name. The method row() returns the current row number, form() returns
the form's name, and context() returns an array of the context's.

A get() and set() method are available for getting and setting object component
values.

Setting the 'table' component defines a default table name to prepend to
any unqualified form variable found in the translation string.

Setting the 'form' component defines the form component to prepend to the
form variables.

Setting the 'form_name' component defines the form name to use in constructing
the form component prepended to form variables. This is different from setting
'form' in that the string 'self.document.' is prepended to the 'form_name'.

Setting the 'target' component defines a hash of names corresponding to
actual form names. If the first component of a string (from C<$> or C<%> to
the first period) matches one of the target names, the actual form name is
substituted into the form variable.

The translate() method scans the input string for form field names beginning
with % (percent) and translates them to a fully qualified form field name.
Form field variable names are translated according to the settings of
table, row, and the contexts listed in @context. Curly braces can be used
to surround the field name like %{name} if the following characters would
be interpreted as part of the name.

There are six options that control the translation of form field
variables.  They are in pairs, opposite of each other. They all have
the 'tr_' prefix.  The options C<tr_fq> and C<tr_form> control if the
translation of the form field is to include the 'self.document.form'
prefix (tr_fq => 1) or not (tr_form => 1).  The options C<tr_value>
and C<tr_field> control if the translation includes the '.value' suffix
(tr_value => 1) or not (tr_field => 1). The C<tr_event> and C<tr_script>
options control if the variable translation should be for a script appearing
in a field's events (tr_event => 1>, or if the variable substitution
is for a script between the <script>...</script> tags (tr_script => 1).
Setting the C<form> or C<form_name> option implies C<tr_script>;

The default option settings is C<tr_fq>, C<tr_field>, and C<tr_event>
which translates field variables to be fully qualified form field objects
intended to be substituted into form field events (onFocus="script"
or onChange="script").

The translate() method also scans the input for form field names beginning
with C<$> (dollar) and translates them to fully qualified form field values.
Using the C<$> substitution is like using C<%> with the options tr_fq => 1
and tr_value => 1.  Setting tr_field => 1 in translate()'s arguments changes
the substitution of C<$> to be just like the C<%> substitution. Another
slight difference using the C<$> substitution is that the variable names
cannot contain a period (.) as the C<%> allows to specify C<%{table.column}>
database names. The C<$> substitution uses the double underline __ to specify
a fully qualified database column name like C<$table__column>, which is the
same as Perl's convention.

Context settings/options can be set in translate() arguments, which override
the current context settings and applies only to that specific translation.

The db_to_form() and the form_to_db() utility routines convert database names
(C<table.column>) to form names (C<table__name>) and form names back to database
names. They both can be called as methods ($Context->db_to_form('name'))
or as normal functions. The form context can be specified to db_to_form()
by using the optional arguments $row and $context.

=head1 TODO

 Need to test the contexts and rows functionality. Forms are not equipped yet.

=head1 CHANGE LOG

 4/16/01  Initial version.
 4/18/01  Incorporated db_to_form() and form_to_db() translation utilities.
 4/20/01  Added $ variable substitution to the form field's value.
          Added tr_event/tr_script translation options for event scripts.
 6/25/01  Added target form substitutions.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

require Exporter;

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(form_to_db db_to_form);

#-------------------------------------------------------------------------------

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my $self = {
		form	  => '',
		form_name => 'forms[0]',# name of form, default is first form
		table     => '',
		row       => 0,
		context   => [],
		tr_fq     => 1,		# translate to fully qualified names
		tr_form   => 0,		# translate with no form name prefix
		tr_field  => 1,		# translate to form field object
		tr_value  => 0,		# translate to form field value
		tr_event  => 1,		# translate using this.form
		tr_script => 0,		# translate using self.document.forms[0]
		target    => undef,
		@_
		};

	return bless $self, $class;
}

sub next { my ($self) = @_; return ++$self->{row}; }

sub push {
	my ($self,$form_field) = @_;
	push @{$self->{context}}, $form_field;
	return (join '___', @{$self->{context}});
}

sub pop {
	my ($self,$form_field) = @_;
	pop @{$self->{context}};
	if (@{$self->{context}}) {
		return (join '___', @{$self->{context}});
	}
	else {
		return '';
	}
}

sub context { my ($self) = @_; return (join '___', @{$self->{context}}); }

sub get { my ($self,$key) = @_; return $self->{$key} || ''; }

sub set {
	my ($self,$key,$val) = @_;
	my $prev = $self->{$key};
	$self->{$key} = $val;
	return $prev;
}

sub translate {
	my $self = shift;
	my $string = shift;
	return '' unless defined $string;

	my $opt = { @_ };		# remaining arguments are options

	# Set translation options
	my ($tr_fq, $tr_value, $tr_event);	# reduced option settings

	$tr_fq = $self->{tr_fq};
	$tr_fq = $opt->{tr_fq}		if defined $opt->{tr_fq};
	$tr_fq = ! $opt->{tr_form}	if defined $opt->{tr_form};
	$tr_fq = 1			if defined $opt->{form};

	$tr_value = $self->{tr_value};
	$tr_value = $opt->{tr_value}	if defined $opt->{tr_value};
	$tr_value = ! $opt->{tr_field}	if defined $opt->{tr_field};

	$tr_event = $self->{tr_event};
	$tr_event = $opt->{tr_event}	if defined $opt->{tr_event};
	$tr_event = ! $opt->{tr_script}	if defined $opt->{tr_script};
	$tr_event = 0			if defined $opt->{form};

	my $form = $opt->{form} || $self->{form};
	$form ||= 'self.document.' . ($opt->{form_name} || $self->{form_name});

	my $target = $opt->{target} || $self->{target} || {};

	# Translation function for each variable found in string.
	my $tr = sub {
		my $str = shift;

		return '' unless defined $str;

		# If first character is \, gobble it and return full string.
		$str =~ s/^\\// and return $str;

		# Remove form variable indicator character % or $.
		$str =~ s/^([%\$])// or return $str;

		# Set tranlation option if $ substitution used.
		# Setting the tr_field option in the translation() arguments
		# overrides the $ substitution from appending '.value'.
		$tr_value = 1 if $1 eq '$' && ! $opt->{tr_field};

		# Remove curly braces {} if they were used
		$str =~ s/^{(.*)}/$1/;

		# Get actual form name and remove form code if a target form.
		my $frm = ($str =~ /^(\w+)\./)
		  ? $target->{$1} || ''
		  : '';
		$str =~ s/$1\.// if $frm;

		# Change posible periods . to double underline __
		$str =~ s/\./__/g;

		# Prepend table name if default specified.
		# Leave display-only fields alone (those beginning with caps)
		my $tbl = $opt->{table} || $self->{table};
		$str = $tbl . "__$str"
		  if ($tbl && $str !~ /__/ && $str !~ /^[A-Z]_/);

		# Prepend row number if greater than zero.
		my $row = $opt->{row} || $self->{row};
		$str = $row . "_$str"
		  if ($row > 0); 
	
		# Append context strings.
		my $context = $opt->{context} || $self->{context};
		$str .= '___' . (join '___', @{$context})
		  if $context && @{$context};

		# Prepend form name.
		if ($frm) {
			$str = "$frm.$str";
		}
		else {
			if ($tr_fq) {
				if ($tr_event || ! $form) {
					$str = "this.form.$str";
				}
				else {
					$str = "$form.$str";
				}
			}
		}

		# Append '.value' if option set
		$str =  "$str.value" if $tr_value;

		return $str;
	};

	# Start translation.

	# Substitute %name or %{name}, ignore \%name or \%{name}
	$string =~ s/(
		     \\?	# grab possible backslash quote
		     %		# indicator for begin of form field name
		     {?		# grab possible left curly {
		     [\w.]+	# grab next alpha-numeric with a possible .
		     }?		# match possible right (closing) curly }
		     )
		    /&$tr($1)/egsx;	# translate for all matches on all lines

	# Substitute $name or ${name}, ignore \$name or \${name}
	$string =~ s/(
		     \\?	# grab possible backslash quote
		     \$		# indicator for begin of form field name
		     {?		# grab possible left curly {
		     [\w.]+	# grab next alpha-numeric with a possible .
		     }?		# match possible right (closing) curly }
		     )
		    /&$tr($1)/gsex;	# translate for all matches on all lines

	# Remove trailing semi-colon if script is for event handlers.
	$string =~ s/\s*;$// if $tr_event;

	return $string;
}

use vars qw($Context);		# global object for db_to_form().

sub db_to_form {
	# This can be called both as a utility routine or as an object method.
	# Use global object if called as a utility routine.
	my $c;
	if (ref($_[0]) || $_[0] eq __PACKAGE__) {
		$c = shift;
	}
	else {
		$c = $Context ||= new Acme::Form::Context;
	}

	my ($db_name, $nrow, $context) = @_;

	$c->set(row => $nrow)        if $nrow;
	$c->set(context => $context) if $context;

	# Prepend a form variable symbol % to db_name and translate it.
	return $c->translate("%$db_name", tr_form=>1) if $db_name;
	return '';
}

sub form_to_db {
	# This can be called both as a utility routine or as an object method.
	shift if ref($_[0]) || $_[0] eq __PACKAGE__;

	my $var_name = shift or return '';

	# Initialize return variables.
	my ($db_name, $nrow, $context) = ($var_name, '', '');

	# Strip sequence number from form variable.
	$db_name =~ s/^(\d+)_// and $nrow = $1;

	# Strip context if master/detail relationship.
	$db_name =~ s/___(.*)// and $context = $1;

	# Translate variable name into database table.field name.
	$db_name =~ s/__/./g;

	return ($db_name,$nrow,$context) if wantarray;
	return $db_name;
}

sub AUTOLOAD {
        my $self = shift;
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;
	return if $method eq 'DESTROY';		# gobble

	return $self->get($method);
}

1;
