package Acme::Form::Filler;

=head1 NAME

Acme::Form::Filler - Fill in form template from database

=head1 SYNOPSIS

    #
    # Fill in all forms in a HTML page.
    #

    use Acme::Form;
    use Acme::Form::Filler;
    our @ISA = qw(Acme::Form::Filler);

    my $self = new Acme::Form;

    $self->fill;
     -or-
    $self->fill(field1=>'value1',field2=>'value2');

    #
    # To fill only one form in a HTML page.
    #

    require Acme::Form::Filler;

    Acme::Form::Filler::fill_form($form_obj);
     -or-
    Acme::Form::Filler::fill_form($form_obj, field1=>'value1',field2=>'value2');


=head1 DESCRIPTION

I<Acme::Form::Filler> is a subclass of I<Acme::Form>. It provides methods to
fill form fields with database (or default) values.

The fill() method fills in all forms in a HTML file using fill_form() on
each form found.

The fill_form() method fills a template with database values. Dropdown lists
of values are substituted fields with the Choose schema properties. Multiple
input fields are substituted for fields with the Input_Array schema properties.

HTML tags of the form:

    <!--$ command string -->

are substitued with the command string's output. (see I<Local::Acme::Extensions>).

The fill_form() method fills forms for three situations depending upon the
C<$request> parameter: New, Edit, or Query. If C<$request> is 'New', then
the form default values are filled in. If the C<$request> is 'Edit', then
the current database values for the form are filled in. And if the C<$request>
is 'Query', the form is filled in to be used to generate lists from the 
search parameters entered in the form fields.

The fill_form() method is used by I<Acme::Form::Saver> to fill in form values to
compare to cgi variables to see which fields changed. When saving form values,
lookup tables are not searched and dropdown lists are not made. The resultant
html is not usable.

The fill_form() (and fill()) method also accepts a printing flag, which if 
true, changes the output of the form to be acceptable for printing. An extra
stylesheet C</styles/form_print.css> is refered to which can change the
appearance of fields and page dividers.

=head1 CHANGE LOG

 5/9/00   Broke out from Acme::Form.
 5/3/01   Changed to use Acme::Base::Schema. Removed field_default().
          Added commands.
 6/22/01  Changed so New records look in CGI arguments for default values too.
          Started to use Acme::HTML::Tag to parse <input> tags, which forces all
	  attribute names to lower case.
 7/14/01  Added Print request.
 7/17/01  Moved command substitution after tag substitution so detail lists
          using <input> tags won't be re-substituted.
 10/24/01 Generalized assignments and implemented choose_assign using
 	  Acme::Form::Assign.
  3/19/02 Added fill() method to fill all forms on web page.
  4/1/02  Use Acme::Form::Vars to store database values to facilitate executing
          Do and Do_filler schema properties.
  4/6/02  Threaded save option/object into fill_form() for Acme::Form::Saver
          to obtain original values for form fields to compare before saving.
  6/20/02 Modified fill_form() to fill forms used for entering query parameters.


=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000,2001  Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

use CGI;
use Acme::HTTP			qw(escape);
use Acme::HTML			qw(escape_html unescape_html display_message);
use Acme::HTML::ListMenus;
use Acme::HTML::InputArray;
use Acme::HTML::Tag;
use Acme::Db;
use Acme::Form::Vars;
use Acme::Form::Schema;
use Acme::Base::Schema;
use Acme::Base::Schema::Join;
use Acme::Base::Schema::Master;
use Acme::SQL::Join;
use Acme::Date			qw(today	timestamp);
use Acme::Config;
use Local::Acme::User		qw(user);
use Local::Acme::Format;
use Local::Acme::Form::Extensions;
use Text::ParseWords;

use Acme::Security;

our $Schema;
our $Primary_Table;
our $Primary_Key;
our $Keyval;
our $ID;
our $Sql;
our $N_Form;
our $Fill_Mode;
our $Init_Scripts;
our $User;
our $Today;
our $Timestamp;
our $Saving;
our $Start;
our $Printing;

#-------------------------------------------------------------------------------
#
# Module for holding initial database values for form fields.
#
#-------------------------------------------------------------------------------

package Acme::Form::DbVars;
our @ISA = qw(Acme::Form::Vars);

#-------------------------------------------------------------------------------
#
#  Filler public methods
#
#-------------------------------------------------------------------------------

package Acme::Form::Filler;

sub fill {
	my $self = shift;
	my %opt = @_;		# rest of arguments are key/value options

	# Note: expanding fill_form() to fill in several forms on a page
	# increased the memory used and the number of passes through the
	# form file so that now it would be better to go to a complete
	# event driven pass through the page like the HTML::Entities from
	# CPAN. (mp 3/22/02)

	my $args      = $opt{args}		||
			$self->{args}		||
			$self->{cgi}		||
			new CGI;
	$args = new CGI $args if ref($args) eq 'HASH';

	my $request   = $opt{request}		||
			$self->{request}	||
			$args->param('request')	||
			'';
	my $title     = $opt{title}		||
			$self->{title}		||
			$args->param('title')	||
			'';
	my $print     = ($opt{print} ? 1 : 0)   ||
			defined $self->{print}  ||
			(defined $args->param('print') ? 1 : 0) ||
			$request =~ /print/i;

	#
	# Set flag if form is to be printed. (no dropdown lists, no buttons)
	#

	local $Printing = $print;

	#
	# Fill forms in HTML page with database values and default values.
	#

	$self->{html} =~ s/<title>\s*?</<title>$title</s if $title;

	my $indicator = lc($request);
	$self->{html} =~ s/<title>(.*?)</<title>$indicator $1</s; # add edit indicator

	#
	# Process each form.
	#

	local $N_Form = 0;	# initialize form number (zero based)

	foreach (		# extract forms from html
		$self->{html} =~ m{
			^(.+?)(?=<form)		# prefix
			(<form.*?</form\s*>)	# first form
			(?:(.*?(?=<form))(<form.*?</form\s*>))*	# other forms
			(.+)$			# postfix
			}six
	) {
		next unless $_;

		if (/<form/i) {
			# Fill each form.
			$self->{html} .= $self->fill_form($_,\%opt);
			$N_Form++;
		}
		else {
			# Initialize first time.
			$self->{html} = '' if $N_Form == 0;

			# Fill stuff outside of <form> tags
			$self->{html} .= $self->fill_fluff($_,\%opt);
		}
	}

	#
	# Change form if printing.
	#

	if ($Printing) {

		#
		# Setup filled form to be printed.
		#

		# Add stylesheet used for printing forms.
		my $styles = Acme::Config->Styles;
		$self->{html} =~ s{(</head>)}
			   {<link rel="stylesheet" href="$styles/form_print.css" type="text/css">\n$1}si;

		# Add print command to <body onload=...> event.
		$self->{html} =~ s{(<body\b.*?)>}
			   {
				my $prcmd = "window.print();window.setTimeout('window.close()',1500)";

				my $tag = $1;
				$tag =~ s/\b(onload=(["']).*?)\2/$1;$prcmd$2/i
				  or
				$tag .= qq[ onload="$prcmd"];
				$tag . '>';
			   }sei;
	}

	return $self->{html};
}

#-------------------------------------------------------------------------------

sub fill_fluff {
	my $self = shift;	# an Acme::Form object
	my ($html) = (@_);

	#
	# Substitute commands that are outside <form> tags.
	#

	$html =~ s/(<!--+\$+\s*.*?\s*--+>)/_tag_command($1,$self->{extend})/egs;

	return $html;
}

#-------------------------------------------------------------------------------

sub fill_form {
	my $self = shift;	# an Acme::Form object
	my ($html,$opt) = @_;

	#
	# Return a form with the fields filled.
	#

	# Note: this routine has become fairly complex because it has three
	# operational modes: New, Edit, or Query. Besides that, there are
	# three operational "environments" in which this routine operates
	# in: Display the form (default), Printing the form, and Saving the
	# form.

	#
	# Determine function parameters
	#

	my $args      = $opt->{args}		||
			$self->{args}		||
			$self->{cgi}		||
			new CGI;
	$args = new CGI $args if ref($args) eq 'HASH';

	my $request   = $opt->{request}		||
			$args->param('request')	||
			$self->{request}	||
			'';
	my $template  = $self->{template};
	my $form      = $opt->{form}		||
			$args->param('form')	||
			$self->{form}		||
			'';
	$form = '' if $template;

	my $table     = $opt->{table}		||
			$args->param('table')	||
			$self->{table}		||
			'';
	my $master    = $args->param("master")	||
			'';
	my $sql       = $opt->{sql}		||
			$args->param('sql')	||
			$self->{sql};
	my $title     = $opt->{title}		||
			$args->param('title')	||
			$self->{title}		||
			'';
	my $print     = ($opt->{print} ? 1 : 0) ||
			(defined $args->param('print') ? 1 : 0) ||
			defined $self->{print}  ||
			$request =~ /print/i;

	#
	# Set flag if form is to be saved. (no output, just field values)
	#

	local $Saving = $opt->{save};

	local $Start  = $opt->{start};	# start time of edit (for Saving)

	#
	# Set flag if form is to be printed. (no dropdown lists, no buttons,...)
	#

	local $Printing = $print;

	#
	# Define and initialize global variables
	#

	local $Schema = $self->{schema} || new Acme::Form::Schema { form => $form };

	# Default primary table is the first table listed in a form file.
	($table) = $Schema->tables if ! $table && $form;

	local $Primary_Table = $table;

	my $t = $Schema->t($Primary_Table)
	  or die("Cannot find the table '$Primary_Table' in the"
	  . ($form ? " form '$form'" : " schema")
	  . " description");


	local $Primary_Key = $t->key;

	local $Init_Scripts = '';	# Special initialization javascript

	local $ID = $t->id;	# key treated differently if it is an ID number

	$self->{has_id_key} = 1 if $ID;	# Set information flag

	#
	# Fill query forms here for they do not need the database values
	# set up below.
	#

	return _fill_query_form($self,$t,$args,$html) if $request =~ /Query/i;

	#
	# Accept a cgi parameter named 'search' as primary key value
	# if primary key parameter is not specified.
	#

	if (! exists $args->{$Primary_Key}
	&&  ! exists $args->{"$Primary_Table.$Primary_Key"}
	&& (my $search = $args->param('search')) ) {
		$args->param($Primary_Key,$search) if ($search =~ /^\w+$/);
	}

	#
	# The existance of the primary key field value determines if the
	# form is to fill in the table for editing or if the form is
	# to be set up for entering new values.
	#

	# Note: for tables with non-ID keys -- keys that can be changed --
	# the key value of a record being updated is stored in a hidden
	# field with a name beginning with "P_", so we check that for
	# a key value too.

	local $Keyval = $args->param("P_${Primary_Table}__${Primary_Key}")
		     || $args->param("${Primary_Table}__${Primary_Key}")
	             || $args->param("${Primary_Table}.${Primary_Key}")
	             || $args->param($Primary_Key);

	local ($Fill_Mode, $User, $Today, $Timestamp);

	#
	# Define record set for holding variables to fill forms.
	#

	my $rs = new Acme::Form::DbVars;

	#
	# Fill record set with values for form fields.
	#

	$request ||= 'Edit' if $Keyval;		# set default if key specified

	if ( $request !~ /New/i 
	&& ( exists $args->{"${Primary_Table}__${Primary_Key}"}
	  || exists $args->{"${Primary_Table}.${Primary_Key}"}
	  || exists $args->{$Primary_Key} ) ) {

		$Fill_Mode = 'Edit';

		$sql ||= _make_sql($t,$args);

		# Check fields that can be updated
		$sql = Acme::Security::clean_sql($sql,'AU');

		$sql = _add_time_fields($sql) if $t->audit;

		if (! _fill_record_set($rs,$sql,$Primary_Table)) {

			#
			# Filling the recordset failed.
			# Check if security restricted access to record.
			#

			my $keyval = Acme::Db->escape_sql($Keyval);
			my $test_keyval = Acme::Db->get(<<End_Of_SQL);
select $Primary_Key
from $Primary_Table
where $Primary_Key = '$keyval'
End_Of_SQL
			if ($test_keyval) {
				die Acme::Security::alert(
  "You do not have privileges to edit this information.\nPlease check with your administrator."
		  					);
			}
			else {
				display_message("No records were found.", "$title",undef,0);
				exit 1;
			}
		}
	}
	else {
		$Fill_Mode = 'New';

		die "Cannot determine which record to edit. A key value was not specified"
		  if $request =~ /Edit/i;

		$sql = Acme::Security::clean_sql($sql,'AI') if $sql;

		# Note: Always fill the record set even if there is no sql.
		# It is necessary to get the current time and date from
		# the database to make the audit trails more accurate, as
		# the database can reside on a different computer with a
		# different time clock than where this script is executed.

		$sql = _add_time_fields($sql) if $t->audit;

		_fill_record_set($rs,$sql,$Primary_Table) if $sql;
	}

	local $Sql = $sql;	# used as flag

	if ($t->audit) {

		#
		# Set audit values to database generated time values.
		#

		$User      = user();
		$Today     = $rs->get('_today');
		$Timestamp = $rs->get('_timestamp');
	}

	#
	# Substitute commands.
	#

	if (! $Saving) {
		my $do = join ';', $t->do, $t->do_filler;
		if ($do) {

			#
			# Evaluate do_filler property's Perl script.
			# Subroutines defined in do_filter can be invoked
			# with the "Do" form filler command, like:
			#
			#  <!--$ do subroutine($arg1,$arg2,$arg3, ...) -->
			#

			# Note: The do_filler property subverts Perl's
			# taint checks as the schema (and form) files
			# are assumed to be protected files.
		
			package Acme::Form::DbVars;
			no strict;
			($do) = $do =~ /^(.*)$/s;	# untaint
			eval $do;
			die "Error evaluating schema's Do property: $@" if $@;
		}

		$html =~ s/(<!--+\$+\s*.*?\s*--+>)/_tag_command($1,$self->{extend})/egs;
	}

	#
	# Substitute <input> tags.
	#

	$html =~ s|(<input\s.*?>)|_tag_input($1,$args)|egsi;

	#
	# Substitute <textarea> tags.
	#

	$html =~ s|(<textarea.*?>.*?</textarea>)|_tag_textarea($1,$args)|egsi;

	#
	# Finish adjusting form with initialization scripts and form tracking
	# fields.
	#

	$Init_Scripts = qq[<script language="JavaScript" type="text/javascript">$Init_Scripts</script>]
	  if $Init_Scripts && ! $Saving;

	my $end_of_form_stuff = $Init_Scripts;

	if ($Printing) {

		#
		# Setup filled form to be printed.
		#

		# Remove calendar buttons.
		$html =~ s|<a\s+href="javascript:displayCalendar.*?</a>||gs;

		# Remove required images.
		$html =~ s|<img src="[^"]*?required.gif".*?>||gs;

		$html =~ s|(</form\s*>)|${end_of_form_stuff}$1|i
		  if $end_of_form_stuff;
	}
	elsif ($Saving) {

		#
		# Import any columns in the SQL query that has not been
		# defined as a form field.
		#

		# The Saver needs to have the joined tables' key values and
		# audit field values even though they may not be input fields
		# on the form. It is needed for entry into the audit tables.
		# This is a quick and dirty method that imports fields from
		# the "database variables object" to the "form saver object"
		# that have not already been defined in the form. It can and
		# does import extra fields. :(

		$Saving->import_var(
				map { ("$_", $rs->get($_)) }
				grep { ! exists $Saving->{var}->{$_} } keys %{$rs->{var}}
				);

		#
		# Import form control variables into form save set
		#

		$Saving->import_var(
				_form    => $form,
				_request => $request,
				_table   => $Primary_Table,
				_key     => $Primary_Key,
				_master  => $master,
				_start   => $Timestamp,
				);
		
	}
	else {
		#
		# Embed request and table fields into form for save.cgi
		#

		my $form = $self->{form} || '';

		$Timestamp ||= timestamp();	# define even if not auditing

		$end_of_form_stuff .= <<EOF;

<input type="hidden" name="_form" value="$form">
<input type="hidden" name="_request" value="$request">
<input type="hidden" name="_table" value="$Primary_Table">
<input type="hidden" name="_key" value="$Primary_Key">
<input type="hidden" name="_master" value="$master">
<input type="hidden" name="_start" value="$Timestamp">
EOF

		if (defined $args->{keep_open}) {
			my $keep_open = $args->{keep_open};
			if ('yes' =~ /\Q$keep_open\E/i) {
				$end_of_form_stuff .= <<EOF;
<script language="JavaScript" type="text/javascript">Keep_Open=true</script>
EOF
			}
			else {
				$end_of_form_stuff .= <<EOF;
<script language="JavaScript" type="text/javascript">Keep_Open=false</script>
EOF
			}
		}

		$html =~ s|(</form\s*>)|$end_of_form_stuff$1|i
		  if $end_of_form_stuff;
	}

	return $html;
}

#-------------------------------------------------------------------------------
#
#  Private methods
#
#-------------------------------------------------------------------------------

sub _tag_input {
	my ($tag,$args) = @_;

	#
	# Substitue values into <input> tags.
	#

	my $tg = new Acme::HTML::Tag $tag;
	my $tag_type = $tg->type;

	#
	# Remove buttons if printing the form.
	#

	if ($Printing) {
		return ''
		  if $tag_type eq 'button'
		  || $tag_type eq 'submit'
		  || $tag_type eq 'reset'
		  || $tag_type eq 'image'
		  || $tag_type eq 'password';
	}
	else {
		return $tag
		  if $tag_type eq 'button'
		  || $tag_type eq 'submit'
		  || $tag_type eq 'reset'
		  || $tag_type eq 'image';
	}

	#
	# Parse table and field names from form field name
	#

	my $tag_name = $tg->name;
	return $tag unless $tag_name;		# ignore tags with out names
	return $tag if $tag_name =~ /^_/;	# ignore private tags

	my ($recno, $table, $field) =
	  $tag_name =~	/^
	  		(?:N(\d+)_)?	# optional N12_ prefix
			(?:(\w*?)__)?	# optional table name
			(\w*)		# required field name
			/x;

	$table ||= $Primary_Table;
	$table =~ s/^[A-Z]_//;		# remove special field indicators (C_, P_)

	if ($Saving && $tag_name !~ /__/) {
		
		#
		# Assure tag_name conforms to "table__field" format.
		# This allows forms with simple field names (not fully
		# qualified) to be saved as long as a table name is
		# provided in the url or the field named "_table".
		#
		
		$tag_name = "${table}__${tag_name}";
	}

	if ($Schema && (my $t = $Schema->t($table))) {
		my $f = $t->{$field};	# can be undefined if field is ID
		$f = $t->display->{$tag_name} if $tag_name =~ /^D_/;

		#
		# Determine tag value
		#

		my $tag_value;

		if ($Fill_Mode eq 'New') {

			if ($t->audit) {

				if ($Saving) {

					#
					# Substitute audit fields with example info.
					#

					# Date and timestamps are entered into 
					# the database by using SQL constants.

					if ($field eq $t->audit_entry) {
						$Saving->set($tag_name => $Today);
						return '';
					}

					if ($field eq $t->audit_update) {
						$Saving->set($tag_name => $Timestamp);
						return '';
					}

					if ($field eq $t->audit_user) {
						$Saving->set($tag_name => $User);
						return '';
					}
				}
				else {
					#
					# Substitute audit fields with example info.
					#

					# Audit date and timestamps are entered into 
					# the database by using SQL constants, so
					# they are the time the record was saved, not
					# this time which is the time the form was
					# rendered.

					return _auditify($Today) if $field eq $t->audit_entry;
					return $Timestamp	 if $field eq $t->audit_update;
					return $User             if $field eq $t->audit_user;
				}
			}


			#
			# Substitute ID field.
			#

			if ($field eq $ID && $table eq $Primary_Table) {
				if ($Saving) {
					$Saving->set($tag_name => undef);
					return '';
				}
				else {

					# Note: it is important that a <span> tag
					# be used and that it is the second node
					# right after the <input> tag. The contents
					# of the <span> can be changed.
					# See idNodeTracker in /scripts/form_save.js

					return $tag .= qq[<span class="id">(Unassigned)</span>];
				}
			}

			#
			# Get default tag value from schema, CGI, database, or
			# field value in form html.
			#

			if ($Saving) {
				DEFAULT: {
					$tag_value = eval $f->default_dynamic if $f;
					last if defined $tag_value;

					$tag_value = $tg->value;
					last if $tag_value;

					$tag_value = '';
				}
			}
			else {
				DEFAULT: {
					$tag_value = $args->param($tag_name);
					last if defined $tag_value;

					$tag_value = eval $f->default_dynamic if $f;
					last if defined $tag_value;

					$tag_value = _db_value($tag_name);
					last if defined $tag_value;

					$tag_value = $tg->value
					   if $tag_type ne 'checkbox'
					   && $tag_type ne 'radio';
					last if $tag_value;

					$tag_value = $args->param("${table}__${field}");
					last if defined $tag_value;

					$tag_value = $args->param("${table}.${field}");
					last if defined $tag_value;

					$tag_value = $args->param("${field}");
					last if defined $tag_value;

					$tag_value = '';
				}
			}
		}
		elsif ($Fill_Mode eq 'Edit') {

			#
			# Get tag value from database.
			#

			$tag_value = _db_value($tag_name) || '';

			#
			# Set display only fields to their default value
			#

			$tag_value = eval $f->default_dynamic if $f && $tag_name =~ /^D_/;

			# Note: detail records list sometimes have Display-only
			# <input> fields with their values already filled in for
			# form computations like TOTAL(). Keep those values next.

			$tag_value ||= $tg->value if $tag_name =~ /^D_/;

			$tag_value ||= '';

			#
			# Fill audit fields
			#

			if ($t->audit) {
				if ($Saving) {

					#
					# Substitute audit fields with example info.
					#

					if ($field eq $t->audit_entry
					||  $field eq $t->audit_update
					||  $field eq $t->audit_user) {
						$Saving->set($tag_name => $tag_value);
						return '';
					}
				}
				else {
					#
					# Substitute audit fields.
					#

					return _auditify($tag_value)
					  if $field eq $t->audit_entry;
					return $tag_value
					  if $field eq $t->audit_update;
					return $tag_value
					  . _make_audit_link(
					  		$table,
							$t->key,
							_db_value(
			($recno ? "N$recno}_" : '') . $table . '__' . $t->key)
								 )
					  if $field eq $t->audit_user;
				}
			}

			#
			# Substitute ID field.
			#

			if ($field eq $ID && $table eq $Primary_Table) {
				if ($Saving) {
					$Saving->set($tag_name => $tag_value);
					return '';
				}
				else {
					# Set hidden field to ID value. Append
					# HTML display of the ID value.

					$tg->set(value => $tag_value);
					return $tg->as_html
					  . qq[<span class="id">$tag_value</span>];
				}
			}

			#
			# Auto-change: change field to CGI value if defined.
			#

			if (! $Saving) {

				# Note: This allows fields to be automatically
				# changed if the field value is specified in
				# the CGI arguments.

				my $change_value;

				CHANGE: {
					$change_value = $args->param($tag_name);
					last if defined $change_value;

					$change_value = $args->param("${table}__${field}");
					last if defined $change_value;

					$change_value = $args->param("${table}.${field}");
					last if defined $change_value;

					# Note: The following is more dangerous as
					# a table name is assumed. Names could 
					# conflict. However, this is more consistant
					# to the search parameters accepted in the url.

					$change_value = $args->param("${field}");
					last if defined $tag_value;
				}

				$tag_value = $change_value if defined $change_value;
			}
		}
		elsif ($Fill_Mode eq 'Query') {

			#
			# Get default tag value only from CGI parameters.
			#

			DEFAULT: {
				$tag_value = $args->param($tag_name);
				last if defined $tag_value;

				$tag_value = $args->param("${table}__${field}");
				last if defined $tag_value;

				$tag_value = $args->param("${table}.${field}");
				last if defined $tag_value;

				$tag_value = $args->param("${field}");
				last if defined $tag_value;

				$tag_value = '';
			}
		}

		if ($Saving) {

			#
			# Save form value to compare with CGI arguments.
			#

			$Saving->set($tag_name => $tag_value);

			#
			# Register checkbox fields so that their "cleared"
			# state is saved to the database. HTML forms do
			# not send the server those checkbox fields that
			# are unchecked.
			#

			$Saving->add_checkbox($tag_name) if $f && $f->checkbox;

			# Skip the rest of the field filling process when saving.
			# It only formats the field values for display.

			return '';
		}

		#
		# Special processing for Combo and Previous support fields
		#

		if ($tag_name =~ /^[CP]_/) {
			$tg->set(value => $tag_value);
			return $tg->as_html;
		}

		#
		# Lookup and fill display fields from search assignments
		#

		if (defined $tag_value
		&& length($tag_value) > 0
		&& ($Sql || $Fill_Mode eq 'New')
		&& $f
		&& ($f->search || $f->choose)) {
			_do_search($f, $tag_name, $tag_value);
		}

		#
		# Substitute dropdown list for input field if defined in schema.
		#

		if (! $Printing
		&& $f
		&& $f->choose
		&& ! $f->hidden
		&& (! $f->readonly || $Fill_Mode eq 'Query')) {
			my $attr = $tg->as_select_attr();
			my $size = $f->choose_size;
			my $d = _choose_dropdown($f,$tag_name,$tag_value,$size,$attr);
			return $d if $d;
		}

		#
		# Set 'checked' property for checkbox and radio buttons.
		#

		if ($tag_type eq 'checkbox' || $tag_type eq 'radio') {
			if (defined $tag_value && $tag_value eq $tg->value) {
				$tg->set(checked => 'yes');
			}
			else {
				$tg->set(checked => 'no');
			}
		}
		else {

			#
			# Redefine tag with new tag value.
			#

			$tg->set(value => $tag_value);
		}

		#
		# Substitute input array for input field if defined in schema
		#

		if ($f && $f->array && $f->input && $Fill_Mode ne 'Query') {

			#
			# Set default table and form to use for input arrays
			# as a safety until all form fields have their table
			# names embedded in the field name.
			#

			$Acme::HTML::InputArray::Table = $Primary_Table if $Primary_Table;

			my $t = Acme::HTML::InputArray->Make($tg->as_html);
			return $t if $t;
		}
	}
	else {
		#
		# No database value to use for tag value.
		#

		my $tag_value = $args->param($tag_name) || $tg->value
		  if $Fill_Mode eq 'New';

		if ($Saving) {
			$Saving->set($tag_name => $tag_value);
			return '';
		}

		$tg->set(value => $tag_value);
	}

	return $tg->as_html;
}

#-------------------------------------------------------------------------------

sub _tag_textarea {
	my ($tag,$args) = @_;

	my $tg = new Acme::HTML::Tag $tag;
	my $tag_name = $tg->name;
	return $tag unless $tag_name;		# ignore tags with out names

	#
	# Parse table and field names from form field name
	#

	my ($recno, $table, $field) =
	  $tag_name =~	/^
	  		(?:N(\d+)_)?	# optional N12_ prefix
			(?:(\w*?)__)?	# optional table name
			(\w*)		# required field name
			/x;

	$table ||= $Primary_Table;

	(my $tag_value) = $tag =~ m|>(.*?)</textarea>|si;

	if ($Saving && $tag_name !~ /__/) {
		
		#
		# Assure tag_name conforms to "table__field" format.
		# This allows forms with simple field names (not fully
		# qualified) to be saved as long as a table name is
		# provided in the url or the field named "_table".
		#
		
		$tag_name = "${table}__${tag_name}";
	}

	if ($Schema && (my $t = $Schema->t($table))) {
		my $f = $t->{$field};	# can be undefined if field is ID
		$f ||= $t->display->{$tag_name} if $tag_name =~ /^D_/;

		#
		# Determine tag value
		#

		$tag_value = unescape_html($tag_value) if $tag_value;

		if ($Fill_Mode eq 'New') {
			#
			# Get default tag value from schema, CGI, database, or
			# field value in form html.
			#

			my $tv = $tag_value;	# save current value

			if ($Saving) {
				DEFAULT: {
					$tag_value = eval $f->default_dynamic if $f;
					last if defined $tag_value;

					$tag_value = $tv;
					last if $tag_value;

					$tag_value = '';
				}
			}
			else {
				DEFAULT: {
					$tag_value = $args->param($tag_name);
					last if defined $tag_value;

					$tag_value = eval $f->default_dynamic if $f;
					last if defined $tag_value;

					$tag_value = _db_value($tag_name);
					last if defined $tag_value;

					$tag_value = $tv;
					last if $tag_value;

					$tag_value = $args->param("${table}__${field}");
					last if defined $tag_value;

					$tag_value = $args->param("${table}.${field}");
					last if defined $tag_value;

					$tag_value = $args->param("${field}");
					last if defined $tag_value;

					$tag_value = '';
				}
			}
		}
		elsif ($Fill_Mode eq 'Edit') {
			#
			# Get tag value from database.
			#

			$tag_value = _db_value($tag_name);

			#
			# Set display only fields to their default value
			#

			$tag_value = eval $f->default_dynamic if $tag_name =~ /^D_/;

			$tag_value ||= '';

		}
		elsif ($Fill_Mode eq 'Query') {

			#
			# Get default tag value only from CGI parameters.
			#

			DEFAULT: {
				$tag_value = $args->param($tag_name);
				last if defined $tag_value;

				$tag_value = $args->param("${table}__${field}");
				last if defined $tag_value;

				$tag_value = $args->param("${table}.${field}");
				last if defined $tag_value;

				$tag_value = $args->param("${field}");
				last if defined $tag_value;

				$tag_value = '';
			}
		}

		if ($Saving) {

			#
			# Save form value to compare with CGI arguments.
			#

			$Saving->set($tag_name => $tag_value);

			# Skip the rest of the field filling process when saving.
			# The rest of the routine only formats the field value
			# for display.

			return '';
		}

		#
		# Lookup and fill display fields from search assignments
		#

		if (defined $tag_value
		&& length($tag_value) > 0
		&& ($Sql || $Fill_Mode eq 'New')
		&& $f
		&& ($f->search || $f->choose)) {
			_do_search($f, $tag_name, $tag_value);
		}

		#
		# Substitute dropdown list for input field if defined in schema.
		#
		# We look in <textarea> tags for dropdown lists or list menus
		# because text arrays of multiple settings get encoded into
		# <textarea> tags if the length of the field is greater than
		# 255, which is often the maximum length of <input> tags.
		#

		if (! $Printing
		&& $f
		&& $f->choose
		&& ! $f->hidden
		&& (! $f->readonly || $Fill_Mode eq 'Query')) {
			my $attr = $tg->as_select_attr();
			my $size = $f->choose_size;
			my $d = _choose_dropdown($f,$tag_name,$tag_value,$size,$attr);
			return $d if $d;
		}
	}
	else {
		#
		# No database value to use for tag value.
		#

		$tag_value = $args->param($tag_name) || $tag_value
		  if $Fill_Mode eq 'New';

		if ($Saving) {
			$Saving->set($tag_name => $tag_value);
			return '';
		}
	}

	$tag_value = escape_html($tag_value);
	$tag =~ s|(<.*?>).*?(</\s*textarea)|$1$tag_value$2|si;

	return $tag;
}

#------------------------------------------------------------------------------

sub _tag_command {
	my ($tag,$extensions) = @_;

	my ($command) = $tag =~ /<!--+\$+\s*(.*?)\s*--+>/s;

	return $tag unless $command;

	COMMAND: {

		#
		# Form Filler Commands
		#

		if ($command =~ s/^Detail\s//) {

			#
			# Process special tag used by Form Builder to
			# tell this Form Filler to list the form's
			# detail records.
			#

			$tag = _detail_list($command) if $Fill_Mode ne 'Query';
			return $tag;
		}

		if ($command =~ s/^Do\s//i) {

			#
			# Evaluate form filler's Do directive.
			# Note: The Do directive subverts Perl's
			# taint checks as the schema (and form) files
			# are assumed to be protected files.
			#
			
			no strict;
			($command) = $command =~ /^(.*)$/s;	# untaint
			$tag = eval "package Acme::Form::DbVars;$command";
			die "Error evaluating schema's Do property. $@" if $@;

			#
			# Set up recursive call to substitute any
			# command tags found in the string just
			# substituted.
			#

			# The number of dollar signs $ in the command
			# string determine the number of substitutions
			# (recursive calls) performed on the tag.

			$tag =~ s/\$//;		# remove first $ in tag.
			$tag =~ s/(<!--+\$+\s*.*?\s*--+>)/_tag_command($1,$extensions)/egs;

			return $tag;
		}

		for my $e (keys %$extensions) {
			if ($command =~ /$e/) {

				#
				# Process programmer defined extensions
				# defined in module constructor.
				#

				my $cmd = $extensions->{$e};

				if ((ref $cmd) =~ /CODE/) {
					$tag .= &$cmd($command);
				}
				else {
					$tag .= eval $cmd;
				}

				#
				# Set up recursive call to substitute any
				# command tags found in the string just
				# substituted.
				#

				# The number of dollar signs $ in the command
				# string determine the number of substitutions
				# (recursive calls) performed on the tag.

				$tag =~ s/\$//;		# remove first $ in tag.
				$tag =~ s/(<!--+\$+\s*.*?\s*--+>)/_tag_command($1,$extensions)/egs;

				return $tag;
			}
		}
	}

	#
	# Check for other site-specific standard commands
	#

	return Local::Acme::Form::Extensions::tag_command($tag,$Printing);
}

#-------------------------------------------------------------------------------

sub _db_value {
	my $field_name = shift;

	#
	# Get database record number from form field name.
	#

	# Return if field_name is a special form field.
	return '' if $field_name =~ /^(N\d+_)?[A-BE-MOQ-Z]_[^_]/;

	# Note: The fields with the following prefixes are filled with database
	# values: C_ (combo), D_ (display), and P_ (previous value).

	# Remove special form field prefixes (keep record number).
	# Ignore uninitialized warnings of $1 which occurs most often.
	no warnings qw(uninitialized);
	$field_name =~ s/^(N\d+_)?[CDP]_/$1/;

	return Acme::Form::DbVars->get($field_name);
}

#-------------------------------------------------------------------------------

sub _add_time_fields {
	my $sql = shift;

	if (! $sql) {
		# Note: Empty SQL statement. Construct one that will return the
		# current date and time from the database. Only one record should
		# be returned. Postgres does not require a "from" statement in
		# their sql, otherwise, some other table must be queried for
		# "one record" consisting of the current date and time.

		return 'select current_date as _today, current_timestamp as _timestamp';
	}

	my $prefix = '';
	$prefix = $1 if $sql =~ s/(.*;\s*)(select\b)/$2/si;

	# Add _today and _timestamp fields to query
	$sql =~ s/(\s+from\s)/,\ncurrent_date as _today,\ncurrent_timestamp as _timestamp$1/si;

	return "$prefix$sql";
}

#-------------------------------------------------------------------------------

sub _fill_record_set {
	my ($rs,$sql,$default_table) = @_;

	$default_table ||= '';

	my $q = Acme::Db->query($sql);

	my ($table, $field, $field_name, %v);

	if ($q->count > 0) {
		$q->next(\%v);

		for (keys %v) {
			# Field names starting with _ are not parsed into
			# table and field names.
			if (/^_/) {
				$field_name = $_;
			}
			else {
				($table, $field) = /(.*?)\.(.*)/
				  or ($table, $field) = ($default_table,$_);

				$field_name = "${table}__${field}";
			}

			$rs->set($field_name => $v{$_});
		}
	}

	if ($q->count > 1) {

		# Note: we limit the number of records read from database
		# to prevent an errant query from gobbling up all memory.

		our $Max_Detail_Rows = Acme::Config->Max_Detail_Rows;

		my $recno = 1;
		do {
			for (keys %v) {
				($table, $field) = /(.*?)\.(.*)/
				  or ($table, $field) = ($default_table,$_);

				# Embed record number into field name
				$field_name = "N${recno}_${table}__${field}";

				$rs->set($field_name => $v{$_});
			}
			last if ++$recno > $Max_Detail_Rows;
		} while $q->next(\%v);
	}

	#
	# Return number of rows stored in record set.
	#

	return $q->count;
}

#-------------------------------------------------------------------------------

sub _make_sql {
	my ($t,$args) = @_;

	my $sql = '';	# return value

	#
	# Get contents of table for specified key value
	#

	my (@tables,@fields);
	push @tables, $Primary_Table;
	push @fields, "$Primary_Table.$ID" if $ID;	# ID fields are special
	push @fields, map { "$Primary_Table.$_" } $t->fields;
	if ($t->audit) {
		push @fields,
		  "$Primary_Table." . $t->audit_entry,
		  "$Primary_Table." . $t->audit_update,
		  "$Primary_Table." . $t->audit_user;

		if ($Saving) {

			#
			# If saving a form, add a field that is the difference
			# between the form's render time and the last update
			# time of record. If the difference is negative, then
			# the record was updated since the form was rendered,
			# so a reconciliation screen needs to be generated and
			# displayed to the user before posting the changes.
			#
			# Note: its important for the column alias to be a
			# private name beginning with an underscore '_' for
			# _fill_record_set().
			#

			push @fields,
"(timestamp '$Start') - $Primary_Table." . $t->audit_update . " as _y_${Primary_Table}";
		}
	}

	#
	# Get fields of joined tables.
	#

	my $join_clause = $t->join;
	my $join_expr = new Acme::SQL::Join;
	my $j;
	if ($join_clause) {
		$j = new Acme::Base::Schema::Join $Primary_Table, $join_clause;
		for my $tab ($j->join_tables) {
			my $jt = $Schema->t($tab);
			next unless $jt;
			push @tables, $tab;
			push @fields, "$tab.". $jt->id if $jt->id;
			push @fields, map { "$tab.$_" } $jt->fields;
			if ($jt->audit) {
				push @fields,
					"$tab." . $jt->audit_entry,
					"$tab." . $jt->audit_update,
					"$tab." . $jt->audit_user;

				if ($Saving) {

					# Save the difference between the form
					# start time and the table's last
					# update time. This is the delta time
					# that if negative determines if the
					# table record needs reconcilliation.

					push @fields,
  "(timestamp '$Start') - $tab." . $jt->audit_update . " as _y_${tab}"
				}
			}
		}

		$join_expr->add_join($j->join_expr($Primary_Table));
	}
	else {
		$join_expr->add_join($Primary_Table);
	}

	#
	# Get fields of master tables.
	#

	my $master_clause = $t->master;
	my $m;
	if ($master_clause) {
		$m = new Acme::Base::Schema::Master $Primary_Table, $master_clause;
		for my $tab ($m->master_tables) {
			my $mt = $Schema->t($tab);
			next unless $mt;
			push @tables, $tab;
			push @fields, "$tab.". $mt->id if $mt->id;
			push @fields, map { "$tab.$_" } $mt->fields;
			if ($mt->audit) {
				push @fields,
					"$tab." . $mt->audit_entry,
					"$tab." . $mt->audit_update,
					"$tab." . $mt->audit_user;

				if ($Saving) {

					# Save the difference between the form
					# start time and the table's last
					# update time.

					push @fields,
  "'$Start'::datetime - $tab." . $mt->audit_update . " as _y_${tab}"
				}
			}
		}

		$join_expr->add_join($m->join_expr($Primary_Table));
	}

	#
	# Construct SQL to retrieve database values
	#

	my $keyval = Acme::Db->escape_sql($Keyval);

	$sql  = "select \n" . (join ",\n", @fields);
	$sql .= "\nfrom " . $join_expr->as_string;
	$sql .= "\nwhere $Primary_Table.$Primary_Key = '$keyval'";

	return $sql;
}

#-------------------------------------------------------------------------------

sub _make_audit_link {
	my ($table, $key, $keyval) = @_;

	our $Audit_Prog = Acme::Config->Audit_Prog;

	return '' unless Acme::Security::allow('Access',$Audit_Prog);

	our $Images = Acme::Config->Images;

    	return <<EOF;

<a href="$Audit_Prog?table=$table&$key=$keyval" onClick="new_target(this,'audit')" title="Display audit list of changes to this record"><img src="$Images/audit.gif" border="0" onMouseOver="this.src='$Images/audit_hilite.gif'" onMouseOut="this.src='$Images/audit.gif'"></a>
EOF
}

#-------------------------------------------------------------------------------

sub _auditify {

	# Return date like the first part of the timestamp.

	my $date = shift or return '';
	my $DS = Acme::Config->Date_Separator;
	my ($m,$d,$y) = split(/\W+/,$date);
	return "$y$DS$m$DS$d";
}

#------------------------------------------------------------------------------
#
# Autoload routines
#
#------------------------------------------------------------------------------

our %Subs = (

#-------------------------------------------------------------------------------
'_choose_dropdown' => <<'END_OF_FUNC',

use Acme::TextArray;
use Acme::Form::Assign;
use Acme::Form::Context;

sub _choose_dropdown {
	my ($f, $form_field_name,$default,$size,$attributes,$format) = @_;

	$default = unescape_html($default);

	$format ||= $f->choose_format;
	$format =~ s/^(["'])(.*)\1$/$2/s;	# strip quotes if found

	my $multi = $f->array ? 'multiple' : '';

	my $select_rows = $f->choose_rows;
	$select_rows ||= 4 if $multi;

	#
	# Adjust parameters if query
	#

	if ($Fill_Mode eq 'Query') {

		#
		# Change query dropdown list to accept multiple values
		#

		$multi ||= 'multiple';
		$select_rows ||= 3;
	}

	$attributes .= qq( size="$select_rows") if $select_rows;

	#
	# Render dropdown list
	#

	my $tbl = $f->choose_tables;

	if ($tbl && Acme::HTML::ListMenus->can($tbl)) {
		return Acme::HTML::ListMenus->$tbl(
						$form_field_name,
						$default,
						$size,
						$attributes,
						$format,
						$multi);
	}

	if (my $menu = $f->choose_menu) {
		my @menu_opt = Text::ParseWords::parse_line('\s+',0,$menu);
		return Acme::HTML::ListMenus->Menu(\@menu_opt,
						$form_field_name,
						$default,
						$size,
						$attributes,
						$format);
	}

	if (my $list = $f->choose_list) {
		my %list_opt = Text::ParseWords::parse_line('\s+',0,$list);
		return Acme::HTML::ListMenus->List(\%list_opt,
						$form_field_name,
						$default,
						$size,
						$attributes,
						$format);
	}

	my $sql = $Fill_Mode eq 'Query'
		? $f->choose_query_sql
		: $f->choose_sql;

	if (! $sql) {

		#
		# Return empty list
		#

		# Note: this allows forms to be built without having to have
		# the database tables defined too.

		return Acme::HTML::ListMenus->List([],
					$form_field_name,
					$default,
					$size,
					$attributes,
					$format,
					$multi);
	}

	if (! $tbl) {
		$sql = Acme::Security::clean_sql($sql);

		return Acme::HTML::ListMenus->List(\$sql,
					$form_field_name,
					$default,
					$size,
					$attributes,
					$format,
					$multi);
	}

	#
	# The following is for queries constructed automatically to choose
	# records from another table(s), which is probably the most common case.
	#

	my $t = $f->table;

	#
	# Define assignments
	#

	my $assign_scripts;	# scripts supporting assignments to form fields
	my $edit_assign_scripts;# scripts for edit mode
	my @columns;		# extra columns needed in query for assignments

	($assign_scripts, $edit_assign_scripts, $attributes, @columns) =
	  _choose_dropdown_assignments($form_field_name, $t, $f, $default, $attributes);

	#
	# Parse sql to add needed columns for assignments and to
	# make sure a default record is selected from the table.
	#

	my $sqlo = new Acme::SQL $sql;

	#
	# Add needed columns referenced in assignments to SQL
	#

	for (@columns) {
		$sqlo->add_column($_) if ! $sqlo->is_column($_);
	}

	if ($Fill_Mode ne 'Query') {

		#
		# Add default value to sql query.
		#

		# This is to accommodate the example where a code is "retired"
		# from the select list of a code table, but the user is editing
		# an old record that references the "retired" code. This allows
		# the old code's description to be displayed along with the
		# code, which would be displayed anyway.

		my $st = $Schema->t($tbl);	# search table properties

		$default  ||= $st->select_default;

		my $default_field = $st->code_field || $st->key;
		my $or_is_default = '';
		if ($default && $multi =~ /multi/i) {
			# define default clause for text array defaults
			my $s = (join "', '", Acme::TextArray::split_array($default));
			$or_is_default = qq[ or $default_field in ('$s')] if $s;
		}
		else {
			$or_is_default = qq[ or $default_field = '$default'] if $default;
		}
		$sqlo->add_where($or_is_default) if $or_is_default;
	}

	#
	# Check query with security.
	#

	$sql = Acme::Security::clean_sqlo($sqlo);

	return $assign_scripts . Acme::HTML::ListMenus->List(\$sql,
					$form_field_name,
					$default,
					$size,
					$attributes,
					$format,
					$multi,
					\@columns) . $edit_assign_scripts;
}

sub _choose_dropdown_assignments {
	my ($form_field_name, $t, $f, $default, $attributes) = @_;

	# This is a helper function to _choose_dropdown(). It formulates
	# the JavaScript code that implements assignments to other
	# fields when the dropdown menu list selection changes.
	#
	# The return values is a list of:
	#	* the JavaScript used to setup and initialize the arrays of
	#	  values used to assign to other fields.
	#	* the JavaScript used to initialize fields to the values
	#	  that corresponds to the database value in Edit mode.
	#	* a modified set of attributes for the <select> tag.
	#	* a list of database columns used in the assignments.
	#
	# A side effect is that the Init_Scripts global is added to if needed.

	my $assign_scripts = '';
	my $edit_assign_scripts = '';
	my @columns;
	$attributes ||= '';

	if ($Fill_Mode eq 'Query') {
		return ($assign_scripts, $edit_assign_scripts, $attributes, @columns);
	}

	#
	# Define assignments
	#

	my $choose_assign         = $f->choose_assign;
	my $choose_assign_options = $f->choose_assign_options;
	my $choose_assign_input   = $f->choose_assign_input;
	my $choose_next	          = $f->choose_next;

	if ($choose_assign || $choose_assign_options || $choose_assign_input || $choose_next) {
		my $context = new Acme::Form::Context table=>$t->name;

		my $asgn = new Acme::Form::Assign
			form	     => 'f',	# synonym for current form
			result	     => 'r',	# synonym for search results
			search_table => $f->choose_table, # table to search
			t	     => $t,	# table object of current form
			context	     => $context;

		my $assignments = '';
		$assignments .= $asgn->assign_options( $choose_assign_options );
		$assignments .= $asgn->assign_fields(  $choose_assign         );
		$assignments .= $asgn->assign_input(   $choose_assign_input   );

		# Make hash of sql columns used in all the assignments to
		# eliminate duplicates.
		my %columns = map { $_, 1 } $asgn->columns;
		@columns = keys %columns;

		#
		# Set focus to next field specified by Choose_next.
		#

		my $focus_next_field = '';
		if ($choose_next) {
			# Get field to go to after a successful search.
			my $frm = qq(self.document.forms["$N_Form"]);
			$focus_next_field = ($choose_next =~ /^[%\$]/)
			  ? $context->translate($choose_next, form=>$frm)
			  : "$frm.".$context->db_to_form($choose_next);
			$focus_next_field .= ".focus()";

			# Need to wrap focus() within setTimeout() in order
			# to get out of event handler before setting focus().
			# Otherwise, effort is thwarted as focus() is reset
			# to source field till onChange event handler finishes.

			$focus_next_field = "setTimeout('$focus_next_field',5);";
			$focus_next_field .= "\n\t\t";	# pretty-print
		}

		#
		# Define support scripts of assignments
		#

		$assignments =~ s/\n/\n\t\t/g;	# pretty-print

		$assign_scripts = <<End_Of_Javascript if $assignments || $choose_next;
/*
 * Choose list assignments and next field
 */

function _choose_onchange_${form_field_name}(me) {
	var f = me.form;
	if (f) {
		var r;
		if (self._result_${form_field_name})
			r = _result_${form_field_name}\[me.selectedIndex];

		if (r) {
			// Assign selection values to fields.
			${assignments}
		}
		// Set focus to next field here.
		${focus_next_field}return true;
	}
	return false;
}
End_Of_Javascript

		#
		# Change attributes to include an onChange event to
		# call the function used to assign values to fields.
		#

		my $onchange=escape_html(
qq[if (! _choose_onchange_${form_field_name}(this)){return false}]
					);

		$attributes ||= '';
		$attributes =~
		  s{\b(onchange\s*=\s*(['"]))(.*?)\2}
		   {$1$onchange$3$2}si
		  or $attributes .= qq[ onChange="$onchange"];

		#
		# Edit mode requires initialization of display-only fields.
		#

		if ($Fill_Mode eq 'Edit' || ($Fill_Mode eq 'New' && $default)) {

			my $edit_assignments = '';
			$edit_assignments .=
			  $asgn->assign_options( $choose_assign_options );
			$edit_assignments .=
			  $asgn->assign_display( $choose_assign         );
			$edit_assignments .=
			  $asgn->assign_input(   $choose_assign_input   );

			if ($edit_assignments) {
				#
				# Define support scripts of assignments
				#

				# Pretty-print formatting
				$edit_assignments =~ s/^/\t\t/;
				chop $edit_assignments;

				$edit_assign_scripts .= <<End_Of_Javascript;
function _choose_init_${form_field_name}(r) {
	// Initialize fields from choose results
	var f = self.document.forms['$N_Form'];
	if (r) {
		// Assign selection values to fields.
$edit_assignments
	}
}
End_Of_Javascript

			$Init_Scripts .= <<End_Of_Javascript;
_choose_init_${form_field_name}(_result_${form_field_name}\[self.document.forms['$N_Form'].${form_field_name}.selectedIndex])
End_Of_Javascript
			}
		}

		if ($assign_scripts) {
			$assign_scripts =
	  qq[<script language="JavaScript" type="text/javascript">\n$assign_scripts</script>];
		}

		if ($edit_assign_scripts) {
			$edit_assign_scripts =
	  qq[<script language="JavaScript" type="text/javascript">\n$edit_assign_scripts</script>];
		}
	}

	return ($assign_scripts, $edit_assign_scripts, $attributes, @columns);
}
END_OF_FUNC

#-------------------------------------------------------------------------------
'_do_search' => <<'END_OF_FUNC',

use Acme::Form::Assign;

sub _do_search {
	my ($f, $field_name, $value) = @_;

	# Do search and initialize display-only fields and select options.

	# This is very similar to the _search_button() routine in the
	# Form Builder. Some differences are:
	#
	#  * Builder sets up search. Here, the result value is already known.
	#  * After assignment routines are built, search is done immediately.
	#  * Only display-only fields of search_assign are processed in Edit mode.
	#  * All fields of search_assign are procesed if a New form.
	#  * Both search_assign and choose_assign properties are used here.
	#
	# The methodology here is to add JavaScript to the form that 
	# has a data structure holding the form field values. It's a data
	# structure very similar to the result data structure of a Search
	# window when it returns values to the calling form. The Search
	# button's return function is then called when the form is initialized
	# to fill in the form fields with the looked up values.

	my $t = $f->table;

	my $search_table = $f->search_table or $f->choose_table or return;
	my $sql          = $f->search_sql   or $f->choose_sql   or return;

	#
	# Initialize form context.
	#

	require Acme::Form::Context;
	my $context = new Acme::Form::Context table=>$t->name;

	#
	# Define assignments
	#

	my $asgn = new Acme::Form::Assign
		form	     => 'f',	# form name to use for target fields
		result	     => 'r',	# object name for result (form) fields
		search_table => $search_table,# table name to search
					# to produce the result
		t	     => $t,	# AcmeBase schema object for the table
					# whose form fields are being filled.
					# Not the table being searched.
		context	     => $context;# Acme::Form::Context object which 
					# controls translation

	my $assignments = '';
	$assignments .= $asgn->assign_options(
					$f->search_assign_options ||
					$f->choose_assign_options
					);
	if ($Fill_Mode eq 'New') {
		# Fill default values from search
		$assignments .= $asgn->assign_fields(
						$f->search_assign ||
						$f->choose_assign
						);
	}
	else {
		# Assign display-only fields only
		$assignments .= $asgn->assign_display(
						$f->search_assign ||
						$f->choose_assign
						);
	}
	$assignments .= $asgn->assign_input(
					$f->search_assign_input ||
					$f->choose_assign_input
					);

	return unless $assignments;
	
	# Note: make sure the fields in the search_assign strings below are
	# in search_sql Often the search_sql is derived from the target
	# table's select property, which normally only selects the fields
	# which are displayed in the list box.

	# Parse $sql and add needed columns to query.
	my $sqlo = new Acme::SQL $sql;

	# Make hash of columns to eliminate duplicates.
	my %columns = map { $_, 1 } $asgn->columns;

	for (keys %columns) {
		$sqlo->add_column($_) if ! $sqlo->is_column($_);
	}

	#
	# Define initial assignment statement for assigning
	# a value to the search field.
	# Assign key value of search table to current form field
	# if the search is of a database table.
	#

	my ($assign_tab, $assign_fld);
	if ($assign_tab = $search_table) {
		my $st = $t->schema->t($assign_tab);
		$assign_fld = $st->key;
	}
	else {
		# Use first column of the query as the column
		# to assign to the search field
		my ($col) = $sqlo->columns;
		($assign_tab, $assign_fld) = $col =~ /(\w+?)\.(\w+)/
		  or do {
			($assign_tab) = $sqlo->tables;	# first table
			$assign_fld = $col;
		}
	}
	my $search_script = "$assign_tab.$assign_fld = '$value'";

	#
	# Do search script
	#

	$sqlo->add_where($search_script);
	$sql = $sqlo->as_sql;

	# Execute search query

	my %result;
	Acme::Db->get($sql,\%result) or return;

	#
	# Define object for search results
	#

	my $scripts = <<End_Of_Script;

/*
 * Search initialization for $field_name
 */

var result_${field_name} = new Object;
End_Of_Script

	my $rf;
	for (keys %columns) {
		($rf = $_) =~ s/\W/__/g;

		$scripts .= qq[result_${field_name}]
			  . qq[['$rf'] = {'value': unescape(']
			  . escape($result{$_})
			  . qq[')};\n];
	}

	# Define Search support scripts

	$scripts .= <<End_Of_Script;

function _search_result_${field_name}(r) {
	// Initialize fields from search results
	var f = self.document.forms['$N_Form'];
	if (r) {
		// Assign selection values to fields.
$assignments
	}
}

_search_result_${field_name}(result_${field_name});
End_Of_Script
	
	$Init_Scripts .= $scripts;

}
END_OF_FUNC

#-------------------------------------------------------------------------------
'_detail_list' => <<'END_OF_FUNC',
use Acme::Lister;
use Acme::SQL;

our $No_Tabindex = Acme::Config->No_Tabindex;
our $Bullet      = Acme::Config->Bullet;


sub _detail_list {
	my ($command) = @_;

	#
	# Construct sql for list of detail records.
	#

	my ($field_name,$detail_table,$detail_field,$table,$field) =
	  $command =~ /(\w+)\s+(\w+)\.(\w+)\s*=\s*(\w+)\.(\w+)/;

	my $schema = $Schema;		# make visible to enclosed sub's
	my $t = $schema->t($table) or die ("Unknown table: $table");
	my $f = $t->f($field);

	my $dt = $schema->t($detail_table);
	my $key = $dt->key;
	my $edit_prog = $dt->edit_prog;

	# Define partial url (sans key value) for editing detail record.
	my $url = "$edit_prog?table=$detail_table&master=$field_name&$key=";

	my $obj = ($f && $f->detail) ? $f : $t;

	my $list_fields = $obj->detail_list_fields;
	my $list_format = $obj->detail_format;
	my $list_header = $obj->detail_header;
	my $list_footer = $obj->detail_footer;
	$list_header = $list_format if $list_header eq 'yes';
	$list_footer = '' 	    if $list_footer eq 'yes';	# no default
	my @total_columns = split /\s*,\s*/, $obj->detail_total_fields;

	#
	# Define assignments
	#

	my $assignments = '';

	my @columns;		# extra sql columns needed for assignments
	my $assign_scripts = '';# scripts actually assigning detail values 

	my $detail_assign         = $obj->detail_assign;
	my $detail_assign_options = $obj->detail_assign_options;
	my $detail_assign_input   = $obj->detail_assign_input;

	if ($detail_assign || $detail_assign_options || $detail_assign_input) {
		my $context = new Acme::Form::Context table=>$t->name;

		my $asgn = new Acme::Form::Assign
			form	     => 'f',	# synonym for current form
			result	     => 'd',	# synonym for search results
			search_table => $detail_table,	# table to search
			t	     => $t,	# table object of current form
			context	     => $context;

		$assignments .= $asgn->assign_options( $detail_assign_options );
		$assignments .= $asgn->assign_fields(  $detail_assign         );
		$assignments .= $asgn->assign_input(   $detail_assign_input   );

		$assignments =~ s/\n/\n\t\t/g;	# pretty-print formatting

		$assign_scripts = <<End_Of_Javascript if $assignments;
<script language="JavaScript" type="text/javascript">
function _detail_assign_${field_name}(f,d) {
	if (f && d) {
		// Assign detail values to fields.
		${assignments}return true;
	}
	return false;
}
</script>
End_Of_Javascript

		# Make hash of sql columns used in all the assignments to
		# eliminate duplicates.
		my %columns = map { $_, 1 } $asgn->columns;
		@columns = keys %columns;
	}

	#
	# Process SQL query for detail records.
	#

	my $sqlo = new Acme::SQL $obj->detail_sql;

	$sqlo->add_where("${detail_table}.${field} = '"
	  . Acme::Db->escape_sql(_db_value($field_name)) . "'");

	# Assure detail table key field is in query
	$sqlo->add_column("$detail_table.$key")
	  if ! $sqlo->is_column("$detail_table.$key");

	# Add needed columns referenced in assignments to SQL
	for (@columns) {
		$sqlo->add_column($_) if ! $sqlo->is_column($_);
	}

	my $detail_sql = Acme::Security::clean_sqlo($sqlo);

	my $has_edit_priv = Acme::Security::allow('Update',
						  $detail_table,
						  $dt->secure,
						  $dt->restrict);
	#
	# Define list object
	#

	my $l = new Acme::Lister sql => $detail_sql, total => \@total_columns;

	#
	# Define helper functions to list detail items.
	#

	my $row_name = $dt->row_name;

	my $before_list = sub {
		package Acme::Lister::Sandbox;

		#
		# Compose list header
		#

		my $header = Acme::Form::Filler::_detail_list_header(
					$list_header,
					$list_fields,
					$detail_table,
					$schema,
					$dt);

		if ($header) {
			# Surround $header with <td> </td> tags if not already.
			if ($header !~ m|^\s*<t[dh]|i
			||  $header =~ s|^\s*<td>(.*)</td>\s*$|$1|i) {
				$header = qq[<th>$header</th>];
			}

			# Prepend a blank column used for edit bullets in list.
			$header = qq[<tr class="detail-list-header">\n<td></td>$header\n</tr>];
		}

		#
		# Insert a blank row to use as a blank example row so JavaScript
		# can add new rows to the form. This version makes the row
		# invisible.
		#

		my $title  = "Edit $row_name.";

		# Format row and evaluate embedded functions
		my $blank_row = Local::Acme::Format::sprintf_cmd($list_format);

		# Surround $blank_row with <td> </td> tags if not already.
		if ($blank_row !~ m|^\s*<td|i
		||  $blank_row =~ s|^\s*<td>(.*)</td>\s*$|$1|i) {
			$blank_row = qq[<td class="detail-item">$blank_row</td>];
		}

		my $attr = Acme::Form::Filler::_detail_list_attr($field_name,$url);

		my $edit_bullet = $has_edit_priv
		  ? qq[<a href="javascript:void 0" title="$title" class="edit-bullet" tabindex="$No_Tabindex">$Bullet</a>]
		  : $Bullet;

		return <<End_Of_HTML;
<script language="JavaScript" type="text/javascript">_detail_url_${field_name}="$url"</script>
<table class="detail-list" cellspacing=0>
<tbody id="_detail_table_${field_name}" class="detail-body">
<tr id="_detail_blank_${field_name}" class="detail-row" style="display:none"$attr>
<td class="detail-link">$edit_bullet</td>
$blank_row
</tr>
$header
End_Of_HTML
	};

	my $after_list = sub {
		package Acme::Lister::Sandbox;

		#
		# Compose list footer
		#

		my $footer = Acme::Form::Filler::_detail_list_footer(
					$list_footer,
					$list_fields,
					$detail_table,
					$schema,
					$dt);

		if ($footer) {
			# Surround $footer with <td> </td> tags if not already.
			if ($footer !~ m|^\s*<td|i
			||  $footer =~ s|^\s*<td>(.*)</td>\s*$|$1|i) {
				$footer = qq[<td>$footer</td>];
			}

			# Prepend a blank column used for edit bullets in list.
			$footer = qq[<tr class="detail-list-footer">\n<td></td>$footer\n</tr>];
		}

		 $footer = Local::Acme::Format::sprintf_cmd($footer) if $footer;

		 return <<End_Of_HTML;
$footer
</tbody>
</table>
$assign_scripts
End_Of_HTML
	};

	my $before_row = sub {
		package Acme::Lister::Sandbox;
		no strict 'refs';

		my $k = "${detail_table}__${key}";
		my $keyval = $$k;

		my $href  = "javascript:_detail_open_${field_name}('$url','$keyval')";

		my $title  = "Edit $row_name.";

		my $attr   = '';
		$attr = Acme::Form::Filler::_detail_list_attr($field_name,$url)
		  if $has_edit_priv;

		# The edit bullet does not have to be a link as the <tr> tag
		# has an onClick event that edits the record. Defining the
		# edit bullet as a link causes 2 edit windows to open (the
		# <tr> event does not cancel the event bubble).

		#my $edit_bullet = $has_edit_priv
		#  ? qq[<a href="$href" title="$title" class="edit-bullet" tabindex="$No_Tabindex">$Bullet</a>]
		#  : $Bullet;
		my $edit_bullet = $Bullet;

		return <<End_Of_HTML;
<tr id="_detail_${field_name}_${keyval}" class="detail-row"$attr>
<td class="detail-link">$edit_bullet</td>
End_Of_HTML
	};

	my $after_row = sub {
		package Acme::Lister::Sandbox;
		return <<End_Of_HTML;

</tr>
End_Of_HTML
	};

	my $every_row = sub {
		package Acme::Lister::Sandbox;
		our @a;		# array of database values for record

		# Format row and evaluate embedded functions
		my $row = Local::Acme::Format::sprintf_cmd($list_format,@a);

		# Surround $row with <td> </td> tags if not already surrounded.
		$row = qq[<td class="detail-item">$row</td>\n]
		  if $row !~ m|^\s*<td|i;

		return $row
	};

	#
	# Construct List
	#

	return
	    &$before_list()
	  . $l->list(	before_row => $before_row,
			after_row  => $after_row,
			every_row  => $every_row,
			)
	 . &$after_list();
}

sub _detail_list_header {
	my ($format,$fields,$table,$schema,$t) = @_;

	# Helper function for building a list header from field names.

	$format ||= '';
	return $format unless $format =~ /\%/;	# not a sprintf() format

	# Merge field titles into format string.

	my ($tab,$fld,$tt,@headers);
	for my $h (split(/\s*,\s*/,$fields)) {
		($tab,$fld) = $h =~ /(\w+)\.(\w+)/ 
		  or ($tab,$fld) = ($table,$h);

		if ($tab eq $table) { $tt = $t; }
		else                { $tt = $schema->t($tab); }

		if ($tt && $tt->{$fld}) {
			push @headers, $tt->{$fld}->title;
		}
		else {
			# Capitalize field name
			# (happens on ID fields)
			$fld =~ s/_/ /g;
			$fld =~ s/\b(\w+)\b/ucfirst(lc($1))/eg;
			push @headers, $fld;
		}
	}

	# Format header and evaluate embedded functions
	return Local::Acme::Format::sprintf_cmd($format,@headers);
}

sub _detail_list_footer {
	my ($format,$fields,$table,$schema,$t) = @_;

	# Helper function for building a list footer.

	# Note: There is no default footer at this time.

	return $format || '';
}

sub _detail_list_attr {
	my ($field_name,$url) = @_;
	
	# Helper function for building <tr> attributes for detail rows.
	# This has the side effect of modifying $Init_Scripts.

	# To turn off, simply return.
	#return '';

	if (! our $detail_script) {
		$detail_script = <<'End_Of_JavaScript';

/*
 * Detail row functions
 */

function hi(row) {
	// Uncomment next to switch class attributes instead of changing styles.
	//if (row) row.className='detail-row-hilite';

	// Use style settings to implement hilites. This is faster in IE5.
	if (row) {
		var rs = row.style;
		rs.color='FF00FF';
		rs.backgroundColor='DDDDDD';
		rs.cursor='hand';
	}
}
function lo(row) {
	// Uncomment next to switch class attributes instead of changing styles.
	//if (row) className='detail-row';

	// Use style settings to implement hilites. This is faster in IE5.
	if (row) {
		var rs = row.style;
		rs.color = "black";
		rs.backgroundColor = "transparent";
		rs.cursor = 'auto';
	}
}
function go(row) {

	row.id.match(/^_detail_(.*)_([^_]+)$/);
	var field_name = RegExp.$1;
	var keyval     = RegExp.$2;
	if (field_name && keyval) {
		eval("_detail_open_"+field_name+"(_detail_url_"+field_name+",keyval)");
	}
}
End_Of_JavaScript

		# Append to global scripts that gets substituted in the header.
		# This only happens once.

		$Init_Scripts .= $detail_script;
	}

	return qq[ onMouseOver="hi(this)" onMouseOut="lo(this)" onClick="go(this)"];
}
END_OF_FUNC

#-------------------------------------------------------------------------------
'_fill_query_form' => <<'END_OF_FUNC',

sub _fill_query_form {
	my ($self, $t, $args, $html) = @_;

	$Fill_Mode = 'Query';		# set global mode

	#
	# Substitute commands.
	#

	my $do = $t->do_query;
	if ($do) {

		#
		# Evaluate do_query property's Perl script.
		# Subroutines defined in do_filter can be invoked
		# with the "Do" form query command, like:
		#
		#  <!--$ do subroutine($arg1,$arg2,$arg3, ...) -->
		#

		# Note: The do_query property subverts Perl's
		# taint checks as the schema (and form) files
		# are assumed to be protected files.
	
		package Acme::Form::DbVars;
		no strict;
		($do) = $do =~ /^(.*)$/s;	# untaint
		eval $do;
		die "Error evaluating schema's Do property: $@" if $@;
	}

	$html =~ s/(<!--+\$+\s*.*?\s*--+>)/_tag_command($1,$self->{extend})/egs;

	#
	# Substitute <input> tags.
	#

	$html =~ s|(<input\s.*?>)|_tag_input($1,$args)|egsi;

	#
	# Substitute <textarea> tags.
	#

	$html =~ s|(<textarea.*?>.*?</textarea>)|_tag_textarea($1,$args)|egsi;

	return $html;
}
END_OF_FUNC
);

#------------------------------------------------------------------------------

sub AUTOLOAD {
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;
	return if $method eq 'DESTROY';		# gobble
	if (exists $Subs{$method}) {
		my $s = $Subs{$method};
		delete $Subs{$method};		# release memory

		if ($s) {
			eval "package $pkg; $s";
			die "${pkg}::$method(): $@" if $@;
			goto &$method;
		}
	}
	return '';				# unknown method
}

1;
