package Acme::Form::Query;

=head1 NAME

Acme::Form::Query - Query information from a database.

=head1 SYNOPSIS

    use Acme::Form;
    use Acme::Form::Query;
    our @ISA = qw(Acme::Form::Query);

    my $self = new Acme::Form;

    $search = $self->query;
      -or-
    ($search, $list_title) = $self->query;

=head1 DESCRIPTION

I<Acme::Form::Query> is a subclass of I<Acme::Form>. It provides a method to
change CGI arguments into an AcmeBase search script. The search script can then
be used by the C<list.cgi> program to perform the query and list the results.

The query() method reads the CGI parameters of numbers, date ranges, and
text patterns. The CGI parameter names are form field names corresponding to
database columns, like "table__column". Names beginning with "_" are ignored
as are "request", "title", "list_title", "list_exclude", "report", and "table",
which are common option parameters specified in the URL.


=head1 CHANGE LOG

 6/17/00  Initial version.
 1/17/02  Added list_exclude to exclude rows saved in a list file from a
	  previous query.
 6/17/02  Repackaged into a method of Acme::Form. Added multiple values
          per field processing.
 6/22/02  Excluded strings within ' quotes from test for operators.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000-2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

use Acme::HTTP		qw(escape);
use Acme::HTML		qw(debug	debug_on);
use Acme::Db;
use Acme::Base::Schema;
use Acme::Base::Schema::Join;
use CGI;

use Acme::Config;

our $DS = Acme::Config->Date_Separator;
our $SS = Acme::Config->Script_Separator || ';';   # either ; or |

#-------------------------------------------------------------------------------
#
#  Query public methods
#
#-------------------------------------------------------------------------------

sub query {
	my $self = shift;
	my %opt = @_;		# rest of arguments are key/value options

	# Uncomment the following to turn on debug output.
	#debug_on();				# turn on debug
	#print "Content-type: text/html\n\n";	# turn on output

	my $args         = $opt{args}			||
			   $self->{args}		||
			   $self->{cgi}			||
			   new CGI;
	$args = new CGI $args if ref($args) eq 'HASH';

	my $table        = $opt{table}			||
			   $self->{table}		||
			   $args->param('_table')	||
			   $args->param('table')	||
			   '';
	my $title        = $opt{title}			||
			   $args->param('title')	||
			   $self->{title}		||
			   '';
	my $list_title   = $opt{list_title}		||
			   $args->param('list_title')	||
			   $self->{list_title}		||
			   '';
	my $list_exclude = $opt{list_exclude}		||
			   $args->param('list_exclude')	||
			   $self->{list_exclude}	||
			   '';
	#
	# Start constructing search script
	#

	my $schema = new Acme::Base::Schema;

	my $t = $schema->t($table)
	  or die "Cannot find a database table named '$table'\n";

	my $key = $t->key;

	$title      ||= $t->title;
	$list_title ||= $t->list_title;

	my $where_clause   = '';	
	my $exclude_script = '';
	my $join_script    = '';
	my %jointables;

	#
	# Define set of operators recognized in query parameters.
	#

	# Standard SQL operators
	my @operators = ('=','<','>','<=','>=','<>','!=','between','and','is null');
		
	# Add database dependent string pattern operators
	my $dbengine = $schema->dbengine;

	my $matches;		# operator for string pattern matches
	my $match_pat;		# regular expression that matches match operator
	my $anchor;		# a '^' character or empty to anchor search pattern

	if ($dbengine =~ /Postgres/i) {
		$matches   = '~*';
		$match_pat = qr/\~\*?/;
		$anchor    = '^';

		push @operators, '~', '~*';
	}
	elsif ($dbengine =~ /MySQL/i) {
		$matches = 'regexp';
		$match_pat = qr/\bregexp\b/i;
		$anchor    = '^';

		push @operators, 'regexp';
	}
	elsif ($dbengine =~ /Informix/i) {
		$matches = 'matches';
		$match_pat = qr/\bmatches\b/i;
		$anchor    = '';

		push @operators, 'matches';
	}

	#
	# Process the non-empty CGI parameters.
	#

	my ($k, $v, $s, $c, $arg, @args);
	my ($recno, $tbl, $fld);
	my ($is_date);

	foreach $k ($args->param()) {

		#
		# Ignore keywords and private variable fields (those beginning
		# with _).
		#

		debug("Checking parameter '$k'");

		next if grep {$k =~ /^$_$/i}
			'_.*',
			'[A-Z]_.*',
			'title',
			'list_title',
			'list_exclude',
			'table',
			'report',
			'request';

		#
		# Parse field name
		#

		($recno, $tbl, $fld) =
		  $k =~ /^(?:N(\d+)_)?		# optional N12_ prefix
			  (?:(\w*?)__)?		# optional table name
			  (\w*)			# required field name
			/x;

		$tbl ||= $table;

		my $qt = $schema->t($tbl);
		my $f = $qt->f($fld);

		#
		# Look for fields that are not empty.
		#

		@args = $args->param($k);

		debug("Process arguments for $k:", \@args);

		if (@args > 1) {
			
			#
			# Multiple settings for an argument implies that each
			# value is a fixed value (not a pattern).
			#

			$s = '';
			if ($f->array) {

				#
				# Make where clause for matching a text array.
				#

				$c = $f->array_separator; 

				for (@args) {
					$s .= " or " if $s;
					$s .= "$tbl.$fld $matches '$c$_$c'";
				}

				debug("Parameter is array, s=$s");
			}
			else {
				#
				# Make where clause for matching text values.
				#

				if ($f->is_number) {
					# No quotes around values needed
					$s = join ', ', @args;
				}
				else {
					# Put quotes around values
					$s = "'" . (join "', '", @args) . "'";
				}
				$s = "$tbl.$fld in ( $s )";

				debug("Parameter is multi-value list, s=$s");
			}


			$where_clause .= " and ( $s )";

			#
			# Track tables used in query.
			#

			$jointables{$tbl}++ if $tbl;

			next;
		}

		$arg = $args[0];

		if (length $arg > 0) {

			#
			# Single value for an argument.
			#

			#
			# Setup query where clause
			#

			if ($dbengine =~ /Postgres/i) {
				# Following is for those users migrating from
				# Informix to Postgres who forget to put the
				# leading period for patterns to 'Match All'.

				$arg =~ s/^\*/.*/;		# match begin *
				$arg =~ s/([^.\\])\*$/$1.*/;	# match final *

				# Anchor pattern match to beginning of string
				# by default to speed up string searches.

				$arg = "^$arg"
				  if $f && $f->is_text && $arg =~ /^\w/
				  && ($arg =~ /[?*+[]/ and $v !~ /$match_pat/);	# pattern 
			}

			debug("Parameter for $k is $arg");

			#
			# Set flag if field is a date field.
			# Account for audit fields do not have a $f structure.
			#

			$is_date = 0;
			$is_date = 1
			  if (($f && $f->is_date)
			  || $fld eq $t->audit_entry
			  || $fld eq $t->audit_update);

			#
			# Split value to account for OR conditions such as "=2 | >10"
			#

			$s = '';
			for $v (split /\|/, $arg) {

				debug("Condition fragment: '$v'");

				#
				# Transform special cases
				#

				if ($is_date) {

					# Change range operator (:)

					$v =~ s/  # Change '1-1-2000':'1-31-2000'
						(['"])	# quote
						(.*?)	# first date
						\1	# matching quote
						\s*:\s*	# : range delimiter
						(['"])	# quote
						(.*?)	# second date
						\3	# matching quote
					      /between '$2' and '$4'/x
					or
					$v =~ s/  # Change 1-1-2000:1-31-2000
						(\d\d?\W\d\d?\W\d\d(?:\d\d)?)
						\s*:\s* # : range delimiter
						(\d\d?\W\d\d?\W\d\d(?:\d\d)?)
					      /between '$1' and '$2'/x
					or
					$v =~ s/  # Change 010100:013100
						(\d\d)(\d\d)(\d\d(?:\d\d)?)
						\s*:\s*
						(\d\d)(\d\d)(\d\d(?:\d\d)?)
					      /between '$1$DS$2$DS$3' and '$4$DS$5$DS$6'/x;
				}
				elsif ($f && $f->is_number) {

					# Change range operator (:)

					$v =~ s/  # Change -12345:67890
						([-+]?\d+(?:\.\d*)?(?:[eE][-+]?\d+)?)
						\s*:\s*
						([-+]?\d+(?:\.\d*)?(?:[eE][-+]?\d+)?)
					      /between $1 and $2/x;
				}

				# Setup a string pattern match.
				$v =~ s/(.*)/"$matches '" . Acme::Db->escape_sql($1) . "'"/e
				  if ($v =~ /[?*+[]/ and $v !~ /$match_pat/);	# pattern 

				# Change a single = equal sign into "is null".
				$v = "$tbl.$fld is null" if ($v =~ /^\s*=\s*$/);

				# Change a "= null" into "is null".
				$v = "$tbl.$fld is null" if ($v =~ /^\s*=\s*null\s*$/i);

				#
				# Test if an operator was specified
				#

				debug("Checking for operator in v: $v");

				# Exclude strings within ' quotes from test by
				# splitting into array on quoted strings

				my $op = 0;	# false
				my @vp = split /('[^']*')/, $v;

				my $vcheck;
				my $count = 0;
				for (@vp) {
					next if $count++ % 2;	# skip quoted strings

					debug("Checking fragment for operator: $_");

					$vcheck = $_;

					$op ||= grep { $vcheck =~ /\Q$_\E/i } @operators;

					debug("Operator found") if $op;

					# Substitute $ with column name
					s/\$/$tbl.$fld/g;

					if ($is_date) {
						# Change mmddyy numbers to dates
						s{\b(\d\d)(\d\d)(\d\d(?:\d\d)?)\b}
						 {$1$DS$2$DS$3}g;

						# Quote dates that are not quoted.
						s{([^"'\d])(\d\d?(\W)\d\d?\3\d\d(\d\d)?)\b}
						 {$1'$2'}g
					}

					debug("Modified fragment: $_");
				}

				# Reassemble from operators and quoted strings

				$v = join '', @vp;	# reassemble $v

				debug("Reassembled v: $v");

				#
				# Quote entire string
				#

				$v = "'" . Acme::Db->escape_sql($v) . "'"
				  if (! $op && $v !~ /['"]/);	# quote'ify

				#
				# Assemble where clause: variable op value
				#

				# Supply default operator =

				$v = "= $v" unless $op;

				# Prepend column name if needed

				my ($firstword) = $v =~ /^\s*([^\w\s']+)/;

				$firstword ||= 'between'  if $v =~ /^\s*between\b/i;
				$firstword ||= 'is null'  if $v =~ /^\s*is null\b/i;
				$firstword ||= "$matches" if $v =~ /^\s*$match_pat\b/i;

				debug("Checking if first word is an operator:".($firstword||''));

				$v = "$tbl.$fld $v"
				  if $firstword
				  && grep { $firstword =~ /$_/ } @operators;


				#-------------------
				#
				#### Example from prototype:
				#### Special processing of display-only fields.
				#### For future consideration (?) ####
				#
				# Special processing of display-only fields.
				# Display only fields have names starting with "D_".
				# The field name is substituted with their formula
				# using real database table fields.
				#
				#
				#SPECIAL: {
				# Example:
				#	if ($k eq "D_disbursement_total") {
				#		# Change key name into equivalent expr
				#		$k =~ s{D_disbursement_total}
				#			{(reservations.deposit + reservations.final_pmt)}g;
				#
				#		# Hack: prepend table name to $k for
				#		# %jointables, see below
				#		$k = "reservations." . $k;
				#		last SPECIAL;
				#	}
				#}
				#-------------------

				
				$s .= " or " if $s;
				$s .= $v;

				debug("Assembled query fragment: s=$s");
			}

			$where_clause .= " and ( $s )";

			#
			# Track tables used in query.
			#

			# the table name is field name up to the period
			$jointables{$tbl}++;
		}
	}

	#
	# Construct join_script.
	#

	my ($jt, $join_clause);
	if (my $jt = new Acme::Base::Schema::Join $table, $t->join) {
		foreach $k (keys %jointables) {
			if ($join_clause = $jt->join_clause($k)) {
				$join_script .= qq(${SS}table_join("$table inner join $k on $join_clause"));
			}
		}
	}

	debug("jointables:",\%jointables,"Assembled join clause: $join_script");

	#-------------------
	#
	#### Example from prototype:
	#### Dependent conditions that need to be accommodated (?)  #####
	#### Three tables: customers -> trips -> reservations
	#
	# All tables are assumed to join customers.customer_id with a field named
	# customer_id.
	#
	#delete $jointables{customers} if (exists $jointables{customers});
	#
	# Assure reservations join trips table if a reservation field specified.
	#$jointables{trips}++ if ($jointables{reservations});
	#
	#foreach $k (keys %jointables) {
	#	$join_script .= 
	#	  "${SS}inner_join(\"$k\",\"customers.customer_id = $k.customer_id\")";
	#}
	#
	# special case, fields from both reservations and trips specified
	#$where_clause .= " and reservations.trip_id = trips.trip_id"
	#  if (exists $jointables{trips} and exists $jointables{reservations});
	#--------------------


	#
	# Construct search script.
	#

	$where_clause =~ s/^ and //;	# remove initial " and " construct

	$exclude_script = "${SS}list_exclude('$list_exclude','$table.$key')" if $list_exclude;

	my $search = "$title$exclude_script$join_script${SS}$where_clause";

	debug("Assembled search script: $search");

	$search = escape($search);

	return $search if ! wantarray;

	#
	# Construct title for query if one not specified.
	#

	unless ($list_title) {
		# Make list header from where clause
		$list_title = $where_clause;
		$list_title =~ s/\b\w+?\.//g;	# remove table names
		$list_title =~ s/^\s*\(//;	# remove beginning (
		$list_title =~ s/\)\s*$//;	# remove trailing (
		$list_title = $t->title . " where $list_title";
	}
	$list_title = escape($list_title);

	return ($search,$list_title);
}

1;
