package Acme::Form::Reconcile;

=head1 NAME

Acme::Form::Reconcile - Construct a reconcilliation form

=head1 SYNOPSIS

    use Acme::Form::Reconcile;

    Acme::Form::Reconcile::display_reconcile_form($Form_RS, $t, $args, $nrec);

=head1 DESCRIPTION

This is a utility package for I<Acme::Form::Saver> module. It was intended
to be dynamically loaded only when needed by the Form Saver. The
display_reconcile_form() function displays a reconcilliation form allowing
the user to select the changes they made on their original form or choose
the existing database values which were entered or updated by somebody else
since the user's form was rendered.

=head1 CHANGE LOG

 6/05/02  Initial version.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

use Acme::HTML			qw(escape_html);
use Acme::Form::Context;
use Acme::Config;

require Exporter;

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(display_reconcile_form);

our $Scripts_Lib = Acme::Config->Scripts_Lib;

sub display_reconcile_form {
	my ($Form_RS, $t, $args, $nrec) = @_;

	#
	# Display a screen with the form values and the current database
	# values and let the user choose which values to save.
	# This is performed for only one record of a table at a time.
	# This routine does not return and exits the program.
	#

	my $table = $t->name;
	my $key   = $t->key;

	#
	# Get the name of the user who last updated the record.
	#

	my $audit_user = $t->name . '__' . $t->audit_user;
	my $user = '';
	if (! $nrec) {
		$user = $Form_RS->get($audit_user);
	}
	elsif ($nrec > 1) {
		$audit_user = "N${nrec}_$audit_user";
		$user = $Form_RS->get($audit_user);
	}
	else {	# $nrec == 1
		# First try finding user without the number prefix.
		$user = $Form_RS->get($audit_user);
		if (! $user) {
			# Try again with numbered record.
			$audit_user = "N${nrec}_$audit_user";
			$user = $Form_RS->get($audit_user);
		}
	}

	my $html = <<End_Of_HTML;
<p>The database information was changed by "$user" since you started editing the $table table. Please indicate which values should be saved to the database: the changes you made or the current database values.</p>

<table class="form">
End_Of_HTML

	my $script = <<End_Of_JavaScript;
<script src="$Scripts_Lib/form_save.js" type="text/javascript"></script>
<script language="JavaScript" type="text/javascript">
/*
 * Reconciliation works by knowing the Form Saver ignores any fields whose name
 * begins with a capital letter and a _ character (e.g. the pattern /[A-Z]_(.*)/).
 * Selecting an answer changes a field's name to a name recognized by the Saver
 * as a database column name, and the other field's name is altered with a 'A_'
 * or 'B_' prefix.
 */

/* 
 * Set the selected input field's name to a real database field name by
 * removing the "A_" or "B_" prefix from the field name.
 */

function reconcile(fld) {
	var n = fld.name;
	var v = fld.value;
	var f = fld.form;

	var field_name = n.replace(/^C_/,'');

	var Aname = 'A_' + field_name;
	var Bname = 'B_' + field_name;

	// To accommodate radio fields, all the elements in the form
	// are searched for names to change to the A selection or the
	// B selection.

	var i, e;
	switch(v) {
	case 'A':
		for (i=0; i < f.elements.length; i++) { 
			e = f.elements[i];
			if (e.name) {
				if (e.name == field_name) { e.name = Bname }
				if (e.name == Aname)      { e.name = field_name }
			}
		}
		break;
	case 'B':
		for (i=0; i < f.elements.length; i++) { 
			e = f.elements[i];
			if (e.name) {
				if (e.name == field_name) { e.name = Aname }
				if (e.name == Bname)      { e.name = field_name }
			}
		}
		break;
	}
}

window.resizeTo(760, 472 + 200);	// add fudge factor to Y
window.name = "Reconcile_" + window.name;

window.onload = function() {
	var f = document['parameters'];
	if (f) {
		f.onsubmit=function() { return form_save(f) }
		focus_first_field(f);
		f.name = f['_form_name'].value;	// for security
		form_is_setup(f,true);
	}
}
window.onunload = function() {
	// Resize reconciliation window to message window size.
	//window.resizeTo(475,270);
	window.resizeTo(475 + 15, 270 + 40); // add IE resize fudge factors
}
</script>
End_Of_JavaScript

	#
	# Build form fragment of input fields of both values with radio
	# buttons allowing user to choose which field to post.
	#

	# Building the reconciliation form fields is very similar to
	# the Form Builder. However, two fields are built, one field
	# to hold the current database value and one to hold the CGI
	# value that the user is trying to post to the database.
	# Another difference is the field's "compute" property is ignored.

	my $Tabindex_Interval = 10;	# Interval between tabindex values
	my $Field_Size = 50;		# Average input field size (advisory)

	#
	# Get current timestamp to overwrite original _start value.
	#

	my $Timestamp = Acme::Form::DbVars->get('_timestamp');	# set to now

	#
	# Setup new form context used to change field variables defined in the
	# schema into actual form field names.
	#

	my $Context = new Acme::Form::Context table=>$table, row=>($nrec || 0);

	#
	# Construct input fields to hold values that need to be reconciled.
	#

	my ($k, $va, $vb, $vref);
	my ($description, $title_str);
	my ($type, $rows, $size, $maxlength, $decimal);
	my ($bgn_nobr, $end_nobr);
	my ($f, $field_name, $field_label, $field_label_str);
	my ($radio, $checkbox, $hidden);
	my ($class, $class_str);
	my ($style, $style_str);
	my ($tabindex, $tabindex_str1, $tabindex_str2, $tabindex_str3, $tabindex_str4);
	my ($validate, $validate_str, $valid_error);
	my ($translate, $translate_str);
	my ($check_str);
	my ($on_focus_str, $on_change_str);

	#
	# Display fields in the order they appear in the form.
	#

	# This is complicated because the keys in %$args are database column
	# names where as the field names in $Form_RS are fully qualified
	# (actual) field names. Build a separate hash of fully qualified
	# field names and the order in which they were defined in the form
	# so that they can be sorted.

	local our %seq;

	for $k (keys %$args) {

		#
		# Get original form field values.
		#

		# Make form field name out of table and field name.
		# Try several ways of making a fully qualilfied field name.
		# Add record number to field name if used.
		# Keep trying until we can find a defined value for a field.

		$nrec ||= 0;

		if ($nrec > 1) {
			$field_name = "N${nrec}_${table}__${k}" if $nrec;
			$vb = $Form_RS->get($field_name);	# get old value
		}
		else {
			# Try first without number prefix.
			$field_name = "${table}__${k}";
			$vb = $Form_RS->get($field_name);	# get old value
			if (! defined $vb) {
				# Try again with numbered record.
				$field_name = "N${nrec}_$field_name";
				$vb = $Form_RS->get($field_name);# get old value
			}
		}

		# Try again because we allowed fields not qualified by a
		# table name when filling the record set above.

		$field_name = $k if ! defined $vb && $nrec != 1;

		#
		# Assign the order that the field was defined in the form.
		#

		$seq{$field_name} = $Form_RS->{var}->{$field_name};
	}
	
	no warnings qw(closure);		# turn off for $seq in byval()

	sub byval { $seq{$a} <=> $seq{$b} }	# for sort

	use warnings qw(closure);		# turn back on

	for $field_name (sort byval keys %seq) {

		#
		# Get the value of the field the user is trying to post.
		#

		($k = $field_name) =~ s/.*__//;	# get table column name

		next if $k eq $t->id;		# ID keys cannot change, ignore

		$vref = $args->{$k};

		$va = $$vref || '';

		#
		# Get the original form field value - the database value.
		#

		$vb = $Form_RS->get($field_name) || '';

		#
		# Build reconciliation fields.
		#

		$f = $t->f($k);

		$description = escape_html($f->description);
		$hidden      = $f->hidden;
		$maxlength   = $f->length;
		$size        = $maxlength > $Field_Size ? $Field_Size : $maxlength;
		$decimal     = $f->decimal;
		$radio       = $f->radio;
		$checkbox    = $f->checkbox;
		$class       = $f->class;
		$style       = $f->style;
		$type        = 'text';
		$type        = 'radio'    if $radio;
		$type        = 'checkbox' if $checkbox;
		$type        = 'file'     if $class eq 'file';
		$type        = 'text'     if $hidden;
		$translate   = $Context->translate($f->translate);
		$validate    = $Context->translate($f->valid);
		$valid_error = $f->valid_error;

		#
		# Compose field label
		#

		if ($field_label = $f->label) {
			$field_label = escape_html($f->title)
			  if $field_label eq 'yes';
		}
		else {
			# This is different from Form Builder.
			# Always have some kind of label for the field.
			if (!$field_label) {
				$field_label = escape_html($f->title);
			}
		}

		$field_label_str = ($field_label =~ / /)
		  ? "<nobr>$field_label</nobr>"
		  : $field_label;

		#
		# Assemble fields' input tags
		#

		# The title parameter produces a pop-up "tool tip" in IE5
		$title_str = $description ? qq[ title="$description"] : '';

		$bgn_nobr = $end_nobr = '';

		$size = $maxlength > $Field_Size ? $Field_Size : $maxlength;

		$tabindex += $Tabindex_Interval;
		$tabindex_str1 = qq[ tabindex="$tabindex"];
		$tabindex += $Tabindex_Interval;
		$tabindex_str2 = qq[ tabindex="$tabindex"];
		$tabindex += $Tabindex_Interval;
		$tabindex_str3 = qq[ tabindex="$tabindex"];
		$tabindex += $Tabindex_Interval;
		$tabindex_str4 = qq[ tabindex="$tabindex"];

		$class_str = '';
		$class_str = qq[ class="$class";] if $class;

		$style_str = '';
		$style_str = qq[ style="$style";] if $style;

		#
		# Set up form field validation routines
		#

		$translate_str = '';
		if ($translate) {
			$translate_str = $translate;
			$translate_str = escape_html($translate_str);
			$translate_str = qq[translate_input(this,'$translate_str')];
		}

		$check_str = '';
		if (! $f->choose) {
			$check_str = qq[check_input(this,'$class',$maxlength,$decimal)];
		}

		$validate_str = '';
		if ($validate) {
			$validate_str = $validate;
			$validate_str = escape_html($validate_str);
			$validate_str = "validate_input(this,($validate_str)";
			$valid_error = escape_html($valid_error);
			$validate_str .= ",'$valid_error'" if $valid_error;
			$validate_str .= ")";
		}

		#
		# Assemble onChange attribute string
		#

		$on_change_str = '';

		if ($type eq 'text' || $type eq 'file') {

			# Assemble onChange clause without any
			# compute string -- only the translation or
			# validation clauses.

			$on_change_str =
			  join " && ", grep { $_ }
			  $translate_str, $check_str, $validate_str;

			$on_change_str &&=
qq[ onChange="if($on_change_str){return true} this.focus();return false"];
		}

		$on_focus_str = '';		# not used here

		#
		# Define input fields
		#

		if ($type eq 'text' || $type eq 'file' || $type eq 'hidden') {

			$rows = $f->rows || 0;

			if ($maxlength < 100 || $rows == 1 || $type eq 'file') {
				# Standard input field
				$html .= qq[
<tr>
  <td class="label">$field_label_str</td>
  <td class="field"><table class="reconcile">
    <tr class="reconcile">
      <td class="reconcile"><input type="radio" name="C_$field_name" value="A" onClick="reconcile(this)" checked$tabindex_str1></td>
      <td>Your changes:</td>
    </tr>
    <tr>
      <td></td>
      <td>$bgn_nobr<input type="$type" name="$field_name" value="$va" size="$size" maxlength="$maxlength"$class_str$style_str$on_change_str$on_focus_str$tabindex_str2$title_str>$end_nobr</td>
    </tr>
    <tr class="reconcile">
      <td class="reconcile"><input type="radio" name="C_$field_name" value="B" onClick="reconcile(this)"$tabindex_str3></td>
      <td>Existing value:</td>
    </tr>
    <tr>
      <td></td>
      <td>$bgn_nobr<input type="$type" name="B_$field_name" value="$vb" size="$size" maxlength="$maxlength"$class_str$style_str$on_change_str$on_focus_str$tabindex_str4$title_str>$end_nobr</td>
    </tr>
  </table>];
			}
			else {
				if (! $rows) {
					# Long text fields use <textarea> fields.
					# Try to size <textarea> intelligently.
					$rows = int(($maxlength+$Field_Size)/$Field_Size);
					if ($rows > 9) { $rows = 6 }
					elsif ($rows > 7) { $rows = 5 }
					elsif ($rows > 5) { $rows = 4 }
					elsif ($rows > 3) { $rows = 3 }
				}

				$html .= qq[
<tr>
  <td class="label">$field_label_str</td>
  <td class="field"><table class="reconcile">
    <tr class="reconcile">
      <td class="reconcile"><input type="radio" name="C_$field_name" value="A" onClick="reconcile(this)" checked$tabindex_str1></td>
      <td>Your changes:</td>
    </tr>
    <tr>
      <td></td>
      <td>$bgn_nobr<textarea name="$field_name" cols="$Field_Size" rows="$rows"$class_str$style_str$on_focus_str$on_change_str$tabindex_str2$title_str>$va</textarea>$end_nobr</td>
    </tr>
    <tr class="reconcile">
      <td class="reconcile"><input type="radio" name="C_$field_name" value="B" onClick="reconcile(this)"$tabindex_str3></td>
      <td>Existing value:</td>
    </tr>
    <tr>
      <td></td>
      <td>$bgn_nobr<textarea name="B_$field_name" cols="$Field_Size" rows="$rows"$class_str$style_str$on_focus_str$on_change_str$tabindex_str4$title_str>$vb</textarea>$end_nobr</td>
    </tr>
  </table>];
			}
		}
		elsif ($type eq 'radio') {
			my @radios = Text::ParseWords::parse_line('\s+',0,$radio);
			my ($rval,$rlab);
			my $checked = '';

			$html .= <<End_Of_HTML;
<tr>
  <td class="label">$field_label_str</td>
  <td class="field"><table class="reconcile">
    <tr class="reconcile">
      <td class="reconcile"><input type="radio" name="C_$field_name" value="A" onClick="reconcile(this)" checked$tabindex_str1></td>
      <td>Your changes:</td>
    </tr>
    <tr>
      <td></td>
      <td><table class="radio-group">
    <tr>
End_Of_HTML
			my $i = 0;
			while ($i < @radios) {
				$rval = $radios[$i++] or last;
				$rlab = $radios[$i++] or last;

				if ($rval eq $va) { $checked = ' checked' }
				else              { $checked = ''         } 


				$html .= <<End_Of_HTML;
      <td class="radio-field"><input type="$type" name="$field_name" value="$rval"$class_str$style_str$on_change_str$tabindex_str2$title_str$checked></td>
      <td class="radio-label">$rlab</td>
End_Of_HTML
			}
			$html .= <<End_Of_HTML;
    </tr>
    </table></td>
    </tr>
    <tr class="reconcile">
      <td class="reconcile"><input type="radio" name="C_$field_name" value="B" onClick="reconcile(this)"$tabindex_str3></td>
      <td>Existing value:</td>
    </tr>
    <tr>
      <td></td>
      <td><table class="radio-group">
    <tr>
End_Of_HTML
			$i = 0;
			while ($i < @radios) {
				$rval = $radios[$i++] or last;
				$rlab = $radios[$i++] or last;

				if ($rval eq $vb) { $checked = ' checked' }
				else              { $checked = ''         } 

				$html .= <<End_Of_HTML;
      <td class="radio-field"><input type="$type" name="B_$field_name" value="$rval"$class_str$style_str$on_change_str$tabindex_str4$title_str$checked></td>
      <td class="radio-label">$rlab</td>
End_Of_HTML
			}
			$html .= <<End_Of_HTML;
    </tr>
    </table></td>
  </tr>
  </table>
End_Of_HTML
		}
		elsif ($type eq 'checkbox') {		
			my $v = $f->checkbox || '';
			my $checka = ($va eq $v) ? ' checked' : '';
			my $checkb = ($vb eq $v) ? ' checked' : '';

			$html .= qq[
<tr>
  <td class="field"><table class="reconcile">
    <tr class="reconcile">
      <td class="reconcile"><input type="radio" name="C_$field_name" value="A" onClick="reconcile(this)" checked$tabindex_str1></td>
      <td>Your changes:</td>
    </tr>
    <tr>
      <td></td>
      <td>$bgn_nobr<input type="$type" name="$field_name" value="$v" $class_str$style_str$on_change_str$tabindex_str2$title_str$checka>$end_nobr</td>
    </tr>
    <tr class="reconcile">
      <td class="reconcile"><input type="radio" name="C_$field_name" value="B" onClick="reconcile(this)"$tabindex_str3></td>
      <td>Existing value:</td>
    </tr>
    <tr>
      <td></td>
      <td>$bgn_nobr<input type="$type" name="B_$field_name" value="$v" $class_str$style_str$on_change_str$tabindex_str4$title_str$checkb>$end_nobr</td>
    </tr>
  </table>];
		}
	
		$html .= qq[</td>\n</tr>\n];
	}

	$html .= <<End_Of_HTML;
</table>
End_Of_HTML

	#
	# Redefine the form rendering timestamp
	#

	$html .= qq[<input type="hidden" name="_start" value="$Timestamp">\n];

	#
	# Display reconcilliation screen
	#

	require Acme::HTML::Parameters;

	Acme::HTML::Parameters::get_parameters(
			window_title => "Reconcile",
			title        => "Reconcile $table",
			html         => $html,
			script       => $script,
			submit       => "  Save  ",
			);
}

1;
