package Acme::Form::Saver;

=head1 NAME

Acme::Form::Saver - Save form information to a database.

=head1 SYNOPSIS

    use Acme::Form;
    use Acme::Form::Saver;
    our @ISA = qw(Acme::Form::Saver);

    my $self = new Acme::Form;

    $self->save;
     -or-
    $self->save(form_name=>'name');

=head1 DESCRIPTION

I<Acme::Form::Saver> is a subclass of I<Acme::Form>. It provides methods to
save form fields to the database.

Acme::Form::Saver also inherits methods from I<Acme::Form::Filler> to fill
a form with values. The filled in form is then compared with the incoming
CGI arguments to determine which fields needs to be saved to the database.
Security checks are performed and audit trails are made if requested in
the schema (e.g. Secure and Audit properties).

The save() method saves the named form or the first form found in the html
template. It returns the number of records saved or 0 if no changes were
made. Errors generate exceptions killing the program.

The save_form() method fills database values changed in a form. Dropdown lists
of values are not built as they are in the Filler. Only the fields' current 
values are needed by the Saver.

Form Filler directives like:

    <!--$ command string -->

are executed by the Saver just as they are by the Filler.

All fields to be added or updated in the database are submitted as one SQL
transaction -- even in forms with fields for more than one database record.
If there is any problem posting any field of any record to the database,
the entire save is rolled back. This allows the user to re-save the 
entire form, which is still rendered on their computer until a form has
been saved or canceled.

=head1 CHANGE LOG

 5/14/00  Initial version
 10/3/01  Changed to Form::Saver.
 4/9/02   Major upgrade: added audit trail, security, reconciliation screen,
          schema 'do ...' and 'onSave' clause processing, and multiple rows.

=head1 TODO

 * Test saving multiple records in form.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000,2002  Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

use CGI;
use Acme::HTTP			qw(escape);
use Acme::HTML			qw(escape_html unescape_html debug debug_on);
use Acme::Db;
use Acme::Db::RecordSet;
use Acme::Form::Vars;
use Acme::Form::Filler;
use Acme::Base::Schema;
use Acme::Base::Schema::Join;
use Acme::Base::Schema::Master;
use Acme::Config;
use Local::Acme::User		qw(user);
use Local::Acme::Format;
use Text::ParseWords;

use Acme::Security;

our $Save_Mode;		# New or Edit depending if key value exists.

our $Dbh;		# Database handle.
our $Schema;		# AcmeBase handle.
our $Primary_Table;	# Name of primary table being saved.
our $Primary_Key;	# Name of primary table's key field.
our $Master;		# Master field on master form with SQL filter value.
our $Form_RS;		# Form variables to compare with CGI variables.
our $CGI_RS;		# CGI variables to compare with Form variables.
our $User;		# User Id who is submitting transaction.
our @After_Save;	# Cached afterSave events.

our @ISA = qw(Acme::Form::Filler);

#-------------------------------------------------------------------------------
#
# Module for holding original values of form fields.
#
#-------------------------------------------------------------------------------

package Acme::Form::FormVars;
our @ISA = qw(Acme::Form::Vars);

#
# Add special functions to track the checkboxes found in the form so that
# blank values can be added to the CGI variables for the "unchecked" checkbox
# fields. Browsers do not send "unchecked" field values to the server.
#

our @Checkboxes;

sub new {
	my $self = shift;
	@Checkboxes = ();	# initialize for mod_perl
	return $self->SUPER::new(@_);
}

sub add_checkbox {
	my $self = shift;
	push @Checkboxes, @_;
}

sub checkboxes {
	my $self = shift;
	return @Checkboxes;
}

#-------------------------------------------------------------------------------
#
# Module for holding new CGI values for form fields.
#
#-------------------------------------------------------------------------------

# This module is used as the sandbox for executing "do" scripts defined in
# the schema. The CGI arguments are assigned global variables here before
# executing the "do" scripts. It's the global variables from this module that
# actually gets posted to the database.

package Acme::Form::CGIVars;
our @ISA = qw(Acme::Form::Vars);

our $Inserting = 0;		# set to 1 if current record is being inserted
our $Updating  = 0;		# set to 1 if current record is being updated

#-------------------------------------------------------------------------------
#
# Module to build SQL statement
#
#-------------------------------------------------------------------------------

package Acme::Form::Saver::SQL;

sub new { 
	my $this = shift;
	my $class = ref($this) || $this;

	my $self = {
		sql => '',		  # Accumulated SQL statements.
		new_records => [],	  # Array of pending inserts to commit.
		detail_result_args => [], # Cached arguments to detail_result()
		detail_result_keys => [], # Cached key values to detail_result()
		};

	return bless $self, $class;
}

sub add {
	my $self = shift;

	# Separate statements with semi-colons.
	$self->{sql} .= join "\n", (map { /;\s*$/ ? $_ : "$_;" } @_), '';
}

sub as_sql {
	my $self = shift;
	return $self->{sql};
}

#-------------------------------------------------------------------------------
#
#  Saver public methods
#
#-------------------------------------------------------------------------------

package Acme::Form::Saver;

#debug_on();

sub save {
	my $self = shift;
	my %opt = @_;		# rest of arguments are key/value options

	my $args      = $opt{args}		||
			$self->{args}		||
			$self->{cgi}		||
			new CGI;
	$args = new CGI $args if ref($args) eq 'HASH';


	my $form_name = $opt{form_name}	  	   ||
			$args->param('_form_name') ||
			$self->{form_name}	   ||
			$args->param('form_name')  ||
			'';
	#
	# Find form to save
	#

	my ($form, $tag, $name);

	my $n_save;			# number of records to save

	my $href = \$self->{html};	# get scalar ref for progressive search
	pos($$href) = 0;		# safety
	while ($$href =~ m|(<form.*?</form.*?>)|gsi) {
		$form = $1;
		if ($form_name) {
			($tag) = $form =~ m/<form\s(.*?)>/si;
			(undef,$name) = $tag =~ /name=(['"])?(.*?)\1\s/si;
			next if $name !~ /^$form_name$/oi;

			$n_save = $self->save_form($form,\%opt) or return;
			last;
		}
		else {
			# Form name not specified. Save first form found.
			$n_save = $self->save_form($form,\%opt) or return;
			last;
		}
	}

	die ("The form named '$form_name' was not found in form file")
	  if ! defined $n_save;

	return $self->{n_save} = $n_save;
}

#-------------------------------------------------------------------------------

sub save_form {
	my $self = shift;	# an Acme::Form object
	my ($html,$opt) = @_;

	#
	# Determine function parameters
	#

	my $args      = $opt->{args}		||
			$self->{args}		||
			$self->{cgi}		||
			new CGI;
	$args = new CGI $args if ref($args) eq 'HASH';

	my $request   = $opt->{request}		||
			$self->{request}	||
			$args->param('_request')||
			$args->param('request')	||
			'';
	my $form      = $opt->{form}		||
			$self->{form}		||
			$args->param('_form')	||
			$args->param('form')	||
			'';
	my $table     = $opt->{table}		||
			$self->{table}		||
			$args->param('_table')	||
			$args->param('table')	||
			'';
	my $master    = $opt->{master}		||
			$self->{master}		||
			$args->param('_master')	||
			$args->param('master')	||
			'';

	local $Dbh = Acme::Db->connect();	# connect to database

	local $Schema = $self->{schema} || new Acme::Form::Schema { form => $form };

	# Default primary table is the first table listed in a form file.
	($table) = $Schema->tables if ! $table && $form;

	local $Primary_Table = $table;

	my $t = $Schema->t($Primary_Table)
	  or die("Cannot find the table '$Primary_Table' in the"
	  . ($form ? " form '$form'" : " schema")
	  . " description");

	local $Primary_Key = $t->key;

	local $Master = $master;		# optional master field name

	local $User = user();			# For audit

	local @After_Save = ();			# Initialized cached Do clauses

	my $n_save;				# Returned number of recs saved

	#
	# Build record set of CGI values to compare with form's original values.
	#

	local $CGI_RS = new Acme::Form::CGIVars;

	my ($arg, @val, $sep);
	for $arg ($args->param()) {
		next if $arg =~ /^_/;		  # Private variables
		if ($arg =~ /^P_/) {		  # Previous key values
			$CGI_RS->import($arg => $args->param($arg));	
			next;
		}
		next if $arg =~ /^[A-Z]_[^_]/;	  # Special form (non-data) fields

		@val = $args->param($arg);
		if (@val > 1) {
			# Make text array out of multiple values.

			$sep = $self->_get_field_sep($arg);
			$_ = $sep . (join $sep, @val) . $sep;

			$CGI_RS->set($arg => $_);
		}
		else {
			$CGI_RS->set($arg => $val[0]);
		}
	}

	debug("save_form: CGI namespace:",$CGI_RS);

	#
	# Build record set of original form field values. Use the Form Filler.
	#

	# Add extra options to the options hash passed to this routine and
	# use it as the options hash for Acme::Form::Filler::fill_form().
	# Pass the start time to the Filler to use to determine if records
	# were updated since the form was rendered.

	local $Form_RS = new Acme::Form::FormVars;

	$opt->{save}  = $Form_RS;		# initial form values record set
	$opt->{table} = $args->param('_table');	# primary table
	$opt->{start} = $args->param('_start');	# time form was rendered

	($opt->{table}) = $opt->{table} =~ /(\w+)/;	# untaint

	$self->fill_form($html, $opt);

	debug("save_form: Form namespace: (initial form values)",$Form_RS);

	#
	# Add blank settings to CGI arguments for all checkboxes that are
	# not defined. Browsers do not send blank field values if a checkbox
	# is unchecked.
	#

	for my $c ($Form_RS->checkboxes) {
		$CGI_RS->set($c => '') unless $CGI_RS->get($c);
	}

	#
	# Check incoming CGI form fields with original form fields.
	#

	$self->_check_security_fields($form,$table,$request);

	#
	# Initialize SQL script used to post the form.
	#

	my $sqls = new Acme::Form::Saver::SQL;

	# Define sql() utility routine so it is available to the code in the
	# schema "do", "beforeSave", "onSave", and "afterSave properties.
	# It constructs the sql statements to be sent to the database engine.
	# Defining the sql() function here allows the use of the above
	# variable $sqls to be used as a global variable but still be dynamic
	# in mod_perl.

	package Acme::Form::CGIVars;
	no warnings qw(misc);		# no warning that $sqls will not stay shared.
	sub sql($;@) { $sqls->add(@_) }

	package Acme::Form::Saver;

	#
	# Do saver commands defined in schema.
	#

	my $do = join ';', $t->do, $t->do_saver;
	if ($do) {

		#
		# Evaluate do_saver property's Perl script.
		# This gets executed before the beforeSave,
		# onSave, and afterSave properties. 
		#

		# Note: The do_saver property subverts Perl's
		# taint checks as the schema (and form) files
		# are assumed to be protected files.
	
		($do) = $do =~ /^(.*)$/s;	# untaint

		package Acme::Form::CGIVars;
		no strict;
		eval $do;
		die "Error evaluating schema's Do property: $@" if $@;
	}

	#
	# Fill record set of the fields with changes and their values.
	#

	my $rs = new Acme::Db::RecordSet;		# data to be saved

	$n_save = $self->_save_set($rs,$sqls,$t,$request)
	  or return 0;					# return if no changes

	debug("Changes need to be saved to database. RecordSet:",$rs);

	if ($n_save) {

		#
		# Execute all SQL statements as one transaction.
		#

		$self->_execute_sql($sqls);

		#
		# Now execute the cached afterSave do clauses.
		#

		if (@After_Save) {
			my $do = join "; ", @After_Save;

			package Acme::Form::CGIVars;
			no strict;
			eval $do;
			die "Error evaluating schema's afterSave property: $@" if $@;
		}
	}

	return $n_save;
}

#-------------------------------------------------------------------------------
#
#  Private methods
#
#-------------------------------------------------------------------------------

sub _check_security_fields {
	my ($self, $form, $table, $request) = @_;

	#
	# Compare the CGI fields with the original Form fields.
	#

	# Form fields do not get built by the Builder unless user has clearance.
	# Likewise, fields do not get filled by the Filler unless the user has
	# clearance. Therefore, CGI fields that are not found in the $Form_RS
	# are fields that the user does not have security clearance for.
	#
	# There is the case where the user is using a pre-built form template
	# that has a field defined on it that the user does not have clearance
	# for, so this routine also checks security for each table/field saved.

	my @notfound = ();
	my @insecure = ();
	my ($recno, $tab, $fld);

	for (keys %{$CGI_RS->{var}}) {

		#
		# Parse field name
		#

		($recno, $tab, $fld) = /^(?:N(\d+)_)?	# optional N12_ prefix
					(?:(\w*?)__)?	# optional table name
					(\w*)		# required field name
					/x;

		$tab ||= $table;	# use primary table if not in field name

		#
		# Check if corresponding field in original form.
		#

		if (! exists $Form_RS->{var}->{$_}) {

			# The CGI field is not in the original form fields.
			# Check if it is a new record added in the browser.
			# There must be a field number prefix (N1_table__field)
			# and there must exist an original form field
			# named N0_table__field. The N0_table_field's are
			# used as templates to build new records and are
			# otherwise ignored.

			if (! defined $recno
			||  ! exists $Form_RS->{var}->{"N0_${tab}__${fld}"}) {
				push @notfound, $_;
			}
		}
		elsif (! _has_clearance($User,$tab,$fld,$Save_Mode)) {
			push @insecure, $_;
		}
	}

	return unless @notfound || @insecure;

	#
	# Send security alert and an error mesage.
	#

	$form ||= "(not specified, table schema used)";
	my $Webadministrator = Acme::Config->Webadministrator;

	my $notfound_msg = qq[The user attempted to insert into the database a field that was not defined in the original form. ];

	my $insecure_msg = qq[The user does not have security clearance to enter or change a database field. ];

	Acme::Security::alert(<<End_Of_Alert);
A problem occurred when saving a form.

$notfound_msg$insecure_msg

   User: $User
   Form: $form
  Table: $table
Request: $request

The fields the user attempted to save are:

	@{[ join "\n\t", @notfound, @insecure ]}

End_Of_Alert

	die <<End_Of_Msg;
A problem occurred when saving your form.

Please contact $Webadministrator.
End_Of_Msg
}

#-------------------------------------------------------------------------------

sub _check_security_insert {
	my ($self, $t) = @_;

	if (my $restrict_where = $t->restrict_where_insert) {
		my $Webadministrator = Acme::Config->Webadministrator;
		Acme::Security::alert(<<End_Of_Alert);
Attempt to insert into a table with a restrict code.
	$restrict_where
Insert rejected.
End_Of_Alert

		die <<End_Of_Msg;
Your request has been denied because you are trying to insert information into a restricted table. Either that or this program has encountered a bug and has become utterly confused.

Please contact $Webadministrator.
End_Of_Msg
	}
}

#-------------------------------------------------------------------------------

sub _has_clearance {
	my ($user,$tab,$fld,$mode) = @_;

	#
	# Check security if user has clearance to insert or update the database.
	#

	# Track last request to speed up security checks.

	our ($prev_user, $prev_tab, $t, $security_code);

	if (! $prev_tab || $prev_tab ne $tab || $user ne $prev_user) {
		$t = $Schema->t($tab);
		$security_code = $t->security_code;
		$prev_tab  = $tab;
		$prev_user = $user;
	}

	return 1 if ! $security_code;		# OK, No security.

	if ($mode eq 'New') {
		return 0 unless Acme::Security::allow('Access',$security_code);
	}
	else {
		return 0 unless Acme::Security::allow('Update',$security_code);
	}

	my $f = $t->f($fld);

	return 0 if !$f && $mode eq 'New' && $fld eq $t->id;  # cannot update ID
	return 1 unless $f;			# should not happen (?)

	if (my $sc = $f->security_code) {
		if ($mode eq 'New') {
			return 0 unless Acme::Security::allow('Access',$sc);
		}
		else {
			return 0 unless Acme::Security::allow('Update',$sc);
		}
	}

	return 1;				# OK
}

#-------------------------------------------------------------------------------

sub _get_field_sep {
	my ($self, $field) = @_;

	#
	# Return the text array separator for a field or undef
	# if the field is not a text array.
	#

	my $sep;

	# Parse field name
	my ($recno, $tab, $fld) =
	  $field =~ /^(?:N(\d+)_)?	# optional N12_ prefix
		     (?:(\w*?)__)?	# optional table name
		     (\w*)		# required field name
		    /x;

	# Hack into Saver module for $Schema global to use to find the
	# "official" array_separator character for this field.

	my ($t,$f);
	if ($Schema && $tab && ($t = $Schema->t($tab)) && ($f = $t->f($fld)) ) {
		return unless $f->array;	# not an array
		$sep = $f->array_separator;
	}
	else {
		$sep = Acme::Config->Array_Separator;
	}
	$sep ||= ';';	# final default

	return $sep;
}

#-------------------------------------------------------------------------------

sub _get_save_mode {
	my ($self, $t, $keyval, $request) = @_;

	#
	# Return save mode (New or Edit) depending on whether a value is set for
	# a key value. Key values need to be set unless it is an ID key field
	# and the request mode is New.
	#
	# Return undef if request is New and record already exists.
	#

	my $table = $t->name;

	my $id = $t->id;	# key treated differently if it is an ID number

	if ($keyval) {
		if ($id) {
			# Key value given for an ID record.
			# Only way to get an ID number is to 
			# previously add it. Do not check for
			# records existance.

			return 'Edit';
		}
		else {
			# Key value given for a non-ID record.
			# Check to see if key exists to determine if
			# Edit mode or New mode.

			my $key = $t->key;
			my $curr_keyval = $Dbh->get(<<EOS);
select $key
from $table
where $key = '$keyval'
EOS
			if (length($curr_keyval || '')) {
				if ($table eq $Primary_Table
				&& $request =~ /new/i) {
					die <<EOF;
$table record already exists with a value of '$key'
EOF
				}
				return 'Edit';
			}
			else {

				return 'New';
			}
		}
	}
	else {
		#
		# Key value NOT specified.
		#

		# Every record must have a key. Only records with ID
		# fields have their key auto-generated.

		die "Required key value was not specified.\n" unless $id;

		return 'New';
	}
}

#-------------------------------------------------------------------------------

sub _save_set {
	my ($self, $rs, $sqls, $t, $request) = @_;

	my ($key, $keyval);	# key field name and value
	my ($prev_keyval);	# previous key value to allow key changes
	my ($fvref, $cvref);	# references to $Form_RS and $CGI_RS variables
	my ($tab,$fld,$recno);	# parts of field name
	my $ndiff = 0;		# number of fields with differences

	my $n_save = 0;		# returned number of records to save

	#
	# Multiple records can be saved from a form if every field of each
	# record has a prefix of N1_ or N2_ and so on. Records with the
	# N0_ prefix are assumed to be pattern records and are ignored.
	# Fields without a prefix are processed first.
	#

	for my $nrec (undef,1..$CGI_RS->{nrec}) {

		$ndiff = 0;

		if (defined $nrec) {
			$key = "N${nrec}_${Primary_Table}__${Primary_Key}";
		}
		else {
			$key = "${Primary_Table}__${Primary_Key}";
		}

		#
		# Tables with non-ID keys uses a "P_$key" field to store the
		# key's previous value so the user can change key values
		# for the record.
		#

		$keyval        = $CGI_RS->get($key);
		$prev_keyval   = $CGI_RS->get("P_$key");
		$ndiff++ if $prev_keyval && $prev_keyval ne $keyval;
		$prev_keyval ||= $keyval;

		#
		# Determine Save_Mode for each table. Join tables can have a
		# Save_Mode of "New" even though the primary table record has
		# a Save_Mode of "Edit".
		#

		local $Save_Mode = $self->_get_save_mode($t,$prev_keyval,$request);

		#
		# Initialize record set with key value if editing record.
		#

		if ($Save_Mode eq 'Edit') {
			
			# Add key value to record set.
			# Note: The $nrec value is lost when _build_sql()
			# assembles the record sets into SQL. Therefore,
			# we now embed the key value here.

			$rs->set($Primary_Table,$Primary_Key,$nrec,$CGI_RS->get_vref($key));
		}

		#
		# Build record set of input arguments that have changed.
		#

		no warnings qw(uninitialized);

		for my $k (keys %{$CGI_RS->{var}}) {

			$k =~ s/^.*:://; # strip package name from variable

			next if $k =~ /^_/;	# ignore form processing fields

			#
			# Parse field name
			#

			my ($recno, $tab, $fld) =
			  $k =~ /^(?:N(\d+)_)?	# optional N12_ prefix
				  (?:(\w*?)__)?	# optional table name
				  (\w*)		# required field name
				 /x;
			$tab ||= $Primary_Table;	# use primary table if not in field name

			if (defined $nrec) {
				next if ! defined $recno;
				next if $nrec != $recno;
			}
			else {
				next if defined $recno;
			}

			$recno ||= 0;
			$recno-- if $recno > 0;	 # make 0 based index

			$fvref = $Form_RS->get_vref($k);# get ref to form variable
			$cvref = $CGI_RS->get_vref($k);	# get ref to cgi variable

			debug("_save_set:"
			  . " $Save_Mode: $k: old '$$fvref' = new '$$cvref' ?");

			#
			# Compare original value with current value.
			# Save only those that differ.
			#
			
			if (($Save_Mode eq 'New'  && length($$cvref) > 0)
			||  ($Save_Mode eq 'Edit' && "$$fvref" ne "$$cvref")) {

				debug("Different!");

				#
				# Build record set to save to database.
				#

				# Note: we are storing a reference to the 
				# variable, not the variable itself.

				debug(
				  "Setting tab=$tab, fld=$fld, recno=$recno, cvref=$$cvref");

				$rs->set($tab,$fld,$recno,$cvref);

				$ndiff++;
			}
		}

		#
		# Assemble SQL statement to save all records.
		#

		$n_save += $self->_build_sql($sqls, $rs, $request) if $ndiff;
	}

	return $n_save;
}

#-------------------------------------------------------------------------------

sub _build_sql {
	my ($self, $sqls, $rs, $request) = @_;

	#
	# Make SQL needed to update database with CGI arguments
	#

	my $n_save = 0;		# returned number of records saved in the SQL

	#
	# Save primary table first
	#

	my $t = $Schema->t($Primary_Table);

	my $j = new Acme::Base::Schema::Join $Primary_Table, $t->join
	  if $t->join;

	my $nrec = 0;
	for my $primary_args ($rs->records($Primary_Table)) {

		$nrec++;

		# Skip the record if there isn't any data arguments.
		# Happens on multiple records with unchanged fields.

		next unless %$primary_args;

		debug("_build_sql: processing primary_args.");

		#
		# Save
		#

		if ($Save_Mode eq 'New') {
			if ($self->_build_sql_new($sqls,$primary_args,$t)) {
				$n_save++;
		  	}
			else {
				return 0;	# something is wrong
			}
		}
		else {	# $Save_Mode eq 'Edit'
			if ($self->_build_sql_edit($sqls,$primary_args,$t,$nrec)) {
				$n_save++;
		  	}
		}

		#
		# Save dependent tables
		#

		for my $table ($rs->tables) {

			next if $table eq $Primary_Table;

			my $t = $Schema->t($table);

			my $m = new Acme::Base::Schema::Master $table, $t->master
			  if $t->master;

			my $dep_nrec = 0;
			for my $args ($rs->records($table)) {

				# Note: This loop should only execute once
				# as only 1 to 1 joins on the screens are
				# accommodated at this time.

				$dep_nrec++;		# (not used yet in this version)
				next unless $args;	# happens with unchanged records.

				debug("_build_sql: processing dependent: $table");

				my $key_field = $t->key;

				# Check join clause.
				if ($j) {
					my $field = $j->field($table,$key_field);

					# Assigned join'ed table's key field to
					# value of primary table's join field.

					$args->{$key_field} = $primary_args->{$field}
					  unless defined $args->{$key_field};
				}

				# Check master clause.
				if ($m) {
					for my $field ($m->fields) {
						if (! $args->{$field}
						&& (my $join_field = $m->master_field($field,$Primary_Table))) {
							$args->{$field} = $primary_args->{$join_field};
						}
					}
				}

				my $keyval = ${$args->{$key_field}};

				# Note: Each record has their own "save mode"
				# depending upon whether their key has a
				# value or not.

				local $Save_Mode = $self->_get_save_mode($t, $keyval, $request);

				#
				# Save
				#

				$n_save += $self->_build_sql_new($sqls,$args,$t)
				  if $Save_Mode eq 'New';

				$n_save += $self->_build_sql_edit($sqls,$args,$t,$nrec)
				  if $Save_Mode eq 'Edit';

				#
				# Cache arguments to _detail_result() routine
				# to be executed later after sql is executed.
				#

				push @{$sqls->{detail_result_args}}, [ $t, $args ]
				  if $Master;
			}
		}

		#
		# Cache arguments to _detail_result() routine
		# to be executed later after sql is executed.
		#

		push @{$sqls->{detail_result_args}}, [ $t, $primary_args ]
		  if $Master;
	}

	return $n_save;
}

#-------------------------------------------------------------------------------

sub _build_sql_new {
	my ($self,$sqls,$args,$t) = @_;

	#
	# Make SQL statment for adding a new record to the database.
	#

	my $table = $t->name;
	my $key = $t->key;

	debug("_build_sql_new: table=$table, args:",$args);

	#
	# Final security checks before insert.
	#

	$self->_check_security_insert($t);

	#
	# Insert new blank record and get key value assigned to it.
	#

	my $r = $Dbh->new_record($table,$key,${$args->{$key}})
	  or die "Cannot make a new database record for table '$table'";

	#
	# Enter the key and the keyval into the record set pointed to
	# by $args so that it is available to dependent joined tables
	# that need to know the value of the joined field.
	#

	my $keyval = $r->keyval;		# assign value to key
	$args->{$key} = \$keyval;		# save key in recordset
						# Note that values are refs.

	#
	# Import new key value into CGI variable space so that it is
	# available to onSave statements when adding new records.
	#

	$CGI_RS->import($key => $keyval);	

	#
	# Construct SQL to save record to database
	#

	$self->_build_sql_script($sqls,$args,$t,$keyval);

	#
	# Cache records needing to be committed after sql is executed.
	#

	push @{$sqls->{new_records}}, $r;

	#
	# Return number of records "inserted".
	#

	return 1;
}

#-------------------------------------------------------------------------------

sub _build_sql_edit {
	my ($self,$sqls,$args,$t,$nrec) = @_;

	#
	# Make SQL statment for editing a record in the database.
	#

	my $table = $t->name;
	my $row_name = $t->row_name;

	debug("_build_sql_edit: table=$table, nrec=$nrec, args:",$args);

	#
	# Check if there are any data fields to post to database.
	# Skip update if the args contain only a record of key field and value.
	#

	my $key = $t->key;
	my $keyval = ${$args->{$key}} if $args->{$key};

	my $prev_keyval = $CGI_RS->get('P_' . ($nrec ? "N${nrec}_" : '') . "${table}__${key}");
	$prev_keyval  ||= $CGI_RS->get("P_${table}__${key}");
	$prev_keyval  ||= $keyval || '';

	debug("_build_sql_edit: keyval=$keyval, prev_keyval=$prev_keyval");

	return 0 if scalar(keys(%$args)) == 1
	  && ( $args->{$key} && ($keyval eq $prev_keyval));

	#
	# Check if record has been updated since the form was rendered.
	#

	# Note: the Form Filler adds a field to the record set when saving.
	# It is the difference between the database record's last_update
	# and the time the form was rendered. This is so the database is
	# the creator of the timestamps and it is the one that computes the
	# difference between the times.

	my $audit = $t->audit;
	my $skip_audit_of_new_master_rec = 0;		# flag

	if ($audit) {

		# Get the user name of the user who previously updated the
		# current record to compare with the name of the user currently
		# updating the record to use to prevent adding an audit record
		# when a user is re-saving a new master record after entering 
		# its detail records.

		my $last_user = $Form_RS->get(
		  _fq_field_prefix($table,$key,$nrec) . $t->audit_user);

		debug("_build_sql_edit: last user='$last_user', current user='$User'");

		# Get delta time values compute by the database when getting
		# the database values used to determine the form's original
		# field values.

		my $delta = Acme::Form::DbVars->get('_y_' . ($nrec ? "N${nrec}_" : '') . $table);
		$delta ||= Acme::Form::DbVars->get("_y_$table") || '';

		debug("_build_sql_edit: delta='$delta'");

		# Note: the computed interval $delta should begin with a minus sign
		# if record was updated since the form was rendered. However, Postgres
		# wants to do fancy output here and formats a string like:
		# 	@ 15 mins ago
		# so the determination whether to present a reconciliation screen
		# or not searches for a minus sign and for the keyword "ago".

		if ($delta =~ /^\s*-/ || $delta =~ /\bago\b/) {

			#
			# The time interval between the time the form was
			# rendered and the record's last update time is negative.
			#

			# Check if current user is the last user who updated
			# the record, and if so, check delta time and make sure
			# it is a "big enough" interval to accommodate the case
			# of adding a new master record, posting it before 
			# adding detail records, and then changing the master
			# record and posting it again after entering the 
			# detail records. The time returned to the master
			# form is the time the record was attempted to be
			# read from the database, not the actual timestamp
			# used when posting the record. There can (is) a
			# difference.

			if ($last_user eq $User
			&& ( $delta eq '@ 1 sec ago'	# Postgres
			  || $delta eq '-00:00:01'
			)) {
				$skip_audit_of_new_master_rec = 1;	# set flag

				debug("_build_sql_edit: reconcilliation is not"
				  . " needed as the current user was the last"
				  . " one to update this record a short time"
				  . " ago.");
			}
			else {
				#
				# Display reconciliation screen from the fields that
				# have changed and then exit the program.
				#

				$self->_reconcile($t, $args, $nrec);
			}
		}
		else {
			# Skip entering an audit record if re-saving a master
			# record after entering its detail records.

			if ($last_user eq $User
			&& ( $delta eq '@ 0'		# Postgres
			  || $delta eq '00:00:00'
			)) {
				$skip_audit_of_new_master_rec = 1;	# set flag

				debug("_build_sql_edit: skip entry of audit record.");
			}
		}
	}
	
	#
	# Assemble SQL for editing table record.
	#

	$self->_build_sql_script($sqls,$args,$t,$prev_keyval);

	#
	# Assemble SQL to saving history of changes to table record.
	#

	$self->_build_sql_audit($sqls,$args,$t,$nrec)
	  if $audit && ! $skip_audit_of_new_master_rec;

	#
	# Compose edit result message
	#

	if ($table eq $Primary_Table) {

		if (my $id = $t->id) {

			#
			# Display ID number of changed record.
			#

			my $keyval = ${$args->{$id}};
			$self->{result} .= <<EOF;
<br>The $row_name record with <nobr>ID #<span style="font-size:large">$keyval</span></nobr> has been changed.
EOF
		}
		else {
			#
			# Standard changed record message.
			#

			$self->{result} .= <<EOF;
<br>The $row_name record has been changed.
EOF
		}
	}
	else {
		#
		# Compose result message for subordinate tables.
		#

		if (! $self->{result}) {

			# Execute following if a confirmation message is wanted
			# for every record updated, even the detail rows.
			# Do only if the primary was not saved (and no message).

			my $nf = $t->number_field;
			my $n = $args->{$nf} if $nf;
			$n ||= '';
			$n = " #$n" if $n;
			$self->{result} .= qq[\n<br>The $row_name record$n has been changed.];
		}
	}

	#
	# Build a "result form" to pass back key and key value to calling window.
	# 

	$self->{result} .= 
qq[<form name="result" onSubmit="return false"><input type="hidden" name="key" value="$key"><input type="hidden" name="keyval" value="$keyval"></form>];

	#
	# Return number of records updated.
	#

	return 1;
}

#-------------------------------------------------------------------------------

sub _build_sql_script {
	my ($self, $sqls, $args, $t, $keyval) = @_;

	#
	# Trigger beforeSave, onSave, and afterSave events and assemble SQL
	# statement for updating a record. This is also used for entering
	# values into a newly inserted blank record.
	#

	my $table = $t->name;
	my $key = $t->key;

	$keyval = ${$args->{$key}} if ! defined $keyval && defined $args->{$key};
	die "Missing value for $key of table $table" if ! defined $keyval;

	my ($audit, $audit_entry, $audit_update, $audit_user);
	if ($audit = $t->audit) {
		$audit_entry  = $t->audit_entry;
		$audit_update = $t->audit_update;
		$audit_user   = $t->audit_user;
	}

	#
	# Trigger beforeSave
	#

	my $do;			# Perl scripts to do
	if ($do = $t->beforeSave) {
		$do = _prep_do($do,$table,$key,$keyval);

		# Note: taint checks are subverted here.
		($do) = $do =~ /^(.*)$/s;	# untaint

		package Acme::Form::CGIVars;
		no strict;
		eval $do;
		die "Error evaluating schema's beforeSave property: $@" if $@;
	}

	#
	# Assemble SQL Update statement
	#

	# Note: The values in %args are references to the real variables.
	# That allows changes to variables in Acme::Form::CGIVars above be
	# immediately available for assembling the SQL below.

	my ($k, $v, $vref, @kv);
	while (($k,$vref) = each %$args) {

		if ($audit) {
			# Skip audit fields. They will be added back in below.
			next if $k eq $audit_entry;
			next if $k eq $audit_update;
			next if $k eq $audit_user;
		}

		# Strip out updates to keys with same value.
		next if $k eq $key && ($Save_Mode eq 'New' || $$vref eq $keyval);

		$v = Acme::Db->escape_sql(Acme::HTML::unescape_html($$vref));

		# Set nulls to temporarily code.
		$v = '#NULL#' unless length $v > 0;

		push @kv, join " = '", $k, $v;
	}

	my $save_script = "update $table set \n" . (join "',\n", @kv) . "'";
	if ($audit) {
		# Add in audit information
		$save_script .= ",\n$audit_entry = current_date"
		  if $Save_Mode eq 'New';
		$save_script .= ",\n$audit_update = current_timestamp";
		$save_script .= ",\n$audit_user = '$User'";
	}
	$save_script .= "\nwhere $key = '$keyval';";
	
	# Substitute null for temporary codes set above.
	$save_script =~ s/= '#NULL#'/= NULL/gs;

	#
	# Add SQL Update statement to SQL script
	#

	$sqls->add($save_script);
	
	#
	# Trigger onSave
	#

	if ($do = $t->onSave) {
		$do = _prep_do($do,$table,$key,$keyval);

		# Note: taint checks are subverted here.
		($do) = $do =~ /^(.*)$/s;	# untaint

		package Acme::Form::CGIVars;
		no strict;
		eval $do;
		die "Error evaluating schema's beforeSave property: $@" if $@;
	}

	#
	# Trigger afterSave
	#

	if ($do = $t->afterSave) {
		$do = _prep_do($do,$table,$key,$keyval);

		# Note: taint checks are subverted here.
		($do) = $do =~ /^(.*)$/s;	# untaint

		#
		# Cache afterSave events until sql is actually executed.
		#

		unshift @After_Save, $do;
	}
}

#-------------------------------------------------------------------------------

sub _build_sql_audit {
	my ($self, $sqls, $args, $t, $nrec) = @_;

	#
	# Assemble SQL script for inserting database record changes into its
	# audit table.
	#

	my $table = $t->name;
	my $key = $t->key;
	my $keyval = ${$args->{$key}};
	debug("_build_sql_audit: table=$table, keyval=$keyval");
	die "Missing value for $key of table $table" if ! defined $keyval;

	my $audit_table  = $t->audit_table;
	my $audit_entry  = $t->audit_entry;
	my $audit_update = $t->audit_update;
	my $audit_user   = $t->audit_user;

	my $audit_type      = Acme::Config->audit_type; 
	my $audit_userstamp = Acme::Config->audit_userstamp;
	my $audit_timestamp = Acme::Config->audit_timestamp;

	# Build hash of fields to exclude from saving in the audit table.
	my %exclude =  map { ($_,1) } split /\s*\W+\s*/, $t->audit_exclude || '';
		
	debug("_build_sql_audit: fields excluded from audit:",\%exclude);

	#
	# Get the prefix to use to turn column names into the fully-qualified
	# form field names. Use the record's key field as the pattern field.
	#

	my $prefix = _fq_field_prefix($table,$key,$nrec);

	#
	# Build two hashes: one for non-null changes to the field values,
	# and one for those fields that were changed to null. Store the
	# old values of the changed fields in the history file.
	#

	# Note: The values in %args are references to the real variables.

	my (%null_args, %nonnull_args);
	my ($k, $v, $vref);

	while (($k,$vref) = each %$args) {
		next if $exclude{$k};
		next if $k eq $key;	# keys treated differently

		
		# Get original form field values.
		#

		$v = $Form_RS->get("$prefix$k");

		#
		# Save values that are being changed in the audit table.
		#

		if ($v) {
			# Store old value for history record.

			$v = Acme::Db->escape_sql(Acme::HTML::unescape_html($v));

			$nonnull_args{$k} = $v;
		}
		else {
			# Use new value as a placeholder for a "null" value in a
			# "null" type history record. This record is used when
			# a null value is changed to a non-null value.

			$v = Acme::Db->escape_sql(Acme::HTML::unescape_html($$vref));

			$null_args{$k} = $v;
		}
	}

	#
	# Assemble SQL Insert into audit history table.
	#

	debug("_build_sql_audit: nonnull_args:",\%nonnull_args);

	if (%nonnull_args) {
		# Set type of record to code for non-null values
		$nonnull_args{$audit_type} = 'V';

		# Get audit_update and audit_user from original form values
		$nonnull_args{$audit_entry}  = $Form_RS->get("$prefix$audit_entry");
		$nonnull_args{$audit_update} = $Form_RS->get("$prefix$audit_update");
		$nonnull_args{$audit_user}   = $Form_RS->get("$prefix$audit_user");

		my $c = (join qq[,\n], keys %nonnull_args)
		  . qq[,\n$key]					# add key
		  . qq[,\n$audit_userstamp]			# add userstamp
		  . qq[,\n$audit_timestamp];			# add timestamp

		my $v = qq[']
		  . (join qq[',\n'], map { Acme::Db->escape_sql($_) } values %nonnull_args)
		  . qq[']
		  . qq[,\n'$keyval']				# add key value
		  . qq[,\n'$User']				# add userstamp
		  . qq[,\ncurrent_timestamp];			# add timestamp

		$sqls->add( qq[insert into $audit_table (\n$c\n) values (\n$v\n);] );
	}

	debug("_build_sql_audit: null_args:",\%null_args);

	if (%null_args) {
		# Set type of record to code for null values
		$null_args{$audit_type} = 'N';

		# Get audit_update and audit_user from original form values
		$null_args{$audit_entry}  = $Form_RS->get("$prefix$audit_entry");
		$null_args{$audit_update} = $Form_RS->get("$prefix$audit_update");
		$null_args{$audit_user}   = $Form_RS->get("$prefix$audit_user");

		my $c = (join qq[,\n], keys %null_args)
		  . qq[,\n$key]					# add key value
		  . qq[,\n$audit_userstamp]			# add userstamp
		  . qq[,\n$audit_timestamp];			# add timestamp

		my $v = qq[']
		  . (join qq[',\n'], map { Acme::Db->escape_sql($_) } values %null_args)
		  . qq[']
		  . qq[,\n'$keyval']				# add key value
		  . qq[,\n'$User']				# add userstamp
		  . qq[,\ncurrent_timestamp];			# add timestamp

		$sqls->add( qq[insert into $audit_table (\n$c\n) values (\n$v\n);] );
	}
}

#-------------------------------------------------------------------------------

sub _fq_field_prefix {
	my ($table, $col, $nrec) = @_;

	#
	# Get the prefix to use to turn column names into the fully-qualified
	# form field names.
	#

	my ($k, $v, $field);

	if ($nrec > 1) {
		$field = "N${nrec}_${table}__${col}" if $nrec;
		$v = $Form_RS->get($field);	# get old value
	}
	else {
		# Try first without number prefix.
		$field = "${table}__${col}";
		$v = $Form_RS->get($field);	# get old value
		if (! defined $v) {
			# Try again with numbered record.
			$field = "N${nrec}_$field";
			$v = $Form_RS->get($field);# get old value
			if (! defined $v) {
				$field = "${table}__${col}";
				$v = $Form_RS->get("${Primary_Table}__${col}");# get old value
			}
		}
	}

	# Try again because we allowed fields not qualified by a
	# table name when filling the record set above.
	# NOTE: This might be superfluous as values saved in $Form_RS
	# are fully-qualified names. See Acme::Form::Filler::_tag_input().

	if (! defined $v && $nrec != 1) {
		$field = $col;
		$v = $Form_RS->get($field);
	}

	my ($prefix) = $field =~ /(.*__)/;

	return $prefix;
}

#-------------------------------------------------------------------------------

sub _prep_do {
	my ($do,$table,$key,$keyval) = @_;

	#
	# Convert "do" property shortcuts and SQL into Perl code.
	#

	# Note: The sandbox where the Perl code is executed has a subroutine
	# sql() defined in it which is used to assemble all the sql statements
	# together to be submitted as one transaction to the database.

	my ($first_word) = $do =~ /^\s*(\w+)/;

	KEYWORD: {

		# Keyword SET updates the current database record.
		if ($first_word =~ /^set$/i) {
			$do = qq[sql(q[update $table $do where $key = '$keyval']);];
			last KEYWORD;
		}

		# SQL keywords are registered to be executed by the database.
		if ($first_word =~ /^(insert|update|delete)$/i) {
			# The $do clause is a full SQL statement.
			$do = "sql(q[$do]);";
			last KEYWORD;
		}

		# Anything else is Perl code and returned intact.
	}
	
	return $do;
}

#-------------------------------------------------------------------------------

sub _execute_sql {
	my ($self,$sqls) = @_;

	#
	# Execute all SQL statements as one transaction.
	#

	debug("_execute_sql: sqls:", $sqls);

	$Dbh->execute($sqls->as_sql)
	  or die "Errors occurred when saving form";
	
	#
	# Commit all new records just added to the database.
	#

	my ($t,$table,$row_name,$id,$key,$keyval,$keyname,$nameval);

	for my $r (@{$sqls->{new_records}}) {
		$r->commit;

		$table  = $r->{table};
		$key    = $r->{key};
		$keyval = $r->{keyval};

		$t = $Schema->t($table);

		$row_name = $t->row_name;

		#
		# Cache the just found key value to use with the
		# cached arguments to detail_result() which were
		# cached in _build_sql().
		#

		push @{$sqls->{detail_result_keys}}, $keyval if $Master;

		#
		# Display results screen
		#

		if ($table eq $Primary_Table) {

			#
			# Compose result message for Primary Table
			#

			if ($id = $t->id) {

				#
				# Display ID number of new record.
				#

				if ($keyname = $t->name_field) {

					#
					# Display name assigned to ID number.
					#

					($nameval) = $Dbh->get(<<EOS);
select $keyname
from $table
where $key = '$keyval'
EOS
					$self->{result} .= <<EOF;
<p>
The $row_name <nobr>ID #<span style="font-size:large">$keyval</span></nobr>
has been assigned to <nobr><span style="font-size:large">$nameval</span></nobr>.
EOF
				}
				else {
					#
					# Display only the ID number added.
					#

					$self->{result} .= <<EOF;
<br>The $row_name record has been added
and assigned <nobr>ID #<span style="font-size:large">$keyval</span></nobr>.
EOF
				}
			}
			else {
				#
				# Display plain result of record added.
				#

				$self->{result} .= <<EOF;
<br>The $row_name record has been added.
EOF
			}

			#
			# Build a "result form" to pass back key, key value,
			# and new timestamp to calling window.
			# 

			my $timestamp = $Form_RS->get('_start') || '';
			$timestamp &&= 
qq[<input type="hidden" name="_timestamp" value="$timestamp">];

			$self->{result} .= 
qq[<form name="result" onSubmit="return false"><input type="hidden" name="key" value="$key"><input type="hidden" name="keyval" value="$keyval">$timestamp</form>];

		}
		else {
			#
			# Compose result message for subordinate tables.
			# Try to use the row index number (1, 2, 3 ...).
			#

			# Execute following if a confirmation message is wanted
			# for every record updated, even the detail rows.
			#my $nf = $t->number_field;
			#my $n = $args->{$nf} if $nf;
			#$n ||= '';
			#$n = " #$n" if $n;
			#$self->{result} .= qq[\n<br>The $row_name record$n has been added.];
		}
	}

	#
	# Compose formatted list of records to hide in the user message window
	# to use to paste into the master form's window. This only applies
	# if this form being saved has a master form.
	#

	my ($detail_args, $detail_key);
	while ($detail_args = pop @{$sqls->{detail_result_args}} ) {
		$detail_key = pop @{$sqls->{detail_result_keys}};
		$self->detail_result(@$detail_args,$detail_key);
	}

	return 1;	# success
}

#-------------------------------------------------------------------------------

sub detail_result {
	my ($self,$t,$args,$keyval) = @_;

	# Return a formatted line of the saved record which can be
	# used to paste into the master record's screen form.

	# Parse master table and field names from master field name
	(my $mn = $Master) =~ s/^[A-Z]_(?=[^_])/$1/;# strip prefix
	$mn =~ s/^\d+_(?=[^_])/$1/;		# strip prefix
	my ($master_table, $master_field) = $mn =~ /(\w*?)__(.*)/;

	my $table  = $t->name;
	my $key    = $t->key;
	$keyval ||= $CGI_RS->get("${table}__${key}");

	my $mt = $Schema->t($master_table);
	my $mf = $mt->f($master_field);

	#
	# Set obj to either the field detail or the table detail object.
	#

	my $obj = ($mf && $mf->detail) ? $mf : $mt;

	#
	# Get format to use to print out record
	#

	my $list_format = $obj->detail_format      || $t->form_list_format;

	#
	# Define SQL to use to retreive the added detail record so that
	# it can be formatted and printed to the result screen.
	#

	my $sqlo = new Acme::SQL $obj->detail_sql;
	$sqlo->add_where("${table}.${key} = '" . Acme::Db->escape_sql($keyval) . "'");

	# Assure detail table key field is in query
	$sqlo->add_column("$table.$key")
	  if ! $sqlo->is_column("$table.$key");

	#
	# Read database
	# 

	my @a = $Dbh->get($sqlo->as_sql);

	#
	# Define bullet with a link to editing detail row.
	#

	my $bullet      = Acme::Config->Bullet;
	my $no_tabindex = Acme::Config->No_Tabindex;

	#
	# Format record into HTML
	#

	my $edit_prog = $t->edit_prog;
	my $row_name  = $t->row_name;

	my $url    = "$edit_prog?table=$table&$key=".escape($keyval)."&master=$Master";
	my $href   = "javascript:_detail_open_${Master}('$url')";
	my $title  = "Edit $row_name.";

	$keyval    = escape_html($keyval);

	no warnings qw(uninitialized);	# turn off warnings for undefs in @a

	my $example_link = qq[<a href="$href" title="$title" class="detail-link" tabindex="$no_tabindex">$bullet</a>];
	my $example_item = Local::Acme::Format::sprintf_cmd($list_format,@a);

	# Surround $example_item with <td> </td> tags if not already surrounded.

	$example_item = qq[<td class="detail-item">$example_item</td>]
	  if $example_item !~ /^\s*<td/i;

	$self->{detail_result} .=
	    qq[<div style="display:none">]
	  . qq[<table>]
	  . qq[<tbody id="dtbody">]
	  . qq[<tr id="_detail_${Master}_$keyval">\n]
	  . qq[<td class="detail-link">$example_link</td>\n]
	  . $example_item
	  . qq[</tr>]
	  . qq[</tbody>]
	  . qq[</table>]
	  . qq[</div>\n];
}

#-------------------------------------------------------------------------------

sub _reconcile {
	my ($self, $t, $args, $nrec) = @_;

	require Acme::Form::Reconcile;

	Acme::Form::Reconcile::display_reconcile_form($Form_RS, $t, $args, $nrec);

	exit 0;		# Terminate program
}

1;
