package Acme::Form::Schema;

=head1 NAME

Acme::Form::Schema - Form specification. A sub-class of Acme::Base::Schema.

=head1 SYNOPSIS

    use Acme::Form::Schema;

    my $fs = new Acme::Form::Schema { form => 'filename' };


=head1 DESCRIPTION

I<Acme::Form::Schema> is a sub-class of I<Acme::Base::Schema>. It allows
a form specification file to override database field definitions.
It can be used to build forms using the exact same syntax used to
describe the database in the schema files.

Any table referenced in the form specification file that is not also
defined in the file is looked up in the default database schema
files.

The form schema file is located in the $Forms directory which is a configurable
variable defined in the site configuration file (../etc/site.conf).

=head1 CHANGE LOG

 3/21/02  Initial version.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

use Acme::Base::Schema;
use Acme::Config;

#------------------------------------------------------------------------------

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my $form_file;		# file name of form schema
	if (defined $_[0]) {
		if (ref $_[0] eq 'HASH') {
			# Decode options hash
			$form_file = $_[0]->{form} if $_[0]->{form};
		}
	}

	#
	# Because Acme::Base::Schema is cached, we must delegate, not inherit.
	# Build an object that contains two Acme::Base::Schema objects, one
	# schema built from the form specification file, the other from the
	# default database schema file(s).
	#
	# Note that because Acme::Base::Schema names its cache after the file
	# name, form specification files end up being cached too. That might not
	# be desired in mod_perl. There is a "Cache no" or a "Nocache" property
	# that can be specified first in the form schema file to prevent the
	# automatic caching.
	#

	my $self = {
		form_schema => ($form_file
				? new Acme::Base::Schema { file => $form_file }
				: undef),
		db_schema   => new Acme::Base::Schema @_,
		};

	# Link this node as a parent to Schema nodes
	$self->{form_schema}->{'.parent'} = $self if $self->{form_schema};
	$self->{db_schema}->{'.parent'} = $self;

	return bless $self, $class;
}

sub schema { return shift }

sub tables {
	my $self = shift;

	# The tables() method returns the tables in the schema.
	# Combine the list from both the form schema file and the database
	# schema. Use a hash to eliminate duplicate names. The second grep
	# is used to eliminate undefs and avoid -w warnings.

	# Note that the list of tables from the database schema are checked
	# with security to see if the user has access to them. The tables
	# listed in the form schema file are NOT checked with security.

	my %seen;
	return grep { $_ if ! $seen{$_}++} grep {$_}
		($self->{form_schema} ? $self->{form_schema}->tables : undef),
		$self->{db_schema}->tables;
}

sub DESTROY {
	my $self = shift;
	$self->{form_schema}->DESTROY(@_) if $self->{form_schema};
	$self->{db_schema}->DESTROY(@_)   if $self->{db_schema};
}

sub AUTOLOAD {
	my $self = shift;
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;

	return $self->{form_schema}->$method(@_) || $self->{db_schema}->$method(@_)
	  if $self->{form_schema};
	return $self->{db_schema}->$method(@_);
}

1;
