package Acme::Form::Vars;

=head1 NAME

Acme::Form::Vars - Access database values for form fields.

=head1 SYNOPSIS

    use Acme::Form::Vars;

    my $var = new Acme::Form::Vars;

    $var->set_var(var_name => var_value);

    # Three ways to access the variables. The first is the most efficient
    # (but less object oriented).

    my $name = $Acme::Form::Vars::customers__customer_name;
    my $name = $var->get_var('customers__customer_name');
    my $name = $var->customers__customer_name;

    # Variable names follow the same naming convention of form field names.
    my $name2 = $Acme::Form::Vars::N2_customers__customer_name;

=head1 DESCRIPTION

This module performs the mapping of form fields names to their corresponding 
database values.

This module does not follow normal "good programming practices" for object
oriented modules. The object's variables are stored as package variables.
The object's constructor and destructor routines initialize and remove the
package variables, but the explicit removal of package variables only needs
to be done in a mod_perl environment.

This module provides a sandbox in which to execute the Acme::Base::Schema
C<Do> or C<Do_filler> (or C<Do_Saver>) properties' code.  The code in the
schema files can reference form field values by accessing Perl variables
with the same name as the field name. For example, the code accesses
the value of the form field with the name of "customers__customer_name"
through the variable C<$customers__customer_name>.  Execution of the C<Do...>
properties occurs before the form field's are filled in. Subroutines defined
in C<Do> or C<Do_filler> can be invoked by the C<Do> form directive
C<<!--$ Do routine($arg1,$arg2,$arg3) -->> placed in the schema files.

=head1 CHANGE LOG

 4/1/02  Initial version.
 4/14/02 Added get_vref() to return references to variables instead of their values.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

#-------------------------------------------------------------------------------
#
# Tie scalar variables.
# Stolen from standard Perl module Tie::ScalerSTD.
#
#-------------------------------------------------------------------------------

use Tie::Scalar;
our @ISA = qw(Tie::Scalar);

sub TIESCALAR {
    my $class = shift;
    my $instance = shift || undef;
    return bless \$instance => $class;
}

sub FETCH {
    return ${$_[0]};
}

sub STORE {
    ${$_[0]} = $_[1];
}

sub DESTROY {
    
    # Note: this gets called for every scalar tied above and for the object
    # created by new() below. Ignore the error "Not a scalar..." when
    # destroying the object.

    eval { undef ${$_[0]} };
}

#-------------------------------------------------------------------------------

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my $self = {
		var     => {},	# hash of form variables defined
		nvar    => 0,	# number of variables defined
		nrec    => 0,	# number of records of variables
		import  => {},	# hash of non-form variables defined (imported)
		nimport => 0,	# number of non-form variables imported
		};

	bless $self, $class;

	$self->import_var(@_) if @_;

	return $self;
}

sub set    { return set_var(@_) };
sub get    { return get_var(@_) };
sub import { return import_var(@_) };

no strict 'refs';

sub import_var {
	my $self = shift;
	my $class = ref($self) || $self;

	my ($k,$v);
	while ($k = shift) {
		$v = shift;
		
		# Track variables imported and their import order.
		$self->{import}->{$k} = ++$self->{nimport};

		# Set variable
		eval qq[tie \$${class}::$k, 'Acme::Form::Vars'; \$${class}::$k = \$v];
	}
}

sub set_var {
	my $self = shift;
	my $class = ref($self) || $self;

	my ($k,$v);
	while ($k = shift) {
		$v = shift;
		
		# Track variables set and their order.
		$self->{var}->{$k} = ++$self->{nvar};

		# Track number of records saved.
		# Record number is the variable's prefix.

		if ($k =~ /^N_?(\d+)_(?!_)/) {
			$self->{nrec} = $1 if $1 > $self->{nrec};
		}

		# Set variable
		eval qq[tie \$${class}::$k, 'Acme::Form::Vars'; \$${class}::$k = \$v];
	}
}

sub get_var {
	my $self = shift;
	my $class = ref($self) || $self;

	# Return specified global variables
	if (wantarray) {
		return map { $_ = "${class}::$_" unless /::/; $$_ } @_;
	}
	else {
		$_ = shift;
		$_ = "${class}::$_" unless /::/;
		return $$_;
	}
}

sub get_vref {
	my $self = shift;
	my $class = ref($self) || $self;

	# Return references to specified global variables
	if (wantarray) {
		return map {$_ = "${class}::$_" unless /::/; \$$_} @_;
	}
	else {
		$_ = shift;
		$_ = "${class}::$_" unless /::/;
		return \$$_;
	}
}

sub AUTOLOAD {
	my $self = shift;
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;

	return $self->get_var($method);
}

1;
