package Acme::HTML;

=head1 NAME

Acme::HTML - HTML User Interface modules for Acme Base Company

=head1 SYNOPSIS

    use Acme::HTML qw(display_message	display_error
    		    escape_html		unescape_html
		    message		error);

    display_message("This is a non-informative message.",	# message
		    "User Message",				# title
		    "clear_form()",				# callback
		    1.5);					# time to live

    display_error("This is an error message.",			# message
		  "Error",					# title
		  "clear_form()";				# callback

    # Or if you want print it yourself:
    print message("This is a non-informative message.",		# message
		  "User Message",				# title
		  "clear_form()",				# callback
		  1.5);						# time to live

    print error("This is an error message.",			# message
		"Error",					# title
		"clear_form()";					# callback


    #
    # Methods used for debug
    #

    use Acme::HTML qw(debug debug_dump);

    debug_on();

    debug("This is a debug message");		# print debug output if $Debug

    debug_dump("My Data:",\%structure);		# dump a memory structure

    debug_stack();				# dump subroutine call stack

    debug_off();

    #
    # Methods used for debug probes
    # 

    # These methods ignore debug on/off settings.

    dump_msg("This is a debug message");	# print debug output

    dump_data(\%struc1,"Struct2",\%struct2);	# dump data structure(s)

    dump_stack();				# dump subroutine call stack

    #
    # HTML string conversion
    #

    my $encoded_string = escape_html( qq(weird "&%" characters) );
    print qq(<input type=text value="$encoded_string">);

    my $string = unescape_html($encoded_string);


=head1 DESCRIPTION

This module sets up a program environment and defines a few common 
utility routines for interacting with the user using HTML. 
Currently (2/2/00), Dynamic HTML version 4.0 is used.

The CGI::Carp module is initialized so that errors get sent to the
user in addition to the error log. 

Display_message() displays a message in a browser window. An "OK" button
is displayed which closes the window when clicked.

If the call back function is defined (a string, not a function reference,
as the function is a JavaScript function, not a Perl function),
the function is called in the calling window when the user presses OK.

The message is displayed only for a second or two (as determined by the
user -- see I<Acme::Options)> unless it is specified in the function call.
Setting the time to 0 keeps the window open until the user manually
closes the window.

Display_error displays an error message much like the display_message()
routine. It is the routine used to display error messages to the browser
when the program calls die() or croak(). Note that the CGI::Carp module
escapes all HTML codes in error messages processed by die() or croak().
Only the plain text is displayed by display_error() -- no links or fun stuff.

The message() and error() return the HTML code for a message window or
error display without printing it.

The escape_html() routine returns the given string encoded as a safe HTML
string -- non-ascii characters are encoded as HTML entities. In other words,
characters like LF are encoded like `&#010;' not `%0A'. Unescape_html()
undoes escape_html().

The debug routines: debug_on(), debug(), and debug_off() control debug output.
The debug() routine checks its arguments and will output them to the browser
appropriately. The dump_data() routine works like debug() except it does
not check if the debug flag is set. It is useful for one-time probes.
The dump_stack() routine dumps the calling stack so you can see which 
routines called which and where.

=head1 CHANGE LOG

 3/22/02 Added debug methods.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1997-2000, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

require Exporter;

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(message		error
		    display_message	display_error
		    escape_html		unescape_html
		    debug_on		debug_off
		    debug		debug_dump		debug_stack
		    dump_msg		dump_data		dump_stack
		   );

use Acme::Options;

use CGI::Carp qw(fatalsToBrowser set_message);	# for error messages

BEGIN {

	# Setup custom user error message handler for CGI::Carp::croak().

	sub compile_error {
		my $msg = shift;

		my($wm) = $ENV{SERVER_ADMIN} ?
		  qq[the webmaster (<a href="mailto:$ENV{SERVER_ADMIN}?subject=Mystery%20Errors">$ENV{SERVER_ADMIN}</a>)] :
		  "this site's webmaster";

		print <<EOF;
<h1>Software Error</h1>
<pre>$msg</pre>
<p>
For help, please send mail to $wm,
giving this error message and the time and date of the error.
</p>
EOF
	}
	set_message(\&compile_error);
}

#------------------------------------------------------------------------------

our $Debug = 0;				# set to 1 to turn on debug output

sub debug_on  { $Debug = 1 }		# turn on debug output
sub debug_off { $Debug = 0 }		# turn off debug output

sub debug { dump_data(@_) if $Debug }

sub debug_dump { dump_data(@_) if $Debug }

sub debug_stack { dump_stack(@_) if $Debug }

#
# The following debug routines can be used for temporary debug probes.
# They do not pay attention to the $Debug flag.
#

sub dump_msg { print "<br> --- ", map { escape_html($_) } @_, "\n" }

sub dump_data {
	for (@_) {
		if (ref $_) {
			# Dump data structure
			require Data::Dumper;
			print qq[<pre style="text-align:left;font-size:9pt">],
			  escape_html(Data::Dumper::Dumper($_)),
			  qq[</pre>\n];
		}
		else {
			# Display message
			dump_msg($_);
		}
	}
}

sub dump_stack {
	print qq[<pre>\nStack Trace:\n];
	my $frame = 1;
	while ( my ($pkg, $file, $line, $sub) = caller($frame++)) {
		print qq[    $file: line $line: ${pkg}::${sub}()\n];
	}
	print qq[</pre>\n];
}

#------------------------------------------------------------------------------

sub message {
	my ($message, $title, $javascript_callback, $open_time, $error_flag) = @_;

	$message ||= "(no message specified by the program)";
	$title ||= "Message";
	$javascript_callback ||= '';	# JavaScript callback function call
					# note: specify both () and arguments
	$error_flag ||= 0;		# hack to allow display_message() to show errors

	if (not defined $open_time) {
		# Get default time to keep window open from user options.
		$open_time = Acme::Options->new->MessageWindowAutoCloseTime || 0;
	}
	$open_time *= 1000;	# change seconds into milliseconds

	#
	# Define different components of message
	#

	my ($message_begin, $message_end);
	if ($error_flag) {

		#
		# Surround error messages with 3 red lines.
		#

		$message_begin = <<EOF;
<table width="100%" border="1" cellspacing="0" cellpadding="2" align="center" bordercolor="red"><tr><td>
  <table width="100%" border="1" cellspacing="0" cellpadding="2" align="center" bordercolor="red"><tr><td>
    <table width="100%" border="1" cellspacing="0" cellpadding="20" align="center" bordercolor="red"><tr><td>
EOF
		$message_end = <<EOF;
    </td></tr></table>
  </td></tr></table>
</td></tr></table>
EOF
	}
	else {
		#
		# Surround informative messages with a black box.
		#

		$message_begin = <<EOF;
<table width=100% border=1 cellspacing=0 cellpadding=10><tr><td>
EOF
		$message_end = <<EOF;
</td></tr></table>
EOF
	}
		
	my $callback = '';
	if ($javascript_callback) {
		my ($javascript_func) = $javascript_callback =~ /(.+?)\s*\(/;

		$callback = <<EOF;
		self.blur();

		if (window.opener
		&& ! window.opener.closed
		&& window.opener.$javascript_func) {
			window.opener.$javascript_callback;
		}
EOF
	}

	my ($timeout,$set_timeouts,$set_timeouts_func) = ('','','');

	if (! $error_flag && ! $Debug) {

		$timeout = $open_time
			 ? "setTimeout('do_exit()',$open_time);"
			 : "// (disabled)";

		$set_timeouts = 'set_timeouts()';

		$set_timeouts_func = <<EOF;

	function set_timeouts() {

		// Allow calling window to specify number of seconds before
		// automatically closing this window in the global variable
		// MessageWindowAutoCloseTime. Setting it to 0 forces window
		// to stay open.

		if (window.opener
		&& ! window.opener.closed
		&& typeof window.opener.MessageWindowAutoCloseTime == 'number') {
			var open_time = window.opener.MessageWindowAutoCloseTime * 1000;
			if (open_time > 0) setTimeout('do_exit()',open_time);
			if (open_time < 0) do_exit();  // exit immediately
			return;
		}

		// Set auto close timeout as per user option setting.
		$timeout
	}
EOF
	}

	#
	# Assemble message HTML
	#

	return <<EOF;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
"http://www.w3.org/TR/html4/loose.dtd">

<html>
<head>
<title>$title</title>

<link rel="stylesheet" href="/styles/message.css" type="text/css">

<script language="JavaScript" type="text/javascript">
	window.document.onkeypress = do_keypress;

	function do_keypress(evt) {
		switch (window.event.keyCode) {
		case 27:	// escape key
			self.close();
			return true;
		case 10:	// ctrl-enter key
			do_exit();
		}
	}

	function do_exit() {
		// terminate and clean up
$callback
		self.close();
	}
$set_timeouts_func
</script>

</head>
<body onLoad="self.focus();f=self.document.forms.length-1;self.document.forms[f].elements[ self.document.forms[f].elements.length-1 ].focus();$set_timeouts">

$message_begin

<div class="message">$message</div>

<form action="javascript:void 0"><!-- safety to terminate active form if any --></form>
<form action="javascript:void 0" onSubmit='do_exit();return false'>
<div class="button"><input type=submit value='  OK  '></div>
</form>

$message_end

</body>
</html>
EOF
}

#------------------------------------------------------------------------------

sub error {
	my ($message,$title,$javascript_callback) = @_;
	$message ||= "An error occurred.";
	$title   ||= "Error";

	# Surround message with <p> tags and make blank lines paragraphs.
	$message = "<p>$message</p>";
	$message =~ s|\n|</p><p>|gs;

	dump_stack() if $Debug;

	return message($message,$title,$javascript_callback,0,"error");
}

#------------------------------------------------------------------------------

sub display_message { print message(@_) }
sub display_error   { print error(@_)   }

#------------------------------------------------------------------------------

#
# Use the fancier display_error routine now.
#

# Note: Setting $SIG{__DIE__} below prevents cascading errors from appearing in
# user's error message. However, it creates an error "Unable to create sub
# named *CORE::die" when called from the Safe module, which is where the 
# search scripts are executed.
#set_message(\&display_error);
set_message( sub {display_error(@_); $SIG{__DIE__} = \&CORE::die;} );


#------------------------------------------------------------------------------

######
# Note: Mysterious bug with CGI escapeHTML in that I have a '&&' translated
# into only one &amp;
#
# CGI escapeHTML() and unescapeHTML() differ from the routines below in that
# non-printing characters are also encoded in escape_html(), and the routines
# below always returns a defined string.
######

#use CGI qw(escapeHTML unescapeHTML);
#
#sub escape_html   { escapeHTML(@_)   || '' }
#sub unescape_html { unescapeHTML(@_) || '' }

sub escape_html {

	# Return string with non-ascii characters encoded into HTML entity
	# characters, such as &#010; for the line-feed character.

	my $s = shift;
	return '' unless defined($s);	# return is always a defined string
	$s =~ s/([^\w !#$%&()*+,.\/:;=?@[\\\]^`{|}~-])/sprintf("&#%03d;",ord($1))/egs;
	return $s;
}

#------------------------------------------------------------------------------

# Only entities mapping to standard ascii characters are decoded here.
our %_html_entities = (
	lt   => '<',
	gt   => '>',
	amp  => '&',
	quot => '"',
);

sub unescape_html {
	my $s = shift;
	return undef unless defined($s);

	# substitute code word constants
	$s =~ s/&(\w+);/$_html_entities{$1} ? $_html_entities{$1} : "&$1;"/egs;

	# substitute number coded constants
	$s =~ s/&#(\d{1,3});/pack('c',$1)/egs;
	return $s;
}
1;
