package Acme::HTML::HideMoreLess;

=head1 NAME

HideMoreLess - Hide document sections with More/Less buttons to display/hide it

=head1 SYNOPSIS

    use Acme::HTML::HideMoreLess qw(start_hiding stop_hiding);

    print start_hiding("block_id_name");
    print "other stuff ..."
    print stop_hiding();

      -or-

    print start_hiding("block_id_name",
    	button_label=>"Instructions",
    	button_description=>"Hide/Show Instructions",
		      );
    print "other stuff ..."
    print stop_hiding();

      -or-

    print start_hiding("block_id_name",
    	more_label=>"Show",
    	more_description=>"Show this section.",
    	less_label=>"Hide",
    	less_description=>"Hide this section.",
		      );
    print "other stuff ..."
    print stop_hiding();

=head1 DESCRIPTION

These routines return the HTML code to display a "More >>" button and
start a block that version 4 browsers will hide from view. All output
to the browser is hidden until stop_hiding() is executed.

Flags are kept so that stop_hiding() can be called repeatedly without
consequence.

Button labels can be specified when calling start_hiding(). The button label is
used to keep the button labels straight when there are more than one group of
more/less buttons are used.

=head1 CHANGE LOG

 04/1/99 Initial version.
 09/6/00 Change button styles and use <button> instead of <a> tags.
         Allow specification of button labels in start_hiding().
 7/7/02  Simplified buttons. Use <a> tags instead of <button>.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1999-2002 Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.000;
require Exporter;

@ISA = qw(Exporter);
@EXPORT_OK = qw(start_hiding stop_hiding);

local %started;		# flags set to 1 when block is hiding
local %labels;		# button labels and descriptions

#------------------------------------------------------------------------------

sub start_hiding {

	my $id = shift;		# button/division ID

	$labels{$id} = {
		button_label => '',	  # init for argument specification
		button_description => '', # init for argument specification
		more_label => 'More',
		more_description => 'Show more information.',
		less_label => 'Less',
		less_description => 'Show less information.',
		@_			  # my argument definitions
	};

	my $l = $labels{$id};
	my $more_label = $l->{button_label} || $l->{more_label};
	my $more_description = $l->{button_description} || $l->{more_description};
	my $less_label = $l->{button_label} || $l->{less_label};
	my $less_description = $l->{button_description} || $l->{less_description};

	$started{$id} = 1;	# set to true

	#
	# Display a "More >>" button and begin a new hidden <div> tag.
	# When the user presses the "More >>" button, it is hidden and
	# a "<< Less" button is displayed along with the block that was
	# hidden.
	#

	return <<EOF;
<!-- Start Hiding -->
<script language="JavaScript" type="text/javascript">
function show_$id() {
	var c;
	c = document.getElementById('${id}_more_button_section');
	c.style.display = "none";
	c = document.getElementById('${id}');
	c.style.display = "block";	
}
function hide_$id(goto_begin) {
	var c;
	c = document.getElementById('${id}');
	c.style.display = "none";       
	c = document.getElementById('${id}_more_button_section');
	c.style.display = "block";

	if (goto_begin) {
		// Scroll to make beginning button visible at bottom of page.
		c.scrollIntoView(false);
		window.scrollBy(0,50);
	}
}
</script>

<div id="${id}_more_button_section" style="display:block">

<!-- Begin More Button -->
<span class="button-bar"><table cellspacing="0" cellpadding="0" border="0"><tr>
 <td><a href="javascript:show_$id()" title="$more_description" tabindex="-1">$more_label&nbsp;<font class="button-chevron">&raquo;</font></a></td>
</tr></table></span>
<!-- End More Button -->
</div>

<div id="$id" style="display:none">

<!-- Begin Less Button -->
<span class="button-bar"><table cellspacing="0" cellpadding="0" border="0"><tr>
 <td><a href="javascript:hide_$id()" title="$less_description" tabindex="-1"><font class="button-chevron">&laquo;</font>&nbsp;$less_label</a></td>
</tr></table></span>
<!-- End Less Button -->

EOF
}

#------------------------------------------------------------------------------

sub stop_hiding {
	my $id = shift;
	my $show_less_button = shift;

	return '' if (defined($id) && ! (defined($started{$id}) && $started{$id}));

	my $button = '';
	my $initialize_less2_script = '';

	if ($show_less_button) {
		my $l = $labels{$id};
		my $less_label = $l->{button_label} || $l->{less_label};
		my $less_description = $l->{button_description} || $l->{less_description};

		return <<EOF;
<!-- Begin Less Button -->
<span class="button-bar"><table cellspacing="0" cellpadding="0" border="0"><tr>
 <td><a href="javascript:hide_$id()" title="$less_description" tabindex="-1"><font class="button-chevron">&laquo;</font>&nbsp;$less_label</a></td>
</tr></table></span>
<!-- End Less Button -->
</div>
<!-- Stop Hiding -->
EOF
	}
	else {
		return "</div>\n";
	}
}

1;
