package Acme::HTML::InputArray;

=head1 NAME

Acme::HTML::InputArray - return HTML code to enter an array of values

=head1 SYNOPSIS

  # Make array input fields from database table values
  print Acme::HTML::InputArray->Make(
  				  'array_field',	# db array field name
				  $default_values,	# initial values
				  $filter,		# array item labels
				  $attributes,		# extra attributes
				  );

=head1 DESCRIPTION

The methods in this package returns HTML code for entering a database array.
A database array is a text string that begins, ends, and is delimited by a 
punctuation character. For example: |1|2|3|

The Make() method constructs input tags from a database schema description
(see I<Acme::Base::Schema>). The $default_values is a text array that stores
default values for the <input> tags. The $filter is a text array of tag names.
They are (should be) a subset of the codes in the array, hash, or sql script
(Make()'s second argument). If present, they are the actual labels that are
displayed for entry. Other fields in the input array are hidden from view and
are disabled so that they do not send values to the server when form is posted.

The Make() method assumes that the form field name has the database table and
field name embedded in it like: 'table__field'. For situations where the
database table name is not a part of the form field name, you can set the
$Acme::HTML::InputArray::Table global variable.

=head2 Note

A limitation with Make() is that it does not work if there are more than
one form in the document unless the generated Javascript is hand edited.
An alternative is to set the global variable $Acme::HTML::InputArray::Form
to the name of the form if known.

=head1 CHANGE LOG

 4/5/01   Initial version.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1998-2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Acme::HTML		qw(escape_html);
use Acme::HTML::Tag;
use Acme::Base::Schema;
use Acme::Db;
use Acme::TextArray;
use Acme::Form::Context;
use Acme::Config;

#
# Define global which can be set to supply a database table name if it is not
# embedded in the $array_field name argument.
#

$Acme::HTML::InputArray::Table = '';

#
# Define global to supply a default name for HTML forms.
#

$Acme::HTML::InputArray::Form = 'forms[0]';

#-------------------------------------------------------------------------------

sub Make {
	my $self = shift;

	my ($array_field,$default,$filter,$attributes) = @_;

	my $tag = new Acme::HTML::Tag $array_field;
	my $tabindex = $tag->tabindex;

	#
	# Extract database table / field name from the form tag name.
	#

	my $context = new Acme::Form::Context
			table => $Acme::HTML::InputArray::Table,
			form  => "self.document.$Acme::HTML::InputArray::Form";

	my $db_name = $context->form_to_db($tag->name);

	my ($table,$field) = $db_name =~ /(.*)\.(.*)/;
	$table ||= $Acme::HTML::InputArray::Table;	# global default
	$field ||= $db_name;
	return unless $table;

	my $s = new Acme::Base::Schema $table or return;

	my $t = $s->t($table) or return;

	my $f = $t->f($field) or return;

	#
	# Collect variables needed to construct <input> tags and controlling
	# JavaScript scripts.
	#

	$default ||= $tag->value || $f->default;

	$filter     ||= $f->input_filter;
	$attributes ||= $f->input_attributes;
	$attributes   = " $attributes" if $attributes; # insert delimiter space
	my $maxlength = $f->input_length;
	my $size      = $maxlength || 10;
	my $decimal   = $f->input_decimal;
	my $class     = $f->input_class || 'input-array';
	my $sep       = $f->array_separator || ';';

	$filter = $context->db_to_form($filter);

	#
	# Get list of codes, and labels for <input> tags
	#

	my @codes = Acme::TextArray::split_array($f->input_array);
	my @labels;
	if (@codes) {
		@labels = Acme::TextArray::split_array($f->input_labels);
	}
	else {
		#
		# Read codes (and possible labels) from database.
		#

		my $sql = $f->input_sql;
		my (@col, $code, $label);
		my $q = Acme::Db->query($sql) or return;
		while ((@col) = $q->next) {
			# Collect field labels from the second element of 
			# the sql fields, as that is usually the column used
			# for description columns in drop down lists.
			push @codes, $col[0];
			push @labels, $col[($#col > 0 ? 1 : $#col)];
		}
	}

	#
	# Set up form field validation routines
	#

	my ($translate, $validate);
	my ($translate_str, $check_str, $validate_str, $compute_str, $form_compute_str);
	my ($on_change_str, $on_focus_str);

	$translate = escape_html($f->input_translate);
	$validate = escape_html($f->input_valid);
	$valid_error = escape_html($f->input_valid_error);

	$translate_str = $translate
	  ? qq(translate_input(input_field,'$translate'))
	  : '';
	
	$check_str = qq(check_input(input_field,'$class',$maxlength,$decimal));

	$validate_str = $validate
	  ? qq(validate_input(input_field,'$validate'))
	    . ($valid_error ? ",$valid_error" : '')
	    . ')'
	  : '';

	# All of the event scripts for each input field of array is consolodated
	# into one JavaScript function. 

	$compute_str = $context->translate($f->input_onchange, tr_script=>1);
	$compute_str =~ s/this/input_field/gs;		# change 'this' keyword
	$compute_str = "($compute_str)" if $compute_str;# surround with ()

	# form_compute is called after a change to an input field below
	#$form_compute_str = qq(form_compute(input_field.form));	# always
	$form_compute_str = qq();	# never (done in ..._onChange() below)

	$on_change_str = join " && ", grep { $_ } $translate_str, $check_str,
						  $validate_str,  $compute_str,
						  $form_compute_str;

	$on_focus_str = $context->translate($f->input_onfocus,tr_script=>1) || 'false';
	$on_focus_str =~ s/this/input_field/gs;		# change 'this' keyword
	$on_focus_str =~ s/;\s*$//s;			# remove trailing ;

	#
	# Assemble input array <input> tags and controlling JavaScript.
	#

	my $tag_name   = $tag->name;
	my $array_name = "I_$tag_name";		# name of JavaScript array

	my $html = qq($array_field<!-- Begin input array for $tag_name --><table class="input-array" cellspacing="0" cellpadding="0">\n  <tr>);

	my $scripts = '';

	my $input_fields = '';
	my $input_labels = '';

	my ($input_name, $input_label, $input_description, $code, $elem);

	for (my $i = 0; $i < @codes; $i++) {

		#
		# Assemble <input> tags
		#

		$input_name = "I_${i}_".$tag->name;
		##$input_label = $labels[$i] || ''; # use description as label
		$input_label = $codes[$i] || '';    # use code as label
		$input_description = $f->input_description || $labels[$i];

		$input_labels .= <<EOF;
    <td><div class="input-array-label" id="l$input_name">$input_label</div></td>
EOF
		$tabindex++;
		$input_fields .= <<EOF;
    <td><div class="input-array-field" id="f$input_name"><input type="text" name="$input_name" value="" class="$class" size="$size" tabindex="$tabindex" onChange="${tag_name}_onChange(this)" onFocus="${tag_name}_onFocus(this)" title="$input_description"$attributes></div></td>
EOF
		#
		# Collect JavaScript fragments
		#

		$elem = $array_name . "[$i]";		# array element
		($code = $codes[$i]) =~ s/'/\\'/g;	# escape quotes

		$scripts .= <<EOF;
$elem = '$code';
EOF
	}

	#
	# Assemble Javascript 
	#

	$scripts = qq(<script language="JavaScript" type="text/javascript">

/*
 * Scripts supporting the input array for $tag_name.
 */

// If this is not the first form in the document, change the following
// statement to specify the name of the form.
// var Form_${tag_name} = 'name_of_form';
var Form_${tag_name} = self.document.${Acme::HTML::InputArray::Form}.name;

// Define the separator character (delimiter) to use for building text arrays.
var Sep_${tag_name} = '${sep}';

// Define an object to provide quick lookup of shown field names, those
// which are in the filter applied by assign_${tag_name}().
var Lookup_${tag_name} = new Object;

var ${array_name} = new Array();
$scripts
function assign_${tag_name}(name_ta, value_ta) {

	// Display only those input fields of an input array which are
	// for entering the codes listed in the name_ta text array argument.
	var names = split_array(name_ta);
	var values = split_array(value_ta);

	// Build object for quick lookup of names and values.
	var i, val;
	Lookup_${tag_name} = new Object;
	for (i = 0; i < names.length; i++) {
		val = values[i] ? values[i] : '';
		Lookup_${tag_name}\[names[i]] = val;
	}
	
	var f = self.document.forms[Form_${tag_name}];

	var fn, fs, ls, e;  // field name, field div id, label div id, element
	for (i = 0; i < ${array_name}.length; i++) {
		// Construct name of input field from index i
		fn = 'I_'+i+'_${tag_name}';	// name of input field
		fs = 'fI_'+i+'_${tag_name}';	// id of field div
		ls = 'lI_'+i+'_${tag_name}';	// id of label div

		if (typeof Lookup_${tag_name}\[${array_name}\[i]] == 'undefined') {
			// Code is not in name_ta
			// Hide and clear input field
			e = f.elements[fn];
			e.value = '';
			e.disabled  = true;

			// Hide field
			e = document.getElementById(fs);
			e.style.display = 'none';
			// Hide label
			e = document.getElementById(ls);
			e.style.display = 'none';
		} else {
			// Code is in name_ta
			// Show input field and set to value
			val = Lookup_${tag_name}\[${array_name}\[i]];
			e = f.elements[fn];
			e.value = val;
			e.disabled  = false;

			// Show field
			e = document.getElementById(fs);
			e.style.display = 'block';
			// Show label
			e = document.getElementById(ls);
			e.style.display = 'block';
		}
	}
}

function set_${tag_name}() {
	// Set contents of ${tag_name} to values of the input array.

	var ta = Sep_${tag_name};	// initialize result text array

	var f = self.document.forms[Form_${tag_name}];

	for (var i = 0; i < ${array_name}.length; i++) {
		if (!(typeof Lookup_${tag_name}\[${array_name}\[i]] == 'undefined')) {
			// Code is in Lookup fields
			ta += f.elements['I_'+i+'_${tag_name}'].value + Sep_${tag_name};
		}
	}
	if (ta == Sep_${tag_name}) ta = '';	// no elements, set to empty
	f.${tag_name}.value = ta;
}

function ${tag_name}_onFocus(input_field) {
	// Common onFocus event handler for all of the array's input fields.
	if ($on_focus_str) {
		set_${tag_name}();	// set controlling field contents
		input_field.select();
		return true;
	} else {
		return false;
	}
}

function ${tag_name}_onChange(input_field) {
	// Common onChange event handler for all of the array's input fields.
	if ($on_change_str) {
		set_${tag_name}();	// set controlling field contents
		return form_compute(document.forms[Form_${tag_name}]);
	} else {
		return false;
	}
}


// Initialize input array.
with (document.forms[Form_${tag_name}]) { assign_input(${tag_name},$filter.value) }

</script>);

	#
	# Assemble final html for input array
	#

	$html .= qq(
$input_labels  </tr>
  <tr>
$input_fields  </tr></table>$scripts<!-- End input array for $tag_name -->);

	return $html;
}

1;
