package Acme::HTML::ListMenus;

=head1 NAME

Acme::HTML::ListMenus - return HTML code for drop down lists

=head1 SYNOPSIS

  use Acme::HTML::ListMenus;

  # Generic list maker interfaces
  print Acme::HTML::ListMenus->List('select * from codes', 'tagname');
  print Acme::HTML::ListMenus->Menu('select code from codes', 'tagname');
    or
  print Acme::HTML::ListMenus->List(\%hashlist, 'tagname');
  print Acme::HTML::ListMenus->List(\@arraylist,'tagname');
  print Acme::HTML::ListMenus->Menu(\@arraylist,'tagname');
  print Acme::HTML::ListMenus->Menu(\%hashlist, 'tagname');
    or
  print Acme::HTML::ListMenus->List('select * from codes', 'tagname');
    or
  print Acme::HTML::ListMenus->List('select * from codes',
  				    'tagname',		# name of control
				    $default,		# default value
				    $size,		# width (char)
				    $attributes,	# <select> attributes
				    $format,		# sprintf() like format
				    $flag,		# 'multi' or 'noundef'
				    \@columns,		# sql names to make
				    );			# assignment statements

    or

  # Make list from database table contents
  print Acme::HTML::ListMenus->Make('table_name',
  				    'tagname',		# name of tag
				    $default,		# default value
				    $size,		# width (char)
				    $attributes,	# <select> attributes
				    $format,		# sprintf() like format
				    $flag		# 'multi' or 'noundef'
				    );

    or

  use Acme::HTML::ListMenus qw(list menu);

  print list(\%hashlist,'tagname',$default,$size,$attributes,$format);
  print menu(\@arraylist,'tagname',$default,$size,$attributes,$format,'multiple');


  Note: This module is derived from I<Local::Acme::ListMenus>. Any methods not
  defined in this module are referred to I<Local::Acme::ListMenus>, which if
  it needs to, assembles a list from database tables using methods defined here.
  This allows sites to customize and override drop-down generated lists by
  defining them in I<Local::Acme::ListMenus>.

=head1 DESCRIPTION

The methods in this package returns HTML code for a selection list of options
that's usually obtained from database control tables. 

The first argument to List() and Menu() can be an SQL statement, an array
of values, or a hash of values and display text.

There is a problem with using hashes in that they loose the order of
their definition and the drop-down lists are displayed in semi-random
order. Therefore, the lists from hashes are sorted by their display
values. However, that is not always desired either, so if the beginning of
every display value is a number followed by a space, the number (and space)
is stripped off before displaying the list. Every display value must be
preceded by a number otherwise all the values are displayed as defined.

The second argument is the name of the HTML select tag.

Default values can be specified. They cause the option which matches the 
default to be displayed in the list. If the default value is not found in
the list, it is added to the end of the list so that it can be displayed. 
This was implemented so that dropdown lists could be used in forms which
edit database records whose values are out of date or not on the current
list of options to display in the dropdown list.

Attributes to the <select> tag can be passed to the routine as the third
argument. 

The C<$size> argument specified the maximum number of characters of the option
descriptions. If the descriptions exceed the maximum size, if specified,
the description is truncated and has "..." appended.

The format of the list is influenced by setting the C<$format> argument. 
It is a character string as used by Perl's sprintf() function.
Examples are: 

	"%s -- %s"
	"%3.2d -- %s"

Different from Perl's sprintf() function, the substituted $format string is
searched for embedded functions to execute that look like &func(arg1,arg2).
This was set up so that mingle() could be used to format text arrays, but
the number of recognized functions can be extended in Local::Acme::Format.

	"&mingle([':',','], %s, %s)"

The C<$flag> argument can have the values of 'multi' or 'noundef'. The
'multi' flag turns the drop-down list into a multi-value list where
the user can select more than one of the options. The 'noundef' flag
setting inhibits the "(Undefined)" option from being one of the list options.

The C<\@columns> argument is a reference to an array of SQL column names.
The columns are assumed to be part of the SQL statement used to generate
the list, but instead of displaying the column values in the list, a
collection of JavaScript statements is generated assigning the column
values to a result array. The resulting array of SQL column values is used
to assign appropriate values to other form fields depending upon which
list item the user selects.

Menu() differs from List() by displaying only the first field in the query
and by not including a "(Undefined)" entry in the list.

The Make() method constructs a list/menu from a database table description
stored in the database schema (see Acme::Base::Schema).

Setting the $multiple flag to the string 'multiple' changes the HTML selection
box from a single value dropdown list to a multiple value selection list.
It also creates a button which allows the user to revert the selection list
to the original values. The default value is parsed into an array of default
values.  (An array is a text string delimited by its quote characters,
i.e. :one:two:three:).

The revert button created by setting the $multiple flag to 'multiple' refers
to a script located in the script file C</scripts/lib/form_validate.js>, so
make sure the following line is included in your forms:

	<script src=/scripts/lib/form_validate.js></script>

The Form Builder places the above line in every form by default.

=head1 CHANGE LOG

 3/98     Initial version.
 2/29/00  Generalized list construction. List definitions moved to
          Local::Acme:ListMenus.
 4/6/00	  Added Make method, to construct a list from the database schema file.
 7/4/00   Added $multiple argument to allow multiple selections.
 10/10/01 Use Local::Acme::Format to substitute commands in sprintf() formats.
 2/25/02  Allow hash values to specify sort order by beginning with a number
          thats stripped off before display.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1998-2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.000;
require Exporter;

@ISA = qw(Local::Acme::ListMenus Exporter);
@EXPORT_OK = qw(list menu);

use Acme::HTML		qw(escape_html);
use Acme::Db;
use Acme::TextArray;
use Local::Acme::ListMenus;
use Local::Acme::Format;
use Acme::Config;

#-------------------------------------------------------------------------------
#
# Utility routine interface
#
#-------------------------------------------------------------------------------

sub menu {
	my ($data,$name,$default,$size,$attributes,$format) = @_;
	$format='%s' unless $format;
	return _list_menu($data,$name,$default,$size,$attributes,$format);
}

sub list {
	return _list_menu(@_);
}

#-------------------------------------------------------------------------------
#
# Object oriented interface
#
#-------------------------------------------------------------------------------

sub Menu {
	my ($class,$data,$name,$default,$size,$attributes,$format) = @_;
	$format='%s' unless $format;
	return _list_menu($data,$name,$default,$size,$attributes,$format,'noundef');
}

sub List {
        my $class = shift;
	return _list_menu(@_);
}

sub Make {
	my ($class,$table,$name,$default,$size,$attributes,$format,$multiple) = @_;

	#
	# Construct the sql to search a database table for a HTML <select> list
	# from the database table properties listed in the schema.
	#

	require Acme::Base::Schema;
	require Acme::SQL;
	require Acme::Security;

	my $s = new Acme::Base::Schema
	  or return _list_menu([],$name,$default,$size,$attributes,$format,$multiple);

	my $t = $s->t($table)
	  or return _list_menu([],$name,$default,$size,$attributes,$format,$multiple);

	$default  ||= $t->select_default;
	$multiple ||= '';

	my $field = $t->code_field || $t->key;
	my $or_is_default = '';
	if ($default && $multiple =~ /multi/i) {
		# define default clause for text array defaults
		my $s = (join "', '", Acme::TextArray::split_array($default));
		$or_is_default = qq[ or $field in ('$s')] if $s;
	}
	else {
		$or_is_default = qq[ or $field = '$default'] if $default;
	}

	my $sqlo = new Acme::SQL $t->select_sql;
	$sqlo->add_where($or_is_default) if $or_is_default;

	my $sql = Acme::Security::clean_sqlo($sqlo);

	$format ||= $t->select_format;
	$format =~ s/^\s*(["'])(.*)\1\s*$/$2/;	# remove possible quotes

	return _list_menu($sql,$name,$default,$size,$attributes,$format,$multiple);
}

#-------------------------------------------------------------------------------
#
# Private routines
#
#-------------------------------------------------------------------------------

sub _list_menu {
	my ($data,$name,$default,$size,$attributes,$format,$flag,$columns) = @_;

	$name       ||= '';
	$default    ||= '';
	$size       ||= '';
	$attributes ||= '';
	$format     ||= '';
	$flag       ||= '';
	$columns    ||= [];

	$attributes &&= " $attributes";

	my $noundef = 0;					# false
	$noundef = 1 if $flag =~ /noundef/i;			# true

	my $multi = '';						# false
	$multi = ' multiple' if $flag =~ /multi/i;		# true

	my @defaults = ();
	if ($default) {

		#
		# Default can be a single value, a text array to be split
		# into values, or a reference to an array of default values.
		#

		if (ref($default) eq 'ARRAY') {
			@defaults = @$default;
		}
		elsif ($multi) {
			@defaults = Acme::TextArray::split_array($default);
		}
		else {
			push @defaults, $default;
		}
	}

	local $index = 0;	# option index
	local %is_default = map { $_, 1 } @defaults;	# flags
	local %selected;				# flags, revert indexes

	#
	# Define helper functions used to format options list.
	#

	my $make_option = sub {
		# Helper function for _list_menu().
		my ($opt, $defaults_arr, $disp, $size) = @_;
		my $sel;

		if ($is_default{$opt}) {
			$sel = ' selected';
			$selected{$opt} = $index;
		}
		else {
			$sel = '';
		}

		# Truncate display text if too long.

		if ($size && length $disp > $size) {
			if ($size < 8) {
				$disp = (substr $disp, 0, $size);
			}
			else {
				$disp = (substr $disp, 0, $size-3) . '...';
			}
		}
			
		$opt = escape_html($opt) || '';
		$disp = escape_html($disp);

		$index++;

		return qq(\n  <option value="$opt"$sel>$disp</option>);
	};

	my $make_display = sub {
		# Helper function for _list_menu().
		my $format = shift;

		local $^W;	# Turn off warnings of uninitialized values
				# from nulls (undefs) read from database.

		return Local::Acme::Format::sprintf_cmd($format,@_) if $format;

		#
		# Generate a default display format.
		#

		if (scalar @_ > 1) {
			$format = '%s -- ' . (join ', ', map {'%s'} @_[1..$#_]);
			return sprintf($format,@_);
		}
		else {
			return sprintf('%s',$_[0]);
		}
	};

	my $make_assignments = sub {
		# Helper function to save other query column values in
		# JavaScript variables.

		my $idx = shift;
		my $val;
		my $assignments = '';
		$assignments .=
		    qq(_result_${name}) . qq([$idx])
		  . qq( = new Object;\n);
		for (@$columns) {
			s/\W/__/g;	# change SQL name to Javascript variable
			$assignments .=
			    qq(_result_${name}) . qq([$idx])
			  . qq(['$_'] = {'value':')
			  . escape_html(${$_})
			  . qq('};\n);
		}
		return $assignments;
	};

	#
	# Initialize assignment scripts.
	#

	# Column values needed to assign to other fields are collected as
	# assignments to JavaScript variables.

	my $assignments = '';		# inititalize
	my $idx = 0;

	#
	# Initialize <select> tag.
	#

	my $select_clause = qq(<select name="$name"$attributes$multi>);

	if (! $noundef && ! $multi) {

		#
		# Make empty option for undefined item.
		#

		$select_clause .= &$make_option('',\@defaults,'(Undefined)',$size);

		$assignments .= &$make_assignments($idx++) if @$columns;
	}

	#
	# Construct options list
	#

	# The type of $data determines if list is an SQL script (scalar),
	# a hash, or an array of values.

	my ($val, $disp, @f);
	my $ref = ref $data;
	if (not $ref or $ref eq 'SCALAR') {
		# Allow both scalar or reference to scalar.
		my $q;
		if ($ref) { $q = Acme::Db->query($$data) }
		else      { $q = Acme::Db->query($data) }

		$q->bind_columns_to_variables() if @$columns;

		while (@f = $q->next) {

			# First column of query is for the option value.
			# Use first column as display column if only column.

			if (scalar @f > 1) { $val = shift @f }
			else 		   { $val = $f[0] }

			$disp = &$make_display($format,@f);
			$select_clause .= &$make_option($val,\@defaults,$disp,$size);
			$assignments .= &$make_assignments($idx++) if @$columns;
		}
	}
	elsif ($ref eq 'HASH') {

		# Sort hash ascending by values

		# Check if every value begins with a number, if so, strip them.
		my $strip_number = 1;
		for (values %$data) {
			if (! /^\d+ /) {
				$strip_number = 0;
				last;
			}
		}

		for (sort { $data->{$a} cmp $data->{$b} } keys %$data) {
			$val = $_;
			$f[0] = $$data{$_};

			$f[0] =~ s/^\d+ // if $strip_number;

			$disp = &$make_display($format,@f);
			$select_clause .= &$make_option($val,\@defaults,$disp,$size);
		}
	}
	elsif ($ref eq 'ARRAY') {
		for (@$data) {
			$val = $_;
			$f[0] = $_;

			$disp = &$make_display($format,@f);
			$select_clause .= &$make_option($val,\@defaults,$disp,$size);
		}
	}

	#
	# Make a selection for any default values that have not yet been listed.
	#

	for (@defaults) {
		if (! exists $selected{$_}) {
			$disp = &$make_display('',$_);
			$select_clause .= &$make_option($_,\@defaults,$disp,$size);
		}
	}

	$select_clause .= "\n</select>";

	if ($multi) {

		#
		# Append a button that reverts settings to original.
		#

		my $no_tabindex = Acme::Config->No_Tabindex;
		my $images      = Acme::Config->Images;
######
## Following uses a plain ol' button.
##		$select_clause .= qq[<input type="button" value="Revert"]
##				. qq[ class="revert-button"]
##				. qq[ onClick="revert(this.form.$name)"]
##				. qq[ tabindex="$no_tabindex"]
##				. qq[>];
######
## Following uses fancy graphic. (might cause form to submit)
##		$select_clause .= qq[<input type="image"]
##				. qq[ style="vertical-align:top"]
##				. qq[ src="$images/undo.gif"]
##				. qq[ title="Undo changes to selections."]
##				. qq[ onMouseOver="this.src='$images/undo_hilite.gif'"]
##				. qq[ onMouseOut="this.src='$images/undo.gif'"]
##				. qq[ onClick="revert(this.form.$name)]
##				. qq[;window.event.cancelBubble=true]
##				. qq[;return false"]
##				. qq[ tabindex="$no_tabindex"]
##				. qq[>];
## Following uses an image on system button. (need to know form object)
##		my $this_form = "self.document.forms[0]";
##		$select_clause .= qq[<button type="button"]
##				. qq[ style="vertical-align:top;padding:0;"]
##				. qq[ title="Undo changes to selections."]
##				. qq[ tabindex="$no_tabindex">]
##				. qq[<img src="$images/undo.gif"]
##				. qq[ onClick="revert($this_form.$name)]
##				. qq[;return false">]
##				. qq[</button>];
## Following uses a CSS image background for an A tag. Uses formOf() function
## to get around having to know the form object. For cross-browser implementation,
## where <a> tag's style needs to be display:block in order to specify the height,
## use a transparent image for forground so buttons don't get rendered when printed.
		my $img = $ENV{HTTP_USER_AGENT} =~ /MSIE/
			? ''
			: qq[<img src="$images/transparent.gif" width="22" height="20" border="0">];
		$select_clause .= qq[<a]
				. qq[ href="javascript:void 0"]
				. qq[ title="Undo changes to selections."]
				. qq[ onClick="revert(formOf(this).$name)"]
				. qq[ class="undo"]
				. qq[ tabindex="$no_tabindex"]
				. qq[>$img</a>];

		$select_clause  = qq[<nobr>$select_clause</nobr>];
	}

	if ($assignments) {

		#
		# Put assignments into <script> tags
		#

		$assignments = <<End_Of_Javascript;
<script language="JavaScript" type="text/javascript">
var _result_${name} = new Array;
$assignments</script>
End_Of_Javascript
	}

	return $assignments . $select_clause;
}

1;
