package Acme::HTML::NavButtons;

=head1 NAME

Acme::HTML::NavButtons - navigation buttons for HTML pages

=head1 SYNOPSIS

    use Acme::HTML::NavButtons;


    $b = new Acme::HTML::NavButtons;
    print $b->buttons();

	...

    print $b->last_buttons();

or more precisely

    $b = Acme::HTML::NavButtons->new(
	name				=> $name,
	header_bottom_button_left	=> $gif-picT1,
	header_bottom_button_right	=> $gif-picT2,
	header_previous_button_left	=> $gif-picT3,
	header_previous_button_right	=> $gif-picT4,
	header_down_button		=> $gif-picT5,
	footer_top_button_left		=> $gif-picB1,
	footer_top_button_right		=> $gif-picB2,
	footer_previous_button_left	=> $gif-picB3,
	footer_previous_button_right	=> $gif-picB4,
	footer_up_button		=> $gif-picB5,
      );

    print $b->setup;
    print $b->first_buttons;

    print $b->buttons;

    print $b->up_button($return_anchor);	# top of section
    print $b->down_button($return_anchor);	# bottom of section

    print $b->nav_button($return_anchor);	# small set of up, down and prev

    print $b->last_buttons;


=head1 DESCRIPTION

The Navigation Buttons provide buttons to jump around within a browser
window. Think of them as a collection of buttons which divide records.
The default pictures for the buttons form a record divider across the
browser window.

All arguments to the routines are optional. They override defaults if
specified.

If you want more than one set of buttons on a page, you must specify a
unique name to the new() method.

The methods setup() and first_buttons() are optional. Setup() is executed
by first_buttons() if it has not been yet executed and first_buttons()
is executed by buttons() if it has not been executed. Setup() returns
JavaScript code supporting the buttons.

The up_button() method displays a small button which goes to the top
of the current section, i.e. goes to the previous button bar. It is
not visually part of the button bars. Pressing the Previous button on
the button bar returns to the location of the up_button, which can be
optionally specified as an argument to the method.

The down_button() method works like the up_button() method except that it
goes to the bottom of the section, or in other words, down to the next
section of navigation buttons.

The nav_button() method returns a small button that consists of three
parts. The upper goes up to the previous section like up_button().
The lower part goes down to the next section like down_button().
The middle of the button goes to the previous location.

Note that the up_button(), down_button(), and nav_button() do not define
anchors as the first_buttons(), buttons(), and last_buttons() methods do.
They use references to anchors defined elsewhere.

All methods (except new()) return HTML code which must be sent to the browser
by the calling program.

A C<print> flag has been added to new() so that the navigation buttons can
be rendered to appear on a printed page. At this time, none of the buttons
are rendered and the html text in C<print_divider> is returned whenever the
buttons() method is called. The default setting of C<print_divider> is the
<HR class="divider"> tag.

A C<$print> flag has been added to new() so that suitable substitution or
omission of navigation buttons can determined by the designer of the buttons.
Quite often, the navigation buttons are used as page dividers which are 
should be mimicked on a printed page.

=head1 CHANGE LOG

 11/98    Initial version.
 7/4/99   Added up_button() and down_button().
 3/17/00  Changed links to reset title that MSIE5 insists on changing.
 10/20/00 Added nav_button() method. Cleaned up JavaScript routines.
 8/10/01  Added $print flag.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1998-2000, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use vars qw($No_Tabindex $Nav_Images);

use Acme::Config;

$No_Tabindex = Acme::Config->No_Tabindex || '-1';
$Nav_Images  = Acme::Config->Nav_Images || '/images';

sub import {}

sub new {
        my $this = shift;
        my $class = ref($this) || $this;

	# default values
	my $self = {
		name			     => "buttons",
		basequery		     => "",
		ndex			     => 0,
		footer_anchor		     => "top",
		previous_footer_anchor	     => "top",
		top_anchor		     => "top",
		bottom_anchor		     => "bottom",
		up_anchor		     => "top",
		nav_anchor		     => "top",
		setup_executed		     => 0,
		first_executed		     => 0,
		first_header_button_border   => "$Nav_Images/first_header_button_border.gif",
		header_bottom_button_left    => "$Nav_Images/header_bottom_button_left.gif",
		header_bottom_button_right   => "$Nav_Images/header_bottom_button_right.gif",
		header_previous_button_left  => "$Nav_Images/header_previous_button.gif",
		header_previous_button_right => "$Nav_Images/header_previous_button.gif",
		header_down_button	     => "$Nav_Images/header_down_button.gif",
		footer_top_button_left	     => "$Nav_Images/footer_top_button_left.gif",
		footer_top_button_right	     => "$Nav_Images/footer_top_button_right.gif",
		footer_previous_button_left  => "$Nav_Images/footer_previous_button.gif",
		footer_previous_button_right => "$Nav_Images/footer_previous_button.gif",
		footer_up_button	     => "$Nav_Images/footer_up_button.gif",
		last_footer_button_border    => "$Nav_Images/last_footer_button_border.gif",
		up_button		     => "$Nav_Images/up_button.gif",
		down_button		     => "$Nav_Images/down_button.gif",
		nav_button		     => "$Nav_Images/nav_button.gif",
		print			     => 0,
		print_divider		     => '<hr class="divider">',
		@_				# set argument values
	};

	return bless $self, $class;
}

#------------------------------------------------------------------------------

sub setup {
	my $self = shift;

	my $base = "http://"
	  . $ENV{SERVER_NAME}
	  . ($ENV{SERVER_PORT} != 80 ? ":$ENV{HTTP_PORT}" : '')
	  . $ENV{SCRIPT_NAME};
	$self->{basequery} = $ENV{QUERY_STRING} ? "$base?$ENV{QUERY_STRING}" : "$base";

	$self->{setup_executed} = 1;	# set flag that this method executed

	# output necessary javascript
	return <<EOF;

<!-- Start Buttons Scripts -->
<script language="JavaScript">

var prevAnchor = "$self->{top_anchor}";
var currDir;	// current direction of navigation (U -- up, D -- down)

function up_anchor(to,from) {
	if (currDir != 'U') {
		// record position to go to with previous button
		currDir = 'U';
		goto_anchor(to,from);
	} else {
		goto_anchor(to);
	}
}

function down_anchor(to,from) {
	if (currDir != 'D') {
		// record position to go to with previous button
		currDir = 'D';
		goto_anchor(to,from);
	} else {
		goto_anchor(to);
	}
}

function previous_anchor(from) {
	// go to previous saved anchor
	goto_anchor(prevAnchor,from);
}
	
function goto_anchor(to, from) {
	if (from) prevAnchor = from;
	self.location.hash = to;

	// force MSIE5 to keep same title on window
	self.document.title = self.document.title;
}
</script>
<!-- End Buttons Scripts -->

EOF
}

#------------------------------------------------------------------------------

sub first_buttons {
	my $self = shift;
	my $id = shift;

	return '' if $self->{print};

	my $s = "";
	$s = setup($self) unless $self->{setup_executed};

	$id = "p" unless ($id);
	$self->{footer_anchor} = $id . ($self->{ndex} + 2); # mimic page numbers

	$self->{ndex} += 1;

	$self->{first_executed} = 1;	# set flag that this method executed
	return $s .
	  qq(<a name="$self->{top_anchor}"></a>\n) .
	  _first_header_buttons_html($self) .
	  _header_buttons_html($self);
}

#------------------------------------------------------------------------------

sub buttons {
	my $self = shift;
	my $id = shift;

	return first_buttons($self) unless $self->{first_executed};

	return $self->{print_divider} if $self->{print};

	my $s = _footer_buttons_html($self);	# assemble footer first

	$self->{previous_footer_anchor} = $self->{footer_anchor};
	
	$id = "p" unless ($id);
	$self->{footer_anchor} = $id . ($self->{ndex} + 2); # mimic page numbers

	$self->{ndex} += 1;

	return $s . _header_buttons_html($self); # append header to footer
}

#------------------------------------------------------------------------------

sub last_buttons {
	my $self = shift;
	my $id = shift;

	return '' if $self->{print};

	return _footer_buttons_html($self) .
	  _last_footer_buttons_html($self) .
	  qq(<a name="$self->{bottom_anchor}"></a>\n\n);
}
	
#------------------------------------------------------------------------------

sub up_button {
	my $self = shift;
	my $return_anchor = shift;

	return '' if $self->{print};

	my $s = "";
	$s = setup($self) unless $self->{setup_executed};

	if ($return_anchor) {
		# return anchor has been already defined by caller, use it.
		$self->{up_anchor} = $return_anchor;
	}
	else {
		# use same anchor index as header/footer buttons
		$self->{up_anchor} = "s" . ($self->{ndex} + 2);
		$self->{ndex} += 1;

		# define return anchor position to be at button location
		$s .= qq(<a name="$self->{up_anchor}"></a>);
	}

	return $s . _up_button_html($self);
}

#------------------------------------------------------------------------------

sub down_button {
	my $self = shift;
	my $return_anchor = shift;

	return '' if $self->{print};

	my $s = "";
	$s = setup($self) unless $self->{setup_executed};

	if ($return_anchor) {
		# return anchor has been already defined by caller, use it.
		$self->{down_anchor} = $return_anchor;
	}
	else {
		# use same anchor index as header/footer buttons
		$self->{down_anchor} = "s" . ($self->{ndex} + 2);
		$self->{ndex} += 1;

		# define return anchor position to be at button location
		$s .= qq(<a name="$self->{down_anchor}"></a>);
	}

	return $s . _down_button_html($self);
}

#------------------------------------------------------------------------------

sub nav_button {
	my $self = shift;
	my $return_anchor = shift;

	return '' if $self->{print};

	my $s = "";
	$s = setup($self) unless $self->{setup_executed};

	if ($return_anchor) {
		# return anchor has been already defined by caller, use it.
		$self->{nav_anchor} = $return_anchor;
	}
	else {
		# use same anchor index as header/footer buttons
		$self->{nav_anchor} = "s" . ($self->{ndex} + 2);
		$self->{ndex} += 1;

		# define return anchor position to be at button location
		$s .= qq(<a name="$self->{nav_anchor}"></a>);
	}

	return $s . _nav_button_html($self);
}

#------------------------------------------------------------------------------
#
# Private methods
#
#------------------------------------------------------------------------------

sub _first_header_buttons_html {
	my $self = shift;

	local $\ = "";	# assure print() does down append newlines

	return <<EOF;

<!-- Start First Buttons Border -->
<div align=center><img
  src='$self->{first_header_button_border}'
  width=2.5%
  height=4
  border=0
  ><img
  src='$self->{first_header_button_border}'
  width=2.5%
  height=4
  border=0
  ><img
  src='$self->{first_header_button_border}'
  width=89.5%
  height=4
  border=0
  ><img
  src='$self->{first_header_button_border}'
  width=2.5%
  height=4
  border=0
  ><img
  src='$self->{first_header_button_border}'
  width=2.5%
  height=4
  border=0
  ></div>
<!-- End First Buttons Border -->
EOF

}

#------------------------------------------------------------------------------

sub _header_buttons_html {
	my $self = shift;

	# Note: there cannot be any spaces or linefeeds between <a> and <img>
	# tags here as the graphics are intended to be continuous.

	local $\ = "";	# assure print() does not append newlines

	return <<EOF;

<!-- Start Down Buttons -->
<div align=center><a href="javascript:goto_anchor('$self->{bottom_anchor}','$self->{previous_footer_anchor}')" tabindex=$No_Tabindex><img
  src='$self->{header_bottom_button_left}'
  alt='Bottom'
  width=2.5%
  height=12
  border=0
  ></a><a href='javascript:previous_anchor("$self->{previous_footer_anchor}")' tabindex=$No_Tabindex><img
  src='$self->{header_previous_button_left}'
  alt='Previous'
  width=2.5%
  height=12
  border=0
  ></a><a href="javascript:down_anchor('$self->{footer_anchor}','$self->{previous_footer_anchor}')" tabindex="$No_Tabindex"><img
  src='$self->{header_down_button}'
  alt='Down'
  width=89.5%
  height=12
  border=0
  ></a><a href='javascript:previous_anchor("$self->{previous_footer_anchor}")' tabindex=$No_Tabindex><img
  src='$self->{header_previous_button_right}'
  alt='Previous'
  width=2.5%
  height=12
  border=0
  ></a><a href="javascript:goto_anchor('$self->{bottom_anchor}','$self->{previous_footer_anchor}')" tabindex=$No_Tabindex><img
  src='$self->{header_bottom_button_right}'
  alt='Bottom'
  width=2.5%
  height=12
  border=0
  ></a></div>
<!-- End Down Buttons -->
EOF

}

#------------------------------------------------------------------------------

sub _footer_buttons_html {
	my $self = shift;

	local $\ = "";	# assure print() does down append newlines

	return <<EOF;

<!-- Start Up Buttons -->
<div align=center>
<a name='$self->{footer_anchor}'></a>
<a href="javascript:goto_anchor('$self->{top_anchor}','$self->{footer_anchor}')" tabindex=$No_Tabindex><img
  src='$self->{footer_top_button_left}'
  alt='Top'
  width=2.5%
  height=12
  border=0
  ></a><a href='javascript:previous_anchor("$self->{footer_anchor}")' tabindex=$No_Tabindex><img
  src='$self->{footer_previous_button_left}'
  alt='Previous'
  width=2.5%
  height=12
  border=0
  ></a><a href="javascript:up_anchor('$self->{previous_footer_anchor}','$self->{footer_anchor}')" tabindex="$No_Tabindex"><img
  src='$self->{footer_up_button}'
  alt='Up'
  width=89.5%
  height=12
  border=0
  ></a><a href='javascript:previous_anchor("$self->{footer_anchor}")' tabindex=$No_Tabindex><img
  src='$self->{footer_previous_button_right}'
  alt='Previous'
  width=2.5%
  height=12
  border=0
  ></a><a href="javascript:goto_anchor('$self->{top_anchor}','$self->{footer_anchor}')" tabindex=$No_Tabindex><img
  src='$self->{footer_top_button_right}'
  alt='Top'
  width=2.5%
  height=12
  border=0
  ></a></div>
<!-- End Up Buttons -->
EOF

}

#------------------------------------------------------------------------------

sub _last_footer_buttons_html {
	my $self = shift;

	# Note: there cannot be any spaces or linefeeds between <a> and <img>
	# tags here as the graphics are intended to be continuous.

	local $\ = "";	# assure print() does not append newlines

	return <<EOF;

<!-- Start Last Buttons Border -->
<div align=center><img
  src='$self->{last_footer_button_border}'
  width=2.5%
  height=4
  border=0
  ><img
  src='$self->{last_footer_button_border}'
  width=2.5%
  height=4
  border=0
  ><img
  src='$self->{last_footer_button_border}'
  width=89.5%
  height=4
  border=0
  ><img
  src='$self->{last_footer_button_border}'
  width=2.5%
  height=4
  border=0
  ><img
  src='$self->{last_footer_button_border}'
  width=2.5%
  height=4
  border=0
  ></div>
<!-- End Last Buttons Border -->
EOF

}

#------------------------------------------------------------------------------

sub _up_button_html {
	my $self = shift;

	return
qq(<!-- Start Up Button --><span class="up-button"><a
  href="javascript:goto_anchor('$self->{previous_footer_anchor}','$self->{up_anchor}')" tabindex="$No_Tabindex"><img
  src="$self->{up_button}"
  alt="Up"
  width="10"
  height="12"
  border="0"
  ></a></span><!-- End Up Button -->);
}

#------------------------------------------------------------------------------

sub _down_button_html {
	my $self = shift;

	return
qq(<!-- Start Down Button --><span class="down-button"><a
  href="javascript:window.document.getElementById('$self->{footer_anchor}').scrollIntoView(false)"
  onClick="prevAnchor = '$self->{down_anchor}'" tabindex=$No_Tabindex><img
  src="$self->{down_button}"
  alt="Down"
  width="10"
  height="12"
  border="0"
  ></a></span><!-- End Down Button -->);
}

#------------------------------------------------------------------------------

sub _nav_button_html {
	my $self = shift;

	return
qq(<!-- Start Navigation Button --><span class="nav-button"><img 
  src="$self->{nav_button}"
  alt="Navigation Button"
  width=10
  height=18
  border=0
  usemap="#navigation_button$self->{nav_anchor}"
  ><map name="navigation_button$self->{nav_anchor}"><area
    shape="rect" coords="0,0,10,5"
    href="javascript:goto_anchor('$self->{previous_footer_anchor}','$self->{nav_anchor}')"
    alt="Up"
    tabindex="$No_Tabindex"><area
    shape="rect" coords="0,5,10,13"
    href="javascript:previous_anchor('$self->{nav_anchor}')"
    alt="Previous"
    tabindex="$No_Tabindex"><area
    shape="rect" coords="0,13,10,18"
    href="javascript:window.document.getElementById('$self->{footer_anchor}').scrollIntoView(false)"
    alt="Down"
    onClick="prevAnchor = '$self->{nav_anchor}'"
    tabindex="$No_Tabindex"></map></span><!-- End Navigation Button -->);
}

1;
