package Acme::HTML::Parameters;

=head1 NAME

Acme::HTML::Parameters - Parse HTML tags.

=head1 SYNOPSIS

    use Acme::HTML::Parameters qw(get_parameters get_confirmation);

    # Get complete html form for entering the parameters.
    $parm_html = Acme::HTML::Parameters->get(	# synonym for parameters()
		title => "Parameter Window Title",
		html  => $html_code_fragment_of_input_tags,
		);
     -or-
    print Acme::HTML::Parameters->parameters(	# synonym for get()
		title  => "Parameter Window Title",
		html   => $html_code_fragment_of_input_tags,
		script => qq[<script src="/scripts/lib/calendar.js"></script>],
		);
     -or-
    print Acme::HTML::Parameters->parameters(	# synonym for get()
		title  => "Parameter Window Title",
		params => \@params,		# search script parameters
		);
     -or-
    get_parameters(				# display (print) form
		title   => "Parameter Window Title",
		html    => $html_code_fragment_of_input_tags,
		exclude => ['field1','field2'],
		);
     -or- (listing all options)
    get_parameters(				# display (print) form
		window_title => 'Enter Parameters',
		title        => 'Customer Search Parameters',
		method       => 'POST',		# or 'GET'
		submit       => ' Search ',	# Submit button text
		reset        => ' Cancel ',	# Reset button text
		other        => '<input type="button" class="button" value="Preview" onClick="do_preview()">'
		exclude      => ['field1','field2'],
		script       => qq[<script src="/scripts/lib/calendar.js"></script>],
		style        => '<link rel="stylesheet" href="/mycalendar.css" type="text/css">',
		instructions => $instructions_html,
		instructions_file => $html_file_path,
		instructions_url  => $url,

		html         => $html_code_fragment_of_input_tags,
		url          => '/cgi-bin/prog.cgi',
		cgi          => $cgi,
		all          => 1;
		);

    # Confirmation message with Yes/No/Cancel buttons.
    $msg_html = Acme::HTML::Parameters->confirm(
		title   => 'Window Title',
		message => 'A message to the user',
		prompt  => 'A prompt or question for the user.',
		answer  => 'return_answer_field_name',
		);
     -or-
    get_confirmation(				#  display (print) form
		title   => 'Window Title',
		message => 'A message to the user',
		prompt  => 'A prompt or question for the user.',
		answer  => 'return_answer_field_name',
		);

    # Utility functions.
    $html .= save_cgi_param_except('answer');
    -or-
    $html .= save_cgi_param_except($cgi,'answer');

    $html .= save_all_param_except($cgi,'answer');

    $html .= param_html($cgi,@param);

    @fields = field_names($html);

=head1 DESCRIPTION

This module defines methods that allow server programs to ask for parameters
from the user, and if necessary, have the user confirm that they want to
perform the task.

An example program flow is like:

=over 4

=item 1

user clicks a link to invoke a server program,

=item 2

the server program displays a form asking for missing parameters,

=item 3

user fills in the form and submits it back to the server,

=item 4

the server checks the user's answers if if needed, sends another form to the user asking
them to confirm their answers,

=item 5

user submits the confirmation to the server which completes the task.

=back

CGI parameters are automatically saved in hidden fields as needed when
displaying the parameter form and confirmation form to the user. Fields
which are not to be passed on can be specified in the C<exclude>
parameter. Parameters which are asked for from the user in the C<html>
argument are automatically excluded from the CGI parameters before saving
them in hidden parameters, so they do not have to be specified in an
C<exclude> parameter.

Url parameters are saved if the CGI parameters are POSTed and also specified
in the url if the C<all> parameter is true (non-zero).

The HTML text for hidden input fields can be obtained directly by using the
utility function save_cgi_param_except(). The function save_all_param_except()
returns all the parameters of save_cgi_param_except() with the addition of
any parameters listed in the url.

Care has been taken to preserve multiple values for a parameter.

CGI objects can be optionally passed to any of the routines. If one is not
specified, a new CGI object is automatically created.

Instead of specifying (or in addition to specifying) the html code you can
specify a C<params> argument which points to an array of Acme::Searcher
parameter arguments -- those specifications which are found in search scripts -- such as:

	$last_name=?t24?Pexton?Last Name
	
	$bgndate=?d?$today?Beginning Date

	$user=?_users?$user?User ID

The parts of the parameter arguments are:

	$field_name=?t24?default?field_label?help

	field_name	-- field name assigned to the <input> field
	=?		-- signifies this is a parameter definition
	t24		-- type and length of field. (text field of 24 spaces)
	?		-- required separator
	default 	-- initial default value assigned to field
	?		-- required separator
	field_label	-- prompt string for field entry
	help		-- help text (optional)

In search scripts, the C<field_name> begins with a C<$> because it is
a variable name. Here, C<field_name> is a name of a field so the C<$>
character is optional.

The field type indicator characters are:

	t 		-- text
	d 		-- date
	h 		-- time
	n 		-- number
	m 		-- US money
	_ 		-- dropdown list, the following word is the table name

The HTTP method of transmitting the form can be specified in the arguments
to parameter(). The default is 'POST' but sometimes the 'GET' method must
be specified to keep arguments in the url, usually when the web page alters
its url and refreshes itself to display its different options.

=head1 EXAMPLE

    use Acme::HTML::Parameters;
    use CGI;

    $cgi = new CGI;
    $file = $cgi->param('file');
    unless ($file) {
        Acme::HTML::Parameters->get_parameters(
                           title => "Enter File Name",
			   param => [ '$file=?t24??File Name' ],
                           );
        exit;
    }
    
    if (-e $file) {
        $overwrite = $cgi->param('overwrite');
        unless ($overwrite =~ /yes/i) {
            Acme::HTML::Parameters->get_confirmation(
                               title   => 'Please Confirm',
                               message => 'The file already exists.',
                               prompt  => 'Do you want to overwrite it?',
                               answer  => 'overwrite',
                               );
            exit;
        }

    ## Process file here ##

    }

=head1 CHANGE LOG

 4/21/01  Initial version.
 8/29/01  Added 'method' argument to specify HTTP method of POST or GET.
 1/12/02  Added param_html() method to assist Acme::Searcher scripts.
 10/23/02 Added save_all_param_except() method.
 12/2/02  Added help text to search script parameter specifications
          and added instructions.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2001,2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

use Acme::HTML			qw(escape_html);
use Acme::HTML::ListMenus;
use Acme::HTML::HideMoreLess	qw(start_hiding		stop_hiding);
use Local::Acme::Form::Banners	qw(common_header	common_footer
				   common_param_header	common_param_footer);
use Local::Acme::Form::Extensions;
use Acme::Config;
use Carp;
use CGI;

require Exporter;

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(get_parameters		parameters
		    get_confirmation		confirm
		    save_cgi_param_except	save_all_param_except
		    field_names);

#------------------------------------------------------------------------------

sub get { return parameters(@_); }

sub get_confirmation { print confirm(@_);    }
sub get_parameters   { print parameters(@_); }

#------------------------------------------------------------------------------

sub parameters {
	# This routine can be used both as a function or a method.
	shift if ref($_[0]) || $_[0] eq __PACKAGE__;

	my $arg = { @_ };

	our $Styles		 = Acme::Config->Styles;
	our $Scripts		 = Acme::Config->Scripts;
	our $Scripts_Lib	 = Acme::Config->Scripts_Lib;

	my $cgi          = $arg->{cgi}		|| new CGI;
	my $window_title = $arg->{window_title}	|| 'Enter Parameters';
	my $title        = $arg->{title}	|| '';
	my $method       = $arg->{method}	|| 'POST';
	my $submit       = exists $arg->{submit} ? $arg->{submit} : ' Search ';
	my $reset        = exists $arg->{reset}  ? $arg->{reset}  : ' Cancel ';
	my $other        = $arg->{other}	|| '';
	my $exclude      = $arg->{exclude}	|| [];
	my $script       = $arg->{script}	|| '';
	my $style        = $arg->{style}	|| '';

	my $html         = $arg->{html}		|| '';

	my $instructions = get_instructions_html($arg->{instructions})
			|| get_instructions_file($arg->{instructions_file})
			|| get_instructions_url($arg->{instructions_url});

	my $url          = $arg->{url};
	if (! $url) {
		# Make url that will call ourselves again with new parameters.
		$url = $cgi->url(-full=>1, -path=>1);
		$url .= '?' . $ENV{QUERY_STRING}
		  if $ENV{REQUEST_METHOD} eq 'POST' && $ENV{QUERY_STRING};
	}

	$html .= param_html($cgi,@{$arg->{params}}) if $arg->{params};

	my $hidden_fields = $arg->{all}
			  ? save_all_param_except($cgi,(field_names($html),@$exclude))
			  : save_cgi_param_except($cgi,(field_names($html),@$exclude));

	#
	# Construct parameter form
	#

# Keep IE6 in IE5 compatablilty mode by not specifying url
#<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
#"http://www.w3.org/TR/html4/loose.dtd">

	my $param = <<EOF;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">

<html>
<head>
<title>$window_title</title>

<link rel="stylesheet" href="$Styles/common.css" type="text/css">
<link rel="stylesheet" href="$Styles/form.css" type="text/css">
$style
<script src="$Scripts/common.js" type="text/javascript"></script>
<script src="$Scripts_Lib/form_validate.js" type="text/javascript"></script>
<script src="$Scripts_Lib/form_events.js" type="text/javascript"></script>
$script
</head>
<body onLoad="focus_first_field()">
EOF

	$param .= common_header($window_title);

	$param .= <<EOF;
<form   name="parameters"
	action="$url"
	enctype="multipart/form-data"
	method="$method"
>
EOF
	$param .= common_param_header($window_title, {
						submit => $submit,
						reset => $reset,
						other => $other,
						});

	$param .= <<EOF;
<h2>$title</h2>

$instructions

$html

$hidden_fields
EOF

	$param .= common_param_footer($window_title, {
						submit => $submit,
						reset => $reset,
						other => $other,
						});

	$param .= <<EOF;
</form>
<br>
EOF
	$param .= common_footer($window_title);

	$param .= <<EOF;
</body>
</html>
EOF

	#
	# Substitute commands.
	#

	$param =~ s/(<!--+\$+\s*.*?\s*--+>)/tag_command($1)/egs;

	return $param;
}

#------------------------------------------------------------------------------

sub confirm {
	# This routine can be used both as a function or a method.
	shift if ref($_[0]) || $_[0] eq __PACKAGE__;

	my $arg = { @_ };

	our $Styles = Acme::Config->Styles;

	my $cgi     = $arg->{cgi}	|| new CGI;
	my $title   = $arg->{title}	|| 'Please Confirm';
	my $message = $arg->{message}	|| 'Please confirm.';
	my $prompt  = $arg->{prompt}	|| 'Click Yes or No.';
	my $answer  = $arg->{answer}	|| 'answer';
	my $method  = $arg->{method}	|| 'POST';

	my $url     = $arg->{url};
	if (! $url) {
		# Make url that will call ourselves again with new parameters.
		$url = $cgi->url(-full=>1, -path=>1);
		$url .= '?' . $ENV{QUERY_STRING}
		  if $ENV{REQUEST_METHOD} eq 'POST' && $ENV{QUERY_STRING};
	}

	my $hidden_fields = $arg->{all}
			  ? save_all_param_except($cgi,$answer)
			  : save_cgi_param_except($cgi,$answer);

	#
	# Construct confirmation form
	#

# Keep IE6 in IE5 compatablilty mode by not specifying url
#<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN"
#"http://www.w3.org/TR/html4/loose.dtd">

	my $confirm = <<EOF;
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">

<html>
<head>
<title>$title</title>
<link rel="stylesheet" href="$Styles/message.css" type="text/css">
<script language="JavaScript" type="text/javascript">
// Assign keypress event handler only if not assigned before
window.document.onkeypress = param_keypress;

function param_keypress(evt) {
        switch (window.event.keyCode) {
        case 27:        // escape key
                window.close();
                return true;

        case 32:                // space key
                /*
                 * Ctrl-Space bar displays the console screen.
                 */
                if (window.event.ctrlKey) {
                        if (typeof window.console == "function") {
                                console();
                                return false;
                        }
                }
                return true;

	case 89:		// 'Y'
	case 121:		// 'y'
		self.document.forms[0].elements['$answer'][0].click();
                return true;

	case 78:		// 'N'
	case 110:		// 'n'
		self.document.forms[0].elements['$answer'][1].click();
		return true;
        }
}
</script>
</head>
<body onLoad=self.document.forms[0].elements[1].focus()>

<form	method="$method"
	action="$url"
>

<div class="warning">
<big>$message</big>
<br>
<br>
<big>$prompt</big>
<br>
<div class="button">
<input type="submit" name="$answer" value=" Yes " accesskey="y">
&nbsp;
<input type="reset" name="$answer" value="  No  " onClick="history.back()" accesskey="n">
&nbsp;
<input type="reset" value="Cancel" onClick="self.close()">
</div>
</div>

$hidden_fields
</form>
</body>
</html>
EOF
	return $confirm;
}

#------------------------------------------------------------------------------

sub save_cgi_param_except {
	# This routine can be used both as a function or a method.
	shift if (ref($_[0]) && ref($_[0]) ne 'CGI') || $_[0] eq __PACKAGE__;
	
	# Look for optional CGI object passed as the first argument.
	my $cgi = shift if ref($_[0]) eq 'CGI';
	$cgi = new CGI unless $cgi;

	#
	# Save current program CGI arguments in hidden input fields
	# except function arguments (e.g. those arguments listed in @_).
	#

	my $hidden_fields = _hidden_fields_except($cgi,@_);

	$hidden_fields = <<EOF if $hidden_fields;
<!-- Begin hidden data fields -->
$hidden_fields
<!-- End hidden data fields -->
EOF

	return $hidden_fields;
}

sub save_all_param_except {
	# This routine can be used both as a function or a method.
	shift if (ref($_[0]) && ref($_[0]) ne 'CGI') || $_[0] eq __PACKAGE__;
	
	# Look for optional CGI object passed as the first argument.
	my $cgi = shift if ref($_[0]) eq 'CGI';
	$cgi = new CGI unless $cgi;

	#
	# Save current program CGI arguments in hidden input fields
	# except function arguments (e.g. those arguments listed in @_).
	# Both the parameters in the URL and in the POSTed fill-out form
	# are saved.
	#

	my $hidden_fields = _hidden_fields_except($cgi,@_)
			  . _url_hidden_fields_except($cgi,@_,$cgi->param());


	$hidden_fields = <<EOF if $hidden_fields;
<!-- Begin hidden data fields -->
$hidden_fields
<!-- End hidden data fields -->
EOF

	return $hidden_fields;
}

sub _hidden_fields_except {
	# Return hidden fields html

	my $cgi = shift;

	my %exclude = map { $_, 1 } @_;			# excluded fields

	my $hidden_fields = '';
	for my $key ($cgi->param()) {
		next if $exclude{$key};
		for ($cgi->param($key)) {
			$hidden_fields .= _hidden_field($key, $_);
		}
	}

	return $hidden_fields;
}

sub _url_hidden_fields_except {
	# Return hidden fields html

	my $cgi = shift;

	my %exclude = map { $_, 1 } @_;			# excluded fields

	my $hidden_fields = '';
	for my $key ($cgi->url_param()) {
		next if $exclude{$key};
		for ($cgi->url_param($key)) {
			$hidden_fields .= _hidden_field($key, $_);
		}
	}

	return $hidden_fields;
}

sub _hidden_field {
	# Function to contruct hidden field html

	my $key = escape_html(shift);
	my $value = escape_html(shift);

	if (length($value) < 255) {
		# hide normal fields in <input> controls
		return qq(<input type=hidden name="$key" value="$value">\n);
	}
	else {
		# hide long fields in <textarea> controls
		return qq(<textarea name="$key" wrap="virtual" tabindex="-1" style="display:none;">$value</textarea>\n);
	}
}

#------------------------------------------------------------------------------

sub field_names {
	# This routine can be used both as a function or a method.
	shift if ref($_[0]) || $_[0] eq __PACKAGE__;

	#
	# Return field names (names of <input> and <textarea> tags) of
	# html text.
	#

	my $html = shift;

	my @fields;		# returned array of names found in html

	#
	# Define helper function to extract names from tags
	# Names are pushed into @fields as they are found.
	#

	my $get_name = sub {
		my $tag = shift;

		if ($tag =~ /\sname\s*=\s*(['"])(.*?)\1/i) {
			# name is surrounded by quotes
			push @fields, $2;
		}
		elsif ($tag =~ /\sname\s*=\s*(.*?)\s/i) {
			# name not surrounded by quotes
			push @fields, $1;
		}

		return $tag;		# return original input
	};

	#
	# Process <input> and <textarea> tags 
	#

	# Look for <input> tags.
	$html =~ s|(<input\s.*?>)|&$get_name($1)|egsi;

	# Look for <textarea> tags.
	$html =~ s|(<textarea.*?>.*?</textarea>)|&$get_name($1)|egsi;

	return @fields;
}

#------------------------------------------------------------------------------

sub tag_command {
	my ($tag) = @_;

	my ($command) = $tag =~ /<!--+\$+\s*(.*?)\s*--+>/s;

	return $tag unless $command;
	return Local::Acme::Form::Extensions::tag_command($tag);
}

#------------------------------------------------------------------------------

sub param_html {

	my $cgi = shift;	# look in cgi for default values

	#
	# Build parameter HTML from array of Acme::Searcher parameters like
	# `$var=?t10?testing?Test Prompt'. The question marks separate field type
	# and size, initial default value, and the prompt for the field entry.
	#

	our $Scripts_Lib	= Acme::Config->Scripts_Lib;
	our $Images		= Acme::Config->Images;
	our $No_Tabindex	= Acme::Config->No_Tabindex;

	my ($var,$def);		# variable/definition parts of parameter stmt
	my ($type, $default, $prompt, $help);	# parts of parameter definition
	my ($size, $list_name, $title_str);

	my $cal_script = '';
	my $html = qq[<table class="form">\n];

	for my $l ( @_) {
		($var,$def) = split /=\?/, $l;
		($type, $default, $prompt, $help) = split /\?/, $def;

		# remove $ from var so that var is now just a name
		$var =~ s/^\s*\$//;

		# Substitute standard variables into default string
		if ($default =~ /\$/) {
			package defaults;

			use Acme::Date		qw(dayadd today);
			use Local::Acme::User	qw(user);

			our $today = today();
			our $user  = user();

			no strict 'refs';

			# Note: This should be done using Safe module like
			# Acme::Searcher. However, this is usually called
			# by Acme::Searcher, which already processed the
			# script this fragment comes from. Actually, it
			# should fill in the default values it already
			# knows and this should only be used by others.

			$default = eval $default;

			# substitute only package variables
			#$default =~ s/(\${?\w+}?)/${$1}/eg;
		}

		#
		# Look in cgi for default value.
		#

		# Note: because the parameter scripts get changed into
		# $cgi->param(...) statements, the cgi values could be 
		# used anyway for the field input even if they are not
		# displayed here.

		$default ||= $cgi->param($var) || '';

		$prompt = "<nobr>$prompt</nobr>" if $prompt =~ / /;

		$html = join "\n", $html,
		  '<tr>',
		  '<td class="label">',
		  "$prompt",
		  '</td>',
		  '<td class="field">'; 

		SWITCH: {
			if ($type =~ /^t/) {	# text
				$title_str = $help ? qq[ title="$help"] : '';
				($size) = $type =~ /t(\d*)/;
				$size ||= 50;
				if ($size < 100) {
					$html .=
qq[<input type="text" name="$var" value="$default" size="$size" onFocus="this.select()" onChange="check_input(this,'text',$size)"$title_str>\n];
				}
				else {
					# Long text fields use <textarea> fields.

					# Try to size <textarea> intelligently.
					my $rows = int((50+$size)/$size);
					if ($rows > 9) { $rows = 6 }
					elsif ($rows > 7) { $rows = 5 }
					elsif ($rows > 5) { $rows = 4 }
					elsif ($rows > 3) { $rows = 3 }

					$html .=
qq[<textarea name="$var" cols="50" rows="$rows" wrap="virtual" onFocus="this.select()" onChange="check_input(this,'text',$size)"$title_str>$default</textarea>\n];
				}
				last SWITCH;
			}

			if ($type =~ /^d/) {	# date
				$title_str = $help ? qq[ title="$help"] : '';
				$html .=
qq[<input type="text" name="$var" value="$default" size="10" class="date" onChange="checkDate(this)" onFocus="this.select()" onChange="check_input(this,'date')"$title_str>\n];

				#
				# Return calendar button image link
				#

				$cal_script ||= qq[<script src="$Scripts_Lib/calendar.js" type="text/javascript"></script>\n];
				my $tabindex_str = qq[\n tabindex="$No_Tabindex"];
				$html .= qq[<a
 href="javascript:displayCalendar(1,'$var',self.document.parameters.$var.value)"
 title="Display a calendar."$tabindex_str
><img
   src="$Images/calendar.gif"
   border=0
   onMouseOver="this.src='$Images/calendar_hilite.gif'"
   onMouseOut="this.src='$Images/calendar.gif'"
></a>];
				last SWITCH;
			}

			if ($type =~ /^h/) {	# time
				$title_str = $help ? qq[ title="$help"] : '';
				$size ||= 10;
				$html .=
qq[<input type="text" name="$var" value="$default" size="$size" onChange="check_input(this,'time')"$title_str>\n];
				last SWITCH;
			}

			if ($type =~ /^n/) {	# number
				$title_str = $help ? qq[ title="$help"] : '';
				($size) = $type =~ /t(\d*)/;
				$size ||= 10;
				$html .=
qq[<input type="text" name="$var" value="$default" size="$size" onFocus="this.select()" onChange="check_input(this,'number',$size)"$title_str>\n];
				last SWITCH;
			}

			if ($type =~ /^m/) {	# money
				$title_str = $help ? qq[ title="$help"] : '';
				($size) = $type =~ /t(\d*)/;
				$size ||= 10;
				$html .=
qq[<input type="text" name="$var" value="$default" size="$size" onFocus="this.select()" onChange="check_input(this,'money',$size)"$title_str>\n];
				last SWITCH;
			}

			if ($type =~ /^_/) {	# drop down list
				($list_name) = $type =~ /_(.*)/;
				$html .= Acme::HTML::ListMenus->$list_name($var,$default);
				$html .= " $help" if $help;
				last SWITCH;
			}
		}

		$html .= "</td>\n</tr>\n";
	}
	$html .= "</table>\n";
	
	return $cal_script . $html;
}

#------------------------------------------------------------------------------

sub get_instructions_html {
	my $instructions = shift or return '';

	#
	# Return instructions html text enclosed in button that shows/hides them.
	#

	my $instr = qq[<div align="left" class="instructions">\n]
	  . start_hiding("instruction_body",
		more_label	 => "Instructions",
		more_description => "Show instructions.",
		less_label	 => "Instructions",
		less_description => "Hide instructions.",
		);

	$instr .= <<EOF;
<div style="border:solid black 1px;padding:2ex;">
$instructions
</div>
EOF
	$instr .= stop_hiding("instruction_body") . qq[</div>\n];

	return <<EOF;
<div style="margin:1ex 0ex;">
$instr
</div>
EOF
}

sub get_instructions_file {
	my $instructions_file = shift or return '';

	open INSTR, '<', $instructions_file or return '';

	#
	# Read instructions file.
	#

	local $/;	# read entire file
	my $instructions = <INSTR>;
	close INSTR;

	#
	# Extract instructions between <body> and </body> tags.
	#

	$instructions =~ s|^.*?<body.*?>||si;
	$instructions =~ s|</body.*||si;

	#
	# Build button
	#

	return get_instructions_html($instructions);
}

sub get_instructions_url {
	my $instructions_url = shift or return '';

	#
	# Return url in a <a> tag.
	#

	return <<EOF;
<a href="$instructions_url" onClick="new_target(this,'help')" title="Show Instructions Window" class="instructions">Instructions</a>
EOF
}

1;
