package Acme::HTML::SearchControl;

=head1 NAME

Acme::HTML::SearchControl - return HTML code for search controls

=head1 SYNOPSIS

  use Acme::HTML::SearchControl;

  my $customer_search_control = new Acme::HTML::SearchControl
  				table	  => 'customers',
				title	  => 'Customers',
  print $customer_search_control;

You can also be very explicit:

  my $customer_search_control = new Acme::HTML::SearchControl
  				table	  => 'customers',
				title	  => 'Customers',

				searches => '/templates/customer_searches.htm', (depreciated)
				reports => '/cgi-bin/myreports.cgi',
				query	=> 'customer_query.htm',
				help	=> 'search_instructions.htm',
				fields	=> 'customers customer_info',
				options	=> 'customer_search',

				view	=> '/cgi-bin/customers/view.pl',
				edit	=> '/cgi-bin/customers/edit.pl',
				list	=> '/cgi-bin/customers/list.pl',
				count	=> '/cgi-bin/customers/list.pl',
				new	=> '/cgi-bin/customers/new.pl',

				include => $extra_html;


=head1 DESCRIPTION

The methods in this package returns HTML code for a form to use to search
a database table.

Only the table parameter is required to display a search control, however,
it is assumed that there exists:

	* a template file of canned searches that has the name of
	  ${table}_canned_searches.htm in the Templates directory.
	  A link is not displayed if the file does not exist.
	  (this has been depreciated in favor of reports. mp 12/4/02)

	* an anchor in the options.htm file with the name of ${table}_options
	  that head a section of user option settings for the search control.
	  The options.cgi program automatically handles the search tables
	  entered into the Search_Tables array in the configuration file.

The form's Javascript also needs to include (hopefully to be automated...):

	* the lib/control.js library of utility functions.
	* the lib/common.js library for functions hilite() and unhilite().

=head1 CHANGE LOG

 8/8/00   Initial version.
 7/21/01  Changed new and edit programs to use pick.cgi to select records.
12/26/01  Added include option for extra buttons and links.
12/4/02   Added reports link, canned searches is now depreciated.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1998-2000, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

use Acme::HTML::ListMenus;
use Acme::Options;
use Acme::Base::Schema;
use Acme::Config;

our $Images    = Acme::Config->Images;		# location of images
our $Templates = Acme::Config->Templates;	# location of form templates
our $Help      = Acme::Config->Help;		# location of instructions

our $Select_Prog = '/cgi-bin/pick.cgi';		# program to select records
						# for View and Edit programs.

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	#
	# Note: this routine returns HTML code, not an object.
	#

	my $args = { @_ };

	# Get user options if not passed in arguments.
	my $opt = $args->{opt} || new Acme::Options;

	# Define simple variables to insert into the here document below.

	# Required parameters:
	my $table = $args->{table};

	# Optional parameters:

	# Note: searches has been depreciated. Only put in link if explicitly
	# set in program arguments.
	#my $searches = $args->{searches} || (
	#	-f "$Templates/${table}_canned_searches.htm"
	#	? "/cgi-bin/searches.cgi?template=${table}_canned_searches.htm"
	#	: '');
	my $searches = $args->{searches}|| '';

	my $reports  = $args->{reports}	|| "/cgi-bin/report.cgi?table=${table}";
	my $lists    = $args->{lists}	|| "/cgi-bin/list.cgi?table=${table}";
	my $query    = $args->{query}	|| "/cgi-bin/query.cgi?table=${table}";
	my $import   = $args->{import}	|| "/cgi-bin/import.cgi?table=${table}";
	my $update   = $args->{update}	|| "/cgi-bin/update.cgi?table=${table}";
	my $help     = $args->{help}	|| "/cgi-bin/help.cgi?file=${Help}/search_instructions.htm";
	my $fields   = $args->{fields}	|| "/cgi-bin/schema.cgi?table=${table}";
	my $options  = $args->{options}	|| "/cgi-bin/options.cgi#${table}_search_options";

	# Get default parameters from database schema:
	my $s = new Acme::Base::Schema;
	my $t = $s->t($table);

	my $title = $args->{title} || $t->title;

	my $view = $args->{view}
	  || $t->view_prog
	  || Acme::Config->View_Prog;
	my $edit = $args->{edit}
	  || $t->edit_prog
	  || Acme::Config->Edit_Prog;
	my $list = $args->{list}
	  || $t->list_prog
	  || Acme::Config->List_Prog;
	my $count = $args->{count}
	  || $t->count_prog
	  || Acme::Config->Count_Prog;
	my $new	= $args->{new}
	  || $t->new_prog
	  || Acme::Config->New_Prog;

	unless ($title) {
		# Capitalize table name
		$title = $table;
		$title =~ s/_+/ /g;			# change _ to spaces
		$title =~ s/\b(\w)/uc($1)/eg;		# capitalize words
	}

	#
	# Build list of left-side links
	#

	my $include = $args->{include} || '';		# extra buttons and links

	my $canned_searches_link = '';
	$canned_searches_link = <<EOF if $searches;
<a href="${searches}" target="${table}_searches" onClick="new_target(this)" class="link">Canned Searches</a><br>
EOF
	my $reports_link = '';
	$reports_link = <<EOF if $reports;
<a href="${reports}" target="_blank" onClick="new_target(this,'report')" class="link">Reports</a>
EOF
	my $lists_link = '';
	$lists_link = <<EOF if $lists;
<a href="${lists}" target="_blank" onClick="new_target(this,'list')" class="link">Lists</a>
EOF
	my $reports_lists_link	= $reports_link && $lists_link
				? "$reports_link/$lists_link<br>"
				: $reports_link
				? "$reports_link<br>"
				: "$lists_link<br>";
	my $import_link = '';
	$import_link = <<EOF if $import;
<a href="${import}" target="${table}_import" onClick="new_target(this,'import')" class="link">Import</a>
EOF
	my $update_link = '';
	$update_link = <<EOF if $update;
<a href="${update}" target="${table}_update" onClick="new_target(this,'update')" class="link">Update</a>
EOF
	my $import_update_link	= $import_link && $update_link
				? "$import_link/$update_link<br>"
				: $import_link
				? "$import_link<br>"
				: "$update_link<br>";
	my $prev_search = $opt->get("${table}Search") || '';

	#
	# Pre-build search history dropdown list
	#

	my @a = $opt->list("${table}Search");
	my $attr = qq(class="history" onChange="if (this.value) document.${table}_control_form.search.value = this.value;document.${table}_control_form.search.focus()" onClick="if (this.value) document.${table}_control_form.search.value = this.value" tabindex="-1");

	my $history_options = Acme::HTML::ListMenus->Menu(\@a,"${table}_history",'',40,$attr,'','noundef');

	#
	# Set search control options
	#

	my $search_enter = $opt->get("${table}SearchEnter");
	my $search_control_enter = $opt->get("${table}SearchControlEnter");

	###########################################################
	## Note: Need to remember to include library:            ##
	##	<script src="/scripts/common.js"></script>       ##
	##	<script src="/scripts/lib/control.js"></script>  ##
	## in the HTML page header.                              ##
	###########################################################

	return <<End_Of_HTML;
<!-- Start ${title} Control -->
<script language="JavaScript" type="text/javascript">
var ${table}SearchEnter          = "${search_enter}";
var ${table}SearchControlEnter   = "${search_control_enter}";
var ${table}_wn = Math.floor(Math.random() * 10000);

function show_${table}_control() {
	document.getElementById('${table}_control').style.display = "block";	
	document.getElementById('${table}_control_open').style.display = "none";
	window.document.${table}_control_form.search.select();
}
function hide_${table}_control() {
	document.getElementById('${table}_control').style.display = "none";
	document.getElementById('${table}_control_open').style.display = "block";
}
function ${table}_dispatch(request) {
	var f = document.${table}_control_form;
	var h = document.${table}_history_form.${table}_history;
	var s = f.search.value;

	f.request.value = request;

	f.target = "${table}_" + ${table}_wn++;
	f.method = "post";

	switch (request) {
	  case "View":
		if (s.length <= 0) {
			alert("You need to enter a search to select which record to view.");
			return false;
		}
		f.save_history.value = enter_history(s,h);

		if (s.match(/^\\s*\\d+\\s*\$/)) {	// search is ID number
			f.action = "${view}";
			f.method = "get";		// for auto-refresh hack
		} else {				// search string
			f.action = "${Select_Prog}";
			f.method = "post";
		}

		open_view("",f.target);
		f.submit();	
		break;

	  case "Edit":
		if (s.length <= 0) {
			alert("You need to enter a search to select which record to edit.");
			return false;
		}
		f.save_history.value = enter_history(s,h);

		if (s.match(/^\\s*\\d+\\s*\$/)) {	// search is ID number
			f.action = "${edit}";
		} else {				// search string
			f.action = "${Select_Prog}";
		}

		open_form("",f.target);
		f.submit();	
		break;

	  case "List":
		if (s.length > 0) f.save_history.value = enter_history(s,h);
		f.action="${list}";
		open_list("",f.target);
		f.submit();
		break;

	  case "Count":
		if (s.length > 0) f.save_history.value = enter_history(s,h);
		f.action="${count}";
		open_list("",f.target);
		f.submit();
		break;

	  case "New":
		var prev_search = f.search.value;
		f.search.value="";	// no search allowed for new records
		f.action="${new}";
		open_form("",f.target);
		f.submit();
		f.search.value = prev_search;
		break;
	}
	f.search.select();	// keep focus in form
}
function enterKey_${table}_control_form() {
	switch (${table}SearchEnter) {
	case "enter":
		return true;	// do nothing
	case "ctrl-enter":
		${table}_dispatch(${table}SearchControlEnter);
		return false;
	case "tab":
		// change focus to form enclosing buttons
		document.${table}_button_form.elements[0].focus();
		return false;
	}
	return;
}
function ctrlEnterKey_${table}_control_form() {
	${table}_dispatch(${table}SearchControlEnter);
	return;
}
</script>

<div id="${table}_control_open" class="control-closed" style="display:block">

<table border="0" cellspacing="0" cellpadding="2" class="control-closed">
  <tr valign="top"> 
    <td width="16"><a
         href="javascript:show_${table}_control()"
	 title="Show ${title} Control"><img
         src="$Images/control.gif"
         name="${table}_control_image"
         border="0"
         onMouseOver="hilite(this)" onMouseOut="unhilite(this)"
        ></a></td>

    <td width="105" align="right" nowrap><a
         href="javascript:show_${table}_control()"
         title="Show ${title} Control"
         onMouseOver="hilite(document.images['${table}_control_image'])"
         onMouseOut="unhilite(document.images['${table}_control_image'])"
         class="plain"
        ><nobr>${title}</nobr></a></td>

    <td width="340" class="search-closed"><a
         href="javascript:show_${table}_control()"
         title="Show ${title} Control"><img
         src="$Images/transparent.gif"
         width="340"
         height="18"
         border="0"
         onMouseOver="hilite(document.images['${table}_control_image'])"
         onMouseOut="unhilite(document.images['${table}_control_image'])"
        ></a></td>

    <td width="105">&nbsp;</td>

    <td width="16"><img src="$Images/blank.gif"></td>
  </tr>
</table>
</div>

<div id="${table}_control" class="control-opened" style="display:none">

<table border="0" cellspacing="0" cellpadding="2" class="control-opened">
  <tr valign="top"> 
    <td width="16"><a
         href="javascript:hide_${table}_control()"
	 title="Hide ${title} Control"><img
         src="$Images/control_open.gif"
         border="0"
         name="${table}_control_open_image"
         onMouseOver="hilite(this)" onMouseOut="unhilite(this)"
        ></a></td>

    <td><table cellspacing="0" cellpadding="0" border="0" width="100%">
    <tr>

      <td width="105" align="right" valign="top" nowrap><a
           href="javascript:hide_${table}_control()"
           title="Hide ${title} Control"
           onMouseOver="hilite(document.images['${table}_control_open_image'])"
           onMouseOut="unhilite(document.images['${table}_control_open_image'])"
  	   class="plain"
          ><nobr>${title}</nobr></a><br>
        $canned_searches_link
        <a href="${query}" target="${table}_query" onClick="new_target(this,'form')" class="link">Construct Search</a><br>
	$reports_lists_link
	$import_update_link
        <a href="${help}" onClick="new_target(this,'help')" class="link">Instructions</a><br>
        <a href="${fields}" onClick="new_target(this,'doc')" class="link">Columns</a><br>
        <a href="${options}" onClick="new_target(this,'options')" class="link">Options</a><br>
      </td>
  
      <td width="340" class="search-open"> 
        <form name="${table}_control_form"
         action="/cgi-bin/list.cgi"
         method="post"
         target="new_${table}"
         style="margin:0px"
        >
  
        <textarea name="search" class="control">${prev_search}</textarea>
  
        <input type="hidden" name="table" value="${table}">
        <input type="hidden" name="request" value="List">
        <input type="hidden" name="save_history" value="yes">
        </form>
  
      </td>
      </tr>
      <tr>
      <td width="105" align="right" valign="middle" nowrap>
        <font class="history">History</font> 
      </td>
      <td class="search_open">
        <form name="${table}_history_form"
         style="margin:2px 0 0 0"
        >
	$history_options
        </form>
      </td>

    </tr>
    </table></td>

    <td width="105"><form name="${table}_button_form"><table border="0" cellspacing="0" cellpadding="1">
        <tr> 
          <td><input type="button" value="View" class="control" accesskey="V" onClick="${table}_dispatch('View')"></td>
        </tr>
        <tr> 
          <td><input type="button" value="Edit" class="control" accesskey="E" onClick="${table}_dispatch('Edit')"></td>
        </tr>
        <tr> 
          <td><input type="button" value="List" class="control" accesskey="L" onClick="${table}_dispatch('List')"></td>
        </tr>
        <tr> 
          <td><input type="button" value="Count" class="control" accesskey="C" onClick="${table}_dispatch('Count')"></td>
        </tr>
        <tr>
          <td><input type="button" value="New" class="control" accesskey="N" onClick="${table}_dispatch('New')"></td>
        </tr>
      </table></form></td>

    <td width="16"><img src="$Images/blank.gif"></td>
  </tr>
</table>
$include
</div>
<!-- End ${title} Control -->

End_Of_HTML
}

1;
