package Acme::HTML::SectionControl;

=head1 NAME

Acme::HTML::SectionControl - return HTML code for section controls

=head1 SYNOPSIS

  use Acme::HTML::SectionControl;

  my $secctl = new Acme::HTML::SectionControl
  			name => 'Orders',
			contents => $prepared_order_html_text;
  print $secctl->as_html;

You can also be very explicit:

  my $secctl = new Acme::HTML::SectionControl
  			name => 'Orders',
			id   => $order_id;
  print $secctl->begin;
  print $secctl->begin_item;
  print $prepared_order_html_text_1;
  print $secctl->end_item;
  print $secctl->begin_item;
  print $prepared_order_html_text_2;
  print $secctl->end_item;
  print $secctl->end;

=head1 DESCRIPTION

Section controls are HTML tags and JavaScript code that surround page sections.
They provide a hide and show buttons that makes the section contents disappear
and reappear.

Item controls have buttons that go to the top of the section.

The as_html() method returns HTML text for the entire section.
The contents of the section must be passed to new() as a named parameter and
there can not be any item controls within the section.

The JavaScript file C<common.js> must be included into the header section of
every HTML page using these controls.

The optional C<id> parameter to new(), if specified, should be a unique C<id>,
one that is not used in any other control sections. If an C<id> is not
specified, an internal counter is used.

=head1 CHANGE LOG

 8/13/00   Initial version.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1998-2000, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Acme::Config;

$Acme::HTML::SectionControl::Images = Acme::Config->Images;	# location of image files

#
# Define global to use for missing IDs.
#

$Acme::HTML::SectionControl::Section_Count = 0;

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my $self = {
		id	=> ++$Acme::HTML::SectionControl::Section_Count,
		name	=> '',
		title	=> '',
		display	=> 0,	# section not initially displayed as default
		contents=> '',	# optional contents of section
		@_
		};

	$self->{title} = _capitalize($self->{name}) unless $self->{title};

	return bless $self, $class;
}

#------------------------------------------------------------------------------

sub as_html {
	my $self = shift;

	# Return HTML for entire section.
	return $self->begin . $self->{contents} . $self->end;
}

#------------------------------------------------------------------------------

sub begin {
	my $self = shift;

	while (my $k = shift) { $self->{$k} = shift; };	# reassign options

	# Return HTML code for the beginning of a section.

	my $id = $self->{id} || '';
	my $name = $self->{name} || '';
	my $title = $self->{title} || '';

	#########################################################
	## Note: Need to remember to include library:
	##	<script src="/scripts/common.js"></script>
	## in the HTML page header.
	#########################################################

	my $html = <<EOF;	# returned string

<!--  Begin ${title} Control Header -->
<script language="JavaScript" type="text/javascript">
function show_${name}${id}_control() {
	var c;
	c = document.getElementById('${name}${id}_control');
	c.style.display = "block";	
	c = document.getElementById('${name}${id}_control_open');
	c.style.display = "none";
}
function hide_${name}${id}_control() {
	var c;
	c = document.getElementById('${name}${id}_control');
	c.style.display = "none";       
	c = document.getElementById('${name}${id}_control_open');
	c.style.display = "block";
}
EOF

	my ($collapsed_display, $expanded_display);

	if ($self->{display}) {
		$collapsed_display="none";
		$expanded_display="block";
	}
	else {
		$collapsed_display="block";
		$expanded_display="none";

		# Section is hidden so track the anchors defined in the
		# section so that the section can be displayed before
		# the browser tries to jump to the anchor.
		# This must be done on the client side as there is no way
		# to find the url hash on the server side at this time (7/99).

		$html .= <<EOF;

if (typeof anchor_defined != "object") anchor_defined = {};
EOF
	}

	$html .= <<EOF;

</script>

<div id="${name}${id}_control_open" style="display:${collapsed_display}; text-align:center; margin-top:3pt">

<table border="0" cellspacing="0" cellpadding="2" width=100%>
  <tr valign="top"> 
    <td width=16><a
         href="javascript:show_${name}${id}_control()"><img
         src='$Images/control.gif'
         alt='Show $title'
	 name='${name}${id}_control_image'
         border=0
         onMouseOver="hilite(this)" onMouseOut="unhilite(this)"
        ></a>
    </td>

    <td width=105 align=left nowrap><nobr><a
         href="javascript:show_${name}${id}_control()"
         title='Show $title'
         onMouseOver="hilite(self.document.images['${name}${id}_control_image'])"
	 onMouseOut="unhilite(self.document.images['${name}${id}_control_image'])"
	 class="plain"
        >$title</a></nobr>
    </td>

    <td>
    </td>

    <td width=16><img src='$Images/transparent.gif'></td>
  </tr>
</table>
</div>

<div id="${name}${id}_control" style="display:${expanded_display}; margin-top:3pt; padding-bottom:12pt">

<table border="0" cellspacing="0" cellpadding="2" width=100%>
  <tr valign="top"> 
    <td width=16><a
         href="javascript:hide_${name}${id}_control()"><img
         src='$Images/control_open.gif'
         alt='Hide $title'
	 name='${name}${id}_control_open_image'
         border=0
         onMouseOver="hilite(this)" onMouseOut="unhilite(this)"
        ></a>
    </td>

    <td width=105 align=left nowrap><nobr><a
         href="javascript:hide_${name}${id}_control()"
         title='Hide $title'
         onMouseOver="hilite(self.document.images['${name}${id}_control_open_image'])"
	 onMouseOut="unhilite(self.document.images['${name}${id}_control_open_image'])"
	 class="plain"
        >$title</a></nobr>
    </td>

    <td><a
         href="javascript:hide_${name}${id}_control()"><img
         src='$Images/transparent.gif'
         alt='Hide $title'
	 width=100%
	 height=100%
         border=0
         onMouseOver="hilite(self.document.images['${name}${id}_control_open_image'])"
	 onMouseOut="unhilite(self.document.images['${name}${id}_control_open_image'])"
        ></a>
    </td>

    <td width=16><img src='$Images/transparent.gif'></td>
  </tr>
</table>
<!--  End ${title} Control Header -->
EOF
	return $html;
}

#------------------------------------------------------------------------------

sub end {
	my $self = shift;

	my $id = $self->{id} || '';
	my $name = $self->{name} || '';
	my $title = $self->{title} || '';

	my $html = <<EOF;	# returned string
<!--  Begin ${title} Control Trailer -->
EOF
	unless ($self->{display}) {
		# Display this section if the url hash jumps to an anchor
		# defined in this section.
		$html .= <<EOF;

<script language="JavaScript" type="text/javascript">
if (anchor_defined[self.location.hash]) show_${name}${id}_control();
</script>
EOF
	}

	$html .= <<EOF;

</div>
<!--  End ${title} Control Trailer -->

EOF
	return $html;
}

#------------------------------------------------------------------------------

sub begin_item {
	my $self = shift;

	my $id = $self->{id};
	my $name = $self->{name};
	my $title = $self->{title};

	return <<EOF;	# returned string
  <!--  Begin ${title} Item Header -->
  <tr>
    <td width=16><img src='$Images/transparent.gif'></td>
    <td colspan=2>
        <hr>
    </td>
    <td width=16 valign=baseline><img src='$Images/transparent.gif'></td>
  </tr>

  <tr>
    <td width=16 valign=baseline><a
         href="javascript:hide_${name}${id}_control()"><img
         src='$Images/small_control_open.gif'
         alt='Hide $title'
         border=0
         onMouseOver="hilite(this)" onMouseOut="unhilite(this)"
        ></a>
    </td>

    <td colspan=2>
  <!--  End ${title} Item Header -->
EOF
}

#------------------------------------------------------------------------------

sub end_item {
	my $self = shift;

	return <<EOF;
  <!--  Begin $self->{title} Item Trailer -->
    </td>
    <td width=16 valign=baseline><img src='$Images/transparent.gif'></td>
  </tr>
  <!--  End $self->{title} Item Trailer -->
EOF
}

#------------------------------------------------------------------------------
#
# Private
#
#------------------------------------------------------------------------------

sub _capitalize {
	my $title = shift;
	$title =~ s/_/ /g;	# change _ to spaces
	$title =~ s/\bid\b/ID/;	# uppercase ID
	$title =~ s/\b(\w+)\b/ucfirst(lc($1))/eg; # capitalize
	return $title;
}

1;
