package Acme::HTML::Tag;

=head1 NAME

Acme::HTML::Tag - Parse HTML tags.

=head1 SYNOPSIS

    use Acme::HTML::Tag;

    $str = '<input type=checkbox name="field_name" value="This is On" checked>';

    my $tag = new Acme::HTML::Tag $str;

    print "tag type = " . $tag->type;

    my $tag_type = $tag->get('type');
    $tag->set(type=>'hidden');

    print $tag->as_html;

There is a utility routine interface too:

    use Acme::HTML::Tag	qw(parse_tag);

    my %tag = parse_tag($str);

returns the hash:

    begin_tag => '<'
    tag       => 'input'
    type      => 'checkbox'
    name      => 'field_name'
    value     => 'This is On'
    checked   => 'yes'
    end_tag   => '>'

=head1 DESCRIPTION

The I<Acme::HTML::Tag> object contains the components of a HTML tag. Properties of
the object are the (lowercase) key words of the tag attributes. Single word
keys that have no value are assigned the value of "yes".

Only the contents between < and > are parsed, not between a beginning
and ending tag. In other words, the tag object holds the contents of
'<textarea class="test">' but not the contents of
'<textarea>testing</textarea>'.

Tag attribute values can be obtained by executing a method of the attribute's
name, like $tag->type. To get an attribute whose name is not known at compile
time use the get() method like $tag->get('type'). Both methods return an
empty string (not undef) if a tag attribute or its value does not exist.

To change the value of an attribute, use the set() method like
$tag->set(type=>'hidden'). It returns the previous value of the attribute.

The as_html() method reconstructs the tag into a string. It has been tested
mostly with form tags <input> and <textarray>. The list of single keyword
settings such as C<checked> or C<readonly> which must be processed specially
is incomplete for other tags.

The as_select_attr() method returns a string of attribute settings suitable
for use in a <select> tag. It is used to convert <input> tags into <select>
tags for dropdown lists. See I<Acme::Form::Filler> and I<Acme::HTML::ListMenus>.

The parse_tag() parses HTML tag and returns a hash of attribue key/value pairs.
Keywords are converted to lowercase before storing in hash. Single word
keywords without a value are assigned the value of "yes".

=head1 CHANGE LOG

 4/9/01  Initial version.
 6/23/01 Added as_select_attr() method to return attribute tags that are NOT
         illegal in <select> tags. Also added begin_tag and end_tag so
         as_html() is more faithful returning the equivilant tag that was
	 parsed, even if XHTML.
 6/3/02  Changed debug statements to Acme::HTML::debug().

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use strict;

require 5.6.0;
require Exporter;

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(parse_tag);

use Acme::HTML	qw(escape_html unescape_html debug debug_on);

#debug_on();

#
# Set order of parameters output in as_html();
#

our @params = ( 'type', 'name', 'id', 'value', 'class' ); 

#
# Build hash of parameters to skip when building a tag in as_html().
#

our %skip = map { $_, 1 } 'tag', 'begin_tag', 'end_tag', @params;

#
# Build hash of boolean parameters to treat specially in as_html().
#

our %single_word_param = map { $_, 1 } 'checked', 'readonly', 'disabled';

#
# Build hash of attributes that are not valid in <select> tags.
#

our %non_select_attr = map { $_, 1 } 'type','name','value','size','rows','readonly','onfocus';

#------------------------------------------------------------------------------
#
# Object methods
#
#------------------------------------------------------------------------------

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my $self = { parse_tag(@_) };

	return bless $self, $class;
}

sub get {
	my ($self,$key) = @_;
	return $self->{$key} || '';
}

sub set {
	my ($self,$key,$val) = @_;
	my $prev = $self->{$key};
	$self->{$key} = $val;
	return $prev;
}

sub as_html {
	my $self = shift;

	my $v = $self->{tag};

	for (@params) {
		$v .= qq[ $_="] . escape_html($self->{$_}) . qq["]
		  if defined $self->{$_};
	}

	#
	# Append remaining parameters to value string.
	#

	for (keys %$self) {
		next if $skip{$_};
		if ($single_word_param{$_}) {
			$v .= " $_" if $self->{$_} && $self->{$_} =~ /yes/i;
			next;
		}
		$v .= qq[ $_="] . escape_html($self->{$_}) . qq["];
	}

	#
	# Return fully constructed tag with delimiters.
	#

	return $self->{begin_tag} . $v . $self->{end_tag};
}

sub as_select_attr {
	my $self = shift;

	#
	# Derive attributes of <select> tag from those of an <input>
	# or <textarea> tags. Define only those that are legal or make
	# sense for a <select> tag. 
	#

	my $v = '';

	for (@params) {
		next if $non_select_attr{$_};
		$v .= qq[ $_="] . escape_html($self->{$_}) . qq["]
		  if defined $self->{$_};
	}

	#
	# Append remaining parameters to value string.
	#

	for (keys %$self) {
		next if $skip{$_};
		next if $non_select_attr{$_};
		if ($single_word_param{$_}) {
			$v .= " $_" if $self->{$_} && $self->{$_} =~ /yes/i;
			next;
		}
		$v .= qq[ $_="] . escape_html($self->{$_}) . qq["];
	}

	#
	# Change textarea's rows= attribute into select's size= attribute.
	#

	$v =~ s/rows=/size=/i;

	#
	# Return string of attributes settings
	#

	return $v;
}

sub AUTOLOAD {
        my $self = shift;
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;
	return if $method eq 'DESTROY';		# gobble

	return $self->get($method);
}

#-------------------------------------------------------------------------------
#
# Utility routine
#
#-------------------------------------------------------------------------------

sub parse_tag {
	my $str = shift;

	my %tags;			# returned hash

	#
	# Save beginning of string to tag delimiter.
	#

	$str =~ s/^(.*?<)//s;
	$tags{begin_tag} = $1 or return;

	#
	# Save end of tag to end of string.
	#

	$str =~ s/(\s*\/?>.*)$//s;
	$tags{end_tag} = $1 or return;

	#
	# Get tag id (eg. input, div, span, ...).
	#

	$str =~ s/(\w+)//s;
	$tags{tag} = lc $1 or return;

	#
	# Gather all parameters with values surrounded by quotes.
	#

	while ($str =~ s/\b(\w+)\s*=\s*(['"])(.*?)\2//s) {
		$tags{lc $1} = unescape_html($3);
	}

	#
	# Gather all parameters with values not surrounded by quotes.
	#

	while ($str =~ s/\b(\w+)\s*=\s*(\w+)//s) {
		$tags{lc $1} = unescape_html($2);
	}

	#
	# Gather all "single word" parameters. Assign "yes" as their value.
	#

	while ($str =~ s/\b(\w+)//s) {
		$tags{lc $1} = 'yes';
	}

	#
	# At this point, there should not be anything left in $str
	# (except spaces, and possibly a trailing / if using XHTML).
	#

	debug("Tag: tags structure: ",\%tags,"Tag: remaining = '$str'");

	return %tags;
}

1;
