package Acme::HTML::UPC;

=head1 NAME

Acme::HTML::UPC - Universal Product Code

=head1 SYNOPSIS

    use Acme::HTML::UPC;

    my $upc_img = Acme::HTML::UPC::upc_img(1234567890,$attr, ...);

    my $upc_src = Acme::HTML::UPC::upc_src(1234567890);
      or
    my $upc_src = Acme::HTML::UPC::upc_src(
    				type => 1,
				mfr  => 12345,
				prod => 67890,
				);

=head1 DESCRIPTION

The upc_src() routine returns the url of a GIF file that holds the
specified UPC code. The UPC code is specified 
by the 5 digit manufacturers number and the 5 digit product number.
The product type used by default when not specified in the arguments is in
the package global variable C<Acme::HTML::UPC::Product_Type>.

The upc_img() routine returns a <img> tag with the C<src> attribute set
to the url returned by upc_src() above. The first arguemnt is a 10 digit
UPC number. The remainder arguments are attributes for the generated <img>
tag.

UPC's are stored in GIF image files with their names corresponding to the
code. A file is looked for first and returned if found. If an existing file
does not exist for the given UPC, one is created using the C<pbmupc> and
C<ppmtogif> programs to make a portable bitmap of the UPC and convert it
into a GIF file. Both programs are by Jef Poskanzer (c) 1989 and are usually
distributed with the GNU utilities.

UPC bitmaps are 230 pixels wide by 175 pixels high.

=head1 CHANGE LOG

 10/3/01  Initial version.
 10/20/01 Changed to include <img> tag so that when empty codes shows nothing
 	  in browser.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000,2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.000;
require Exporter;

@ISA = qw(Exporter);
@EXPORT_OK = qw(upc_img upc_src);

use Acme::Config;

$Base         = Acme::Config->Base;			# base url
$DocumentRoot = Acme::Config->DocumentRoot;		# corresponds to Base
$UPC_Images   = Acme::Config->UPC_Images;		# UPC images directory
$Product_Type = Acme::Config->Product_Type || 1;	# default product type

sub upc_src {

	my ($type,$mfr,$prod);
	if (@_ > 1) { 
		my $opt = { @_ };
		$type = $opt->{type} || $Product_Type;
		$mfr  = $opt->{mfr};
		$prod = $opt->{prod};
	}
	else {
		my $upc = shift;

		return '' unless ($upc =~ /^\d+$/);

		# Parse UPC components from argument
		($type,$mfr,$prod) = $upc =~ /^(\d)(\d\d\d\d\d)(\d\d\d\d\d)$/
		  or ($mfr,$prod) = $upc =~ /^(\d\d\d\d\d)(\d\d\d\d\d)$/;
		$type = $Product_Type unless defined $type;
	}

	return '' if !defined $type || !defined $mfr || !defined $prod;

	my $upc  = "$type$mfr$prod";
	my $url  = "$Base$UPC_Images/$upc.gif";
	my $file = "$DocumentRoot$UPC_Images/$upc.gif";

	return $url if -r $file;
		
	#
	# Create image file of UPC
	#

	delete @ENV{ qw(IFS CDPATH ENV BASH_ENV) };     # untaint
	$ENV{PATH} = '/bin:/usr/bin';                   # untaint

	system("/usr/bin/pbmupc -s1 '$type' '$mfr' '$prod' | /usr/bin/ppmtogif > $file 2>/dev/null");

	return $url if -r $file;
	return '';
}

sub upc_img {

	my $upc = shift;

	my $url = upc_src($upc) or return '';

	my $attr = join ' ', @_;
	$attr &&= " $attr";

	return qq[<img src="$url"$attr>];
}

1;
