package Acme::Lister;

=head1 NAME

Acme::Lister - List processing methods

=head1 SYNOPSIS

    use Acme::Lister;

    $l = new Acme::Lister (
	title        => 'My Title',	# list title
    	sql          => $sql,		# sql to execute
	format       => $format,	# a sprintf() type format
    	);

    $l->print;

or to be more specific by supplying routines to place a title and format rows:

    use Acme::Lister;

    $l = new Acme::Lister (
    	sql          => $sql,		# sql to execute

	# Following $statements can be strings to eval or sub references.
	before_list => $statements,	# execute statements before list
	after_list  => $statements,	# execute statements after list
	before_row  => $statements,	# execute statements before row
	after_row   => $statements,	# execute statements after row
	every_row   => $statements,	# execute statements on every row

	sandbox      => "package_name", # optional package where list produced
    	);

    $l->print;

or what is usually a little clearer by placing before and after clauses in the
print method. It allows before_group/after_group clauses to be attached to 
column names (note that before_group/after_group are arrays, not hashes,
to preserve their order):

    $l = new Acme::Lister (
	title        => 'Title',	# list title
    	sql          => $sql,		# sql to execute
	format      => $format,		# format used if every_row not defined
    	);

    $l->print (
	before_list   => $statements,	# execute statements before list
	after_list    => $statements,	# execute statements after list

	before_group => [ col1 => $statements,
	                  col2 => $statements,
		 	],
	after_group  => [ col1 => $statements,
	                  col2 => $statements,
			],

	before_row   => $statements,	# execute statements before row
	after_row    => $statements,	# execute statements after row
	every_row    => $statements,	# execute statements on every row
    );

or to produce the list in memory, use the list() method which is synonomous
with print(). Limits can be specified in either the list constructor or in
the print() method. The $statements can be strings to be eval'ed, references
to strings to eval, or references to functions to execute.

    $l = new Acme::Lister (
	title        => 'Title',	# list title
    	sql          => $sql,		# sql to execute
	format       => $format,	# format used if every_row not defined
    	);

    my $list = $l->list (		# return list
	before_list  => $statements,	# execute statements before list
	after_list   => $statements,	# execute statements after list

	before_group => [ col1 => $statements,
	                  col2 => $statements,
			],
	after_group  => [ col1 => $statements,
	                  col2 => $statements,
			],

	before_row   => $statements,	# execute statements before row
	after_row    => $statements,	# execute statements after row
	every_row    => $statements,	# execute statements on every row

	limits       => { 100  => $statements,
			  200  => $statements,
			},
    );



=head1 DESCRIPTION

Acme::Lister provides a generic interface to printing lists from
SQL statements.

A sandbox is used to construct lists as global variables are used extensively.
The sandbox package name is I<Acme::Lister::Sandbox> if one is not
specified in the arguments. The C<before_group/after_group> and C<every_row>
statements and subroutines are all executed within the sandbox. A global array
of database values is stored in the array @a in Acme::Lister::Sandbox
for each record which can be used in lieu of field name variables. Your own
sandbox (e.g. a package) can be specified in the new() constructor.

This module inherits the I<Acme::Db> object to access the database. The
enhancements this module gives over I<Acme::Db> are: 1) executing the query
in a sandbox, 2) easier method to specify "before group" and "after group"
control breaks, 3) initializing globals for mod_perl, and 4) triggers
events when specified limits are reached.

The C<before_group> and C<after_group> statements are executed when values
in the named columns change. It is best if the query sorts the information
on the same columns. The print() method allows C<before_group/after_group>
statements to be specified all at once in an array.

Columns can be specified in the C<before_group> and C<after_group> by name
or by number. Numbered columns begin with 1. They are numbered the same
as SQL's C<order by> clause. That is different than the numbers used in
Acme::Db C<before_group/after_group> methods which are 0 based.

Totals are kept for columns specified in the C<total> parameter, which
is an array of column names. The totals are kept in a special global hash
named %_Totals in the sandbox's namespace. That hash is known to the total()
routine located in the I<Acme::Lister::Functions> that can be used in a
format string processed by Local::Acme::Format::sprintf_cmd(), which is used
to format the list's footers (and headers and all rows).

Limits can be specified in the hash pointed to by the C<limit> parameter.
The hash has numeric keys. The hash values are statements or references
to functions which are executed when a limit is reached. If a reference
to a function is used, the function is called with the current record
count as its sole argument.

Letting the print() method call the C<before_list> and C<after_list> statements
makes the first record or last record values available to those routines.

The list() method works the same as the print() method except it produces the
list in memory instead of printing it as it is produced. The before_*() and
after_*() statements/routines specified in the arguments should return their
results so they can be collected in the list() method.

=head1 CHANGE LOG

 6/3/01   Initial version.
 6/19/01  Changed print() method to accept most of the list arguments.
 6/21/01  Added list() method to produce list in memory.
 8/3/01   Removed clean_sql() security checks. The routines who are creators
          of sql should do their own checking with security.
 8/25/02  Added totals.
 12/27/02 Added limits.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000-2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Acme::Db;
use Acme::Config;
use Carp;

@ISA = ( 'Acme::Db' );

#------------------------------------------------------------------------------

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	#
	# Create our name space which we will attach to the Acme::Db object.
	#

	my $lister = {
			total => undef,
			group => undef,
			};

	#
	# Create Db object and add our component (lister) to it.
	#

	my $self = new Acme::Db lister => $lister;

	bless $self, $class;

	$self->_set_list_options(@_);

	$self->query($self->{sql});

	#
	# Initialize sandbox of any previously defined globals.
	#

	if (! $self->{sandbox}) {
		$self->{sandbox} = 'Acme::Lister::Sandbox';
		$self->initialize_sandbox();
	}

	#
	# Bind array elements to sql column name variables in the sandbox.
	#

	$self->bind_columns_to_variables($self->{sandbox});

	#
	# Initialize totals accumulators if being used.
	#

	$self->_init_totals();

	return $self;
}

#------------------------------------------------------------------------------

sub print {
	my $self = shift;

	#
	# Print list as it is constructed.
	#

	$self->_print_list('print',@_);
}

#------------------------------------------------------------------------------

sub list {
	my $self = shift;

	#
	# Return list in memory.
	#

	return $self->_print_list('return',@_) || '';
}

#------------------------------------------------------------------------------

sub initialize_sandbox {
	my $self = shift;

	#
	# Delete globals of previously created list by removing all entries
	# in sandbox's symbol table except the global array @a used to hold
	# values read from the database. 
	#

	$sandbox = $self->{sandbox};
	eval qq[
		for (keys %${sandbox}::) {
			delete \$${sandbox}::{\$_} unless \$_ eq 'a';
		}
	];
}

#------------------------------------------------------------------------------
#
# Private methods
#
#------------------------------------------------------------------------------

sub _set_list_options {
	my $self = shift;

	my ($k, $v);
	while ($k = shift) {

		$v = shift;

		if ($k eq 'before_group') {

			#
			# Process before_group options
			#

			while (my $col = shift @$v, my $statement = shift @$v) {
				# Turn column number into 0 based numbers.
				$col -= 1 if $col =~ /^\d+$/;

				$self->before_group($col,$statement);
			}

			next;
		}

		if ($k eq 'after_group') {

			#
			# Process after_group options
			#

			while (my $col = shift @$v, my $statement = shift @$v) {
				# Turn column number into 0 based numbers.
				$col -= 1 if $col =~ /^\d+$/;

				$self->after_group($col,$statement);
			}

			next;
		}

		#
		# Change before_list/after_list triggers into
		# before_query/after_query triggers. They must be the same,
		# otherwise before_group triggers would execute before the
		# before_list trigger.
		#

		if ($k eq 'before_list') {

			$self->before_query($v);
			next;
		}

		if ($k eq 'after_list') {

			$self->after_query($v);
			next;
		}


		if ($k eq 'total') {

			$self->{total} = $v;	# pointer to an array
			next;
		}


		if ($k eq 'sandbox') {
			# Sandbox must be defined in constructor new() if used.
			croak("Cannot redefine sandbox from '$self->{sandbox}' to '$v'")
			  if $self->{sandbox} && $self->{sandbox} ne $v;
			
			# fall through
		}

		#
		# Set option
		#

		$self->{$k} = $v;
	}
}

#------------------------------------------------------------------------------

sub _init_totals {
	my $self = shift;

	#
	# Initialize hash to accumulate totals for columns.
	#

	if ($self->{total} && @{$self->{total}}) {
		eval q[%{].$self->{sandbox}.q[::_Totals} = (); ];

		my $init_totals = '';

		$init_totals  = qq[package $self->{sandbox};\n];
		$init_totals .= qq[no strict;\n];
		$init_totals .= qq[no warnings qw(uninitialized);\n];

		my ($col,$var);
		for $col (@{$self->{total}}) {
			($col) = $col =~ /([\w.]+)/;	# untaint
			($var = $col) =~ s/\./__/g;
			$init_totals .= qq[\$_Totals{'$col'} = 0;\n];
		}

		eval $init_totals;
	}
}

#------------------------------------------------------------------------------

sub _print_list {
	my $self = shift;

	#
	# Procedure for producing a list in memory or to print it.
	#

	# This is a utility method for both the print() and list() methods.
	# The before_*(), after_*(), and every_row() routines need to return
	# their results to be collected in the $list variable, whereas the
	# methods for the print() method should print their result and not
	# return a value.

	my $cmd = shift;	# either 'print' or 'return'

	#
	# Setup default routines to return list header if needed.
	#

	if (! $self->{before_list} && ($title = $self->{title})) {
		$self->{before_list} = q[no warnings qw(uninitialized);] .
				       $cmd . q[ "<h3>$title</h3>\n";];
	}

	#
	# Process arguments.
	#

	$self->_set_list_options(@_);

	#
	# Setup default routine to return list rows if needed.
	#

	if (! $self->{every_row} && ! $self->{before_group} && ! $self->{after_group}) {

		#
		# Define a "every_row" function if one is not defined.
		#

		if (my $format = $self->{format}) {

			#
			# Use format specified in list options.
			#

			$format =~ s/"/\\"/g; # escape " characters in $format
			$self->{every_row} = q[no warnings qw(uninitialized);] .
					     $cmd . q[ sprintf_cmd("].$format.q[",@a);];
		}
		else {
			#
			# Simply list values separated by a space character.
			#

			$self->{every_row} = q[no warnings qw(uninitialized);] .
					     $cmd . q[ (join ' ', @a);];
		}
	}

	#
	# Construction function to compute totals
	#

	my $compute_totals_func;	# code to execute to compute totals
	if ($self->{total} && @{$self->{total}}) {

		$self->_init_totals;	# might be repeating if done in new()

		#
		# Define code to execute for every row to update the totals
		#

		my $compute_totals;
		$compute_totals  = qq[sub {\n];
		$compute_totals .= qq[package $self->{sandbox};\n];
		$compute_totals .= qq[no strict;\n];
		$compute_totals .= qq[no warnings qw(uninitialized);\n];

		my ($col,$var);
		for $col (@{$self->{total}}) {
			($col) = $col =~ /([\w.]+)/;	# untaint
			($var = $col) =~ s/\./__/g;
			$compute_totals .= qq[\$_Totals{'$col'} += \$$var;\n];
		}
		$compute_totals .= qq[}\n];

		$compute_totals_func = eval $compute_totals;
	}

	#
	# Set limits to the list.
	#

	my $limits;
	if (exists $self->{limits}) {

		$limits = $self->{limits};
	}
	else {

		#
		# Set default limits for all lists that is specified in
		# the site configuration file (.../etc/site.conf).
		#

		require Acme::Lister::Limit;

		my $limit_abort = Acme::Config->List_Limit_Abort;
		my $limit_warn  = Acme::Config->List_Limit_Warn;

		$limits = $self->{limits} = {};
		$limits->{$limit_abort} = \&Acme::Lister::Limit::abort_func	if $limit_abort;
		$limits->{$limit_warn}  = \&Acme::Lister::Limit::warn_func	if $limit_warn;
	}

	#
	# Set array reference in sandbox to hold database values.
	#

	my $aref = eval q[\@{].$self->{sandbox}.q[::a}];

	#
	# Use Acme::Db private routine _do() to execute code in the sandbox.
	#

	no warnings qw(uninitialized);

	my $list = '';
	my $count = 0;
	my $stmt;
	while ($self->next($aref)) {
		$count++;

		#
		# Compute totals.
		#

		&$compute_totals_func if $compute_totals_func;

		#
		# Render list.
		#

		$list .= $self->_do($self->{sandbox},$self->{before_row})
		  if $self->{before_row};

		$list .= $self->_do($self->{sandbox},$self->{every_row})
		  if $self->{every_row};

		$list .= $self->_do($self->{sandbox},$self->{after_row})
		  if $self->{after_row};
		
		#
		# Test list limits.
		#

		if ($limits && ($stmt = $limits->{$count}) ) {

			$self->_do($self->{sandbox},$stmt,$count);

			if ($@) {

				#
				# Terminate list
				#

				# Note: after_list function changed to after_query
				# above in _set_list_options().

				$list .= $self->_do($self->{sandbox},$self->{after_query})
				  if $self->{after_query};
				
				last;
			}
		}
	}

	return $list if $list;
	return;
}

1;
