package Acme::Lister::Filter;

=head1 NAME

Acme::Lister::Filter - Make SQL to filter list

=head1 SYNOPSIS

    use Acme::Lister::Filter;

    my $fltr = new Acme::Lister::Filter
     -or-
    my $fltr = new Acme::Lister::Filter cgi=>$cgi,
    					temp_table=>"t_ids",
					key=>"customers.customer_id",
					expect=>$nrows;

    # Use for filters with a selector filter (table/key queries).
    my $sql = $fltr->as_sql . <<End_Of_SQL;
 select *
 from customers join t_ids on customers.customer_id = t_ids.customer_id;
 End_Of_SQL

    my $srch = $fltr->{searcher};  # Acme::Searcher object of cgi search script

    my $temp_table = $fltr->{temp_table};  # name of temporary table created

    my $temp_key = $fltr->{temp_key};      # name of temporary table's key column

    $fltr->confirm_expectations($count_of_rows);

=head1 DESCRIPTION

This module constructs a SQL statement from the C<sql>, C<search>,
C<selector>, C<include>, C<exclude>, C<select_all> and C<invert_all>
CGI parameters. The parameters are those produced by the C<list.cgi> program.

The SQL statement performs the query, applies the filters, and creates a
temporary table that has one column in it, that of key values for records
that match the criteria.

The name of the temporary table can be specified when creating the filter and
it can be retrieved by using the temp_table() method.

The search script gets parsed and evaluated by the as_sql() method. The
I<Acme::Searcher> object of the parsed search script can be retrieved by 
the searcher() method.

The fully qualified name of the key field needs to be specified to this filter.
It usually is inferred from the CGI C<table> parameter, but it can be specified
by the C<key> cgi parameter or a C<key> parameter to new().

The sole column in the temporary table has the same column name as the key.

Note that there isn't any actual action taken by this module. It only
returns a SQL script that is intended to be attached as a prefix onto your
SQL statement, of which they would then be submitted to the database engine
as one query for execution.

If the C<table> parameter is specified in the CGI or in the arguments to 
new(), then the default filters and selectors from the schema are applied
if they are not specified in the CGI.

The CGI arguments can be explicitly set in new() by passing it an argument
C<args> that points to a hash that has values for C<search>, C<selector>,
C<include>, and the other parameters.

=head2 Expectations

This filter module reconstructs the search parameters displayed by the
list.cgi program.  There will always be a possibility that the filter
does not give the number of rows displayed in the list, especially when a
Acme::Lister::Selector filter is used, and especially when the selector
is filtered on a regular expression that is evaluated by Perl when listed
and here by the database engine.

In the spirit of "allow but monitor", use the C<expected> argument to new() to
record the number of selected rows anticipated. After the list is processed,
call the confirm_expectations() method with the actual number of rows.
It will issue an error message and send email to the webmaster notifying
them of a discrepancy as soon as possible.

=head3 Caveat

Many of the high speed unloads/loads and updates to the database by the
database engine seldom return the number of rows affected to the program
in a standard or useful way, thwarting the use of confirm_expectations().

=head1 CHANGE LOG

 2/25/02  Initial version, an assimilation of various prototype routines.
 8/30/02  Moved filter processing to client JavaScript, which modifies the
          search parameter before being processed here, simplifying things.
 12/2/02  Incorporated new Acme::Lister::Selector with selector conditions.
 12/10/12 Added confirm_expectations() for error checking.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2002  Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use CGI;
use Acme::HTML		qw(unescape_html);
use Acme::Base::Schema;
use Acme::SQL;
use Acme::Searcher;
use Acme::Lister::Selector;
use Acme::Config;

#------------------------------------------------------------------------------

sub new {
	my $proto = shift;	# caller object (should be class/package name)
	my $class = ref($proto) || $proto;

	my $self = {
			cgi		=> undef,
			primary_table	=> undef,
			key		=> undef,
			sql		=> undef, 	# can be in cgi too
			searcher	=> undef,
			expect		=> undef,	# expected number of rows
			temp_table	=> 't_keyval',
			sel		=> undef,	# Lister::Selector object
			@_
			};

	$self->{cgi} ||= new CGI $self->{args} if ref($self->{args}) eq 'HASH';
	$self->{cgi} ||= new CGI;

	$self->{sql} ||= $self->{cgi}->param('sql');
	$self->{key} ||= $self->{cgi}->param('key');

	my $table = $self->{table} || $self->{cgi}->param('table');

	$self->{primary_table} ||= $table;

	if (! $self->{primary_table} && $self->{key}) {
		# Get table name from fully qualified key column name.
		($self->{primary_table}) = $self->{key} =~ /^(\w+)\./;
	}

	if (! ($self->{sql} && $self->{key})) {
		
		if (! $self->{primary_table}) {
			die "A table name or a fully qualified key name is required for"
			  . " filtering database records. Error"
			  if ! $self->{key};

			die "A table name or a sql query is required for"
			  . " filtering database records. Error"
			  if ! $self->{sql};
		}

		#
		# Get key from schema
		#

		my $schema = new Acme::Base::Schema;
		my $t = $schema->t($self->{primary_table})
		  or die "The database table '$self->{primary_table}' is not recognized";

		#
		# Set key column
		#

		$self->{key} ||= $self->{primary_table} . '.' . $t->key;

		#
		# Define SQL query from table
		#

		$self->{sql} ||= $t->list_sql;
	}
		
	#
	# Get table's selector/rejector clauses
	#

	$self->{sel} = new Acme::Lister::Selector
				table	=> $primary_table,
				cgi	=> $self->{cgi};

	#
	# Set temporary table name if not specified.
	#

	$self->{temp_table} ||= $Temp_Table;

	return bless $self, $class;
}

#-------------------------------------------------------------------------------

sub temp_table {
	my $self = shift;
	return $self->{temp_table};
}

#-------------------------------------------------------------------------------

sub temp_key {
	my $self = shift;

	(my $temp_key	= $self->{key}) =~ s/.*\.//;	# remove table name from key
	return "$self->{temp_table}.$temp_key";
}

#-------------------------------------------------------------------------------

sub searcher {
	my $self = shift;
	return $self->{searcher};	# return Acme::Searcher object
}

#-------------------------------------------------------------------------------

sub as_sql {
	my $self = shift;

	my $search	= $self->{cgi}->param('search');
	my $exclude	= $self->{cgi}->param('exclude');
	my $include	= $self->{cgi}->param('include');
	my $invert_all  = $self->{cgi}->param('invert_all');
	my $select_all  = $self->{cgi}->param('select_all');
	my $temp_table	= $self->{temp_table};
	(my $temp_key	= $self->{key}) =~ s/.*\.//;	# remove table name from key

	$invert_all     = '' unless 'yes' =~ /\Q$invert_all\E/i; # make boolean
	$select_all     = '' unless 'yes' =~ /\Q$select_all\E/i; # make boolean

	#
	# Unescape key values in $include and $exclude.
	#

	# Note: this is normally not needed as contents of form fields are
	# quoted and unquoted automatically. However, the $include/$exclude
	# lists of key values gets their values from the ID attribute of
	# the HTML check marks, which needs to be escaped there. So the
	# values entered into the $include/$exclude form fields are already
	# escaped values which need to be unescaped here.

	$include = unescape_html($include);
	$exclude = unescape_html($exclude);

	#
	# Construct SQL script to select the chosen rows.
	# Only select the key value and place results into a temporary file.
	#

	my $sqlo = new Acme::SQL $self->{sql};

	my $key = $self->{key} =~ /\./
		? $self->{key}
		: "$self->{primary_table}.$self->{key}";

	$sqlo->set_columns($key);
	$sqlo->set_into_temp($self->{temp_table});
	$sqlo->set_distinct();

	#
	# Alter query according to selector and invert all/select all settings
	#

	my $selector_clause = $self->{sel}->where_clause;

	if ($selector_clause && ! $select_all) {

		if (! $invert_all) {
			# Normal case
			$sqlo->add_where($selector_clause);
		}
		else {
			# Inverted
			$sqlo->add_where("not ($selector_clause)");
		}

		#
		# Add join to a selector column's table (if needed)
		#

		my $join_clause = $self->{sel}->join_clause;
		$sqlo->add_join($join_clause) if $join_clause;
	}
	$sqlo->sort('');	# no need to sort temporary file

	my $sql;
	if ($search) {

		#
		# Evaluate search script.
		#

		$self->{searcher} = new Acme::Searcher $search, $self->{cgi};

		#
		# Combine sql and search script.
		#

		$sql = $self->{searcher}->as_sql($sqlo);
	}
	else {
		$sql = $sqlo->as_sql;
	}

	$sql .= ";\n";	# terminate sql statement

	#
	# Alter the sql query according to the user checks/unchecks.
	#

	if ($select_all && $invert_all) {

		# It seems strange to delete all the selected rows, but
		# some are added back if listed in "include" parameter.
		# We should forego the above "select" statement and just
		# do a "create temp table $temp_table ($temp_key $temp_key_type);"
		# but we would have to put in logic to determine type of key.

		$sql .= "delete from $temp_table where (1 = 1);\n";
	}

	#
	# Delete those IDs excluded by the user.
	#

	# There are two methods used depending upon the length of the
	# $include or $exclude clauses. If they are suspected of being
	# too big to put into a "col in ('val1','val2')" type construct
	# for a SQL where clause because of limits the database engine
	# imposes on statement size, they are broken up into a bunch
	# of insert/delete statements to execute individually.

	my $Max_Len = 500;

	if ($exclude and ! $invert_all) {

		if (length $exclude < $Max_Len) {
			$sql .= <<End_Of_SQL;
delete from $temp_table
where $temp_key in ($exclude);
End_Of_SQL
		}
		else {
			for my $val (split /,/, $exclude) {
				$sql .= <<End_Of_SQL;
delete from $temp_table where $temp_key = $val;
End_Of_SQL
			}
		}
	}

	if ($include and $invert_all) {

		if (length $include < $Max_Len) {
			$sql .= <<End_Of_SQL;
delete from $temp_table
where $temp_key in ($include);
End_Of_SQL
		}
		else {
			for my $val (split /,/, $include) {
				$sql .= <<End_Of_SQL;
delete from $temp_table where $temp_key = $val;
End_Of_SQL
			}
		}
	}

	#
	# Append extra IDs "included" by the user
	#

	# There are two methods used depending upon the length of the
	# $include or $exclude clauses here too. Like above, if they are
	# suspected of being too big,they are broken up into a bunch
	# of insert/delete statements to execute individually.

	if ($include and ! $invert_all) {

		if (length $include < $Max_Len) {
			$sql .= <<End_Of_SQL;
insert into $temp_table ($temp_key) 
select $self->{key}
from $self->{primary_table}
where $self->{key} in ($include);
End_Of_SQL
		}
		else {
			for my $val (split /,/, $include) {
				$sql .= <<End_Of_SQL;
insert into $temp_table ($temp_key) value ($val);
End_Of_SQL
			}
		}
	}

	if ($exclude and $invert_all) {

		if (length $exclude < $Max_Len) {
			$sql .= <<End_Of_SQL;
insert into $temp_table ($temp_key) 
select $self->{key}
from $self->{primary_table}
where $self->{key} in ($exclude);
End_Of_SQL
		}
		else {
			for my $val (split /,/, $exclude) {
				$sql .= <<End_Of_SQL;
insert into $temp_table ($temp_key) value ($val);
End_Of_SQL
			}
		}
	}

	#
	# Return SQL script.
	#

	return $sql;
}

#------------------------------------------------------------------------------

sub confirm_expectations {
	my $self = shift;

	my ($count,$expect) = @_;
	$expect = $self->{expect} if ! defined $expect;

	die "Lister::Filter: confirm_expectations: No expectations were set."
	  if ! defined $expect;
	
	if ($count != $expect) {

		require Acme::Db;
		require Data::Dumper;

		# Get function used to select records from Lister::Selector.
		my $sel = $self->{sel};
		my $func = $sel->{selector}
			 ? $sel->_selector_func()
			 : $sel->{table}
			 ? $sel->_selector_func_old_style()
			 : 'sub {return 1}';		# select all

		Acme::Db->error(<<EOF . Data::Dumper::Dumper($self->{cgi}));
Program: $0
Lister::Filter did not return the expected number of rows $self->{expect}.

Actual number of rows processed was $count.

Selector function for list:

$func

CGI parameters:

EOF
		die "The number of database rows processed was not the number expected.\n"
		  . " There were $expect rows listed."
		  . " The number of rows processed was $count.\n"
		  . " Please check your results.";
	}
}

1;
