package Acme::Lister::Limit;

=head1 NAME

Acme::Lister::Limit - Methods supporting list limits

=head1 SYNOPSIS

    use Acme::Lister::Limit;

    our $l = new Acme::Lister (
		sql          => $sql,
		sandbox      => 'list',
		);

    $l->print(
		before_list  => \&display_list_header,
		after_list   => \&display_list_footer,
		every_row    => \&display_list_item,
		limits       => {
				100  => \&Acme::Lister::Limit::warn_func,
				1000 => \&Acme::Lister::Limit::abort_func
				},
		);


=head1 DESCRIPTION

The functions here support limits in lists rendered by I<Acme::Lister>.

Two functions are used as function references passed to the lister
object. The functions are called when a limit is reached and a warning 
needs to be issued and/or the list processing terminated.

The warn_func() function emails a warning to the Web administrators
but continues processing the list. The user does not get any indication
that an email warning has been issued.

The abort_func() function emails an error message to the Web administrators
and then terminates the list processesing. When abort_func() dies, the
C<Acme::Lister> discontinues reading any more records from the database
and prints its summary before terminating.

=head1 CHANGE LOG

 12/28/02  Initial version.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2002  Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Acme::Security::Alert;

#------------------------------------------------------------------------------

sub abort_func {
	my $limit = shift;

	my $to;		# alternate email addresses

	my $subject = qq[Aborting: Limit of $limit for $list_id list has been exceeded.];

	my $msg = <<End_Of_Msg;
The $list_id list has exceeded its limit of $limit records.
The list has been terminated.
End_Of_Msg

	Acme::Security::alert($msg, ($to || ''), $subject);
	die $msg;
}

#------------------------------------------------------------------------------

sub warn_func {
	my $limit = shift;

	my $to;		# alternate email addresses

	my $subject = qq[Warning: Limit of $limit for $list_id list has been reached.];

	my $msg = <<End_Of_Msg;
The $list_id list has reached a limit of $limit records and is generating this notice.
The list is continuing to be rendered. This is only a warning.
End_Of_Msg

	Acme::Security::alert($msg, ($to || ''), $subject);
}

1;
