package Acme::Lister::Selector;

=head1 NAME

Acme::Lister::Selector - Make SQL to select items of a list

=head1 SYNOPSIS

    use Acme::Lister::Selector;

    my $sel = new Acme::Lister::Selector
     -or-
    my $sel = new Acme::Lister::Selector cgi=>$cgi,
					 selector=>"address_usage_codes.usage";
     -or-
    my $sel = new Acme::Lister::Selector table=>'address_usage_codes',
					 selector=>"address_usage_codes.usage",
					 selector_cond=>'eq',
					 selector_param=>'y';

    print $checked if $sel->is_checked;

    my $sql = 'select * from ' . $sel->join_clause . 'where '. $sel->where_clause;

    # Print text explaining the selector conditions
    print $sel->explanation;

    # List conditions and their descriptions
    for (Acme::Lister::Selector->conditions) {
    	print "$_: ", Acme::Lister::Selector->description($_), "\n";
    }


=head1 DESCRIPTION

A list selector is a special kind of filter for lists. It is used to display
a checkbox next to a row on a list which can be "checked" (e.g. selected)
or "unchecked" (unselected) by the user.

A list selector consists of a database column name, a condition (logical
operation), and zero to two parameters.

The selector condition is evaluated in two realms: Perl (when displaying the
list), and SQL (when filtering the list for a list operation such as Export).

This module is used to a) determine if each record listed should be "checked"
or "unchecked", and b) construct a SQL where clause for selecting records
for a list.

If the C<table> parameter is specified in the CGI or in the arguments to
new(), then the default selector from the schema is applied if it is 
not specified in the CGI.

The arguments can be explicitly set when invoking new() by passing it values
for C<selector>, C<cond>, and C<param1>, C<param2>, and C<param3>. The
CGI parameters C<selector>, C<selector_cond>, and C<selector_param1>,
C<selector_param2>, and C<selector_param3> are queried if C<selector> is
not specified in the arguments to new().

If C<selector> is not specified in either the arguments to new() or in the
CGI parameters, but C<table> is specified in either new() or CGI, then the
selector and its parameters are obtained from the table's schema file.

The explanation() method returns a sentence or two explaining the selector
filter -- its column and its condition.

The where_clause() method returns a SQL where clause that filters the records
so that all those records in the list that were initially "unchecked" are
excluded from the query.

The join_clause() method returns a SQL join expression. Schema files for
a table can specify joins to other tables, and if so, columns in a joined
table can be used as a selector. The join_clause() method returns the
appropriate join expression for the table containing the selector column.

=head1 CHANGE LOG

 12/2/02  Initial version, from accumulated code bits.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2002  Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use CGI;
use Acme::HTML		qw(debug	debug_on);
use Acme::Base::Schema;
use Acme::Base::Schema::Join;
use Acme::Date		qw(strtodate	datetostr   strtotime	timetostr);
use Acme::Config;

our @Selector_Conditions = (
			'null',
			'not null',
			'eq',
			'ne',
			'lt',
			'le',
			'gt',
			'ge',
			'between',
			'not between',
			'in',
			'not in',
			'matches',
			'not matches',
			);

our %Selector_Desc = (
			'null'		=> 'is empty (null)',
			'not null'	=> 'is not empty',
			'eq'		=> 'is equal to',
			'ne'		=> 'does not equal',
			'lt'		=> 'is less than',
			'le'		=> 'is less than or equal to',
			'gt'		=> 'is greater than',
			'ge'		=> 'is greater than or equal to',
			'between'	=> 'is in between the values',
			'not between'	=> 'is not in between the values',
			'in'		=> 'is in the set of values',
			'not in'	=> 'is not in the set of values',
			'matches'	=> 'matches the pattern',
			'not matches'	=> 'does not match the pattern',
			);

our %Selector_Params = (
			"null"		=> 0,
			"not null"	=> 0,
			"eq"		=> 1,
			"ne"		=> 1,
			"lt"		=> 1,
			"le"		=> 1,
			"gt"		=> 1,
			"ge"		=> 1,
			"between"	=> 2,
			"not between"	=> 2,
			"in"		=> 3,
			"not in"	=> 3,
			"matches"	=> 1,
			"not matches"	=> 1,
			);

our %Perl_Ops = (	# used to compare numbers
			'eq'		=> '==',
			'ne'		=> '!=',
			'lt'		=> '<',
			'le'		=> '<=',
			'gt'		=> '>',
			'ge'		=> '>=',
			'matches'	=> '=~',
			'not matches'	=> '!~',
			);
our %MySQL_Ops = (
			'eq'		=> '=',
			'ne'		=> '!=',
			'lt'		=> '<',
			'le'		=> '<=',
			'gt'		=> '>',
			'ge'		=> '>=',
			'matches'	=> 'regexp',
			'not matches'	=> 'not regexp',
			);
our %Postgres_Ops = (
			'eq'		=> '=',
			'ne'		=> '!=',
			'lt'		=> '<',
			'le'		=> '<=',
			'gt'		=> '>',
			'ge'		=> '>=',
			'matches'	=> '~',
			'not matches'	=> '!~',
			);
our %Informix_Ops = (
			'eq'		=> '=',
			'ne'		=> '!=',
			'lt'		=> '<',
			'le'		=> '<=',
			'gt'		=> '>',
			'ge'		=> '>=',
			'matches'	=> 'matches',
			'not matches'	=> 'not matches',
			);

# Uncomment the following to turn on debug output.
#debug_on();				# turn on debug

#------------------------------------------------------------------------------

sub new {
	my $proto = shift;	# caller object (should be class/package name)
	my $class = ref($proto) || $proto;

	my $self = {
			cgi	     => undef,

			selector     => undef,
			cond         => undef,
			param1       => undef,
			param2       => undef,
			param3       => undef,
			desc         => undef,
			table        => undef,	# primary table which might be
						# the same as the selector table

			sandbox      => (caller)[0],

			@_,

			# result values
			where_clause => undef,
			join_clause  => undef,

			# table info from schema variables
			_tbl         => undef,	# derived from selector
			_col         => undef,	# derived from selector
			_is_num      => undef,	# true if selector is a number
			_is_date     => undef,	# true if selector is a date
			_is_time     => undef,	# true if selector is a time value
			_t	     => undef,	# schema's table object 
			_f	     => undef,	# schema's field object for selector
			_dbengine    => undef,	# database engine for SQL

			_check_func  => undef,  # function for testing each row
			_got_table_info => 0,	# flag
			};

	$self->{cgi} ||= new CGI $self->{args} if ref($self->{args}) eq 'HASH';
	$self->{cgi} ||= new CGI;

	# Fill in missing arguments from CGI.
	$self->{selector} ||= $self->{cgi}->param('selector')		|| '';
	$self->{cond}	  ||= lc ($self->{cgi}->param('selector_cond')	|| '');
	$self->{param1}	  ||= $self->{cgi}->param('selector_param1')	|| '';
	$self->{param2}	  ||= $self->{cgi}->param('selector_param2')	|| '';
	$self->{param3}	  ||= $self->{cgi}->param('selector_param3')	|| '';
	# Only get description from my arguments or table schema, not cgi.
	#$self->{desc}	  ||= $self->{cgi}->param('selector_desc');
	($self->{selector}) = $self->{selector} =~ /([\w.()]*)/; # untaint
	($self->{cond})     = $self->{cond}     =~ /([a-z ]*)/;	 # untaint
	# Note: untaint param's later when it is known what type they should be.
	$self->{table}	  ||= $self->{cgi}->param('primary_table')
			  ||  $self->{cgi}->param('table')
			  ||  $self->{cgi}->url_param('table')
			  ||  ($self->{selector}) =~ /(\w+)\./	# parse selector as column
			  ||  ($self->{selector}) =~ /(\w+)__/;	# parse selector as field
	$self->{selector} ||= '(none)' if ! $self->{table};

	debug("Selector: new(): self:", \$self);

	return bless $self, $class;
}

#-------------------------------------------------------------------------------

sub selector {
	my $self = shift;

	#
	# Return the selector column and possibly all the conditions.
	#

	if (! $self->{selector}) {
		$self->_get_table_info() unless $self->{_got_table_info};
	}

	if (wantarray) {
		return (
			$self->{selector},
			$self->{cond},
			$self->{param1},
			$self->{param2},
			$self->{param3},
			);
	}
	else {
		return $self->{selector};
	}
}

#-------------------------------------------------------------------------------

sub explanation {
	my $self = shift;

	if (! $self->{selector}) {
		$self->_get_table_info() unless $self->{_got_table_info};
	}

	if (! $self->{selector} || $self->{selector} eq '(all)') {
		return "All rows are initially checked.\n";
	}
	elsif ($self->{selector} eq '(none)') {
		return "A selector filter is not being used. All rows are selected.\n";
	}

	return $self->{desc} if $self->{desc};

	my $name = shift || $self->{selector};	# allow optional name for selector
	my $cond = $self->{cond};

	my $prefix = qq[Rows are initially checked if the column "$name" $Selector_Desc{$cond}];

	if ($Selector_Params{$cond} == 0) {
		return qq[$prefix.\n];
	}

	if ($Selector_Params{$cond} == 1) {
		return qq[$prefix $self->{param1}.\n]	if $self->is_number;
		return qq[$prefix "$self->{param1}".\n];
	}
	if ($Selector_Params{$cond} == 2) {
		return qq[$prefix $self->{param1} and $self->{param2}.\n]
		  if $self->is_number;
		return qq[$prefix "$self->{param1}" and "$self->{param2}".\n]
	}
	if ($Selector_Params{$cond} == 3) {
		return qq[$prefix:\n$self->{param3}]	if $self->is_number;
		return qq[$prefix:\n"$self->{param3}"];
	}
	return 'There is no explanation.';	# should not happen
}

#-------------------------------------------------------------------------------

sub is_number {
	my $self = shift;

	#
	# Return true (1) if selector column is a number column.
	#

	$self->_get_table_info() unless $self->{_got_table_info};

	return $self->{_is_num};
}

#-------------------------------------------------------------------------------

sub is_checked {
	my $self = shift;

	#
	# Return true (1) if record should be checked or unchecked.
	#

	# Note: The where clause returned by the where_clause() method below
	# must exclude any rows that this routine returns "false" for.

	return &{$self->{_check_func}} if $self->{_check_func};

	#
	# Construct function to use to check each row.
	#

	# Note: This method looks in the callers package variables for values
	# to the current record.

	$self->_get_table_info() unless $self->{_got_table_info};

	my $func = ($self->{selector}
			&& $self->{selector} ne '(none)'
			&& $self->{selector} ne '(all)')
		 ? $self->_selector_func()
		 : 'sub {return 1}';		# select all

	#
	# Compile function.
	#

	$self->{_check_func} = eval $func;

	die "Selector function did not compile.\n$@\n$func" if $@;

	#
	# Execute function.
	#

	return &{$self->{_check_func}};
}

#-------------------------------------------------------------------------------

sub join_clause {
	my $self = shift;

	$self->_get_table_info() unless $self->{_got_table_info};

	return $self->{join_clause} || '';
}

#-------------------------------------------------------------------------------

sub where_clause {
	my $self = shift;

	return $self->{where_clause} if $self->{where_clause};

	#
	# Assemble a SQL where clause.
	#

	# Note: The where clause assembled here must be "the same" as the
	# results of is_checked() above. This where clause must exclude
	# all the rows that the list would initially display as "unchecked".

	$self->_get_table_info() unless $self->{_got_table_info};

	my $selector = $self->{selector};
	my $cond = $self->{cond};

	return $self->{where_clause} = '' 		# select all
	  if ! $selector || $selector eq '(all)' || $selector eq '(none)';
		
	#
	# Assemble a SQL where clause from selector and its conditions.
	#

	# Get operator used by database engine for the condition.

	my $op	= $self->{_dbengine} =~ /Postgres/i
		? $Postgres_Ops{$cond}
		: $self->{_dbengine} =~ /MySQL/i
		? $MySQL_Ops{$cond}
		: $self->{_dbengine} =~ /Informix/i
		? $Informix_Ops{$cond}
		: $cond;


	if ($Selector_Params{$cond} == 0) {

		$self->{where_clause} = qq[$selector is $cond\n];
		return $self->{where_clause};
	}

	if ($Selector_Params{$cond} == 1) {

		if ($self->{_is_num}) {
			$self->{where_clause} = qq[$selector $op $self->{param1}\n]
		}
		elsif ($self->{_is_time}) {
			my $param = timetostr(strtotime($self->{param1}));
			$self->{where_clause} = qq[$selector $op '$param'\n];
		}
		elsif ($self->{_is_date}) {
			my $param = datetostr(strtodate($self->{param1}));
			$self->{where_clause} = qq[$selector $op '$param'\n];
		}
		else {
			$self->{where_clause} = qq[$selector $op '$self->{param1}'\n]
		}

		return $self->{where_clause};
	}

	if ($Selector_Params{$cond} == 2) {

		if ($self->{_is_num}) {
			$self->{where_clause} = qq[$selector $cond $self->{param1} and $self->{param2}\n];
		}
		elsif ($self->{_is_time}) {
			my $param1 = timetostr(strtotime($self->{param1}));
			my $param2 = timetostr(strtotime($self->{param2}));
			$self->{where_clause} = qq[$selector $cond '$param1' and '$param2'\n];
		}
		elsif ($self->{_is_date}) {
			my $param1 = datetostr(strtodate($self->{param1}));
			my $param2 = datetostr(strtodate($self->{param2}));
			$self->{where_clause} = qq[$selector $cond '$param1' and '$param2'\n];
		}
		else {
			$self->{where_clause} = qq[$selector $cond '$self->{param1}' and '$self->{param2}'\n];
		}

		return $self->{where_clause};
	}

	if ($Selector_Params{$cond} == 3) {

		my @param = split /\s*,\s*/, $self->{param3};

		return $self->{where_clause} = '' if ! @param;

		# Remove quotes
		s/^(['"])(.*)\1$/$2/ for @param;

		if (! $self->{_is_num}) {
			# Requote values 
			$_ = "'$_'" for @param;
		}

		my $param = join ", ", @param;

		$self->{where_clause} = qq[$selector $cond ($param)];

		return $self->{where_clause};
	}

	return $self->{where_clause} || '';
}

#------------------------------------------------------------------------------

sub conditions {
        # This routine can be used both as a function or a method.
	shift if ref($_[0]) || $_[0] eq __PACKAGE__;

	return @Selector_Conditions;
}

sub description {
        # This routine can be used both as a function or a method.
        shift if ref($_[0]) || $_[0] eq __PACKAGE__;

	my $cond = shift;
	return $Selector_Desc{$cond} || $cond;
}

#------------------------------------------------------------------------------
#
# Private routines
#
#------------------------------------------------------------------------------

sub _get_table_info {
	my $self = shift;

	#
	# Check for selector code words.
	# 

	if ($self->{selector}
	&& ($self->{selector} eq '(none)' || $self->{selector} eq '(all)')) {
		$self->{_got_table_info} = 1;	# no table info to be got
		debug("Selector: _get_table_info(): selector = $self->{selector}");
		return;
	}

	#
	# Get table information from schema.
	#

	my $schema = new Acme::Base::Schema;

	$self->{_dbengine} = $schema->dbengine;

	my $t = $self->{_t} = $schema->t($self->{table}) 
	  or die "Unknown table '$self->{table}'";

	if (! $self->{selector}) {
		if ($self->{selector} = $t->list_selector) {

			#
			# Get selector conditions
			#

			$self->{cond}	= $t->list_selector_cond	|| 'not null';
			$self->{desc}	= $t->list_selector_description;
			$self->{param1} = $t->list_selector_param1;
			$self->{param2} = $t->list_selector_param2;
			$self->{param3} = $t->list_selector_param3;

		}
		else {
			#
			# Select all records of a table query if not specified.
			#

			$self->{selector} = '(all)';
		}

		#
		# Check again for selector code words.
		# 

		if ($self->{selector} eq '(none)' || $self->{selector} eq '(all)') {
			$self->{_got_table_info} = 1;	# no table info to be got
			debug("Selector: _get_table_info(): table selector = $self->{selector}");
			return;
		}
	}

	#
	# Fully qualify selector column.
	#

	$self->{selector} = "$self->{table}.$self->{selector}"
	  unless $self->{selector} =~ /[.(]/;

	my ($tbl,$col) = $self->{selector} =~ /(\w+)\.(\w+)/
	or ($tbl,$col) = $self->{selector} =~ /(\w+)__(\w+)/
	or ($tbl,$col) = ($self->{table},$self->{selector});

	$self->{_tbl} = $tbl;
	$self->{_col} = $col;
	my $f = $self->{_f} = $t->f($col);

	if (! $f) {
		
		#
		# Column was not found in table. Look in joined tables.
		#

		if (my $join_clause = $t->join) {

			my $j = new Acme::Base::Schema::Join $self->{table}, $join_clause;
			my ($fld, $join_tab, $join_fld, $jt);

			for ($j->joins) {
				next unless $_->{join_table} eq $tbl;

				$fld      = $_->{field};
				$join_tab = $_->{join_table};
				$join_fld = $_->{join_field};

				if ($jt = $schema->t($join_tab)) {
					if ($f = $jt->f($col)) {
						
						#
						# Found column
						#

						$self->{_f} = $f;
						$self->{join_clause} = $join_clause;
						last;
					}
				}
			}
		}
	}

	if ($f) {
		$self->{_is_num}  = $f->is_number;
		$self->{_is_date} = $f->is_date;
		$self->{_is_time} = $f->is_time;
	}
	elsif ($col eq $t->audit_entry) {
		$self->{_is_num}  = 0;
		$self->{_is_date} = 1;
		$self->{_is_time} = 0;
	}
	elsif ($col eq $t->audit_update) {
		$self->{_is_num}  = 0;
		$self->{_is_date} = 0;
		$self->{_is_time} = 1;
	}

	$self->{_got_table_info} = 1;

	no warnings qw(uninitialized);
	debug("Selector: _get_table_info():",
			" selector = $self->{selector}",
			" cond = $self->{cond}",
			" param1 = $self->{param1}",
			" param2 = $self->{param2}",
			" param3 = $self->{param3}",
			" table = $self->{table}",
			" sandbox = $self->{sandbox}",
			" where_clause = $self->{where_clause}",
			" join_clause = $self->{join_clause}",
			" _tbl = $self->{_tbl}",
			" _col = $self->{_col}",
			" _is_num = $self->{_is_num}",
			" _is_date = $self->{_is_date}",
			" _is_time = $self->{_is_time}",
			" _dbengine = $self->{_dbengine}",
			" _got_table_info = $self->{_got_table_info}",
			);
}

sub _selector_func {
	my $self = shift;

	#
	# Return a pointer to a function that returns 1 or 0 if row
	# is to be checked (selected) or not.
	#

	# A function is built here depending upon the selector type and
	# its conditions.

	my $pkg = $self->{sandbox};

	($pkg) = $pkg =~ /([\w:]+)/s;		# untaint

	my $func = <<EOF;
sub {
	package $pkg;
	no strict;
	no warnings;
EOF

	#
	# Assemble function.
	#

	my $selector = $self->{selector};
	my $cond = $self->{cond};
	($cond) = $cond =~ /([\w ]+)/;		# untaint

	# Change Perl operator to work on number values if needed.
	my $op	= ($self->{_is_num} || $self->{_is_date} || $self->{_is_time})
		? $Perl_Ops{$cond}
		: $cond;

	# Change sql column into a perl variable name.
	my $var = $self->{selector};
	$var =~ s/\W/__/g;
	($var) =~ /(\w+)/;		# untaint
	$var = '$' . $var;

	$func .= qq[print "selector value ='$var'";\n] if $Acme::HTML::Debug;

	PARAMS: {
		if ($Selector_Params{$cond} == 0) {

			# Is null condition

			if ($cond =~ /not/) {
				$func .= qq[return (defined $var && $var ne '') ? 1 : 0;\n];
			}
			else {
				$func .= qq[return (!defined $var || $var eq '') ? 1 : 0;\n];
			}
			last;
		}

		if ($Selector_Params{$cond} == 1) {

			# Equals, Less than, Greater than ... conditions.

			die qq[The "matches" condition cannot be used with a]
			  . qq[ number, time, or date type of data.]
			  . qq[ You need to change the selector column condition.]
			  if $cond =~ /matches/
			  && ( $self->{_is_num} || $self->{_is_time} || $self->{_is_date});

			if ($self->{_is_num}) {
				my ($param) = $self->{param1} =~ m|([\deE.+-]+)|; # untaint

				$func .= qq[return $var $op $param;\n];
			}
			elsif ($self->{_is_time}) {
				my ($param) = $self->{param1} =~ m|([\w\s/.:-]+)|; # untaint

				$func = <<EOF;
use Acme::Date;
require "bigint.pl";
my \$p = Acme::Date::strtotime('$param');
$func
my \$v = Acme::Date::strtotime($var);
return main::bcmp(\$v,\$p) $op 0;
EOF
			}
			elsif ($self->{_is_date}) {
				my ($param) = $self->{param1} =~ m|([\w\s/.:-]+)|; # untaint

				$func = <<EOF;
use Acme::Date;
$func
return Acme::Date::daydiff('$param',$var) $op 0;
EOF
			}
			else {
				# Following is for text.
				my ($param) = $self->{param1} =~ /(.+)/;	# untaint

				$param	= $cond =~ /matches/
					? 'm' . _quotify($param)
					: 'q' . _quotify($param);

				$op	= $cond =~ /matches/
					? $Perl_Ops{$cond}
					: $cond;

				$func .= qq[return $var $op $param;\n];
			}
			last;
		}

		if ($Selector_Params{$cond} == 2) {

			# Between condition.
			
			my $not = $cond =~ /not/ ? qq[not ] : '';

			if ($self->{_is_num}) {
				my ($param1) = $self->{param1} =~ m|([\deE.+-]+)|; # untaint
				my ($param2) = $self->{param2} =~ m|([\deE.+-]+)|; # untaint

				$func .= qq[return $not($var >= $param1 && $var <= $param2)\n];
			}
			elsif ($self->{_is_time}) {
				my ($param1) = $self->{param1} =~ m|([\w\s/.:-]+)|; # untaint
				my ($param2) = $self->{param2} =~ m|([\w\s/.:-]+)|; # untaint

				$func = <<EOF;
use Acme::Date;
require "bigint.pl";
my \$p1 = Acme::Date::strtotime('$param1');
my \$p2 = Acme::Date::strtotime('$param2');
$func
my \$v = Acme::Date::strtotime($var);
return $not(main::bcmp(\$p1,\$v) <= 0 && main::bcmp(\$v,\$p2) <= 0);
EOF
			}
			elsif ($self->{_is_date}) {
				my ($param1) = $self->{param1} =~ m|([\w\s/.:-]+)|; # untaint
				my ($param2) = $self->{param2} =~ m|([\w\s/.:-]+)|; # untaint

				$func = <<EOF;
use Acme::Date;
$func
return $not(Acme::Date::daydiff('$param1',$var) >= 0
	 && Acme::Date::daydiff($var,'$param2') >= 0);
EOF
			}
			else {
				my ($param1) = $self->{param1} =~ /(.+)/;	# untaint
				my ($param2) = $self->{param2} =~ /(.+)/;	# untaint

				$param1 = 'q' . _quotify($param1);
				$param2 = 'q' . _quotify($param2);

				$func .= qq[return $not($var ge $param1 && $var le $param2);\n];
			}
			last;
		}

		if ($Selector_Params{$cond} == 3) {
			
			# In a set of values condition.

			my @param = split /\s*,\s*/, $self->{param3};

			if (! @param) {
				$func .= qq[return 0;];	# no parameters, uncheck all
				last;
			}

			if ($self->{_is_num}) {
				@param = map { m|([\deE.+-]+)| } @param; # untaint

			}
			elsif ($self->{_is_date}) {
				@param = map { m|([\w\s/.:-]+)| } @param;	# untaint

				# Put quotes around date values to compare as strings.
				$_ = "'" . datetostr(strtodate($_)) . "'" for @param;
			}
			elsif ($self->{_is_time}) {
				@param = map { m|([\w\s/.:-]+)| } @param;	# untaint

				# Put quotes around time values to compare as strings.
				$_ = "'" . timetostr(strtotime($_)) . "'" for @param;
			}
			else {
				@param = map { /(.*)/ } @param;		# untaint

				for (@param) {
					# Remove quotes
					s/^(['"])(.*)\1$/$2/;
					# Requote values 
					$_ = 'q' . _quotify($_);
				}
			}

			$func = "\n"
			  . q[our %_selectors = map { $_,1 } ]
			  . (join ", ", @param)
			  . ";\n"
			  . $func
			  . qq[return \$_selectors{$var} || 0;\n];

			last;
		}
		
		$func .= "return 1;";	# should not happen
	}

	$func .= '}';		 # end of function definition
	
	debug("Selector: _selector_func():",\$func);

	return $func;
}

sub _quotify {
	# Put quotes around the argument which do not appear within it.
	
	# Try paring quotes (), [], {}, and <> first.
	$_ = shift;
	return qq[($_)] unless /[)(]/;
	return qq{[$_]} unless /[\]\[]/;
	return qq[{$_}] unless /[}{]/;
	return qq{<$_>} unless /[><]/;

	# Try single characters.
	my $val = $_;
	for ( '!', '#', '%', '&', '*', '+', ',', ':', '=', '?', '^', '|', '~') {
		return qq[$_$val$_]  unless /$_/;
	}

	die "Cannot find a quote character for selector parameter '$val'."
	  . "Entry is unsafe";
}

1;
