package Acme::Options;

=head1 NAME

Acme::Options - Save and retreive user option settings

=head1 SYNOPSIS

    use Acme::Options;

    $o = new Acme::Options;
      or
    $o = new Acme::Options $user;

    $myoption = $o->get('myoption');
      or
    $myoption = $o->myoption;

    $prev_setting = $o->set('myoption' => 'new setting');

    $o->save;

    
    #
    # Circular lists are maintained by the following routines
    #

    $list = "Example";

    $o->enter($list,$item);

    @list_items = $o->list($list);

    $last_item_entered = $o->$list;

    $max_items_in_list = $o->get("${list}Max");
      which can be changed by
    $last_max = $o->set("${list}Max" => 15);


=head1 DESCRIPTION

These routines save and retreive user options for use in the Acme system.

Options are stored in a user's directory that is configured in the site
configuration file and returned by the user_folder() routine in the
I<Local::Acme::Users> module. Although the option files are private for
each user, they should be able to be read (and the path to them searchable)
by the database administration accounts.

Different option file located elsewhere can be specified by the C<file>
parameter to the new() method. Table list options are stored in such files.

No error messages are returned -- only an empty list of option settings.

Normally, you do not need to worry about saving options set in the program.
However, in Perl v5.005_03, everything works if the object variable is a
lexical (my) variable, but not for global variables. The CGI object gets
DESTROYed before this DESTROY.

Circular lists can be stored in with options. I find them handy to store
user's previous entries in a field, such as previous search strings.

=head1 CHANGE LOG

 11/13/98 Initial version.
 2/20/00  Changed to use CGI.pm. Allow for storage areas to be somewhere besides
	  the user's home directory.
 3/16/00  Merged History functions into this package.
 6/14/01  Changed get() to return an empty string if option not defined.
 4/13/02  Updated to 'use strict'.
 7/15/02  Changed new() to read/save/initialize non-user option files.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1998, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

use CGI;
use Acme::Date		qw(timestamp);
use Local::Acme::User	qw(user_folder);
use Acme::Config;

#------------------------------------------------------------------------------

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my $self = {
		file		=> '',
		user		=> '',
		save_options	=> 0,		# flag
		init_file	=> undef,	# file name of initial settings
		cgi		=> undef,
		@_,
	};

	if ($self->{file}) {
		# Untaint file name
		($self->{file}) = $self->{file} =~ /^([-\@\w. :\/\\]+)$/;
	}
	else {
		# Retreive standard user options.
		$self->{file} = _option_file_name($self->{user});
		$self->{init_file} ||= Acme::Config->Options_File;
	}

	if (open my $opt, '<', $self->{file}) {
		$self->{cgi} = new CGI $opt;	# read saved settings
		close $opt;
	}
	else {
		local our $initial_options = {};

		if ($self->{init_file}) {
			do $self->{init_file} or die "$!";
		}

		$self->{cgi} = new CGI $initial_options;
	}

	return bless $self, $class;
}

sub save {
	my $self = shift;
	$self->{save_options} = 1;	# raise flag

	# Note: Must force actual save here with perl 5.005_03. Cannot
	# trust DESTROY saving options when $self is assigned to a global.
	$self->_save_options();

	return 1;
};

sub get {
	my ($self,$opt) = @_;
	my $o = $self->{cgi}->param($opt);
	return (defined $o ? $o : '');
}

sub set {
	my ($self,$opt,$val) = @_;
	my $prev = $self->{cgi}->param($opt);
	$self->{cgi}->param($opt => $val);
	$self->{save_options} = 1;
	return $prev;
}

#------------------------------------------------------------------------------
#
#  Lists
#
#------------------------------------------------------------------------------

sub enter {
	my ($self, $list, $item) = @_;
	return unless $item;

	# Store item in in a circular list in user's options.
	# Don't store it if item is the same as the last one saved.

	$self->set($list => $item);	# save current list head as its own item

	my $max = $self->get("${list}Max");
	if ($max eq '') { $max = 10; $self->set("${list}Max" => $max) }

	my $top = $self->get("${list}Top");	# current top of list
	if ($top eq '') { $top = -1; $self->set("${list}Top" => $top) }

	if ($top < 0 || $item ne $self->get("$list$top")) {
		my $n = ($top + 1) % $max;
		$self->set("${list}$n" => $item);
		$self->set("${list}Top" => $n);
	}

	return $self->save;
}

#------------------------------------------------------------------------------

sub list {
	my ($self, $list) = @_;

	# List items in list.

	my $max = $self->get("${list}Max");
	return if $max eq '';

	my $top = $self->get("${list}Top"); # current top of list
	return if $top eq '';

	$max ||= 10;			# default
	$top ||= 0;

	my ($n,@ra);
	for ($n = $top; $n >= 0; $n--) {
		push @ra, $self->get("$list$n");
	}

	for ($n = $max; $n > $top; $n--) {
		push @ra, $self->get("$list$n");
	}

	return @ra;
}

#------------------------------------------------------------------------------
#
#  Private 
#
#------------------------------------------------------------------------------

sub _option_file_name {
	return user_folder(@_) . '/' . Acme::Config->Options_File_Name
}

sub _save_options {
	my $self = shift;

	if (open my $opt, '>', $self->{file}) {
		$self->{cgi}->save($opt);
		close($opt);
	}
	$self->{save_options} = 0;		# reset flag
}
	
sub DESTROY {
	my $self = shift;
	$self->_save_options() if $self->{save_options} && $self->{cgi};
}

sub AUTOLOAD {
        my $self = shift;
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;

	return $self->{cgi}->param($method) || '';	# get option value
}

1;
