package Acme::Parse;

=head1 NAME

    Parse - Parse routines

=head1 SYNOPSIS

    use Acme::Parse qw(parse_args parse_args_string);

    local (%args = parse_args()) or die;
	or
    local %args;
    parse_args(\%args) or die;
	or
    parse_args(\%args, $@)

=head1 DESCRIPTION

Parse_args() parses program arguments. Arguments are name=value pairs.
They can be specified in the routine arguments (@_), on the command line
(@ARGV), or they can be in the environment variable QUERY_STRING if
the environment variable REQUEST_METHOD equals GET, or they can be read
from the standard input if REQUEST_METHOD equals POST.

If a reference to a hash is passed to this routine as the first argument,
it is filled. The hash is returned to the calling program.

Parse_args() saves the query string actually used for parsing in the
global variable query_string which may be accessed explicitly by
$Acme::Parse::query_string or it can be accessed using the method
parse_args_string(). The query_string is saved so that it could be
used again intact -- even if it was read from the standard input.

Parse_args() is included here for historical purposes. It is depreciated
in favor of the now standard CGI module.

=head1 CHANGE LOG

 11/13/1998 Initial version.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1998, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.000;
require Exporter;

@ISA = qw(Exporter);
@EXPORT_OK = qw(parse_args parse_args_string);

$query_string = "";		# define global 

#------------------------------------------------------------------------------

sub parse_args {

	my $hashref;
	if (ref $_[0] eq "HASH") {
		$hashref = shift;
	}
	else {
		$hashref = {};
	}

	my (@key_value_pairs, $key_value, $key, $value);
	my $request_method;

	$query_string = "";		# clear global

	if ($#_ >= 0) {			# check for routine arguments
		$query_string = join "&", @_;
	}
	elsif ($#ARGV >= 0) {		# check for command line options
		# parse command line arguments
		$query_string = join "&", @ARGV;
	}
	elsif (defined($ENV{REQUEST_METHOD})) {		# check cgi arguments
		$request_method = "$ENV{REQUEST_METHOD}";

		if ($request_method eq "GET") {
			$query_string = $ENV{QUERY_STRING};
		} elsif ($request_method eq "POST") {
			read (STDIN, $query_string, $ENV{CONTENT_LENGTH});
		} # no error if unrecognized request method
	}

	return 0 if not $query_string;

	# Parse arguments
	@key_value_pairs = split /&/, $query_string;

	foreach $key_value (@key_value_pairs) {
		($key, $value) = split /=/, $key_value;

		# change pluses to spaces
		$key =~ tr/+/ /;
		$value =~ tr/+/ /;

		# change %nn symbols to corresponding hex characters
		$key =~ s/%([\dA-F][\dA-F])/pack("C", hex($1))/ieg;
		$value =~ s/%([\dA-F][\dA-F])/pack("C", hex($1))/ieg;

		if (defined($$hashref{$key})) {
##			# append value to existing key
##			$$hashref{$key} = join "\0", $$hashref{$key}, $value;
			# overwrite existing key
			$$hashref{$key} = $value;
		} else {
			# create new key
			$$hashref{$key} = $value;
		}
	}
	return %$hashref;	# succesful
}

sub parse_args_string {
	return $query_string;
}

1;
