package Acme::SQL;

=head1 NAME

    Acme::SQL - Parse SQL query statements

=head1 SYNOPSIS

    use Acme::SQL;

    my $sqlo = new Acme::SQL "select c1,c2 from t where c1 = 'sample'";
      -or-
    my $sqlo = Acme::SQL->parse(
    			"select c1,c2 from t where c1 = 'sample'",
			dbengine => 'Postgres'
			);

    # Information methods

    print $sqlo->columns;
    print $sqlo->tables;

    print $sqlo->dbcolumns;
    print $sqlo->dbtables;

    print $sqlo->prefix;
    print $sqlo->select;
    print $sqlo->from;
    print $sqlo->where;
    print $sqlo->group_by;
    print $sqlo->having;
    print $sqlo->order_by;
    print $sqlo->into_temp;
    print $sqlo->postfix;


    # Alteration methods

    $sqlo->add_column("col_name") if ! $sqlo->is_column("col_name");

    $sqlo->set_columns('col1','col2','col3', ...);

    $sqlo->set_into_temp('temp_table');

    $sqlo->set_distinct();

    $sqlo->set_limit(1000);

    $sqlo->add_join("table1 left join table2 on table1.col = table2.col");
      -or-
    $sqlo->add_join("table") unless $sqlo->is_table("table");

    $sqlo->add_where("c1 = 'sample'");

    $sqlo->sort("col1,col2");	# use column names
      -or-
    $sqlo->sort(1,2,3,4);	# use column position (1 is first column)


    # Result method

    print $sqlo->as_sql;

=head1 EXAMPLE

    # Combine user search script with a sql script and list the results.

    use Acme::Db;
    use Acme::SQL;
    use Acme::Searcher;
    use CGI;

    our $cgi         = new $CGI;
    my $sql          = $cgi->param('sql');
    my $sqlo         = new Acme::SQL $sql;
    my $search       = $cgi->param('search');
    my $srch         = new Acme::Searcher $search, $cgi;
    my $combined_sql = $srch->as_sql($sqlo);
    my $q            = Acme::Db->query($combined_sql);
    $q->bind_columns_to_variables();
    while ( $q->next ) {
	print "$_ = $$_\n" foreach $sqlo->columns;
    }

=head1 DESCRIPTION

Parse() parses a (possibly multi-statement) sql script, the last clause being
a SELECT clause. It parses out the field names (aliases), the table names
(aliases), the where clause, the group by clause, and the order by clause.
It produces an object suitable for input into the I<Acme::Searcher->sql>
module.

INSERT, UPDATE, DELETE, or any other non-SELECT SQL statements are not parsed.
The last statement of the SQL input must be a SELECT statement, only it
is parsed.  The other statements are considered to be part of the prefix
and are ignored. The SELECT statement can contain sub SELECTs.

The column() method returns an array of column names.  Column synonyms are
used as the column name if specified in the SQL script.

The tables() method returns an array of table names. Table synonyms are used
as the table name is specified in the SQL script.

The dbcolumns() method returns an array of actual database column names. It
is useful for checking security clearances. It can return an * (asterisk) as
a field name. Likewise, the dbtables() method returns an array of actual
database table names.

The is_column() method checks if it's argument is in the columns selected
by the sql query. The add_column() appends a column to the list of columns
selected in the query. The set_columns() method sets the columns to select.
The old list of columns is lost.

The add_where() method adds a condition to the current WHERE clause.
The add_join() method adds a join table and condition (a join expression).

The sort() method sets the C<order by> clause. The old C<order by> clause
is returned.

The as_sql() method reassembles the SQL object as a string which can be
executed by the database engine.

=head1 NOTE

The SQL keywords C<Union>, C<Intersect>, and C<Except> are recoginzed only
if they are the first word on a line. The table and column names are parsed
from just the first C<Select> clause, before the C<Union>, C<Intersect>,
or C<Except> keywords. The add_column() method does not work properly at
this time with SQL statements having any of the above keywords.

=head1 TODO

Formally define a simple proper subset of SQL92 that can be easily parsed
(for security) and extend it to transfer database information to/from text
files in a "standard" way.

SQL commands besides SELECT need to be accommodated. See Acme::Db.

Convert to a real parser.

=head1 CHANGE LOG

 11/13/98 Initial version.
  6/13/00 Added parse_sql_fields() and parse_sql_tables() to assist security.
  9/30/00 Repackaged into object/methods, changing parse_sql_fields() and
          parse_sql_tables() into dbcolumns() and dbtables() methods. Also
	  added as_sql() method.
  4/14/01 Added postfix field to object to hold remainder of compound 
          statements, such as those with UNION clauses.
  4/16/01 Added is_column() and add_column() methods to check if a column is
          listed in the query and to add a new column to the list of columns.
  10/3/01 Added add_where() method to add a search condition to the sql.
 11/26/01 Added Acme::SQL::Join object to parse SQL92 join expressions.
          Added add_join() method.
  1/17/02 Added sort() method to set the sql "order by" clause.
  8/11/02 Added set_columns() method to set the list of columns to select.
  8/19/02 Enhanced performance by reversing SQL before matching patterns.
 10/20/02 Added set_distinct() to make sql return only distinct values.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000,2001 Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

use Acme::SQL::Join;
use Acme::Config;
use Text::ParseWords;

sub new { return parse(@_) }

#
# Define reversed patterns for parse()
#

our $into_temp = qr/\s+(?:yraro)pmet\s+otni\b/i;
our $limit     = qr/\s+timil\b/i;
our $order_by  = qr/\s+yb\s+redro\b/i;
our $postfix   = qr/\s(tpecxe|tcesretni|noinu)[ \t]*\n/i;
our $having    = qr/\s+gnivah\b/i;
our $group_by  = qr/\s+yb\s+puorg\b/i;
our $where     = qr/\s+erehw\b/i;
our $from      = qr/\s+morf\b/i;
our $select    = qr/\s+tceles\b/i;

sub parse {
        my $this = shift;
        my $class = ref($this) || $this;

	my $sql = shift;

	my $self = { @_ };

	$self->{dbengine} ||= Acme::Config->DBENGINE;

	# The algorithm for parsing an SQL statement works from the end
	# to the beginning.  To speed up the parse, we reverse the sql and
	# then work from the beginning to the end to match patterns that
	# are also reversed. This doubled the performance from the old way.

	# Trim beginning semi-colons.
	$sql =~ s/^[\s;]*//;

	$sql = reverse $sql;

	# Trim trailing (now beginning) semi-colons.
	$sql =~ s/^[\s;]*//;

	my $prefix;
	if (my @s = Text::ParseWords::parse_line(';',1,$sql)) {
		$sql = shift @s;
		$prefix = reverse join ";", '', @s;
	}

	bless $self, $class;

	return $self unless $sql;

	#
	# Parse from end to beginning, which is beginning to end now that the
	# SQL statement has been reversed.
	#

	my $into_temp_clause;
	$into_temp_clause	= reverse $1 if $self->{dbengine} =~ /Informix/i
	  && $sql =~ m/(.*?)$into_temp/gcsi;

	my $limit_clause	= reverse $1 if $sql =~ m/\G(.*?)$limit/gcs;

	my $order_by_clause	= reverse $1 if $sql =~ m/\G(.*?)$order_by/gcs;

	my $postfix_clause      = reverse $1 if $sql =~ m/\G(.*)$postfix/gcs;
	my $having_clause       = reverse $1 if $sql =~ m/\G(.*?)$having/gcs;
	my $group_by_clause     = reverse $1 if $sql =~ m/\G(.*?)$group_by/gcs;

	# Make search for where clause non-greedy (greedy in reverse) to
	# accommodate sub-selects.
	my $where_clause        = reverse $1 if $sql =~ m/\G(.*)$where/gcs;

	my $from_clause         = reverse $1 if $sql =~ m/\G(.*?)$from/gcs;

	$into_temp_clause	= reverse $1 if $self->{dbengine} =~ /Postgres/i
	  && $sql =~ m/(.*)$into_temp/gcs;

	my $select_clause       = reverse $1 if $sql =~ m/\G(.*?)$select/gcs;

	$self->{prefix}		= $prefix		|| '';
	$self->{select}		= $select_clause	|| '';
	$self->{from}		= $from_clause		|| '';
	$self->{where}		= $where_clause		|| '';
	$self->{group_by}	= $group_by_clause	|| '';
	$self->{having}		= $having_clause	|| '';
	$self->{postfix}	= $postfix_clause	|| '';
	$self->{order_by}	= $order_by_clause	|| '';
	$self->{limit}		= $limit_clause		|| '';
	$self->{into_temp}	= $into_temp_clause	|| '';

	$self->{join}           = Acme::SQL::Join->new($from_clause);

	# Note: we delay parsing the select clause for a "distinct" keyword
	# until needed. This retains any SQL comments for the log files if
	# the parsing of the select clause is not needed.

	$self->{distinct}	= '';	# delay parsing select clause for this
	$self->{is_select_parsed} = 0;	# flag to delay parse

	return $self
}

# Delegate to Acme::SQL::Join object.
sub tables        { return shift->{join}->tables;        }
sub dbtables      { return shift->{join}->dbtables;      }
sub realtable     { return shift->{join}->realtable(@_); }
sub primary_table { return shift->{join}->primary_table; }
sub joins         { return shift->{join}->joins;         }
sub add_join      { return shift->{join}->add_join(@_);  }
sub is_table      { return shift->{join}->is_table(@_);  }

sub columns {
	my $self = shift;

	#
	# Extract field names or synonyms from query script.
	#

	my @f;			# returned list of parsed fields

	$self->_parse_select() unless $self->{is_select_parsed};

	# Define regular expression to match matching parentheses (even if nested).
	# From the Camel book, 3rd edition, p214.

	our $matched_paren = qr{
				  \(
				  (?:
				    (?> [^()]+ )
				  |
				    (??{ $matched_paren })
				  )*
				  \)
				}x;

	#
	# Split fields on comma's. Watch for functions, recombine comma separated
	# arguments if found.
	#

	my $prefix = '';
	for (Text::ParseWords::parse_line(',', 1, $self->{select})) {
		s/^\s+//;  # remove possible white space at beginning
		s/\s+$//;  # remove possible white space at end

		$prefix .= $_;
		# Combine with next value if we're within a function call ().
		if ($prefix =~ /\(/ && $prefix !~ /$matched_paren/s) {
			$prefix .= ',';		# put comma delimiter back
			next;
		}
		$_ = "$prefix";
		$prefix = '';

		# Remove synonym if specified.
		# Look for keyword 'as'.
 		s/^.+\s+as\s+([^)\s]+)$/$1/si;

		# Remove Informix style synonym.
		# Space between field name and synonym.
 		s/[^-+*\/]+\s+([^)\s]+)$/$1/s;

		push @f, $_;
	}

	return @f;
}

sub dbcolumns {
	my $self = shift;

	#
	# Extract actual column names from query script.
	#

	my @f;			# returned list of parsed fields

	$self->_parse_select() unless $self->{is_select_parsed};

	# Split fields on comma's
	for my $expr (Text::ParseWords::parse_line(',', 1, $self->{select})) {
		$expr =~ s/^\s+//;  # remove possible white space at beginning
		$expr =~ s/\s+$//;  # remove possible white space at end

		# Remove synonym if specified.
		# Look for keyword 'as'.
 		$expr =~ s/^(.+)\s+as\s+[^)\s]+$/$1/si;

		# Remove Informix style synonym.
		# Space between field name and synonym.
 		$expr =~ s/([^-+*\/]+?)\s+[^)\s]+$/$1/s;

		# Remove keywords for literals date, time, timestamp, interval
		next if $expr =~ /^(date|interval|time(stamp)?) /i;

		# Ignore quoted strings.
		next if $expr =~ /^(['"]).*\1\s*$/s;

		if ($expr =~ /^((\w+)\.)?\*$/) {
			# * star field
			push @f, $expr;
			next;
		}

		# Split on operators || + - * and /
		for (split(/(?:\|\|)|\+|-|\*|\//,$expr)) {
			# Look into function calls

			s/^\s+//;  # remove possible white space at beginning
			s/\s+$//;  # remove possible white space at end

			# Note: we should parse each function call,
			# but look only for SQL type functions and
			# just check the first argument for the rest.
			if (/.*\((.*)\)/) {
				# Check for function args like: a FROM col
				my $a = $1;
				if ($a =~ s/.*from\s+//i) {
					push @f, $a;
					next;
				}
				# Check for function args like: a IN col
				if ($a =~ s/.*in\s+//i) {
					push @f, $a;
					next;
				}
				# Check for function args like: a AS col
				if ($a =~ s/.*as\s+//i) {
					push @f, $a;
					next;
				}

				# Check only first argument in function call.
				$a =~ s/\s*,.*//;
				push @f, $a;
				next;
			}

			# Remove Postgres casts.
			s/::\w+//;

			# Remove keywords for literals date, time, timestamp, interval
			next if $expr =~ /^(date|interval|time(stamp)?) /i;

			# Ignore quoted strings.
			next if /^(['"]).*\1$/s;

			# Ignore numbers.
			next if /^[\d.+-e]+$/;

			# Ignore keyword null.
			next if /^null$/i;

			push @f, $_;
		}
	}
	return @f;
}

sub is_column {
	my ($self, $col) = @_;

	$self->_parse_select() unless $self->{is_select_parsed};

	# Return true if $col is a column in the query.
	return $self->{select} =~ /\b$col\b/;
}

sub add_column {
	my $self = shift;

	# Note: This method does not work properly with SQL statements
	# having the keywords Union, Intersect, or Except.

	$self->_parse_select() unless $self->{is_select_parsed};
	$self->{select} .= ", $_" foreach @_;
}

sub set_columns {
	my $self = shift;
	return if ! @_;

	$self->_parse_select() unless $self->{is_select_parsed};
	$self->{select} = join ',', @_;
}

sub set_into_temp {
	my $self = shift;
	return if ! @_;

	$self->{into_temp} = shift;
}

sub set_distinct {
	my $self = shift;
	$self->_parse_select() unless $self->{is_select_parsed};
	$self->{distinct} = 'distinct ';
}

sub set_limit {
	my $self = shift;
	return if ! @_;

	$self->{limit} = shift;
}

sub add_where {
	my $self = shift;

	my $where = $self->{where};
	for (@_) {
		$where &&= "( $where )";
		$where &&= "$where and "
		  unless /^\s*(and|or)\b/i;
		$where .= $_;
	}
	$self->{where} = $where;
}

sub sort {
	my $self = shift;

	my $old_order_by = $self->{order_by};

	$self->{order_by} = join ',', @_ if @_;

	return $old_order_by;
}

sub as_sql {
	my $self = shift;

	if ($self->{postfix} && $self->{order_by}) {

		#
		# Translate "order by" clause into column numbers
		# to pass SQL syntax when a "union", "intersect", or
		# "except" statement is used.
		#

		my $i = 0;
		my %ncol = map { $_, $i++ } $self->dbcolumns;
		my @cols = map { $ncol{$_} || $_ } split /\s*,\s*/, $self->{order_by};
		$self->{order_by} = join q[, ], @cols;
	}

	my $nl = "\n";
	my $s = '';
	$s .= $self->{prefix}				if $self->{prefix};
	$s .= "${nl}select "     . $self->{distinct} . $self->{select};
	$s .= "${nl}into temp "  . $self->{into_temp}	if $self->{into_temp}
	  && $self->{dbengine} =~ /Postgres/i;
	$s .= "${nl}from " . (
	   $self->{join}->{is_altered}
	   ? $self->{join}->as_string	# construct new clause
	   : $self->{from}		# use original clause
	   );
	$s .= "${nl}where "      . $self->{where}	if $self->{where};
	$s .= "${nl}group by "   . $self->{group_by}	if $self->{group_by};
	$s .= "${nl}having "     . $self->{having}	if $self->{having};
	$s .= $self->{postfix}				if $self->{postfix};
	$s .= "${nl}order by "   . $self->{order_by}	if $self->{order_by};
	$s .= "${nl}limit "      . $self->{limit}	if $self->{limit}
	  && ($self->{dbengine} =~ /Postgres/i || $self->{dbengine} =~ /MySQL/i);
	$s .= "${nl}into temp "  . $self->{into_temp}	if $self->{into_temp}
	  && $self->{dbengine} =~ /Informix/i;
	return $s;
}

#-------------------------------------------------------------------------------
#
# Private 
#
#-------------------------------------------------------------------------------

sub _parse_select {
	my $self = shift;

	$self->{select} = _strip_comments($self->{select});

	# Ensure 'distinct' is set if in the select clause.
	$self->set_distinct()
	  if $self->{select} =~ s/^\s*unique\b//si
	  || $self->{select} =~ s/^\s*distinct\b//si;

	$self->{is_select_parsed} = 1;	# set flag 
}

sub _strip_comments {
	local $_ = shift;	# string to strip of comments
	s|/\*.*?\*/||sg;	# remove comments between /* and */
	s|--.*$||mg; 	  	# remove comments between -- and end of line
	return $_;
}

#-------------------------------------------------------------------------------

sub AUTOLOAD {
        my $self = shift;
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;
	return if $method eq 'DESTROY';		# gobble

	# Return object parameter if same name as method name.
        return $self->{$method} if exists $self->{$method};
        return;           	# unknown parameter/method
}

1;
