package Acme::SQL::Join;

=head1 NAME

Acme::SQL::Join - Parse SQL join expressions

=head1 SYNOPSIS

    use Acme::SQL::Join;

    my $j = new Acme::SQL::Join $sqlo->{from};
      -or-
    my $join_clause = "customers left join customer_info using (customer_id)";
    my $j = Acme::SQL::Join->parse($join_clause);

    print $j->tables;
    print $j->dbtables;

    $j->add_join("table1 left join table2 on table1.col = table2.col");
      -or-
    $j->add_join("table") unless $j->is_table("table");

    print $j->as_string;
      -or-
    print "select * from ", $j->as_string, " where id = 2;";
      -or-
    print "select * from ", $j->join_clause, " where id = 2", $j->join_where;

=head1 DESCRIPTION

This object parses a SQL92 C<From> clause into an array of join expressions.

The tables() method returns an array of table names. Table synonyms are used
as the table name if specified in the SQL script.

The dbtables() method returns an array of actual database table names.
A table might be listed more than once if a table is part of a self join.
The elements returned by tables() are unique. Each element of the list returned
by dbtables() has a corresponding element in the list returned by tables().

The add_join() method adds a join expression to the array of joins.

The join_clause() and the join_where() methods are used for older database
engines that separate the join tables from the join conditions. Currently,
join_clause() is a synonym for the as_string() method and join_where() is a
stub returning an empty string. However, they are used in the Acme system
to accommodate those database engines which don't use SQL92 join expressions
in the future.

The as_string() method reassembles the Join object as a string which can be
used as a C<From> clause in an SQL statement.

This module is very simplistic. It does not try to track the kind of joins, just
which tables are used and what their alias names are (if any) so that 
correct variable names (table__field) can be constructed by the system.

At least that was the intention. The SQL92 standard for join expressions, and
especially Postgres's implementation of the standard for join expressions, is
unfriendly to pattern matching code. It is incredibly bad at allowing
a program to "add" a join table if the join is to be an outer join -- the most
common type of join in my "real world".

To accommodate join expressions, the join clause is parsed into parallel arrays
of terms. The terms of the join expressions described as:

    ( left-table join-type right-table ON conditional-expression ) AS range-variable
      -or-
    left-table join-type right-table AS range_var USING ( column-list )
      -or like Postgres -
    left-table join-type right-table AS range_var ON conditional-expression

Each element of the array of joins is "atomic", able to stand alone as
the sole join expression in a SQL From clause. Aliases (range-variables)
are separated into corresponding arrays. The as_string() method
re-combines the array of join expressions and the array of aliases back into
a string that can be used as a (Postgres) SQL C<From> clause that looks 
something like:

	((table left join other_table on ...) left join other table on ...)

This allows the add_join() to be able to add a join table/condition to the
SQL statement.

Debug output and a dump of the object can be obtained by setting the $Debug
package variable to 1.

=head1 CHANGE LOG

 11/26/01 Initial version.
 11/27/02 Added is_table().
 12/05/02 Added rudimentary consolidation of redundant joins in as_string().

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000,2001 Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

# Sometimes you need to output the HTTP header for debug info.
#print "Content-type: text/html\n\n";

our $Debug = 0;
sub debug { print "<br> --- ", @_, "\n" if $Debug }
sub debug_dump {
	return unless $Debug;
	require Data::Dumper;
	print "<pre>",Data::Dumper::Dumper(@_),"</pre>";
}
sub debug_dump_callers {
	return unless $Debug;
	my $s = '';
	my $i = 1;
	print "<br>Called by:";
	while ( my ($package, $filename, $line, $subroutine) = caller($i++) ) {
		$s .= "\n<br>$subroutine at line $line";
	}
	print $s;
}

sub new {
        my $this = shift;
        my $class = ref($this) || $this;

	my $self = {
		joins           => [],	# array of (original) join strings
		ltables         => [],	# array of joins' left tables (lvalues)
		rtables         => [],	# array of joins' right tables
		alias           => [],	# array of join expression aliases
		join_conds      => [],	# array of join conditions (on, using)
		join_types      => [],	# array of join types (left join, ...)
		tables          => {},	# hash of table names and counts
		dbtables        => {},	# hash of db table names and counts
		realtable       => {},	# hash of table names for every alias
		primary_table   => '',	# first left table found in all clauses
		curr_left_table => '',	# left table of join, working variable
		is_altered      => 0,	# true if join expression altered
		};

	bless $self, $class;

	return $self->parse(@_);
}

sub parse {
	my $self = shift;

        return new($self,@_) unless ref($self);

	my $join_clause = shift;

	return $self unless $join_clause;

	# Build arrays of the join clauses, the tables referenced, and the
	# table aliases used.

	$join_clause =~ s|/\*.*?\*/||sg;# remove comments between /* and */
	$join_clause =~ s|--.*$||mg;	# remove comments between -- and end of line

	my $alias;

	$self->{curr_left_table} = '';

	debug("Beginning to parse SQL From clause: $join_clause");
	debug_dump_callers();

	for my $expr (Text::ParseWords::parse_line(',', 1, $join_clause)) {
		$expr =~s/^\s+//s;	# remove possible white space at beginning
		$expr =~s/\s+$//s;	# remove possible white space at end

		debug("Starting to parse: $expr");

		$self->parse_expr($expr);
	}

	debug_dump($self) if $Debug;

	return $self;
}

# Define regular expression to match matching parentheses (even if nested).
# From the Camel book, 3rd edition, p214.

our $matched_paren = qr{
			  \(
			  (?:
			    (?> [^()]+ )
			  |
			    (??{ $matched_paren })
			  )*
			  \)
			}x;

sub parse_expr {
	my $self = shift;
	my ($expr) = @_;

	#
	# Recursive routine to parse individual join expressions like:
	#
	#   ((cust left join cust_info on ...) left join cust_contact on ...)
	#
	# or table expressions like:
	#
	#   customers, trips, reservations
	#
	# The "left table" of a join expression cannot be determined until the
	# lowest level of recursion and cannot be assigned until we are popping up.
	#
	# The "left table" of a table expression is the first table in a list
	# of tables, so it is implied.
	#

	my ($left_table, $right_table);
	my $alias = '';

	if ($expr =~ s/\s+[aA][sS]\s+(\w+)\s*$//) {
		# Get alias at end of expr
		$alias = $1;
	}

	$expr =~ s/^\(\s*(.*)\s*\)$/$1/s;	# strip off enclosing ()

	if ($expr =~ s/^($matched_paren)//s) {
		# Recurse
		debug("Recurse: parse_expr: $1");

		$self->parse_expr($1);

		debug("Pop, continuing to parse.");

		# Recall saved $left_table.

		$left_table = $self->{curr_left_table};
	}
	else {
		# Strip off first word as the left table.
		$expr =~ s/(\w+)\s*//;
		$left_table = $1;

		# Save $left_table so it is available as we "pop" up
		# from recursive call.

		$self->{curr_left_table} = $left_table;

		# First "left" table is the implied left table for all
		# other join expressions if they do not specify a left
		# table -- like an expression consisting of just a 
		# table name of an implied inner join.

		# Track left table name unless it has an alias
		if (! $self->{primary_table}) {
			debug("assigning primary table='$left_table', alias='$alias'");

			$self->{primary_table} = $left_table;

			my $range_var = $alias || $left_table;
			$self->{tables}->{$range_var} += 1;

			# Keep hash of table/alias cross reference
			$self->{realtable}->{$range_var} = $left_table;
		}
	}

	my ($join_expr,$join_type,$join_cond);

	if ($expr) {
		# If here, we are parsing a "join-expression".

		debug("join-expression: '$expr'");

		if ($expr =~ s/(\w+)\s+[aA][sS]\s+(\w+)(\s*)/$1$3/) {
			# Get alias (can be in middle of expr)
			$right_table = $1;
			$alias = $2;
		}

		($join_type,$join_cond) = $expr =~ /(.*)\s+(on\s+.*)/si    or
		($join_type,$join_cond) = $expr =~ /(.*)\s+(using\s+.*)/si or
		($join_type,$join_cond) = ($expr,'');

		# Last word is table, rest is join type
		$join_type =~ s/\s*(.*)\s+(\w+)\s*$/$1/s;
		$right_table ||= $2;
	}
	else {
		# If here, we are parsing a "table-expression".

		$right_table = $left_table;
		$alias     ||= $right_table;
		$left_table  = $self->{primary_table};	# re-assign
		$join_type   = 'inner join';
		$join_cond   = '';

		debug("table-expression:"
		  . " left_table='$self->{primary_table}',"
		  . " right_table=$right_table,"
		  . " alias='$alias'");
	}

	$alias       ||= $right_table;

	$join_expr  = "$left_table $join_type $right_table";
	$join_expr .= " as $alias"  if $alias && $alias ne $right_table;
	$join_expr .= " $join_cond" if $join_cond;

	# Keep hash of table/alias cross reference
	$self->{realtable}->{$alias} = $right_table;

	push @{$self->{joins}},      $join_expr;
	push @{$self->{ltables}},    $left_table;
	push @{$self->{rtables}},    $right_table;
	push @{$self->{alias}},      $alias;
	push @{$self->{join_types}}, $join_type;
	push @{$self->{join_conds}}, $join_cond;

	no warnings qw(uninitialized);

	$self->{tables}->{$alias}         += 1;
	$self->{dbtables}->{$right_table} += 1;
	$self->{dbtables}->{$left_table}  += 1;

	debug("assigning:"
	  . "\n<br> join_expr='$join_expr'"
	  . "\n<br> left_table='$left_table'"
	  . "\n<br> right_table='$right_table'"
	  . "\n<br> alias='$alias'"
	  . "\n<br> join_type='$join_type'"
	  . "\n<br> join_cond='$join_cond'"
	  ) if $Debug;
}

sub is_table {
	my $self = shift;
	my ($tbl) = @_;
	return $self->{tables}->{$tbl} ? 1 : 0;
}

sub tables {
	my $self = shift;
	return keys %{$self->{tables}};
}

sub dbtables {
	my $self = shift;
	return keys %{$self->{dbtables}};
}

sub realtable {
	my $self = shift;
	my ($alias) = @_;
	return $self->{realtable}->{$alias};
}

sub primary_table {
	my $self = shift;
	return $self->{primary_table};
}

sub joins {
	my $self = shift;
	return @{$self->{joins}};
}

sub add_join {
	my $self = shift;
	$self->{is_altered} = 1;	# set flag
	return $self->parse(@_);
}

sub as_string {
	my $self = shift;

	# Search through the array of join expressions and combine all those
	# that join the same left table. In other words, group all similar joins
	# together.

	# Define return string (a complete From clause w/o 'From' keyword).
	my $s = '';			# returned string of join expressions.

	my ($lvalue, $rvalue);		# left and right parts of $s expressions

	my $joins   = $self->{joins};		# array ref
	my $conds   = $self->{join_conds};	# array ref
	my $alias   = $self->{alias};		# array ref
	my $types   = $self->{join_types};	# array ref
	my $ltables = $self->{ltables};		# array ref
	my $rtables = $self->{rtables};		# array ref

	my $is_last_occurrance_of_table = 0;	# little flag
	my %done;				# more flags
	my %table_done;				# more flags

	my $join_expr_table;
	my ($expr, $left_table, $right_table, $synonym);

	#
	# Mark any redundant joins as already done. (use brute force)
	#

	for (my $i = 0; $i < @$joins; $i++) {
		next if $done{$i};
		for (my $j = $i+1; $j < @$joins; $j++) {
			next if $done{$j};

			if ($$rtables[$i] eq $$rtables[$j]
			&&  $$ltables[$i] eq $$ltables[$j]) {
				
				# Two join conditions are on the same tables.

				# Try to be smart about which join condition
				# to mark as done. Inner joins have precedence,
				# followed by any other join type, followed by
				# a table join (where the join condition is in
				# the where clause).

				if ($$types[$i] =~ /inner/i) {
					# Inner join takes precedence.
					$done{$j} = 1;
				}
				elsif ($$types[$j] =~ /inner/i) {
					# Inner join takes precedence.
					$done{$i} = 1;
				}
				elsif (! $$conds[$i]) {
					# Table expression
					$done{$i} = 1;
				}
				elsif (! $$conds[$j]) {
					# Table expression
					$done{$j} = 1;
				}
				elsif (lc($$types[$i]) eq lc($$types[$j])) {
					# Both join types the same.
					# Do nothing, let an error be triggered.
				}

				# Note: other combinations slip through and
				# will cause errors with the database engine,
				# as it should.

				last;
			}
		}

	}

	#
	# Assemble join-expressions into From clause
	#

	for (my $i = 0; $i < @$joins; $i++) {
		next if $done{$i};

		($lvalue, $rvalue) = ('','');

		$join_expr_table = $$ltables[$i];

		for (my $j = $i; $j < @$joins; $j++) {
			next if $done{$j};
			next if ! $$conds[$j];	# not a join-expression

			$left_table  = $$ltables[$j];
			$right_table = $$rtables[$j];

			next if $left_table ne $join_expr_table;

			$expr    = $$joins[$j];
			$synonym = $$alias[$j];
			$synonym = '' if $synonym eq $self->{realtable}->{$synonym};

			$expr =~ s/\w+\s+//; # remove left table

			$lvalue .= "(";
			$rvalue .= " $expr)\n";

			$is_last_occurrance_of_table = 1; # initialize to true
			for (my $k = $j+1; $k < @$ltables; $k++) {
				next if $done{$k};
				if ($left_table eq $$ltables[$k]) {
					$is_last_occurrance_of_table = 0; # false
					debug("join->as_string(),"
					  . " NOT is_last_occurance_of_table '$left_table'")
					  if $Debug;
					last;
				}
			}

			if ($is_last_occurrance_of_table) {
				# now put $left_table in expression.

				$lvalue .= "$left_table\n";

				$table_done{$left_table} = 1;
			}

			$done{$j} = 1;
		}

		$s .= ', ' if $s;
		$s .= "$lvalue$rvalue";
	}

	#
	# Prepend table-expressions to From clause
	#

	my $table_expr = '';		# table expressions prepended to $s
	for (my $i = 0; $i < @$joins; $i++) {
		next if $done{$i};
		next if $$conds[$i];	# not a table-expression

		$left_table  = $$ltables[$i];
		$right_table  = $$rtables[$i];

		next if $table_done{$left_table};

		$table_expr .= ", " if $table_expr;
		$table_expr .= $right_table;
		$table_expr .= " as $synonym" if $synonym;
	}

	if ($table_expr) {
		$s = ", $s" if $s;
		$s = "$table_expr$s";
	}

	debug("join->as_string(), returning: <pre>$s</pre>") if $Debug;

	return $s;
}

sub join_clause { return as_string(@_) }

sub join_where { return '' }

1;
