package Acme::Searcher;

=head1 NAME

    Search - Assemble SQL script for searches

=head1 SYNOPSIS

    use Acme::Searcher;
    use Acme::Db;

    my $script = Acme::Searcher::preprocess($user_input, {
	RecordSetName	   => "Customers",
	IDField		   => "customer_id",
	IDFieldName	   => "ID",
	CodeField	   => "user",
	CodeFieldName	   => "User",
	NameField	   => "name",
	NameFieldName      => "Name",
	PhoneField	   => "phone",
	PhoneFieldName     => "Phone",
	EmailField	   => "email",
	EmailFieldName     => "Email",
	FirstNameField	   => "first_name",
	FirstNameFieldName => "First Name",
	LastNameField	   => "last_name",
	LastNameFieldName  => "Last Name",
	});


    my $srch = new Acme::Searcher $script, \%args;
     -or-
    my $srch = Acme::Searcher->parse($script, \%args)
      or die "Error evaluating search.";

    if ($srch->uses_table("trips")) {
	$extra_columns = ", trips.depart_date";
    }

    my $sql = $srch->as_sql( {
	title	  => "A Sample Search",
	select	  => "customers.customer_id, customers.name $extra_columns",
	from	  => "customers",
	where	  => "entry_date = today",
	order_by  => "name",
	into_temp => "table_name",
	dbengine  => "Postgres",
    } );

    my $q = Acme::Db->query($sql);
	...

    print $srch->title;
    print $srch->annotate;

=head1 EXAMPLE

    #
    # Combine user search script with a SQL script and list the results.
    #

    use Acme::Db;
    use Acme::SQL;
    use Acme::Searcher;
    use CGI;

    my $cgi          = new CGI;
    my $sql          = $cgi->param('sql');
    my $search       = $cgi->param('search');
    my $sqlo         = new Acme::SQL $sql;
    my $srch         = new Acme::Searcher $search, $cgi;
    my $combined_sql = $srch->as_sql($sqlo);
    my $q            = Acme::Db->query($combined_sql);
    $q->bind_columns_to_variables();
    while ( $q->next ) {
	foreach $c ($sqlo->columns) { print "$c = $$c\n" }
    }

=head1 DESCRIPTION

The Search module assembles a SQL script based on parameters
supplied by the programmer with parameters supplied by the user.

A search is constructed with a script as defined in
.../html/help/search_instructions.htm.

=head2 The Preprocessor

The preprocess() module takes user input and constructs a search script
consisting of a search title, and a SQL C<where> clause fragment.
It takes several options, most of which are the names of the fields to
search and their titles. The field titles are used to construct search
titles, like `Customer ID matching 12345', with `Customer ID' being the
field title.

If the user's input looks like an ID number and the IDField option is
specified, then a SQL C<where> clause is constructed searching the IDField.
A search string looks like an ID number if it is all digits -- a number,
or if it is a comma (or plus) separated list of numbers.

If the user's input looks like a name and the NameField option is
specified, then a SQL C<where> clause is constructed searching the field(s)
specified in NameField. A search string from the user looks like a name
search if if has both capital letters and lower case letters, or if it
has magic characters, or if the entire search string is enclosed in
quotes -- either single or double ('").

The FirstNameField and LastNameField options can specify two name fields
to search if the user's input has a comma in it, in which case the pattern
before the comma is to match the field named by LastNameField, and the
pattern after the comma is to match the FirstNameField field.

If the user's input looks like a code and the CodeField option is
specified, then a SQL C<where> clause is constructed searching the CodeField
field(s). A search string looks like a code if it is all lowercase letters or
all upper case letters, possibly with number, but without magic characters.

If the user's input looks like a phone number and the PhoneField option is
specified, then a SQL C<where> clause is constructed searching the PhoneField
field(s). A search string looks like a phone number if it consists digits with
dashes and/or parenthesis.

The PhoneMatch option can be set to 'begin', "equal', or 'end' (the default)
to anchor phone number matches to the beginning of the field, exactly equal
to the field contents, or matches the end of the field.  Anchoring the phone
number search to the end of the field allows the user to search for phone
numbers without the area codes or international access codes, but it is the
slowest as it cannot make use of database indexes.

If the user's input looks like an email address and the EmailField option is
specified, then a SQL C<where> clause is constructed searching the EmailField
field(s). A search string looks like an email if it has the @ character in 
the middle of it. Magic characters can be used to search for patterns.

Search patterns cannot contain magic characters if they are to be recognized
for searching the PhoneField or CodeField fields. However, there are two
options PhoneFieldMatch and CodeFieldMatch that can be set to one of the
constants: 'begin', 'end', or 'equal'. They allow the search to match the
field's beginning, the field's ending, whether the search must match the
field exactly. For performance reasons, it is best to set these options to 
'begin' or 'equal'. The default value for CodeFieldMatch is 'begin'. The
default value for PhoneFieldMatch is 'end', allowing phone numbers with out
area codes or international access codes to match.

Titles are also made for the lists depending upon the fields searched.

=head2 Search Constructor

A I<Acme::Searcher> object can be constructed with the new() or parse()
methods.  The search script is passed along with a hash of variable
name/values to substitute into the script variables. Script variables are
like Perl scalar variables -- they begin with a $. The argument hash can be
a CGI object, and usually is.

A third argument to new() or parse() methods is a flag which if set to
non-zero, prints out a CGI header ('Content-type: text/html') before trying
to display a parameter entry form. Normally, the CGI header is assumed to be
output before evaluating the search script, so displaying a parameter entry for
script parameters (e.g. $name=?t20?Mack?Enter Name) does not need a CGI header
to be output first. However, the C<pick.cgi> program issues a CGI redirect
command when searches produce only one record, so the output of the normal
CGI header is delayed until after the search is evaluated. In that case,
the CGI header has to be output first before displaying a parameter entry
form to the user.  That is what the third argument C<do_http_header> is for.

=head2 Search Parameters

If parameters are needed from the user (as determined by search script)
a form will be displayed to the user. The form will recall the program with
all the original parameters and the new parameters entered by the user.

=head2 SQL Method

Finally, the SQL script is assembled using the sql() method. It takes it's
parameters of sql clauses and assembles them with the sql clauses from
from the search script and produces a SQL query script. The SQL script
can then be executed using the Acme::Db::query() method.

=head2 List Files

A search script can join the sql query with list files. Lists are files
which contain one field value per line, as would be produced by the sql
"unload to file_name select mykey from mytable". Lists can be specified in the
search script by identifying a file name and a database field name. Multiple
lists can join the database field.  The join is an "inner join". A temporary
table of unique values is created for the list so the query can join it. The
name assigned to the temporary table is constructed from the field name by
prepending "zz" to it. The temp table field name is the same as the database
field name.

=head2 Informative Methods

The program can retrieve and inspect parts of the search -- the
title, the C<where> clause, and a C<join> clause if the script joined other
tables or lists. The title(), where_clause(), and join_clause() methods
provide access to that information.

The uses_table() method allows the program to determine if extra columns
or a different order clause should be used if the search script joined
to other tables.

Debug output and a dump of the object can be obtained by setting the
$Acme::Searcher::Debug variable to 1.

=head1 CHANGE LOG

 5/97     Initial version.
 11/20/98 Assembled routines into this library.
 3/16/00  Allow CGI objects to be passed directly to parse().
 6/11/00  Added code_field search field if user input is all one case.
 3/10/01  Added Safe processing.
 11/25/01 Added EmailField and PhoneField search field parameters.
          Made NameField, PhoneField, and EmailField accept a comma delimited
	  list of search fields.
 1/11/02  Restored search parameters, convert to Postgres (SQL92 subset) join
          expressions, added list_exclude() search function.
 9/29/02  Fixed subtle parsing bug by putting \' in constructed titles.
 11/26/02 Changed list files to be able to join non-integer keys.
 12/05/02 Parameter screens look in cgi for "instructions" parameter.
          Added $annotate, which like $title, can be set in script and
	  later retreived using $srch->annotate.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1998-2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

use Acme::HTML;
use Acme::HTML::Parameters;	# to get script parameters from user
use Acme::SQL::Join;		# to add join clauses together
use Acme::Date qw(today dayadd daydiff);# reference for search scripts
use Local::Acme::User qw(user);	# user name
use Acme::Config;
use Text::ParseWords;		# to split scripts into statements
use Carp;			# for error messages
use Safe;			# eval's used for variable substitutions, ...
use CGI;

our $SS = Acme::Config->Script_Separator || ';';	# either ; or |
our $SSP = '\s*[|;]\s*';	# search script pattern used to parse scripts
				# allow use of old | and new ; separator

our $DS = '/';			# directory name separator (/ \ or :)

our $Debug = 0;
sub debug { print "<br> --- ", @_, "\n" if $Debug }
#print "Content type: text/html\n\n";

#------------------------------------------------------------------------------
#
# Search script pre-processor functions
#
#------------------------------------------------------------------------------

sub preprocess {
	# This routine can be used both as a function or a method.
	shift if ref($_[0]) || $_[0] eq __PACKAGE__;

	my $user_input = shift;		# search entered by user
	my $opt = shift;		# options hash reference

	my $search = $user_input;	# returned search script
	my $title  = '';

	our $Db_Engine = $opt->{dbengine} || Acme::Config->DBENGINE;
	our $Postgres = $Db_Engine =~ /Postgres/i;
	our $MySQL    = $Db_Engine =~ /MySQL/i;
	our $Informix = $Db_Engine =~ /Informix/i;

	our $Match_Op = $Postgres
		      ? '~'
		      : $MySQL
		      ? 'regexp'
		      : $Informix
		      ? 'matches'
		      : '~';

	# The following fields are defined for the options hash $opt.
	# The IDField (if the search parameter is numeric) or the
	# NameField (if the search is pattern matching) must be present.
	# FirstNameField and FirstNameFieldName are optional synonyms for
	# NameField and NameFieldName, suitable for when defining
	# LastNameField and LastNameFieldName.
	# The CodeField, PhoneField, and EmailField options have been
	# added. They and IDField and NameField can be lists of
	# fields to search.
	#
	#	RecordSetName	   -- name of record set (capitalized)
	#			      (used for titles)
	#	IDField		   -- database name of numeric ID field
	#	IDFieldName	   -- name of ID field (capitalized)
	#
	#	CodeField	   -- database name of code field
	#	CodeFieldName	   -- name of code field (capitalized)
	#    
	#	NameField	   -- database name of full name field
	#	NameFieldName	   -- name of full name field (capitalized)
	#    
	#	PhoneField	   -- database name of phone field
	#	PhoneFieldName	   -- name of phone field (capitalized)
	#    
	#	EmailField	   -- database name of email field
	#	EmailFieldName	   -- name of email field (capitalized)
	#    
	#	FirstNameField	   -- database name of first name field
	#	FirstNameFieldName -- name of first name field (capitalized)
	#	LastNameField	   -- database name of last name field
	#	LastNameFieldName  -- name of last name field (capitalized)


	my $id_field    = $opt->{IDField};
	my $code_field  = $opt->{CodeField};
	my $phone_field = $opt->{PhoneField};
	my $email_field = $opt->{EmailField};

	my $name_field  = $opt->{NameField};

	return ''
	  unless ($id_field || $name_field || $code_field || $phone_field || $email_field);

	#
	# Set up default options
	#

	my $record_set_name = "Search";
	$record_set_name = $opt->{RecordSetName}		|| '';

	my $id_field_name  = $opt->{IDFieldName}	
	  || _mktitle($id_field)
	  || '';
	my $code_field_name  = $opt->{CodeFieldName}	
	  || _mktitle($code_field)
	  || '';
	my $phone_field_name = $opt->{PhoneFieldName}
	  || _mktitle($phone_field)
	  || '';
	my $email_field_name = $opt->{EmailFieldName}
	  || _mktitle($email_field)
	  || '';

	#
	# Name fields are different - there can be only NameField, or there
	# can be FirstNameField and LastNameField to be used together.
	#

	$name_field = $opt->{FirstNameField}
	  if defined $opt->{FirstNameField};

	my $name_field_name = $opt->{FirstNameFieldName}
	  || $opt->{NameFieldName}
	  || _mktitle($name_field)
	  || ''; 

	my $last_name_field      = $opt->{LastNameField}	|| '';
	my $last_name_field_name = $opt->{LastNameFieldName}	|| '';

	#
	# Change user input into formatted search scripts for parse().
	# A semi-colon (;) can now be used as a separator instead of a pipe (|).
	#

	# There are 6 type of searches:
	# - Lookup using an ID number.
	#
	# - Simple search matching the code name(s).
	#   Codes are all uppercase or all uppercase strings. 
	#   The field(s) searched are those listed in CodeField option.
	#
	# - Simple search matching phone number(s).
	#   Phone numbers are all digits with -(). character separators.
	#   The field(s) searched are those listed in PhoneField option.
	#
	# - Simple search matching email address(es).
	#   Email addresses are strings with a @ character.
	#   The field(s) searched are those listed in EmailField option.
	#
	# - Simple search matching the name, or if specified, the last name
	#   and possibly the first name if that is specified too.
	#
	# - Complex search specifying a SQL where clause (without the where
	#   keyword) which is optionally preceded with a script. The script
	#   is used to substitute variables and parameters into the SQL
	#   where clause. Script statements are separated with | characters.
	#   The script title is the text up to the first pipe (|) (unless the
	#   text is a script statement), and the sql where clause is the text
	#   from the last | to the end of the script. The simplest script is
	#   therefore: "title|where_clause". The semi-colon character can be
	#   used instead of the pipe (|) character.

	SWITCH: {
                if ($search =~ s/^(['"])(.*)\1\s*$/$2/s) {
                        # Simple search.
                        # Search is a quoted string to match a name.
                        # Quotes are removed before trying to match name.

                        my $nf = $last_name_field ? $last_name_field : $name_field;
			die <<EOF unless $nf;
A name field to search has not been specified.
In order to search for this pattern: $search 
a non-numeric field has to be specified as a name field.

This is a problem for whoever set up your database.
EOF
			($title = $search) =~ s/'/\\'/g; # Quote ' quotes
			$title = qq($record_set_name matching '$title');

                        $search = Acme::Db->escape_sql($search);
			$search = $title.$SS._mkwhere($nf,$Match_Op,"^$search.*");
			return $search;
                }

		if ($search =~ /$SSP/) {
			# Complex search
			# Search has an unquoted | or ; delimiter.
			# Search is a script.

			return $search;
		}

		if (
		  $search =~ /[<>=!~]/ or
		  $search =~ /\bin\s+\(/ or
		  $search =~ /\bis\b/ or
		  $search =~ /\bbetween\b/ or
		  $search =~ /\blike\b/ or
		  $search =~ /\bexists\b/) {
			# Complex search.
			# Search is a where clause.
			# Evaluate where clause as a null script.
			# Allow the substitution of default variables and field
			# values in the where clause.

			$title = qq($record_set_name where $search);
			$search = qq($title$SS$search);
			return $search;
		}

		if ($search =~ /\s*\d+([,+\s]+\d+)+\s*/s) {
			# Complex search.
			# Search is a space, comma, or plus separated list of
			# ID numbers.

			die <<EOF unless $id_field;
A ID field to search has not been specified.
In order to search for this pattern: $search 
a numeric field has to be specified.

This is a problem for whoever set up your database.
EOF
			$title = qq($id_field_name matching $search);

			$search =~ s/^\s+//s;	# remove beginning whitespace
			$search =~ s/\s+$//s;	# remove trailing whitespace
			$search =~ s/[,+\s]+/, /gs; # change spaces to commas
			$search =~ s/, $//s;	# remove trailing comma (fluff)

			$search = $title.$SS._mkwhere($id_field," in ($search)");
			return $search;
		}

		if ($search =~ /^\s*(\d+)\s*$/s) {

			# Lookup search.
			# Search is an ID number.

			die <<EOF unless $id_field || $name_field;
The database field to use to lookup numbers has not been specified.

This is a problem for whoever set up your database.
EOF
			if ($id_field) {
				# redefine search
				$title = qq($id_field_name matching $1);
				$search = $title.$SS._mkwhere($id_field,'=',$1);
			}
			else {
				$title = qq($name_field_name matching $1);
				$search = $title.$SS._mkwhere($name_field,'=',$1);
			}
			return $search;
		}

		if ($last_name_field && $search =~ /^\s*(\w+),[ \t]*(\w+)\s*/s) {
			# Simple search.
			# Search is a match of last name and first name.
			# Note: if last_name_field is used, name_field can have
			# only one field name (not a comma separated list).

			($title = $search) =~ s/'/\\'/g; # Quote ' quotes
			$title = qq($record_set_name matching '$title');
			my $lastname = Acme::Db->escape_sql($1);
			my $name = Acme::Db->escape_sql($2);
			$search =
	  qq($title$SS$last_name_field $Match_Op '^$1.*' and $name_field $Match_Op '^$2.*');
			return $search;
		}

		if ($phone_field && $search =~ /^\s*(\d{0,5}([().-]\d{1,5})+)\s*$/s) {
			# Simple search.
			# Search is a match of phone field and input 
			# is all numbers with -() or . delimiters.

			$title = qq($record_set_name matching '$search');

			# Look for options to determine if the match is to be
			# at the beginning of the field or the end of the field,
			# or if the pattern must be equal to the field contents.

			my $match = $opt->{PhoneFieldMatch} || 'end';
			my ($bgn,$end);

			if ($match eq 'begin') {
				$bgn = '^';
				$end = '.*';
			}
			elsif ($match eq 'end') {
				$bgn = '.*';
				$end = '$';
			}
			elsif ($match eq 'equal') {
				$bgn = '^';
				$end = '$';
			}

			$search = $title.$SS._mkwhere($phone_field,$Match_Op,"$bgn$1$end");
			return $search;
		}

		if ($email_field && $search =~ /^\s*(.+\@.+)\s*/s) {
			# Simple search.
			# Search is a match of email field and input 
			# is a pattern with a @ character in it.

			$title = qq($record_set_name matching '$search');
			$search = $title.$SS._mkwhere($email_field,$Match_Op,"^$1");
			return $search;
		}

		if ($code_field && $search =~ /^\s*([a-z\d]+|[A-Z\d_-]+)\s*$/s) {
			# Simple search.
			# Search is a match of code field and input 
			# is either all uppercase or all lowercase.
			# Numbers are also allowed.
			# The - and _ characters are allowed if not first.

			$title = qq($record_set_name matching '$search');

			# Look for options to determine if the match is to be
			# at the beginning of the field or the end of the field,
			# or if the pattern must be equal to the field contents.

			my $match = $opt->{CodeFieldMatch} || 'begin';
			my ($bgn,$end);

			if ($match eq 'begin') {
				$bgn = '^';
				$end = '.*';
			}
			elsif ($match eq 'end') {
				$bgn = '.*';
				$end = '$';
			}
			elsif ($match eq 'equal') {
				$bgn = '^';
				$end = '$';
			}

			$search = $title.$SS._mkwhere($code_field,$Match_Op,"$bgn$1$end");
			return $search;
		}

		# Default
		# Simple search.
		# Assume search is a match of last name if defined.

		my $nf = $last_name_field ? $last_name_field : $name_field;

			die <<EOF unless $nf;
A name field to search has not been specified.
In order to search for this pattern: $search 
a non-numeric field has to be specified.

This is a problem for whoever set up your database.
EOF
		($title = $search) =~ s/'/\\'/g; # Quote ' quotes
		$title = qq($record_set_name matching '$title');

		$search =~ s/^\s+//s;		# remove beginning whitespace
		$search =~ s/\s+$//s;		# remove trailing whitespace

		if ($search eq '*') {
			# syntactic sugar
			$search = '.*';		# change to match all records
			$title = qq(All $record_set_name);
		}

		$search = Acme::Db->escape_sql($search);
		$search = $title.$SS._mkwhere($nf,$Match_Op,"^$search.*");
		return $search;
	}
}

sub _mkwhere {
	my ($field_list, $op, $expr) = @_;

	# Helper function for composing where clauses on more than one field.

	my $where = '';
	my $count = 0;
	for (Text::ParseWords::parse_line('\s*,\s*',1,$field_list)) {
		$where .= " or " if $where;
		$where .= "$_ $op";
		$where .= " '$expr'" if defined $expr;
		$count++;
	}
	$where = "( $where )" if $count > 1;

	return $where;
}

sub _mktitle {
	my ($name_list) = @_;

	# Helper function for composing titles from field lists

	my $title = '';
	for (Text::ParseWords::parse_line('\s*,\s*',1,$name_list)) {
		$title .= " or " if $title;
		$title .= ucfirst lc $_;
	}

	return $title;
}

#------------------------------------------------------------------------------
#
# Search script methods
#
#------------------------------------------------------------------------------

sub new { return parse(@_) }

sub parse {
        my $this = shift;
        my $class = ref($this) || $this;

	my $script = shift;
	my $args   = shift;	# variable argument/value pairs or CGI object
	my $do_http_header = shift;  # print cgi header when getting parameters

	#
	# Convert arguments hash to a CGI object if not one already.
	#

	$args ||= {};
	$args = new CGI $args if ref $args ne 'CGI';

	#
	# Define object
	#

	my $self = {
		title		  => '',
		where_clause	  => '',
		order_by_clause   => '',
		join_clause	  => '',
		join_where_clause => '',	# used if join conditions need to be
						# appended to the where clause
		cgi		  => $args,	# values of script parameters/variables
		script_lines	  => [],	# parsed lines of search script
		join		  => undef,	# Acme::SQL::Join object
		joins		  => [],	# array of script's join conditions
		lists		  => {},	# key is field name,
						# value is file names array
		exclude_lists	  => {},	# key is field name,
						# value is file names array
		do_http_header    => ($do_http_header || 0),	# flag
	};

	bless $self, $class;

	$self->_evaluate_script($script) if $script;

	return $self;
}

#------------------------------------------------------------------------------

sub as_sql {
	my $self = shift;

	my $oref = shift;	# options hash reference
	my $s;			# returned sql string

	our $Db_Engine = $oref->{dbengine} || Acme::Config->DBENGINE;
	our $Postgres = $Db_Engine =~ /Postgres/i;
	our $MySQL    = $Db_Engine =~ /MySQL/i;
	our $Informix = $Db_Engine =~ /Informix/i;

	my $j = new Acme::SQL::Join $oref->{from};

	#
	# Initially load lists into database
	#

	$s  = _load_lists($self,$j->primary_table);
	$s .= _load_exclude_lists($self,$j->primary_table);

	#
	# Construct join clause
	#

	$j->add_join($_) foreach @{$self->{joins}};

	$self->{join_clause}       = $j->join_clause;
	$self->{join_where_clause} = $j->join_where;

	#
	# Assemble SQL code
	#

	my $where = $oref->{where} || "1 = 1";	# default value
	my $and   = $self->{where_clause}      ? "and " : "";
	my $join  = $self->{join_where_clause} ? "and " : "";

	$s .= <<EOS if $oref->{into_temp} && $MySQL;
create temporary table $oref->{into_temp}
EOS
	$s .= <<EOS;
select $oref->{distinct} $oref->{select}
EOS
	$s .= <<EOS if $oref->{into_temp} && $Postgres;
into temp $oref->{into_temp}
EOS
	$s .= <<EOS;
from $self->{join_clause}
where $where
 $join $self->{join_where_clause}
 $and $self->{where_clause}
EOS
	$s .= "\ngroup by $oref->{group_by}" if $oref->{group_by};

	$s .= "\nhaving $oref->{having}" if $oref->{having};

	$s .= "\norder by" if $self->{order_by_clause} or $oref->{order_by};
	$s .= " $self->{order_by_clause}" if $self->{order_by_clause};
	$s .= "," if $self->{order_by_clause} and $oref->{order_by};
	$s .= " $oref->{order_by}" if $oref->{order_by};

	$s .= <<EOS if $oref->{into_temp} && $Informix;
into temp $oref->{into}
EOS
	debug("Searcher::sql: <pre>$s</pre>");

	return $s;
}

#------------------------------------------------------------------------------

sub uses_table {
	my $self = shift;
	my $table = shift;

	return grep { $table eq $_ } $self->{join}->dbtables if $self->{join};

	# A primary table has not been deduced from an sql script and a join
	# structure has not been built yet. Scan collected join clause 
	# fragments for table references.

	my $jtab;
	for (@{$self->{joins}}) {
		($jtab) = /\bjoin\s+(\w)/i;
		return 1 if $jtab eq $table;
	}
	return;
}

#------------------------------------------------------------------------------

sub title {
	my $self = shift;
	return $self->{title} || $self->{script_title};
}

#------------------------------------------------------------------------------

sub annotate {
	my $self = shift;
	return $self->{annotate};
}

#------------------------------------------------------------------------------

sub where_clause {
	my $self = shift;

	my $and = $self->{join_where_clause} ? "\n and " : "";

	return $self->{where_clause} . $and . $self->{join_where_clause};
}

#------------------------------------------------------------------------------

sub order_by_clause {
	my $self = shift;
	return $self->{order_by_clause} || '1';
}

#------------------------------------------------------------------------------

sub join_clause {
	my $self = shift;
	return '' unless $self->{join};
	return $self->{join}->join_clause;
}


#------------------------------------------------------------------------------
#
# Private methods for script processing
#
#------------------------------------------------------------------------------

sub _evaluate_script {
	my $self = shift;
	my $script = shift;

	debug("Original Script:<pre>$script</pre>\n");

	#
	# Set up secure area to execute search script
	#

	$sandbox::today = today();
	$sandbox::user  = user();
	%sandbox::param = ();

	# Hack into the sandbox the Searcher object reference for the search
	# script functions to use to store their stuff.
	$sandbox::SELF_REF = $self;	# for script funtions

	my $sandbox = new Safe 'sandbox';

	$sandbox->share_from('Acme::Date',   [ 'today','dayadd','daydiff' ]);

	#
	# Preprocess the script. Look for parameters.
	#

	my @params = $self->_preprocess_script($script,$sandbox);

	#
	# Define parameter values from cgi variables for the param() function.
	# This is how form field values should get substituted into the search
	# clause on canned searches as it is safer than the arguments_script
	# below. The arguments_script is more convenient in canned search forms.
	#

	my $parameters_script = "%param = ();";
	my $p;
	foreach (@params) {
		($p) = $self->{cgi}->param($_) =~ /(.*)/;	# untaint
		$parameters_script .= qq{\n\$param{$_}=q{$p};};
	}

	#
	# Change cgi arguments to simple arguments assignments.
	# This is how form field values get substituted into the
	# search clause on canned searches.
	#

	my $arguments_script = "";
	my ($arg, $val);
	foreach $arg ($self->{cgi}->param()) {
		next if ($arg eq 'search');
		$val = $self->{cgi}->param($arg) || '';
		$arg =~ s/\W/__/g;	# Change SQL column name to a variable
		$arguments_script .= "\$$arg = q{$val};\n"
	}

	($arguments_script) = $arguments_script =~ /(.*)/s;	# untaint

	#
	# Reinitialize structures built in script during preprocessing.
	#

	$self->{joins} = [];
	$self->{lists} = {};
	$self->{exclude_lists} = {};

	#
	# Construct perl script to eval
	#

	# Note: the following substitution for variables s/\$(\w+)/${$1}/eg
	# does not work on 'my' variables, only globals. However, the
	# substitution as recommended in the Camel book s/(\$\w+)/$1/eeg
	# does work on 'my' variables and globals, however, it triggers
	# a Taint exception when the evaluated string gets re-evaluated
	# again the second time.

	$script = join "\n",
	  q{local $^W;},	# turn off -w for unitialized variables
	  qq{local \$search_str = q{$self->{script_where}};},
	  qq{local \$title = q{$self->{script_title}};},
	  q{local $annotate = '';},
	  $arguments_script, 
	  $parameters_script, 
	  @{$self->{script_lines}},			# search script
	  q{$title =~ s/\$(\w+)/${$1}/eg;},		# substitute variables
	  q{$search_str =~ s/\$(\w+)/${$1}/eg;},	# substitute variables
	  q{return ("$title", "$search_str","$annotate");};

	debug("Evaluate Script:<pre>$script</pre>");

	#
	# Execute script
	#

	(
	  $self->{title},
	  $self->{where_clause},
	  $self->{annotate},
	) = $sandbox->reval($script);

	croak "Errors found evaluating search script. $@" if $@;

	#
	# The user can now specify an optional sort clause by using the key
	# words "order by" in their search (where) clause. Separate it out
	# here as the final operation of evaluating the script.
	#

	$self->{order_by_clause} = '';
	if ($self->{where_clause} =~ s/\border by\b(.*)//) {
		$self->{order_by_clause} = $1;
	}
}

#------------------------------------------------------------------------------

sub _preprocess_script() {
	my $self = shift;
	my ($script,$sandbox) = @_;

	# Scan the script for all parameter assignments ($var =? ...).
	# Change the parameter assignment in the script to use the param()
	# function to retreive the values entered by the user.
	
	#
	# Strip white space from script
	#

	$script =~ s/^\s+//;
	$script =~ s/\s+$//;

	#
	# Parse search script lines
	#

	@{$self->{script_lines}} = Text::ParseWords::parse_line($SSP,1,$script);

	# Strip optional last line where clause from script
	$self->{script_where} = pop @{$self->{script_lines}} || '';

	# Strip optional first line title from script
	$self->{script_title} = shift @{$self->{script_lines}} || '';

	#
	# Scan script for parameter statements
	#

	my @params;	# returned parameters
	my @get_params;	# parameters to get from the user

	my ($var,$def);	# variable, definition parts of parameter stmt
	my ($type, $default, $prompt, $help);	# parts of parameter definition
	my $var_name;

	my $parameters_script = "%param = ();";

	foreach (@{$self->{script_lines}}) {
		if (/=\?/) {

			#
			# Process parameter definition.
			#

			($var,$def) = split /=\?/;
			($type, $default, $prompt, $help) = split /\?/, $def;

			if ($default && ! $prompt) {
				# Only 2 fields in parameter definition.
				# Make the default field the optional field.
				$prompt = $default;
				$default = "";
			}

			($var_name = $var) =~ s/^\$//;	# remove '$' from var

			push @get_params, $_;	# track parameters to get

			#
			# Build parameter values. Use default values.
			#

			$parameters_script .= qq{\n\$param{$var_name}="$default";};

			#
			# Change this "parameter definition" script statement
			# into a "get cgi parameter" statement.
			#

			$_ = "$var=param('$var_name')";
		}

		if (/\bparam\((['"])(\w+)\1/) {
			push @params, $2;	# track cgi parameters to assign
		}

		$_ .= ';';  # append semi-colon delimiters to each line (Perl)
	}

	#
	# Construct script
	#

	my $prep_script = join "\n",
	  q{local $^W;},	# turn off -w for unitialized variables
	  $parameters_script,
	  @{$self->{script_lines}},			# search script
	  q{my $title = q[} . $self->{script_title} . q{];},
	  q{$title =~ s/\$(\w+)/${$1}/eg;},		# substitute variables
	  q{return $title;};

	debug("Preprocess Script:<pre>$prep_script</pre>\n");

	#
	# Execute preprocessor script
	#

	my $title = $sandbox->reval($prep_script);

	croak "Searcher: Errors were found in search script:\n $@" if $@;

	#
	# Redefine search script with changed parameter assignment statements
	#

	if (@get_params) {
		# Search parameters are specified in the script.

		#
		# Change search CGI parameter to new preprocessed script.
		#

		my $new_script = join $SS,
		  ($self->{script_title} || ''),
		  (map {s/;$//; $_} grep {$_} @{$self->{script_lines}}),
		  $self->{script_where};

		$self->{cgi}->param('search',$new_script);

		#
		# Note: _get_script_parameters() will exit this
		# program after parameter form is displayed.
		# This program will re-execute when the user
		# submits the form.
		#

		$self->_get_script_parameters($self->{script_title},@get_params);

		exit;
	}

	return @params;
}

#------------------------------------------------------------------------------

sub _get_script_parameters() {
	my $self = shift;
	my $title = shift;

	#
	# Display parameter entry form and exit.
	#

	print $self->{cgi}->header(-type=>'text/html') if $self->{do_http_header};
	Acme::HTML::Parameters->get_parameters(
		title   => $title,
		cgi     => $self->{cgi},
		params  => [ @_ ],

		# Look in cgi for instructions for parameter screen.
		instructions      => ($self->{cgi}->param('instructions')      || ''),
		instructions_file => ($self->{cgi}->param('instructions_file') || ''),
		instructions_url  => ($self->{cgi}->param('instructions_url')  || ''),
		);
}

#------------------------------------------------------------------------------
#
# Private methods to help assembling SQL statements to load lists
#
#------------------------------------------------------------------------------

use Acme::Base::Schema;		# to find table key types in list_join()

sub _load_lists {
	# Return sql to load lists specified in script.

	my $self = shift;
	my $primary_table = shift;	# the table the list belongs to

	my $s = '';			# returned sql for loading list files

	our ($Informix, $Postgres, $MySQL);

	my $l = $self->{lists};
	if (%$l) {

		#
		# Import lists into database tables for joining
		#

		# All the lists joining a field are loaded into a temporary
		# table, and then a combined table of unique entries is
		# made to join with the query.

		our $List_Space = Acme::Config->List_Space;

		# Assure table name conforms to valid Informix names
		# (18 char. max, no periods)
		my $max_table_name_length = 16;	# maximum length - 2 for prefix

		my ($temp_table, $temp_work_table);
		my %temp_table_created;		# flags
		my $offset;
		my ($field, $file);
		my ($tbl,$col,$t,$f,$key,$key_type);

		my $schema = new Acme::Base::Schema;

		for $field (keys %$l) {

			# Issue error message if script function was called 
			# with out a field spec. This is an error message
			# from list_join that was delayed until now.
			die "The field to join with the list needs to be"
			  . " specified in the list_join() search script"
			  . " function"
			  if ! $field;

			#
			# Assure field name is fully qualified.
			#

			($tbl,$col) = $field =~ /(.*)\.(.*)/
			  or
			($tbl,$col) = ($primary_table,$field);

			$t = $schema->t($tbl)
			  or die "The database table '$tbl' is not recognized";
			
			$key = $t->key;
			if ($f = $t->{$key}) {
				
				#
				# Set temporary table key type to be the same
				# as the joined table's key type.
				#

				$key_type = $f->type;
			}
			else {

				#
				# Key is an ID. Set key type for temporary
				# table to be an integer type.
				#

				$key_type = 'integer';	# key is an ID number
			}

			#
			# Make name of temporary working table and data table.
			# The name is keyed off the join'ed table's name. All
			# lists associated with that table is loaded.
			#

			$temp_table = $field;
			$temp_table =~ s/\W//;	# remove non-word characters
			$offset = length $temp_table;
			$offset = $max_table_name_length
			  if $offset > $max_table_name_length;
			$temp_table = substr($temp_table, -$offset);
			$temp_work_table = "zy" . $temp_table;
			$temp_table = "zz" . $temp_table;

			#
			# Create temporary table.
			#

			$s .= "create temp table $temp_work_table (zzkey $key_type);\n"
			  unless $temp_table_created{$field};
			$temp_table_created{$field} = 1;	# true

			#
			# Load all files of keys into working table
			#

			for $file (@{$self->{lists}->{$field}}) {
				$file = $List_Space . ($file =~ /^$DS/ ? '' : '/') . $file;

				$Informix and $s .=
		"load from \"$file\" insert into $temp_work_table(zzkey);\n";

				$Postgres and $s .=
		"copy $temp_work_table from '$file';\n";

				$MySQL and $s .=
		"load data infile '$file' into table $temp_work_table (zzkey);\n";
			}

			#
			# Copy unique keys into a temp table to use in sql
			#

			$Informix and $s .=
	"select distinct zzkey from $temp_work_table into temp $temp_table;\n";

			$Postgres and $s .=
	"select distinct zzkey into temp $temp_table from $temp_work_table;\n";

			$MySQL and $s .=
	"create temp table $temp_table select distinct zzkey from $temp_work_table;\n";

			#
			# Adjust SQL's where clause to exclude keys in temp table
			#

			push @{$self->{joins}},
			  "$tbl inner join $temp_table on $tbl.$key = $temp_table.zzkey";
		}
	}

	return $s;
}

#------------------------------------------------------------------------------

sub _load_exclude_lists {

	# Return sql to load lists specified in script.
	# These lists contain ID numbers to EXCLUDE from resultant list.

	my $self = shift;
	my $primary_table = shift;	# the table the list belongs to

	my $s = "";			# returned sql for loading list files

	our ($Informix, $Postgres, $MySQL);

	my $l = $self->{exclude_lists};
	if (%$l) {

		#
		# Import lists into database tables for joining
		#

		# All the lists joining a field are loaded into a temporary
		# table, and then a combined table of unique entries is
		# made to join with the query.

		our $List_Space = Acme::Config->List_Space;

		# Assure table name conforms to valid Informix names
		# (18 char. max, no periods)
		my $max_table_name_length = 16;	# maximum length - 2 for prefix

		my ($temp_table, $temp_work_table);
		my %temp_table_created;		# flags
		my $offset;
		my ($field, $file);
		my ($tbl,$col,$t,$f,$key,$key_type);

		my $schema = new Acme::Base::Schema;

		for $field (keys %$l) {

			# Issue error message if script function was called 
			# with out a field spec. This is an error message
			# from list_exclude that was delayed until now.
			die "The field to join with the list needs to be"
			  . " specified in the list_exclude() search script"
			  . " function"
			  if ! $field;

			#
			# Assure field name is fully qualified.
			#

			($tbl,$col) = $field =~ /(.*)\.(.*)/
			  or
			($tbl,$col) = ($primary_table,$field);

			$t = $schema->t($tbl)
			  or die "The database table '$tbl' is not recognized";
			
			$key = $t->key;
			if ($f = $t->{$key}) {
				
				#
				# Set temporary table key type to be the same
				# as the joined table's key type.
				#

				$key_type = $f->type;
			}
			else {

				#
				# Key is an ID. Set key type for temporary
				# table to be an integer type.
				#

				$key_type = 'integer';	# key is an ID number
			}

			#
			# Make name of temporary working table and data table.
			# The name is keyed off the join'ed table's name. All
			# lists associated with that table is loaded.
			#

			$temp_table = $field;
			$temp_table =~ s/\W//;	# remove non-word characters
			$offset = length $temp_table;
			$offset = $max_table_name_length
			  if $offset > $max_table_name_length;
			$temp_table = substr($temp_table, -$offset);
			$temp_work_table = "zw" . $temp_table;
			$temp_table = "zx" . $temp_table;

			#
			# Create temporary table.
			#

			$s .= "create temp table $temp_work_table (zxkey $key_type);\n"
			  unless $temp_table_created{$field};
			$temp_table_created{$field} = 1;	# true

			#
			# Load all files of keys into working table
			#

			for $file (@{$self->{exclude_lists}->{$field}}) {
				$file = $List_Space . ($file =~ /^$DS/ ? '' : '/') . $file;

				$Informix and $s .=
		"load from \"$file\" insert into $temp_work_table(zxkey);\n";

				$Postgres and $s .=
		"copy $temp_work_table from '$file';\n";

				$MySQL and $s .=
		"load data infile '$file' into table $temp_work_table (zxkey);\n";
			}

			#
			# Copy unique keys into a temp table to use in sql
			#

			$Informix and $s .=
	"select distinct zxkey from $temp_work_table into temp $temp_table;\n";

			$Postgres and $s .=
	"select distinct zxkey into temp $temp_table from $temp_work_table;\n";

			$MySQL and $s .=
	"create temp table $temp_table select distinct zxkey from $temp_work_table;\n";

			#
			# Adjust SQL's where clause to exclude keys in temp table
			#

			if ($self->{where_clause}) {
				$self->{where_clause} =
				  "((" . $self->{where_clause} . ")\n   and " .
				  "($field not in (select zxkey from $temp_table)))";
			}
			else {
				$self->{where_clause} =
				  "($field not in (select zxkey from $temp_table))";
			}
		}

	}

	return $s;
}

#--------------------------------------------------------------------
#
# Script functions which can be used in search scripts.
#
#--------------------------------------------------------------------

package sandbox;

sub param {
	my $p = shift;

	# This is a pseudo user script function. It can be used by
	# the user/scripter in the search script, but it is usually
	# used by the Preprocessor to substitute for a script parameter
	# statement (e.g. $s=?t10?Mack?Enter first name;) once the
	# parameter has been entered by the user.

	return $sandbox::param{$p};
}

#--------------------------------------------------------------------

sub table_join {
	# Join a table
	my $join_expr = shift;
	push @{ $sandbox::SELF_REF->{joins} }, $join_expr;
}

#--------------------------------------------------------------------

sub list_join {
	# Join a table field to a list of values.
	my $file = shift;	# file name of list values
	my $join_field = shift;	# database field whose values must match the
				# list's value. The field name can be 
				# qualified by the table name (table.field).

	# Delay error message until we're out of the Safe module.
	# It spawns another fluff error because of Acme::HTML set_message()
	# statement.
	#die "The field to join with the list needs to be specified in the"
	#  . " list_join() search script function"
	#  if ! $join_field;

	# Define field pointer to an array if not already there.
	$sandbox::SELF_REF->{lists}->{$join_field} ||= [];

	push @{ $sandbox::SELF_REF->{lists}->{$join_field} }, $file;
}

#--------------------------------------------------------------------

sub list_exclude {
	# Exclude list of values from selection.
	my $file = shift;	# file name of list values
	my $join_field = shift;	# database field whose values must NOT match the
				# list's value. The field name can be 
				# qualified by the table name (table.field).

	# Delay error message until we're out of the Safe module.
	# It spawns another fluff error because of Acme::HTML set_message()
	# statement.
	#die "The field to join with the list needs to be specified in the"
	#  . " list_exclude() search script function"
	#  if ! $join_field;

	# Define field pointer to an array if not already there.
	$sandbox::SELF_REF->{exclude_lists}->{$join_field} ||= [];

	push @{ $sandbox::SELF_REF->{exclude_lists}->{$join_field} }, $file;
}

1;
