package Acme::Security;

=head1 NAME

Acme::Security - Security routines

=head1 SYNOPSIS

    use Acme::Security;
      or 
    use Acme::Security 'access,update';
      or
    require Acme::Security;

    Acme::Security::check('access,update');
      or
    die Acme::Security::alert "You do not have privileges to execute this program"
      unless Acme::Security::allow('Update');

    $table = 'customers';
    Acme::Security::check('access,update',$table);
      or
    die Acme::Security::alert "You do not have privileges to update $table."
      unless Acme::Security::allow('Update',$table);

    # To check a sql query with security privileges on tables and fields.
    $sql = 'select employee_name, salary from employees'; 
    $sql = Acme::Security::clean_sql($sql);
      or
    my $sqlo = new Acme::SQL $sql;
    $sql = Acme::Security::clean_sqlo($sqlo);

=head1 DESCRIPTION

The check() routine checks if the specified privilege is allowed and alerts
the web administrator and dies if not. The message displayed to the user can
be changed by setting $Acme::Security::Message just before running check(). The
message is reset to the default message after check() runs.

The alert() routine emails a message to the web administrator and
to the webmaster. It returns the original message so that it can be used like:

	die Acme::Security::alert($message);

The allow() routine checks if the specified privilege is allowed.
The deny() method is opposite of allow().

Privileges are specified by a string's beginning character.

	A -- access
	I -- insert
	U -- update
	D -- delete

The optional second argument to check() or allow() supplies a security code,
(an identifier code) to  use to check privileges. If not supplied, the
program's name is used.

The C<use Acme::Security 'Access,Update'> statement automatically check()'s
user privileges during compile time, however Perl's compile time error message
clouds the the error message output by check(). Messages are cleaner if
check() is executed during run time.

The I<Acme::Security::User> package is used to actually check privileges.

The clean_sql() routine checks accessed tables and columns in a sql query
and either dies or modifies the query according to the user's privileges.

The clean_sqlo() routine is like clean_sql() except it works on a Acme::SQL
object directly, which is a parsed sql query object that clean_sql() uses.
It is sometimes more convenient to use, especially if a query was combined
with a I<Acme::Searcher> script.

=head1 CHANGE LOG

 6/13/00  Initial version.
 7/25/00  Changed to define symbols in caller's name space
 3/17/01  Added check() which throws an exception if there is no privileges.
          Removed superfluous interface routines.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

$Acme::Security::Default_Message = "Access denied.\n";
$Acme::Security::Message = '';

use Acme::Security::User;
use Acme::Security::Alert;

#------------------------------------------------------------------------------

sub import { shift; for (@_) { check($_) } }

sub check($@) {
	die alert($Acme::Security::Message || $Acme::Security::Default_Message)
	  unless allow(@_);
	$Acme::Security::Message = '';	# reset
	return 1;
}

sub allow($@) {
	# Convert input access code like "Insert,Update,Delete" into "IUD".
	my $p = (join '', map {uc(substr($_,0,1))} split /\W+/, shift) || 'A';
	return Acme::Security::User::allow($p,@_);
}

sub deny($@) {
	return ! allow(shift,@_);
}

sub alert(@) {
	return Acme::Security::Alert::security_alert(@_);
}

sub clean_sql($;$) {
	require Acme::Security::SQL;
	return  Acme::Security::SQL::clean_sql(@_);
}

sub clean_sqlo($;$) {
	require Acme::Security::SQL;
	return  Acme::Security::SQL::clean_sqlo(@_);
}

1;
