package Acme::Security::Access;

=head1 NAME

Acme::Security::Access - Perl module interface database field definitions

=head1 SYNOPSIS

    use Acme::Security::Access;

    my ($allow,$deny) = Acme::Security::Access->privileges;

=head1 DESCRIPTION

Acme::Security::Access builds two data structures, hashes, of security
information from a file of user names and security codes. 

The file is organized according to security code (which is a program name,
table name, or an arbitrary code defined in schema), whereas the C<allow> and
C<deny> data structures are built organized by user login name.

Security codes can be the string 'ALL', a full pathname, a pathname relative
to /cgi-bin/, a table name, or a security code specified in the schema for a
table or field.

Privilege codes are a comma separated list of the words Access, Insert, Update,
and Delete. They can also be specified by just the first capitalized character
of each word like: AIUD.

Privilege codes are converted into bit codes by OR'ing the constants Access,
Insert, Update, or Delete OR'd together.

The user name is the name returned by Local::Acme::User::user(), which is
usually their login name.

The process for determining if the user has a privilege is:

	Privilege is allowed if user ID privilege is found in $allow.
	Privilege is denied if user ID privilege is found in $deny.

	Privilege is allowed if user's group ID(s) privilege is found in $allow.
	Privilege is denied if user's group ID(s) privilege is found in $deny.

	Privilege is allowed if the ID 'ALL' privilege is found in $allow.
	Privilege is denied if the ID 'ALL' privilege is found in $deny.

	Otherwise, privilege is granted.


=head1 CHANGE LOG

 10/14/01 Initial version.
 9/11/02  Added groups.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000,2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

use Acme::Security::Alert;
use Acme::Cache;
use Acme::Config;
use Text::ParseWords;

# Debug Note: The checking of security access often occurs before the HTTP
# header has been output, so it is safest to print out the header here first.
#print "Content-type: text/html\n\n" unless $ENV{MOD_PERL};

our $Debug=0;
sub debug { print "<br> --- ", @_, "\n" if $Debug }

#------------------------------------------------------------------------------

sub new {
	my $this = shift;
	my $class = ref($this) || $this;

	my $self = Acme::Cache->get(
		file        => (Acme::Config->Access_File || '../etc/access.conf'),
		constructor => \&add,
		class       => __PACKAGE__,
		);

	return bless $self, $class;
}

sub add {
	my ($self,$filename) = @_;

	#
	# Define object to cache if needed.
	#

	if (! ref($self)) {
		my $class = $self;
		$self = {};
		bless $self, $class;

		$self->{allow}  = {};	# structure of allow privileges
		$self->{deny}   = {};	# structure of deny privileges
		$self->{groups} = {};	# structure of groups
	}

	debug("\nReading access file: $filename\n");

	#
	# Read/Process access description file
	#

	my ($prefix, $data, $key, $value, $here);
	my $curr_code  = '';	# current security code(s)
	my $curr_allow = '';	# current allow access code
	my $curr_deny  = '';	# current deny access code
	my $curr_group = '';	# current group

	open(ACCESS, '<', $filename)
	  or die Acme::Security::Alert::security_alert("Cannot open access file '$filename'. $!");

	local $/ = "\n";		# set input line separator.
	while (<ACCESS>) {		# read line
		chomp;

		last if /^\s*__END__\b/;

		# Ignore comment lines.
		next if /^\s*#/;	

		# Continue to nextline if last chararacter on the line is \.
		if (s/\\$//) { $_ .= <ACCESS>; redo }

		# Strip comments from end of line.
		s/([^\\])#.*/$1/;	

		# Unescape escaped # comment delimiter.
		s/[\\]#/#/g;

		# Strip trailing white space.
		s/\s+$//;		

		# Continue to nextline if last chararacter on the line is \.
		# This test is repeated after a comment at the end is stripped.
		if (s/\\$//) { $_ .= <ACCESS>; redo }

		# Strip white space at the beginning of the line.
		($prefix, $data) = m/^(\s*)(.*)/;

		# Ignore empty lines.
		next unless $data;	

		# Split line into key value pair.
		($key, $value) = $data =~ m/([^\s]+)\s*(.*)/;

		# Strip possible quotes from value.
		$value =~ s/^(["'])(.*)\1$/$2/;

		#
		# Check if value comes from a here document (<<WORD)
		#

		if ($value =~ /^<<(.*)/) {
			$here = $1;
			$value = '';
			# Read file until a line with only contents of $here.
			while (my $v = <ACCESS>) {
				chomp $v;
				last if $v eq $here;
				$value .= "$v\n";
			}
			chomp $value;
		}

		debug("\n>>data='$data', key='$key', value='$value' ");

		#
		# Check for keywords
		#

		if ($key =~ /^(secure|restrict)$/i) {

			#
			# Keyword 'secure' or 'restrict'
			#

			$curr_code = $value;
			$curr_allow = $curr_deny = $curr_group = '';

			debug("keyword 'secure': setting curr_code='$curr_code':'$value'");

			next;
		}

		if ($key =~ /^allow$/i) {

			#
			# Keyword 'allow'
			#

			($curr_allow,$value) = _extract_priv($value);
			$curr_deny = $curr_group = '';

			debug("keyword 'allow': curr_code='$curr_code', allow='$curr_allow'");
			die Acme::Security::Alert::security_alert(
			  "Bad privilege code: $curr_allow, from line: $data")
			  unless $curr_allow;

			$self->add_allow($curr_code,$curr_allow,$value) if $value;
			next;
		}

		if ($key =~ /^deny$/i) {

			#
			# Keyword 'deny'
			#

			($curr_deny,$value) = _extract_priv($value);
			$curr_allow = $curr_group = '';

			debug("keyword 'deny': curr_code='$curr_code', deny='$curr_deny'");
			die Acme::Security::Alert::security_alert(
			  "Bad privilege code: $curr_deny, from line: $data")
			  unless $curr_deny;

			$self->add_deny($curr_code,$curr_deny,$value) if $value;
			next;
		}

		if ($key =~ /^group$/i) {

			#
			# Keyword 'group'
			#

			$value =~ s/^(\S+)\s*//;

			$curr_group = $1;
			$curr_code = $curr_allow = $curr_deny  = '';

			debug("keyword 'group': group='$curr_group'");
			next unless $curr_group;

			$self->add_group($curr_group,$value) if $value;
			next;
		}

		#
		# Add security settings
		#

		if ($curr_allow) {
			# Add allow privileges
			debug("  append allow '$curr_code: $curr_allow'; '$data'");
			$self->add_allow($curr_code,$curr_allow,$data);
		}
		elsif ($curr_deny) {
			# Add deny privileges
			debug("  append deny '$curr_code: $curr_deny'; '$data'");
			$self->add_deny($curr_code,$curr_deny,$data);
		}
		elsif ($curr_group) {
			# Add user groups
			debug("  append group '$curr_group: '; '$data'");
			$self->add_group($curr_group,$data);
		}
		else {
			# Add to list of codes
			$curr_code .= ' ' if $curr_code;
			$curr_code .= $data;
		}
	}

	close(ACCESS);

	return $self;		# return object as a success value
}

sub add_allow {
	my $self = shift;
	my ($curr_code,$curr_allow,$value) = @_;

	# $curr_code is a list of security codes
	# $curr_allow is a bitmap of privileges (Access,Insert,Update,Delete)
	# $value is a list of user names

	for my $user (Text::ParseWords::parse_line('[\s,;]+',0,$value)) {
		$user = 'ALL' if $user =~ /^all$/i;
		for my $code (Text::ParseWords::parse_line('[\s,;]+',0,$curr_code)) {
			$code = 'ALL' if $code =~ /^all$/i;
			$self->{allow}->{$user}->{$code} = $curr_allow;
		}
	}
}

sub add_deny {
	my $self = shift;
	my ($curr_code,$curr_deny,$value) = @_;

	# $curr_code is a list of security codes
	# $curr_deny is a bitmap of privileges (Access,Insert,Update,Delete)
	# $value is a list of user names

	for my $user (Text::ParseWords::parse_line('[\s,;]+',0,$value)) {
		$user = 'ALL' if $user =~ /^all$/i;
		for my $code (Text::ParseWords::parse_line('[\s,;]+',0,$curr_code)) {
			$code = 'ALL' if $code =~ /^all$/i;
			$self->{deny}->{$user}->{$code} = $curr_deny;
		}
	}
}

sub add_group {
	my $self = shift;
	my ($curr_group,$value) = @_;

	# $curr_group is the name of the group
	# $value is a list of user names

	for my $user (Text::ParseWords::parse_line('[\s,;]+',0,$value)) {
		$self->{groups}->{$curr_group}->{$user} = 1;
	}
}

sub access_groups {
	my $self = shift;
	$self = new($self) if !ref($self);

	my @groups;	# returned list of group names

	my $access_groups = $self->{groups};
	for my $user (@_) {
		push @groups, grep { $access_groups->{$_}->{$user} } keys %$access_groups;
	}
	return @groups;
}

sub privileges {
	# Return pointers to the allow and deny privilege structures.
	my $self = shift;
	$self = new($self) if !ref($self);
	return ($self->{allow},$self->{deny});
}

#------------------------------------------------------------------------------
# Private functions
#------------------------------------------------------------------------------

# Constants
use constant Access => 0x0001;
use constant Insert => 0x0002;
use constant Update => 0x0004;
use constant Delete => 0x0008;

sub _extract_priv {
	# Extract privilege code from line.
	# Return privilege and line remainder.
	my ($line) = @_;

	$line =~ s/^(['"])(.*?)\1\s*//
	  or
	$line =~ s/^((\S+))\s*//;

	my $code = $2;

	# Change privilege like 'AIUD' into bitmap.
	my $priv = 0;
	if ($code =~ /^all$/i) {
		$priv = Access | Insert | Update | Delete;
	}
	else {
		# Convert access code like "Insert,Update,Delete" into "IUD".
		$code = join '', map {uc(substr($_,0,1))} split /\W+/, $code
		  if $code !~ /^[AIUD]+$/;

		$priv |= Access if index($code,'A') >= 0;
		$priv |= Insert if index($code,'I') >= 0;
		$priv |= Update if index($code,'U') >= 0;
		$priv |= Delete if index($code,'D') >= 0;
	}

	return ($priv,$line);
}

1;
