package Acme::Security::Alert;

=head1 NAME

Acme::Security::Alert - Security routines

=head1 SYNOPSIS

    require Acme::Security::Alert;

    die Acme::Security::Alert::security_alert(
      "You do not have privileges to execute this program")
      unless Acme::Security::allow('update');


=head1 DESCRIPTION

This is a utility function for Acme::Security.

The security_alert() routine emails a message to the web administrator and
to the webmaster. It returns the original message so that it can be used like:

	die security_alert($message);

=head1 CHANGE LOG

 7/25/00  Initial version.
 3/17/01  Repackaged and streamlined.
 12/26/02 Add alternative addresses and subject.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000-2002 Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Acme::HTTP		qw(unescape);
use Acme::Config;
use Local::Acme::User	qw(user);
use Sys::Hostname;

$Webmaster        = Acme::Config->Webmaster;
$Webadministrator = Acme::Config->Webadministrator;

#------------------------------------------------------------------------------

sub security_alert(@) {
	my ($msg,$to,$subject) = @_;

	no warnings qw(uninitialized);

	my $time = localtime;
	my $user = user();
	my $host = eval { hostname() } || $ENV{SERVER_NAME} || '';
	my $arguments = "'" . (join "'", @ARGV) . "'";
	my $request_method = "'" . ($ENV{REQUEST_METHOD} || '') . "'";
	my $query_string = "'" . ($ENV{QUERY_STRING} || '') . "'";
	my $dqs = unescape($query_string);
	my $decoded_query_string = ($dqs ne $query_string) ? $dqs : $query_string;

	#
	# Create alert message
	#

	my $alert_message = <<End_of_Msg;

$msg

---------------------------------------------------------------------- 

            Time $time
            User $user
            Host $host
         Program $0
       Arguments $arguments
  Request_method $request_method
    Query_string $decoded_query_string
End_of_Msg

	#
	# Show stack trace on who called this program.
	#

	$alert_message .= "\nStack Trace:\n";
	my $frame = 1;
	while ( my ($pkg, $file, $line, $sub) = caller($frame++)) {
		$alert_message .= "    $file: line $line: $pkg::$sub()\n";
	}

	#
	# Show environment program was executed under.
	#

	$alert_message .= "\nEnvironment:\n"
	. (join "", map {++$c % 2 ? $_.'=' : $_."\n"} %ENV);

	#
	# Untaint variables.
	#

	# Note: We want to send 'truth' to the administrator, so there
	# is a question of how much to filter out of the $alert_message
	# in order to pass the taint test.

	($alert_message) = $alert_message =~ /(.+)/s;

	delete @ENV{ qw(IFS CDPATH ENV BASH_ENV) };     # untaint
	$ENV{PATH} = '/bin:/usr/bin';                   # untaint

	#
	# Send message to administrators.
	#
	
	my $saved_mail_err = '';
	if (open(SENDMAIL,"| /usr/sbin/sendmail -oi -t")) {

		#
		# Send to administrators or send a copy to the administrators
		# if an alternate address was specified in the arguments.
		#

		my $cc = '';
		if ($to) {
			$cc = "\nCC: $Webadministrator; $Webmaster";
		}
		else {
			$to = "$Webadministrator; $Webmaster";
		}

		if ($subject) {
			
			#
			# Prepend the user and host name to subject.
			#

			$subject = "Security: $user: $subject";
		}
		else {
			#
			# Use program name as the subject.
			#

			($subject = $0) =~ s|.*cgi-bin/||;
			$subject = "Security: $user using $subject";
		}

		print SENDMAIL <<End_of_Msg;
From: $Webmaster
To: $to$cc
Subject: $subject
Mime-Version: 1.0
Content-type: text/plain

End_of_Msg
		print SENDMAIL $alert_message;
		close(SENDMAIL);
	}
	else {
		$saved_mail_err = $!;
	}

	#
	# Save message in the security log.
	#

	my $security_log = Acme::Config->Security_Log;
	my $file_exists = (-f $security_log);

	if (open(LOG,'>>',$security_log)) {
		print LOG <<End_of_Msg;

######################################################################
End_of_Msg
		print LOG "\nEmail notification was NOT sent: $saved_mail_err\n"
		  if $saved_mail_err;
		print LOG $alert_message, "\n";
		close(LOG);

		chmod 0622, $security_log if ! $file_exists;
	}

	# Note: CGI::Carp does not include the user's name on the timestamp
	# it prepends for the server log (!!). We append it as a list element
	# so calling program need not bother with it if it does not want to.

	return ($msg,"\n\n$user\@$host\n") if wantarray;
	return $msg;
}

1;
