package Acme::Security::SQL;

=head1 NAME

Acme::Security::SQL - Library of routines for securing sql

=head1 SYNOPSIS

    require Acme::Security::SQL

    $scode = '1';	# allow only 1 query statement (no semi-colons)
    $scode = 'u';	# allow update statement
    $scode = 'udi';	# allow update, delete, and/or insert statements
    
    $sql = Acme::Security::SQL::clean_sql($sql,$scode);


=head1 DESCRIPTION

Check_sql() returns a "clean" sql statement that removes any fields the user
is not allowed to query, update, delete, or insert.

Actually, at this time, fields are not removed as the program dies if user
is not allowed. However, this routine has the side effect of expanding any
* in the field list to the fully qualified fields.

The privilege code used by check_sql allows the programmer to specify if
sql statements can contain UPDATE, INSERT, or DELETE statements. It also
can specify how many sql statements can be contained in the script.

The privilege code is a string  that can contain the following characters:

	u -- Allow UPDATE statement
	d -- Allow DELETE statement
	i -- Allow INSERT statement

The code can also contain a number which specifies the number of sql
statement. For exemple:

	ud4

is a code allowing up to 4 sql statements which may be UPDATE or DELETE
statements.

If a privilege code is not specified, the default behavior is to allow
any number of statements in the script, but none of them can be INSERT
(or SELECT INTO TABLE), UPDATE, or DELETE statements.

Temporary tables are allowed. Only they can be created or dropped. They
can be filled with SELECT INTO TEMP or CREATE TEMP TABLE AS statements.
Using the INSERT statement to fill the temporary tables will fail unless
the I privilage is specified.

If the user has restrict codes that apply, the where clause is modified with
added clause(s) that restrict the selection to records the user can see.

Errors (security breeches) generates two error messages, one displayed to the
user, and another more complete message is emailed to the site administrator.

=head1 NOTE

As with all other security utility routines, these routines should be accessed
indirectly through the routines in the I<Acme::Security> module.

=head1 CHANGE LOG

 6/12/00  Initial version.
 9/30/00  Use Acme::SQL module to parse sql and retain synonyms in query.
 6/10/01  Added restrict_where clauses.
 7/31/01  Added union statement processing.
 10/7/01  Added clean_sqlo() to clean Acme::SQL object.
 1/15/02  Threaded the accommodation of temporary tables into code.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.6.0;
use strict;

use Acme::SQL;
use Acme::Base::Schema;
use Acme::Security::Alert;
use Acme::Config;
use Text::ParseWords;

our $Webadministrator = Acme::Config->Webadministrator;

our $Debug = 0;
sub debug { print "<br> --- ", @_, "\n" if $Debug }
sub debug_dump {
	return unless $Debug;
	require Data::Dumper;
	print "<pre>",Data::Dumper::Dumper(@_),"</pre>";
}

our %Temp_Table;	# hash to track temporary tables defined

#------------------------------------------------------------------------------

sub clean_sql {
	my ($sql, $pcode) = @_;

	$pcode ||= 'A';			# default Access privilage code

	my $clean_sql = '';		# returned sql

	my ($clean_sql_stmt, $select_count, $sql_stmt_prefix);

	local %Temp_Table = ();	# hash of recognized temporary tables


	my $die_msg = <<End_Of_Msg;
Your request has been denied because you are trying to access or alter restricted information.  Either that or this program has encountered a bug and has become utterly confused.

Please contact $Webadministrator.
End_Of_Msg

	my $count = 0;		# tells if parsed item a delimiter or statement
	my $stmt_count = 0;	# number of SQL statements (between ; characters)

	for my $sql_stmt (Text::ParseWords::parse_line('\s*;\s*','delimiters',$sql)) {
		if ($count++ % 2) {
			# Attach original delimiter (;) to sql.
			$clean_sql .= $sql_stmt if $sql_stmt;
			next;
		}

		next unless $sql_stmt;		# ignore empty statements

		$stmt_count++;
		$sql_stmt_prefix = '';

		#
		# Check for LOAD and UNLOAD statements
		# PostgreSQL uses COPY...FROM and COPY...TO
		#

		if ($sql_stmt =~ /^copy\s+(\w+)\s+to\s/i) {

			die_alert($die_msg,
			  'Following SQL has no privileges for a COPY clause.',
			  $sql, $pcode)
			  if $pcode !~ /A/i;

			# Allow COPY table TO 'filename'
			$clean_sql .= $sql_stmt;
			next;
		}

		if ($sql_stmt =~ /^copy\s+(\w+)\s+from\s/i) {

			# Allow COPY table FROM 'filename' for temp tables
			die_alert($die_msg,
			  'Following SQL has an COPY ... FROM clause.',
			  $sql, $pcode)
			  if $pcode !~ /I/i && ! $Temp_Table{$1};

			$clean_sql .= $sql_stmt;
			next;
		}


		#
		# Check for CREATE and DROP statements
		#

		if (my ($temp,$table) = $sql_stmt =~ /^create\s+(\w+)\s+(\w+)/i) {

			# Allow CREATE only if for a temporary table
			die_alert($die_msg,
			  'Following SQL has an CREATE clause.',
			  $sql,$pcode)
			  if $temp !~ /^temp(?:orary)?$/i || $table !~ /^table$/i;

			# Track temporary tables 
			$sql_stmt =~ /^create\s+\w+\s+\w+\s+(\w+)/i;
			$Temp_Table{$1} = 1;

			# Check CREATE...AS SELECT... by stripping off
			# the prefix and parsing the SELECT clause
			# and then reassembling CREATE statement below.

			if ($sql_stmt =~ s/(create.*?)(select\b)/$2/i) {
				$sql_stmt_prefix = $1;

				# pass SELECT part to clean_sqlo() for parsing
			}
			else {
				$clean_sql .= $sql_stmt;
				next;
			}
		}

		if ($sql_stmt =~ /^drop\s+table\s+(\w+)/i) {

			# Allow DROP TABLE only for a temporary table
			die_alert($die_msg,
			  'Following SQL has an DROP clause.',
			  $sql,$pcode)
			  if ! $Temp_Table{$1};

			# Clear temp table flag
			$Temp_Table{$1} = undef;

			$clean_sql .= $sql_stmt;
			next;
		}

		#
		# Check for INSERT, UPDATE, DELETE, and a special case of INSERT,
		# SELECT INTO TABLE clause.
		#

		if ($sql_stmt =~ /\binsert\s+into\s+(\w+)/i) {

			# Allow INSERT of temporary tables
			die_alert($die_msg,
			  'Following SQL has an INSERT clause.',
			  $sql,$pcode)
			  if $pcode !~ /I/i && ! $Temp_Table{$1};

			# Should check user privileges on table here...

			# Check INSERT...SELECT by stripping off the
			# INSERT prefix and parsing the SELECT clause
			# and then reassembling INSERT below.

			if ($sql_stmt =~ s/(insert.*?)(select\b)/$2/i) {
				$sql_stmt_prefix = $1;

				# pass SELECT part to clean_sqlo() for parsing
			}
			else {
				$clean_sql .= $sql_stmt;
				next;
			}
		}

		if ($sql_stmt =~ /\bupdate\s+(\w+)/i) {

			# Allow UPDATE of temporary tables
			die_alert($die_msg,
			  'Following SQL has an UPDATE clause.',
			  $sql,$pcode)
			  if $pcode !~ /U/i && ! $Temp_Table{$1};

			# Should check user privileges on table here...

			# Check UPDATE...SELECT by stripping off the
			# UPDATE prefix and parsing the SELECT clause
			# and then reassembling UPDATE below.

			if ($sql_stmt =~ s/(update.*?)(select\b)/$2/i) {
				$sql_stmt_prefix = $1;

				# pass SELECT part to clean_sqlo() for parsing
			}
			else {
				$clean_sql .= $sql_stmt;
				next;
			}
		}

		if ($sql_stmt =~ /\bdelete\s+from\s+(\w+)/i) {

			# Allow DELETE from temporary tables
			die_alert($die_msg,
			  'Following SQL has an DELETE clause.',
			  $sql,$pcode)
			  if $pcode !~ /D/i && ! $Temp_Table{$1};

			# Should check user privileges on table here...

			$clean_sql .= $sql_stmt;
			next;
		}

		if (my ($temp,$table) = $sql_stmt =~ /\binto\s+(\w+)\s+(\w+)\b/i) {

			# Require 'I' privilege to SELECT INTO table
			# Ignore 'SELECT INTO TEMP' clauses
			if ($temp =~ /^temp(orary)?$/i) {

				# Track temporary tables 
				if ($table =~ /table/i) {
					# Get next word as temp table name.
					($table) = $sql_stmt =~ /\binto\s+\w+\s+\w+\s+(\w+)/i;
				}
				$Temp_Table{$table} = 1;
			}
			else {
				die_alert($die_msg,
				  'Following SQL has a SELECT INTO TABLE clause.',
				  $sql,$pcode)
				  if $pcode !~ /I/i;
			}

			# pass through SQL as clean_sqlo() can parse SELECTs
		}

		#
		# Check the number of statements submitted if pcode restricts them.
		#

		if ($pcode =~ /(\d+)/) {

			die_alert($die_msg,
			  'Following SQL has too many statements.',
			  $sql,$pcode)
			  if $stmt_count > $1;
		}

		#
		# At this point, only SELECT statements can pass
		#

		if ($sql_stmt !~ /^\s*select\s/i) {

			# Allow only SELECT beyond this point
			die_alert($die_msg,
			  'Following SQL has an unrecognized or illegal statement.',
			  $sql,$pcode);
		}

		$clean_sql_stmt = '';
		$select_count = 0;
		for (split /(\s+(?:union|except|intersect)\s+(?:all\s+)?)/i, $sql_stmt) {

			if ($select_count++ % 2) {
				$clean_sql_stmt .= $_;	# 'union ...' connector string
				next;
			}

			#
			# Clean each SELECT clause 
			#

			my $sqlo = new Acme::SQL $_;
			$clean_sql_stmt .= clean_sqlo($sqlo,$pcode);
		}

		# Re-assemble CREATE...AS SELECT... or INSERT INTO...SELECT...
		# or UPDATE...SELECT... statements.

		$clean_sql .= $sql_stmt_prefix . $clean_sql_stmt;
	}

	%Temp_Table = ();	# clear so it won't be used again in clean_sqlo

	return $clean_sql;
}

sub clean_sqlo {
	my ($sqlo,$pcode) = @_;

	$pcode ||= 'A';			# default Access

	my ($table, $field, $tbl, $fld, $t, $f);

	my $schema = new Acme::Base::Schema;

	#
	# Verify tables are valid
	#

	my @tables   = $sqlo->tables;
	my @dbtables = $sqlo->dbtables;
	my $restrict_where = '';

	for $tbl (@dbtables) {

		next if $Temp_Table{$tbl};	# skip temporary tables

		$t = $schema->t($tbl);
		if (not defined $t) {
			die_alert(
			  "The table '$tbl' "
			  . "referred to in the query is not a defined table.\n"
			  . "Is it spelled correctly in the schema?\nError",
			  "Table not defined: '$tbl'.",
			  $sqlo->as_sql,$pcode);
		}

		#
		# Collect restrict where clauses of each table for user.
		#

		$restrict_where .= ($restrict_where ? ' ' : '') .
		  ($t->restrict_where
		    || ($pcode =~ /U/i && $t->restrict_where_update)
		    || ($pcode =~ /D/i && $t->restrict_where_delete)
		  );

		#
		# Die if restrict code restricts insert of records.
		# This is the same as a Secure code restricting insert.
		#

		if ($pcode =~ /I/i && $t->restrict_where_insert) {
			my $die_msg = <<End_Of_Msg;
Your request has been denied because you are trying to insert information into a restricted table. Either that or this program has encountered a bug and has become utterly confused.

Please contact $Webadministrator.
End_Of_Msg
			my $alert_msg = <<End_Of_Alert;
Attempt to insert into a table with a restrict code.
	$restrict_where
Insert rejected.
End_Of_Alert
			die_alert($die_msg,$alert_msg,$sqlo->as_sql,$pcode);
		}
	}

	#
	# Build hash of fields accessed
	#

	my %tblh;
	my $rt;
	my $star_field_found = 0;	# flag
	for $fld ($sqlo->dbcolumns) {
		if ($fld =~ /\*/) {
			$star_field_found++;
			next;
		}

		# Parse fully qualified column name into table and field names.
		# Watch for functions by excludedin ( characters.

		($table,$field) = $fld =~ /([^(]*?)\.(.*)/;	# no functions

		if (not defined $table) {

			# Not a fully qualified field name.
			# Check if field is defined in the primary table first
			# because @dbtables is not neccessarily in the order
			# defined in the SQL From clause.

			$field = $fld;
			$table = $sqlo->primary_table;

			$t = $schema->t($table);
			next unless $t;		# temporary table checked above
			$f = $t->f($field);
			if (not (defined $f or $field eq $t->id) ) {

				# Search schema for first table with field $f.
				for (@dbtables) {
					$t = $schema->t($_);
					$f = $t->f($field);
					if (defined $f or $field eq $t->id) {
						$table = $_;
						last;
					}
				}
			}
		}

		if (not defined $table) {
			die_alert(
			  "A query field '$fld' is not defined.\n"
			  . "Is it spelled correctly in the schema?\nError",
			  "Field not defined: '$fld'.",
			  $sqlo->as_sql,$pcode);
		}

		#
		# Build hash of table names and fields selected for comparison
		# with set of allowed tables and fields.
		#

		if (! ($rt = $sqlo->realtable($table))) {
			
			die_alert(
			  "The table named '$table' was not specified in the SQL query.\n"
			  . "An error was found parsing the SQL column '$fld'",
			  "The table named '$table' was not specified in the SQL query.",
			  $sqlo->as_sql,
			  $pcode);
		}

		$tblh{$rt} ||= {};
		$tblh{$rt}->{$field} = 1;
	}

	#
	# Compare fields in the sql to those that user can access.
	#

	my @badfields = ();
	for $tbl (keys %tblh) {

		# Note: at this point, all the tables in %tblh are valid
		# database tables or temporary tables.

		$t = $schema->t($tbl) or next;

		if ($t->secure) {
			# Build hash of fields user can access.
			my %sec = map {($_,1)} $t->fields_secure;

			# Compare hashes of fields
			for $fld (keys %{$tblh{$tbl}}) {
				push @badfields, "$tbl.$fld"
				  unless exists $sec{$fld};
			}
		}
	}

	if (@badfields) {
		my $badfields = join "\n\t", @badfields;
		my $die_msg = <<End_Of_Msg;
Your request has been denied because you are trying to access or alter restricted information.  Either that or this program has encountered a bug and has become utterly confused.

Please contact $Webadministrator.
End_Of_Msg
		my $alert_msg = <<End_Of_Alert;
Attempt to access database fields:
	$badfields
End_Of_Alert
		die_alert($die_msg,$alert_msg,$sqlo->as_sql,$pcode);

	}

	#
	# If here, script has passed all security checks.
	# Expand * and table.* columns into list of fields
	#

	if ($star_field_found) {

		# Helper function for expanding table.* and * columns.
		my $expand = sub {
			my @expandedfields;
			for my $table (shift || @tables) {
				# Expand '*' into list of fields
				my $t = $schema->t($sqlo->realtable($table));
				if (! $t) {
					# Temporary table 
					# Pass on * field to db engine.
					push @expandedfields, "$table.*";
					next;
				}
				my $idfield = $t->id;
				unshift @expandedfields, map {"$table.$_"}
				  $t->audit_entry, $t->audit_update, $t->audit_user
				  if $t->audit;
				unshift @expandedfields, map {"$table.$_"} $t->fields_secure;
				unshift @expandedfields, map {"$table.$_"} $idfield if $idfield;
			}
			return join ",\n", @expandedfields;
		};

		# Strip out any comments between /* and */ as a quick hack
		# to prevent * star field expansion from expanding /* stars.

		$sqlo->{select} =~ s|/\*.*?\*/||sg;

		# Change {select} component of $sqlo object with * fields
		# into fully qualified names.

		$sqlo->{select} =~ s/((\w+)\.)?\*/&$expand($2)/ge;
	}

	#
	# Add restrict where clauses if restrict codes defined.
	#

	$sqlo->{where} = '('
	  . ($sqlo->{where} || '1 = 1')
	  . ") $restrict_where"
	  if $restrict_where;

	#
	# Return re-assembled sql.
	#

	debug('clean_sql: sqlo:');
	debug_dump($sqlo);

	return $sqlo->as_sql;
}

sub die_alert {
	my ($die_msg, $alert_msg, $sql,$pcode) = @_;

	#
	# Die and send an email message to a systems administrator.
	#

	Acme::Security::Alert::security_alert(<<End_Of_Alert);
$die_msg

--------

$alert_msg
SQL rejected.
Privilege code = '$pcode'.

$sql
End_Of_Alert

	die $die_msg;
}

1;
