package Acme::Security::User;

=head1 NAME

Acme::Security::User - check user's security privileges

=head1 SYNOPSIS

    use Acme::Security::User;

    die unless &Acme::Security::User::allow_access();
    die unless &Acme::Security::User::allow_insert();
    die unless &Acme::Security::User::allow_update();
    die unless &Acme::Security::User::allow_delete();

      or

    die if &Acme::Security::User::deny_access();
    die if &Acme::Security::User::deny_insert();
    die if &Acme::Security::User::deny_update();
    die if &Acme::Security::User::deny_delete();

      or

    die unless &Acme::Security::User::allow('IUD','customers','orders','invoices');
    die if &Acme::Security::User::deny('IUD','customers','orders','invoices');

=head1 DESCRIPTION

These routines check to see if the current user has a privilege or 
restrictions applied to the given security code. If the security code
is not given, the current program name is used. Program names are absolute
path names or are path names relative to .../cgi-bin/.

Privileges are determined by a user's ID or group ID that the user belongs to.
The user ID is determined by the Local::Acme::User::user_security_id() module
and the group IDs are determined by the Local::Acme::User::user_groups() module.
The user_security_id() is usually the same as Local::Acme::User::user() unless
there are multiple guests using one user account.

Privileges are stored in two hashes.  A list of allowed privileges are
stored in the %allow hash.  A list of restrictions are stored in the
%deny hash.

The hash keys are the user's ID, group ID, or the string ALL. The hash
values point to a hash of security_codes / privileges. The security
codes are arbitrary. They can be used to identify a table or a group of
tables restricted in the schema, but they usually are program names. The
routines use the program name as a default security code if the security
code is not specified.

The routines searches the 2 hashes and stops at the first match:

	Privilege is allowed if user ID privilege is found in %allow.
	Privilege is denied if user ID privilege is found in %deny.

	Privilege is allowed if user's group ID(s) privilege is found in %allow.
	Privilege is denied if user's group ID(s) privilege is found in %deny.

	Privilege is allowed if the ID 'ALL' privilege is found in %allow.
	Privilege is denied if the ID 'ALL' privilege is found in %deny.

	Otherwise, privilege is granted.

The string ALL can be used to match all users or all security codes in the 
%allow or %deny hashes.

Four constants identify security privileges:

	Access
	Insert
	Update
	Delete

They are binary constants that can be OR'ed together to make a privilege value.

=head1 NOTE

These routines should be accessed indirectly by programs through the
Acme::Security module.

=head1 CHANGE LOG

 5/1/99   Initial version.
 7/24/00  Change to allow_*() and deny_*() routines. Added ALL constant.
 9/1/00   Added allow() to check multiple privileges at once and changed
          _allow_privilege() to check list of codes.
 3/17/01  Made security codes derived from program names relative to ../cgi-bin.
10/15/01  Changed to use Acme::Security::Access to read etc/access.conf
          configuration file of security settings.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000,2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Acme::Security::Access;
use Local::Acme::User qw(user user_security_id user_groups);

# Constants
use constant Access => 0x0001;
use constant Insert => 0x0002;
use constant Update => 0x0004;
use constant Delete => 0x0008;

#------------------------------------------------------------------------------

sub import {}

sub allow {
	my $p = shift;

	# Change privilege constant like 'AIUD' into privilege constant.
	my $priv = 0;
	$priv |= Access if index($p,'A') >= 0;
	$priv |= Insert if index($p,'I') >= 0;
	$priv |= Update if index($p,'U') >= 0;
	$priv |= Delete if index($p,'D') >= 0;

	return unless $priv;	# unknown privilege
   	return _allow_privilege($priv,@_);
}

sub deny { return ! allow(@_) }

sub allow_access { return _allow_privilege(Access, @_) }
sub allow_insert { return _allow_privilege(Insert, @_) }
sub allow_update { return _allow_privilege(Update, @_) }
sub allow_delete { return _allow_privilege(Delete, @_) }

sub deny_access { return ! _allow_privilege(Access, @_) }
sub deny_insert { return ! _allow_privilege(Insert, @_) }
sub deny_update { return ! _allow_privilege(Update, @_) }
sub deny_delete { return ! _allow_privilege(Delete, @_) }

#------------------------------------------------------------------------------
#
# Private methods
#
#------------------------------------------------------------------------------

sub _allow_privilege {
	my ($priv,@security_codes) = @_;

	# Security codes can be a comma-separated string of security codes.

	# Security code is program base name if not specified.
	($security_codes[0] = $0) =~ s|.*cgi-bin/|| unless @security_codes;

	my $user    = user();
	my $user_id = user_security_id();

	my $user_access = new Acme::Security::Access;

	my ($allow,$deny) = $user_access->privileges;

	my @ids = ($user_id, user_groups($user), $user_access->access_groups($user), 'ALL');

CODE:	for my $sc (map { split /[,; ]+/, $_ } @security_codes) {
		for my $id (@ids) {
			if (defined $allow->{$id}) {
				if (defined $allow->{$id}->{$sc}) {
					return if ($allow->{$id}->{$sc} & $priv) != $priv;
					next CODE;
				}
				if (defined $allow->{$id}->{ALL}) {
					return if ($allow->{$id}->{ALL} & $priv) != $priv;
					next CODE;
				}
			}

			if (defined $deny->{$id}) {
				if (defined $deny->{$id}->{$sc}) {
					return if $deny->{$id}->{$sc} & $priv;
				}
				if (defined $deny->{$id}->{ALL}) {
					return if $deny->{$id}->{ALL} & $priv;
				}
			}
		}
	}

	return 1;
}

1;
