package Acme::TextArray;

=head1 NAME

Acme::TextArray - TextArray routines for Acme

=head1 SYNOPSIS

    use Acme::TextArray qw(mingle);

    my @array = split_array($textarray) if is_text_array($textarray);

    print mingle([':', ','],$textarray1, $textarray2);

=head1 DESCRIPTION

The I<Acme::TextArray> module is for utilites that process text arrays.

A "text array" is a character string that contains multiple values. A separator
character is used to delimit the values and is also used to quote the string.
For example: ;one;two;three;

The is_text_array() returns true if the argument is a text array.

The split_array() function splits a character string into an array of
parts.

The mingle() function returns a string that is a concatonation of each
element of the text array arguments separated by the specified delimiters.

For example:

	mingle([': ', ', '],';one;two;three;',';cat;dogs;birds;')

produces the string

	one cat: two dogs, three birds

=head1 CHANGES

 10/7/01  Initial version.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1997-2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.000;
require Exporter;

@ISA = qw(Exporter);
@EXPORT_OK = qw(is_text_array split_array mingle);

use Acme::Config;
my $tmpdir = Acme::Config->Temp_Dir || '/var/tmp';

#------------------------------------------------------------------------------

sub is_text_array {
	$s = shift or return;
	return ($s =~ /^(\W)(.*)\1$/);
}

#------------------------------------------------------------------------------

sub split_array {

	# Split string only if it begins and ends with the same delimiter
	# character. Trailing null fields are also split and returned. The
	# original string is returned as a one item array if no delimiter
	# characters are found.

	$s = shift or return;
	if ($s =~ /^(\W)(.*)\1$/) {
		return split /[$1]/, $2, -1;
	}
	return ($s);
}

#-------------------------------------------------------------------------------

sub mingle {
	my $delim;
	if (ref $_[0] eq 'ARRAY') {
		$delim = shift;
	}
	$delim ||= [', '];	# default

	my $rv = '';		# return value

	my @arr;		# array of arrays
	for (my $i = 0; $i < @_; $i++) {
		$arr[$i] = [];
		push @{$arr[$i]}, split_array($_[$i]);
	}

	my $v;
	my $d = '';
	for (my $j = 0; $j < @{$arr[0]}; $j++) {
		for (my $i = 0; $i < @arr; $i++) {
			$v = $j < @{$arr[$i]}
			   ? $arr[$i][$j]
			   : '';
			$rv .= $d . $v;

			$d = $i < @{$delim}
			   ? $delim->[$i]
			   : $delim->[@{$delim}];
		}
	}

	return $rv;
}

1;
