package Local::Acme::Banners;

=head1 NAME

Local::Acme::Banners - HTML page header and footer banners

=head1 SYNOPSIS

    use Local::Acme::Banners qw(common_header common_footer);

    print Local::Acme::Banners::common_header($title, $print)

    print Local::Acme::Banners::common_footer($title, $print)

=head1 DESCRIPTION

The common_header() returns the HTML code which should appear at the head of
the page. The common_footer() returns the code which should appear at the
end of the page.

If the C<$print> flag is set, the header and footer are to be rendered for a
printed page.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000-2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use strict;

require 5.6.0;
require Exporter;

use Acme::DailyPics	qw(pic_of_day);
use Acme::Options;
use Acme::Config;

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(common_header common_footer common_scripts);

our $Scripts		= Acme::Config->Scripts;
our $Images		= Acme::Config->Images;
our $WWW		= Acme::Config->WWW;
our $DBTITLE		= Acme::Config->DBTITLE;
our $Company_Banner	= Acme::Config->Company_Banner;
our $Company_Logo	= Acme::Config->Company_Logo;

our $Options_Prog	= '/cgi-bin/options.cgi';

#------------------------------------------------------------------------------

sub company_banner {
	return qq[<img
     src="$Company_Banner"
     name='internet_button'
     alt='Internet'
     border=0
    >] if $Company_Banner;

	# Make company banner
	return qq[<nobr>$DBTITLE</nobr>];
}

#------------------------------------------------------------------------------

our $common_scripts_done = 0;		# flag

sub common_scripts {
	my ($print) = @_;
	$common_scripts_done = 1;	# set flag

	# Define console program if alternate is used.
	my $console_prog = Acme::Config->Console_Prog;
	my $condef = $console_prog =~ m/console.cgi/
		   ? ''	# standard console
		   : qq(\n<script language="JavaScript">consoleProg="$console_prog"</script>);

	return qq(<script src="$Scripts/common.js" type="text/javascript"></script>$condef);
}

#------------------------------------------------------------------------------

sub common_menus {
	my ($print) = @_;
	return '' if $print;
	return <<EOS;
  <tr>
  <td colspan=5 class="header-menu"><a href="javascript:console()" class="headbutton" title="Open the console window.">Console</a>
  <a href="$WWW" onClick="new_target(this)" class="headbutton" title="Open the company web site.">Internet</a>
  <a href="$Options_Prog" onClick="new_target(this,'options')" class="headbutton" title="Open user options window.">Options</a>
  <a href="about:blank" class="headbutton" target="_blank" title="Open a new blank window.">Window</a></td>
  </tr>
EOS
}

#------------------------------------------------------------------------------

sub common_header {
	my ($title,$print) = @_;

	$title ||= '';
	$title =~ s/^(["'])(.*)\1$/$2/s;		# strip possible quotes

	my $pic;
	my $opt = new Acme::Options;
	if (! $print
	&& defined $opt->DailyPicture
	&& $opt->DisplayDailyPicture =~ /yes/i) {
		$pic   = $opt->DailyPicture;
	}
	$pic ||= "$Images/transparent.gif";		# default default

	my $scripts = ! $common_scripts_done ? common_scripts($print) : '';

	my $menus   = common_menus($print);

	my $banner  = company_banner();

	if (defined($title) && $title ne "") {

		#
		# Banner with title.
		#

		return <<EOS;

<!-- Start common header -->
$scripts
<table class="header" cellpadding="0" cellspacing="0" align="center" width="100%">
  <tr>
<!-- Console Button -->
  <td class="header-left" width=80><a href="javascript:console()" title="Console"><img
     src='$Company_Logo'
     name='acme_button'
     alt='Console'
     width=80
     border=0
    ></a></td>
  
<!-- Blank Button -->
  <td class="header-spacer"><a href="about:blank" target="_blank" title="New Window"><img
     src='$Images/transparent.gif'
     alt='New Window'
     width=50
     height=50
     border=0
    ></a></td>

<!-- Internet Button -->
  <td class="header-center"><a href="$WWW" onClick="new_target(this)" class="banner" title="Internet">$banner</a><span class="title">$title</span></td>
  
<!-- Blank Button -->
  <td class="header-spacer"><a href="about:blank" target="_blank" title="New Window"><img
     src='$Images/transparent.gif'
     alt='New Window'
     width=50
     height=50
     border=0
    ></a></td>

<!-- Options Button -->
  <td class="header-right" width=80><a href="$Options_Prog" onClick="new_target(this,'options')" title="Options"><img
     src='$pic'
     name='options'
     alt='Options'
     width=80
     border=0
    ></a></td>
  </tr>

$menus
</table>
<!-- End common header -->

EOS
	}
	else {
		#
		# Banner without title.
		#

		return <<EOS;

<!-- Start common header -->
$scripts
<table class="header" cellpadding="0" cellspacing="0" align="center" width="100%">
  <tr valign=middle>
<!-- Console Button -->
  <td class="header-left" width=62><a href="javascript:console()" title="Console"><img
     src='$Company_Logo'
     name='acme_button'
     alt='Console'
     width=62
     border=0
    ></a></td>

<!-- Blank Button -->
  <td class="header-spacer"><a href="about:blank" target="_blank" title="New Window"><img
     src='$Images/transparent.gif'
     alt='New Window'
     width=50
     height=50
     border=0
    ></a></td>

<!-- Internet Button -->
  <td class="header-center"><a href="$WWW" onClick="new_target(this)" class="banner" title="Internet">$banner</a></td>

<!-- Blank Button -->
  <td class="header-spacer"><a href="about:blank" target="_blank" title="New Window"><img
     src='$Images/transparent.gif'
     alt='New Window'
     width=50
     height=50
     border=0
    ></a></td>

<!-- Options Button -->
  <td class="header-right" width=80><a href="$Options_Prog" onClick="new_target(this,'options')" title="Options"><img
     src='$pic'
     name='options'
     alt='Options'
     width=62
     border=0
    ></a></td>
  </tr>

$menus
</table>
<!-- End common header -->

EOS
	}
}

#------------------------------------------------------------------------------

sub common_footer {
	my ($title,$print) = @_;
	return <<EOS;

<!-- Start common footer -->
<!-- End common footer -->
EOS
}

1;
