package Local::Acme::Callbacks;

=head1 NAME

Callbacks - Local routines for listing callback messages.

=head1 SYNOPSIS

    use Local::Acme::Callbacks;
    
    my $bgndate = '1/1/2001';
    my $bgndate = '1/1/2001';
    my $user    = 'mack';

    my $sql = Local::Acme::Callbacks->sql($bgndate, $enddate, $user);

    my $l = new Acme::Lister sql => $sql,
    			     every_row => \&callbacks::display_callback,
			     sandbox   => 'callbacks';
    $l->print;

=head1 DESCRIPTION

The C<sql> method returns the SQL to use to select log records from the database
that need callback messages sent to users. It allows local installations to
include callback messages from other tables besides the customer_log table.

The C<$user> parameter is a pattern match parameter. It can be a user ID or
it can be a pattern like `.*'.

Do NOT include an ORDER BY clause at the end of the SQL as the programs will
add that before executing.

The callbacks::display_callback() routine, placed here in this file, displays
each row selected by the sql() method. These two routines together
select and print the rows for callbacks -- both the reports and email messages.
They are the routines which tend to be customized for each site.

This is a semi-automatic way to provide custom callback lists to sites, until
the production of callbacks and log messages are further standardized.

=head1 CHANGE LOG

1/25/02 Initial version.
7/5/02  Changed links to work when Javascript is not available (e.g. email).
	Added user_sql() to select all user IDs that have callbacks.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use Acme::Config;

our $Base = Acme::Config->Base;			# base url

#------------------------------------------------------------------------------

sub user_sql {
	# This routine can be used both as a function or a method.
	shift if ref($_[0]) || $_[0] eq __PACKAGE__;

	my ($bgndate, $enddate) = @_;

	#
	# This routine returns all the user IDs that have callbacks messages.
	# The where clause is almost identical as sql()'s below.
	#

	return <<End_Of_SQL;
	select distinct
		customer_log.user_id
	from customer_log, customers
	where customer_log.callback_date between '$bgndate' and '$enddate'
	  and customer_log.respond_date is null
	  and customer_log.customer_id = customers.customer_id
	  and customer_log.user_id is not null

	union

	select distinct
		supplier_log.user_id
	from supplier_log, suppliers
	where supplier_log.callback_date between '$bgndate' and '$enddate'
	  and supplier_log.respond_date is null
	  and supplier_log.supplier_id = suppliers.supplier_id
	  and supplier_log.user_id is not null
End_Of_SQL
}

#------------------------------------------------------------------------------

sub sql {
	# This routine can be used both as a function or a method.
	shift if ref($_[0]) || $_[0] eq __PACKAGE__;

	my ($bgndate, $enddate, $user) = @_;

	return <<End_Of_SQL;
	select
		'customer_log'                as table,
		customers.customer_name       as sort_order,
		customer_log.log_id           as log_id,
		customer_log.log_type         as log_type,
		customer_log.customer_id      as master_id,
		customer_log.log              as log,
		customer_log.callback_date    as callback_date,
		customer_log.user_id          as user_id,
		customer_log.message          as message,
		customers.name                as name,
		customers.company             as company,
		customers.city                as city,
		customers.state               as state,
		customers.country             as country,
		customers.phone               as phone,
		customers.company_phone       as phone2,
		customers.fax                 as fax,
		customers.email               as email,
		customers.customer_type       as master_type
	from customer_log, customers
	where customer_log.callback_date between '$bgndate' and '$enddate'
	  and customer_log.respond_date is null
	  and customer_log.user_id ~ '$user'
	  and customer_log.customer_id = customers.customer_id

	union all

	select
		'supplier_log'                as table,
		supplier_name                 as sort_order,
		supplier_log.log_id           as log_id,
		supplier_log.log_type         as log_type,
		supplier_log.supplier_id      as master_id,
		supplier_log.log              as log,
		supplier_log.callback_date    as callback_date,
		supplier_log.user_id          as user_id,
		supplier_log.message          as message,
		suppliers.name                as name,
		suppliers.company             as company,
		suppliers.city                as city,
		suppliers.state               as state,
		suppliers.country             as country,
		suppliers.phone               as phone,
		suppliers.phone2              as phone2,
		suppliers.fax                 as fax,
		suppliers.email               as email,
		suppliers.supplier_type       as master_type
	from supplier_log, suppliers
	where supplier_log.callback_date between '$bgndate' and '$enddate'
	  and supplier_log.user_id ~ '$user'
	  and supplier_log.respond_date is null
	  and supplier_log.supplier_id = suppliers.supplier_id
End_Of_SQL
}

#------------------------------------------------------------------------------

package callbacks;

our $wno = 0;				# window number counter
our $wid;				# window id from wno

sub display_callback {

	no strict;
	no warnings qw(uninitialized once);

	print "<li>";

	my ($acknowledge_tag, $respond_tag, $view_tag);

	# Acknowledge tag is the same for all tables.
	$wid = 'w' . $wno++;
	$acknowledge_tag = qq[<a href="$Base/cgi-bin/callbacks/acknowledge.cgi?table=$table&log_id=$log_id" onClick="open_message('Acknowledging callback.','','','$wid')" target="$wid" title="Acknowledge callback without entering a response.">];

	if ($table eq "customer_log") {

		$wid = 'w' . $wno++;
		$respond_tag = qq[<a href="$Base/cgi-bin/customers/log/respond.pl?log_id=$log_id" onClick="open_log_response('','$wid')" target="$wid" title="Respond to callback message.">];

		$wid = 'w' . $wno++;
		$view_tag = qq[<a href="$Base/cgi-bin/customers/view.cgi?customer_id=$master_id" onClick="open_view('','$wid')" target="$wid" title="View customer information">];
	}
	elsif ($table eq "supplier_log") {

		$wid = 'w' . $wno++;
		$respond_tag = qq[<a href="$Base/cgi-bin/supplier_log_respond.pl?log_id=$log_id" onClick="open_log_response('','$wid')" target="$wid" title="Respond to callback message.">];

		$wid = 'w' . $wno++;
		$view_tag = qq[<a href="$Base/cgi-bin/suppliers/view.pl?supplier_id=$master_id" onClick="open_view('','$wid')" target="$wid" title="View supplier information">];
	}

	print
	  qq[$callback_date: $user_id: ],
	  qq[<span style="font-size:7pt">(],
	  qq[${respond_tag}respond</a>],
	  qq[ | ],
	  qq[${acknowledge_tag}acknowledge</a>],
	  qq[)</span>],
	  qq[<br>\n];

	print
	  $view_tag,
	  "<big>",
	  $name,
	  ($name && $company ? ', ' : ''),
	  $company,
	  "</big>",
	  "</a>",
	  " (", $master_id, ", <b>", $master_type, "</b>)",
	  " ", $city,
	  " ", $state ? $state : " ",
	  ($country && $country ne 'United States') ? " $country" : " ";

	print " <nobr><i>p:</i> ", $phone,  "</nobr>" if $phone;
	print " <nobr><i>p:</i> ", $phone2, "</nobr>" if $phone2;
	print " <nobr><i>f:</i> ", $fax,    "</nobr>" if $fax;

	print qq{ <a href="mailto:$email" title="Send email">$email</a>}
	  if ($email);
	print "<br>\n";

	print ($message ? $message : $log) . "\n";
}

1;
