package Local::Acme::Config;

=head1 NAME

Local::Acme::Config - Custom configuration module

=head1 DESCRIPTION

This module provides custom configuration settings, usually those settings
determined programmically, or settings not specified in the site
configuration file.

Settings in this module are refered to if they are not defined in the
Acme::Config module or in the site configuration file that it
processes (e.g. ../etc/site.conf).

=cut

use strict;
require 5.6.0;

#
# Note: Arrays and hashes are best constructed as anonymous structures
# with pointers, e.g.:
#     $myhash = { 'key' => 'val' };  not  %myhash = ( 'key' => 'val' );

#
# Define databases
#

# This following is a scheme that allows the use of multiple databases.
# 
# Databases's are organized by database name. The dbschema field identifies the
# database schema (description) file or directory name containing schema files,
# one for each table in the database. The schema files contain the
# meta-database information used to define database properties and
# relationships.
# The dbengine field defines the database engine used. It helps the programs
# assemble non-standard SQL statements specific to the different databases.
# Other fields are used as default values if not specified in programs.

our $Databases = {
	'acme' => {	# database name
		'title'    => 'Acme Base Company',
		'schema'   => '/home/acme/www/schema/acme',
		'dbengine' => 'Postgres',	    # PostgreSQL database
		'host'     => '',		    # not used
		'port'     => '',		    # not used
		'connect'  => 'dbname=acme',	    # PostgreSQL connect string
		'user'     => 'acme',		    # all use same name
		'password' => 'acmepassword',	    # all use same password
		}

#	'acme' => {	# database name
#		'title'    => 'Acme Company',
#		'schema'   => '/home/acme/www/schema/acme',
#		'dbengine' => 'Postgres',	    # PostgreSQL database
#		'host'     => '',		    # url of host
#		'port'     => '',		    # access port on url of db
#		'connect'  => 'dbi:Pg:dbname=acme', # DBI Data Source Name
#		'user'     => 'acmeuser',	    # all use same name
#		'password' => 'acmepass',	    # all use same password
#		}
	};

#
# Set up default values for connecting to the database.
#

our $DBNAME    = $ENV{DBNAME}    || 'acme'; # default database name
our $DBCONNECT = $ENV{DBCONNECT} || '';	    # default DBI connect string

sub DBTITLE   { $Databases->{$_[1] || $DBNAME}->{title} }
sub DBSCHEMA  { $Databases->{$_[1] || $DBNAME}->{schema} }
# Alternate method of locating DBSCHEMA by using database name.
#sub DBSCHEMA  { Acme::Config->Schema_Space . '/' . $DBNAME }
sub DBENGINE  { $Databases->{$_[1] || $DBNAME}->{dbengine} }
sub DBHOST    { $Databases->{$_[1] || $DBNAME}->{host} }
sub DBPORT    { $Databases->{$_[1] || $DBNAME}->{port} }
sub DBCONNECT { $DBCONNECT || $Databases->{$_[1] || $DBNAME}->{connect} }
sub DBUSER    { $Databases->{$_[1] || $DBNAME}->{user}
		or require Local::Acme::User && Local::Acme::User::user() }
sub DBPASS    { $Databases->{$_[1] || $DBNAME}->{password}
		or require Local::Acme::User && Local::Acme::User::user_passwd() }

#
# Provide a personal space for user's list files.
#

sub List_Space {
	use Local::Acme::User 'user';
	my $dir = Acme::Config->Company_Space . "/lists";
	my $usrdir = "$dir/".user();
	mkdir $dir,    0777 or warn "Cannot create $dir"    unless -d $dir;
	mkdir $usrdir, 0777 or warn "Cannot create $usrdir" unless -d $usrdir;
	return  $usrdir;
}

#
# Audit information constants
#

# Following fields appear in data tables.
our $audit_entry	= 'entry_date';		# name of entry date audit field
our $audit_update	= 'last_update';	# name of timestamp audit field
our $audit_user		= 'last_update_by';	# name of user name audit field

# Audit table name made out prefix and the data table's name.
our $audit_table_prefix = 'y_';			# prefix of table's audit table

# Following fields appear only in the audit tables.
our $audit_type		= 'audit_type';		# audit record type field name
our $audit_timestamp    = 'audit_time';		# time of audit record entry
our $audit_userstamp    = 'audit_user';		# user entering the audit record

# Database column types for above dates, times, and user ids.
our $audit_entry_type  = 'date';		# database type
our $audit_update_type = 'timestamp';		# database type
our $audit_user_type   = 'varchar(14)';		# database type

#------------------------------------------------------------------------------

sub AUTOLOAD {
        my $self = shift;
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;

	no strict 'refs';

	# Return global variable if set
	return $$method if defined $$method;
	if (defined %$method) {
		return %$method if wantarray;
		return \%$method;
	}
	if (defined @$method) {
		return @$method if wantarray;
		return \@$method;
	}
        return undef;           # unknown parameter/method
}

1;
