package Local::Acme::File;

=head1 NAME

File - Local file utility routines for Acme

=head1 SYNOPSIS

    use Local::Acme::File  qw(copy_to_www_server 	copy_to_file_server 
    			copy_from_www_server	copy_from_file_server);
    
    copy_to_www_server($from, $to, $options);
    copy_from_www_server($from, $to, $options);

    copy_to_file_server($from, $to, $options);
    copy_from_file_server($from, $to, $options);

=head1 DESCRIPTION

Copy_to_www_server() copies a file to a destination on the Web server.
The destination root is the root location of files for the Web server.
The first 2 parameters are the local file name (from) and the target file
name (to). The remaining parameters are options as described below.

Copy_from_www_server() copies a file from a location on the Web server.
The web server root is the root location of files for the Web server.
The first 2 parameters are the remote file name (from) and the target file
name (to). The remaining parameters are options as described below.

Copy_to_file_server() copies a file to a destination on the file server.
The destination root is the root location of Acme files for the file server.
The first 2 parameters are the local file name (from) and the target file
name (to). The remaining parameters are options as described below.

Copy_from_file_server() copies a file from a location on the file server.
The location root is the root location of Acme files for the file server.
The first 2 parameters are the remote file name (from) and the target file
name (to). The remaining parameters are options as described below.

Options can match the following keywords:
       binary	-- sets copy mode to binary
       ascii	-- sets copy mode to ascii
       makedir	-- makes directories on the target if needed
   
Options can also be of the form keyword=value where the
keywords are:
       server	-- file server
       user	-- user id to us to log on
       pass	-- password

Copy_from_www_server() and copy_from_file_server() ignore the makedir
option. It is assumed that file locations exist on the local server
before calling the routine.

=head1 CHANGE LOG

 11/19/98 Initial version.
  2/25/00 Renamed package to Local::Acme::File
	  Imports constants from Acme::Config.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 1998-2000 Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.000;
require Exporter;

@ISA = qw(Exporter);
@EXPORT_OK = qw(	copy_to_www_server 	copy_to_file_server 
			copy_from_www_server	copy_from_file_server);

use Acme::Config;

#------------------------------------------------------------------------------

sub copy_to_www_server {
	my $from = shift;
	my $to   = shift;

	return 0 unless ($from and $to);

	# set options
	my $server = Acme::Config->WWW;
	my $user   = Acme::Config->WWWUSER;
	my $pass   = Acme::Config->WWWPASS;
	my $copymode = "b";	# default binary
	my $makedir  = 0;	# default no
	for (@_) {
		/binary/i and $copymode = "b";
		/ascii/i  and $copymode = "a";
		/makedir(ectory)*/i and $makedir = 1;
		/^(.*?)=["']*(.*)["']*$/ and eval (${$1} = $2);
	}

	($d,$f) = ($to =~ m|(.*/)(.*)|);	# parse into directory / file
	$d =~ s|^/||;		# no absolute paths allowed
	$d =~ s|/$||;		# remove trailing /

	delete @ENV{ qw(IFS CDPATH ENV BASH_ENV) };     # untaint
	$ENV{PATH} = '/bin:/usr/bin';                   # untaint

	open FTP,"| ftp -n '$server' >/dev/null 2>&1";
	print FTP "user $user $pass\n";
	print FTP "binary\n" if ($copymode eq "b");
	print FTP "ascii\n"  if ($copymode eq "a");
	for (split "/", $d) {
		print FTP "mkdir \"$_\"\n" if $makedir;
		print FTP "cd \"$_\"\n";
	}
	print FTP "put \"$from\" \"$f\"\n";
	print FTP "quit\n";
	close FTP;

	return 1;
}

#------------------------------------------------------------------------------

sub copy_from_www_server {
	my $from = shift;
	my $to   = shift;

	return 0 unless ($from and $to);

	# set options
	my $server = Acme::Config->WWW;
	my $user   = Acme::Config->WWWUSER;
	my $pass   = Acme::Config->WWWPASS;
	my $copymode = "b";	# default binary
	my $makedir  = 0;	# default no
	for (@_) {
		/binary/i and $copymode = "b";
		/ascii/i and $copymode  = "a";
		/makedir(ectory)*/i and $makedir = 1;
		/^(.*?)=["']*(.*)["']*$/ and eval (${$1} = $2);
	}

	delete @ENV{ qw(IFS CDPATH ENV BASH_ENV) };     # untaint
	$ENV{PATH} = '/bin:/usr/bin';                   # untaint

	open FTP,"| ftp -n '$server' >/dev/null 2>&1";
	print FTP "user $user $pass\n";
	print FTP "binary" if ($copymode eq "b");
	print FTP "ascii"  if ($copymode eq "a");
	print FTP "get \"$from\" \"$to\"\n";
	print FTP "quit\n";
	close FTP;

	return 1;
}

#------------------------------------------------------------------------------

sub copy_to_file_server {
	my $from = shift;
	my $to   = shift;

	return 0 unless ($from and $to);

	# set options
	my $server = Acme::Config->FILESERVER;
	my $user   = Acme::Config->FILESERVERUSER;
	my $pass   = Acme::Config->FILESERVERPASS;
	my $copymode = "b";	# default binary
	my $makedir  = 0;	# default no
	for (@_) {
		/binary/i and $copymode = "b";
		/ascii/i  and $copymode = "a";
		/makedir(ectory)*/i and $makedir = 1;
		/^(.*?)=["']*(.*)["']*$/ and eval (${$1} = $2);
	}

	($d,$f) = ($to =~ m|(.*/)(.*)|);	# parse into directory / file
	$d =~ s|^/||;		# no absolute paths allowed
	$d =~ s|/$||;		# remove trailing /

	delete @ENV{ qw(IFS CDPATH ENV BASH_ENV) };     # untaint
	$ENV{PATH} = '/bin:/usr/bin';                   # untaint

	open FTP,"| ftp -n '$server' >/dev/null 2>&1";
	print FTP "user $user $pass\n";
	print FTP "binary" if ($copymode eq "b");
	print FTP "ascii"  if ($copymode eq "a");
	for (split "/", $d) {
		print FTP "mkdir \"$_\"\n" if $makedir;
		print FTP "cd \"$_\"\n";
	}
	print FTP "put \"$from\" \"$f\"\n";
	print FTP "quit\n";
	close FTP;

	return 1;
}

#------------------------------------------------------------------------------

sub copy_from_file_server {
	my $from = shift;
	my $to   = shift;

	return 0 unless ($from and $to);

	# set options
	my $server = Acme::Config->FILESERVER;
	my $user   = Acme::Config->FILESERVERUSER;
	my $pass   = Acme::Config->FILESERVERPASS;
	my $copymode = "b";	# default binary
	my $makedir  = 0;	# default no
	for (@_) {
		/binary/i and $copymode = "b";
		/ascii/i and $copymode  = "a";
		/makedir(ectory)*/i and $makedir = 1;
		/^(.*?)=["']*(.*)["']*$/ and eval (${$1} = $2);
	}

	($d,$f) = ($to =~ m|(.*/)(.*)|);	# parse into directory / file
	$d =~ s|^/||;		# no absolute paths allowed
	$d =~ s|/$||;		# remove trailing /

	delete @ENV{ qw(IFS CDPATH ENV BASH_ENV) };     # untaint
	$ENV{PATH} = '/bin:/usr/bin';                   # untaint

	open FTP,"| ftp -n '$server' >/dev/null 2>&1";
	print FTP "user $user $pass\n";
	print FTP "binary" if ($copymode eq "b");
	print FTP "ascii"  if ($copymode eq "a");
	print FTP "get \"$from\" \"$to\"\n";
	print FTP "quit\n";
	close FTP;

	return 1;
}

1;
