package Local::Acme::Form::Banners;

=head1 NAME

Local::Acme::Form::Banners - HTML page header and footer banners for forms

=head1 SYNOPSIS

    use Local::Acme::Form::Banners qw(common_form_header common_form_footer
    				    common_scripts);

    print common_scripts();

    print common_header();

    print common_form_header();
     -or-
    print common_form_header($title, {
	submit=>' Search ',
	reset=>' Cancel '
	});

    print common_form_footer();

    print common_param_header($title, {
	submit=>' Save  ',
	reset=>' Cancel '
	other=>'<input type=button value=" Lookup " onclick="do_lookup()">'
	});

    print common_param_footer();

    print common_query_header();

    print common_query_footer();

    print common_footer();

=head1 DESCRIPTION

The common_header() function returns the HTML code which should appear at
the head of the page. The common_footer() returns the code which should
appear at the end of the page.

The common_form_header() function returns the HTML code which should
appear at the head of the form. It is rendered inside the <form> tags so
that submit buttons can be placed in the header. A title can be specified
for the form along with an options hash that can change the labels used
on the Save and Cancel buttons.

The common_form_footer() function returns the code which should appear at
the end of the form.

The common_query_header() function returns the HTML code which should appear
at the head of a query form. Like common_form_header(), it is rendered inside
the <form> tags.  The common_query_footer() function returns the HTML code
which should appear at the end of the form.

The common_param_header() function is very similar to the common_form_header.
I can be thought of as a "light weight" header which is to appear on screens
asking for parameters before executing the program. Parameter screens are
conceptually similar to dialog boxes.

The common_scripts() function returns scripts that are neccessary for all
forms with these headers and footers.

=head1 CHANGE LOG

 4/19/00 Initial version.
 6/16/00 Added common_query_header() and common_query_footer().
 1/11/02 Added options to common_form_header() to allow different button labels.
 8/8/02  Added common_param_header/footer to separate common_form_header/footer
         from screens asking for search and report parameters. Added "other"
	 parameter to common_form_header for adding extra buttons to header.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000-2002, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.000;
require Exporter;

@ISA = qw(Exporter);
@EXPORT_OK = qw(
		common_header		common_footer
		common_form_header	common_form_footer
		common_param_header	common_param_footer
		common_query_header	common_query_footer
		);

use Acme::Options;
use Acme::Config;

$Scripts      = Acme::Config->Scripts;
$Images       = Acme::Config->Images;
$Company_Icon = Acme::Config->Company_Icon;

#------------------------------------------------------------------------------

$common_scripts_done = 0;		# flag

sub common_scripts {
		$common_scripts_done = 1;
		return qq(<script src="$Scripts/common.js" type="text/javascript"></script>);
}

#------------------------------------------------------------------------------

sub common_header {
	return "<!--\$ First Navigation Buttons -->\n";
}

sub common_footer {
	return "<!--\$ Last Navigation Buttons -->\n";
}

#------------------------------------------------------------------------------

sub common_form_header {
	my $title = shift || '';
	my $opt   = shift || {};

	# Set text of submit and reset buttons
	my $submit = exists $opt->{submit} ? $opt->{submit} : '  Save  ';
	my $reset  = exists $opt->{reset}  ? $opt->{reset}  : ' Cancel ';
	$title =~ s/^(["'])(.*)\1$/$2/s;		# strip possible quotes

	my $scripts = $common_scripts_done ? '' : common_scripts();

	my $submit_button = '';
	$submit_button = qq[<input type="submit" value="$submit" class="button">]
	  if $submit;

	my $reset_button = '';
	$reset_button = qq[<input type="button" value="$reset" onClick="if(typeof self.form_cancel=='function'){form_cancel(this.form)}else{self.close()}" class="button">]
	  if $reset;

	my $other_button = $opt->{other} || '';

	return <<EOS;

<!-- Start common form header -->
$scripts
<table width="100%" border="1">
<tr>
<td><table cellpadding="0" cellspacing="0" border="0" width="100%">
<tr>
 <td><table cellspacing="0" cellpadding="2" style="margin:6pt 0pt; background-color:white; width:100%; height:35">
   <tr valign="middle">
   <td width="15"><a href="javascript:console()" tabindex="-1"><img
     src="$Company_Icon"
     name="console"
     alt="Console"
     border="0"
    ></a></td>
   <td><span class="title">$title</span></td>
 </tr>
 </table></td>
<td align="right"><table cellspacing="0" cellpadding="2" style="margin:6pt 0pt; background-color:white; width:100%; height:35">
  <tr>
   <td align="right">$other_button $submit_button $reset_button</td>
  </tr>
 </table></td>
</tr>
</table></td></tr></table>
<!-- End common form header -->

EOS
}

#------------------------------------------------------------------------------

sub common_form_footer {
	my $title = shift || '';
	my $opt   = shift || {};
	return <<EOS;
<!-- Start common form footer -->
<!-- End common form footer -->
EOS
}


#------------------------------------------------------------------------------

sub common_param_header {
	return common_form_header(@_);		# use the same as a normal form.
}

#------------------------------------------------------------------------------

sub common_param_footer {
	return common_form_footer(@_);		# use the same as a normal form.
}

#------------------------------------------------------------------------------

sub common_query_header {
	my $title = shift || '';

	$title =~ s/^(["'])(.*)\1$/$2/s;		# strip quotes

	my $scripts = common_scripts() unless $common_scripts_done;

	return <<EOS;

<!-- Start common query header -->
$scripts
<table width="100%" border="1">
<tr>
<td><table cellpadding="0" cellspacing="0" border="0" width="100%">
<tr>
 <td><table cellspacing="0" cellpadding="2" style="margin:6pt 0pt; background-color:white; width:100%; height:35">
   <tr valign="middle">
   <td width="15"><a href="javascript:console()" tabindex="-1"><img
     src="$Company_Icon"
     name="console"
     alt="Console"
     border="0"
    ></a></td>
   <td><span class="title">$title</span></td>
 </tr>
 </table></td>
<td align="right"><table cellspacing="0" cellpadding="2" style="margin:6pt 0pt; background-color:white; width:100%; height:35">
  <tr>
   <td align="right"><input type=submit name="request" value=" Count " class="button">
   <input type=submit name="request" value="  List   " class="button">
   <input type="reset" value=" Reset " class="button">
   <input type="reset" value=" Close " onClick="self.close()" class="button"></td>
  </tr>
 </table></td>
</tr>
</table></td></tr></table>
<!-- End common query header -->

EOS
}

#------------------------------------------------------------------------------

sub common_query_footer {
	my $title = shift || '';
	return <<EOS;
<!-- Start common query footer -->
<!-- End common query footer -->
EOS
}

1;
