package Local::Acme::Form::Extensions;

=head1 NAME

Local::Acme::Form::Extensions - HTML page header and footer banners for forms

=head1 SYNOPSIS

    use Local::Acme::Form::Extensions;

    $tag = Local::Acme::Form::Extensions->tag_command($tag);

=head1 DESCRIPTION

The tag_command() is executed whenever a command string is found in a form
template that is being processed to be displayed to the user.

A command is a specially formatted XML/HTML comment like:

	<!--$ command string -->

The command string is an arbitrary text string that can be substituted by
tag_command().

A flag $printing is passed to tag_command() to denote when the form is being
prepared for printing.

=head1 CHANGE LOG

 4/1/01   Initial version.
 5/9/01   Incorporate all commands here.
 7/14/01  Added printing flag.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.000;
require Exporter;

@ISA = qw(Exporter);
@EXPORT_OK = qw(tag_command);

use Acme::HTML::NavButtons;
use Local::Acme::Banners;
use Local::Acme::Form::Banners;

#
# Static variables
#

local $NavButtons;
local $NavAnchors = 0;

#------------------------------------------------------------------------------

sub tag_command {
	my ($tag,$printing) = @_;

	my ($command) = $tag =~ /<!--+\$+\s*(.*?)\s*--+>/s;

	# The number of dollar signs $ in the command string determine the
	# number of substitutions (recursive calls) performed on the tag.

	$tag =~ s/\$//;		# remove first $ in tag.

	return $tag unless $command;

	# Note: When adding commands below, it is important to APPEND results
	# to $tag. Recursion stops if entire contents of $tag is replaced.

	COMMAND: {

		#
		# Headers and Footers
		#

		if ($command =~ s/\s*Form Header\s*//i) {
			$tag .= Local::Acme::Form::Banners::common_form_header($command);
			last COMMAND;
		}
		if ($command =~ s/\s*Form Footer\s*//i) {
			$tag .= Local::Acme::Form::Banners::common_form_footer($command);
			last COMMAND;
		}
		if ($command =~ s/\s*Form Page Header\s*//i) {
			$tag .= Local::Acme::Form::Banners::common_header($command);
			last COMMAND;
		}
		if ($command =~ s/\s*Form Page Footer\s*//i) {
			$tag .= Local::Acme::Form::Banners::common_footer($command);
			last COMMAND;
		}

		if ($command =~ s/\s*Page Header\s*//i) {
			$tag .= Local::Acme::Banners::common_header($command);
			last COMMAND;
		}
		if ($command =~ s/\s*Page Footer\s*//i) {
			$tag .= Local::Acme::Banners::common_footer($command);
			last COMMAND;
		}

		#
		# Navigation Buttons
		#

		if ($command =~ /First Nav.*Button/i) {
			last COMMAND if $printing;
			$NavButtons ||= new Acme::HTML::NavButtons;
			$tag .= $NavButtons->first_buttons();
			last COMMAND;
		}
		if ($command =~ /Last Nav.*Button/i) {
			last COMMAND if $printing;
			$NavButtons ||= new Acme::HTML::NavButtons;
			$tag .= $NavButtons->last_buttons();
			last COMMAND;
		}
		if ($command =~ /Nav.*Buttons/i) {
			if ($printing) {
				$tag .= qq[<hr class="section">];
				last COMMAND if $printing;
			}
			$NavButtons ||= new Acme::HTML::NavButtons;
			$tag .= $NavButtons->buttons();
			last COMMAND;
		}
		if ($command =~ /Nav.*Button/i) {
			last COMMAND if $printing;
			$NavButtons ||= new Acme::HTML::NavButtons;
			my $anchor = 'anchor'.++$NavAnchors;
			$tag .= qq(<a name="$anchor">)
			  . $NavButtons->nav_button("$anchor");
			last COMMAND;
		}
		if ($command =~ /Up Button/i) {
			last COMMAND if $printing;
			$NavButtons ||= new Acme::HTML::NavButtons;
			my $anchor = 'anchor'.++$NavAnchors;
			$tag .= qq(<a name="$anchor">)
			  . $NavButtons->up_button("$anchor");
			last COMMAND;
		}
		if ($command =~ /Down Button/i) {
			last COMMAND if $printing;
			$NavButtons ||= new Acme::HTML::NavButtons;
			my $anchor = 'anchor'.++$NavAnchors;
			$tag .= qq(<a name="$anchor">)
			  . $NavButtons->down_button("$anchor");
			last COMMAND;
		}

		######
		## Example
		######
		if ($command =~ /Test/i) {
			$tag .= "\n" . " Testing Testing Testing";
			last COMMAND;
		}
	}

	# Set up recursive call to substitute any command tags found in
	# the string just substituted.

	$tag =~ s/(<!--+\$+\s*.*?\s*--+>)/tag_command($1,$printing)/egs;

	return $tag;
}

1;
