package Local::Acme::Format;

=head1 NAME

Format - Process format specification strings

=head1 SYNOPSIS

    use Local::Acme::Format  qw(sprintf_cmd);
    
    my $formatted_list = sprintf_cmd('A format %s with &func("%s")',@a)
      or
    print Local::Acme::Format::sprintf_cmd("%s %s &mingle([':',','],'%s','%s) %s",
	$id, $name, ";small;medium;large;", ";3;3;3;", "9");

=head1 DESCRIPTION

The sprintf_cmd function formats the arguments using Perl's sprintf() function.
It then scans the formatted string for functions to execute. The functions
are patterns  which begin with the C<&> character and end with C<)>, like
this: C<&function("arg1","arg2")>.

In the interests of safety, only the functions that are listed in the
C<%Func> hash are executed.

=head1 CHANGE LOG

10/12/01 Initial version.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2001, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

require 5.000;
require Exporter;

@ISA = qw(Exporter);
@EXPORT_OK = qw(sprintf_cmd);

use Acme::TextArray	qw(split_array	mingle);
use Acme::Money		qw(dollars dollars_us dollars_can commafy);
use Acme::Date		qw(mdate meddate ldate longdate
			   short_date_range long_date_range);

use Acme::HTML::UPC	qw(upc_img upc_src);

#
# Enter the names of the functions to be recognized in format strings here.
#

%Func = (
	'split_array'		=> 1,
	'mingle'		=> 1,

	'dollars'		=> 1,
	'dollars_us'		=> 1,
	'dollars_can'		=> 1,
	'commafy'		=> 1,

	'mdate'			=> 1,
	'meddate'		=> 1,
	'ldate'			=> 1,
	'longdate'		=> 1,
	'short_date_range'	=> 1,
	'long_date_range'	=> 1,

	'if'			=> 1,

	'total'			=> 1,

	'upc_img'		=> 1,
	'upc_src'		=> 1,
	);

#------------------------------------------------------------------------------

sub sprintf_cmd {
	my $fmt = shift;

	# Format string using Perl's sprintf() function and then evaluate
	# any functions like "&func(arg,...)" embedded in the resulting string.

	no warnings qw(uninitialized);	# ignore undefs in @_

	my $s = sprintf($fmt,@_);	# format string
	my $r;

        # Define regular expression to match matching parentheses (even if nested).
        # From the Camel book, 3rd edition, p214.

        our $matched_paren = qr{
                                  \(
                                  (?:
                                    (?> [^()]+ )
                                  |
                                    (??{ $matched_paren })
                                  )*
                                  \)
                                }x;

	$s =~  s/
		(\&(\w+)$matched_paren)		# match &func(args) string
		/
		if (exists $Func{$2}) {
			$r = eval $1;		# execute function
			if (defined $r) { $r }	# return result
			else            { '' }
		}
		else {
			$1;			# return original string intact
		}
		/gsex;

	return $s;
}

#------------------------------------------------------------------------------
#
# Format functions
#
#------------------------------------------------------------------------------

sub if {
	my ($cond,$true_str,$false_str) = @_;

	if ($cond) {
		return $true_str || '';
	}
	else {
		return $false_str || '';
	}
}

sub total {
	my ($pkg) = caller 2;		# caller's caller
	return eval <<End_Of_Func;
package $pkg;
require Acme::Lister::Functions;
total(\@_);
End_Of_Func
}

1;
