package Local::Acme::ListMenus;

=head1 NAME

Local::Acme::ListMenus - return HTML code for drop down lists

=head1 SYNOPSIS

    use Local::Acme::ListMenus;

    # User name list
    print Local::Acme::ListMenus->users();

    # Custom list with all arguments.
    print Local::Acme::ListMenus->custom($name,$default,$size,$attributes,$format,$multiple);

=head1 DESCRIPTION

The methods in this package returns HTML code for a selection list of options
usually obtained from database control tables. These are the company specific
routines which augment those avalable in Acme::HTML::ListMenus.

This module automatically refers unknown methods to Acme::HTML::ListMenus->Make()
method to create a list from a database table's definitions with the same name
as the undefined method name. Its a convention in the library routines to 
refer to methods in this module whenever they want a list made from a database
table so that local installations can override default lists.

The first argument is the name of the HTML select tag.

Default values can be specified. They cause the option which matches the 
default to be displayed in the list. If the default value is not found in
the list, it is added to the end of the list so that it can be displayed. 
This was implemented so that dropdown lists could be used in forms which
edit database records whose values are out of date or not on the current
list of options to display in the dropdown list.

The size argument specified the maximum number of characters of the option
descriptions. If the descriptions exceed the maximum size, if specified,
the description is truncated and has "..." appended.

Attributes to the <select> tag can be passed to the routine as the fourth
argument. 

The format of the list is influenced by setting the $format argument. 
It is a character string similar to format strings used by Perl's sprintf()
function. It is actually processed by Local::Acme::Format::sprintf_cmd() method
so the formatted string is scanned for functions that are recognized by
I<Local::Acme::Format> and those function results are also substituted.
Examples are: 

	'%s -- %s'
	'%3.2d -- %s'

The $multiple flag set to true is to change the list from a single value
drop-down list into a multi-line list that can have multiple items selected.
Note that as a general rule, single value drop-down lists are converted
to multi-line lists in Query forms so multiple values can be queried at once.

=head1 CHANGE LOG

 02/00  Initial version.
 7/3/02 Removed defined() method. Use UNIVERSAL->can() instead.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use strict;
require 5.6.0;

use Acme::HTML qw(escape_html);
use Acme::HTML::ListMenus;

#------------------------------------------------------------------------------
#
# Autoload routines
#
#------------------------------------------------------------------------------

#
# Declare the methods that are dynamically loaded when needed.
#

sub users;
sub gender;

#
# Define autoloaded methods.
#

our %Subs = (

#----------------------------------------
'users' => <<'END_OF_FUNC',
use Local::Acme::User	qw(user_names);
sub users {
	my ($class,$name,$default,$size,$attributes,$format,$multi) = @_;

	my %names = user_names();
	
	if ($default and not $names{$default}) { $names{$default} = $default; }

	return Acme::HTML::ListMenus->List(\%names,$name,$default,$size,$attributes,$format,$multi);
}
END_OF_FUNC

#
# Option lists from constant values
#

#----------------------------------------
'gender' => <<'END_OF_FUNC',
sub gender {
	my ($class,$name,$default,$size,$attributes,$format,$multi) = @_;

	my %gender = (
		'M' => 'Male',
		'F' => 'Female'
		);
	
	# Add default value to list if not in above contants.
	if ($default and not $gender{$default}) { $gender{$default} = ''; }

	return Acme::HTML::ListMenus->List(\%gender,$name,$default,$size,$attributes,$format,$multi);
}
END_OF_FUNC
);

#------------------------------------------------------------------------------

sub AUTOLOAD {
        my ($pkg, $method) = our $AUTOLOAD =~ /(.*)::(.*)/;
	return if $method eq 'DESTROY';	# gobble

	if (exists $Subs{$method}) {
		my $s = $Subs{$method};
		delete $Subs{$method};	# release memory

		if ($s) {
			eval "package $pkg; $s"; # compile method
			return if $@;	# no error if eval fails
			goto &$method;	# execute method
		}

		return;			# unknown parameter/method
	}
	else {
		shift;			# remove our class name from args

		#
		# Make list from schema definitions of database.
		#

		return Acme::HTML::ListMenus->Make($method,@_);
	}
}

1;

__END__

################################################################################
#
# Example list using sql
#
################################################################################
#
# Insert following into $Subs hash
#

#----------------------------------------
'ad_sources' => <<'END_OF_FUNC',
sub ad_sources {
	my ($class,$name,$default,$size,$attributes,$format) = @_;

	my $or_is_default = '';
	$or_is_default = qq(or ad_source = '$default') if $default;

	# Note: The first field of the sql is used for the return value for the
	# lists made by Acme::HTML::ListMenus->List.

	my $sql = <<EOS;
select ad_source, ad_source, ad_source_type, ad_source_name
from ad_sources
where (retired is null or retired != 'Y') $or_is_default
order by ad_source
EOS

	return Acme::HTML::ListMenus->List($sql,$name,$default,$size,$attributes,$format);
}
END_OF_FUNC
