package Local::Acme::User;

=head1 NAME

Local::Acme::User - return of user names and user's information for the Acme system

=head1 SYNOPSIS

    use Local::Acme::User qw(user_names
    			user user_name user_group user_home user_folder user_email);

    %names = user_names();

    $user = user();

    $user_id = user_security_id();

    $name = user_name($user);

    $passwd = user_passwd($user);

    $home = user_home($user);

    $space = user_space($user);

    $folder = user_folder($user);

    $email = user_email($user);

    $group_id = user_group($user);

    @group_ids = user_groups($user);

=head1 DESCRIPTION

User_names() returns a list of names allowed to access the Acme system.

User() returns the current user's id. The CGI environment variable REMOTE_USER
is checked first. If it does not exist, the current user id is used. This
user name can be used in file and directory names.

User_security_id() returns the user id to use to check security privilages.
Usually, user_security_id() returns the same id as user() unless guest
users are allowed. In that case, user() returns the user id to use to
identify the user in log files, temporary file names, and such whereas
user_security_id() returns the user id to use to check privilages.

User_name() returns the full name of the given user, or the current user if
a user is not given.

User_passwd returns the user's password.

User_home() returns the user's home directory if they have a login account
on the system, or it returns user_space().

User_space() returns the path of a directory where the user's folder is
placed. It is used for users who do not have local logins (home directories).

User_folder() returns the path of a directory where the system can store user
settings and other files. It is a directory that resides in the directory
returned by user_space(). It must either already exist, or the server must
be able to create it in the directory returned by user_space(). This is
done so that if the directory returned by user_space() is the user's home
directory, the server won't need to have read/write permissions on the
user's home directory. The path to the user's folder must be searchable by
the administrative accounts that do nightly chores, otherwise the user's
option setting are ignored.

User_email() returns the email address of the given user, or the current
user if a user is not given. It is a full email address that includes the
user's name, like: My Name <myname@somewhere.com>

User_group() returns the user's primary security group they belong to.

User_groups() returns all the user's security groups they belong to. At this
time, only the primary security group is returned.

=head1 TODO

 Modify user_groups() to list all groups user belongs to.

=head1 CHANGE LOG

 11/13/98  Initial version.
 02/01/00  Added user_space().
 10/20/01  Moved excluded users list to Acme::Config (etc/site.conf)
 11/08/01  Added user_folder() to be sub-folder of user_space().
 12/30/02  Added user_security_id() to accommodate guests.

=head1 AUTHOR

Mack Pexton <mack@acmebase.org>

=head1 COPYRIGHT

Copyright 2000, Mackley F. Pexton.  All rights reserved.

This is open software and is licensed under the terms of the GNU General Public License.

=cut

use strict;

require 5.000;
require Exporter;

@Local::Acme::User::ISA = qw(Exporter);
@Local::Acme::User::EXPORT_OK = qw(
	user_names user user_security_id user_name user_passwd
	user_group user_groups user_home user_space user_folder
	user_email);

use Acme::Cache;
use Acme::Config;

#------------------------------------------------------------------------------

sub user_names {

	# Return a list of user id and user name pairs

	my $user_names = Acme::Cache->get(
		file        => (Acme::Config->User_File || '../etc/passwd'),
		constructor => \&_read_passwd_file,
		);

	return %$user_names;
}

sub _read_passwd_file {
	my ($file) = @_;

	my $user_names = {};

	%$user_names = %{Acme::Config->Include_Users};
	my %exclude_users = map { $_, 1 } @{Acme::Config->Exclude_Users};

	my ($login,$passwd,$uid,$gid,$gcos,$home,$shell);

	open PASSWD, '<', $file;
	while (<PASSWD>) {
		chomp;
		($login,$passwd,$uid,$gid,$gcos,$home,$shell) = split /:/;
		next if $exclude_users{$login} || ($file eq '/etc/passwd' && $uid < 500);
		$user_names->{$login} = $gcos || $login;	
	}
	close PASSWD;

	return $user_names;
}

#------------------------------------------------------------------------------

sub user {

	# Check if CGI user first

	if (defined($ENV{REMOTE_USER})) {
		# Untaint the user name, allow only word characters
		my ($u) = $ENV{REMOTE_USER} =~ /(\w+)/;	 # untaint user name
		return $u;
	}
	else {
		# Use effective user ID, not the real user ID.
		return (getpwuid($>))[0];
	}
}

#------------------------------------------------------------------------------

sub user_security_id {

	return user();
}


#------------------------------------------------------------------------------

sub user_name {
	my $user = shift || user();
	my ($user_id, $passwd, $uid, $gid, $quota, $comment, $name, $home) = getpwnam($user);

	return $name;
}

#------------------------------------------------------------------------------

sub user_passwd {
	my $user = shift || user();
	my ($user_id, $passwd, $uid, $gid, $quota, $comment, $name, $home) = getpwnam($user);

	return $passwd;
}

#------------------------------------------------------------------------------

sub user_group {
	my $user = shift || user();
	my ($user_id, $passwd, $uid, $gid, $quota, $name, $gcos, $home) = getpwnam($user);

	my $group_name = getgrgid($gid);

	return $group_name;
}

#------------------------------------------------------------------------------

sub user_groups {

	# Return all security groups the user belongs to.
	## TODO
	return user_group(@_);	# return user's primary group
}

#------------------------------------------------------------------------------

sub user_home {
	my $user = shift || user();
	my ($user_id, $passwd, $uid, $gid, $quota, $name, $gcos, $home) = getpwnam($user);
	return $home if $home;
	return user_space();	# default if no home
}

#------------------------------------------------------------------------------

sub user_space {
	my $user = shift || user();

	# Place for user's folder.

	my $space = Acme::Config->User_Space . '/' . $user;
	unless (-d $space) {
		# Try to make directory.

		# Do not die if mkdir fails. Sometimes this gets called by
		# an admin user, and they should not be creating directories.

		my $old_umask = umask 0711;
		mkdir $space;
		umask $old_umask;
	}
	return $space;
}

#------------------------------------------------------------------------------

sub user_folder {
	my $user = shift || user();

	# Place for user's persistant storage and saved lists and reports.

	my $folder = user_space($user) . '/' . Acme::Config->User_Folder;
	unless (-d $folder) {
		# Try to make directory.

		# Do not die if mkdir fails. Sometimes this gets called by
		# an admin user. Security is restrictive.

		my $old_umask = umask 0711;
		mkdir $folder;
		umask $old_umask;
	}
	return $folder;
}

#------------------------------------------------------------------------------

sub user_email {
	my $user = shift || user();

	my $name = user_name($user);
	my $email = $user . '@' . Acme::Config->Email_Send_Domain;
	$email = "$name <$email>" if $name;
	return $email;
}

1;
